/*
 * linux/fs/acl_ea.c
 *
 * (extended attribute representation of access control lists)
 *
 * Copyright (C) 2000 by Andreas Gruenbacher, <a.gruenbacher@computer.org>
 */

#include <linux/version.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/fs.h>
#include <linux/ext_attr.h>
#include <linux/acl_ea.h>
#include <asm/uaccess.h>

EXPORT_SYMBOL(acl_ea_get_inode);
EXPORT_SYMBOL(acl_ea_set_inode);

static int
acl_ea_get(
	struct inode *inode,
	const char *name,
	void *value,
	size_t size,
	int flags);
static int
acl_ea_set(
	struct inode *inode,
	const char *name,
	void *value,
	size_t size,
	int flags);
static posix_acl_t *
acl_from_ea(
	const void *value,
	size_t size);
static void *
acl_to_ea(
	const posix_acl_t *acl,
	size_t *size);

/*
 * handlers for the "$acl" and "$defacl" extended attributes
 */
static struct ext_attr_handler acl_ea_access_handler = {
	name:		ACL_EA_ACCESS,
	get_inode:	acl_ea_get,
	set_inode:	acl_ea_set,
};

static struct ext_attr_handler acl_ea_default_handler = {
	name:		ACL_EA_DEFAULT,
	get_inode:	acl_ea_get,
	set_inode:	acl_ea_set,
};

/*
 * get the "$acl" or "$defacl" attribute of an inode.
 */
static int
acl_ea_get(
	struct inode *inode,
	const char *name,
	void *value,
	size_t size,
	int flags)
{
	struct super_block *sb = inode->i_sb;

	if (!sb)
		return -ENOENT;
	if (!IS_POSIX_ACL(inode))
		return -ENOTSUP;
	return ext_attr_get(inode, name, value, size, flags);
}

/*
 * set the "$acl" or "$defacl" attribute of an inode, and update the
 * per-inode cache.
 */
static int
acl_ea_set(
	struct inode *inode,
	const char *name,
	void *value,
	size_t size,
	int flags)
{
	struct super_block *sb = inode->i_sb;
	posix_acl_t *acl;
	int type, error;

	if (!sb)
		return -ENOENT;
	if (!IS_POSIX_ACL(inode))
		return -ENOTSUP;
	if (IS_RDONLY(inode))
		return -EROFS;
	if (IS_IMMUTABLE(inode) || IS_APPEND(inode))
		return -EPERM;
	if ((current->fsuid != inode->i_uid) && !capable(CAP_FOWNER))
		return -EPERM;
	if (value) {
		if (flags & EXT_ATTR_FLAG_USER) {
			void *value_copy;

			flags &= ~EXT_ATTR_FLAG_USER;
			if (size > acl_ea_size(
				   ACL_EA_MAX_ENTRIES))
				return -EINVAL;
			value_copy = kmalloc(size, GFP_KERNEL);
			if (!value_copy)
				return -ENOMEM;
			error = copy_from_user(value_copy, value, size);
			if (error) {
				kfree(value_copy);
				return error;
			}
			acl = acl_from_ea(value_copy, size);
			kfree(value_copy);
		} else
			acl = acl_from_ea(value, size);
		if (IS_ERR(acl))
			return PTR_ERR(acl);
		else if (acl) {
			error = posix_acl_valid(acl);
			if (error)
				goto release_and_out;
		}
	} else
		acl = NULL;
	if (!strcmp(name, ACL_EA_ACCESS))
		type = ACL_TYPE_ACCESS;
	else
		type = ACL_TYPE_DEFAULT;
	error = acl_ea_set_inode(inode, type, acl);

release_and_out:
	posix_acl_release(acl);
	return error;
}

/*
 * Convert from extended attribute to in-memory representation.
 */
static posix_acl_t *
acl_from_ea(
	const void *value,
	size_t size)
{
	const char *end = (char *)value + size;
	int n, count;
	posix_acl_t *acl;

	if (!value)
		return NULL;
	if (size < sizeof(acl_ea_header))
		 return ERR_PTR(-EINVAL);
	if (((acl_ea_header *)value)->a_version !=
	    cpu_to_le32(ACL_EA_VERSION))
		return ERR_PTR(-EINVAL);
	value = (char *)value + sizeof(acl_ea_header);
	count = acl_ea_count(size);
	if (count < 0)
		return ERR_PTR(-EINVAL);
	if (count == 0)
		return NULL;
	acl = posix_acl_alloc(count);
	if (!acl)
		return ERR_PTR(-ENOMEM);
	for (n=0; n < count; n++) {
		acl_ea_entry *entry =
			(acl_ea_entry *)value;
		if ((char *)value + sizeof(acl_ea_entry_short) > end)
			goto fail;
		acl->a_entries[n].e_tag  = le16_to_cpu(entry->e_tag);
		acl->a_entries[n].e_perm = le16_to_cpu(entry->e_perm);
		switch(acl->a_entries[n].e_tag) {
			case ACL_USER_OBJ:
			case ACL_GROUP_OBJ:
			case ACL_MASK:
			case ACL_OTHER:
				value = (char *)value +
					sizeof(acl_ea_entry_short);
				acl->a_entries[n].e_id = ACL_UNDEFINED_ID;
				break;

			case ACL_USER:
			case ACL_GROUP:
				value = (char *)value + sizeof(acl_ea_entry);
				if ((char *)value > end)
					goto fail;
				acl->a_entries[n].e_id =
					le32_to_cpu(entry->e_id);
				break;
		}
	}
	if (value != end)
		goto fail;
	return acl;

fail:
	posix_acl_release(acl);
	return ERR_PTR(-EINVAL);
}

/*
 * Convert from in-memory to extended attribute representation.
 */
static void *
acl_to_ea(
	const posix_acl_t *acl,
	size_t *size)
{
	acl_ea_header *ext_acl;
	char *e;
	int n;

	*size = acl_ea_size(acl->a_count);
	ext_acl = (acl_ea_header *)kmalloc(*size, GFP_KERNEL);
	if (!ext_acl)
		return ERR_PTR(-ENOMEM);
	ext_acl->a_version = cpu_to_le32(ACL_EA_VERSION);
	e = (char *)ext_acl + sizeof(acl_ea_header);
	for (n=0; n < acl->a_count; n++) {
		acl_ea_entry *entry = (acl_ea_entry *)e;
		entry->e_tag  = cpu_to_le16(acl->a_entries[n].e_tag);
		entry->e_perm = cpu_to_le16(acl->a_entries[n].e_perm);
		switch(acl->a_entries[n].e_tag) {
			case ACL_USER:
			case ACL_GROUP:
				entry->e_id =
					cpu_to_le32(acl->a_entries[n].e_id);
				e += sizeof(acl_ea_entry);
				break;
			default:
				e += sizeof(acl_ea_entry_short);
				break;
		}
	}
	return (char *)ext_acl;
}

/*
 * Get the access or default ACL of an inode in in-memory representation.
 */
posix_acl_t *
acl_ea_get_inode(
	struct inode *inode,
	int type)
{
	const char *name;
	char *value;
	posix_acl_t *acl, **p_acl;
	const size_t size = acl_ea_size(
	                    ACL_EA_MAX_ENTRIES);
	int retval;

	switch(type) {
		case ACL_TYPE_ACCESS:
			name = ACL_EA_ACCESS;
			p_acl = &inode->i_acl;
			break;

		case ACL_TYPE_DEFAULT:
			name = ACL_EA_DEFAULT;
			p_acl = &inode->i_default_acl;
			break;

		default:
			return ERR_PTR(-EINVAL);
	}
	if (*p_acl != POSIX_ACL_NOT_CACHED)
		return posix_acl_dup(*p_acl);
	value = kmalloc(size, GFP_KERNEL);
	if (!value)
		return ERR_PTR(-ENOMEM);

	/* acl_ea_get_inode may be called from permission(),
	   in which case the kernel lock is not held in 2.4.x kernels. */
	lock_kernel();
	retval = ext_attr_get(inode, name, value, size, 0);
	unlock_kernel();

	if (retval == -ENOATTR || retval == -ENOSYS)
		acl = NULL;
	else if (retval < 0)
		acl = ERR_PTR(retval);
	else {
		acl = acl_from_ea(value, retval);
		if (!IS_ERR(acl))
			*p_acl = posix_acl_dup(acl);
	}
	kfree(value);
	return acl;
}

/*
 * Set the access or default ACL of an inode.
 */
int
acl_ea_set_inode(
	struct inode *inode,
	int type,
	posix_acl_t *acl)
{
	const char *name;
	void *value = NULL;
	posix_acl_t **p_acl;
	int size, error;

	switch(type) {
		case ACL_TYPE_ACCESS:
			name = ACL_EA_ACCESS;
			p_acl = &inode->i_acl;
			if (acl) {
				error = posix_acl_equiv_mode(
					acl, &inode->i_mode);
				if (error < 0)
					return error;
				else if (error == 0)
					acl = NULL;
			}
			break;

		case ACL_TYPE_DEFAULT:
			name = ACL_EA_DEFAULT;
			p_acl = &inode->i_default_acl;
			if (!S_ISDIR(inode->i_mode))
				return acl ? -EINVAL : 0;
			break;

		default:
			return -EINVAL;
	}
 	if (acl) {
		if (acl->a_count > ACL_EA_MAX_ENTRIES)
			return -ERANGE;
		value = acl_to_ea(acl, &size);
		if (IS_ERR(value))
			return (int)PTR_ERR(value);
	}

	/* The kernel lock may be held already. */
	lock_kernel();
	error = ext_attr_set(inode, name, value, size, 0);
	unlock_kernel();

	if (value)
		kfree(value);
	if (!error) {
		if (*p_acl && *p_acl != POSIX_ACL_NOT_CACHED)
			posix_acl_release(*p_acl);
		*p_acl = posix_acl_dup(acl);
	}
	return error;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
int
#else
static int __init
#endif
init_posix_acl(void)
{
	register_ext_attr_handler(&acl_ea_access_handler);
	register_ext_attr_handler(&acl_ea_default_handler);
	return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)

static void __exit
exit_posix_acl(void)
{
	unregister_ext_attr_handler(&acl_ea_access_handler);
	unregister_ext_attr_handler(&acl_ea_default_handler);
}

module_init(init_posix_acl)
module_exit(exit_posix_acl)

#endif  /* kernel >= 2.3.x */
