/*
 * linux/fs/ext2/ext_attr.c
 *
 * Copyright (C) 2000 by Andreas Gruenbacher, <a.gruenbacher@computer.org>
 *
 * Fix by Harrison Xing <harrison@mountainviewdata.com>.
 */

/*
 * Extended attributes are stored on disk blocks that allocated outside of
 * any inode. The i_file_acl field is them made to point to this allocated
 * block. If all extended attributes of an inode are identical, these
 * inodes may share the same extended attribute block. Such situations
 * are automatically detected by keeping a cache of recent attribute block
 * numbers and hashes over the block's contents.
 *
 *
 * Extended attribute block layout:
 *
 *   +------------------+
 *   | header           |
 *    entry 1          | |
 *   | entry 2          | | growing downwards
 *   | entry 3          | v
 *   | four null bytes  |
 *   | . . .            |
 *   | value 1          | ^
 *   | value 3          | | growing upwards
 *   | value 2          | |
 *   +------------------+
 *
 * The block header is followed by multiple entry descriptors. These entry
 * descriptors are variable in size, and alligned to EXT2_EXT_ATTR_PAD
 * byte boundaries. The entry descriptors are sorted by attribute name,
 * so that two extended attribute blocks can be compared efficiently.
 *
 * Attribute values are aligned to the end of the block, stored in
 * no specific order. They are also padded to EXT2_EXT_ATTR_PAD byte
 * boundaries. No additional gaps are left between them.
 */

#include <linux/version.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/locks.h>
#include <linux/slab.h>
#include <linux/ext2_fs.h>
#include <linux/ext_attr.h>
#include <linux/ext2_ext_attr.h>
#include <linux/mbcache.h>
#include <linux/quotaops.h>
#include <asm/semaphore.h>
#include <asm/uaccess.h>
#include <linux/compatmac.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
# define mark_buffer_dirty(bh) mark_buffer_dirty(bh, 1)
#endif

#define HDR(bh) ((struct ext2_ext_attr_header *)((bh)->b_data))
#define ENTRY(ptr) ((struct ext2_ext_attr_entry *)(ptr))
#define FIRST_ENTRY(bh) ENTRY(HDR(bh)+1)
#define IS_LAST_ENTRY(entry) (*(__u32 *)(entry) == 0)

#ifdef EXT2_EXT_ATTR_DEBUG
# define ea_idebug(inode, f...) do { \
		printk(KERN_DEBUG "inode %s:%ld: ", \
			kdevname(inode->i_dev), inode->i_ino); \
		printk(f); \
		printk("\n"); \
        } while (0)
# define ea_bdebug(bh, f...) do { \
		printk(KERN_DEBUG "block %s:%ld: ", \
			kdevname(bh->b_dev), bh->b_blocknr); \
		printk(f); \
		printk("\n"); \
        } while (0)
#else
# define ea_idebug(f...)
# define ea_bdebug(f...)
#endif
#define ea_error(inode, f...) do { \
		printk(KERN_ERR "inode %s:%ld: ", \
			kdevname(inode->i_dev), inode->i_ino); \
		printk(f); \
		printk("\n"); \
	} while (0)

#ifdef CONFIG_EXT2_FS_EXT_ATTR_SHARING

static int
ext2_ext_attr_cache_insert(
	struct buffer_head *bh);
static struct buffer_head *
ext2_ext_attr_cache_find(
	struct inode *inode,
	struct ext2_ext_attr_header *header);
static void
ext2_ext_attr_cache_remove(
	struct buffer_head *bh);
static inline void ext2_ext_attr_hash_entry(
	struct ext2_ext_attr_header *header,
	struct ext2_ext_attr_entry *entry);
static void ext2_ext_attr_rehash(
	struct ext2_ext_attr_header *header,
	struct ext2_ext_attr_entry *entry);
static int ext2_ext_attr_check_block(
	struct buffer_head *bh);
static int
ext2_ext_attr_cmp(
	struct ext2_ext_attr_header *header1,
	struct ext2_ext_attr_header *header2);

static struct mb_cache *ext2_ext_attr_cache /* = NULL */;
DECLARE_MUTEX(ext2_ext_attr_sem);

static inline int
ext2_ext_attr_lock(
	void)
{
	if (down_interruptible(&ext2_ext_attr_sem))
		return -EINTR;
	return 0;
}

static inline void
ext2_ext_attr_unlock(
	void)
{
	up(&ext2_ext_attr_sem);
}

#else
# define ext2_ext_attr_cache_insert(bh) 0
# define ext2_ext_attr_cache_find(inode, header) NULL
# define ext2_ext_attr_cache_remove(bh) while(0) {}
# define ext2_ext_attr_rehash(header, entry) while(0) {}
# define ext2_ext_attr_lock() 0
# define ext2_ext_attr_unlock() while(0) {}
#endif

/*
 * ext2_get_ext_attr()
 *
 * Copy an extended attribute or a list of attribute names into the buffer
 * provided, or compute the buffer size required for those two operations.
 * Buffer is NULL to compute the size of the buffer required. Name is NULL
 * to retrieve the list of extended attribute names defined for this inode.
 * If the EXT_ATTR_FLAG_USER flag is set, buffer is assumed to be a user
 * space pointer.
 *
 * Returns a negative error number on failure, or the number of bytes
 * used / required on success.
 */
int
ext2_get_ext_attr(
	struct inode *inode,
	const char *name,
	void *buffer,
	size_t buffer_size,
	int flags)
{
	struct buffer_head *bh = NULL;
	struct ext2_ext_attr_entry *entry;
	unsigned int block, size;
	char *end;
	int error;

	ea_idebug(inode, "name=%s, buffer=%p, buffer_size=%d",
		  name, buffer, buffer_size);

	if (!inode->i_sb)
		return -ENOENT;
	if (!IS_EXT_ATTR(inode))
		return -ENOTSUP;
	if (!inode->u.ext2_i.i_file_acl)
		return -ENOATTR;
	block = inode->u.ext2_i.i_file_acl;
	ea_idebug(inode, "reading block %d", block);
	bh = bread(inode->i_dev, block, inode->i_sb->s_blocksize);
	if (!bh)
		return -EIO;
	end = bh->b_data + bh->b_size;
	if (HDR(bh)->h_magic != cpu_to_le32(EXT2_EXT_ATTR_MAGIC) ||
	    HDR(bh)->h_blocks != cpu_to_le32(1)) {
bad_block:
		ea_error(inode, "bad block %d", block);
		error = -EIO;
		goto cleanup;
	}
	if (name) {
		/* find named attribute */
		int name_len = strlen(name);

		error = -ERANGE;
		if (name_len == 0 || name_len > 255)
			goto cleanup;
		entry = FIRST_ENTRY(bh);
		while (!IS_LAST_ENTRY(entry)) {
			struct ext2_ext_attr_entry *next =
				EXT2_EXT_ATTR_NEXT(entry);
			if ((char *)next >= end)
				goto bad_block;
			if (name_len == entry->e_name_len &&
			    memcmp(name, entry->e_name, name_len) == 0)
				goto found;
			entry = next;
		}
		error = -ENOATTR;
		goto cleanup;
	found:

		/* return value of attribute */
		if (entry->e_value_block != 0)
			goto bad_block;
		size = le32_to_cpu(entry->e_value_size);
		/* n.b. overflow test only needed on 32bit architecture */
		if ((int)size < 0 ||
		    le16_to_cpu(entry->e_value_offs) + size >
		    inode->i_sb->s_blocksize)
			goto bad_block;
	} else {
		/* return list of attribute names */
		size = 1;  /* trailing 0 character */
		entry = FIRST_ENTRY(bh);
		while (!IS_LAST_ENTRY(entry)) {
			struct ext2_ext_attr_entry *next =
				EXT2_EXT_ATTR_NEXT(entry);
			if ((char *)next >= end)
				goto bad_block;
			size += entry->e_name_len + 1;
			entry = next;
		}
	}
	error = ext2_ext_attr_cache_insert(bh);
	if (error)
		ea_idebug(inode, "cache insert failed with %d", error);
	if (!buffer) {
		error = size;
		goto cleanup;
	} else {
		error = -ERANGE;
		if (size > buffer_size)
			goto cleanup;
	}

	if (flags & EXT_ATTR_FLAG_USER) {
		error = verify_area(VERIFY_WRITE, buffer, size);
		if (error)
			goto cleanup;
		if (name) {
			char *value = bh->b_data +
				le16_to_cpu(entry->e_value_offs);
			error = copy_to_user(buffer, value, size);
			if (error)
				goto cleanup;
		} else {
			char *b = (char *)buffer;
			entry = FIRST_ENTRY(bh);
			while (!IS_LAST_ENTRY(entry)) {
				error = copy_to_user(b, entry->e_name,
				                     entry->e_name_len);
				if (error)
					goto cleanup;
				b += entry->e_name_len;
				put_user('\0', b++);
				entry = EXT2_EXT_ATTR_NEXT(entry);
			}
			put_user('\0', b);
		}
	} else {
		if (name) {
			char *value = bh->b_data +
				le16_to_cpu(entry->e_value_offs);
			memcpy(buffer, value, size);
		} else {
			char *b = (char *)buffer;
			entry = FIRST_ENTRY(bh);
			while (!IS_LAST_ENTRY(entry)) {
				memcpy(b, entry->e_name, entry->e_name_len);
				b += entry->e_name_len;
				// ### is this better than *... = 0
				// on Alpha etc.?
				memset(b++, '\0', 1);
				entry = EXT2_EXT_ATTR_NEXT(entry);
			}
			memset(b, '\0', 1);
		}
	}
	error = size;

cleanup:
	brelse(bh);

	return error;
}


/*
 * ext2_set_ext_attr()
 *
 * Create, replace or remove an extended attribute for this inode. Buffer
 * is NULL to remove an existing extended attribute, and non-NULL to
 * either replace an existing extended attribute, or create a new extended
 * attribute. The flags EXT_ATTR_FLAG_EXISTS and EXT_ATTR_FLAG_CREATE
 * specify that an extended attribute must exist and must not exist
 * previous to the call, respectively. If the EXT_ATTR_FLAG_USER flag is
 * set, buffer is assumed to be a user space pointer.
 *
 * Returns 0, or a negative error number on failure.
 */
int
ext2_set_ext_attr(
	struct inode *inode,
	const char *name,
	void *buffer,
	size_t buffer_len,
	int flags)
{
	struct buffer_head *old_bh = NULL, *new_bh = NULL;
	struct ext2_ext_attr_header *header = NULL;
	struct ext2_ext_attr_entry *here, *last;
	unsigned int name_len, blocks;
	int min_offs, found = 0, error;
	char *end;
	
	ea_idebug(inode, "name=%s, buffer=%p, buffer_len=%d",
	          name, buffer, buffer_len);

	if (IS_RDONLY(inode))
		return -EROFS;
	if (!inode->i_sb)
		return -ENOENT;
	if (!IS_EXT_ATTR(inode))
		return -ENOTSUP;
	if (buffer == NULL)
		buffer_len = 0;
	if (name == NULL)
		return -EINVAL;
	name_len = strlen(name);
	if (name_len == 0 || name_len > 255 ||
	    buffer_len > inode->i_sb->s_blocksize)
		return -ERANGE;
	error = ext2_ext_attr_lock();
	if (error)
		return error;
	if (inode->u.ext2_i.i_file_acl) {
		int block = inode->u.ext2_i.i_file_acl, free;

		old_bh = bread(inode->i_dev, block, inode->i_sb->s_blocksize);
		error = -EIO;
		if (!old_bh)
			goto cleanup;
		header = HDR(old_bh);
		end = old_bh->b_data + old_bh->b_size;
		if (HDR(old_bh)->h_magic != cpu_to_le32(EXT2_EXT_ATTR_MAGIC) ||
		    HDR(old_bh)->h_blocks != cpu_to_le32(1)) {
bad_block:
			ea_error(inode, "bad block %d", block);
			error = -EIO;
			goto cleanup;
		}
		ea_idebug(inode, "block %d refcount is %d", block,
		          le32_to_cpu(HDR(old_bh)->h_refcount));
		/* find named attribute */
		min_offs = old_bh->b_size;
		last = FIRST_ENTRY(old_bh);
		while (!IS_LAST_ENTRY(last)) {
			struct ext2_ext_attr_entry *next =
				EXT2_EXT_ATTR_NEXT(last);
			if ((char *)next >= end)
				goto bad_block;
			if (last->e_value_block == 0 &&
			    last->e_value_size != 0) {
				int offs = le16_to_cpu(last->e_value_offs);
				if (offs < min_offs)
					min_offs = offs;
			}
			if (name_len <= last->e_name_len) {
				int cmp = memcmp(name, last->e_name, name_len);
				if (cmp == 0 && name_len == last->e_name_len) {
					found = 1;
					break;
				} else if (cmp < 0)
					break;
			} else {
				if (memcmp(name, last->e_name,
				           last->e_name_len) < 0)
					break;
			}
			last = next;
		}
		here = last;
		while (!IS_LAST_ENTRY(last)) {
			struct ext2_ext_attr_entry *next =
				EXT2_EXT_ATTR_NEXT(last);
			if ((char *)next >= end)
				goto bad_block;
			if (last->e_value_block == 0 &&
			    last->e_value_size != 0) {
				int offs = le16_to_cpu(last->e_value_offs);
				if (offs < min_offs)
					min_offs = offs;
			}
			last = next;
		}

		free = min_offs - ((char*)last - (char*)HDR(old_bh)) -
			sizeof(__u32);
		if (found) {
			error = -EEXIST;
			if (flags & EXT_ATTR_FLAG_CREATE)
				goto cleanup;
			if (here->e_value_block == 0 &&
			    here->e_value_size != 0) {
				unsigned int size =
					le32_to_cpu(here->e_value_size);
				/* n.b. overflow test only needed on 32bit
				   architecture */
				if ((int)size < 0 ||
				    le16_to_cpu(here->e_value_offs) + size > 
				    inode->i_sb->s_blocksize)
					goto bad_block;
			}
			free += EXT2_EXT_ATTR_SIZE(
			        le32_to_cpu(here->e_value_size));
		} else {
			error = -ENOATTR;
			if (flags & EXT_ATTR_FLAG_EXISTS)
				goto cleanup;
			if (buffer != NULL)
				free -= EXT2_EXT_ATTR_LEN(name_len);
		}
		free -= EXT2_EXT_ATTR_SIZE(buffer_len);
		error = -ENOSPC;
		if (free < 0)
			goto cleanup;
		if (HDR(old_bh)->h_refcount == cpu_to_le32(1)) {
			ext2_ext_attr_cache_remove(old_bh);
			new_bh = old_bh;
		} else {
			int delta;

			header = kmalloc(old_bh->b_size, GFP_KERNEL);
			error = -ENOMEM;
			if (header == NULL)
				goto cleanup;
			memcpy(header, HDR(old_bh), old_bh->b_size);
			header->h_refcount = cpu_to_le32(1);
			delta = (char *)header - old_bh->b_data;
			here = ENTRY((char *)here + delta);
			last = ENTRY((char *)last + delta);
		}
	} else if (buffer != NULL) {
		int free;

		error = -ENOATTR;
		if (flags & EXT_ATTR_FLAG_EXISTS)
			goto cleanup;

		free = inode->i_sb->s_blocksize -
			sizeof(struct ext2_ext_attr_header) - sizeof(__u32);
		free -= EXT2_EXT_ATTR_LEN(name_len);
		free -= EXT2_EXT_ATTR_SIZE(buffer_len);
		error = -ENOSPC;
		if (free < 0)
			goto cleanup;

		header = kmalloc(inode->i_sb->s_blocksize, GFP_KERNEL);
		error = -ENOMEM;
		if (header == NULL)
			goto cleanup;
		memset(header, '\0', inode->i_sb->s_blocksize);
		end = (char *)header + inode->i_sb->s_blocksize;
		header->h_magic = cpu_to_le32(EXT2_EXT_ATTR_MAGIC);
		header->h_blocks = header->h_refcount = cpu_to_le32(1);
		last = here = ENTRY(header+1);
		min_offs = inode->i_sb->s_blocksize;
	} else {
		/* request to remove a nonexistent attribute */
		error = (flags & EXT_ATTR_FLAG_EXISTS) ? -ENOATTR : 0;
		goto cleanup;
	}

	if (found) {
		/* remove value */
		if (here->e_value_block == 0 && here->e_value_size != 0) {
			char *first_value = (char *)header + min_offs;
			int offs = le16_to_cpu(here->e_value_offs);
			char *value = (char *)header + offs;
			size_t size = EXT2_EXT_ATTR_SIZE(
				le32_to_cpu(here->e_value_size));
			memmove(first_value + size, first_value,
			        value - first_value);
			memset(first_value, '\0', size);
			here->e_value_offs = 0;
			min_offs += size;

			/* adjust offsets */
			last = ENTRY(header+1);
			while (!IS_LAST_ENTRY(last)) {
				int o = le16_to_cpu(last->e_value_offs);
				if (last->e_value_block == 0 && o < offs)
					last->e_value_offs =
						cpu_to_le16(o + size);
				last = EXT2_EXT_ATTR_NEXT(last);
			}
		}
		if (buffer == NULL) {
			if (EXT2_EXT_ATTR_NEXT(ENTRY(header+1)) == last) {
				/* block now is empty */
				inode->u.ext2_i.i_file_acl = 0;
				new_bh = NULL;
				goto update_inode;
			} else {
				/* remove name */
				int size = EXT2_EXT_ATTR_LEN(name_len);
				last = ENTRY((char *)last - size);
				memmove(here, (char*)here + size,
					(char*)last - (char*)here);
				memset(last, '\0', size);
			}
		}
	} else if (buffer != NULL) {
		/* insert new name */
		int size = EXT2_EXT_ATTR_LEN(name_len);
		int rest = (char *)last - (char *)here;
		memmove((char *)here + size, here, rest);
		memset(here, '\0', size);
		here->e_name_len = name_len;
		memcpy(here->e_name, name, name_len);
	}
	if (buffer != NULL) {
		/* fill in value */
		here->e_value_size = cpu_to_le32(buffer_len);
		if (buffer_len != 0) {
			size_t size = EXT2_EXT_ATTR_SIZE(buffer_len);
			char *value = (char *)header + min_offs - size;
			here->e_value_offs =
				cpu_to_le16((char *)value - (char *)header);
			memset(value + size - EXT2_EXT_ATTR_PAD, '\0',
			       EXT2_EXT_ATTR_PAD); /* clear pad bytes */
			if (flags & EXT_ATTR_FLAG_USER) {
				error = copy_from_user(value, buffer,
						       buffer_len);
				if (error)
					goto cleanup;
			} else
				memcpy(value, buffer, buffer_len);
		}
	}
	ext2_ext_attr_rehash(header, here);

	/* update filesystem */
	if (new_bh == NULL) {
		/* no previous block, or previous block has been cloned */
		new_bh = ext2_ext_attr_cache_find(inode, header);
		if (new_bh != NULL) {
			if (old_bh != new_bh) {
				lock_super(inode->i_sb);
				error = DQUOT_ALLOC_BLOCK(inode->i_sb,
				                          inode, 1);
				unlock_super(inode->i_sb);
				if(error)
					goto cleanup;
				HDR(new_bh)->h_refcount = cpu_to_le32(
				      le32_to_cpu(HDR(new_bh)->h_refcount) + 1);
				ea_idebug(inode, "re-using block %ld "
				          "(refcount now is %d)",
					  new_bh->b_blocknr,
					  le32_to_cpu(HDR(new_bh)->h_refcount));
			}
		}
		if (new_bh == NULL) {
			int block = le32_to_cpu(inode->i_sb->u.ext2_sb.s_es->
				s_first_data_block) +
				inode->u.ext2_i.i_block_group *
				EXT2_BLOCKS_PER_GROUP(inode->i_sb);
			/* ### better goal block? */
			block = ext2_new_block(inode, block, 0, 0, &error);
			if (error)
				goto cleanup;
			ea_idebug(inode, "allocating block %d", block);

			new_bh = getblk(inode->i_dev, block,
			                inode->i_sb->s_blocksize);
			if (!new_bh) {
				ext2_free_blocks(inode, block, 1);
				error = -EIO;
				goto cleanup;
			}
			memcpy(new_bh->b_data, header, new_bh->b_size);
			mark_buffer_uptodate(new_bh, 1);
			ext2_ext_attr_cache_insert(new_bh);
			
			if (!EXT2_HAS_COMPAT_FEATURE(inode->i_sb,
			                        EXT2_FEATURE_COMPAT_EXT_ATTR)) {
				lock_super(inode->i_sb);
				inode->i_sb->u.ext2_sb.s_es->s_feature_compat |=
				      cpu_to_le32(EXT2_FEATURE_COMPAT_EXT_ATTR);
				mark_buffer_dirty(inode->i_sb->u.ext2_sb.s_sbh);
				inode->i_sb->s_dirt = 1;
				unlock_super(inode->i_sb);
			}
		}
	} else if (old_bh == new_bh) {
		new_bh = ext2_ext_attr_cache_find(inode, header);
		if (new_bh) {
			HDR(new_bh)->h_refcount = cpu_to_le32(
				le32_to_cpu(HDR(new_bh)->h_refcount) + 1);
		} else {
			new_bh = old_bh;
			ext2_ext_attr_cache_insert(new_bh);
		}
	}

	if (new_bh) {
		/* update the new block */
		ea_idebug(inode, "updating block %ld", new_bh->b_blocknr);
		mark_buffer_dirty(new_bh);
		if (IS_SYNC(inode)) {
			ll_rw_block(WRITE, 1, &new_bh);
			wait_on_buffer(new_bh);
			error = -EIO;
			if (buffer_req(new_bh) && !buffer_uptodate(new_bh))
				goto cleanup;
		}
		inode->u.ext2_i.i_file_acl = new_bh->b_blocknr;
	} else
		inode->u.ext2_i.i_file_acl = 0;

update_inode:
	blocks = inode->i_blocks;
	if (old_bh)
		blocks -= (inode->i_sb->s_blocksize >> 9);
	if (new_bh)
		blocks += (inode->i_sb->s_blocksize >> 9);
	inode->i_blocks = blocks;
	inode->i_ctime = CURRENT_TIME;
	/* inode->i_version = ++global_event; */
	mark_inode_dirty(inode);
	error = 0;
	if (IS_SYNC(inode))
		error = ext2_sync_inode(inode);

	if (!error && old_bh && old_bh != new_bh) {
		unsigned int refcount = le32_to_cpu(HDR(old_bh)->h_refcount);
		if (refcount == 1) {
			/* free the old block */
			ea_idebug(inode, "freeing block %ld",
			          old_bh->b_blocknr);
			HDR(old_bh)->h_magic = 0;
			ext2_free_blocks(inode, old_bh->b_blocknr, 1);
		} else {
			/* decrement refcount only */
			ea_idebug(inode, "block %ld refcount now is %d",
			          old_bh->b_blocknr, refcount);
			HDR(old_bh)->h_refcount = cpu_to_le32(refcount - 1);
			lock_super(inode->i_sb);
			DQUOT_FREE_BLOCK(inode->i_sb, inode, 1);
			unlock_super(inode->i_sb);
		}
		mark_buffer_dirty(old_bh);  /* no need to sync */
	}

cleanup:
	if (new_bh && new_bh != old_bh)
		brelse(new_bh);
	if (!old_bh || header != HDR(old_bh))
		kfree(header);
	if (old_bh)
		brelse(old_bh);
	ext2_ext_attr_unlock();

	return error;
}

/*
 * ext2_ext_attr_free_inode()
 *
 * Free extended attribute resources associated with this inode. This
 * is called immediately before an inode is freed.
 */
void
ext2_ext_attr_free_inode(
	struct inode *inode)
{
	struct buffer_head *bh;
	unsigned int block = inode->u.ext2_i.i_file_acl;

	if (!block)
		return;
	if (ext2_ext_attr_lock())
		return;
	bh = bread(inode->i_dev, block, inode->i_sb->s_blocksize);
	if (!bh) {
		ea_error(inode, "block %d: I/O error", block);
		goto cleanup;
	}
	if (HDR(bh)->h_magic != cpu_to_le32(EXT2_EXT_ATTR_MAGIC) ||
	    HDR(bh)->h_blocks != cpu_to_le32(1)) {
		ea_error(inode, "bad block %d", block);
		brelse(bh);
		goto cleanup;
	}
	if (HDR(bh)->h_refcount == cpu_to_le32(1)) {
		ext2_ext_attr_cache_remove(bh);
		ext2_free_blocks(inode, block, 1);
		HDR(bh)->h_magic = 0;
		mark_buffer_dirty(bh);
		/* no need to sync */
		bforget(bh);
	} else {
		HDR(bh)->h_refcount = cpu_to_le32(
			le32_to_cpu(HDR(bh)->h_refcount) - 1);
		mark_buffer_dirty(bh);
		if (IS_SYNC(inode)) {
			ll_rw_block(WRITE, 1, &bh);
			wait_on_buffer(bh);
		}
		brelse(bh);
		lock_super(inode->i_sb);
		DQUOT_FREE_BLOCK(inode->i_sb, inode, 1);
		unlock_super(inode->i_sb);
	}
	inode->u.ext2_i.i_file_acl = 0;
	inode->i_blocks -= (inode->i_sb->s_blocksize >> 9);

cleanup:
	ext2_ext_attr_unlock();
}

/*
 * ext2_ext_attr_put_super()
 *
 * This is called when a partition is unmounted.
 */
void
ext2_ext_attr_put_super(
	struct super_block *sb)
{
#ifdef CONFIG_EXT2_FS_EXT_ATTR_SHARING
	mb_cache_shrink(ext2_ext_attr_cache, sb->s_dev);
#endif
}

#ifdef CONFIG_EXT2_FS_EXT_ATTR_SHARING

/*
 * ext2_ext_attr_cache_insert()
 *
 * Create a new entry in the extended attribute cache, and insert
 * it unless such an entry is already in the cache.
 *
 * Returns 0, or a negative error number on failure.
 */
static int
ext2_ext_attr_cache_insert(
	struct buffer_head *bh)
{
	unsigned int hash = le32_to_cpu(HDR(bh)->h_hash);
	struct mb_cache_entry *ce;
	int error;

	ce = mb_cache_entry_alloc(ext2_ext_attr_cache);
	if (!ce)
		return -ENOMEM;
	error = mb_cache_entry_insert(ce, bh->b_dev, bh->b_blocknr, &hash);
	if (error) {
		mb_cache_entry_free(ce);
		if (error == -EBUSY) {
			ea_bdebug(bh, "already in cache (%d cache entries)",
				  atomic_read(&ext2_ext_attr_cache->
				                                c_entry_count));
			error = 0;
		}
	} else {
		ea_bdebug(bh, "inserting [%x] (%d cache entries)", hash,
			  atomic_read(&ext2_ext_attr_cache->c_entry_count));
		mb_cache_entry_release(ce);
	}
	return error;
}

/*
 * ext2_ext_attr_cache_find()
 *
 * Find an identical extended attribute block.
 *
 * Returns a pointer to the block found, or NULL if such a block was
 * not found or an error occurred.
 */
static struct buffer_head *
ext2_ext_attr_cache_find(
	struct inode *inode,
	struct ext2_ext_attr_header *header)
{
	unsigned int hash = le32_to_cpu(header->h_hash);
	struct mb_cache_entry *ce;

	if (!header->h_hash)
		return NULL;  /* never share */
	ea_idebug(inode, "looking for identical blocks in the cache [%x]",
	          hash);
	ce = mb_cache_entry_find_first(ext2_ext_attr_cache, 0,
	                               inode->i_dev, hash);
	while (ce) {
		struct buffer_head *bh = bread(inode->i_dev, ce->e_block,
		                               inode->i_sb->s_blocksize);
		ea_idebug(inode, "checking block %ld", ce->e_block);
		if (!bh) {
			ea_error(inode, "read error block %ld", ce->e_block);
			goto next_match;
		}
		if (ext2_ext_attr_check_block(bh) < 0) {
			ea_error(inode, "bad block %ld", ce->e_block);
			brelse(bh);
			goto next_match;
		}
		if (le32_to_cpu(HDR(bh)->h_refcount) >
		    EXT2_EXT_ATTR_REFCOUNT_MAX) {
			ea_idebug(inode, "block %ld refcount %d too high",
			          ce->e_block,
				  le32_to_cpu(HDR(bh)->h_refcount));
			goto next_match;
		}
		if (!ext2_ext_attr_cmp(header, HDR(bh))) {
			mb_cache_entry_release(ce);
			return bh;
		} else
			brelse(bh);
next_match:
		ce = mb_cache_entry_find_next(ce, 0, inode->i_dev, hash);
	}
	return NULL;
}

/*
 * ext2_ext_attr_cache_remove()
 *
 * Remove the cache entry of a block from the cache. Called when a
 * block becomes invalid.
 */
static void
ext2_ext_attr_cache_remove(
	struct buffer_head *bh)
{
	struct mb_cache_entry *ce;

	ce = mb_cache_entry_get(ext2_ext_attr_cache, bh->b_dev, bh->b_blocknr);
	if (ce) {
		ea_bdebug(bh, "removing (%d cache entries remaining)",
			  atomic_read(&ext2_ext_attr_cache->c_entry_count)-1);
		mb_cache_entry_free(ce);
	} else 
		ea_bdebug(bh, "no cache entry");
}

#define NAME_HASH_SHIFT 5
#define VALUE_HASH_SHIFT 16

/*
 * ext2_ext_attr_hash_entry()
 *
 * Compute the hash of an extended attribute.
 */
static inline void ext2_ext_attr_hash_entry(
	struct ext2_ext_attr_header *header,
	struct ext2_ext_attr_entry *entry)
{
	unsigned int hash = 0;
	char *name = entry->e_name;
	int n;

	for (n=0; n < entry->e_name_len; n++) {
		hash = (hash << NAME_HASH_SHIFT) ^
		       (hash >> (8*sizeof(hash) - NAME_HASH_SHIFT)) ^
		       *name++;
	}

	if (entry->e_value_block == 0 && entry->e_value_size != 0) {
		__u32 *value = (__u32 *)((char *)header +
			le32_to_cpu(entry->e_value_offs));
		for (n = (entry->e_value_size + EXT2_EXT_ATTR_ROUND) >>
		     EXT2_EXT_ATTR_PAD_BITS; n; n--) {
			hash = (hash << VALUE_HASH_SHIFT) ^
			       (hash >> (8*sizeof(hash) - VALUE_HASH_SHIFT)) ^
			       le32_to_cpu(*value++);
		}
	}
	entry->e_hash = cpu_to_le32(hash);
}

#undef NAME_HASH_SHIFT
#undef VALUE_HASH_SHIFT

#define BLOCK_HASH_SHIFT 16

/*
 * ext2_ext_attr_rehash()
 *
 * Re-compute the extended attribute hash value after an entry has changed.
 */
static void ext2_ext_attr_rehash(
	struct ext2_ext_attr_header *header,
	struct ext2_ext_attr_entry *entry)
{
	struct ext2_ext_attr_entry *here;
	unsigned int hash = 0;
	
	ext2_ext_attr_hash_entry(header, entry);
	here = ENTRY(header+1);
	while (!IS_LAST_ENTRY(here)) {
		if (!here->e_hash) {
			hash = 0;
			break;
		}
		hash = (hash << BLOCK_HASH_SHIFT) ^
		       (hash >> (8*sizeof(hash) - BLOCK_HASH_SHIFT)) ^
		       le32_to_cpu(here->e_hash);
		here = EXT2_EXT_ATTR_NEXT(here);
	}
	header->h_hash = cpu_to_le32(hash);
}

#undef BLOCK_HASH_SHIFT

/*
 * ext2_ext_attr_check_block()
 *
 * Check the validity of an extended attribute block.
 *
 * Returns 0, or a negative error number on errors.
 */
static int ext2_ext_attr_check_block(
	struct buffer_head *bh)
{
	struct ext2_ext_attr_entry *entry;
	char *end = bh->b_data + bh->b_size;

	if (HDR(bh)->h_magic != cpu_to_le32(EXT2_EXT_ATTR_MAGIC) ||
	    HDR(bh)->h_blocks != cpu_to_le32(1))
		goto bad_block;

	entry = FIRST_ENTRY(bh);
	while (!IS_LAST_ENTRY(entry)) {
		struct ext2_ext_attr_entry *next =
			EXT2_EXT_ATTR_NEXT(entry);
		if ((char *)next >= end)
			goto bad_block;
		if (entry->e_name_index != 0 || entry->e_value_block != 0)
			goto bad_block;
		if (entry->e_value_size != 0) {
			unsigned int offs = le16_to_cpu(entry->e_value_offs);
			unsigned int size = le32_to_cpu(entry->e_value_size);

			/* n.b. overflow test only needed on
			        32bit architecture */
			if ((int)size < 0 || offs + size > bh->b_size)
				goto bad_block;
		}
		entry = next;
	}
	return 0;

bad_block:
	return -EIO;
}

/*
 * ext2_ext_attr_cmp()
 *
 * Compare two extended attribute blocks for equality.
 *
 * Returns 0 if the blocks are equal, 1 if they differ, and
 * a negative error number on errors.
 */
static int
ext2_ext_attr_cmp(
	struct ext2_ext_attr_header *header1,
	struct ext2_ext_attr_header *header2)
{
	struct ext2_ext_attr_entry *entry1, *entry2;

	entry1 = ENTRY(header1+1);
	entry2 = ENTRY(header2+1);
	while (!IS_LAST_ENTRY(entry1)) {
		if (IS_LAST_ENTRY(entry2))
			return 1;
		if (entry1->e_hash != entry2->e_hash ||
		    entry1->e_name_len != entry2->e_name_len ||
		    entry1->e_value_size != entry2->e_value_size ||
		    memcmp(entry1->e_name, entry2->e_name, entry1->e_name_len))
			return 1;
		if (entry1->e_value_block != 0 || entry2->e_value_block != 0)
			goto bad_block;
		if (memcmp((char *)header1 + le16_to_cpu(entry1->e_value_offs),
		           (char *)header2 + le16_to_cpu(entry2->e_value_offs),
		           le32_to_cpu(entry1->e_value_size)))
			return 1;

		entry1 = EXT2_EXT_ATTR_NEXT(entry1);
		entry2 = EXT2_EXT_ATTR_NEXT(entry2);
	}
	if (!IS_LAST_ENTRY(entry2))
		return 1;
	return 0;

bad_block:
	return -EIO;
}

#endif  /* CONFIG_EXT2_FS_EXT_ATTR_SHARING */

int ext2_ext_attr_init(void)
{
#ifdef CONFIG_EXT2_FS_EXT_ATTR_SHARING
	ext2_ext_attr_cache = mb_cache_create("ext2_ext_attr", NULL,
		sizeof(struct mb_cache_entry) +
		sizeof(struct mb_cache_entry_index), 1, 61);
	if (!ext2_ext_attr_cache)
		return -ENOMEM;
#endif  /* CONFIG_EXT2_FS_EXT_ATTR_SHARING */
	return 0;
}

void ext2_ext_attr_done(void)
{
#ifdef CONFIG_EXT2_FS_EXT_ATTR_SHARING
	mb_cache_destroy(ext2_ext_attr_cache);
#endif  /* CONFIG_EXT2_FS_EXT_ATTR_SHARING */
}
