/*
 * linux/fs/ext_attr.c  (extended attribute handling)
 *
 *  Copyright (C) 2000 by Andreas Gruenbacher <a.gruenbacher@computer.org>
 */

/*
 * Here we have the extended attribute handler management, system calls,
 * and VFS functions.
 */

#include <linux/version.h>
#include <asm/uaccess.h>
#include <linux/ctype.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/smp_lock.h>
#include <linux/ext_attr.h>
#include <linux/module.h>

#ifdef CONFIG_FS_EXT_ATTR

EXPORT_SYMBOL(ext_attr_get);
EXPORT_SYMBOL(ext_attr_set);

static LIST_HEAD(system_attr);

int register_ext_attr_handler(struct ext_attr_handler *handler)
{
	list_add(&handler->link, &system_attr);
	return 0;
}

void unregister_ext_attr_handler(struct ext_attr_handler *handler)
{
	list_del(&handler->link);
}

static int ext_attr_inode(struct inode *, int, const char *, void *, size_t);

#ifdef CONFIG_FS_USER_EXT_ATTR

static int ext_attr_user_get(struct inode *, const char *, void *, size_t, int);
static int ext_attr_user_set(struct inode *, const char *, void *, size_t, int);

/* handler for user-defined attributes */
static struct ext_attr_handler ext_attr_user_handler = {
	get_inode:	ext_attr_user_get,
	set_inode:	ext_attr_user_set,
};

#endif  /* CONFIG_FS_USER_EXT_ATTR */

/* find the handler for an attribute name */
static struct ext_attr_handler *dispatch_handler(const char *name)
{
	struct list_head *l;

#ifdef CONFIG_FS_USER_EXT_ATTR
	if (((*name >= 'A' && *name <= 'Z') ||
	     (*name >= 'a' && *name <= 'z')))
		return &ext_attr_user_handler;
#endif  /* CONFIG_FS_USER_EXT_ATTR */

	for (l = system_attr.next; l != &system_attr; l = l->next) {
		struct ext_attr_handler *handler =
			list_entry(l, struct ext_attr_handler, link);
		if (!strcmp(name, handler->name))
			return handler;
	}
	return ERR_PTR(-ENOTSUP);
}

/*
 * System call given a path name (see ext_attr_path(2)).
 */
asmlinkage int
sys_ext_attr_path(const char *path, int cmd,
                  const char *name, void *value, size_t size)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,99)
	struct dentry *dentry;
	int error;

	lock_kernel();
	dentry = namei(path);
	error = PTR_ERR(dentry);
	if (!IS_ERR(dentry)) {
		error = ext_attr_inode(dentry->d_inode,
				       cmd, name, value, size);
		dput(dentry);
	}
	unlock_kernel();
#else
	struct nameidata nd;
	int error;

	error = user_path_walk(path, &nd);
	if (!error) {
		lock_kernel();
		error = ext_attr_inode(nd.dentry->d_inode, cmd,
		                       name, value, size);
		unlock_kernel();
		path_release(&nd);
	}
#endif
	return error;
}

/*
 * System call given a file descriptor (see ext_attr_fd(2)).
 */
asmlinkage int
sys_ext_attr_fd(int fd, int cmd,
                const char *name, void *value, size_t size)
{
	struct file *file;
	int error;

	lock_kernel();
	file = fget(fd);
	if (!file) {
		error = -EBADF;
		goto out;
	}
	if (!file->f_dentry || !file->f_dentry->d_inode) {
		error = -ENOENT;
		goto fput_and_out;
	}
	error = ext_attr_inode(file->f_dentry->d_inode, cmd, name, value, size);

fput_and_out:
	fput(file);
out:
	unlock_kernel();
	return error;
}

/*
 * Common part of sys_ext_attr_path and sys_ext_attr_fd
 */
static int ext_attr_inode(struct inode *inode, int cmd,
                           const char *name, void *value, size_t size)
{
	struct ext_attr_handler *handler;
	char name_copy[EXT_ATTR_NAME_MAX+1];
	int flags, error;

	if (S_ISLNK(inode->i_mode))
		return -EINVAL;  /* shouldn't get here */

	if (name) {
		error = strncpy_from_user(name_copy, name, sizeof(name_copy));
		if (error < 0)
			return error;
		if (error < 1 || error > EXT_ATTR_NAME_MAX)
			return -ERANGE;
		name = name_copy;
	}

#ifdef EXT_ATTR_DEBUG
	printk(KERN_DEBUG "cmd=%d, name=%s\n", cmd, name);
#endif
	
	/* The high bits of cmd contain the flags */
	flags = (cmd & ~0xFFFF) | EXT_ATTR_FLAG_USER;
	cmd &= 0xFFFF;

	switch(cmd) {
		case EXT_ATTR_GET:
			if (name) {
				handler = dispatch_handler(name);
				if (IS_ERR(handler))
					error = PTR_ERR(handler);
				else if (!handler->get_inode)
					error = -ENOTSUP;
				else
					error = handler->get_inode(inode,
						name, value, size, flags);
			} else
				error = ext_attr_get(inode,
					NULL, value, size, flags);
			break;

		case EXT_ATTR_GET_SIZE:
			if (name) {
				handler = dispatch_handler(name);
				if (IS_ERR(handler))
					error = PTR_ERR(handler);
				else if (!handler->get_inode)
					error = -ENOTSUP;
				else
					error = handler->get_inode(inode,
						name, NULL, 0, 0);
			} else
				error = ext_attr_get(inode, NULL, NULL, 0, 0);
			break;

		case EXT_ATTR_SET:
			if (!name)
				return -EINVAL;
			if (size > EXT_ATTR_VALUE_MAX)
				return -ERANGE;
			handler = dispatch_handler(name);
			if (IS_ERR(handler))
				error = PTR_ERR(handler);
			else if (!handler->set_inode)
				error = -ENOTSUP;
			else
				error = handler->set_inode(inode,
					name, value, size, flags);
			break;

		default:
			return -EINVAL;
	}

	return error;
}

/*
 * get an extended attribute of an inode
 */
int ext_attr_get(struct inode *inode, const char *name,
                 void *value, size_t size, int flags)
{
	int error = -ENOTSUP;

	if (inode->i_op && inode->i_op->get_ext_attr) {
		error = inode->i_op->get_ext_attr(inode, name,
		                                  value, size, flags);
	}
	return error;
}

/*
 * set an extended attribute of an inode
 */
int ext_attr_set(struct inode *inode, const char *name,
                 void *value, size_t size, int flags)
{
	int error = -ENOTSUP;

	if (inode->i_op && inode->i_op->set_ext_attr) {
		error = inode->i_op->set_ext_attr(inode, name,
                                                  value, size, flags);
	}
	return error;
}

#ifdef CONFIG_FS_USER_EXT_ATTR

/*
 * get an extended user attribute of an inode
 */
static int ext_attr_user_get(struct inode *inode, const char *name,
                             void *value, size_t size, int flags)
{
	int error;
	
	if (!IS_EXT_ATTR_USER(inode))
		return -ENOTSUP;
	error = permission(inode, MAY_READ);
	if (error)
		return error;
	return ext_attr_get(inode, name, value, size, flags);
}

/*
 * set an extended user attribute of an inode
 */
static int ext_attr_user_set(struct inode *inode, const char *name,
                             void *value, size_t size, int flags)
{
	if (!IS_EXT_ATTR_USER(inode))
		return -ENOTSUP;
	if (IS_RDONLY(inode))
		return -EROFS;
	if (IS_IMMUTABLE(inode) || IS_APPEND(inode))
		return -EPERM;
	if (S_ISCHR(inode->i_mode) || S_ISBLK(inode->i_mode))
		return -EPERM;
	else {
		int error = permission(inode, MAY_WRITE);

		if (error)
			return error;
	}
	return ext_attr_set(inode, name, value, size, flags);
}

#endif  /* CONFIG_FS_USER_EXT_ATTR */

#else  /* CONFIG_FS_EXT_ATTR */

/*
 * System call given a path name (see ext_attr_path(2)).
 */
asmlinkage int
sys_ext_attr_path(const char *path, int cmd,
                  const char *name, void *value, size_t size)
{
	return -ENOSYS;
}

/*
 * System call given a file descriptor (see ext_attr_fd(2)).
 */
asmlinkage int
sys_ext_attr_fd(int fd, int cmd,
                const char *name, void *value, size_t size)
{
	return -ENOSYS;
}

#endif  /* CONFIG_FS_EXT_ATTR */
