/*
 *  linux/fs/nfs4/util.c
 *
 *  Copyright (c) 2000 The Regents of the University of Michigan.
 *  All rights reserved.
 *
 *  Kendrick Smith <kmsmith@umich.edu>
 *  Andy Adamson <andros@umich.edu>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#define __NO_VERSION__
#include <linux/config.h>
#include <linux/module.h>

#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/dcache.h>
#include <linux/nfs4/nfs4_prot.h>
#include <linux/nfs4/nfs4_debug.h>
#include <linux/nfs4/nfs4.h>

static struct {
	int	nfserr;
	int	syserr;
} nfs_errtbl[] = {
	{ NFS4_OK, 0 },
	{ NFS4ERR_PERM, -EPERM },
	{ NFS4ERR_NOENT, -ENOENT },
	{ NFS4ERR_IO, -EIO },
	{ NFS4ERR_NXIO, -ENXIO },
	{ NFS4ERR_ACCES, -EACCES },
	{ NFS4ERR_EXIST, -EEXIST },
	{ NFS4ERR_XDEV, -EXDEV },
	{ NFS4ERR_MLINK, -EMLINK },
	{ NFS4ERR_NODEV, -ENODEV },
	{ NFS4ERR_NOTDIR, -ENOTDIR },
	{ NFS4ERR_ISDIR, -EISDIR },
	{ NFS4ERR_INVAL, -EINVAL },
	{ NFS4ERR_FBIG, -EFBIG },
	{ NFS4ERR_NOSPC, -ENOSPC },
	{ NFS4ERR_ROFS, -EROFS },
	{ NFS4ERR_MLINK, -EMLINK },
	{ NFS4ERR_NAMETOOLONG, -ENAMETOOLONG },
	{ NFS4ERR_NOTEMPTY, -ENOTEMPTY },
#ifdef EDQUOT
	{ NFS4ERR_DQUOT, -EDQUOT },
#endif
	{ NFS4ERR_STALE, -ESTALE },
	{ NFS4ERR_DENIED, -EAGAIN },
	{ NFS4ERR_SYMLINK, -ELOOP },
	{ -1, -EIO }
};

int
kernerrno(nfsstat4 err)
{
	int i;
	
	/* XXX : not the optimal algorithm, but will do for now! */
	for (i = 0; nfs_errtbl[i].nfserr != -1; i++) {
		if (nfs_errtbl[i].nfserr == err)
			return nfs_errtbl[i].syserr;
	}
	nfsv4_printk(level2, "kernerrno: unrecognized nfs err: %d\n", err);
	/* if no entry exists for this error, use EIO by default */
	return -EIO;
}

int
nfserrno(int errno)
{
	int i;

	/* XXX : not the optimal algorithm, but will do for now! */
	for (i = 0; nfs_errtbl[i].nfserr != -1; i++) {
		if (nfs_errtbl[i].syserr == errno)
			return nfs_errtbl[i].nfserr;
	}
	nfsv4_printk(level2, "nfserrno: unrecongnized errno: %d\n", errno);
	/* if no entry exists for this error, use NFS4ERR_IO by default */
	return NFS4ERR_IO;
}

int
utf8cmp(utf8string *s1, utf8string *s2)
{
	return (s1->utf8string_len != s2->utf8string_len) ||
		memcmp(s1->utf8string_val, s2->utf8string_val, s1->utf8string_len);
}

int
utf8dup(utf8string *dest, utf8string *src)
{
	int status = 0;
	
	if ((dest->utf8string_len = src->utf8string_len) > 0) {
		if ((dest->utf8string_val = NFS4_ALLOC(src->utf8string_len)))
			memcpy(dest->utf8string_val, src->utf8string_val, src->utf8string_len);
		else {
			printk("nfsd: couldn't alloc in utf8dup!\n");
			dest->utf8string_len = 0;
			dest->utf8string_val = NULL;
			status = NFS4ERR_RESOURCE;
		}
	}
	else
		dest->utf8string_val = NULL;
	
	return status;
}

void
ip_to_clientaddr4(u32 ip_addr, clientaddr4 *cp)
{
	/* XXX : ??? */
	/* XXX : eliminate C-style strings from the protocol */
	cp->r_netid = NULL;
	cp->r_addr = NULL;
}

#if NFS4_DEBUG
void
hex_dump(void *ptr, int nbytes)
{
	int i;
	unsigned char *cptr = (unsigned char *) ptr;

	if (nbytes < 0) {
		printk("hex_dump: nbytes < 0!\n");
		return;
	}
	for (i = 0; i < nbytes; i++) {
		printk("%2x", *cptr++);
		if ((i % 24) == 23)
			printk("\n");
		else if ((i % 4) == 3)
			printk("   ");
		else
			printk(" ");
	}
	printk("\n");
}

void
print_rpcsec_gss_info(rpcsec_gss_info *ptr)
{
  int i;
  char *p;

  printk("NFSV4: print_rpcsec_gss_info\n");
  /*secinfo4*/
 
	printk("rpcsec_gss_info:\n");
  printk("     sec_info4 len: %d\n sec_info4 val:",
	    ptr->oid.sec_oid4_len);
  p = ptr->oid.sec_oid4_val;
	printk("     ");
  for(i=0;i<ptr->oid.sec_oid4_len;i++)
    printk("%c",*p++);
  printk("     qop: %d\n",ptr->qop);
  printk("     service: %d\n",ptr->service);
}

void
print_bm(bitmap4 *bm)
{
	int i;
	
	printk("[len %d]", bm->bitmap4_len);
	for (i = 0; i < bm->bitmap4_len; i++)
		printk(" %x", bm->bitmap4_val[i]);
}

void
print_fh(nfs_fh4 *fh)
{
	hex_dump(fh->nfs_fh4_val, fh->nfs_fh4_len);
}

void
print_fattr(fattr4 *fp)
{
	printk("bitmap: ");
	print_bm(&fp->attrmask);
	printk(", attrs: ");
	hex_dump(fp->attr_vals.attrlist4_val, fp->attr_vals.attrlist4_len);
}

void
print_lockowner(nfs_lockowner4 *lp)
{
	hex_dump(lp->owner.owner_val, lp->owner.owner_len);
}

void
print_pathname(pathname4 *p)
{
	int i;
	
	for (i = 0; i < p->pathname4_len; i++)
		printk("%s%.*s", i ? "/" : "",
		       (int)p->pathname4_val[i].utf8string_len,
		       p->pathname4_val[i].utf8string_val);
}

#define HASH_SIZE  256

static u32
ptr_hval(void *ptr)
{
	return opaque_hashval(&ptr, sizeof(void *)) % HASH_SIZE;
}

struct mem_hashent {
	int nbytes;
	char *file;
	int line;
	void *ptr;
	struct mem_hashent *next;
};
static struct mem_hashent *mem_hashtbl[HASH_SIZE];

#define HISTORY_LEN   100
struct ref_hashent {
	void *ptr;
	char *file[HISTORY_LEN];
	int line[HISTORY_LEN];
	int change[HISTORY_LEN];
	int refcount;
	struct ref_hashent *next;
};
static struct ref_hashent *dentry_hashtbl[HASH_SIZE];
static struct ref_hashent *inode_hashtbl[HASH_SIZE];

/* when a filename is entered into the history tables, it may exist in the static memory of
   a different module, which can be unloaded seperately from this one.  therefore, we must
   copy the string.  on the other hand, we want to avoid excessive copying.  we keep a hash
   table of filenames, so that if a duplicate filename is seen, we can return the old copy
   instead of making a new one. */

struct filename_hashent {
	int len;
	char *name;
	struct filename_hashent *next;
};
static struct filename_hashent *filename_hashtbl[HASH_SIZE];

static struct semaphore check_sema;

static char *
dup_filename(char *name)
{
	char *retval;
	struct filename_hashent *p;
	int len = strlen(name);
	int hashval = opaque_hashval(name, len) % HASH_SIZE;

	down(&check_sema);
	for (p = filename_hashtbl[hashval]; p; p = p->next)
		if ((p->len == len) && !memcmp(p->name, name, len)) {
			up(&check_sema);
			return p->name;
		}
	if (!(p = kmalloc(sizeof(*p), GFP_KERNEL))) {
		printk("!!! couldn't kmalloc filename_hashent !!!\n");
		return "unknown";
	}
	if (!(retval = kmalloc(len+1, GFP_KERNEL))) {
		printk("!!! couldn't kmalloc filename !!!\n");
		kfree(p);
		return "unknown";
	}
	memcpy(retval, name, len+1);
	p->len = len;
	p->name = retval;
	p->next = filename_hashtbl[hashval];
	filename_hashtbl[hashval] = p;
	up(&check_sema);
	return retval;
}

static void
adjust_history(struct ref_hashent *p, char *file, int line, int change)
{
	int i;
	for (i = HISTORY_LEN-1; i >= 1; i--) {
		p->file[i] = p->file[i-1];
		p->line[i] = p->line[i-1];
		p->change[i] = p->change[i-1];
	}
	p->file[0] = file;
	p->line[0] = line;
	p->change[0] = change;
	p->refcount += change;
}

static struct ref_hashent *
find_ref_hashent(void *ptr, struct ref_hashent **headp)
{
	struct ref_hashent *p;
	for (p = *headp; p; p = p->next)
		if (p->ptr == ptr)
			break;
	if (!p)
		if ((p = kmalloc(sizeof(*p), GFP_KERNEL))) {
			int i;
			for (i = 0; i < HISTORY_LEN; i++)
				p->change[i] = 0;
			p->refcount = 0;
			p->ptr = ptr;
			p->next = *headp;
			*headp = p;
		}
	return p;
}

static void
print_history(struct ref_hashent *p)
{
	int i;
	printk("%p", p->ptr);
	for (i = 0; (i < HISTORY_LEN) && p->change[i]; i++) {
		if (p->change[i] == 1)
			printk(" GET");
		else if (p->change[i] == -1)
			printk(" PUT");
		else
			printk(" BAD CHANGE: %d?!", p->change[i]);
		printk(" %s:%d", p->file[i], p->line[i]);
	}
}

static void
INITIALIZE(void)
{
	int i;
	init_MUTEX(&check_sema);
	for (i = 0; i < HASH_SIZE; i++) {
		mem_hashtbl[i] = NULL;
		dentry_hashtbl[i] = NULL;
		inode_hashtbl[i] = NULL;
		filename_hashtbl[i] = NULL;
	}
}

void *
__NFS4_ALLOC(int nbytes, char *file, int line)
{
	struct mem_hashent *p;
	void *ptr;
	u32 hashval;

	if (nbytes == 0) {
		printk("ZERO SIZE REQUESTED in NFS4_ALLOC (source %s:%d)!\n",
		       file, line);
		return NULL;
	}
	if (nbytes < 0) {
		printk("NEGATIVE SIZE REQUESTED in NFS4_ALLOC (source %s:%d)!\n",
		       file, line);
		return NULL;
	}
	if (nbytes > 10000) {
		printk("%d BYTES REQUESTED in NFS4_ALLOC (source %s:%d)!\n",
		       nbytes, file, line);
		return NULL;
	}
		
	file = dup_filename(file);
	if (!(p = kmalloc(sizeof(*p), GFP_KERNEL))) {
		printk("HASH KMALLOC FAILED in NFS4_ALLOC!\n");
		return NULL;
	}
	if (!(ptr = kmalloc(nbytes, GFP_KERNEL))) {
		printk("MAIN KMALLOC FAILED in NFS4_ALLOC!\n");
		printk("%s:%d, %d bytes\n", file, line, nbytes);
		return NULL;
	}
	hashval = ptr_hval(ptr);
	p->nbytes = nbytes;
	p->file = file;
	p->line = line;
	p->ptr = ptr;
	down(&check_sema);
	p->next = mem_hashtbl[hashval];
	mem_hashtbl[hashval] = p;
	up(&check_sema);
	return ptr;
}

void
__NFS4_FREE(void *ptr, char *file, int line)
{
	u32 hashval = ptr_hval(ptr);
	struct mem_hashent *p, **pp;

	file = dup_filename(file);
	down(&check_sema);
	pp = &mem_hashtbl[hashval];
	for (p = mem_hashtbl[hashval]; p; p = p->next) {
		if (p->ptr == ptr)
			break;
		pp = &p->next;
	}
	if (!p) {
		up(&check_sema);
		printk("BAD KFREE of ptr %p! (source: %s:%d)\n", ptr, file, line);
	}
	else {
		*pp = p->next;
		up(&check_sema);
		kfree(p);
		kfree(ptr);
	}
}

static void
DUP(struct dentry *dentry, char *file, int line)
{
	struct ref_hashent *p;

#if 0
	dget(dentry);
#endif
	file = dup_filename(file);
	down(&check_sema);
	if (!(p = find_ref_hashent(dentry, &dentry_hashtbl[ptr_hval(dentry)]))) {
		printk("FAILED TO ALLOC REF HASHENT!\n");
		return;
	}
	adjust_history(p, file, line, 1);
#if 0   /* XXX: bakeoff */
	if (p->refcount > 5) {
		printk("BAD DGET? ");
		print_history(p);
	}
#endif
	up(&check_sema);
}

static void
DDOWN(struct dentry *dentry, char *file, int line)
{
	struct ref_hashent *p;

#if 0
	dget(dentry);
#endif
	file = dup_filename(file);
	down(&check_sema);
	if (!(p = find_ref_hashent(dentry, &dentry_hashtbl[ptr_hval(dentry)]))) {
		printk("FAILED TO ALLOC REF HASHENT!\n");
		return;
	}
	adjust_history(p, file, line, -1);
	if (p->refcount < 0) {
		printk("BAD DPUT! ");
		print_history(p);
	}
	up(&check_sema);
}

static void
IUP(struct inode *inode, char *file, int line)
{
	struct ref_hashent *p;

#if 0
	iget(inode->i_sb, inode->i_ino);
#endif
	file = dup_filename(file);
	down(&check_sema);
	if (!(p = find_ref_hashent(inode, &inode_hashtbl[ptr_hval(inode)]))) {
		printk("FAILED TO ALLOC REF HASHENT!\n");
		return;
	}
	adjust_history(p, file, line, 1);
	if (p->refcount > 5) {
		printk("BAD IGET? ");
		print_history(p);
	}
	up(&check_sema);
}

static void
IDOWN(struct inode *inode, char *file, int line)
{
	struct ref_hashent *p;

#if 0
	iget(inode->i_sb, inode->i_ino);
#endif
	file = dup_filename(file);
	down(&check_sema);
	if (!(p = find_ref_hashent(inode, &inode_hashtbl[ptr_hval(inode)]))) {
		printk("FAILED TO ALLOC REF HASHENT!\n");
		return;
	}
	adjust_history(p, file, line, -1);
	if (p->refcount < 0) {
		printk("BAD IPUT! ");
		print_history(p);
	}
	up(&check_sema);
}

struct inode *__IGET(struct super_block *sb, unsigned long ino, char *file, int line)
{
	struct inode *i;
	nfsv4_printk(check, "          iget: %s:%d\n", file, line);
	i = iget(sb, ino);
	if (i)
		IUP(i, file, line);
	return i;
}

void __IPUT(struct inode *inode, char *file, int line)
{
	IDOWN(inode, file, line);
	iput(inode);
}

struct inode * __IGRAB(struct inode *inode, char *file, int line)
{
	struct inode *i;
	nfsv4_printk(check, "          igrab: %s:%d\n", file, line);
	i = igrab(inode);
	if (i == inode)
		IUP(i, file, line);
	else
		printk("IGRAB RETURN MISMATCH %s:%d?!\n", file, line);
	return i;
}

struct dentry *__DGET(struct dentry *dentry, char *file, int line)
{
	struct dentry *d;
	nfsv4_printk(check, "          dget: %s:%d\n", file, line);
	d = dget(dentry);
	if (dentry) {
		if (d == dentry)
			DUP(d, file, line);
		else
			printk("DGET RETURN MISMATCH %s:%d?!\n", file, line);
	}
	return d;
}

struct dentry *__DGET_LOCKED(struct dentry *dentry, char *file, int line)
{
	struct dentry *d;
	nfsv4_printk(check, "          dget_locked: %s:%d\n", file, line);
	d = dget_locked(dentry);
	if (d == dentry)
		DUP(d, file, line);
	else
		printk("DGET_LOCKED RETURN MISMATCH %s:%d?!\n", file, line);
	return d;
}

struct dentry *__D_ALLOC(struct dentry *dentry, const struct qstr *str, char *file, int line)
{
	struct dentry *d;
	nfsv4_printk(check, "          d_alloc: %s:%d\n", file, line);
	d = d_alloc(dentry, str);
	if (d == dentry)
		DUP(d, file, line);
	else
		printk("D_ALLOC RETURN MISMATCH %s:%d?!\n", file, line);
	return d;
}

struct dentry *__D_ALLOC_ROOT(struct inode *inode, char *file, int line)
{
	struct dentry *d;
	nfsv4_printk(check, "          d_alloc_root: %s:%d\n", file, line);
	d = d_alloc_root(inode);
	if (d)
		DUP(d, file, line);
	return d;
}

void __DPUT(struct dentry *dentry, char *file, int line)
{
	if (dentry) {
		dput(dentry);
		DDOWN(dentry, file, line);
	}
}

struct dentry *__D_LOOKUP(struct dentry *dentry, struct qstr *str, char *file, int line)
{
	struct dentry *d;
	nfsv4_printk(check, "          d_lookup: %s:%d\n", file, line);
	d = d_lookup(dentry, str);
	if (d)
		DUP(d, file, line);
	return d;
}

struct dentry *__LOOKUP_ONE(const char *str, struct dentry *dentry, char *file, int line)
{
	struct dentry *d;
	nfsv4_printk(check, "          lookup_one: %s:%d\n", file, line);
	d = lookup_one(str, dentry);
	if (d)
		DUP(d, file, line);
	return d;
}

int __FOLLOW_DOWN(struct vfsmount **mntp, struct dentry **dentryp, char *file, int line)
{
	struct dentry *d = *dentryp;
	int res;
	nfsv4_printk(check, "          follow_one: %s:%d\n", file, line);
	if ((res = follow_down(mntp, dentryp))) {
		DDOWN(d, file, line);
		DUP(*dentryp, file, line);
	}
	return res;
}

static void
COMPLAIN(void)
{
	int i;
	struct mem_hashent *p, *P;
	struct ref_hashent *q, *Q;
	struct filename_hashent *r, *R;
	down(&check_sema);
	printk("Checking for memory leaks...\n");
	for (i = 0; i < HASH_SIZE; i++) {
		for (p = mem_hashtbl[i]; p; p = P) {
			printk("MEMORY LEAK %d bytes at %p (source: %s:%d)\n",
			       p->nbytes, p->ptr, p->file, p->line);
			kfree(p->ptr);
			P = p->next;
			kfree(p);
		}
	}

	printk("Checking dentry refcounts...\n");
	for (i = 0; i < HASH_SIZE; i++) {
		for (q = dentry_hashtbl[i]; q; q = Q) {
#if 0      /* XXX: bakeoff */
			if (q->refcount > 0) {
				printk("DCOUNT LEAK ");
				print_history(q);
			}
#endif
			Q = q->next;
			kfree(q);
		}
	}

	printk("Checking inode refcounts...\n");
	for (i = 0; i < HASH_SIZE; i++) {
		for (q = inode_hashtbl[i]; q; q = Q) {
			if (q->refcount > 0) {
				printk("ICOUNT LEAK ");
				print_history(q);
			}
			Q = q->next;
			kfree(q);
		}
	}

	for (i = 0; i < HASH_SIZE; i++)
		for (r = filename_hashtbl[i]; r; r = R) {
			R = r->next;
			kfree(r->name);
			kfree(r);
		}

	up(&check_sema);
}

#endif  /* NFS4_DEBUG */

/*
 *  Kernel module setup
 */

#ifdef MODULE

EXPORT_NO_SYMBOLS;
MODULE_AUTHOR("UM-CITI");

int
init_module(void)
{
#if NFS4_DEBUG
	INITIALIZE();
	rpc_register_sysctl();
#endif
	gss_init();
	printk(KERN_INFO "Installing nfs4\n");
	return 0;
}

void
cleanup_module(void)
{
	gss_shutdown();
#if NFS4_DEBUG
	rpc_unregister_sysctl();
	COMPLAIN();
#endif
	printk(KERN_INFO "Uninstalling nfs4\n");
}

#endif  /* MODULE */
