/*
 *  linux/fs/nfs4/xdr_clnt.c
 *
 *  Copyright (c) 2000 The Regents of the University of Michigan.
 *  All rights reserved.
 *
 *  Kendrick Smith <kmsmith@umich.edu>
 *  Andy Adamson <andros@umich.edu>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
*/

#include <linux/rpc.h>

static bool_t xdrclnt_getlong(XDR *xdrs, long *lp);
static bool_t xdrclnt_putlong(XDR *xdrs, long *lp);
static bool_t xdrclnt_getbytes(XDR *xdrs, caddr_t buf, u_int len);
static bool_t xdrclnt_putbytes(XDR *xdrs, caddr_t buf, u_int len);
static u_int xdrclnt_getpostn(XDR *xdrs);
static bool_t xdrclnt_setpostn(XDR *xdrs, u_int len);
static int32_t *xdrclnt_inline(XDR *xdrs, u_int len);
static void xdrclnt_destroy(XDR *xdrs);

static struct xdr_ops xdrclnt_ops = {
	xdrclnt_getlong,
	xdrclnt_putlong,
	xdrclnt_getbytes,
	xdrclnt_putbytes,
	xdrclnt_getpostn,
	xdrclnt_setpostn,
	xdrclnt_inline,
	xdrclnt_destroy
};

struct xdrclnt_private {
	unsigned int *rpc_iov_len;
	__kernel_size_t *iov_len;
};

int
xdrclnt_create(XDR *xdrs, struct rpc_rqst *req, u32 *p, enum xdr_op op)
{
	struct xdrclnt_private *priv;
	u32 len, buflen;
	
	xdrs->x_op = op;
	xdrs->x_ops = &xdrclnt_ops;
	
	switch (op) {
	case XDR_ENCODE:
		if (!(priv = kmalloc(sizeof(*xdrs->x_private), GFP_KERNEL))) {
			printk("kmalloc failed in xdrclnt_create()!\n");
			return FALSE;
		}
		if (req->rq_snr != 1) {
			printk("multiple iovecs in xdrclnt_create()!\n");
			return FALSE;
		}
		
		len = (char *)p - (char *)req->rq_svec->iov_base;
		buflen = req->rq_svec->iov_len;
		if ((len < 0) || (len >= buflen)) {
			printk("bad lengths in xdrclnt_create()!\n");
			return FALSE;
		}
		req->rq_slen = len;
		req->rq_svec->iov_len = len;
		priv->rpc_iov_len = &req->rq_slen;
		priv->iov_len = &req->rq_svec->iov_len;

		xdrs->x_private = (caddr_t) priv;
		xdrs->x_base = (caddr_t) p;
		xdrs->x_handy = buflen - len;
		break;
	case XDR_DECODE:
		/* We need to determine how much data was received, to test for buffer
		   overflows before they happen.  Unfortunately, both lengths handed down
		   from the rpc layer, req->rq_rlen and req->rq_rvec[0].iov_len, seem to
		   be garbage!  The only reliable way to get the length appears to be
		   venturing into the "private" part of the request structure, and using
		   req->rq_task->tk_status.  Since this is only an ad hoc procedure which
		   seems to work, not something guaranteed to work by the RPC layer, we
		   subject the length to some consistency checks. */
		
		if (req->rq_rnr != 1) {
			printk("xdrclnt_create: multiple iovecs in xdrclnt_create()!\n");
			return FALSE;
		}
		if (!req->rq_task) {
			printk("xdrclnt_create: req->rq_task is NULL!\n");
			return FALSE;
		}
		
		len = (char *)p - (char *)req->rq_rvec->iov_base;
		buflen = req->rq_task->tk_status;
		if ((len < 0) || (len >= buflen)) {
			printk("bad lengths in xdrclnt_create()!\n");
			return FALSE;
		}

		xdrs->x_private = NULL;
		xdrs->x_base = (caddr_t) p;
		xdrs->x_handy = buflen - len;
		break;
	case XDR_FREE:
		xdrs->x_private = NULL;
		xdrs->x_base = NULL;
		xdrs->x_handy = 0;
		break;
	}

	return TRUE;
}

static bool_t
xdrclnt_getlong(XDR *xdrs, long *lp)
{
	if ((xdrs->x_handy -= 4) < 0) {
		printk("svc_getlong(): buffer underflow!\n");
		return FALSE;
	}
	*lp = ntohl(*(u32 *)xdrs->x_base);
	xdrs->x_base += 4;
	return TRUE;
}

static bool_t
xdrclnt_putlong(XDR *xdrs, long *lp)
{
	struct xdrclnt_private *p = (struct xdrclnt_private *) xdrs->x_private;

	if ((xdrs->x_handy -= 4) < 0) {
		printk("xdrclnt_putlong(): buffer overflow!\n");
		return FALSE;
	}
	*(u32 *)xdrs->x_base = htonl(*lp);
	xdrs->x_base += 4;
	*p->rpc_iov_len += 4;
	*p->iov_len += 4;
	return TRUE;
}

static bool_t
xdrclnt_getbytes(XDR *xdrs, caddr_t buf, u_int len)
{
	if ((xdrs->x_handy -= len) < 0) {
		printk("xdrclnt_getbytes(): buffer underflow!\n");
		return FALSE;
	}
	memcpy(buf, xdrs->x_base, len);
	xdrs->x_base += len;
	return TRUE;
}

static bool_t
xdrclnt_putbytes(XDR *xdrs, caddr_t buf, u_int len)
{
	struct xdrclnt_private *p = (struct xdrclnt_private *) xdrs->x_private;

	if ((xdrs->x_handy -= len) < 0) {
		printk("xdrclnt_putbytes(): buffer overflow!\n");
		return FALSE;
	}
	memcpy(xdrs->x_base, buf, len);
	xdrs->x_base += len;
	*p->rpc_iov_len += len;
	*p->iov_len += len;
	return TRUE;
}

static u_int
xdrclnt_getpostn(XDR *xdrs)
{
	/* never gets used, so left unimplemented for now */
	return 0;
}

static bool_t
xdrclnt_setpostn(XDR *xdrs, u_int len)
{
	/* never gets used, so left unimplemented for now */
	return FALSE;
}

static int32_t *
xdrclnt_inline(XDR *xdrs, u_int len)
{
	/* never gets used, so left unimplemented for now */
	return NULL;
}

static void
xdrclnt_destroy(XDR *xdrs)
{
	switch (xdrs->x_op) {
	case XDR_ENCODE:
		kfree((struct xdrclnt_private *) xdrs->x_private);
		break;
	case XDR_DECODE:
	case XDR_FREE:
		break;
	}
}
