/*
 *  linux/fs/nfs4/xdr_svc.c
 *
 *  Copyright (c) 2000 The Regents of the University of Michigan.
 *  All rights reserved.
 *
 *  Kendrick Smith <kmsmith@umich.edu>
 *  Andy Adamson <andros@umich.edu>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
*/

#include <linux/rpc.h>

static bool_t xdrsvc_getlong(XDR *xdrs, long *lp);
static bool_t xdrsvc_putlong(XDR *xdrs, long *lp);
static bool_t xdrsvc_getbytes(XDR *xdrs, caddr_t buf, u_int len);
static bool_t xdrsvc_putbytes(XDR *xdrs, caddr_t buf, u_int len);
static u_int xdrsvc_getpostn(XDR *xdrs);
static bool_t xdrsvc_setpostn(XDR *xdrs, u_int len);
static int32_t *xdrsvc_inline(XDR *xdrs, u_int len);
static void xdrsvc_destroy(XDR *xdrs);

static struct xdr_ops xdrsvc_ops = {
	xdrsvc_getlong,
	xdrsvc_putlong,
	xdrsvc_getbytes,
	xdrsvc_putbytes,
	xdrsvc_getpostn,
	xdrsvc_setpostn,
	xdrsvc_inline,
	xdrsvc_destroy
};

void
xdrsvc_create(XDR *xdrs, struct svc_rqst *rqstp, enum xdr_op op)
{
	struct svc_buf *sbuf = NULL;
	
	xdrs->x_op = op;
	xdrs->x_ops = &xdrsvc_ops;
	
	switch (op) {
	case XDR_ENCODE:
		sbuf = &rqstp->rq_resbuf;
		break;
	case XDR_DECODE:
		sbuf = &rqstp->rq_argbuf;
		break;
	case XDR_FREE:
	}
	
	xdrs->x_private = (caddr_t) sbuf;
	if (sbuf)
		xdrs->x_handy = (sbuf->buflen << 2) - ((char *)sbuf->buf - (char *)sbuf->base);
	else
		xdrs->x_handy = 0;
}

static bool_t
xdrsvc_getlong(XDR *xdrs, long *lp)
{
	struct svc_buf *sbuf = (struct svc_buf *) xdrs->x_private;

	if ((xdrs->x_handy -= 4) < 0) {
		printk("svc_getlong(): buffer underflow!\n");
		return FALSE;
	}
	
	*lp = ntohl(*sbuf->buf++);
	return TRUE;
}

static bool_t
xdrsvc_putlong(XDR *xdrs, long *lp)
{
	struct svc_buf *sbuf = (struct svc_buf *) xdrs->x_private;

	if ((xdrs->x_handy -= 4) < 0) {
		printk("svc_putlong(): buffer overflow!\n");
		return FALSE;
	}
	*sbuf->buf++ = htonl(*lp);
	return TRUE;
}

static bool_t
xdrsvc_getbytes(XDR *xdrs, caddr_t buf, u_int len)
{
	struct svc_buf *sbuf = (struct svc_buf *) xdrs->x_private;

	if ((xdrs->x_handy -= len) < 0) {
		printk("svc_getbytes(): buffer underflow!\n");
		return FALSE;
	}	
	memcpy(buf, sbuf->buf, len);
	sbuf->buf = (u32 *) ((char *)sbuf->buf + len);
	return TRUE;
}

static bool_t
xdrsvc_putbytes(XDR *xdrs, caddr_t buf, u_int len)
{
	struct svc_buf *sbuf = (struct svc_buf *) xdrs->x_private;

	if ((xdrs->x_handy -= len) < 0) {
		printk("svc_putlong(): buffer overflow!\n");
		return FALSE;
	}
	memcpy(sbuf->buf, buf, len);
	sbuf->buf = (u32 *) ((char *)sbuf->buf + len);
	return TRUE;
}

static u_int
xdrsvc_getpostn(XDR *xdrs)
{
	struct svc_buf *sbuf = (struct svc_buf *) xdrs->x_private;

	return (char *)sbuf->buf - (char *)sbuf->base;
}

static bool_t
xdrsvc_setpostn(XDR *xdrs, u_int pos)
{
	struct svc_buf *sbuf = (struct svc_buf *) xdrs->x_private;
	char *end, *buf;

	end = (char *)sbuf->buf + xdrs->x_handy;
	buf = (char *)sbuf->base + pos;
	if (buf > end) {
		printk("svc_setpostn(): position set past end of buffer!\n");
		return FALSE;
	}
	
	sbuf->buf = (u32 *)buf;
	xdrs->x_handy = end - buf;
	return TRUE;
}

static int32_t *
xdrsvc_inline(XDR *xdrs, u_int len)
{
	/* never gets used, so left unimplemented for now */
	return NULL;
}

static void
xdrsvc_destroy(XDR *xdrs)
{
	return;
}
