/*
 * linux/fs/nfs4fs/acl.c
 *
 * Jake Moilanen <jmoilane@umich.edu>
 * Kendrick Smith <kmsmith@umich.edu>
 * Andy Adamson <andros@umich.edu>
 * Copyright (C) 2001 The Regents of the University of Michigan
 *
 */

#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/mm.h>
#include <linux/pagemap.h>
#include <linux/fs.h>
#include <linux/acl_ea.h>
#include <linux/ext_attr.h>
#include <asm/uaccess.h>

#include <linux/nfs4/nfs4_prot.h>
#include <linux/nfs4/nfs4_debug.h>
#include <linux/nfs4/nfs4.h>
#include <linux/nfs4/nfs4fs.h>

#ifdef NFS4ACL

int nfs4_posix_acl_permission(struct inode *inode, int mask);
posix_acl_t *nfs4_acl_get_inode(struct inode *inode, int type);
int nfs4_acl_set_inode(struct inode *inode, int type, posix_acl_t *acl);
int nfs4_get_ext_attr(struct inode *inode, const char *name, void *buffer, 
		                             size_t buffer_size, int flags);
int nfs4_set_ext_attr(struct inode *inode, const char *name, void *buffer, 
		                             size_t buffer_size, int flags);
static void * acl_to_ea(const posix_acl_t *acl, size_t *size);


int
nfs4_posix_acl_permission(struct inode *inode, int mask)
{       
       return posix_acl_permission(inode, mask);
} 

posix_acl_t *
nfs4_acl_get_inode(struct inode *inode, int type)
{
	printk("In nfs4_acl_get_inode\n");

	return acl_ea_get_inode(inode, type);
}

int 
nfs4_acl_set_inode(struct inode *inode, int type, posix_acl_t *acl)
{
	printk("In nfs4_acl_set_inode\n");

	return acl_ea_set_inode(inode, type, acl);
}

static u32 acl_bmval[] = { 0x00000000 };
bitmap4 nfs4_acl_bitmap = { 1, acl_bmval };
/* To do
 * check if name is NULL,
 * check if buffer_size = 0,
 * check if not an user land pointer
 * make sure acl entries are in order
 * do a link to acl_to_ea and not have it in this file
 * fix case where there isn't any acl's
 * change size of fattr to be correct (different file)
 */
int
nfs4_get_ext_attr(struct inode *inode, const char *name, void *buffer, 
		                             size_t buffer_size, int flags)
{
	struct compound4args    cargs;
	struct compound4res     cres;
	struct nfs4fs_ctx 	ctx;
	struct nfs4fs_putfh     putfh;
	struct nfs4fs_getattr   getattr;
	struct nfs4fs_getfh     getfh;
	struct nfs4fs_fattr 	fattr;
	posix_acl_t		*acl;
	void 			*p;
	int 			status;
	

	
	printk("In nfs4_get_ext_attr\n");
	printk("name=%s, buffer=%p, buffer_size=%d, flags=%d\n",
			                  name, buffer, buffer_size, flags);
	if (list_empty(&inode->i_dentry)){
		printk("nfs4_get_ext_attr: inode has empty dentry\n");
		goto out_res;             // XXX get an error code
	}

	cargs.sb = inode->i_sb;
	cargs.tag = "nfs4_get_ext_attr()";
	ctx.pt_sb = inode->i_sb;
	ctx.pt_dentry = list_entry(inode->i_dentry.next, struct dentry, d_alias);
	putfh.dentry = list_entry(inode->i_dentry.next, struct dentry, d_alias);
	getattr.flags = 0;
	*nfs4_acl_bitmap.bitmap4_val |= (1 << FATTR4_ACL);
	*nfs4_acl_bitmap.bitmap4_val |= (1 << FATTR4_TYPE);
	*nfs4_acl_bitmap.bitmap4_val |= (1 << FATTR4_CHANGE);
	*nfs4_acl_bitmap.bitmap4_val |= (1 << FATTR4_SIZE);
	*nfs4_acl_bitmap.bitmap4_val |= (1 << FATTR4_FSID);
	*nfs4_acl_bitmap.bitmap4_val |= (1 << FATTR4_FILEID);
	printk("nfs4_acl_bitmap.bitmap4_val = %d\n", nfs4_acl_bitmap.bitmap4_val);
	getattr.bm = &nfs4_acl_bitmap;
	getfh.nonfatal = 0;

	if (nfs4_setup_compound(&cargs))
		goto out_dput;
        if (nfs4_setup_putfh(&cargs, &putfh))
                goto out_args;
        if (nfs4_setup_getattr(&cargs, &getattr))
                goto out_args;

        if (nfs4_call_compound(&cargs, &cres, &ctx))
                goto out_args;
	if (nfs4_handle_putfh(&cres, &putfh))
                goto out_res;
        if (nfs4_handle_getattr(&cres, &getattr))
                goto out_res;

	printk("nfs4_get_ext_attr: just got back from nfs_handle_getattr\n");
	fattr = getattr.fattr;

	/* check if there are acls to begin with */
	if (fattr.fa_nacl){
		printk("nfs4_get_ext_attr: At least 1 acl\n");
		acl = nfsv4_acl_to_posix_acl(fattr.fa_acl, fattr.fa_nacl);
	} else {
		printk("nfs4_get_ext_attr: 0 acl's\n");
		acl = 0;
		return 0;
	}

	p = acl_to_ea(acl, &buffer_size);
	status = copy_to_user(buffer, p, buffer_size);
	if (status)
		goto out_res;

	printk("Success???\n");
out_dput:
	printk("out_dput\n");
out_args: 
	printk("out_args\n");
out_res:
	printk("out_res\n");
	
	return buffer_size;
}

int 
nfs4_set_ext_attr(struct inode *inode, const char *name, void *buffer, 
		                             size_t buffer_size, int flags)
{
	printk("In nfs4_set_ext_attr\n");
	return 0;
}

/*
 * Convert from in-memory to extended attribute representation.
 */
static void *
acl_to_ea(
        const posix_acl_t *acl,
        size_t *size)
{
        acl_ea_header *ext_acl;
        char *e;
        int n;

        *size = acl_ea_size(acl->a_count);
        ext_acl = (acl_ea_header *)kmalloc(*size, GFP_KERNEL);
        if (!ext_acl)
                return ERR_PTR(-ENOMEM);
        ext_acl->a_version = cpu_to_le32(ACL_EA_VERSION);
        e = (char *)ext_acl + sizeof(acl_ea_header);
        for (n=0; n < acl->a_count; n++) {
                acl_ea_entry *entry = (acl_ea_entry *)e;
                entry->e_tag  = cpu_to_le16(acl->a_entries[n].e_tag);
                entry->e_perm = cpu_to_le16(acl->a_entries[n].e_perm);
                switch(acl->a_entries[n].e_tag) {
                        case ACL_USER:
                        case ACL_GROUP:
                                entry->e_id =
                                        cpu_to_le32(acl->a_entries[n].e_id);
                                e += sizeof(acl_ea_entry);
                                break;
                        default:
                                e += sizeof(acl_ea_entry_short);
                                break;
                }
        }
        return (char *)ext_acl;
}
#endif
