/*
 * fs/nfs4fs/callback.c
 *
 * (C) 2001 Kendrick Smith <kmsmith@umich.edu>
 */

#define __NO_VERSION__
#include <linux/config.h>
#include <linux/module.h>

#include <linux/sched.h>
#include <linux/smp.h>
#include <linux/smp_lock.h>
#include <linux/fs.h>
#include <linux/pagemap.h>
#include <linux/sunrpc/stats.h>
#include <linux/sunrpc/svc.h>
#include <linux/sunrpc/svcsock.h>

#include <linux/nfs4/nfs4_prot.h>
#include <linux/nfs4/nfs4_debug.h>
#include <linux/nfs4/nfs4.h>
#include <linux/nfs4/nfs4fs.h>

#define CB_BASE_PROGRAM		0x0400beef
#define CB_VERSION		1
#define CB_BASE_PORT		31337
#define CB_MAX			16
#define CB_BUFSIZE		1024  /* XXX: total overkill for now */
#define CB_XDRSIZE		1024  /* XXX: total overkill for now */

#define ALLOWED_SIGS    (sigmask(SIGKILL))
#define SHUTDOWN_SIGS   (sigmask(SIGKILL) | sigmask(SIGINT) | sigmask(SIGQUIT))

static u32			current_program = CB_BASE_PROGRAM;

static struct {
	struct super_block	*sb;
	struct svc_serv 	*serv;
	struct svc_program	program;
	struct svc_stat		stat;
	struct semaphore	start;
	wait_queue_head_t	wait;
	pid_t			pid;
	unsigned int		reserved : 1;
} callback_vector[CB_MAX];

static spinlock_t 		callback_lock = SPIN_LOCK_UNLOCKED;
static void			cb_main(struct svc_rqst *);

static struct svc_version *cb_version[] = {
	NULL,
	&nfs4_cb_version1,
};

int
nfs4_create_callback(struct super_block *sb)
{
	struct svc_program *program;
	struct svc_stat *stat;
	struct svc_serv *serv;
	unsigned short port;
	int i;
	int status;

	NFS4_CALLBACK_PORT(sb) = 0;

	/*
	 * Try to reserve a callback slot for the callback on this superblock.
	 */
	spin_lock(&callback_lock);
	for (i = 0; i < CB_MAX; i++)
		if (!callback_vector[i].reserved) {
			if (callback_vector[i].sb || callback_vector[i].serv)
				BUG();
			callback_vector[i].reserved = 1;
			break;
		}
	spin_unlock(&callback_lock);
	
	status = -EBUSY;
	if (i == CB_MAX) {
		printk(KERN_WARNING "nfs4: couldn't reserve a callback slot\n");
		goto out;
	}
	port = CB_BASE_PORT + i;

	/*
	 * Try to create an RPC service for the callback.
	 */
	program = &callback_vector[i].program;
	stat = &callback_vector[i].stat;
	program->pg_prog = current_program++;
	program->pg_lovers = CB_VERSION;
	program->pg_hivers = CB_VERSION;
	program->pg_nvers = sizeof(cb_version) / sizeof(cb_version[0]);
	program->pg_vers = cb_version;
	program->pg_name = "nfs4_cb";
	program->pg_stats = stat;
	memset(stat, 0, sizeof(*stat));
	stat->program = program;

	status = -ENOMEM;
	if (!(serv = svc_create(program, CB_BUFSIZE, CB_XDRSIZE))) {
		printk(KERN_WARNING "nfs4: couldn't create callback service\n");
		goto out_slot;
	}
	if ((status = svc_makesock(serv, IPPROTO_UDP, port))) {
		printk(KERN_WARNING "nfs4: couldn't create callback socket\n");
		goto out_serv;
	}

	/*
	 * Fork the callback thread, and wait for it to start.
	 */
	if ((status = svc_create_thread(cb_main, serv))) {
		printk(KERN_WARNING "nfs4: couldn't create callback thread\n");
		goto out_serv;
	}
	down(&callback_vector[i].start);

	if (!callback_vector[i].pid)
		BUG();
	
	/*
	 * Success!  Update callback table and exit..
	 */
	spin_lock(&callback_lock);
	callback_vector[i].sb = sb;
	callback_vector[i].serv = serv;
	spin_unlock(&callback_lock);

	NFS4_CALLBACK_PORT(sb) = CB_BASE_PORT + i;
	status = 0;

out:
	return status;
out_serv:
	svc_destroy(serv);
out_slot:
	spin_lock(&callback_lock);
	callback_vector[i].reserved = 0;
	spin_unlock(&callback_lock);
	goto out;
}

void
nfs4_get_callback(struct super_block *sb, u32 *progp, char *netid, char *uaddr)
{
	u32 prog = 0;
	unsigned short port;
	int i;
	
	spin_lock(&callback_lock);
	port = NFS4_CALLBACK_PORT(sb);
	if (port >= CB_BASE_PORT && port < CB_BASE_PORT + CB_MAX) {
		i = port - CB_BASE_PORT;
		if (callback_vector[i].sb != sb || !callback_vector[i].serv)
			BUG();
		prog = callback_vector[i].program.pg_prog;
	}
	spin_unlock(&callback_lock);

	*progp = prog;
	if (prog) {
		memcpy(netid, "udp", 4);
		sprintf(uaddr, "%s.%d.%d", NFS4_IP_ADDR(sb), port >> 8, port & 255);
	}
	else {
		*netid = 0;
		*uaddr = 0;
	}
}

void
nfs4_destroy_callback(struct super_block *sb)
{
	struct svc_serv *serv;
	unsigned short port;
	int i;

	port = NFS4_CALLBACK_PORT(sb);
	if (!port)
		return;
	if (port < CB_BASE_PORT || port >= CB_BASE_PORT + CB_MAX)
		BUG();
	i = port - CB_BASE_PORT;

	spin_lock(&callback_lock);
	if ((!callback_vector[i].reserved) ||
	    (!callback_vector[i].serv) ||
	    (callback_vector[i].sb != sb))
		BUG();
	serv = callback_vector[i].serv;
	callback_vector[i].serv = NULL;
	callback_vector[i].sb = NULL;
	spin_unlock(&callback_lock);

	kill_proc(callback_vector[i].pid, SIGKILL, 1);
	current->sigpending = 0;
	interruptible_sleep_on_timeout(&callback_vector[i].wait, 5 * HZ);
	
	spin_lock(&callback_lock);
	if (!callback_vector[i].pid)
		callback_vector[i].reserved = 0;
	else
		printk(KERN_WARNING "nfs4: warning: callback thread failed to exit!\n");
	spin_unlock(&callback_lock);
	
	svc_destroy(serv);
}

int
nfs4_callback_sb(struct svc_rqst *rqstp, struct super_block **sbp)
{
	struct svc_serv *serv = rqstp->rq_server;
	u32 prog = rqstp->rq_prog;
	unsigned short port = rqstp->rq_addr.sin_port;
	int i;
	int status;

	if (port < CB_BASE_PORT || port >= CB_BASE_PORT + CB_MAX) {
		printk("CALLBACK: out-of-range port?\n");
		return NFS4ERR_INVAL;
	}
	i = port - CB_BASE_PORT;

	status = NFS4ERR_INVAL;
	spin_lock(&callback_lock);
	if ((callback_vector[i].reserved) &&
	    (callback_vector[i].program.pg_prog == prog) &&
	    (callback_vector[i].serv == serv)) {
		*sbp = callback_vector[i].sb;
		status = NFS4_OK;
	}
	spin_unlock(&callback_lock);

	return status;
}

/*
 * This is the callback kernel thread.
 */
static void
cb_main(struct svc_rqst *rqstp)
{
	struct svc_serv *serv = rqstp->rq_server;
	u32 prog = serv->sv_program->pg_prog;
	unsigned int signo;
	int status;
	int i;
	
	MOD_INC_USE_COUNT;
	lock_kernel();
	
	spin_lock(&callback_lock);
	for (i = 0; i < CB_MAX; i++) {
		if (callback_vector[i].program.pg_prog == prog)
			break;
	}
	spin_unlock(&callback_lock);

	if (i == CB_MAX)
		BUG();
	callback_vector[i].pid = current->pid;
	up(&callback_vector[i].start);

	daemonize();
	sprintf(current->comm, "nfs4_cb");

	for (;;) {
		spin_lock_irq(&current->sigmask_lock);
		siginitsetinv(&current->blocked, SHUTDOWN_SIGS);
		recalc_sigpending(current);
		spin_unlock_irq(&current->sigmask_lock);

		do {
			status = svc_recv(serv, rqstp, MAX_SCHEDULE_TIMEOUT);
		} while (status == -EAGAIN);
		
		if (status < 0)
			break;

		/*
		 * Process requests with signals blocked.
		 */
		spin_lock_irq(&current->sigmask_lock);
		siginitsetinv(&current->blocked, ALLOWED_SIGS);
		recalc_sigpending(current);
		spin_unlock_irq(&current->sigmask_lock);
		svc_process(serv, rqstp);
	}

	if (status != -EINTR)
		printk(KERN_WARNING "nfs4: callback thread terminating on error %d\n", status);
	else {
		for (signo = 1; signo <= _NSIG; signo++)
			if (sigismember(&current->pending.signal, signo) &&
			    !sigismember(&current->blocked, signo))
				break;
		printk(KERN_WARNING "nfs4: callback thread terminating on signal %d\n", signo);
	}

	callback_vector[i].pid = 0;
	wake_up(&callback_vector[i].wait);
	
	svc_exit_thread(rqstp);
	MOD_DEC_USE_COUNT;
}

/*
 * 
 */
#define DELEG_HASH_BITS			8
#define DELEG_HASH_SIZE			(1 << DELEG_HASH_BITS)
#define DELEG_HASH_MASK			(DELEG_HASH_SIZE - 1)

static struct list_head			delegation_hashtbl[DELEG_HASH_SIZE];

#define delegation_hashval(sb, stateid)					\
(( opaque_hashval((sb), sizeof(struct super_block *)) +			\
   opaque_hashval((stateid), sizeof(stateid4)) ) & DELEG_HASH_MASK)

void
nfs4_accept_delegation(struct inode *inode, u32 type, stateid4 new_stateid)
{
	struct super_block *sb = inode->i_sb;
	unsigned int hashval;
	struct nfs4fs_delegation *dp, *new_dp;

	printk("AAA accept_delegation: starting %p\n", inode);
	
	if (type != OPEN_DELEGATE_READ && type != OPEN_DELEGATE_WRITE)
		BUG();
	
	/*
	 * Note: We do _not_ claim a reference to the inode in dp->dl_inode!
	 * Instead, we return the delegation in the inode's ->clear_inode()
	 * method.  This has the advantage that extra logic for returning
	 * infrequently-used delegations is not needed: it is encapsulated
	 * in the logic for purging infrequently-used inodes.
	 *
	 * XXX: I believe that it is impossible for this routine to race
	 * against clear_inode(), because the caller has to hold a reference
	 * to the inode.  Can someone confirm this assertion?
	 */
	if (!(new_dp = NFS4_ALLOC(sizeof(*new_dp)))) {
		nfsv4_printk(level1, "nfs4_accept_delegation: out of memory!\n");
		goto do_delegreturn;
	}
	new_dp->dl_sb = sb;
	new_dp->dl_inode = inode;
	new_dp->dl_type = type;
	memcpy(new_dp->dl_stateid, new_stateid, sizeof(stateid4));

	hashval = delegation_hashval(sb, new_stateid);

	/*
	 * We refuse to hold multiple delegations for the same inode.
	 * If a delegation is already held, we either return it or return
	 * the current one, depending on which has the stronger delegation
	 * type.  Note that if we sleep, we have to drop the spinlock and
	 * retry the check on the inode's delegation.
	 */
	for (;;) {
		spin_lock(&nfs4_wreq_lock);
		dp = NFS4_DELEGATION(inode);
		if (!dp) {
			list_add(&new_dp->dl_hash, &delegation_hashtbl[hashval]);
			NFS4_DELEGATION(inode) = new_dp;
			spin_unlock(&nfs4_wreq_lock);
			printk("AAA accept_delegation: success %p\n", inode);
			return;
		}
		printk("AAA accept_delegation: suspicious code path...\n");
		if (type == OPEN_DELEGATE_WRITE) {
			list_del(&dp->dl_hash);
			NFS4_DELEGATION(inode) = NULL;
			spin_unlock(&nfs4_wreq_lock);
			printk("AAA accept_delegation: returning old delegation\n");
			nfs4_do_delegreturn(sb, dp->dl_stateid);
			NFS4_FREE(dp);
			continue;
		}
		spin_unlock(&nfs4_wreq_lock);
		goto free_do_delegreturn;
	}

	/* NOTREACHED */
	BUG();

free_do_delegreturn:
	NFS4_FREE(new_dp);
do_delegreturn:
	printk("AAA accept_delegation: returning delegation immediately\n");
	nfs4_do_delegreturn(sb, new_stateid);
}

/*
 * This routine looks up a delegation by stateid.
 * Note: call with spinlock held!
 */
struct nfs4fs_delegation *
nfs4_find_delegation(struct super_block *sb, stateid4 stateid)
{
	unsigned int hashval = delegation_hashval(sb, stateid);
	struct list_head *l;
	struct nfs4fs_delegation *dp;

	list_for_each(l, &delegation_hashtbl[hashval]) {
		dp = list_entry(l, struct nfs4fs_delegation, dl_hash);
		if (dp->dl_sb == sb && !memcmp(dp->dl_stateid, stateid, sizeof(stateid4)))
			goto out;
	}

	printk("AAA nfs4_find_delegation: not found!\n");
	dp = NULL;
	
out:
	return dp;
}

/*
 * This routine is called by renewd to determine whether a delegation
 * exists on the client.  If one does exist, its stateid is written
 * out through the 'stateid' pointer.
 */
int
nfs4_exists_delegation(struct super_block *sb, stateid4 stateid)
{
	int i;
	struct list_head *l;
	struct nfs4fs_delegation *dp;
	int gotit = 0;
	
	/*
	 * Implemented very inefficiently for now (bakeoff);
	 * need to improve this by keeping all delegations on
	 * a per-superblock list!
	 */
	spin_lock(&nfs4_wreq_lock);
	for (i = 0; i < DELEG_HASH_SIZE; i++) {
		list_for_each(l, &delegation_hashtbl[i]) {
			dp = list_entry(l, struct nfs4fs_delegation, dl_hash);
			if (dp->dl_sb != sb)
				continue;
			memcpy(stateid, dp->dl_stateid, sizeof(stateid4));
			gotit = 1;
			break;
		}
	}
	spin_unlock(&nfs4_wreq_lock);

	return gotit;
}

void
nfs4fs_init_callbacks(void)
{
	int i;
	
	memset(callback_vector, 0, sizeof(callback_vector));
	for (i = 0; i < CB_MAX; i++) {
		init_MUTEX_LOCKED(&callback_vector[i].start);
		init_waitqueue_head(&callback_vector[i].wait);
	}
	for (i = 0; i < DELEG_HASH_SIZE; i++)
		INIT_LIST_HEAD(&delegation_hashtbl[i]);
}
