/*
 * fs/nfs4fs/callback_xdr.c
 *
 * (C) 2001 Kendrick Smith <kmsmith@umich.edu>
 */

#include <linux/sched.h>
#include <linux/smp.h>
#include <linux/smp_lock.h>
#include <linux/fs.h>
#include <linux/pagemap.h>
#include <linux/sunrpc/svc.h>
#include <linux/nfsd/cache.h>

#include <linux/nfs4/nfs4_prot.h>
#include <linux/nfs4/nfs4_debug.h>
#include <linux/nfs4/nfs4.h>
#include <linux/nfs4/nfs4fs.h>

/*
 * The first part of this file is devoted to
 * routines to process delegation callbacks.
 */

#define CB_ISIZE	1
#define CB_MAXTAGLEN	64
#define CB_MAXOP	256

struct cb_arg {
	u32		opnum;
	union {
		struct {
			
		} getattr;
		struct {
			stateid4	stateid;
			int		truncate;
		} recall;
	} u;
};
#define recall_stateid		u.recall.stateid
#define recall_truncate		u.recall.truncate


struct cb_args {
#if NFS4_DEBUG
	u32		taglen;
	char		tag[CB_MAXTAGLEN];
#endif
	u32		minorversion;
	u32		nops;
	struct cb_arg	*args;
	struct cb_arg	iargs[CB_ISIZE];
};

struct cb_rep {
	u32		opnum;
	u32		status;
};

struct cb_reps {
#if NFS4_DEBUG
	u32		taglen;
	char		tag[CB_MAXTAGLEN + 1];
#endif
	u32		status;
	u32		nops;
	struct cb_rep	*reps;
	struct cb_rep	ireps[CB_ISIZE];
};

static int
cb_decode_null(struct svc_rqst *rqstp, u32 *p, void *dummy)
{
	return 1;
}

static int
cb_proc_null(struct svc_rqst *rqstp, void *argp, void *resp)
{
	struct super_block *sb;
	return nfs4_callback_sb(rqstp, &sb);
}

static int
cb_encode_null(struct svc_rqst *rqstp, u32 *p, void *dummy)
{
	struct svc_buf *buf = &rqstp->rq_resbuf;

	buf->len = p - buf->base;
	return 1;
}

#define CHECK(condition)	\
if (!(condition)) {		\
	nfsv4_printk(level1, "xdr error in callback: line %d\n", __LINE__); \
	goto xdr_error;		\
}

static int
cb_decode_compound(struct svc_rqst *rqstp, u32 *p, struct cb_args *argp)
{
	struct svc_buf *buf = &rqstp->rq_argbuf;
	u32 *end = buf->base + buf->buflen;
	u32 len, xdrlen;
	struct cb_arg *arg;
	int i;

	printk("AAA cb_decode_compound: starting\n");
	argp->args = argp->iargs;
	
	/*
	 * Get tag length.
	 */
	CHECK(p + 1 <= end);
	len = ntohl(*p++);
	xdrlen = XDR_QUADLEN(len);
	CHECK(len <= CB_MAXTAGLEN);

	/*
	 * Read tag, minor version, array length
	 */
	CHECK(p + xdrlen + 2 <= end);
#if NFS4_DEBUG
	argp->taglen = len;
	memcpy(argp->tag, p, len);
#endif
	p += xdrlen;
	argp->minorversion = ntohl(*p++);
	argp->nops = ntohl(*p++);

	if (argp->nops > CB_ISIZE) {
		argp->args = NFS4_ALLOC(argp->nops * sizeof(struct cb_arg));
		CHECK(argp->args);
	}

	/*
	 * Read the individual ops.
	 */
	for (i = 0; i < argp->nops; i++) {
		arg = &argp->args[i];
		CHECK(p + 1 <= end);
		arg->opnum = ntohl(*p++);
		
		switch (arg->opnum) {
		case OP_CB_GETATTR:
			/*
			 * XXX: not yet!
			 */
			CHECK(0);
			
		case OP_CB_RECALL:
			/*
			 * Get stateid, truncate flag, fhlen.
			 */
			CHECK(p + XDR_QUADLEN(sizeof(stateid4)) + 2 <= end);
			memcpy(arg->recall_stateid, p, sizeof(stateid4));
			p += XDR_QUADLEN(sizeof(stateid4));
			arg->recall_truncate = ntohl(*p++);
			len = ntohl(*p++);
			xdrlen = XDR_QUADLEN(len);
			CHECK(len <= NFS4_FHSIZE);

			/*
			 * Just ignore the filehandle for now.
			 */
			CHECK(p + xdrlen <= end);
			p += xdrlen;
			
			break;
			
		default:
			CHECK(0);
		}
	}

	printk("AAA cb_decode_compound: done\n");
	return 0;

xdr_error:
	if (argp->args != argp->iargs && argp->args) {
		NFS4_FREE(argp->args);
		argp->args = NULL;
	}
	return 1;
}

static int
cb_proc_compound(struct svc_rqst *rqstp, struct cb_args *argp, struct cb_reps *repp)
{
	struct super_block *sb;
	struct cb_arg *arg;
	struct cb_rep *rep;
	int i;
	struct nfs4fs_delegation *dp;
	int rpc_status, nfs_status;

	repp->reps = repp->ireps;
	
	rpc_status = htonl(RPC_PROG_UNAVAIL);
	if (nfs4_callback_sb(rqstp, &sb))
		goto out;

#if NFS4_DEBUG
	repp->taglen = argp->taglen;
	memcpy(repp->tag, argp->tag, argp->taglen);
#endif
	
	rpc_status = rpc_success;
	nfs_status = NFS4_OK;
	
	if (argp->nops > CB_ISIZE) {
		repp->reps = NFS4_ALLOC(argp->nops * sizeof(struct cb_rep));
		if (!repp->reps) {
			/*
			 * Failing this allocation is kind of awkward...
			 * We have to fake the entire response.
			 */
			repp->reps = repp->ireps;
			rep = &repp->ireps[0];
			rep->opnum = argp->args[0].opnum;
			rep->status = NFS4ERR_RESOURCE;
			repp->nops = 1;
			repp->status = NFS4ERR_RESOURCE;
			goto out;
		}
	}
	
	for (i = 0; i < argp->nops && !nfs_status; i++) {
		arg = &argp->args[i];
		rep = &repp->reps[i];
		
		rep->opnum = arg->opnum;
		switch (arg->opnum) {
			
		case OP_CB_GETATTR:
			BUG();
			
		case OP_CB_RECALL:
			spin_lock(&nfs4_wreq_lock);
			dp = nfs4_find_delegation(sb, arg->recall_stateid);
			if (dp) {
				list_del(&dp->dl_hash);
				NFS4_DELEGATION(dp->dl_inode) = NULL;
			}
			spin_unlock(&nfs4_wreq_lock);

			nfs_status = NFS4ERR_BAD_STATEID;
			if (dp) {
				/*
				 * XXX: What is the most sensible thing to do
				 * if nfs4_do_delegreturn() fails?
				 */
				nfs4_do_delegreturn(sb, dp->dl_stateid);
				NFS4_FREE(dp);
				nfs_status = NFS4_OK;
			}
			
			break;
			
		default:
			BUG();
		}

		rep->status = nfs_status;
	}

	repp->status = nfs_status;
	repp->nops = i;
	
out:
	if (argp->args != argp->iargs && argp->args) {
		NFS4_FREE(argp->args);
		argp->args = NULL;
	}
	return rpc_status;
}

static int
cb_encode_compound(struct svc_rqst *rqstp, u32 *p, struct cb_reps *repp)
{
	struct svc_buf *buf = &rqstp->rq_resbuf;
	u32 *end = buf->base + buf->buflen;
	u32 xdrlen;
	struct cb_rep *rep;
	int i;
	int status;
		
	buf->len = p - buf->base;   /* paranoid */

	/*
	 * Encode status, tag, resarray length.
	 */
#if NFS4_DEBUG
	xdrlen = XDR_QUADLEN(repp->taglen);
	CHECK(p + xdrlen + 3 <= end);
	*p++ = htonl(repp->status);
	*p++ = htonl(repp->taglen);
	memcpy(p, repp->tag, repp->taglen);
	p += xdrlen;
	*p++ = htonl(repp->nops);
#else
	CHECK(p + 3 <= end);
	*p++ = htonl(repp->status);
	*p++ = htonl(repp->taglen);
	*p++ = htonl(repp->nops);
#endif

	/*
	 * Encode individual ops.
	 */
	for (i = 0; i < repp->nops; i++) {
		rep = &repp->reps[i];
		/*
		 * Encode opnum, status.
		 */
		CHECK(p + 2 <= end);
		*p++ = htonl(rep->opnum);
		*p++ = htonl(rep->status);

		/*
		 * When we implement CB_GETATTR, there will
		 * be more stuff here...
		 */
	}
	
	buf->len = p - buf->base;   /* paranoid */
	status = 1;
	
out:
	if (repp->reps != repp->ireps && repp->reps) {
		NFS4_FREE(repp->reps);
		repp->reps = NULL;
	}
	return status;

xdr_error:
	nfsv4_printk(level1, "cb_encode_compound: failed to XDR response!\n");
	status = 0;
	goto out;
}

/*
 * XXX: The 1024's here are total overkill; fix later...
 */
#define PROC(proc)			\
{	(svc_procfunc) cb_proc_##proc,	\
	(kxdrproc_t) cb_decode_##proc,	\
	(kxdrproc_t) cb_encode_##proc,	\
	NULL,				\
	1024,			       	\
	1024,				\
	0,				\
	RC_NOCACHE			\
}

static struct svc_procedure cb_v1_procedures[] = {
	PROC(null),
	PROC(compound)
};

struct svc_version nfs4_cb_version1 = {
	1,							/* version number */
	sizeof(cb_v1_procedures) / sizeof(cb_v1_procedures[0]),	/* no. of procedures */
	cb_v1_procedures,					/* procedure table */
	NULL							/* dispatcher */
};

/*
 * The second part of this file contains a
 * rotine to return a delegation by stateid.
 */

static void
delegreturn_handler(struct rpc_task *task)
{
	struct nfs4_async_compound *cp = (struct nfs4_async_compound *)task->tk_calldata;
	struct compound4args *argp = &cp->ac_arg;
	struct compound4res *resp = &cp->ac_res;

	printk("AAA delegreturn_handler: starting\n");
	
	if (&cp->ac_task != task)
		BUG();
	if (task->tk_status < 0) {
		nfsv4_printk(open2, "delegreturn_handler: task->tk_status < 0!\n");
		goto err;
	}
	if (nfs4_async_prehandler(cp, NULL)) {
		nfsv4_printk(open2, "delegreturn_handler: async_prehandler failed!\n");
		goto err;
	}
	if (nfs4_handle_delegreturn(resp)) {
		nfsv4_printk(open2, "delegreturn_handler: handle_delegreturn failed!\n");
		goto err_res;
	}

	nfs4_release_compoundres(resp);
	nfs4_release_compoundargs(argp);
	printk("AAA delegreturn_handler: success\n");
	return;

err_res:
	nfs4_release_compoundres(resp);
err:
	printk(KERN_WARNING "nfs4: warning: couldn't return delegation!\n");
	nfs4_release_compoundargs(argp);
	return;
}

int
nfs4_do_delegreturn(struct super_block *sb, stateid4 stateid)
{
	struct nfs4_async_compound *cp;
	struct rpc_task *task;
	struct compound4args *argp;

	printk("AAA do_delegreturn: starting\n");
	
	if (!(cp = nfs4_async_create(sb))) {
		printk("AAA return_delegation: out of memory!\n");
		goto err;
	}
	
	task = &cp->ac_task;
	argp = &cp->ac_arg;
	argp->sb = sb;
	argp->tag = "delegreturn";
	
	if (nfs4_setup_compound(argp)) {
		printk("AAA: return_delegation: setup_compound() failed!\n");
		goto err_task;
	}
	if (nfs4_setup_delegreturn(argp, stateid)) {
		printk("AAA: return_delegation: setup_delegreturn() failed!\n");
		goto err_argp;
	}
	
	task->tk_action = nfs4_async_call_compound;
	task->tk_exit = delegreturn_handler;
	rpc_execute(task);
	printk("AAA do_delegreturn: async task started...\n");
	return NFS4_OK;

err_argp:
	nfs4_release_compoundargs(argp);
err_task:
	rpc_release_task(task);
err:
	printk(KERN_WARNING "nfs4: warning: couldn't return delegation!\n");
	return NFS4ERR_RESOURCE;
}
