/*
 *  linux/fs/nfs4/decode.c
 *
 *  Copyright (c) 2001 The Regents of the University of Michigan.
 *  All rights reserved.
 *
 *  Kendrick Smith <kmsmith@umich.edu>
 *  Andy Adamson <andros@umich.edu>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/mm.h>
#include <linux/pagemap.h>
#include <asm/uaccess.h>

#include <linux/nfs4/nfs4_prot.h>
#include <linux/nfs4/nfs4_debug.h>
#include <linux/nfs4/nfs4.h>
#include <linux/nfs4/nfs4fs.h>

#if 1
#define seqid_mutating_err(err)	    ((err) == 0)
#else
#define seqid_mutating_err(err)		\
(((err) != NFS4ERR_STALE_CLIENTID) &&	\
 ((err) != NFS4ERR_BAD_SEQID) &&	\
 ((err) != NFS4ERR_STALE_STATEID) &&	\
 ((err) != NFS4ERR_BAD_STATEID))
#endif

/* XXX : is there a better way to handle the dynamic read_buf()'s in this file? */
/* XXX : one side effect of the way things are done here is that if there is extra data
   at the end of an RPC reply, we don't consider this to be an XDR error.  Is this OK? */

#define READDIR_NBYTES  64
#define LINK_MAXNAME    1024

static int open_confirm(struct nfs4fs_open *open);

/*
 *  Primitives for XDR decoding are:
 *    copy_buf() - copies the next n bytes out of the XDR stream into a preallocated buffer
 *    read_buf() - returns a pointer to a block of memory containing the next n bytes from
 *                 the XDR stream.  The caller must pass in a "backup buffer" of size n.
 *                 If feasible, the return value of read_buf() will return a pointer which
 *                 points directly into the XDR stream.  If this is not feasible, the data
 *                 will be copied into the backup buffer, and a pointer to the backup buffer
 *                 returned.
 *    skip_buf() - skips over the next n bytes in the XDR decode stream
 *
 *  All XDR decode routines are written in terms of these primitives, without direct use
 *  to the iov_* fields, so if a different internal representation of the output buffer is
 *  needed, it is only necessary to change the primitives.
 */

static int
copy_buf(struct compound4res *rp, int nbytes, char *buf)
{
	int buflen;
	int status;
	int xdrlen = XDR_LEN(nbytes);

	/* if nbytes is not a multiple of 4, automatically advances buffer past any XDR
	   "baggage" */
	
	for (;;) {
		buflen = rp->end - rp->p;
		if (buflen >= xdrlen) {
			memcpy(buf, rp->p, nbytes);
			rp->p += xdrlen;
			status = 0;
			goto out;
		}
		if (rp->iov_current == rp->iov_nr-1) {
			nfsv4_printk(xdr2, "copy_buf: buffer overflow!\n");
			status = -EIO;
			goto out;
		}
		if (buflen >= nbytes) {
			memcpy(buf, rp->p, nbytes);
			nbytes = 0;   /* XXX : make sure memcpy will work if nbytes==0 */
		}
		else {
			memcpy(buf, rp->p, buflen);
			nbytes -= buflen;
		}
		xdrlen -= buflen;
		buf += buflen;
		rp->iov_current++;
		rp->p = rp->iov[rp->iov_current].iov_base;
		rp->end = rp->p + rp->iov[rp->iov_current].iov_len;
	}

out:
	return status;
}

static inline u32 *
read_buf(struct compound4res *rp, int nbytes, char *buf)
{
	u32 *p;

#if NFS4_DEBUG
	if (nbytes & 3) {
		printk("!!! read_buf: nbytes not divisible by 4 !!!\n");
		return NULL;
	}
#endif

	if (rp->p + nbytes <= rp->end) {
		p = (u32 *)rp->p;
		rp->p += nbytes;
	}
	else
		p = copy_buf(rp, nbytes, buf) ? NULL : (u32 *)buf;
	return p;
}

static inline int
skip_buf(struct compound4res *rp, int nbytes)
{
	int buflen;

#if NFS4_DEBUG
	if (nbytes & 3) {
		printk("!!! skip_buf: nbytes not divisible by 4 !!!\n");
		return -EIO;
	}
#endif
	
	for (;;) {
		buflen = rp->end - rp->p;
		if (nbytes <= buflen) {
			rp->p += nbytes;
			return 0;
		}
		if (rp->iov_current == rp->iov_nr-1) {
			nfsv4_printk(xdr2, "skip_buf: buffer overflow!\n");
			return -EIO;
		}
		nbytes -= buflen;
		rp->iov_current++;		
		rp->p = rp->iov[rp->iov_current].iov_base;
		rp->end = rp->p + rp->iov[rp->iov_current].iov_len;
	}
}

static void
handle_change_info4(u32 *p, struct inode *inode)
{
	u64 changeid;
	
	if (*p++) {     /* operation is atomic */
		changeid = (u64)htonl(*p++) << 32;
		changeid |= htonl(*p++);
		if (changeid == NFS4_CACHE_CHANGEID(inode)) {
			changeid = (u64)htonl(*p++) << 32;
			changeid |= htonl(*p++);
			NFS4_CACHE_CHANGEID(inode) = changeid;
			NFS4_CACHE_JIFFIES(inode) = jiffies;
			if (NFS4_UPDATE_TIMEOUT(inode)) {
				unsigned long timeout = NFS4_CACHE_TIMEOUT(inode);
				unsigned long maxtimeout = NFS4_MAXTIMEOUT(inode);
				timeout <<= 1;
				if (timeout > maxtimeout)
					timeout = maxtimeout;
				NFS4_CACHE_TIMEOUT(inode) = timeout;
				NFS4_CACHE_TIMEOUT_JIFFIES(inode) = jiffies;
			}
			return;
		}
	}
	
	/* operation was non-atomic, or changeids failed to match */
	nfs4_zap_caches(inode);
}

/*
 *  read_attrs: a routine to read the result of a GETATTR operation
 */

static inline int
read_attrs(struct compound4res *rp, struct nfs4fs_fattr *nfp)
{
	u32 bmlen;
        u32 *bmval;
	u32 *p;
	int status;
	char sbuf[32];
	char bbuf[8];
	char *buf = sbuf;
	int buflen = sizeof(sbuf);
	int i;
	int attrlen, len, xdrlen;
	u32 flag;

#if NFS4_DEBUG
	rp->savep = rp->p;
#endif
	
	/* XXX : need more sanity checking on attributes read! */
	/* In particular, a test to make sure the server didn't return attrs we can't
	   parse */

	if (!(p = read_buf(rp, 4, buf))) {
		nfsv4_printk(getattr2, "read_attrs: no bmlen!\n");
		goto xdr_error;
	}
	bmlen = ntohl(*p++);
	nfsv4_printk(getattr, "read_attrs: starting, read bitmap len %d\n", bmlen);
	
	if (bmlen <= 2) {
		if (!(bmval = read_buf(rp, bmlen << 2, bbuf))) {
			nfsv4_printk(getattr2, "read_attrs: no bmval!\n");
			goto xdr_error;
		}
		/* ntohl()'ing of bitmap components will take place later */
	}
	else {
		/* weird case, presumably an xdr error, but go ahead and try to parse it.. */
		if (!(bmval = read_buf(rp, 8, bbuf))) {
			nfsv4_printk(getattr2, "read_attrs: no bmval [2]!\n");
			goto xdr_error;
		}
		for (i = 2; i < bmlen; i++) {
			if (!(p = read_buf(rp, 4, buf))) {
				nfsv4_printk(getattr2, "read_attrs: no bmval[3]!\n");
				goto xdr_error;
			}
			if (*p) {   /* no need to ntohl(), since we only care if nonzero */
				nfsv4_printk(getattr2, "read_attrs: bad bitmap!\n");
				goto xdr_error;
			}
		}
	}

	if (!(p = read_buf(rp, 4, buf))) {
		nfsv4_printk(getattr2, "read_attrs: no attrlen!\n");
		goto xdr_error;
	}
	attrlen = ntohl(*p++);
	nfsv4_printk(getattr, "read_attrs: read attribute len %d\n", attrlen);
#if 0
  if (attrlen > 100)
    attrlen = 100;
	hex_dump(rp->p, attrlen);
#endif 

	nfp->fa_valid = 0;
	
#define OVERFLOW_CHECK(n)  do {		\
	if ((attrlen -= (n)) < 0) {	\
		nfsv4_printk(getattr2, "read_attrs: xdr error on line %d\n", __LINE__); \
		goto xdr_error;		\
	}			       	\
} while (0)
#define PREPARE_BUF(n)     do {		\
	if (!(p = read_buf(rp, (n), buf))) { \
		nfsv4_printk(getattr2, "read_attrs: xdr error on line %d\n", __LINE__); \
		goto xdr_error;		\
	}				\
} while (0)
#define PREPARE_DBUF(n)    do {				\
	if (buflen < (n)) {				\
		if (buf != sbuf)			\
			NFS4_FREE(buf);			\
		if (!(buf = NFS4_ALLOC(n)))		\
			goto nomem;			\
		buflen = (n);				\
	}						\
	if (!(p = read_buf(rp, (n), buf))) {		\
		nfsv4_printk(getattr2, "read_attrs: xdr error on line %d\n", __LINE__); \
		goto xdr_error;				\
	}						\
} while (0)
#define READ64(x)   do {		\
	x = (u64)ntohl(*p++) << 32;	\
	x |= ntohl(*p++);		\
} while (0)

#if 0
/* XXX : need to correctly handle 64-bit time_t */
#define READTIME(x) do {		\
	if (*p) {			\
		x = ~(u32)0;		\
		p += 3;			\
	}				\
	else {				\
		p++;			\
		x = ntohl(*p);		\
		p += 2;			\
	}				\
} while (0)
#else
#define READTIME(x) do {		\
	p++;				\
	x = ntohl(*p);			\
	p += 2;				\
} while (0)
#endif
			    
	/*
	 * parse first bitmap block
	 */
	if (!bmlen)
		goto done;
	bmval[0] = ntohl(bmval[0]);
	if (bmval[0] & 0x00000002) {    /* FATTR4_TYPE */
		OVERFLOW_CHECK(4);
		PREPARE_BUF(4);
		nfp->fa_type = ntohl(*p++);
		nfsv4_printk(getattr, "read_attrs: type=%d\n", (u32)nfp->fa_type);
		nfp->fa_valid |= FA_TYPE;
	}
	if (bmval[0] & 0x00000008) {    /* FATTR4_CHANGE */
		OVERFLOW_CHECK(8);
		PREPARE_BUF(8);
		READ64(nfp->fa_changeid);
		nfsv4_printk(getattr, "read_attrs: changeid=%d\n", (u32)nfp->fa_changeid);
		nfp->fa_valid |= FA_CHANGEID;
	}
	if (bmval[0] & 0x00000010) {    /* FATTR4_SIZE */
		OVERFLOW_CHECK(8);
		PREPARE_BUF(8);
		READ64(nfp->fa_size);
		nfsv4_printk(getattr, "read_attrs: size=%d\n", (u32)nfp->fa_size);
		nfp->fa_valid |= FA_SIZE;
	}
	if (bmval[0] & 0x00000100) {    /* FATTR4_FSID */
		OVERFLOW_CHECK(16);
		PREPARE_BUF(16);
		READ64(nfp->fa_fsid_major);
		READ64(nfp->fa_fsid_minor);
		nfsv4_printk(getattr, "read_attrs: fsid=0x%Lx/0x%Lx\n",
			     nfp->fa_fsid_major, nfp->fa_fsid_minor);
		nfp->fa_valid |= FA_FSID;
	}
	if (bmval[0] & 0x00000400) {    /* FATTR4_LEASE_TIME */
		OVERFLOW_CHECK(4);
		PREPARE_BUF(4);
		nfp->fa_lease_time = ntohl(*p++);
		nfp->fa_valid |= FA_LEASE_TIME;
	}
        if (bmval[0] & 0x00001000) {    /* FATTR4_ACL */
#ifdef NFS4ACL
		printk("Doing FATTR4_ACL decode\n");
	        OVERFLOW_CHECK(4);
	        PREPARE_BUF(4);
	        len = ntohl(*p++);
	        nfsv4_printk(getattr, "read_attrs: num ace=%d\n", len);
	        nfp->fa_nacl = len;
		if (!len) 		/* no acls returned */
			goto acl_out;
			
		nfp->fa_acl = NFS4_ALLOC(sizeof(struct nfs4_acl) * len);
	       	for (i = 0; i < nfp->fa_nacl; i++) {
			nfsv4_printk(getattr, "---------------\n");
	       		OVERFLOW_CHECK(4);
	                PREPARE_BUF(4);
	                nfp->fa_acl[i].type = ntohl(*p++);
	                nfsv4_printk(getattr, "read_attrs: acl type=%d\n", nfp->fa_acl[i].type);
			OVERFLOW_CHECK(4);
			PREPARE_BUF(4);
			nfp->fa_acl[i].flag = ntohl(*p++);
			nfsv4_printk(getattr, "read_attrs: acl flag=%d\n", nfp->fa_acl[i].flag);
			OVERFLOW_CHECK(4);
			PREPARE_BUF(4);
			nfp->fa_acl[i].mask = ntohl(*p++);
			nfsv4_printk(getattr, "read_attrs: acl mask=%d\n", nfp->fa_acl[i].mask);
			OVERFLOW_CHECK(4);
			PREPARE_BUF(4);
			len = ntohl(*p++);
			nfsv4_printk(getattr, "read_attrs: acl owner len=%d\n", len);
			if (len > GSS_NAME_MAX)
				goto xdr_error;
			xdrlen = XDR_QUADLEN(len) << 2;
			OVERFLOW_CHECK(xdrlen);
			PREPARE_DBUF(xdrlen);
			printk("xdrlen: %d\n", xdrlen);

			flag = nfp->fa_acl[i].flag;
			if (!memcmp(p, "OWNER@", 6)){
				printk("OWNER@\n");
				nfp->fa_acl[i].flag |= ACE4_OWNER;
	 			nfp->fa_acl[i].id = -1;			
			} else if (!memcmp(p, "GROUP@", 6)) {
				printk("GROUP@\n");
				nfp->fa_acl[i].flag |= ACE4_GROUP;
				nfp->fa_acl[i].id = -1;
			} else if (!memcmp(p, "EVERYONE@", 9)) {
				printk("EVERYONE@\n");
				nfp->fa_acl[i].flag |= ACE4_EVERYONE;
				nfp->fa_acl[i].id = -1;
			} else {
				if (flag & ACE4_IDENTIFIER_GROUP) {
					printk("ACE4_ID_GROUP\n");
					if ((status = gss_get_num(GSS_GROUP, len, (char *)p, &nfp->fa_acl[i].id))){
						nfsv4_printk(getattr2, "read_attrs: gss_get_num failed!\n");
						nfsv4_printk(getattr2, "read_attrs: gss_get_num tried %s\n", (char *) p);
				   		goto out;
					}
				} else {
					printk("REG USER\n");
					printk("user: %s\n", (char *)p);
					printk("len: %d\n", len);
					printk("p: %p\n", p);
					printk("nfp->fa_acl[i].id: %p\n", &nfp->fa_acl[i].id);
					printk("nfp->fa_acl[i].id: %d\n", nfp->fa_acl[i].id);
					printk("read_attr: going into gss_get_num\n");
					if ((status = gss_get_num(GSS_OWNER, len, (char *)p, &nfp->fa_acl[i].id))){
						nfsv4_printk(getattr2, "read_attrs: gss_get_num failed!\n");
						nfsv4_printk(getattr2, "read_attrs: gss_get_num tried %s\n", (char *) p);
					        goto out;
					}
				}
				printk("Decode: Out of gss_get_num\n");
			}	
			
			nfsv4_printk(getattr, "read_attrs: acl uid=%d\n", nfp->fa_acl[i].id);
		}
		nfp->fa_valid |= FA_ACL;
acl_out:
		printk("Out of FATTR4_ACL decode\n");
#else
		// XXX send back an error

#endif // #ifdef NFS4ACL

	}
	if (bmval[0] & 0x00100000) {    /* FATTR4_FILEID */
		OVERFLOW_CHECK(8);
		PREPARE_BUF(8);
		READ64(nfp->fa_fileid);
		nfsv4_printk(getattr, "read_attrs: fileid=%Ld\n", nfp->fa_fileid);
		nfp->fa_valid |= FA_FILEID;
	}
	if (bmval[0] & 0x00400000) {    /* FATTR4_FILES_FREE */
		OVERFLOW_CHECK(8);
		PREPARE_BUF(8);
		READ64(nfp->fa_ffree);
		nfsv4_printk(getattr, "read_attrs: files_free=0x%Lx\n", nfp->fa_ffree);
		nfp->fa_valid |= FA_FFREE;
	}
	if (bmval[0] & 0x00800000) {    /* FATTR4_FILES_TOTAL */
		OVERFLOW_CHECK(8);
		PREPARE_BUF(8);
		READ64(nfp->fa_ftotal);
		nfsv4_printk(getattr, "read_attrs: files_tot=0x%Lx\n", nfp->fa_ftotal);
		nfp->fa_valid |= FA_FTOTAL;
	}
	if (bmval[0] & 0x08000000) {    /* FATTR4_MAXFILESIZE */
		OVERFLOW_CHECK(8);
		PREPARE_BUF(8);
		READ64(nfp->fa_maxfilesize);
		nfsv4_printk(getattr, "read_attrs: maxfilesize=0x%Lx\n", nfp->fa_maxfilesize);
		nfp->fa_valid |= FA_MAXFILESIZE;
	}
	if (bmval[0] & 0x20000000) {    /* FATTR4_MAXNAME */
		OVERFLOW_CHECK(4);
		PREPARE_BUF(4);
		nfp->fa_maxname = ntohl(*p++);
		nfsv4_printk(getattr, "read_attrs: maxname=%d\n", nfp->fa_maxname);
		nfp->fa_valid |= FA_MAXNAME;
	}
	if (bmval[0] & 0x40000000) {    /* FATTR4_MAXREAD */
		OVERFLOW_CHECK(8);
		PREPARE_BUF(8);
		READ64(nfp->fa_maxread);
		nfsv4_printk(getattr, "read_attrs: maxread=0x%Lx\n", nfp->fa_maxread);
		nfp->fa_valid |= FA_MAXREAD;
	}
	if (bmval[0] & 0x80000000) {    /* FATTR4_MAXWRITE */
		OVERFLOW_CHECK(8);
		PREPARE_BUF(8);
		READ64(nfp->fa_maxwrite);
		nfsv4_printk(getattr, "read_attrs: maxwrite=0x%Lx\n", nfp->fa_maxwrite);
		nfp->fa_valid |= FA_MAXWRITE;
	}

	/*
	 *  parse second bitmap block
	 */
	if (bmlen == 1)
		goto done;
	bmval[1] = ntohl(bmval[1]);
	if (bmval[1] & 0x00000002) {    /* FATTR4_MODE */
		OVERFLOW_CHECK(4);
		PREPARE_BUF(4);
		nfp->fa_mode = ntohl(*p++);
		nfsv4_printk(getattr, "read_attrs: mode=0%o\n", nfp->fa_mode);
		nfp->fa_valid |= FA_MODE;
	}
	if (bmval[1] & 0x00000008) {    /* FATTR4_NUMLINKS */
		OVERFLOW_CHECK(4);
		PREPARE_BUF(4);
		nfp->fa_nlink = ntohl(*p++);
		nfsv4_printk(getattr, "read_attrs: nlinks=0%o\n", nfp->fa_nlink);
		nfp->fa_valid |= FA_NLINK;
	}
	if (bmval[1] & 0x00000010) {    /* FATTR4_OWNER */
		OVERFLOW_CHECK(4);
		PREPARE_BUF(4);
		len = ntohl(*p++);
		nfsv4_printk(getattr, "read_attrs: owner len=%d\n", len);
		if (len > GSS_NAME_MAX) {
			nfsv4_printk(getattr2, "read_attrs: name too long!\n");
			goto xdr_error;
		}
		xdrlen = XDR_QUADLEN(len) << 2;
		OVERFLOW_CHECK(xdrlen);
		PREPARE_DBUF(xdrlen);
		if ((status = gss_get_num(GSS_OWNER, len, (char *)p, &nfp->fa_uid))) {
			nfsv4_printk(getattr2, "read_attrs: gss_get_num failed!\n");
			goto out;
		}
		nfsv4_printk(getattr, "read_attrs: uid=%d\n", (int)nfp->fa_uid);
		nfp->fa_valid |= FA_UID;
	}
	if (bmval[1] & 0x00000020) {    /* FATTR4_OWNER_GROUP */
		OVERFLOW_CHECK(4);
		PREPARE_BUF(4);
		len = ntohl(*p++);
		nfsv4_printk(getattr, "read_attrs: group len=%d\n", len);
		if (len > GSS_NAME_MAX) {
			nfsv4_printk(getattr2, "read_attrs: name too long!\n");
			goto xdr_error;
		}
		xdrlen = XDR_QUADLEN(len) << 2;
		OVERFLOW_CHECK(xdrlen);
		PREPARE_DBUF(xdrlen);
		if ((status = gss_get_num(GSS_GROUP, len, (char *)p, &nfp->fa_gid))) {
			nfsv4_printk(getattr2, "read_attrs: gss_get_num failed!\n");
			goto out;
		}
		nfsv4_printk(getattr, "read_attrs: gid=%d\n", (int)nfp->fa_gid);
		nfp->fa_valid |= FA_GID;
	}
	if (bmval[1] & 0x00000200) {    /* FATTR4_RAWDEV */
		OVERFLOW_CHECK(8);
		PREPARE_BUF(8);
		nfp->fa_rdev_major = ntohl(*p++);
		nfp->fa_rdev_minor = ntohl(*p++);
		nfsv4_printk(getattr, "read_attrs: rdev=%d %d\n",
			     nfp->fa_rdev_major, nfp->fa_rdev_minor);
		nfp->fa_valid |= FA_RDEV;
	}
	if (bmval[1] & 0x00000400) {    /* FATTR4_SPACE_AVAIL */
		OVERFLOW_CHECK(8);
		PREPARE_BUF(8);
		READ64(nfp->fa_savail);
		nfsv4_printk(getattr, "read_attrs: savail=0x%Lx\n", nfp->fa_savail);
		nfp->fa_valid |= FA_SAVAIL;
	}
	if (bmval[1] & 0x00000800) {    /* FATTR4_SPACE_FREE */
		OVERFLOW_CHECK(8);
		PREPARE_BUF(8);
		READ64(nfp->fa_sfree);
		nfsv4_printk(getattr, "read_attrs: sfree=0x%Lx\n", nfp->fa_sfree);
		nfp->fa_valid |= FA_SFREE;
	}
	if (bmval[1] & 0x00001000) {    /* FATTR4_SPACE_TOTAL */
		OVERFLOW_CHECK(8);
		PREPARE_BUF(8);
		READ64(nfp->fa_stotal);
		nfsv4_printk(getattr, "read_attrs: stotal=0x%Lx\n", nfp->fa_stotal);
		nfp->fa_valid |= FA_STOTAL;
	}
	if (bmval[1] & 0x00008000) {    /* FATTR4_TIME_ACCESS */
		OVERFLOW_CHECK(12);
		PREPARE_BUF(12);
		READTIME(nfp->fa_atime);
		nfsv4_printk(getattr, "read_attrs: atime=%d\n", (int)nfp->fa_atime);
		nfp->fa_valid |= FA_ATIME;
	}
	if (bmval[1] & 0x00040000) {    /* FATTR4_TIME_CREATE */
		OVERFLOW_CHECK(12);
		PREPARE_BUF(12);
		READTIME(nfp->fa_ctime);
		nfsv4_printk(getattr, "read_attrs: ctime=%d\n", (int)nfp->fa_ctime);
		nfp->fa_valid |= FA_CTIME;
	}
	if (bmval[1] & 0x00200000) {    /* FATTR4_TIME_MODIFY */
		OVERFLOW_CHECK(12);
		PREPARE_BUF(12);
		READTIME(nfp->fa_mtime);
		nfsv4_printk(getattr, "read_attrs: mtime=%d\n", (int)nfp->fa_mtime);
		nfp->fa_valid |= FA_MTIME;
	}

#undef READTIME
#undef READ64
#undef PREPARE_DBUF
#undef PREPARE_BUF
#undef OVERFLOW_CHECK
	
	if (attrlen) {
		nfsv4_printk(getattr2, "read_attrs: buffer underflow!\n");
		goto xdr_error;
	}
	
done:
	status = 0;	
out:
	if (buf != sbuf)
		NFS4_FREE(buf);
out_nofree:
	nfsv4_printk(getattr, "read_attrs returning %d\n", status);
	return status;
xdr_error:
	nfsv4_printk(getattr2, "read_attrs: xdr error!\n");
	printk("WHOLE PACKET\n");
	hex_dump(rp->startp, (char *)rp->end - (char *)rp->startp);
	printk("ATTRS STARTED HERE:\n");
	hex_dump(rp->savep, (char *)rp->end - (char *)rp->savep);
	nfp->fa_valid = 0;
	status = -EIO;
	goto out;
nomem:
	nfsv4_printk(getattr2, "read_attrs: out of memory!\n");
	nfp->fa_valid = 0;
	status = -ENOMEM;
	goto out_nofree;
}

/*
 *  Some helper functions for nfs4_handle_getattr(), more or less lifted from the v3 code.
 *  XXX : come back and make sure this is what we want!
 */

static int
nfs4_find_actor(struct inode *inode, unsigned long ino, void *opaque)
{
	struct nfs4fs_fattr *fattr = (struct nfs4fs_fattr *)opaque;
	
	return	(NFS4_FSID_MAJOR(inode) == fattr->fa_fsid_major) &&
		(NFS4_FSID_MINOR(inode) == fattr->fa_fsid_minor) &&
		(NFS4_FILEID(inode) == fattr->fa_fileid);
}

static void
nfs_invalidate_inode(struct inode *inode)
{
	umode_t save_mode = inode->i_mode;

	/* Invalidate, but do not unhash, the inode */
	make_bad_inode(inode);
	inode->i_mode = save_mode;
	nfs4_zap_caches(inode);
	NFS4_CACHE_JIFFIES(inode) = 0;
}

/*
 *  The actual XDR handler routines
 */

int
nfs4_handle_compound(struct compound4args *argp, struct compound4res *resp,
		     struct nfs4fs_ctx *ctx)
{
	u32 *p;
	unsigned int taglen;
	char buf[8];
	int status = 0;
#if NFS4_DEBUG
	char tag[100];
	resp->startp = resp->p;
#endif
	
	/*
	 *  Decode the top level status, tag, and number of operations.
	 */
	if (!(p = read_buf(resp, 8, buf)))
		goto xdr_error;
	resp->toplevel_status = ntohl(*p++);
	taglen = ntohl(*p++);
	if (taglen > 100) {
		printk("nfs4_handle_compound: xdr error in tag length(%d)!\n",
		       taglen);
		goto xdr_error;
	}
	else {
		if (copy_buf(resp, taglen, tag))
			goto xdr_error;
	}
	
	if (!(p = read_buf(resp, 4, buf)))
		goto xdr_error;
	resp->nops = ntohl(*p++);
	
	nfsv4_printk(level2, "nfs4_handle_compound(%.*s): server returned nfserr=%d, "
		     "nops=%d\n", (int)taglen, tag, resp->toplevel_status, resp->nops);
	if ((resp->nops <= 0) || (resp->nops > argp->nops))
		goto xdr_error;
	
	/* Process and release the seqid.  Currently, seqid increments only on success. */
	if (argp->seqid_holder) {
		if (seqid_mutating_err(resp->toplevel_status) || (resp->nops > argp->seqid_index))
			argp->seqid_holder->lo_seqid++;
		if (argp->flags & CA_SEQID_RELEASE) {
			up(&argp->seqid_holder->lo_sema);
			argp->flags &= ~CA_SEQID_LOCKED;
		}
	}

	/* Process and release the stateid */
	if (argp->stateid_holder && (argp->flags & CA_STATEID_RELEASE)) {
		nfs4_up_read(&argp->stateid_holder->fi_sema);
		argp->flags &= ~CA_STATEID_LOCKED;
	}

	/* Process and reset the lease */
	if (argp->renew_index && (!resp->toplevel_status ||
				  (resp->nops > argp->renew_index))) {
		if (NFS4_LAST_RENEWAL(argp->sb) < argp->timestamp)
			NFS4_LAST_RENEWAL(argp->sb) = argp->timestamp;
	}
	
	switch (resp->toplevel_status) {
	case 0:
		break;
		/*
		 *  Certain errors must be processed specially; their handlers are here.
		 *  NOTE:  We don't do seqid/stateid processing here, so we must be able to
		 *  guarantee that no compound which returns one of these errors mutates a
		 *  seqid or stateid!
		 */

#if 0   /* XXX: special handlers not stable yet, commented out for release */
	case NFS4ERR_GRACE:
	case NFS4ERR_STALE_CLIENTID:
	case NFS4ERR_STALE_STATEID:
		/* This special return value means that the caller must start
		   recoverd and wait for the grace period to end. */
		status = -EREBOOT;
		if (argp->seqid_holder && (argp->flags & CA_SEQID_LOCKED)) {
			up(&argp->seqid_holder->lo_sema);
			argp->flags &= ~CA_SEQID_LOCKED;
		}
		if (argp->stateid_holder && (argp->flags & CA_STATEID_LOCKED)) {
			nfs4_up_write(&argp->stateid_holder->fi_sema);
			argp->flags &= ~CA_STATEID_LOCKED;
		}
		goto out;
	case NFS4ERR_WRONGSEC:
		/* This special return value means that the caller must call
		   nfs4_do_secinfo() and retry the operation. */
		status = -EWRONGSEC;
		goto out;
		
	/* XXX: Need to add handler for ERR_OLD_STATEID in case this is returned on a
	 * READ/WRITE, as a result of {pre,post}_file_data discrepancy in OPEN (see
	 * comment in nfs4_lookup_open())
	 */
#endif
		
	default:
		break;
	}
	
	status = 0;     /* if we make it here, handle_compound() will succeed */
	
out:
	return status;
xdr_error:
	status = -EIO;
	nfsv4_printk(xdr2, "nfs4_handle_compound: xdr error!\n");
	goto out;
}

int
nfs4_handle_simple_op(struct compound4res *rp, nfs_opnum4 op)
{
	u32 *p;
	char buf[8];
	int status;

	nfsv4_printk(xdr, "nfs4_handle_simple_op: called with op=%d\n", op);
	
	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	if (*p++ != htonl(op))
		goto xdr_error;
	if ((status = *p++)) {
		nfsv4_printk(level2, "nfs4_handle_simple_op(%d): nfserr=%d\n",
			     op, ntohl(status));
		status = kernerrno(ntohl(status));
		goto out;
	}

out:
	return status;
xdr_error:
	nfsv4_printk(xdr2, "nfs4_handle_simpleop(%d): xdr error!\n", op);
	status = -EIO;
	goto out;
}

int
nfs4_handle_close(struct compound4res *rp, struct nfs4fs_close *close)
{
	u32 *p;
	struct nfs4fs_file_data *file_data = close->data;
	char buf[16];
	int status;

	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	if (*p++ != close->op)
		goto xdr_error;
	if ((status = *p++)) {
		nfsv4_printk(open2, "nfs4_handle_close: CLOSE/OPEN_DOWNGRADE failed, "
			     "nfserr=%d\n", ntohl(status));
		status = kernerrno(ntohl(status));
		goto out;
	}

	if (!(p = read_buf(rp, sizeof(stateid4), buf)))
		goto xdr_error;
	memcpy(file_data->fi_stateid, p, sizeof(stateid4));
	p += XDR_QUADLEN(sizeof(stateid4));
#if 0
	nfsv4_printk(stateid, "nfs4_handle_close: received stateid 0x%Lx, lockowner %d\n",
		     file_data->fi_stateid, file_data->fi_lockowner->lo_id);
#endif
	file_data->fi_nopen--;
	if (file_data->fi_nopen) {
		file_data->fi_nread -= close->read;
		file_data->fi_nwrite -= close->write;
	}
	else {
		file_data->fi_nread = 0;
		file_data->fi_nwrite = 0;
		list_del_init(&file_data->fi_read_map);
		list_del_init(&file_data->fi_write_map);
	}
	
out:
	return status;
xdr_error:
	nfsv4_printk(xdr2, "nfs4_handle_close: xdr error!\n");
	status = -EIO;
	goto out;
}

int
nfs4_handle_commit(struct compound4res *rp, struct nfs4fs_commit *commit)
{
	u32 *p;
	char buf[8];
	int status;
	
	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	if (*p++ != htonl(OP_COMMIT))
		goto xdr_error;
	if ((status = *p++)) {
		nfsv4_printk(write2, "nfs4_handle_commit: nfserr=%d!\n", ntohl(status));
		status = kernerrno(ntohl(status));
		goto out;
	}

	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	memcpy(commit->verifier, p, 8);
	status = 0;
	
out:
	return status;
xdr_error:
	nfsv4_printk(xdr2, "nfs4_handle_commit: xdr error!\n");
	status = -EIO;
	goto out;
}

int
nfs4_handle_create(struct compound4res *rp, struct nfs4fs_create *create)
{
	struct dentry *dentry = create->dentry;
	u32 *p;
	char buf[20];
	int status;

	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	if (*p++ != htonl(OP_CREATE))
		goto xdr_error;
	if ((status = *p++)) {
		nfsv4_printk(create2, "nfs4_handle_create: nfserr=%d!\n", ntohl(status));
		if (status == htonl(NFS4ERR_EXIST)) {
			/* file already exists? */
			d_invalidate(dentry);  /* XXX : what if it fails? */
			status = -EEXIST;
		}
		else
			status = kernerrno(ntohl(status));
		goto out;
	}

	if (!(p = read_buf(rp, 20, buf)))
		goto xdr_error;
	handle_change_info4(p, dentry->d_parent->d_inode);

	/* XXX: should we return error if these fail, or invalidate the dentry and return
	   success? */
	if (create->type == NF4DIR)
		if ((status = nfs4_handle_setattr(rp)))
			goto out;
	if ((status = nfs4_handle_getattr(rp, &create->getattr)))
		goto out;
	nfs4_handle_getfh(rp, &create->getfh);      /* ignore return value */
	
out:
	return status;
xdr_error:
	nfsv4_printk(xdr2, "nfs4_handle_create: xdr error!\n");
	status = -EIO;
	goto out;
}

/* attributes which are required on a GETATTR.
   XXX : right now, we require a lot of attributes to be returned.. later, think about how
   to relax this requirement. */
#define POSITIVE_MASK	(FA_CHANGEID | FA_TYPE | FA_FSID | FA_FILEID | FA_SIZE)
#define NEGATIVE_MASK	(FA_CHANGEID | FA_TYPE | FA_FSID | FA_FILEID | FA_SIZE | FA_MODE | \
			 FA_NLINK | FA_UID | FA_GID | FA_RDEV | FA_ATIME |  FA_MTIME)
#define MISSING_ATTR(fattr, mask)   (((fattr)->fa_valid & (mask)) != (mask))

static umode_t typemode[] = {
	0,           /* bad */
	S_IFREG,     /* NF4REG */
	S_IFDIR,     /* NF4DIR */
	S_IFBLK,     /* NF4BLK */
	S_IFCHR,     /* NF4CHR */
	S_IFLNK,     /* NF4LNK */
	S_IFSOCK,    /* NF4SOCK */
	S_IFIFO      /* NF4FIFO */
};
#define TYPE_MAX     NF4FIFO

int
nfs4_handle_getattr(struct compound4res *rp, struct nfs4fs_getattr *getattr)
{
	u32 *p;
	char buf[8];
	struct dentry *dentry = getattr->dentry;
	struct inode *inode = dentry->d_inode;
	struct nfs4fs_fattr *fattr = &getattr->fattr;
	unsigned long ino;
	int status;

	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	if (*p++ != htonl(OP_GETATTR))
		goto xdr_error;
	if ((status = *p++)) {
		nfsv4_printk(getattr2, "nfs4_handle_getattr: nfserr=%d\n", ntohl(status));
		status = kernerrno(ntohl(status));
		goto out;
	}

	if ((status = read_attrs(rp, fattr)))
		goto out;
	if (getattr->flags & FL_NOREVAL)
		goto out;

	dentry = getattr->dentry;
	inode = dentry->d_inode;
	nfsv4_printk(reval, "nfs4_handle_getattr(%.*s/%.*s): starting, dentry is %s\n",
		     (int)dentry->d_parent->d_name.len, dentry->d_parent->d_name.name,
		     (int)dentry->d_name.len, dentry->d_name.name,
		     inode ? "positive" : "negative");

	/* Do some sanity checks on the attributes sent. */
	if (MISSING_ATTR(fattr, POSITIVE_MASK)) {
		nfsv4_printk(reval2, "nfs4_handle_getattr: missing attr! (0x%x/0x%x)\n",
			     fattr->fa_valid, POSITIVE_MASK);
		goto out_drop;
	}
	if ((fattr->fa_type <= 0) || (fattr->fa_type > TYPE_MAX)) {
		nfsv4_printk(reval2, "nfs4_handle_getattr: unrecognized type %d!\n",
			     fattr->fa_type);
		goto out_invalidate;
	}
	if (fattr->fa_valid & FA_MODE) {
		/* some servers (e.g. NetApp) don't return the type bits in the mode.. */
		if (!(fattr->fa_mode & S_IFMT))
			fattr->fa_mode |= typemode[fattr->fa_type];
		
		if (typemode[fattr->fa_type] != (fattr->fa_mode & S_IFMT)) {
			nfsv4_printk(reval2, "nfs4_handle_getattr: type-mode mismatch!\n");
			nfsv4_printk(reval2, "(type=%d, mode=0%o)\n", fattr->fa_type,
				     fattr->fa_mode);
			goto out_invalidate;
		}
	}
	else {
		/* some servers don't support UNIX mode bits */
		fattr->fa_mode = typemode[fattr->fa_type] | S_IALLUGO;
		fattr->fa_valid |= FA_MODE;
	}
	
	if (!inode) {
	get_new_inode:
		if (MISSING_ATTR(fattr, NEGATIVE_MASK)) {
			nfsv4_printk(reval2, "nfs4_handle_getattr: missing attr!\n"
				     "fa_valid=0x%x, NEGATIVE_MASK=%0x\n",
				     fattr->fa_valid, NEGATIVE_MASK);
			goto out_drop;
		}
		status = -EIO;
		ino = (unsigned long)fattr->fa_fileid;
		for (;;) {
			if (!(inode = iget4(dentry->d_sb, ino, nfs4_find_actor, fattr))) {
				nfsv4_printk(reval2, "nfs4_handle_getattr: iget4 failed!\n");
				goto out_drop;
			}
			if (!is_bad_inode(inode))
				break;
			nfsv4_printk(reval, "nfs4_handle_getattr: bad inode, retrying\n");
			remove_inode_hash(inode);
			iput(inode);
		}		
		d_instantiate(dentry, inode);
		
		/*
		 * Check whether the mode has been set, as we only want to
		 * do this once. (We don't allow inodes to change types.)
		 */
		if (inode->i_mode == 0) {
			/* XXX: hard to believe, but inode->i_flock = NULL does not appear to
			   be done in the VFS!  need to audit the VFS to see which fields
			   don't come preinitialized.. */
			inode->i_mode = fattr->fa_mode;
			inode->i_size = fattr->fa_size;
			inode->i_flock = NULL;
			NFS4_FSID_MAJOR(inode) = fattr->fa_fsid_major;
			NFS4_FSID_MINOR(inode) = fattr->fa_fsid_minor;
			NFS4_FILEID(inode) = fattr->fa_fileid;
			NFS4_CACHE_JIFFIES(inode) = jiffies;
			NFS4_CACHE_CHANGEID(inode) = fattr->fa_changeid;
			NFS4_CACHE_TIMEOUT(inode) = NFS4_MINTIMEOUT(inode);
			NFS4_CACHE_TIMEOUT_JIFFIES(inode) = jiffies;
			NFS4_DELEGATION(inode) = NULL;
			NFS4_SNAPSHOT(inode) = 0;
			init_MUTEX(&NFS4_LOCKING_SEMAPHORE(inode));
			INIT_LIST_HEAD(&NFS4_READ_MAPPINGS(inode));
			INIT_LIST_HEAD(&NFS4_WRITE_MAPPINGS(inode));
			NFS4_FLUSHD_NEXTSCAN(inode) = jiffies + NFS4_WRITEBACK_LOCKDELAY;
			INIT_LIST_HEAD(&NFS4_FLUSHD_LIST(inode));
			INIT_LIST_HEAD(&NFS4_FLUSHD_READ(inode));
			INIT_LIST_HEAD(&NFS4_FLUSHD_WRITE(inode));
			INIT_LIST_HEAD(&NFS4_FLUSHD_DIRTY(inode));
			INIT_LIST_HEAD(&NFS4_FLUSHD_COMMIT(inode));
			NFS4_FLUSHD_NREAD(inode) = 0;
			NFS4_FLUSHD_NDIRTY(inode) = 0;
			NFS4_FLUSHD_NCOMMIT(inode) = 0;
			NFS4_FLUSHD_DIRTYBYTES(inode) = 0;
			INIT_LIST_HEAD(&NFS4_RECOVERY_LIST(inode));
		
			if (S_ISREG(inode->i_mode)) {
				nfsv4_printk(reval, "nfs4_handle_getattr: new file\n");
				inode->i_op = &nfs4_file_inode_operations;
				inode->i_fop = &nfs4_file_operations;
				inode->i_data.a_ops = &nfs4_file_aops;
			}
			else if (S_ISDIR(inode->i_mode)) {
				nfsv4_printk(reval, "nfs4_handle_getattr: new directory\n");
				inode->i_op = &nfs4_dir_inode_operations;
				inode->i_fop = &nfs4_dir_operations;
			}
			else if (S_ISLNK(inode->i_mode)) {
				nfsv4_printk(reval, "nfs4_handle_getattr: new symlink\n");
				inode->i_op = &nfs4_symlink_inode_operations;
			}
			else {
				nfsv4_printk(reval, "nfs4_handle_getattr: new device\n");
				/* Why so? Because we want revalidate for devices/FIFOs, and
				 * that's precisely what we have in nfs_file_inode_operations.
				 * XXX : is this right?
				 */
				inode->i_op = &nfs4_file_inode_operations;
				/* XXX : is this really what we want?!  setting the rdev # of
				   the inode based on untrusted values pulled down from the
				   server, and turning the inode loose in the VFS, seems like
				   a bad idea!  come back to this and rethink it carefully.. */
				init_special_inode(inode, inode->i_mode,
						   MKDEV(fattr->fa_rdev_major,
							 fattr->fa_rdev_minor));
			}
			goto fill_inode;
		}
		else {
			nfsv4_printk(reval, "nfs4_handle_getattr: instantiating old ino %ld\n",
				     inode->i_ino);
			goto check_inode;
		}
	}

	if ((NFS4_FSID_MAJOR(inode) != fattr->fa_fsid_major) ||
	    (NFS4_FSID_MINOR(inode) != fattr->fa_fsid_minor) ||
	    (NFS4_FILEID(inode) != fattr->fa_fileid)) {
		if (getattr->flags & FL_CLOBBER_OK) {
			d_delete(dentry);
			/* XXX: normally, d_delete() makes the dentry negative.  but if
			   the dentry is in use by someone else, d_delete() will leave the
			   dentry positive and drop it from the dcache. we need to ensure that
			   this race condition does not happen.  for now, we check to make
			   sure that d_delete() worked as expected; if it didn't, we fall
			   through and return an error. */
			if (!dentry->d_inode)
				goto get_new_inode;
			printk("!!! nfs4_handle_getattr: bad lookup clobber !!!\n");
		}
		printk(KERN_INFO "nfs4: inode number mismatch\n"
		       "expected (0x%Lx/0x%Lx/%Ld), got (0x%Lx/0x%Lx/%Ld)\n",
		       (long long)NFS4_FSID_MAJOR(inode),
		       (long long)NFS4_FSID_MINOR(inode),
		       (long long)NFS4_FILEID(inode),
		       (long long)fattr->fa_fsid_major,
		       (long long)fattr->fa_fsid_minor,
		       (long long)fattr->fa_fileid);		
		/* XXX: should we make_bad_inode() here? */
		goto out_drop;
	}
	
	if (is_bad_inode(inode)) {
		/* XXX : how can we prevent this path from being taken? */
		nfsv4_printk(reval2, "nfs4_handle_getattr: bad inode!\n");
		goto out_drop;
	}

	
check_inode:
	/* Make sure the inode's type hasn't changed. */
	if ((inode->i_mode & S_IFMT) != (fattr->fa_mode & S_IFMT)) {
		printk(KERN_INFO "nfs4_refresh_inode: inode type change!\n"
		       "was 0%o, now is 0%o\n", inode->i_mode, fattr->fa_mode);
		goto out_invalidate;
	}
	/* If we make it this far, the call will succeed. */
	/* If inode hasn't changed, exit early. */
	/* XXX: The size comparison ought to be redundant, but we need it because our Linux
	   server cheats and returns a second-resolution mtime in place of the changeid.
	   Eventually, we should be able remove it (and even remove the size attribute from
	   the positive bitmap). */
	if ((fattr->fa_changeid == NFS4_CACHE_CHANGEID(inode)) &&
	    (fattr->fa_size == inode->i_size)) {
		nfsv4_printk(reval, "nfs4_handle_getattr: inode is unchanged\n");
		NFS4_CACHE_JIFFIES(inode) = jiffies;
		if (MISSING_ATTR(fattr, NEGATIVE_MASK))
			goto out;
		if (NFS4_UPDATE_TIMEOUT(inode)) {
			unsigned long timeout = NFS4_CACHE_TIMEOUT(inode);
			unsigned long maxtimeout = NFS4_MAXTIMEOUT(inode);
			timeout <<= 1;
			if (timeout > maxtimeout)
				timeout = maxtimeout;
			NFS4_CACHE_TIMEOUT(inode) = timeout;
			NFS4_CACHE_TIMEOUT_JIFFIES(inode) = jiffies;
			nfsv4_printk(reval, "nfs4_handle_getattr: cache_timeout=%ld\n",
				     timeout);
		}
	}
	else {
		/* Inode has changed; better update accordingly. */
		nfsv4_printk(reval, "nfs4_handle_getattr: inode changed, zapping caches\n");
		NFS4_CACHE_CHANGEID(inode) = fattr->fa_changeid;
		inode->i_size = fattr->fa_size;
		nfs4_zap_caches(inode);

		if (MISSING_ATTR(fattr, NEGATIVE_MASK)) {
			/* no way to bring inode up-to-date */
			nfsv4_printk(reval, "nfs4_handle_getattr: forcing revalidation\n");
			NFS4_CACHE_JIFFIES(inode) = 0;
			goto out;
		}
	}
	
	inode->i_mode = fattr->fa_mode;
	
fill_inode:
	inode->i_nlink = fattr->fa_nlink;
	inode->i_uid = fattr->fa_uid;
	inode->i_gid = fattr->fa_gid;
	inode->i_rdev = (S_ISCHR(inode->i_mode) || S_ISBLK(inode->i_mode)) ?
		MKDEV(fattr->fa_rdev_major, fattr->fa_rdev_minor) : 0;
	inode->i_atime = fattr->fa_atime;
	inode->i_mtime = fattr->fa_mtime;
	inode->i_ctime = fattr->fa_mtime;
	status = 0;
	
out:
	nfsv4_printk(reval, "nfs4_handle_getattr: returning status %d\n", status);
	return status;
out_invalidate:
	if (inode) {
		printk(KERN_INFO "nfs4: invalidating inode\n");
		nfs_invalidate_inode(inode);
	}
out_drop:
	printk(KERN_INFO "nfs4_handle_getattr: dropping dentry\n");
	d_invalidate(dentry);  /* XXX : make dentry negative? what if d_invalidate() fails? */
	nfs4_drop_fh(dentry);
	status = (getattr->flags & FL_NONFATAL) ? 0 : -EIO;
	goto out;
xdr_error:
	nfsv4_printk(xdr2, "nfs4_handle_getattr: xdr error!\n");
	status = -EIO;
	goto out;
}

int
nfs4_handle_getfh(struct compound4res *rp, struct nfs4fs_getfh *getfh)
{
	u32 *p;
	char buf[8];
	struct dentry *dentry = getfh->dentry;
	struct nfs4fs_dentry_data *ddata;
	unsigned int len;
	int status;

	nfsv4_printk(xdr, "nfs4_handle_getfh: starting\n");
	
	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	if (*p++ != htonl(OP_GETFH))
		goto xdr_error;
	if ((status = *p++)) {
		nfsv4_printk(fh2, "nfs4_handle_getfh: nfserr=%d!\n", ntohl(status));
		status = kernerrno(ntohl(status));
		goto out;
	}

	if (!(p = read_buf(rp, 4, buf)))
		goto xdr_error;
	len = ntohl(*p++);
	
	if (len > NFS4_FHSIZE) {
		nfsv4_printk(xdr2, "nfs4_handle_getfh: length too long!\n");
		goto xdr_error;
	}
	if (!(ddata = (struct nfs4fs_dentry_data *)dentry->d_fsdata)) {
		if (!(dentry->d_fsdata = ddata = NFS4_ALLOC(DDATA_SIZE(len))))
			goto out_nomem;
		ddata->fh_allocated = len;
	}
	else if (ddata->fh_allocated < len) {
		NFS4_FREE(ddata);
		if (!(dentry->d_fsdata = ddata = NFS4_ALLOC(DDATA_SIZE(len))))
			goto out_nomem;
		ddata->fh_allocated = len;
	}
	ddata->fh_len = len;
	if (copy_buf(rp, len, ddata->fh_val))
		goto xdr_error;
	nfsv4_printk(fh, "nfs4_handle_getfh(%.*s): got filehandle\n",
		     (int)dentry->d_name.len, dentry->d_name.name);
	nfsv4_hexdump(fh, ddata->fh_val, ddata->fh_len);
	
out:
	return status;
xdr_error:
	nfsv4_printk(xdr2, "nfs4_handle_getfh: xdr error!\n");
	status = -EIO;
	goto out;
out_nomem:
	nfsv4_printk(fh2, "nfs4_handle_getfh: out of memory!\n");
	status = (getfh->nonfatal) ? 0 : -ENOMEM;
	goto out;
}

int
nfs4_handle_link(struct compound4res *rp, struct nfs4fs_link *link)
{
	struct dentry *dentry = link->dentry;
	struct inode *inode = link->inode;
	u32 *p;
	char buf[20];
	int status;

#if NFS4_DEBUG
	if (dentry->d_inode) {
		printk("!!! nfs4_handle_link: dentry not negative !!!\n");
		return -EIO;
	}
#endif
	
	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	if (*p++ != htonl(OP_LINK))
		goto xdr_error;
	if ((status = *p++)) {
		nfsv4_printk(link2, "nfs4_handle_link: LINK failed, nfserr=%d\n",
			     ntohl(status));
		status = kernerrno(ntohl(status));
		goto out;
	}

	if (!(p = read_buf(rp, 20, buf)))
		goto xdr_error;
	handle_change_info4(p, dentry->d_parent->d_inode);

	atomic_inc(&inode->i_count);
	d_instantiate(dentry, inode);
	inode->i_nlink++;
	status = 0;

out:
	return status;
xdr_error:
	nfsv4_printk(xdr2, "nfs4_handle_link: xdr error!\n");
	status = -EIO;
	goto out;
}

/* size of fields following the owner in a LOCK4denied structure */
#if TYPE_IN_LOCK_DENIED
#  define DENIED_LEN     20   /* offset, length, type */
#else
#  define DENIED_LEN     16   /* offset, length */
#endif

int
nfs4_handle_lock(struct compound4res *rp, struct nfs4fs_lock *lock)
{
	struct nfs4fs_file_data *file_data = lock->file_data;
	u32 namelen;
	u32 *p;
	char buf[16];
	int status;
	
	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	if (*p++ != htonl(OP_LOCK))
		goto xdr_error;
	status = *p++;
	if (!status) {
		if (!(p = read_buf(rp, sizeof(stateid4), buf)))
			goto xdr_error;
		memcpy(file_data->fi_stateid, p, sizeof(stateid4));
		p += XDR_QUADLEN(sizeof(stateid4));
#if 0
		nfsv4_printk(stateid, "nfs4_handle_lock: received stateid 0x%Lx "
			     "lockowner %d\n", file_data->fi_stateid,
			     file_data->fi_lockowner->lo_id);
#endif
		goto out;
	}
	if (status == htonl(NFS4ERR_DENIED)) {
		status = -EAGAIN;
		/* skip LOCK4denied struture */
		if (!(p = read_buf(rp, 12, buf)))
			goto xdr_error;
		p += 2;
		namelen = ntohl(*p++);
		if (skip_buf(rp, DENIED_LEN + (XDR_QUADLEN(namelen) << 2)))
			goto xdr_error;
		goto out;
	}
	else if (status == htonl(NFS4ERR_LOCK_RANGE)) {
		printk(KERN_WARNING "nfs4: server returned NFS4ERR_LOCK_RANGE\n");
		status = -EIO;
		goto out;
	}
	nfsv4_printk(lock2, "nfs4_handle_lock: LOCK returned nfserr=%d\n", ntohl(status));
	status = kernerrno(ntohl(status));
	
out:
	return status;
xdr_error:
	nfsv4_printk(lock2, "nfs4_handle_lock: xdr error!\n");
	status = -EIO;
	goto out;
}

int
nfs4_handle_lockt(struct compound4res *rp, struct nfs4fs_lockt *lockt)
{
	struct file_lock *lock = lockt->lock;
	struct super_block *sb = lockt->lockowner->lo_sb;
	clientid4 clientid;
	unsigned int namelen;
	struct nfs4fs_lockowner *lockowner;
	u64 start, length;
	u32 *p;
	char buf[DENIED_LEN + 4];
	int status;

	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	if (*p++ != htonl(OP_LOCKT))
		goto xdr_error;
	status = *p++;
	if (!status) {
		lock->fl_type = F_UNLCK;
		goto out;
	}
	if (status != htonl(NFS4ERR_DENIED)) {
		nfsv4_printk(lock2, "nfs4_handle_lockt: LOCKT returned nfserr=%d\n",
			     ntohl(status));
		status = kernerrno(ntohl(status));
		goto out;
	}

	/* Lock was denied by server; write conflicting lock info into file_lock structure.
	   The VFS expects the following fields to be set: lock->fl_type, lock->fl_pid,
	   lock->fl_start, lock->fl_end. */

	if (!(p = read_buf(rp, 12, buf)))    /* clientid, lockowner length */
		goto xdr_error;
	clientid = (u64)ntohl(*p++) << 32;
	clientid |= ntohl(*p++);
	namelen = ntohl(*p++);
	if (clientid == NFS4_CLIENTID(sb)) {
		if (namelen != 4)
			goto xdr_error;
		if (!(p = read_buf(rp, 4 + DENIED_LEN, buf)))
			goto xdr_error;
		if ((lockowner = nfs4_get_lockowner_by_id(*p++))) {
			lock->fl_pid = lockowner->lo_pid;
			nfs4_put_lockowner(lockowner);
		}
		else {
			printk(KERN_INFO "nfs4: received unknown lockowner id from server\n");
			lock->fl_pid = pidof_rpciod();
		}
	}
	else {
		if (skip_buf(rp, XDR_QUADLEN(namelen) << 2))
			goto xdr_error;
		if (!(p = read_buf(rp, DENIED_LEN, buf)))
			goto xdr_error;
		lock->fl_pid = pidof_rpciod();
	}
	start = (u64)ntohl(*p++) << 32;
	start |= ntohl(*p++);
	length = (u64)ntohl(*p++) << 32;
	length |= ntohl(*p++);
	lock->fl_start = start;
	if (LOFF_OVERFLOW(start, length))
		lock->fl_end = ~(u64)0;
	else
		lock->fl_end = start + length - 1;

#if TYPE_IN_LOCK_DENIED
	if (ntohl(*p++) & 1)      /* READ_LT, READW_LT */
		lock->fl_type = F_RDLCK;
	else
		lock->fl_type = F_WRLCK;
#else
	/* No way to tell... */
	lock->fl_type = F_WRLCK;
#endif
	status = 0;

out:
	return status;
xdr_error:
	nfsv4_printk(xdr2, "nfs4_handle_lockt: xdr error!\n");
	status = -EIO;
	goto out;
}

int
nfs4_handle_locku(struct compound4res *rp, struct nfs4fs_lock *lock)
{
	struct nfs4fs_file_data *file_data = lock->file_data;
	u32 *p;
	char buf[16];
	int status;

	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	if (*p++ != htonl(OP_LOCKU))
		goto xdr_error;
	status = *p++;
	
	if (status) {
		if (status == htonl(NFS4ERR_LOCK_RANGE)) {
			status = -EIO;
			printk(KERN_WARNING "nfs4: server returned NFS4ERR_LOCK_RANGE\n");
			goto out;
		}
		nfsv4_printk(lock2, "nfs4_handle_locku: LOCKU returned nfserr=%d\n",
			     ntohl(status));
		status = kernerrno(ntohl(status));
		goto out;
	}
	if (!(p = read_buf(rp, sizeof(stateid4), buf)))
		goto xdr_error;
	memcpy(file_data->fi_stateid, p, sizeof(stateid4));
	p += XDR_QUADLEN(sizeof(stateid4));
#if 0
	nfsv4_printk(stateid, "nfs4_handle_locku: received stateid 0x%Lx, lockowner %d\n",
		     file_data->fi_stateid, file_data->fi_lockowner->lo_id);
#endif
	
out:
	return status;
xdr_error:
	nfsv4_printk(lock2, "nfs4_handle_lock: xdr error!\n");
	status = -EIO;
	goto out;
}

int
nfs4_handle_lookup(struct compound4res *rp, struct nfs4fs_lookup *lookup)
{
	u32 *p;
	struct dentry *dentry = lookup->dentry;
	struct inode *inode;
	char buf[8];
	int status;

	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	if (*p++ != htonl(OP_LOOKUP))
		goto xdr_error;
	if ((status = *p++)) {
		if (status == htonl(NFS4ERR_NOENT)) {
			nfsv4_printk(lookup, "nfs4_handle_lookup: file does not exist\n");
			inode = dentry->d_inode;
			if (inode)
				nfs4_zap_caches(inode);
			nfs4_drop_fh(dentry);
			d_delete(dentry);
			status = -ENOENT;
			/* XXX: the infamous usage count race condition strikes again */
			if (dentry->d_inode)
				goto out_zap;
			if (lookup->flags & FL_NOENT_OK) {
				status = 0;
				goto done;
			}
		}
		else {
			nfsv4_printk(level2, "nfs4_handle_lookup: nfserr=%d!\n",
				     ntohl(status));
			status = kernerrno(ntohl(status));
		}
		goto out_zap;
	}

	if ((status = nfs4_handle_getattr(rp, &lookup->getattr)))
		goto out;
	if (lookup->did_getfh)
		nfs4_handle_getfh(rp, &lookup->getfh);   /* ignore return value */

done:
	dentry->d_time = jiffies;
out:
	return status;
out_zap:
	nfs4_zap_caches(dentry->d_parent->d_inode);
	goto out;
xdr_error:
	nfsv4_printk(xdr2, "nfs4_handle_lookup: xdr error!\n");
	status = -EIO;
	goto out;
}

int
nfs4_handle_open(struct compound4args *argp, struct compound4res *rp, struct nfs4fs_open *open)
{
	u32 *p;
	u32 rflags;
	char buf[48];
	struct dentry *dentry = open->dentry;
	struct nfs4fs_file_data *file_data = open->pre_file_data;
	stateid4 stateid;
	u32 delegation_type = OPEN_DELEGATE_NONE;
	stateid4 delegation_stateid;
	int delegation_recall;
	int delegation_acelen;
	int status;

#if NFS4_DEBUG
	/* At the top of handle_open(), we should have the seqid semaphore and (if file_data
	 * is non-NULL), the the stateid semaphore in file_data.
	 */
	if (!(argp->flags & CA_SEQID_LOCKED)) {
		printk("!!! handle_open: seqid is not locked !!!\n");
		return -EIO;
	}
	if (file_data && !(argp->flags & CA_STATEID_LOCKED)) {
		printk("!!! handle_open: stateid is not locked !!!\n");
		return -EIO;
	}
#endif
	
	nfsv4_printk(open, "nfs4_handle_open: starting\n");
	open->post_file_data = NULL;
	
	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	if (*p++ != htonl(OP_OPEN))
		goto xdr_error;
	if ((status = *p++)) {
		nfsv4_printk(open2, "nfs4_handle_open: OPEN failed, nfserr=%d\n",
			     ntohl(status));
		status = kernerrno(ntohl(status));
		goto out;
	}

	if (!(p = read_buf(rp, sizeof(stateid4), buf)))
		goto xdr_error;
	memcpy(stateid, p, sizeof(stateid4));
	p += XDR_QUADLEN(sizeof(stateid4));

	if (!(p = read_buf(rp, 20, buf)))   /* changeinfo */
		goto xdr_error;
	handle_change_info4(p, open->dentry->d_parent->d_inode);

	if (!(p = read_buf(rp, 16, buf)))   /* rflags, verf, delegation type */
		goto xdr_error;
	rflags = ntohl(*p++);
	/* XXX : optimize out if OPEN_CONFIRM flag not set? */
	memcpy(&open->verifier, p, 8);
	p += 2;

	delegation_type = ntohl(*p++);
	switch (delegation_type) {
	case OPEN_DELEGATE_NONE:
		break;
	case OPEN_DELEGATE_READ:
		/*
		 * Get stateid, recall flag, and the first four words of the ACE
		 * (this gets the ACE principal's length, but leaves the principal).
		 * XXX: For now, the ACE is ignored.
		 */
		printk("AAA handle_open: OPEN_DELEGATE_READ xdr starting\n");
		if (!(p = read_buf(rp, sizeof(stateid4) + 20, buf)))
			goto xdr_error;
		memcpy(delegation_stateid, p, sizeof(stateid4));
		p += XDR_QUADLEN(sizeof(stateid4));
		delegation_recall = ntohl(*p++);
		p += 3;
		delegation_acelen = ntohl(*p++);

		/*
		 * Get the ACE principal.
		 */
		if (!read_buf(rp, XDR_LEN(delegation_acelen), buf))
			goto xdr_error;
		printk("AAA handle_open: OPEN_DELEGATE_READ xdr done\n");
		break;

	case OPEN_DELEGATE_WRITE:
		/*
		 * Get stateid, recall flag, limit, and the first four words of the ACE
		 * (this gets the ACE principal's length, but leaves the principal).
		 * XXX: For now, the limit and ACE are ignored.
		 */
		printk("AAA handle_open: OPEN_DELEGATE_WRITE xdr starting\n");
		if (!(p = read_buf(rp, sizeof(stateid4) + 32, buf)))
			goto xdr_error;
		memcpy(delegation_stateid, p, sizeof(stateid4));
		p += XDR_QUADLEN(sizeof(stateid4));
		delegation_recall = ntohl(*p++);
		p += 6;
		delegation_acelen = ntohl(*p++);
		
		/*
		 * Get the ACE principal.
		 */
		if (!read_buf(rp, XDR_LEN(delegation_acelen), buf))
			goto xdr_error;
		printk("AAA handle_open: OPEN_DELEGATE_WRITE xdr done\n");
		break;
		
	default:
		goto xdr_error;
	}

	if ((status = nfs4_handle_getattr(rp, &open->getattr))) {
		nfsv4_printk(open, "nfs4_handle_open: nfs4_handle_getattr failed!\n");
		goto out_deleg;
	}
	if ((status = nfs4_handle_getfh(rp, &open->getfh))) {
		nfsv4_printk(open, "nfs4_handle_open: nfs4_handle_getfh failed!\n");
		goto out_deleg;
	}
#if NFS4_DEBUG
	/* XXX: how can we eliminate this test? */
	if (!dentry->d_inode) {
		printk("!!! nfs4_handle_open: dentry still negative !!!\n");
		status = -EIO;
		goto out_deleg;
	}
#endif
	
	/*
	 * If we don't have a file_data structure yet (because this OPEN created the file),
	 * now is the time to get one.  XXX: This is a little inefficient, since
	 * nfs4_get_file_data() will lookup the lockowner, which we already know...
	 * maybe improve later?
	 *
	 * Note: The refcount and semaphore handling here is a little delicate.  Our
	 * convention is that handle_open() always claims a reference for the file_data
	 * stored in ->post_file_data, whether or not this is the same as ->pre_file_data.
	 * Regarding semaphores, we "take over" both the seqid and stateid semaphores by
	 * clearing both _LOCKED flags from argp->flags.  If open_confirm() is called, it
	 * is responsible for releasing the semaphores.  Otherwise, we do it here.
	 */
	if (file_data) {
		if (dentry->d_inode == file_data->fi_inode) {
			atomic_inc(&file_data->fi_count);
			open->post_file_data = file_data;
			goto got_file_data;
		}
		/* oops, grabbed wrong file_data prior to the OPEN */
		nfs4_up_write(&file_data->fi_sema);
	}
	open->post_file_data = file_data = nfs4_get_file_data(dentry);
	status = -ENOMEM;
	if (!file_data) {
		nfsv4_printk(open2, "nfs4_handle_open: couldn't alloc file data!\n");
		goto out_deleg;
	}
	nfs4_down_write(&file_data->fi_sema);
	
got_file_data:
	memcpy(file_data->fi_stateid, stateid, sizeof(stateid4));
	argp->flags &= ~(CA_SEQID_LOCKED | CA_STATEID_LOCKED);       /* "take over" */
	if (!(rflags & OPEN4_RESULT_CONFIRM)) {
		nfsv4_printk(open, "nfs4_handle_open: confirm flag is not set\n");
		nfs4_up_write(&file_data->fi_sema);
		up(&open->lockowner->lo_sema);
		file_data->fi_nopen++;
		if (open->flags & FMODE_READ)
			file_data->fi_nread++;
		if (open->flags & FMODE_WRITE)
			file_data->fi_nwrite++;
	}
	else {
		/*
		 * XXX: Should we return the delegation if OPEN_CONFIRM fails?
		 * If we decide "no", need to change the out_deleg branch to
		 * be conditional on the RESULT_CONFIRM flag.
		 */
		nfsv4_printk(open, "nfs4_handle_open: confirm flag is set\n");
		if ((status = open_confirm(open)))
			goto out_deleg;
	}

	status = 0;
	if (delegation_type != OPEN_DELEGATE_NONE) {
		/*
		 * XXX: For now, we "accept" the delegation, then return
		 * it to the server.  This is a very temporary situation!
		 */
		struct inode *inode = file_data->fi_inode;
		struct nfs4fs_delegation *dp;
		
		printk("AAA handle_open: accepting delegation %p\n", inode);
		nfs4_accept_delegation(inode, delegation_type, delegation_stateid);

#if 0
		printk("AAA handle_open: unlinking delegation %p\n", inode);
		spin_lock(&nfs4_wreq_lock);
		dp = NFS4_DELEGATION(inode);
		if (dp) {
			list_del(&dp->dl_hash);
			NFS4_DELEGATION(inode) = NULL;
		}
		spin_unlock(&nfs4_wreq_lock);
		
		if (dp) {
			printk("AAA handle_open: returning delegation\n");
			nfs4_do_delegreturn(inode->i_sb, dp->dl_stateid);
			NFS4_FREE(dp);
		}
		else
			printk("AAA handle_open: oops! no delegation to return?!\n");
#endif
	}
	
out:
	nfsv4_printk(open, "nfs4_handle_open: returning status %d\n", status);
	return status;
out_deleg:
	if (delegation_type != OPEN_DELEGATE_NONE)
		nfs4_do_delegreturn(open->lockowner->lo_sb, delegation_stateid);
	goto out;
xdr_error:
	nfsv4_printk(xdr2, "nfs4_handle_open: xdr error!\n");
	status = -EIO;
	goto out;
}

int
nfs4_handle_open_confirm(struct compound4res *rp, struct nfs4fs_open *open)
{
	u32 *p;
	struct nfs4fs_file_data *file_data = open->post_file_data;
	char buf[16];
	int status;

	nfsv4_printk(open, "nfs4_handle_open_confirm: starting\n");
	
	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	if (*p++ != htonl(OP_OPEN_CONFIRM))
		goto xdr_error;
	if ((status = *p++)) {
		nfsv4_printk(open2, "nfs4_handle_open_confirm: OPEN_CONFIRM failed, "
			     "nfserr=%d\n", ntohl(status));
		status = kernerrno(ntohl(status));
		goto out;
	}

	if (!(p = read_buf(rp, sizeof(stateid4), buf)))
		goto xdr_error;

	/* The OPEN is finished; update the stateid and open counts */
	memcpy(file_data->fi_stateid, p, sizeof(stateid4));
	p += XDR_QUADLEN(sizeof(stateid4));
#if 0
	nfsv4_printk(stateid, "nfs4_handle_open_confirm: received stateid 0x%Lx\n",
		     file_data->fi_stateid);
#endif
	file_data->fi_nopen++;
	if (open->flags & FMODE_READ)
		file_data->fi_nread++;
	if (open->flags & FMODE_WRITE)
		file_data->fi_nwrite++;
	status = 0;
	
out:
	nfsv4_printk(open, "nfs4_handle_open_confirm: returning status %d\n", status);
	return status;
xdr_error:
	nfsv4_printk(xdr2, "nfs4_handle_open_confirm: xdr error!\n");
	status = -EIO;
	goto out;
}

int
nfs4_handle_putfh(struct compound4res *rp, struct nfs4fs_putfh *putfh)
{
	u32 *p;
	char buf[8];
	int i;
	int status;

	/* XXX: need recovery code for NFS4ERR_STALE_FH */
	
	nfsv4_printk(xdr, "nfs4_handle_putfh: called with op=%d, nlookups=%d\n", putfh->op,
		     putfh->nlookups);
	
	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	if (*p++ != htonl(putfh->op))
		goto xdr_error;
	if ((status = *p++)) {
		nfsv4_printk(xdr2, "nfs4_handle_putfh: nfserr=%d!\n", ntohl(status));
		status = kernerrno(ntohl(status));
		goto out;
	}

	for (i = 0; i < putfh->nlookups; i++) {
		if (!(p = read_buf(rp, 8, buf)))
			goto xdr_error;
		if (*p++ != htonl(OP_LOOKUP))
			goto xdr_error;
		if ((status = *p++)) {
			nfsv4_printk(xdr2, "nfs4_handle_putfh: nfserr=%d!\n", ntohl(status));
			status = kernerrno(ntohl(status));
			goto out;
		}
	}

out:
	return status;
xdr_error:
	nfsv4_printk(xdr2, "nfs4_handle_putfh: xdr error!\n");
	status = -EIO;
	goto out;
}

int
nfs4_handle_simple_read(struct compound4res *rp, struct nfs4fs_simple_read *read)
{
	u32 *p;
	char buf[8];
	u32 bytes_read;
	int status;
	
	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	if (*p++ != htonl(OP_READ))
		goto xdr_error;
	if ((status = *p++)) {
		nfsv4_printk(read2, "nfs4_handle_read: READ failed, nfserr=%d\n",
			     ntohl(status));
		status = kernerrno(ntohl(status));
		goto out;
	}
	
	if (!(p = read_buf(rp, 8, buf)))
		goto out;
	read->eof = ntohl(*p++);
	bytes_read = read->bytes_read = ntohl(*p++);
	if (bytes_read > read->length)
		goto xdr_error;
	if (copy_buf(rp, bytes_read, read->buf))
		goto xdr_error;

out:
	return status;
xdr_error:
	nfsv4_printk(xdr2, "nfs4_handle_read: xdr error!\n");
	status = -EIO;
	goto out;
}

int
nfs4_handle_read(struct compound4res *rp, struct nfs4fs_read *read)
{
	u32 *p;
	char buf[8];
	int status;
	u32 bytes_read;
	struct nfs4_page *req;
	int i;
	
	nfsv4_printk(xdr, "nfs4_handle_read: starting\n");

	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	if (*p++ != htonl(OP_READ))
		goto xdr_error;
	if ((status = *p++)) {
		nfsv4_printk(read2, "nfs4_handle_read: READ failed, nfserr=%d\n",
			     ntohl(status));
		status = kernerrno(ntohl(status));
		goto out;
	}
	
	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	read->eof = ntohl(*p++);
	bytes_read = read->bytes_read = ntohl(*p++);
	if (bytes_read > read->npages * PAGE_CACHE_SIZE)
		goto xdr_error;

	i = 0;
	while (bytes_read > 0) {
		req = read->pages[i++];
		nfs4_kmap_request(req);
		if (bytes_read < PAGE_CACHE_SIZE) {
			/* XXX: zero out rest of page here?? */
			status = copy_buf(rp, bytes_read, req->wb_buf);
			bytes_read = 0;
		}
		else {
			status = copy_buf(rp, PAGE_CACHE_SIZE, req->wb_buf);
			bytes_read -= PAGE_CACHE_SIZE;
		}
		nfs4_kunmap_request(req);
		if (status)
			goto xdr_error;
	}
	
	nfsv4_printk(xdr, "nfs4_handle_read: eof=%d, bytes_read=%d\n", read->eof,
		     read->bytes_read);

out:
	return status;
xdr_error:
	nfsv4_printk(xdr2, "nfs4_handle_read: xdr error!\n");
	status = -EIO;
	goto out;
}

int
nfs4_handle_readdir(struct compound4res *rp, struct nfs4fs_readdir *readdir)
{
	static unsigned int dt_array[] = {
		DT_UNKNOWN,
		DT_REG,
		DT_DIR,
		DT_BLK,
		DT_CHR,
		DT_LNK,
		DT_SOCK,
		DT_FIFO
	};

	u32 *p;
	u32 bmlen;
	u32 bmval;
	char sbuf[16];                /* Backup buffer for read_buf() */
	char *buf = readdir->buf;     /* Current position in page */
	struct page *page;
	unsigned long index;
	int namelen;
	int expected_attrlen;
	int got_type;
	unsigned int dtype;
	u64 fileid;
	int status;
#if NFS4_DEBUG
	char *name;
#endif

	nfsv4_printk(readdir, "nfs4_handle_readdir: starting, page index=%d\n",
		     (int)readdir->page->index);
	readdir->bad_verifier = 0;
	
	if (!(p = read_buf(rp, 8, sbuf)))
		goto xdr_error;
	if (*p++ != htonl(OP_READDIR))
		goto xdr_error;
	if ((status = *p++)) {
		nfsv4_printk(readdir2, "nfs4_handle_readdir: nfserr=%d!\n", ntohl(status));
		if (status == NFS4ERR_STALE) {
			readdir->bad_verifier = 1;
			status = -EIO;
			goto out;
		}
		status = kernerrno(ntohl(status));
		goto out;
	}

	/* Read verifier, first entry pointer */
	if (!(p = read_buf(rp, 12, sbuf)))
		goto xdr_error;
	memcpy(p, readdir->verifier, 8);
	p += 2;
	
	while (*p++) {   /* loop invariant: at top, p points to next entry4 pointer. */
		nfsv4_printk(xdr, "nfs4_handle_readdir: reading entry\n");
		
		/* Read cookie and entry name length.  */
		if (!(p = read_buf(rp, 12, sbuf)))
			goto xdr_error;
		memcpy(&readdir->cookie, p, 8);   /* cookie is kept in network byte order */
		p += 2;
		namelen = ntohl(*p++);
		/* Note: the space taken up by this entry will be
		   1(next) + 4(namelen) + namelen + 4(type) + 8(fileid) + 1(next)
		   = 18+namelen bytes. */
		
		/* XXX : currently, we return -ENAMETOOLONG if a single entry is too big
		   to fit on a page.  Need to fix this! */
		status = -ENAMETOOLONG;
		if (18 + namelen > PAGE_SIZE) {
			nfsv4_printk(readdir2, "nfs4_handle_readdir: namelen=%d, too long\n",
				     namelen);
			goto out;
		}

		/* If there is not enough room on the page, get the next page. */
		if ((buf + (18 + namelen)) > readdir->end) {
			*buf++ = 1;    /* set "next" field */
			page = readdir->page;
			kunmap(page);
			UnlockPage(page);
			put_page(page);

			/* XXX: what if PAGE_CACHE_SIZE != PAGE_SIZE? */
			index = page->index + 1;
			status = -ENOMEM;   /* XXX: right error? */
			nfsv4_printk(readdir, "nfs4_handle_readdir: advancing to page %ld\n",
				     index);
			if (!(page = grab_cache_page(readdir->inode->i_mapping, index)))
				goto out;
			buf = kmap(page);
			if (!buf) {
				UnlockPage(page);
				put_page(page);
				readdir->page = NULL;
				goto out;
			}
			readdir->page = page;
			readdir->buf = buf;
			readdir->end = buf + PAGE_CACHE_SIZE;
		}
		*buf++ = 2;    /* set "next" field */

		/* Copy entry name length and name into page.  XXX: watch out!  I think that
		   some "filldir" functions may expect a null-terminated name.. */
		*(u32 *)buf = namelen;
		buf += 4;
#if NFS4_DEBUG
		name = buf;
#endif
		if (copy_buf(rp, namelen, buf))
			goto xdr_error;
		buf += namelen;

		/* Read bitmap length */
		if (!(p = read_buf(rp, 4, sbuf)))
			goto xdr_error;
		bmlen = ntohl(*p++);
		if (bmlen > 2)          /* XXX: for now, we consider this an error */
			goto xdr_error;
		
		/* Read bitmap, attrlen.  XXX: no support yet for RDATTR_ERROR attribute! */
		if (!(p = read_buf(rp, (bmlen+1) << 2, sbuf)))
			goto xdr_error;
		expected_attrlen = 0;
		got_type = 0;
		if (!bmlen)
			goto bitmap_done;
		bmval = ntohl(*p++);
		if (bmval & 0xffeffffd) {
			nfsv4_printk(xdr2, "nfs4_handle_readdir: bad bmval[0]!\n");
			goto xdr_error;
		}
		if (bmval & 0x00000002) {    /* FATTR4_TYPE is set */
			got_type = 1;
			expected_attrlen += 4;
		}
		if (bmval & 0x00100000)      /* FATTR4_FILEID is set */
			expected_attrlen += 8;
		else {
			/* XXX: for now, this is considered to be an error */
			printk("!!! nfs4_handle_readdir: no fileid !!!\n");
			goto xdr_error;
		}
		if (bmlen == 1)
			goto bitmap_done;
		if (*p++) {
			nfsv4_printk(xdr2, "nfs4_handle_readdir: bad bmval[1]!\n");
			goto xdr_error;
		}

	bitmap_done:
		if (ntohl(*p++) != expected_attrlen) {
			nfsv4_printk(readdir2, "nfs4_handle_readdir: bad attrlen!\n");
			goto xdr_error;
		}

		/* Read attribute buf and nextentry (to restart the loop). */
		if (!(p = read_buf(rp, expected_attrlen + 4, sbuf))) {
			nfsv4_printk(xdr2, "nfs4_handle_readdir: couldn't read attrs!\n");
			goto xdr_error;
		}
		if (got_type) {
			dtype = ntohl(*p++);
			dtype = ((dtype < NF4REG) || (dtype > NF4FIFO)) ? DT_UNKNOWN :
				dt_array[dtype];
		}
		else
			dtype = DT_UNKNOWN;
		*(u32 *)buf = dtype;          /* Set "type" in page cache */
		buf += 4;

		fileid = (u64)ntohl(*p++) << 32;
		fileid |= ntohl(*p++);
		*(u64 *)buf = fileid;    /* Set "ino" in page cache */
		buf += 8;
		readdir->buf = buf;
		nfsv4_printk(readdir, "nfs4_handle_readdir: decoded \"%.*s\", type=%d, "
			     "fileid=%Ld, cookie=0x%Lx\n", (int)namelen, name, dtype, fileid,
			     readdir->cookie);
	}

	if (!(p = read_buf(rp, 4, sbuf)))
		goto xdr_error;
	readdir->eof = ntohl(*p++);
	status = 0;

out:
	nfsv4_printk(readdir, "nfs4_handle_readdir: returning status %d\n", status);
	return status;
xdr_error:
	nfsv4_printk(readdir2, "nfs4_handle_readdir: XDR error!\n");
	status = -EIO;
	goto out;
}

int
nfs4_handle_readlink(struct compound4res *rp, struct nfs4fs_readlink *readlink)
{
	u32 *p;
	int linklen;
	char buf[8];
	int status;
	
	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	if (*p++ != htonl(OP_READLINK))
		goto xdr_error;
	if ((status = *p++)) {
		nfsv4_printk(readlink2, "nfs4_handle_readlink: LINK failed, nfserr=%d\n",
			     ntohl(status));
		status = kernerrno(ntohl(status));
		goto out;
	}
	
	if (!(p = read_buf(rp, 4, buf)))
		goto xdr_error;
	linklen = ntohl(*p++);
	if (linklen <= 0)
		goto xdr_error;
	if (linklen <= readlink->nbytes) {
		if ((status = copy_buf(rp, linklen, readlink->buf)))
			goto out;
		readlink->buf[linklen] = 0;            /* null-terminate */
	}
	else {
		/* XXX : do we really want to silently truncate, or should we return
		   -ENAMETOOLONG?  Maybe make this a mount option? */
		if ((status = copy_buf(rp, readlink->nbytes, readlink->buf)))
			goto out;
		readlink->buf[readlink->nbytes] = 0;   /* null-terminate */
		/* advance pointer past remaining bytes.  Note that XDR baggage from
		   readlink->nbytes has already been skipped. */
		if ((status = skip_buf(rp, linklen - XDR_LEN(readlink->nbytes))))
			goto out;
	}
	
out:
	return status;
xdr_error:
	nfsv4_printk(xdr2, "nfs4_handle_readlink: xdr error!\n");
	status = -EIO;
	goto out;
}

int
nfs4_handle_remove(struct compound4res *rp, struct inode *dir_inode)
{
	u32 *p;
	char buf[20];
	int status;

	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	if (*p++ != htonl(OP_REMOVE))
		goto xdr_error;
	if ((status = *p++)) {
		nfsv4_printk(remove2, "nfs4_handle_remove: REMOVE returned nfserr=%d\n",
			     ntohl(status));
		status = kernerrno(ntohl(status));
		goto out;
	}
	
	if (!(p = read_buf(rp, 20, buf)))
		goto xdr_error;
	if (dir_inode)
		handle_change_info4(p, dir_inode);
	status = 0;
	
out:
	return status;
xdr_error:
	nfsv4_printk(level2, "nfs4_handle_remove: xdr error!\n");
	status = -EIO;
	goto out;
}

int
nfs4_handle_rename(struct compound4res *rp, struct nfs4fs_rename *rename)
{
	u32 *p;
	struct dentry *old_dir_dentry = rename->old_dentry->d_parent;
	struct dentry *new_dir_dentry = rename->new_dentry->d_parent;
	char buf[20];
	int status;

	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	if (*p++ != htonl(OP_RENAME))
		goto xdr_error;
	if ((status = *p++)) {
		nfsv4_printk(remove2, "nfs4_handle_remove: RENAME returned nfserr=%d\n",
			     ntohl(status));
		status = kernerrno(ntohl(status));
		goto out;
	}
	
	if (!(p = read_buf(rp, 20, buf)))
		goto xdr_error;
	handle_change_info4(p, old_dir_dentry->d_inode);

	if (old_dir_dentry != new_dir_dentry) {
		if (!(p = read_buf(rp, 20, buf)))
			goto xdr_error;
		handle_change_info4(p, new_dir_dentry->d_inode);
	}
	else
		if (skip_buf(rp, 20))
			goto xdr_error;
	status = 0;

out:
	return status;
xdr_error:
	nfsv4_printk(level2, "nfs4_handle_remove: xdr error!\n");
	status = -EIO;
	goto out;
}

int
nfs4_handle_setattr(struct compound4res *rp)
{
	u32 *p;
	char buf[12];
	u32 bmlen;
	int status;

	/* Note: the SETATTR op returns a bitmap in the response even if it fails. */

	nfsv4_printk(xdr, "nfs4_handle_setattr: starting\n");
	
	if (!(p = read_buf(rp, 12, buf)))
		goto xdr_error;
	if (*p++ != htonl(OP_SETATTR))
		goto xdr_error;
	status = *p++;
	bmlen = ntohl(*p++);
	if (bmlen > 2)      /* XXX : for now, this is an error */
		goto xdr_error;
	if (skip_buf(rp, bmlen << 2))
		goto xdr_error;
	if (status) {
		nfsv4_printk(setattr2, "nfs4_handle_setattr: nfserr=%d!\n", ntohl(status));
		status = kernerrno(ntohl(status));
		goto out;
	}

out:
	nfsv4_printk(xdr, "nfs4_handle_setattr: returning status %d\n", status);
	return status;
xdr_error:
	nfsv4_printk(xdr2, "nfs4_handle_setattr: xdr error!\n");
	status = -EIO;
	goto out;
}

int
nfs4_handle_secinfo(struct compound4res *rp, struct nfs4fs_secinfo *secinfo)
{
	u32 *p;
	char buf[16];
	int len,flavorlen,i,flavor=0;
	struct super_block *sb = secinfo->sb;
	int status;

	nfsv4_printk(level2, "nfs4_handle_secinfo\n");
	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	if (*p++ != ntohl(OP_SECINFO))
		goto xdr_error;
	status = *p++;
	if (status) {
		nfsv4_printk(secinfo, "nfs4_handle_secinfo: SECINFO failed, "
			     "nfserr=%d\n", ntohl(status));
		status = kernerrno(ntohl(status));
		goto out;
	}
	if (!(p = read_buf(rp, 4, buf)))
		goto xdr_error;
	len = ntohl(*p++);
  nfsv4_printk(secinfo,"nfs4_handle_secinfo num flavors %d\n",len);
	for(i=0;i<len;i++){
		if (!(p = read_buf(rp, 4, buf)))
			goto xdr_error;
		flavor = ntohl(*p++);
    switch(flavor){
    case RPC_AUTH_NULL:
      nfsv4_printk(secinfo,"nfs4_handle_secinfo RPC_AUTH_NULL flavor \n");
      break;
    case RPC_AUTH_UNIX:
      nfsv4_printk(secinfo,"nfs4_handle_secinfo RPC_AUTH_UNIX flavor \n");
      break;
    case RPC_AUTH_GSS:
      nfsv4_printk(secinfo,"nfs4_handle_secinfo RPC_AUTH_GSS flavor \n");
		/* opaque data - not xdr'ed */
		/* sec_oid4 */
			if (!(p = read_buf(rp, 4, buf)))
				goto xdr_error;
			flavorlen = ntohl(*p++);
      nfsv4_printk(secinfo,"nfs4_handle_secinfo flavorlen %d \n",flavorlen);

		/* for now, don't parse, use hardcoded default so skip oid ,qop4, 
		 * and service.  eventually, we will compare the QOP values 
		 * to the static client state struct that
		 * contains the QOP supported by this client.
		 */
			if (!(p = read_buf(rp, flavorlen, buf)))
				goto xdr_error;
			p+= flavorlen;

      nfsv4_printk(secinfo,"nfs4_handle_secinfo rpcsec_gss_info skipped \n");
      break;
    default:
      nfsv4_printk(secinfo,"NFSV4: do_secinfo INVALID flavor %d\n",flavor);      break;
    }
  }
/* XXX  the following  assumes the hard_coded default qop */
  if (NFS4_CLNT(sb) != NULL &&
      NFS4_CLNT(sb)->cl_auth->au_ops->au_flavor != flavor) {
    nfsv4_printk(secinfo, "nfs4_handle_secinfo: changing to "
           "flavor %d\n", flavor);
    /* Set the mech, service, QOP, based on negotiated secinfo. */
    rpcauth_destroy(NFS4_CLNT(sb)->cl_auth);
    rpcauth_create(flavor, NFS4_CLNT(sb));
	}

out:
	return status;
xdr_error:
	nfsv4_printk(xdr2, "nfs4_handle_secinfo: xdr error!\n");
	status = -EIO;
	goto out;
}

int
nfs4_handle_setclientid(struct compound4res *rp, struct nfs4fs_setclientid *setclientid)
{
	struct compound4args cargs;
	struct compound4res cres;
	u32 *p;
	char buf[16];
	int len;
	int status;

	setclientid->in_use = 0;
	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	if (*p++ != htonl(OP_SETCLIENTID))
		goto xdr_error;
	status = *p++;
	if (status == htonl(NFS4ERR_CLID_INUSE)) {
		if (!(p = read_buf(rp, 4, buf)))     /* skip first string in clientaddr */
			goto xdr_error;
		len = XDR_LEN(ntohl(*p++));
		if (skip_buf(rp, len))
			goto xdr_error;
		if (!(p = read_buf(rp, 4, buf)))     /* skip second string */
			goto xdr_error;
		len = XDR_LEN(ntohl(*p++));
		if (skip_buf(rp, len))
			goto xdr_error;
		setclientid->in_use = 1;
		status = -EIO;
		goto out;
	}
	if (status) {
		nfsv4_printk(clientid2, "nfs4_handle_setclientid: SETCLIENTID failed, "
			     "nfserr=%d\n", ntohl(status));
		status = kernerrno(ntohl(status));
		goto out;
	}
	
	if (!(p = read_buf(rp, 16, buf)))
		goto xdr_error;
	setclientid->clientid = (u64)htonl(*p++) << 32;
	setclientid->clientid |= (u64)htonl(*p++);

	cargs.sb = setclientid->sb;
	cargs.tag = "handle_setclientid()";
	if ((status = nfs4_setup_compound(&cargs)))
		goto out;
	if ((status = nfs4_setup_setclientid_confirm(&cargs, (char *)p)))
		goto out_args;
	if ((status = nfs4_call_compound(&cargs, &cres,(struct nfs4fs_ctx *)NULL)))
		goto out_args;
	if ((status = nfs4_handle_setclientid_confirm(&cres, setclientid)))
		goto out_res;
	p += 2;
	
out_res:
	nfs4_release_compoundres(&cres);
out_args:
	nfs4_release_compoundargs(&cargs);
out:
	return status;
xdr_error:
	nfsv4_printk(xdr2, "nfs4_handle_setclientid: xdr error!\n");
	status = -EIO;
	goto out;
}

int
nfs4_handle_setclientid_confirm(struct compound4res *rp,
				struct nfs4fs_setclientid *setclientid)
{
	u32 *p;
	char buf[8];
	int status;

	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	if (*p++ != htonl(OP_SETCLIENTID_CONFIRM))
		goto xdr_error;
	if ((status = *p++)) {
		if (status == NFS4ERR_CLID_INUSE) {
			nfsv4_printk(clientid2, "nfs4_handle_setclientid: clid in use!\n");
			setclientid->in_use = 1;
			status = -EIO;
			goto out;
		}
		nfsv4_printk(clientid2, "nfs4_handle_setclientid: nfserr=%d!\n",
			     ntohl(status));
		status = kernerrno(ntohl(status));
	}
	
out:
	return status;
xdr_error:
	nfsv4_printk(xdr2, "nfs4_handle_setclientid_confirm: xdr error!\n");
	status = -EIO;
	goto out;
}

int
nfs4_handle_simple_write(struct compound4res *rp, struct nfs4fs_simple_write *write)
{
	u32 how;
	u32 *p;
	char buf[8];
	int status;

	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	if (*p++ != htonl(OP_WRITE))
		goto xdr_error;
	if ((status = *p++)) {
		nfsv4_printk(write2, "nfs4_handle_write: WRITE failed, nfserr=%d\n",
			     ntohl(status));
		status = kernerrno(ntohl(status));
		goto out;
	}

	if (!(p = read_buf(rp, 8, buf)))
		goto xdr_error;
	write->bytes_written = ntohl(*p++);
	if (write->bytes_written > write->length)
		goto xdr_error;
	how = ntohl(*p++);
	if (how > FILE_SYNC4)
		goto xdr_error;
	if (how == UNSTABLE4) {
		static unsigned long  complain;
		if (time_before(complain, jiffies)) {
			printk(KERN_WARNING "nfs4: server failed to commit data!\n");
			complain = jiffies + 300 * HZ;
		}
	}

	/* ignore verifier */
	if (skip_buf(rp, 8))
		goto xdr_error;
	
out:
	return status;
xdr_error:
	nfsv4_printk(xdr2, "nfs4_handle_write: xdr error!\n");
	status = -EIO;
	goto out;
}

int
nfs4_handle_write(struct compound4res *rp, struct nfs4fs_write *write)
{
	struct nfs4_page *req;
	u32 bytes_written;
	u32 *p;
	char buf[20];
	int i;
	int status = 0;

	for (i = 0; i < write->npages; i++) {
		req = write->pages[i];
		if (!(p = read_buf(rp, 8, buf)))
			goto xdr_error;
		if (*p++ != htonl(OP_WRITE))
			goto xdr_error;
		if ((status = *p++)) {
			nfsv4_printk(write2, "nfs4_handle_write: WRITE failed, nfserr=%d\n",
				     ntohl(status));
			status = kernerrno(ntohl(status));
			goto out;
		}

		if (!(p = read_buf(rp, 16, buf)))
			goto xdr_error;
		bytes_written = ntohl(*p++);
		if (bytes_written > req->wb_bytes)
			goto xdr_error;
		if (bytes_written < req->wb_bytes) {
			static unsigned long complain;
			if (time_before(complain, jiffies)) {
				printk(KERN_WARNING
				       "nfs4: server wrote less than requested!\n");
				complain = jiffies + 300 * HZ;
			}
			/* Can't do anything about it right now except throw an error. */
			status = -EIO;
			goto out;
		}
		
		req->wb_how_written = ntohl(*p++);
		if (req->wb_how_written > FILE_SYNC4)
			goto xdr_error;
		if (write->how != UNSTABLE4 && req->wb_how_written == UNSTABLE4) {
			static unsigned long  complain;

			if (time_before(complain, jiffies)) {
				printk(KERN_WARNING
				       "nfs4: server failed to commit data!\n");
				complain = jiffies + 300 * HZ;
			}
		}

		memcpy(req->wb_verifier, p, 8);
	}
	
out:
	return status;
xdr_error:
	nfsv4_printk(xdr2, "nfs4_handle_write: xdr error!\n");
	status = -EIO;
	goto out;
}

static int
open_confirm(struct nfs4fs_open *open)
{
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_putfh putfh;
	int status;

#if NFS4_DEBUG
	if (!open->lockowner || !open->post_file_data) {
		printk("!!! open_confirm(): bad NULL pointer !!!\n");
		return -EIO;
	}
#endif
	
	/* The seqid and stateid semaphores in open->{lockowner,post_file_data} have already
	 * been claimed; we are responsible for releasing them.
	 */
	nfsv4_printk(open, "open_confirm: starting\n");
	putfh.dentry = open->dentry;
	cargs.sb = open->dentry->d_sb;
	cargs.tag = "open_confirm()";
	
	if ((status = nfs4_setup_compound(&cargs)))
		goto out_release;
	if ((status = nfs4_setup_putfh(&cargs, &putfh)))
		goto out_args_release;
	if ((status = nfs4_setup_open_confirm(&cargs, open)))
		goto out_args_release;

	status = nfs4_call_compound(&cargs, &cres,(struct nfs4fs_ctx *)NULL);
	if (status == -EREBOOT) {
		/* XXX: Need to wait for recoverd, then check whether it succeeded in
		 * renewing the OPEN.
		 */
		status = -EIO;
		printk("!!! open_confirm(): server reboot, can't handle this case yet !!!\n");
		goto out_args;
	}
	if (status)
		goto out_args;
	if ((status = nfs4_handle_putfh(&cres, &putfh)))
		goto out_res;
	if ((status = nfs4_handle_open_confirm(&cres, open)))
		goto out_res;

out_res:
	nfs4_release_compoundres(&cres);
out_args:
	nfs4_release_compoundargs(&cargs);
out:
	nfsv4_printk(open, "open_confirm: returning status %d\n", status);
	return status;
out_args_release:
	nfs4_release_compoundargs(&cargs);
out_release:
	up(&open->lockowner->lo_sema);
	nfs4_up_write(&open->post_file_data->fi_sema);
	goto out;
}

#if NFS4_DEBUG
void
debug_dump(struct compound4res *rp, int nbytes)
{
	int curr = rp->iov_current;
	char *p = rp->p;
	char *end = rp->end;
	int buflen;
		
	for (;;) {
		buflen = end - p;
		if (buflen >= nbytes) {
			hex_dump(p, nbytes);
			break;
		}
		hex_dump(p, buflen);
		if (curr == rp->iov_nr-1)
			break;
		curr++;
		nbytes -= buflen;
		p = rp->iov[curr].iov_base;
		end = rp->p + rp->iov[curr].iov_len;
	}
}
#endif
