/*
 * linux/fs/nfs4fs/dir.c
 *
 * Andy Adamson  <andros@umich.edu>
 * Kendrick Smith  <kmsmith@umich.edu>
 * Copyright (C) 2001 The Regents of the University of Michigan
 *
 */
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/mm.h>
#include <linux/pagemap.h>
#include <linux/fs.h>

#include <linux/nfs4/nfs4_prot.h>
#include <linux/nfs4/nfs4_debug.h>
#include <linux/nfs4/nfs4.h>
#include <linux/nfs4/nfs4fs.h>
#ifdef NFS4ACL
#include <linux/acl_ea.h>
#endif

/* XXX : one important thing that needs to be done is to understand all semaphores used by
   the dcache, and make sure we're using them consistently.  also, audit use of lock_kernel();
   we inherited a bunch of lock_kernel()'s from the v3 code.. are they really needed?  */

/* XXX : If an rpc reply is shorter than the allocated buffer size
   (NFS4_BUFSIZE+RPC_SLACK_SPACE), it will silently be truncated, causing the RPC decode to
   fail.  We should always be able to guarantee that, for any COMPOUND we send, the reply
   is guaranteed to be smaller than this.  How can we do this in general? */

/* XXX : we need a better way to guarantee that dentry filehandles will persist across
   blocking kmalloc()'s */

int nfs4_dentry_revalidate(struct dentry *dentry, int flags);
static void nfs4_dentry_release(struct dentry *dentry);
static struct dentry *nfs4_lookup(struct inode *dir_inode, struct dentry *dentry);
static int nfs4_readdir(struct file *filp, void *buf, filldir_t filler);
static int nfs4_link(struct dentry *old_dentry, struct inode *dir_inode,
		     struct dentry *dentry);
static int nfs4_symlink(struct inode *dir_inode, struct dentry *dentry, const char *name);
static int nfs4_unlink(struct inode *dir_inode, struct dentry *dentry);
static int nfs4_mkdir(struct inode *dir_i, struct dentry *dentry, int mode);
static int nfs4_rename(struct inode *old_dir_inode, struct dentry *old_dentry,
		       struct inode *new_dir_inode, struct dentry *new_dentry);

struct dentry_operations nfs4_dentry_operations = {
	d_revalidate: nfs4_dentry_revalidate,
	d_release: nfs4_dentry_release,
	d_iput: nfs4_silly_dentry_iput
};

struct inode_operations nfs4_dir_inode_operations = {
	create: nfs4_create,
	lookup: nfs4_lookup,
	revalidate: nfs4_inode_revalidate,
	link: nfs4_link,
	unlink: nfs4_unlink,
	symlink: nfs4_symlink,
	mkdir: nfs4_mkdir,
	rmdir: nfs4_remove,
	rename: nfs4_rename,
	setattr: nfs4_setattr,
	lookup_open: nfs4_lookup_open,
#ifdef NFS4ACL
	permission: nfs4_posix_acl_permission,
	get_posix_acl: nfs4_acl_get_inode,
	set_posix_acl: nfs4_acl_set_inode,
	get_ext_attr: nfs4_get_ext_attr,
	set_ext_attr: nfs4_set_ext_attr
#endif
	/* XXX : also need mknod, permission */
	/* XXX : what's the deal with the getattr op?  as near as i can tell, this is not
	   used anywhere! */
	
	
};

struct file_operations nfs4_dir_operations = {
	/* XXX : set owner field? */
	read: generic_read_dir,
	readdir: nfs4_readdir,
	/* XXX : need open, release? */
};

/* Code common to nfs4_dentry_revalidate(), nfs4_lookup() */
int
nfs4fs_do_lookup(struct dentry *dentry)
{
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_putfh putfh;
	struct nfs4fs_getattr reval;
	struct nfs4fs_lookup lookup;
	struct nfs4fs_ctx ctx;
	int status;

	cargs.sb = dentry->d_sb;
	cargs.tag = "do_lookup()";
	ctx.pt_sb = dentry->d_sb;
	ctx.pt_dentry = dentry->d_parent;
	putfh.dentry = dentry->d_parent;
	lookup.flags = dentry->d_inode ? 0 : FL_NOENT_OK;
	lookup.dentry = dentry;
	lookup.name = &dentry->d_name;
	if ((status = nfs4_setup_compound(&cargs)))
		goto out;
	if ((status = nfs4_setup_putfh(&cargs, &putfh)))
		goto out_args;
	if ((status = nfs4_setup_reval(&cargs, &reval)))
		goto out_args;
	if ((status = nfs4_setup_lookup(&cargs, &lookup)))
		goto out_args;

	if ((status = nfs4_call_compound(&cargs, &cres,&ctx)))
		goto out_args;
	if ((status = nfs4_handle_putfh(&cres, &putfh)))
		goto out_res;
	if ((status = nfs4_handle_reval(&cres, &reval)))
		goto out_res;
	if ((status = nfs4_handle_lookup(&cres, &lookup)))
		goto out_res;
	
out_res:
	nfs4_release_compoundres(&cres);
out_args:
	nfs4_release_compoundargs(&cargs);
out:
	return status;
}

static inline int
dentry_need_reval(struct dentry *dentry, int flags)
{
	struct inode *dir_inode;
	unsigned long timeout;

	if (!NFS4_HAVE_FH(dentry))
		return 1;
	dir_inode = dentry->d_parent->d_inode;
	if (!NFS4_CACHE_JIFFIES(dir_inode))
		return 1;
	if (NFS4_DELEGATION(dir_inode))
		return 0;
	timeout = NFS4_CACHE_TIMEOUT(dir_inode);
	if (!(flags & LOOKUP_CONTINUE) && (CURRENT_TIME < (dir_inode->i_mtime + 15*60)))
		timeout = 1*HZ;
	return (jiffies > dentry->d_time + timeout);
}

/* d_op->revalidate(): called by the VFS when doing a pathname lookup, to revalidate each
 * component of the pathname.  Dentry will be hashed, but may be positive or negative.
 *   The "flags" parameter consists of flags of the form LOOKUP_FOLLOW, etc.  XXX : what
 * should we do with these?  (v3 does almost nothing)
 *   Locks held: none.  We grab the BKL for the duration of the operation.  XXX : is this
 * really necessary, or can we do better?
 */

int
nfs4_dentry_revalidate(struct dentry *dentry, int flags)
{
	struct inode *inode = dentry->d_inode;
	int status;

	nfsv4_printk(level2, "nfs4_dentry_revalidate: starting with dentry=\"%.*s\", "
		     "flags=%d\n", (int)dentry->d_name.len, dentry->d_name.name, flags);
	nfsv4_printk(reval, "nfs4_dentry_revalidate: jiffies=%ld, dtime=%ld, timeout=%ld\n",
		     jiffies, dentry->d_time, NFS4_CACHE_TIMEOUT(dentry->d_parent->d_inode));
	
	lock_kernel();

	if (!dentry_need_reval(dentry, flags)) {
		nfsv4_printk(reval, "nfs4_dentry_revalidate: dentry will not be refreshed\n");
		goto done;
	}
	nfsv4_printk(reval, "nfs4_dentry_revalidate: dentry will be refreshed\n");
	
	if (IS_ROOT(dentry)) {
		/* special behavior in this case; revalidate without lookup */
		if (nfs4_inode_revalidate(dentry))
			goto out_bad;
		goto done;
	}
	if (inode && is_bad_inode(inode)) {
		nfsv4_printk(reval2, "nfs4_dentry_revalidate: bad inode!\n");
		goto out_bad;
	}
	if (nfs4fs_do_lookup(dentry))
		goto out_bad;
	
done:
	status = 1;  /* success */	
out:
	unlock_kernel();
	nfsv4_printk(level2, "nfs4_dentry_revalidate: returning %s\n\n",
		     status ? "success" : "failure");
	return status;
out_bad:
	status = 0;  /* failure */
	goto out;
}

/* d_op->d_release(): called when the VFS deallocates a dentry */

static void
nfs4_dentry_release(struct dentry *dentry)
{
	if (dentry->d_fsdata) {
		NFS4_FREE(dentry->d_fsdata);
		dentry->d_fsdata = NULL;    /* paranoia */
	}
}

/* i_op->lookup(): called just after the VFS creates a new dentry, to instantiate an inode.
     The dentry parameter will be negative and unhashed, but its parent should be set.  XXX :
   is the parent always set?  Important: if the lookup fails because the file does not exist,
   the VFS assumes that ->lookup() will instantiate a NULL inode, and return success.  Another
   tricky point: the return value of this function can either be an error pointer, a NULL
   pointer (which indicates success), or the dentry which was passed in (which also indicates
   success, but the VFS assumes that dget() has been called on this dentry!).  To avoid messing
   up dentry refcounts, it seems safest to always return an error status, which is what we've
   done.  If this function is successful, the dentry should be hashed, and dentry->d_op should
   be set to &nfs4_dentry_operations.
     Locks held: BKL, dir_inode->i_sem
     XXX : make sure there are no race conditions associated with instantiating a negative
   dentry (this has been bugging me for awhile)
*/

static struct dentry *
nfs4_lookup(struct inode *dir_inode, struct dentry *dentry)
{
	int status;

	nfsv4_printk(level2, "nfs4_lookup(%.*s/%.*s): starting\n",
		     (int)dentry->d_parent->d_name.len, dentry->d_parent->d_name.name,
		     (int)dentry->d_name.len, dentry->d_name.name);

	status = nfs4fs_do_lookup(dentry);
	if (!status) {
		d_rehash(dentry);
		dentry->d_op = &nfs4_dentry_operations;
	}
	nfsv4_printk(level2, "nfs4_lookup: returning status %d\n\n", status);
	return ERR_PTR(status);
}

/* i_op->readdir(): we all know and love it!
     Note: a filldir_t looks like
   int filler(void *buf, const char *name, int namlen, off_t offset, ino_t ino,
              unsigned int d_type).
   d_type can be one of DT_DIR, etc.  XXX : should we bother setting this?  we can do it
   without much extra overhead by requesting a type attribute in the readdir.. on the other
   hand, the v3 code just uses DT_UNKNOWN for all entries!
     Contrary to what it says in the v3 nfsd code, i_op->readdir() does _not_ have to call
   the 'filldir' function with a null string to indicate that the last entry has been read!
     The page cache stores entries in the form
        next(1 byte), namelen, name (no XDR "baggage"), type (in form DT_*), ino (64 bit)
     next==0 if EOF, next==1 if more entries on next page, next==2 if more entries on same
   page.
     Locks held: BKL, dir_inode->i_sem, dir_inode->i_zombie
   XXX : second-pass readdir implementation.  issues:
     o client always caches entire directories at a time, this may be a problem for
       pathologically large, or rapidly changing, directories
     o need to understand what happens to pages when not in active use.  our implementation
       assumes that once a page is put in the page cache, it is there for good..
*/

/* space taken up by fields other than the filename in a page cached directory entry.
   4(namelen) + 4(type) + 8(ino) = 16 bytes. */
#define NFS4_READDIR_PAGESKIP         16

static int
readdir_prepare_pagecache(struct file *filp)
{
	static u32 readdir_bmval[] = { 0x00100002 };    /* FATTR4_TYPE, FATTR4_FILEID */
	static bitmap4 readdir_bitmap = { 1, readdir_bmval };
	struct dentry *dentry = filp->f_dentry;
	struct inode *inode = dentry->d_inode;
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_putfh putfh;
	struct nfs4fs_readdir readdir;
	struct dentry *parent_dentry;
	struct page *page;
	char *buf, *end;
	int try = 0;
	int status = 0;

#if 0
	if (NFS4_SNAPSHOT(inode) && !NFS4_NEED_REVAL(inode))
		goto out;
#endif
	nfsv4_printk(level2, "nfs4_readdir calling nfs4_inode_revalidate\n\n");
	if ((status = nfs4_inode_revalidate(dentry)))
		goto out;
	if (NFS4_SNAPSHOT(inode))
		goto out;

	nfsv4_printk(readdir, "nfs4_readdir: reading directory snapshot into page cache\n");
	
	/* Get a dentry for the parent directory.  A bit tricky if we're at the mountpoint of
	   the NFS filesystem.  XXX: should we actually traverse the mountpoint, or consider
	   our root to be its own parent?  What do other filesystems do? */
	
	parent_dentry = dentry->d_parent;
	if (parent_dentry == dentry) {
		struct vfsmount *mnt = mntget(filp->f_vfsmnt);
		struct dentry *d = dget(dentry);
		do {
			if (!follow_up(&mnt, &d))
				break;
		} while (d->d_parent == d);
		parent_dentry = d->d_parent;
		dput(d);
		mntput(mnt);
	}

try_again:
	
	/* Read directory contents into page cache, using several READDIR ops if
	   necessary.  According to the NFSv4 protocol, the server will not return entries
	   for '.' and '..', so we write those entries into the page cache "by hand"
	   before starting the READDIR loop. */

	/* XXX: think about what bits might be set for this page.  PG_error? */
	status = -ENOMEM;
	if (!(page = grab_cache_page(inode->i_mapping, 0))) {
		nfsv4_printk(readdir, "nfs4_readdir: couldn't grab initial page!\n");
		goto out;
	}
	buf = kmap(page);
	if (!buf) {
		nfsv4_printk(readdir, "nfs4_readdir: couldn't kmap initial page!\n");
		goto out_unlock;
	}
	end = buf + PAGE_CACHE_SIZE;

	/* Write entry '.' into the page cache */
	*buf++ = 2;                                     /* next */
	*(u32 *)buf = 1;                                /* namelen */
	buf += 4;
	*buf++ = '.';                                   /* name */
	*(u32 *)buf = DT_DIR;                           /* type */
	buf += 4;
	*(u64 *)buf = inode->i_ino;                     /* ino */
	buf += 8;
	
	/* Write entry '..' into the page cache */
	*buf++ = 2;                                     /* next */
	*(u32 *)buf = 2;                                /* namelen */
	buf += 4;
	*buf++ = '.';                                   /* name */
	*buf++ = '.';                                   /* name */
	*(u32 *)buf = DT_DIR;                           /* type */
	buf += 4;
	*(u64 *)buf = parent_dentry->d_inode->i_ino;    /* ino */
	buf += 8;

	cargs.sb = dentry->d_sb;
	cargs.tag = "nfs4_readdir()";
	putfh.dentry = dentry;
	readdir.count = NFS4_DTSIZE(inode);
	readdir.bm = &readdir_bitmap;
	readdir.inode = inode;
	readdir.cookie = 0;
	memset(readdir.verifier, 0, sizeof(readdir.verifier));
	readdir.page = page;
	readdir.buf = buf;
	readdir.end = end;
	
	do {
		if ((status = nfs4_setup_compound(&cargs)))
			goto out_unmap;
		if ((status = nfs4_setup_putfh(&cargs, &putfh)))
			goto out_args;
		if ((status = nfs4_setup_readdir(&cargs, &readdir)))
			goto out_args;
		if ((status = nfs4_call_compound(&cargs, &cres, NULL)))
			goto out_args;
		if ((status = nfs4_handle_putfh(&cres, &putfh)))
			goto out_res;
		if ((status = nfs4_handle_readdir(&cres, &readdir))) {
			if (readdir.bad_verifier) {
				/* a race condition with another client caused one of a
				   sequence of READDIRs to fail!  clear the page caches and
				   start over. */
				status = -EIO;
				if (++try >= NFS4_READDIR_MAXTRIES)   /* safeguard */
					goto out;
				page = readdir.page;
				if (page) {
					kunmap(page);
					UnlockPage(page);
					put_page(page);
				}
				nfs4_zap_caches(inode);
				goto try_again;
			}
			goto out_res;
		}
		nfs4_release_compoundres(&cres);
		nfs4_release_compoundargs(&cargs);
	} while (!readdir.eof);
	*readdir.buf = 0;                                      /* set final "next" field */
	NFS4_SNAPSHOT(inode) = 1;

out_unmap:
	page = readdir.page;
	if (page) {
		kunmap(page);
out_unlock:
		UnlockPage(page);
		put_page(page);
	}
out:
	return status;
out_res:
	nfs4_release_compoundres(&cres);
out_args:
	nfs4_release_compoundargs(&cargs);
	goto out_unmap;
}

static int
nfs4_readdir(struct file *filp, void *opaque, filldir_t filler)
{
	struct dentry *dentry = filp->f_dentry;
	struct inode *inode = dentry->d_inode;
	loff_t fpos = filp->f_pos;
	unsigned int index;
	unsigned int offset;
	struct page *page;
	char *start, *buf, *name, *end;
	unsigned int namelen, type;
	ino_t ino;
	int status = 0;

	nfsv4_printk(level2, "nfs4_readdir(%.*s/%.*s): starting, fpos=%Ld\n",
		     (int)dentry->d_parent->d_name.len, dentry->d_parent->d_name.name,
		     (int)dentry->d_name.len, dentry->d_name.name, fpos);
	
	if (fpos > OFFT_OFFSET_MAX) {
		nfsv4_printk(readdir2, "nfs4_readdir: fpos > OFFT_OFFSET_MAX\n");
		goto out_overrun;
	}
	if ((status = readdir_prepare_pagecache(filp)))
		goto out;
	
	index = fpos >> PAGE_SHIFT;
	offset = fpos & (PAGE_SIZE - 1);
	nfsv4_printk(readdir, "nfs4_readdir: starting on page %d\n", index);
	
	if (!(page = find_lock_page(inode->i_mapping, index))) {  /* read past EOF */
		nfsv4_printk(readdir2, "nfs4_readdir: page not found\n");
		goto out_overrun;
	}

	/* Try to locate the entry on the page.  If we end up in the middle of an entry,
	   issue a warning and start with the following entry. */

	start = buf = kmap(page);   /* XXX: can this fail? */
	end = buf + offset;
	while (buf < end) {
		switch (*buf) {
		case 0:   /* EOF */
			goto out_unmap_overrun;
		case 1:   /* next entry is on next page */
			goto read_stale;
		case 2:   /* next entry is on this page; skip it */
			buf++;
			namelen = *(u32 *)buf;
			buf += (NFS4_READDIR_PAGESKIP + namelen);
		}
	}
	if (buf == end)
		goto read_entry;
	
read_stale:
	printk(KERN_INFO "nfs4: warning: stale fpos pointer\n");
	
read_entry:
	fpos = (index << PAGE_SHIFT) + (buf - start);
	filp->f_pos = fpos;
	switch (*buf) {
	case 0:   /* EOF */
		nfsv4_printk(readdir, "nfs4_readdir: at eof\n");
		break;
	case 1:   /* advance to next page */
		nfsv4_printk(readdir, "nfs4_readdir: advancing to next page\n");
		kunmap(page);
		UnlockPage(page);
		put_page(page);
		index++;
		if (!(page = find_lock_page(inode->i_mapping, index))) {
#if NFS4_DEBUG
			printk("!!! nfs4_readdir: couldn't find page %d!!!\n", index);
#endif
			status = -EIO;
			goto out;
		}
		start = buf = kmap(page);   /* XXX: can this fail? */
		goto read_entry;
	case 2:   /* read entry from this page */
		buf++;
		namelen = *(u32 *)buf;
		name = buf + 4;
		buf = name + namelen;
		type = *(u32 *)buf;
		buf += 4;
		ino = *(u64 *)buf;
		buf += 8;
		nfsv4_printk(readdir, "nfs4_readdir: filling \"%.*s\", fpos=%Ld, type=%d, "
			     "ino=%ld\n", (int)namelen, name, fpos, type, ino);
			     
		if (filler(opaque, name, namelen, fpos, ino, type)) {
			nfsv4_printk(readdir, "nfs4_readdir: filler returned error\n");
			break;
		}
		goto read_entry;
	default:
		printk(KERN_WARNING "nfs4: warning: directory page corrupted\n");
		break;
	}

	kunmap(page);
	UnlockPage(page);
	put_page(page);
out:
	nfsv4_printk(level2, "nfs4_readdir: returning status %d\n\n", status);
	return status;
out_unmap_overrun:
	kunmap(page);
	UnlockPage(page);
	put_page(page);
out_overrun:
	printk(KERN_INFO "nfs4: warning: stale fpos pointer (past EOF)\n");
	goto out;
}

/* i_op->link()
 *   The new dentry can be either positive or negative, but will always be hashed.
 *   In nfsv3, the link operation didn't return a filehandle, so the new dentry had to be
 * dropped even if the link succeeded (to force a new lookup, which would get the fh).  In v4,
 * we can improve on this ugly situation by piggybacking a LOOKUP-GETFH with the LINK.
 *   XXX : one problem here is that the VFS always tries a lookup for the new filename before
 * doing the link.  It would be nice if we could find a way to coalesce the COMPOUND containing
 * the LOOKUP with the COMPOUND containing the LINK.
 *   Locks held: BKL, dir_inode->i_zombie
 */

static int
nfs4_link(struct dentry *old_dentry, struct inode *dir_inode, struct dentry *dentry)
{
	struct inode *inode = old_dentry->d_inode;
	struct dentry *dir_dentry = dentry->d_parent;
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_putfh putfh1, putfh2;
	struct nfs4fs_link link;
	struct nfs4fs_lookup lookup;
	struct nfs4fs_getattr reval;
	struct nfs4fs_ctx ctx;

	int status;

	nfsv4_printk(level2, "nfs4_link: starting on %.*s->%.*s/%.*s\n",
		     (int)old_dentry->d_name.len, old_dentry->d_name.name,
		     (int)dir_dentry->d_name.len, dir_dentry->d_name.name,
		     (int)dentry->d_name.len, dentry->d_name.name);
	
	/* XXX : I couldn't find this test anywhere in the VFS.  Does it really have to be
	   done here?? */
	if (!(inode->i_mode & S_IFREG)) {
		printk("nfs4_link: not a regular file!\n");
		status = -EINVAL;
		goto out;
	}
	
#if 0
	/* XXX : is this test a good idea? */
	if ((NFS4_DFSID_MAJOR(old_dentry) != NFS4_FSID_MAJOR(dir_inode)) ||
	    (NFS4_DFSID_MINOR(old_dentry) != NFS4_FSID_MINOR(dir_inode))) {
		status = -EXDEV;
		goto out;
	}
#endif

	cargs.sb = dir_inode->i_sb;
	cargs.tag = "nfs4_link()";
	ctx.pt_sb = dir_inode->i_sb;
	ctx.pt_dentry = dentry;
	putfh1.dentry = old_dentry;
	putfh2.dentry = dir_dentry;
	link.dentry = dentry;
	lookup.flags = FL_CLOBBER_OK;
	lookup.dentry = dentry;
	lookup.name = &dentry->d_name;
	if ((status = nfs4_setup_compound(&cargs)))
		goto out;
	if ((status = nfs4_setup_putfh(&cargs, &putfh1)))
		goto out_args;
	if ((status = nfs4_setup_savefh(&cargs)))
		goto out_args;
	if ((status = nfs4_setup_putfh(&cargs, &putfh2)))
		goto out_args;
	if ((status = nfs4_setup_link(&cargs, &link)))
		goto out_args;
	if ((status = nfs4_setup_reval(&cargs, &reval)))
		goto out_args;
	if ((status = nfs4_setup_lookup(&cargs, &lookup)))
		goto out_args;

	nfs4_zap_caches(dir_inode);
	if ((status = nfs4_call_compound(&cargs, &cres,&ctx))) {
		d_invalidate(dentry);
		goto out_args;
	}
	if ((status = nfs4_handle_putfh(&cres, &putfh1)))
		goto out_res;
	if ((status = nfs4_handle_savefh(&cres)))
		goto out_res;
	if ((status = nfs4_handle_putfh(&cres, &putfh2)))
		goto out_res;
	if ((status = nfs4_handle_link(&cres, &link)))
		goto out_res;
	if (nfs4_handle_reval(&cres, &reval))
		goto out_res;
	if (nfs4_handle_lookup(&cres, &lookup))
		goto out_res;

out_res:
	nfs4_release_compoundres(&cres);
out_args:
	nfs4_release_compoundargs(&cargs);
out:
	nfsv4_printk(link, "nfs4_link: returning status %d\n\n", status);
	return status;
}

/*
 * code common to i_op->symlink(), i_op->mkdir()
 */
static int
__nfs4_create(struct nfs4fs_create *create)
{
	struct dentry *dentry = create->dentry;
	struct dentry *dir_dentry = dentry->d_parent;
	struct inode *dir_inode = dir_dentry->d_inode;
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_putfh putfh;
	struct nfs4fs_getattr reval;
	int status;

	cargs.sb = dentry->d_sb;
	cargs.tag = "__nfs4_create()";
	putfh.dentry = dir_dentry;
	if ((status = nfs4_setup_compound(&cargs)))
		goto out;
	if ((status = nfs4_setup_putfh(&cargs, &putfh)))
		goto out_args;
	if ((status = nfs4_setup_reval(&cargs, &reval)))
		goto out_args;
	if ((status = nfs4_setup_create(&cargs, create)))
		goto out_args;

	nfs4_zap_caches(dir_inode);
	if ((status = nfs4_call_compound(&cargs, &cres,(struct nfs4fs_ctx *)NULL)))
		goto out_args;
	if ((status = nfs4_handle_putfh(&cres, &putfh)))
		goto out_res;
	if ((status = nfs4_handle_reval(&cres, &reval)))
		goto out_res;
	if ((status = nfs4_handle_create(&cres, create)))
		goto out_res;

out_res:
	nfs4_release_compoundres(&cres);
out_args:
	nfs4_release_compoundargs(&cargs);
out:
	return status;	
}

/* i_op->symlink()
     Extremely similar to i_op->link(), but the COMPOUND needed to implement it is more
   straightforward.  Note that no attributes need to be set for symlinks.
     Locks held: BKL, dir_inode->i_zombie
 */

static int
nfs4_symlink(struct inode *dir_inode, struct dentry *dentry, const char *name)
{
	struct nfs4fs_create create;
	int status;

	nfsv4_printk(level2, "nfs4_symlink: starting on %.*s/%.*s -> \"%s\"\n",
		     (int)dentry->d_parent->d_name.len, dentry->d_parent->d_name.name,
		     (int)dentry->d_name.len, dentry->d_name.name, name);
	
	/* XXX : check superblock to see if server allows symlinks, and return error
	   immediately if not?? */

	create.dentry = dentry;
	create.type = NF4LNK;
	create.u.linktext = name;
	status = __nfs4_create(&create);
	
	nfsv4_printk(level2, "nfs4_symlink: returning status %d\n\n", status);
	return status;
}

/* i_op->mkdir(): essentially identical to i_op->symlink(). */

static int
nfs4_mkdir(struct inode *dir_inode, struct dentry *dentry, int mode)
{
	struct nfs4fs_create create;
	int status;
	
	/* XXX : is this necessary? */
	mode = (mode & S_IALLUGO) | S_IFDIR;
	
	nfsv4_printk(level2, "nfs4_mkdir: creating directory %.*s/%.*s, mode=%d\n",
		     (int)dentry->d_parent->d_name.len, dentry->d_parent->d_name.name,
		     (int)dentry->d_name.len, dentry->d_name.name, mode);

	create.dentry = dentry;
	create.type = NF4DIR;
	create.u.mode = mode;
	status = __nfs4_create(&create);

	nfsv4_printk(level2, "nfs4_mkdir: returning status %d\n\n", status);
	return status;
}

/* For the time being, we can actually use the same function, nfs4_remove(), for both
   i_op->unlink() and i_op->rmdir()!
   
   i_op->unlink()
     (XXX: This comment is out of date, since we now sillyrename the file.. rewrite)
     This routine doesn't have to manipulate the dentry at all, just return success or failure.
   If it returns success, the VFS will call d_delete(), which takes care of everything; if it
   returns failure, the VFS does nothing.
     In nfsv3, this was a somewhat complex operation for three reasons: (1) special semantics
   for sillyrenamed files, (2) temporary unhashing of the dentry while the delete is attempted,
   (3) failure if dentry->d_count > 1.  In v4, all of these can be eliminated.  We no longer
   have sillyrenames.  Temporary unhashing of the dentry seems like a race condition waiting
   to happen; suppose the operation fails and the dentry is rehashed, but in the meantime
   someone has done a new lookup and a new dentry has been instantiated (since i_op->lookup()
   doesn't require dir_inode->i_zombie(), a lookup and an unlink can proceed concurrently).
   We will end up with duplicate dentries in the VFS; while I'm not sure if this can actually
   lead to kernel oops or similar disaster, the law of unintended consequences makes this seem
   like a bad idea.  Finally, returning failure if dentry->d_count > 1 seems unnecessary, since
   d_delete() tests this condition and knows how to handle it.  In summary, this is a very
   simple operation in v4; just try to unlink the file, and return success or failure
   accordingly.
     Locks held: BKL, dir_inode->i_zombie
     
   i_op->rmdir()
     The semantics of this routine are identical to those of i_op->unlink()
     Note: the VFS does not check to see whether the directory is empty; that is the
   responsibility of the lower-level filesystem.  XXX : do a local check to see if directory
   contains anything before contacting the server?
     Locks held: BKL, dir_inode->i_zombie, dentry->d_inode->i_zombie
*/

static int
nfs4_unlink(struct inode *dir_inode, struct dentry *dentry)
{
	int status = 0;
#if 0
	int rehash = 0;
#endif

	nfsv4_printk(level2, "nfs4_unlink(%.*s/%.*s): starting\n",
		     (int)dentry->d_parent->d_name.len, dentry->d_parent->d_name.name,
		     (int)dentry->d_name.len, dentry->d_name.name);
	
	if (!NFS4_SILLY_RENAME(dir_inode->i_sb)) {
		nfsv4_printk(remove, "nfs4_unlink: silly rename disabled, removing file\n");
		status = nfs4_remove(dir_inode, dentry);
		goto out;
	}

#if 0
	/*
	 *  Unhash the dentry while we remove the file..
	 */
	if (!d_unhashed(dentry)) {
		d_drop(dentry);
		rehash = 1;
	}
#endif

	if (dentry->d_flags & DCACHE_NFSFS_RENAMED)
		nfsv4_printk(remove, "nfs4_unlink: already sillyrenamed, no-op\n");
	else if (atomic_read(&dentry->d_count) == 1) {
		nfsv4_printk(remove, "nfs4_unlink: dentry not in use, removing file\n");
		status = nfs4_remove(dir_inode, dentry);
	}
	else {
		nfsv4_printk(remove, "nfs4_silly_unlink: silly renaming file\n");
		status = nfs4_sillyrename(dir_inode, dentry);
	}

#if 0
	if (rehash)
		d_rehash(dentry);
#endif
	
out:
	nfsv4_printk(level2, "nfs4_unlink(): returning status %d\n", status);
	return status;
}

int
nfs4_remove(struct inode *dir_inode, struct dentry *dentry)
{
	struct dentry *dir_dentry = dentry->d_parent;
	struct nfs4fs_putfh putfh;
	struct compound4args cargs;
	struct compound4res cres;
	int status;

#if NFS4_DEBUG
	if (dir_dentry->d_inode != dir_inode) {
		nfsv4_printk(remove2, "nfs4_remove: inode mismatch!!\n");
		status = -EINVAL;
		goto out;
	}
#endif

	nfsv4_printk(level2, "nfs4_remove: starting on %.*s/%.*s\n",
		     (int)dir_dentry->d_name.len, dir_dentry->d_name.name,
		     (int)dentry->d_name.len, dentry->d_name.name);
	
	/* XXX : make sure there are no pending writes? */

	cargs.sb = dentry->d_sb;
	cargs.tag = "nfs4_remove()";
	putfh.dentry = dir_dentry;
	if ((status = nfs4_setup_compound(&cargs)))
		goto out;
	if ((status = nfs4_setup_putfh(&cargs, &putfh)))
		goto out_args;
	if ((status = nfs4_setup_remove(&cargs, &dentry->d_name)))
		goto out_args;

	nfs4_zap_caches(dir_inode);
	if ((status = nfs4_call_compound(&cargs, &cres,(struct nfs4fs_ctx *)NULL))) {
		/* we can't be sure whether the remove succeeded, so drop the dentry to force
		   a new lookup. */
		d_invalidate(dentry);
		nfs4_drop_fh(dentry);
		goto out_args;
	}
	if ((status = nfs4_handle_putfh(&cres, &putfh)))
		goto out_res;
	if ((status = nfs4_handle_remove(&cres, dir_inode)))
		goto out_res;
	if (dentry->d_inode)
		dentry->d_inode->i_nlink--;   /* XXX: is this right? */
	
out_res:
	nfs4_release_compoundres(&cres);
out_args:
	nfs4_release_compoundargs(&cargs);
out:
	nfsv4_printk(level2, "nfs4_remove: returning status %d\n\n", status);
	return status;
}

/* i_op->rename() : the mother of all vnode ops!
     We are guaranteed that old_dentry is positive, and that if new_dentry is positive, it
   is a directory iff old_dentry is.
     The v3 code for a rename suffered from 2 unpleasantries:
   (1) code didn't deal well with the case where the target dentry was in use, either
       returning -EUBSY or going through the sillyrename() cruft
   (2) the target dentry was unhashed for the duration of the operation, and rehashed
       if the rename turned out to fail.  this seems like a race condition waiting to
       happen (see comment in nfs4_unlink above).
   I have changed things so that the rename always proceeds normally.  If there are
   references to the target after the rename, the dentry is unhashed (so that no new
   references will be created, and the last dput() will delete the dentry), and the
   inode marked bad, so that all I/O on the inode will fail.  XXX : is making the
   inode bad really right??
     Note: for some reason, if the target is an existing directory, the VFS will rehash the
   target dentry, even if i_op->rename() unhashed it!  I really don't understand this..
     If i_op->rename() succeeds, the VFS will call d_move() to perform the needed dentry
   manipulations.
     Locks held: BKL, old_dir_inode->i_sem, new_dir_inode->i_sem, old_dir_inode->i_zombie,
   new_dir_inode->i_zombie.  also, if new_dentry is an existing directory,
   new_dentry->d_inode->i_zombie.
*/

/*
 *  We now do sillyrename of the target file.  This means that we have to enable the
 *  FS_ODD_RENAME hack in the VFS.
 */
static int
nfs4_rename(struct inode *old_dir_inode, struct dentry *old_dentry,
	    struct inode *new_dir_inode, struct dentry *new_dentry)
{
	struct dentry *old_dir_dentry = old_dentry->d_parent;
	struct dentry *new_dir_dentry = new_dentry->d_parent;
	struct dentry *temp_dentry = NULL;
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_putfh putfh1, putfh2;
	struct nfs4fs_rename rename;
	struct nfs4fs_getattr getattr;
	struct nfs4fs_lookup lookup;
	int status;

#if NFS4_DEBUG
	status = -EINVAL;
	if (!old_dir_dentry || !new_dir_dentry) {
		printk("nfs4_rename: no dir dentry!\n");
		goto out;
	}
	if ((old_dir_dentry->d_inode != old_dir_inode) ||
	    (new_dir_dentry->d_inode != new_dir_inode)) {
		printk("nfs4_rename: inode mismatch!\n");
		goto out;
	}
	if (!old_dentry->d_inode) {
		printk("nfs4_rename: source dentry negative!\n");
		goto out;
	}
#endif

	/* XXX: do we really need to unhash the target dentry?  I think that grabbing the
	   inode semaphore takes care of this problem.. double-check */
	
	nfsv4_printk(level2, "nfs4_rename: starting on %.*s/%.*s->%.*s/%.*s\n",
		     (int)old_dir_dentry->d_name.len, old_dir_dentry->d_name.name,
		     (int)old_dentry->d_name.len, old_dentry->d_name.name,
		     (int)new_dir_dentry->d_name.len, new_dir_dentry->d_name.name,
		     (int)new_dentry->d_name.len, new_dentry->d_name.name);
	
	if (!NFS4_SILLY_RENAME(old_dir_inode->i_sb)) {
		nfsv4_printk(remove, "nfs4_unlink: silly rename disabled, doing rename\n");
		goto do_rename;
	}
	if (!new_dentry->d_inode) {
		nfsv4_printk(rename, "nfs4_rename: target dentry negative, doing rename\n");
		goto do_rename;
	}

	/* XXX: needed?  this is in v3... */
	status = -EBUSY;
	if (S_ISDIR(new_dentry->d_inode->i_mode)) {
		nfsv4_printk(rename2, "nfs4_rename: target is directory?!\n");
		goto out;
	}
	
	if (atomic_read(&new_dentry->d_count) == 1) {
		nfsv4_printk(rename, "nfs4_rename: target not in use, doing rename\n");
		goto do_rename;
	}

	/*
	 *  Try to sillyrename the target.  XXX: we don't have good error handling for the
	 *  case where the sillyrename succeeds but the real rename fails.  Should we try
	 *  to "unsillyrename" the file back to its original name?
	 */
	nfsv4_printk(rename, "nfs4_rename: silly renaming target first...\n");
	status = -ENOMEM;
	if (!(temp_dentry = d_alloc(new_dentry->d_parent, &new_dentry->d_name))) {
		nfsv4_printk(rename2, "nfs4_rename: couldn't alloc new dentry!\n");
		goto out;
	}
	if ((status = nfs4_sillyrename(new_dir_inode, new_dentry))) {
		nfsv4_printk(rename2, "nfs4_rename: nfs4_sillyrename failed!\n");
		goto out;
	}
	temp_dentry->d_op = &nfs4_dentry_operations;
	d_instantiate(temp_dentry, NULL);
	new_dentry = temp_dentry;
	nfsv4_printk(rename, "nfs4_rename: done with silly rename, now doing real rename..\n");

do_rename:
	/*
	 * prune child dentries and writebacks if needed...
	 */
	if (atomic_read(&old_dentry->d_count) > 1) {
		nfs4_sync_file(old_dentry->d_inode, NULL, 0, 0, FLUSH_WAIT);
		shrink_dcache_parent(old_dentry);
	}
	
	cargs.sb = old_dir_inode->i_sb;
	cargs.tag = "nfs4_rename()";
	putfh1.dentry = old_dir_dentry;
	rename.old_dentry = old_dentry;
	rename.new_dentry = new_dentry;
	getattr.flags = FL_NONFATAL;
	getattr.bm = &nfs4_negative_bitmap;
	lookup.flags = FL_CLOBBER_OK;
	lookup.dentry = old_dentry;
	lookup.name = &new_dentry->d_name;
	if ((status = nfs4_setup_compound(&cargs)))
		goto out;
	if ((status = nfs4_setup_putfh(&cargs, &putfh1)))
		goto out_args;
	if ((status = nfs4_setup_savefh(&cargs)))
		goto out_args;
	if (old_dir_dentry != new_dir_dentry) {
		/* small optimization: only need to PUTFH the new directory if it differs
		   from the old one. */
		putfh2.dentry = new_dir_dentry;
		if ((status = nfs4_setup_putfh(&cargs, &putfh2)))
			goto out_args;
	}
	if ((status = nfs4_setup_rename(&cargs, &rename)))
		goto out_args;
	if ((status = nfs4_setup_getattr(&cargs, &getattr)))
		goto out_args;
	if ((status = nfs4_setup_lookup(&cargs, &lookup)))
		goto out_args;

	nfs4_zap_caches(old_dir_inode);
	nfs4_zap_caches(new_dir_inode);
	if ((status = nfs4_call_compound(&cargs, &cres,(struct nfs4fs_ctx *)NULL))) {
		nfsv4_printk(rename2, "nfs4_rename: nfs4_call_compound failed!\n");
		/* we don't know whether the rename succeeded, so drop both dentries to force
		   new lookups.  XXX : there is one case where this won't work, namely when the
		   target is an existing directory (see the comments above).. what can we do
		   about this?! */
		d_invalidate(old_dentry);
		d_invalidate(new_dentry);
		nfs4_drop_fh(new_dentry);
		goto out_args;
	}
	if ((status = nfs4_handle_putfh(&cres, &putfh1)))
		goto out_res;
	if ((status = nfs4_handle_savefh(&cres)))
		goto out_res;
	if (old_dir_dentry != new_dir_dentry) {
		if ((status = nfs4_handle_putfh(&cres, &putfh2)))
			goto out_res;
	}
	if ((status = nfs4_handle_rename(&cres, &rename))) {
		d_invalidate(old_dentry);
		d_invalidate(new_dentry);
		goto out_res;
	}

	/*
	 *  Since we have enabled FS_ODD_RENAME in the VFS, ->remove() is now responsible for
	 *  calling d_move() to update the dcache after the rename.  Ugh!
	 */
	d_move(old_dentry, new_dentry);
	
	if (nfs4_handle_getattr(&cres, &getattr))
		goto out_clearfh;
	if (nfs4_handle_lookup(&cres, &lookup))
		goto out_clearfh;

out_res:
	nfs4_release_compoundres(&cres);
out_args:
	nfs4_release_compoundargs(&cargs);
out:
	if (temp_dentry)
		dput(temp_dentry);
	nfsv4_printk(level2, "nfs4_rename: returning status %d\n\n", status);
	return status;
out_clearfh:
	nfs4_drop_fh(old_dentry);
	goto out_res;
}
