/*
 *  linux/fs/nfs4/encode.c
 *
 *  Copyright (c) 2001 The Regents of the University of Michigan.
 *  All rights reserved.
 *
 *  Kendrick Smith <kmsmith@umich.edu>
 *  Andy Adamson <andros@umich.edu>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/mm.h>
#include <linux/pagemap.h>

#include <linux/nfs4/nfs4_prot.h>
#include <linux/nfs4/nfs4_debug.h>
#include <linux/nfs4/nfs4.h>
#include <linux/nfs4/nfs4fs.h>

static int nfs4_force_getfh(struct dentry *dentry);

/* This is the maximum attribute buffer size on a SETATTR, not including owner/group fields
   (which are variable length).  1(mode) + 2(size) + 4(atime) + 4(mtime) = 11. */
#define SETATTR_STATICWORDS   11

/*
 *  Primitives for XDR encoding are:
 *    reserve_space() - returns a pointer to a contiguous block of memory to be XDR'ed into,
 *                      starting at the current value of the XDR encode pointer, if such a
 *                      block is available.
 *    adjust_args()   - sets the value of the XDR encode pointer
 *    send_buf()      - puts an opaque buffer into the XDR stream, hopefully without a copy
 *
 *  All XDR encode routines are written in terms of these primitives, without direct use
 *  to the iov_* fields, so if a different internal representation of the output buffer is
 *  needed, it is only necessary to change the primitives.
 */

static inline u32 *
reserve_space(struct compound4args *cp, unsigned int nwords)
{
	if (cp->p + nwords <= cp->end)
		/* fast track */
		return cp->p;
	
	if (cp->iov_nr == NFS4FS_MAX_IOV)
		return NULL;

#if NFS4_DEBUG
	/* XXX : do we need this test? */
	if (nwords > NFS4_IOV_NWORDS)
		return NULL;
#endif

	FINALIZE_IOVEC(cp);
	if (!(cp->p = NFS4_ALLOC(NFS4_IOV_NBYTES)))
		return NULL;
	cp->end = cp->p + NFS4_IOV_NWORDS;
	cp->iov[cp->iov_nr].iov_base = (char *)cp->p;
	cp->iov_static[cp->iov_nr] = 0;
	cp->iov_nr++;
	return cp->p;
}

static inline void
adjust_args(struct compound4args *cp, u32 *p)
{
#if NFS4_DEBUG
	/* make sure XDR encode pointer is within allowed range */
	if (((char *)p < (char *)cp->p) || ((char *)p > (char *)cp->end)) {
		printk("!!! nfs4_adjust_args: out-of-range pointer !!!\n");
		return;
	}
	/* make sure XDR encode pointer is kept "aligned", i.e. always advanced by a number of
	   bytes divisible by 4 */
	if (((char *)p - (char *)cp->p) & 3) {
		printk("!!! nfs4fs: unaligned pointer in adjust_args !!!\n");
		return;
	}
#endif
	cp->p = p;
}

static inline int
send_buf(struct compound4args *cp, unsigned int buflen, char *buf)
{
	/* note: if buflen is not divisible by 4, the extra XDR "baggage" will automatically
	   be sent along by nfs4_enc_compound() */
	
	if (cp->iov_nr == NFS4FS_MAX_IOV)
		return -ENOMEM;
	FINALIZE_IOVEC(cp);
	cp->iov[cp->iov_nr].iov_base = buf;
	cp->iov_static[cp->iov_nr] = 1;
	cp->end = cp->p = (u32 *)(buf + buflen);
	cp->iov_nr++;
	return 0;
}

/*
 *  Setup routines 
 */

/* type, changeid, size, fsid, fileid */
static u32 positive_bmval[] = { 0x0010011a };
bitmap4 nfs4_positive_bitmap = { 1, positive_bmval };

/* type, changeid, size, fsid, fileid, mode, nlink, owner, group,
   rawdev, ctime, atime, mtime */
static u32 negative_bmval[] = { 0x0010011a, 0x0020823a };
bitmap4 nfs4_negative_bitmap = { 2, negative_bmval };

static int
__nfs4_setup_lookup(struct compound4args *cp, struct qstr *name)
{
	u32 *p;
	int len = name->len;
	int quadlen = XDR_QUADLEN(len);
	int status;

	nfsv4_printk(level2, "    setting up op #%d: LOOKUP\n", cp->nops);
	if (!(p = reserve_space(cp, quadlen + 3)))
		goto out_nomem;
	*p++ = htonl(OP_LOOKUP);
	*p++ = htonl(1);  /* we only do single-component lookups */
	*p++ = htonl(len);
	memcpy(p, name->name, len);
	p += quadlen;
	adjust_args(cp, p);
	cp->nops++;
	status = 0;
	
out:
	return status;
out_nomem:
	nfsv4_printk(xdr2, "setup_lookup: out of memory!\n");
	status = -ENOMEM;
	goto out;
}

int
nfs4_setup_compound(struct compound4args *cp)
{
	u32 *p;
	int status;
#if NFS4_DEBUG
	static spinlock_t taglock = SPIN_LOCK_UNLOCKED;
	static unsigned int tagid = 0;
	static char tagbuf[100];
	int taglen;
#endif

	cp->flags = CA_FAIL_RETRY;
	cp->current_fh = NULL;
	cp->save_fh = NULL;
	cp->seqid_holder = NULL;
	cp->stateid_holder = NULL;
	if (!(cp->iov[0].iov_base = NFS4_ALLOC(NFS4_IOV_NBYTES)))
		goto out_nomem;
	cp->iov_nr = 1;
	cp->iov_static[0] = 0;
	cp->p = (u32 *)cp->iov[0].iov_base;
	cp->end = cp->p + NFS4_IOV_NWORDS;
	cp->nops = 0;
	cp->seqidp = NULL;
	cp->nstateids = 0;
	cp->renew_index = 0;
	
	/* Tag is XDR encoded first.. */
#if NFS4_DEBUG
	if (!(p = reserve_space(cp, 112)))
		goto out_free;
	spin_lock(&taglock);
	sprintf(tagbuf, "%s:%d", cp->tag, tagid++);
	taglen = strlen(tagbuf);
	*p++ = htonl(taglen);
	memcpy(p, tagbuf, taglen);
	p += XDR_QUADLEN(taglen);
	spin_unlock(&taglock);
#else
	if (!(p = reserve_space(3)))
		goto out_free;
	*p++ = 0;
#endif

	/* Then minor version and argarray length.. */
	*p++ = htonl(NFS4_MINOR_VERSION);
	cp->nopsp = p;
	*p++ = 0;   /* set argarray_len to 0, just to be paranoid */
	adjust_args(cp, p);
	status = 0;

out:
	return status;
out_nomem:
	nfsv4_printk(xdr2, "setup_compound: out of memory!\n");
	cp->iov_nr = 0;
	status = -ENOMEM;
	goto out;
out_free:
	nfsv4_printk(xdr2, "setup_compound: not enough space in iovec!\n");
	NFS4_FREE(cp->p);
	cp->iov_nr = 0;
	status = -EIO;
	goto out;
}

int
nfs4_setup_simple_op(struct compound4args *cp, nfs_opnum4 op)
{
	u32 *p;
	int status;

	if (op == OP_PUTROOTFH)
		nfsv4_printk(level2, "    setting up op #%d: PUTROOTFH\n", cp->nops);
	else if (op == OP_READLINK)
		nfsv4_printk(level2, "    setting up op #%d: READLINK\n", cp->nops);
	else
		nfsv4_printk(level2, "    setting up op #%d: ???\n", cp->nops);
	
	if (!(p = reserve_space(cp, 1)))
		goto out_nomem;
	*p++ = htonl(op);
	adjust_args(cp, p);
	cp->nops++;
	status = 0;

out:
	return status;
out_nomem:
	nfsv4_printk(xdr2, "nfs4_setup_simpleop: out of memory!\n");
	status = -ENOMEM;
	goto out;
}

int
nfs4_setup_close(struct compound4args *cp, struct nfs4fs_close *close)
{
	u32 *p;
	struct nfs4fs_file_data *fdata = close->data;
	int access;
	int status;

#if NFS4_DEBUG
	if (!fdata || !fdata->fi_nopen) {
		printk("!!! nfs4_setup_close: no opened file !!!\n");
		return -EIO;
	}
#endif

	if ((status = nfs4fs_wait_on_recoverd(cp->sb)))
		goto out;
	if (!(p = reserve_space(cp, 10)))
		goto out_nomem;

	if (fdata->fi_nopen == 1) {
		nfsv4_printk(level2, "    setting up op #%d: CLOSE\n", cp->nops);
		*p++ = close->op = htonl(OP_CLOSE);
		cp->seqidp = p;
		p++;
		cp->stateidp[0] = p;
		p += XDR_QUADLEN(sizeof(stateid4));
	}
	else {
		access = 0;
		if (fdata->fi_nread > close->read)
			access |= OPEN4_SHARE_ACCESS_READ;
		if (fdata->fi_nwrite > close->write)
			access |= OPEN4_SHARE_ACCESS_WRITE;
		nfsv4_printk(level2, "    setting up op #%d: OPEN_DOWNGRADE\n", cp->nops);
		*p++ = close->op = htonl(OP_OPEN_DOWNGRADE);
		cp->stateidp[0] = p;
		p += XDR_QUADLEN(sizeof(stateid4));
		cp->seqidp = p;
		p++;
		*p++ = htonl(access);
		*p++ = 0;                 /* we don't use share deny bits */
	}
	adjust_args(cp, p);
	cp->nops++;
	cp->flags = CA_SEQID_RELEASE;
	cp->seqid_holder = fdata->fi_lockowner;
	cp->nstateids = 1;
	cp->stateid_holder = fdata;
	cp->seqid_index = cp->nops;
	cp->renew_index = cp->nops;
	
	status = 0;

out:
	return status;
out_nomem:
	nfsv4_printk(xdr2, "nfs4_setup_close: out of memory!\n");
	status = -ENOMEM;
	goto out;
}

int
nfs4_setup_commit(struct compound4args *cp, struct nfs4fs_commit *commit)
{
	u32 *p;
	int status;

	if (list_empty(&commit->pages))
		BUG();
	
	if (!(p = reserve_space(cp, 4)))
		goto out_nomem;
	nfsv4_printk(level2, "    setting up op #%d: COMMIT\n", cp->nops);
	*p++ = htonl(OP_COMMIT);
	*p++ = htonl((u32)(commit->start >> 32));
	*p++ = htonl((u32)(commit->start));
	*p++ = htonl((u32)(commit->length));
	cp->nops++;
	adjust_args(cp, p);
	status = 0;
	
out:
	return status;
out_nomem:
	nfsv4_printk(xdr2, "nfs4_setup_commit: out of memory!\n");
	status = -ENOMEM;
	goto out;
}

int
nfs4_setup_create(struct compound4args *cp, struct nfs4fs_create *create)
{
	u32 *p;
	int status;
	int len, quadlen;

	nfsv4_printk(level2, "    setting up op #%d: CREATE\n", cp->nops);

	len = create->dentry->d_name.len;
	quadlen = XDR_QUADLEN(len);
	if (!(p = reserve_space(cp, quadlen + 3)))
		goto out_nomem;
	*p++ = htonl(OP_CREATE);
	*p++ = htonl(len);
	memcpy(p, create->dentry->d_name.name, len);
	p += quadlen;
	*p++ = htonl(create->type);
	adjust_args(cp, p);
	
	switch (create->type) {
	case NF4DIR:
		nfsv4_printk(level2, "    setting up op #%d: SETATTR\n", cp->nops+1);
		if (!(p = reserve_space(cp, 8)))
			goto out_nomem;
		*p++ = htonl(OP_SETATTR);
		*p++ = 0;		/* stateid */
		*p++ = 0;		/* stateid */
		*p++ = htonl(2);	/* bitmap */
		*p++ = 0;		/* bitmap */
		*p++ = htonl(0x00000002);	/* bitmap -- set mode */
		*p++ = htonl(4);	/* attrvals */
		*p++ = htonl(create->u.mode);	/* attrvals */
		adjust_args(cp, p);
		cp->nops += 2;
		break;
	case NF4LNK:
		len = strlen(create->u.linktext);
		quadlen = XDR_QUADLEN(len);
		if (!(p = reserve_space(cp, quadlen + 1)))
			goto out_nomem;
		*p++ = htonl(len);
		memcpy(p, create->u.linktext, len);
		p += quadlen;
		adjust_args(cp, p);
		cp->nops++;
		break;
	default:
		printk("!!! nfs4_setup_create: bad type !!!\n");
		status = -EINVAL;
		goto out;
	}
	cp->current_fh = create->dentry;

	create->getattr.flags = FL_NONFATAL;
	create->getattr.bm = &nfs4_negative_bitmap;
	create->getfh.nonfatal = 1;
	if ((status = nfs4_setup_getattr(cp, &create->getattr)))
		goto out;
	if ((status = nfs4_setup_getfh(cp, &create->getfh)))
		goto out;
	
out:
	return status;
out_nomem:
	nfsv4_printk(xdr2, "setup_create: out of memory!\n");
	status = -ENOMEM;
	goto out;
}

int
nfs4_setup_delegreturn(struct compound4args *cp, stateid4 stateid)
{
	u32 *p;
	int status;
	
	nfsv4_printk(level2, "    setting up op #%d: DELEGRETURN\n", cp->nops);
	p = reserve_space(cp, 1 + XDR_QUADLEN(sizeof(stateid4)));
	if (!p)
		goto out_nomem;
	*p++ = htonl(OP_DELEGRETURN);
	memcpy(p, stateid, sizeof(stateid4));
	p += XDR_QUADLEN(sizeof(stateid4));
	adjust_args(cp, p);
	cp->nops++;
	status = 0;

out:
	return status;
out_nomem:
	nfsv4_printk(xdr2, "setup_delegreturn: out of memory!\n");
	status = -ENOMEM;
	goto out;
}

int
nfs4_setup_getattr(struct compound4args *cp, struct nfs4fs_getattr *getattr)
{
	bitmap4 *bm = getattr->bm;
	u32 *p;
	int i;
	int status;

#if NFS4_DEBUG
	if (!cp->current_fh && !(getattr->flags & FL_NOREVAL)) {
		printk("!!! setup_getattr: no fh !!!\n");
		return -EIO;
	}
#endif

	nfsv4_printk(level2, "    setting up op #%d: GETATTR\n", cp->nops);

	if (!(p = reserve_space(cp, 2 + bm->bitmap4_len)))
		goto out_nomem;
	*p++ = htonl(OP_GETATTR);
	*p++ = htonl(bm->bitmap4_len);
	for (i = 0; i < bm->bitmap4_len; i++)
		*p++ = htonl(bm->bitmap4_val[i]);
	adjust_args(cp, p);
	cp->nops++;
	getattr->dentry = cp->current_fh;
	status = 0;

out:
	return status;
out_nomem:
	nfsv4_printk(xdr2, "nfs4_setup_getattr: out of memory!\n");
	status = -ENOMEM;
	goto out;
}

int
nfs4_setup_getfh(struct compound4args *cp, struct nfs4fs_getfh *getfh)
{
	u32 *p;
	int status;
	
#if NFS4_DEBUG
	if (!cp->current_fh) {
		printk("!!! setup_getfh: no fh !!!\n");
		return -EIO;
	}
#endif
	nfsv4_printk(level2, "    setting up op #%d: GETFH\n", cp->nops);

	getfh->dentry = cp->current_fh;

	if (!(p = reserve_space(cp, 1)))
		goto out_nomem;
	*p++ = htonl(OP_GETFH);
	adjust_args(cp, p);
	cp->nops++;
	status = 0;
	
out:
	return status;
out_nomem:
	nfsv4_printk(xdr2, "setup_getfh: out of memory!\n");
	status = -ENOMEM;
	goto out;
}

int
nfs4_setup_link(struct compound4args *cp, struct nfs4fs_link *link)
{
	u32 *p;
	struct qstr *name = &link->dentry->d_name;
	int quadlen = XDR_QUADLEN(name->len);
	int status;

#if NFS4_DEBUG
	if (!cp->current_fh || !cp->save_fh || !cp->save_fh->d_inode) {
		printk("!!! setup_link: missing fh, or negative dentry !!!\n");
		return -EIO;
	}
#endif

	nfsv4_printk(level2, "    setting up op #%d: LINK\n", cp->nops);

	if (!(p = reserve_space(cp, quadlen + 2)))
		goto out_nomem;
	*p++ = htonl(OP_LINK);
	*p++ = htonl(name->len);
	memcpy(p, name->name, name->len);
	p += quadlen;
	adjust_args(cp, p);
	cp->nops++;
	link->inode = cp->save_fh->d_inode;
	status = 0;
	
out:
	return status;
out_nomem:
	nfsv4_printk(xdr2, "setup_link: out of memory!\n");
	status = -ENOMEM;
	goto out;
}

int
nfs4_setup_lock(struct compound4args *cp, struct nfs4fs_lock *lock)
{
	struct nfs4fs_file_data *file_data = lock->file_data;
	struct nfs4fs_lockowner *lockowner = file_data->fi_lockowner;
	u64 length;
	u32 *p;
	int status;

	if ((status = nfs4fs_wait_on_recoverd(cp->sb)))
		goto out;
	if (!(p = reserve_space(cp, 12)))
		goto out_nomem;
	nfsv4_printk(level2, "    setting up op #%d: LOCK(%Ld-%Ld) lockowner=%d\n",
		     cp->nops, lock->start, lock->end, lockowner->lo_id);
	length = (lock->end != OFFSET_MAX) ? (lock->end - lock->start + 1) : ~(u64)0;
	
	*p++ = htonl(OP_LOCK);
	*p++ = htonl(lock->type);
	cp->seqidp = p;
	p++;
	*p++ = htonl(lock->reclaim);
	cp->stateidp[0] = p;
	p += XDR_QUADLEN(sizeof(stateid4));
	*p++ = htonl((u32)(lock->start >> 32));
	*p++ = htonl((u32)(lock->start));
	*p++ = htonl((u32)(length >> 32));
	*p++ = htonl((u32)(length));

	nfsv4_printk(lock, "nfs4_setup_lock: owner=%d type=%d start=0x%Lx length=0x%Lx\n",
		     lockowner->lo_id, lock->type, lock->start, length);
	
	adjust_args(cp, p);
	cp->nops++;
	cp->flags = CA_SEQID_RELEASE;
	cp->seqid_holder = lockowner;
	cp->nstateids = 1;
	cp->stateid_holder = file_data;
	cp->seqid_index = cp->nops;
	cp->renew_index = cp->nops;
	status = 0;
	
out:
	return status;
out_nomem:
	nfsv4_printk(lock2, "nfs4_setup_lock: out of memory!\n");
	status = -EIO;
	goto out;
}

int
nfs4_setup_lockt(struct compound4args *cp, struct nfs4fs_lockt *lockt)
{
	struct nfs4fs_lockowner *lockowner = lockt->lockowner;
	struct file_lock *lock = lockt->lock;
	struct super_block *sb = lockowner->lo_sb;
	clientid4 clientid;
	u64 length;
	u32 *p;
	int status;

	if ((status = nfs4fs_wait_on_recoverd(cp->sb)))
		goto out;
	nfsv4_printk(level2, "    setting up op #%d: LOCKT\n", cp->nops);
	
	if (!NFS4_HAVE_CLIENTID(sb) && (status = nfs4_do_setclientid(sb))) {
		nfsv4_printk(lock2, "nfs4_setup_lockt: couldn't get a clientid!\n");
		goto out;
	}
	clientid = NFS4_CLIENTID(sb);

	if (!(p = reserve_space(cp, 10)))
		goto out_nomem;
	*p++ = htonl(OP_LOCKT);
	
	switch (lock->fl_type) {    /* encode nfs_lock_type4 */
	case F_RDLCK:
		*p++ = htonl(READ_LT);
		break;
	case F_WRLCK:
		*p++ = htonl(WRITE_LT);
		break;
#if NFS4_DEBUG
	default:
		printk("!!! bad type in nfs4_setup_lockt() !!!\n");
		return -EIO;
#endif
	}
	length = (lock->fl_end != OFFSET_MAX) ? (lock->fl_end - lock->fl_start + 1) : ~(u64)0;
	
	*p++ = htonl((u32)(clientid >> 32));
	*p++ = htonl((u32)(clientid));
	*p++ = htonl(4);
	*p++ = lockowner->lo_id;  /* lockowner ids are always presented in host byte order */
	*p++ = htonl((u32)(lock->fl_start >> 32));
	*p++ = htonl((u32)(lock->fl_start));
	*p++ = htonl((u32)(length >> 32));
	*p++ = htonl((u32)(length));

	nfsv4_printk(lock, "nfs4_setup_lockt: owner=%d type=%d start=%Ld length=%Ld\n",
		     lockowner->lo_id, lock->fl_type, lock->fl_start, length);
	
	adjust_args(cp, p);
	cp->nops++;
	status = 0;

out:
	return status;
out_nomem:
	nfsv4_printk(lock2, "nfs4_setup_lockt: out of memory!\n");
	status = -ENOMEM;
	goto out;
}

int
nfs4_setup_locku(struct compound4args *cp, struct nfs4fs_lock *lock)
{
	struct nfs4fs_file_data *file_data = lock->file_data;
	struct nfs4fs_lockowner *lockowner = file_data->fi_lockowner;
	u64 length;
	u32 *p;
	int status;

	/* XXX: almost identical to nfs4_setup_lock.. maybe extract common code into its own
	   function? */

	if ((status = nfs4fs_wait_on_recoverd(cp->sb)))
		goto out;
	if (!(p = reserve_space(cp, 11)))
		goto out_nomem;
	nfsv4_printk(level2, "    setting up op #%d: LOCKU(%Ld-%Ld) lockowner=%d\n",
		     cp->nops, lock->start, lock->end, lockowner->lo_id);
	length = (lock->end != OFFSET_MAX) ? (lock->end - lock->start + 1) : ~(u64)0;
	
	*p++ = htonl(OP_LOCKU);
	*p++ = htonl(lock->type);
	cp->seqidp = p;
	p++;
	cp->stateidp[0] = p;
	p += XDR_QUADLEN(sizeof(stateid4));
	*p++ = htonl((u32)(lock->start >> 32));
	*p++ = htonl((u32)(lock->start));
	*p++ = htonl((u32)(length >> 32));
	*p++ = htonl((u32)(length));

	nfsv4_printk(lock, "nfs4_setup_locku: owner=%d type=%d start=%Ld length=%Ld\n",
		     lockowner->lo_id, lock->type, lock->start, length);
	
	adjust_args(cp, p);
	cp->nops++;
	cp->flags = CA_SEQID_RELEASE;
	cp->seqid_holder = lockowner;
	cp->nstateids = 1;
	cp->stateid_holder = file_data;
	cp->seqid_index = cp->nops;
	cp->renew_index = cp->nops;
	status = 0;
	
out:
	return status;
out_nomem:
	nfsv4_printk(lock2, "nfs4_setup_locku: out of memory!\n");
	status = -EIO;
	goto out;	
}

int
nfs4_setup_lookup(struct compound4args *cp, struct nfs4fs_lookup *lookup)
{
	struct dentry *dentry = lookup->dentry;
	struct nfs4fs_getattr *getattr = &lookup->getattr;
	int flags = lookup->flags;
	int status;
	
	/* Set up LOOKUP. */
	if ((status = __nfs4_setup_lookup(cp, lookup->name)))
		goto out;
	cp->current_fh = dentry;
	
	/* Set up coalesced GETATTR-GETFH. */
	getattr->flags = flags;
	if ((flags & FL_CLOBBER_OK) || (!dentry->d_inode)) {
		getattr->bm = &nfs4_negative_bitmap;
		lookup->did_getfh = 1;   /* XXX: should depend on fh volatility */
	}
	else {
		getattr->bm = NFS4_NEED_REVAL(dentry->d_inode) ? &nfs4_negative_bitmap :
			&nfs4_positive_bitmap;
		lookup->did_getfh = !NFS4_HAVE_FH(dentry);
	}
	if ((status = nfs4_setup_getattr(cp, getattr)))
		goto out;
	if (lookup->did_getfh && ((status = nfs4_setup_getfh(cp, &lookup->getfh))))
		goto out;

out:
	return status;
}

int
nfs4_setup_open(struct compound4args *cp, struct nfs4fs_open *open)
{
	u32 *p;
	struct super_block *sb = open->dentry->d_sb;
	struct nfs4fs_lockowner *lockowner = open->lockowner;
	struct nfs4fs_file_data *file_data = open->pre_file_data;
	clientid4 clientid;
	int len, quadlen;
	int owner_quadlen = 0;     /* no "might be used uninitialized in this function" */
	int group_quadlen = 0;
	struct gss_cacheent *owner, *group;
	int status;

#if NFS4_DEBUG
	if (!lockowner) {
		printk(KERN_ERR "!!! nfs4_setup_open(): NULL lockowner !!!\n");
		return -EIO;
	}
#endif
	
	/* XXX: Does the OPEN operation renew state? */
	
	if ((status = nfs4fs_wait_on_recoverd(cp->sb)))
		goto out;
	nfsv4_printk(level2, "    setting up op #%d: OPEN(%.*s/%.*s), lockowner %d\n",
		     cp->nops, (int)open->dentry->d_parent->d_name.len,
		     open->dentry->d_parent->d_name.name, (int)open->dentry->d_name.len,
		     open->dentry->d_name.name, lockowner->lo_id);
	
	if (!NFS4_HAVE_CLIENTID(sb) && (status = nfs4_do_setclientid(sb))) {
		nfsv4_printk(open2, "nfs4_setup_open: couldn't get a clientid!\n");
		goto out;
	}
	clientid = NFS4_CLIENTID(sb);

	if (file_data)
		cp->stateid_holder = file_data;
		
	if (!(p = reserve_space(cp, 1)))
		goto out_nomem;
	*p++ = htonl(OP_OPEN);
	adjust_args(cp, p);
	
	switch (open->claim_type) {
	case CLAIM_NULL:
		len = open->dentry->d_name.len;
		quadlen = XDR_QUADLEN(len);
		if (!(p = reserve_space(cp, quadlen + 3)))
			goto out_nomem;
		*p++ = htonl(CLAIM_NULL);
		*p++ = htonl(1);   /* we don't do multi-component open's */
		*p++ = htonl(len);
		memcpy(p, open->dentry->d_name.name, len);
		p += quadlen;
		break;
	case CLAIM_PREVIOUS:
		if (!(p = reserve_space(cp, 1)))
			goto out_nomem;
		*p++ = 0;    /* XXX: mysterious "delegate type" field.. what is it? */
		break;
#if NFS4_DEBUG
	default:
		printk("!!! bad claim type in nfs4_setup_open() !!!\n");
		status = -EIO;
		goto out;
#endif
	}
	adjust_args(cp, p);
	
	if (open->flags & O_CREAT) {
		if ((status = gss_get_name(GSS_OWNER, current->fsuid, &owner))) {
			nfsv4_printk(xdr2, "setup_open: get_name failed!\n");
			goto out;
		}
		if ((status = gss_get_name(GSS_GROUP, current->fsgid, &group))) {
			nfsv4_printk(xdr2, "setup_open: get_name failed!\n");
			gss_put(owner);
			goto out;
		}
		owner_quadlen = XDR_QUADLEN(owner->name.len);
		group_quadlen = XDR_QUADLEN(group->name.len);
		if (!(p = reserve_space(cp, owner_quadlen + group_quadlen + 11))) {
			gss_put(owner);
			gss_put(group);
			goto out_nomem;
		}

		*p++ = htonl(OPEN4_CREATE);
		*p++ = (open->flags & O_EXCL) ? htonl(GUARDED4) : htonl(UNCHECKED4);
		*p++ = htonl(2);
		if (open->flags & O_TRUNC) {		/* bitmap, attrlen, size attribute */
			*p++ = htonl(0x00000010);
			*p++ = htonl(0x00000032);
			*p++ = htonl((owner_quadlen + group_quadlen + 5) << 2);
			*p++ = 0;
			*p++ = 0;
		}
		else {					/* bitmap, attrlen */
			*p++ = 0;
			*p++ = htonl(0x00000032);
			*p++ = htonl((owner_quadlen + group_quadlen + 3) << 2);
		}
		*p++ = htonl(open->mode);
		*p++ = htonl(owner->name.len);
		memcpy(p, owner->name.name, owner->name.len);
		p += owner_quadlen;
		*p++ = htonl(group->name.len);
		memcpy(p, group->name.name, group->name.len);
		p += group_quadlen;
		gss_put(owner);
		gss_put(group);
	}
	else {
		if (!(p = reserve_space(cp, 1)))
			goto out_nomem;
		*p++ = htonl(OPEN4_NOCREATE);
	}
	adjust_args(cp, p);

	if (!(p = reserve_space(cp, 7)))
		goto out_nomem;
	*p++ = htonl((u32)(clientid >> 32));
	*p++ = htonl((u32)(clientid));
	*p++ = htonl(4);
	*p++ = lockowner->lo_id;   /* don't bother htonl'ing, since opaque to server */
	cp->seqidp = p;
	p++;
	*p++ = htonl(open->flags & 3);
	*p++ = 0;
	adjust_args(cp, p);
	cp->nops++;
	cp->flags = 0;
	cp->current_fh = open->dentry;
	cp->seqid_holder = lockowner;
	cp->seqid_index = cp->nops + 1;
	
	open->getattr.flags = FL_CLOBBER_OK;
	open->getattr.bm = &nfs4_negative_bitmap;
	open->getfh.nonfatal = 1;
	if ((status = nfs4_setup_getattr(cp, &open->getattr)))
		goto out;
	status = nfs4_setup_getfh(cp, &open->getfh);

out:
	return status;
out_nomem:
	nfsv4_printk(xdr2, "nfs4_setup_open: out of memory!\n");
	status = -ENOMEM;
	goto out;
}

int
nfs4_setup_open_confirm(struct compound4args *cp, struct nfs4fs_open *open)
{
	u32 *p;
	int status;

	/* XXX: Does the OPEN_CONFIRM operation renew state? */
	/* The passed in lockowner and file_data already have their semaphores claimed */

	if ((status = nfs4fs_wait_on_recoverd(cp->sb)))
		goto out;
	nfsv4_printk(level2, "    setting up op #%d: OPEN_CONFIRM\n", cp->nops);
	
	if (!(p = reserve_space(cp, 4)))
		goto out_nomem;
	*p++ = htonl(OP_OPEN_CONFIRM);
	cp->seqidp = p;
	p++;
	memcpy(p, open->verifier, 8);
	p += 2;
	adjust_args(cp, p);
	cp->nops++;
	cp->flags = CA_SEQID_LOCKED | CA_STATEID_LOCKED | CA_SEQID_RELEASE | CA_NORETRY;
	cp->seqid_holder = open->lockowner;
	cp->seqid_index = cp->nops;
	cp->stateid_holder = open->post_file_data;
	status = 0;
	
out:
	return status;
out_nomem:
	nfsv4_printk(xdr2, "nfs4_setup_open_confirm: out of memory!\n");
	status = -ENOMEM;
	goto out;	
}

int
nfs4_setup_putfh(struct compound4args *cp, struct nfs4fs_putfh *putfh)
{
	struct dentry *dentry = putfh->dentry;
	struct dentry *parent;
	struct nfs4fs_dentry_data *data;
	int quadlen;
	u32 *p;
	int status;

	if (!dentry->d_parent)
		BUG();

	nfsv4_printk(fh, "setup_putfh: starting, fh=%s, parent_fh=%s, mountpoint=%s\n",
		     NFS4_HAVE_FH(putfh->dentry) ? "yes" : "no",
		     NFS4_HAVE_FH(dentry->d_parent) ? "yes" : "no",
		     (dentry == dentry->d_parent) ? "yes" : "no");
	
	if (NFS4_HAVE_FH(dentry)) {
		nfsv4_printk(level2, "    setting up op #%d: PUTFH(%.*s/%.*s)\n", cp->nops,
			     (int)dentry->d_parent->d_name.len, dentry->d_parent->d_name.name,
			     (int)dentry->d_name.len, dentry->d_name.name);
		data = (struct nfs4fs_dentry_data *)dentry->d_fsdata;
		quadlen = XDR_QUADLEN(data->fh_len);
		if (!(p = reserve_space(cp, quadlen + 2)))
			goto out_nomem;
		*p++ = htonl(OP_PUTFH);
		*p++ = htonl(data->fh_len);
		memcpy(p, data->fh_val, data->fh_len);
		p += quadlen;
		adjust_args(cp, p);
		cp->nops++;
		putfh->op = OP_PUTFH;
		putfh->nlookups = 0;
		nfsv4_printk(fh, "nfs4_setup_putfh(%.*s): put filehandle\n",
			     (int)dentry->d_name.len, dentry->d_name.name);
		nfsv4_hexdump(fh, data->fh_val, data->fh_len);
	}
	else {
		nfsv4_printk(fh, "setup_putfh: no fh cached, doing lookup\n");
		parent = dentry->d_parent;
		if (parent == dentry) {		/* at mountpoint */
			int plen = NFS4_PLEN(dentry->d_sb);
			struct qstr *pval = NFS4_PVAL(dentry->d_sb);
			int i;
			
			if (!(p = reserve_space(cp, 1)))
				goto out_nomem;
			nfsv4_printk(level2, "    setting up op #%d: PUTROOTFH\n", cp->nops);
			*p++ = htonl(OP_PUTROOTFH);
			adjust_args(cp, p);
			cp->nops++;
			for (i = 0; i < plen; i++)
				if ((status = __nfs4_setup_lookup(cp, &pval[i])))
					goto out;
			putfh->op = OP_PUTROOTFH;
			putfh->nlookups = plen;
		}
		else {
			if (!NFS4_HAVE_FH(parent)) {  /* yikes, parent fh not cached either! */
				if ((status = nfs4_force_getfh(parent)))
					goto out;
#if NFS4_DEBUG
				if (!(NFS4_HAVE_FH(parent))) {
					printk("!!! force_getfh didn't get fh !!!\n");
					status = -EIO;
					goto out;
				}
#endif
			}
			/* now we are ensured that the parent fh is cached, so PUTFH it and
			   do a LOOKUP. */
			data = (struct nfs4fs_dentry_data *)parent->d_fsdata;
			quadlen = XDR_QUADLEN(data->fh_len);
			if (!(p = reserve_space(cp, quadlen + 2)))
				goto out_nomem;
			nfsv4_printk(level2, "    setting up op #%d: PUTFH(%.*s/%.*s)\n",
				     cp->nops, (int)parent->d_parent->d_name.len,
				     parent->d_parent->d_name.name,
				     (int)parent->d_name.len, parent->d_name.name);
			*p++ = htonl(OP_PUTFH);
			*p++ = htonl(data->fh_len);
			memcpy(p, data->fh_val, data->fh_len);
			p += quadlen;
			adjust_args(cp, p);
			cp->nops++;
			if ((status = __nfs4_setup_lookup(cp, &dentry->d_name)))
				goto out;
			putfh->op = OP_PUTFH;
			putfh->nlookups = 1;
		}
	}

	cp->current_fh = dentry;
	status = 0;
	
out:
	nfsv4_printk(fh, "setup_putfh: returning status %d\n", status);
	return status;
out_nomem:
	nfsv4_printk(xdr2, "nfs4_setup_putfh: out of memory!\n");
	status = -ENOMEM;
	goto out;
}

static inline int
__setup_read(struct compound4args *cp, struct nfs4fs_file_data *file_data,
	     loff_t start, u32 length)
{
	u32 *p;
	int status;
	
	if ((status = nfs4fs_wait_on_recoverd(cp->sb)))
		goto out;
	
	nfsv4_printk(level2, "    setting up op #%d: READ\n", cp->nops);
	if (!(p = reserve_space(cp, 8)))
		goto out_nomem;
	*p++ = htonl(OP_READ);
	cp->stateidp[0] = p;
	p += XDR_QUADLEN(sizeof(stateid4));
	*p++ = htonl((u32)(start >> 32));
	*p++ = htonl((u32)(start));
	*p++ = htonl(length);
	adjust_args(cp, p);
	cp->nops++;
	cp->flags = CA_STATEID_RELEASE;
	cp->renew_index = cp->nops;
	cp->nstateids = 1;
	cp->stateid_holder = file_data;
	status = 0;

out:
	return status;
out_nomem:
	nfsv4_printk(xdr2, "setup_read: out of memory!\n");
	status = -ENOMEM;
	goto out;
}

int
nfs4_setup_simple_read(struct compound4args *cp, struct nfs4fs_simple_read *read)
{
	return __setup_read(cp, read->file_data, read->start, read->length);
}

int
nfs4_setup_read(struct compound4args *cp, struct nfs4fs_read *read)
{
	if (!read->npages || read->npages > NFS4_MAXIO)
		BUG();

	return __setup_read(cp, read->pages[0]->wb_fdata,
			    ((loff_t) read->pages[0]->wb_page->index) << PAGE_CACHE_SHIFT,
			    read->npages * PAGE_CACHE_SIZE);
}

int
nfs4_setup_readdir(struct compound4args *cp, struct nfs4fs_readdir *readdir)
{
	bitmap4 *bm = readdir->bm;
	u32 *p;
	int i;
	int status;

	nfsv4_printk(level2, "    setting up op #%d: READDIR\n", cp->nops);

	if (!(p = reserve_space(cp, 10)))
		goto out_nomem;
	nfsv4_printk(readdir, "nfs4_setup_readdir: cookie=0x%Lx\n", readdir->cookie);
	*p++ = htonl(OP_READDIR);
	memcpy(p, &readdir->cookie, 8);   /* cookie is kept in network byte order */
	p += 2;
	memcpy(p, readdir->verifier, 8);
	p += 2;
	*p++ = htonl(readdir->count >> 4);  /* meaningless "dircount" field */
	*p++ = htonl(readdir->count);
	*p++ = htonl(bm->bitmap4_len);
	for (i = 0; i < bm->bitmap4_len; i++)
		*p++ = htonl(bm->bitmap4_val[i]);
	adjust_args(cp, p);
	cp->nops++;
	status = 0;
	
out:
	return status;
out_nomem:
	nfsv4_printk(xdr2, "setup_readdir: out of memory!\n");
	status = -ENOMEM;
	goto out;
}

int
nfs4_setup_remove(struct compound4args *cp, struct qstr *name)
{
	u32 *p;
	int len = name->len;
	int quadlen = XDR_QUADLEN(len);
	int status;

	nfsv4_printk(level2, "    setting up op #%d: REMOVE(%.*s)\n", cp->nops,
		     (int)name->len, name->name);

	if (!(p = reserve_space(cp, quadlen + 2)))
		goto out_nomem;
	*p++ = htonl(OP_REMOVE);
	*p++ = htonl(len);
	memcpy(p, name->name, len);
	p += quadlen;
	adjust_args(cp, p);
	cp->nops++;
	status = 0;

out:
	return status;
out_nomem:
	nfsv4_printk(xdr2, "setup_remove: out of memory!\n");
	status = -ENOMEM;
	goto out;
}

int
nfs4_setup_rename(struct compound4args *cp, struct nfs4fs_rename *rename)
{
	u32 *p;
	int old_len = rename->old_dentry->d_name.len;
	int old_quadlen = XDR_QUADLEN(old_len);
	int new_len = rename->new_dentry->d_name.len;
	int new_quadlen = XDR_QUADLEN(new_len);
	int status;

	nfsv4_printk(level2, "    setting up op #%d: RENAME(%.*s->%.*s)\n", cp->nops,
		     (int)rename->old_dentry->d_name.len, rename->old_dentry->d_name.name,
		     (int)rename->new_dentry->d_name.len, rename->new_dentry->d_name.name);

	if (!(p = reserve_space(cp, old_quadlen + new_quadlen + 3)))
		goto out_nomem;
	*p++ = htonl(OP_RENAME);
	*p++ = htonl(old_len);
	memcpy(p, rename->old_dentry->d_name.name, old_len);
	p += old_quadlen;
	*p++ = htonl(new_len);
	memcpy(p, rename->new_dentry->d_name.name, new_len);
	p += new_quadlen;
	adjust_args(cp, p);
	cp->nops++;
	status = 0;
	
out:
	return status;
out_nomem:
	nfsv4_printk(xdr2, "setup_rename: out of memory!\n");
	status = -ENOMEM;
	goto out;
}

int
nfs4_setup_renew(struct compound4args *cp, struct nfs4fs_file_data *fp)
{
	u32 *p;
	int status;

	nfsv4_printk(level2, "   setting up op #%d: RENEW\n", cp->nops);
	if (!(p = reserve_space(cp, 5)))
		goto out_nomem;
	*p++ = htonl(OP_RENEW);
	cp->stateidp[0] = p;
	p += XDR_QUADLEN(sizeof(stateid4));
	adjust_args(cp, p);
	cp->nops++;
	cp->flags = CA_STATEID_RELEASE | CA_NORETRY;
	cp->nstateids = 1;
	cp->stateid_holder = fp;
	cp->renew_index = cp->nops;
	status = 0;

out:
	return status;
out_nomem:
	nfsv4_printk(xdr2, "setup_renew: out of memory!\n");
	status = -ENOMEM;
	goto out;
}

int
nfs4_setup_restorefh(struct compound4args *cp)
{
	u32 *p;
	int status;

	nfsv4_printk(level2, "    setting up op #%d: RESTOREFH\n", cp->nops);

	if (!(p = reserve_space(cp, 1)))
		goto out_nomem;
	*p++ = htonl(OP_RESTOREFH);
	adjust_args(cp, p);
	cp->nops++;
	cp->current_fh = cp->save_fh;
	status = 0;

out:
	return status;
out_nomem:
	nfsv4_printk(xdr2, "nfs4_setup_restorefh: out of memory!\n");
	status = -ENOMEM;
	goto out;
}

int
nfs4_setup_reval(struct compound4args *cp, struct nfs4fs_getattr *reval)
{
	struct dentry *dentry = cp->current_fh;
	
#if NFS4_DEBUG
	if (!dentry || !dentry->d_inode) {
		printk("!!! nfs4_setup_reval: no file !!!\n");
		return -EIO;
	}
#endif

	/* XXX: used to be NFS4_NEED_REVAL, not NFS4_UPDATE_TIMEOUT, but our server has
	   an issue with the FATTR4_CHANGE attribute. */
	
	reval->flags = FL_NONFATAL;
	reval->bm = NFS4_UPDATE_TIMEOUT(dentry->d_inode) ? &nfs4_negative_bitmap :
		&nfs4_positive_bitmap;
	return nfs4_setup_getattr(cp, reval);
}

int
nfs4_setup_savefh(struct compound4args *cp)
{
	u32 *p;
	int status;
	
#if NFS4_DEBUG
	if (!cp->current_fh) {
		printk("!!! setup_savefh: current fh not set !!!\n");
		return -EIO;
	}
#endif
	nfsv4_printk(level2, "    setting up op #%d: SAVEFH\n", cp->nops);
	
	if (!(p = reserve_space(cp, 1)))
		goto out_nomem;
	*p++ = htonl(OP_SAVEFH);
	adjust_args(cp, p);
	cp->nops++;
	cp->save_fh = cp->current_fh;
	status = 0;
	
out:
	return status;
out_nomem:
	nfsv4_printk(xdr2, "setup_savefh: out of memory!\n");
	status = -ENOMEM;
	goto out;
}

int
nfs4_setup_setattr(struct compound4args *cp, struct iattr *iap,
		   struct nfs4fs_file_data *fdata)
{
	u32 *p;
	int reserve_length;
	struct gss_cacheent *owner, *group;
	int owner_quadlen = 0;    /* no "might be used uninitialized in this function" */
	int group_quadlen = 0;
	u32 *bmval, *attrlenp, attrlen;
       	int status;

	/* XXX: should we let a SETATTR through during the grace period if it doesn't change
	   the file size? */
	if ((status = nfs4fs_wait_on_recoverd(cp->sb)))
		goto out;
	nfsv4_printk(level2, "    setting up op #%d: SETATTR\n", cp->nops);

	reserve_length = SETATTR_STATICWORDS + 9;
	if (iap->ia_valid & ATTR_UID) {
		if ((status = gss_get_name(GSS_OWNER, iap->ia_uid, &owner))) {
			nfsv4_printk(setattr2, "setup_setattr: get_name failed!\n");
			goto out;
		}
		owner_quadlen = XDR_QUADLEN(owner->name.len);
		reserve_length += (owner_quadlen + 1);
	}
	if (iap->ia_valid & ATTR_GID) {
		if ((status = gss_get_name(GSS_GROUP, iap->ia_gid, &group))) {
			nfsv4_printk(setattr2, "setup_setattr: get_name failed!\n");
			goto out_owner;
		}
		group_quadlen = XDR_QUADLEN(group->name.len);
		reserve_length += (group_quadlen + 1);
	}

	if (!(p = reserve_space(cp, reserve_length)))
		goto out_nomem;
	*p++ = htonl(OP_SETATTR);
	if (fdata) {
		cp->stateidp[0] = p;
		p += XDR_QUADLEN(sizeof(stateid4));
		cp->nstateids = 1;
		cp->stateid_holder = fdata;
		cp->flags = CA_STATEID_RELEASE;
		cp->renew_index = cp->nops + 1;
	}
	else {
		*p++ = 0;
		*p++ = 0;
	}
	*p++ = htonl(2);
	bmval = p;
	*p++ = 0;     /* bitmap val is set to 0, then bits are set as attribute buffer */
	*p++ = 0;     /* is filled in.. */
	attrlenp = p;
	attrlen = 0;
	p++;          /* atttibute buf length is skipped, filled in last.. */
	if (iap->ia_valid & ATTR_SIZE) {
		nfsv4_printk(setattr, "nfs4_setup_setattr: setting size=%d\n",
			     (int)iap->ia_size);
		bmval[0] |= htonl(0x00000010);
		attrlen += 8;
		*p++ = htonl((u32)(iap->ia_size >> 32));
		*p++ = htonl((u32)(iap->ia_size));
	}
	if (iap->ia_valid & ATTR_MODE) {
		nfsv4_printk(setattr, "nfs4_setup_setattr: setting mode=0%o\n", iap->ia_mode);
		bmval[1] |= htonl(0x00000002);
		attrlen += 4;
		*p++ = htonl((u32)(iap->ia_mode));
	}
	if (iap->ia_valid & ATTR_UID) {
		nfsv4_printk(setattr, "nfs4_setup_setattr: setting uid=%d\n", iap->ia_uid);
		bmval[1] |= htonl(0x00000010);
		attrlen += ((owner_quadlen << 2) + 4);
		*p++ = htonl(owner->name.len);
		memcpy(p, owner->name.name, owner->name.len);
		p += owner_quadlen;
	}
	if (iap->ia_valid & ATTR_GID) {
		nfsv4_printk(setattr, "nfs4_setup_setattr: setting gid=%d\n", iap->ia_gid);
		bmval[1] |= htonl(0x00000020);
		attrlen += ((group_quadlen << 2) + 4);
		*p++ = htonl(group->name.len);
		memcpy(p, group->name.name, group->name.len);
		p += group_quadlen;
	}
	if (iap->ia_valid & ATTR_ATIME) {
		/* XXX: is this best, or should we always set to server time? */
		nfsv4_printk(setattr, "nfs4_setup_setattr: setting atime=%d\n",
			     (int)iap->ia_atime);
		bmval[1] |= htonl(0x00010000);
#if 0
		attrlen += 16;
		*p++ = htonl(SET_TO_CLIENT_TIME4);
		*p++ = 0;
		*p++ = htonl((u32)(iap->ia_atime));
		*p++ = 0;
#else
		attrlen += 4;
		*p++ = htonl(SET_TO_SERVER_TIME4);
#endif
	}
	if (iap->ia_valid & ATTR_MTIME) {
		nfsv4_printk(setattr, "nfs4_setup_setattr: setting mtime=%d\n",
			     (int)iap->ia_mtime);
		bmval[1] |= htonl(0x00400000);
#if 0
		attrlen += 16;
		*p++ = htonl(SET_TO_CLIENT_TIME4);
		*p++ = 0;
		*p++ = htonl((u32)(iap->ia_mtime));
		*p++ = 0;
#else
		attrlen += 4;
		*p++ = htonl(SET_TO_SERVER_TIME4);
#endif
	}
	*attrlenp = htonl(attrlen);
	adjust_args(cp, p);
	cp->nops++;
	status = 0;

out_group:
	if (iap->ia_valid & ATTR_GID)
		gss_put(group);
out_owner:
	if (iap->ia_valid & ATTR_UID)
		gss_put(owner);
out:
	nfsv4_printk(xdr, "nfs4_setup_setattr: returning status %d\n", status);
	return status;
out_nomem:
	nfsv4_printk(xdr2, "nfs4_setup_setattr: out of memory!\n");
	status = -ENOMEM;
	goto out_group;
}

int
nfs4_setup_secinfo(struct compound4args *cp, struct nfs4fs_secinfo *secinfo)
{
	u32 *p;
	int quadlen = XDR_QUADLEN(secinfo->name->len);
	int status;

	nfsv4_printk(level2, "    setting up op #%d: SECINFO\n", cp->nops);
	
	nfsv4_printk(secinfo,"nfs4_setup_secinfo\n");
	if (!(p = reserve_space(cp, quadlen + 2)))
		goto out_nomem;
	adjust_args(cp, p);
	*p++ = htonl(OP_SECINFO);
	*p++ = htonl(secinfo->name->len);
	memcpy(p, secinfo->name->name, secinfo->name->len);
	p += quadlen;
	adjust_args(cp, p);
	cp->nops++;
	status = 0;

out:
	nfsv4_printk(secinfo,"nfs4_setup_secinfo returns status %d\n",status);
	return status;
out_nomem:
	nfsv4_printk(xdr2, "nfs4_setup_secinfo: out of memory!\n");
	status = -ENOMEM;
	goto out;
}

int
nfs4_setup_setclientid(struct compound4args *cp, struct nfs4fs_setclientid *setclientid)
{
	u32 *p;
	struct timeval tv;
	int len1, len2, len3;
	int quadlen1, quadlen2, quadlen3;
	int status;

	/* XXX : for now, callback fields are not used. */
	
	nfsv4_printk(level2, "    setting up op #%d: SETCLIENTID\n", cp->nops);
	
	get_fast_time(&tv);

	len1 = setclientid->namlen;
	len2 = strlen(setclientid->callback_netid);
	len3 = strlen(setclientid->callback_universal_address);
	quadlen1 = XDR_QUADLEN(len1);
	quadlen2 = XDR_QUADLEN(len2);
	quadlen3 = XDR_QUADLEN(len3);
	
	if (!(p = reserve_space(cp, quadlen1 + quadlen2 + quadlen3 + 7)))
		goto out_nomem;
	*p++ = htonl(OP_SETCLIENTID);
	*p++ = (u32)tv.tv_sec;     /* verifier -- don't bother htonl()ing, since opaque */
	*p++ = (u32)tv.tv_usec;    /* to server anyway */
	*p++ = htonl(len1);
	memcpy(p, setclientid->name, len1);
	p += quadlen1;
	*p++ = htonl(setclientid->callback_prog);
	*p++ = htonl(len2);
	memcpy(p, setclientid->callback_netid, len2);
	p += quadlen2;
	*p++ = htonl(len3);
	memcpy(p, setclientid->callback_universal_address, len3);
	p += quadlen3;
	adjust_args(cp, p);

	cp->nops++;
	status = 0;

out:
	return status;
out_nomem:
	nfsv4_printk(xdr2, "nfs4_setup_setclientid: out of memory!\n");
	status = -ENOMEM;
	goto out;
}

int
nfs4_setup_setclientid_confirm(struct compound4args *cp, verifier4 verifier)
{
	u32 *p;
	int status;

	nfsv4_printk(level2, "    setting up op #%d: SETCLIENTID_CONFIRM\n", cp->nops);
	
	if (!(p = reserve_space(cp, 3)))
		goto out_nomem;
	*p++ = htonl(OP_SETCLIENTID_CONFIRM);
	memcpy(p, verifier, 8);
	p += 2;
	adjust_args(cp, p);
	cp->nops++;
	status = 0;

out:
	return status;
out_nomem:
	nfsv4_printk(xdr2, "nfs4_setup_setclientid_confirm: out of memory!\n");
	status = -ENOMEM;
	goto out;
}

static inline int
__setup_write(struct compound4args *cp, struct nfs4fs_file_data *file_data,
	      loff_t start, u32 how, u32 length, char *data)
{
	u32 *p;
	int status;
	
	if ((status = nfs4fs_wait_on_recoverd(cp->sb)))
		goto out;
	
	nfsv4_printk(level2, "    setting up op #%d: WRITE(%Ld..%Ld)\n",
		     cp->nops, start, start + length);
	if (!(p = reserve_space(cp, 9)))
		goto out_nomem;
	*p++ = htonl(OP_WRITE);
	cp->stateidp[cp->nstateids++] = p;
	p += XDR_QUADLEN(sizeof(stateid4));
	*p++ = htonl((u32)(start >> 32));
	*p++ = htonl((u32)(start));
	*p++ = htonl(how);
	*p++ = htonl(length);
	adjust_args(cp, p);
	if ((status = send_buf(cp, length, data))) {
		nfsv4_printk(xdr2, "setup_write: send_buf failed!\n");
		goto out;
	}
	cp->nops++;
	cp->flags = CA_STATEID_RELEASE;
	cp->stateid_holder = file_data;
	if (!cp->renew_index)
		cp->renew_index = cp->nops;

out:
	return status;
out_nomem:
	nfsv4_printk(xdr2, "setup_write: out of memory!\n");
	status = -ENOMEM;
	goto out;
}

int
nfs4_setup_simple_write(struct compound4args *cp, struct nfs4fs_simple_write *write)
{
	return __setup_write(cp, write->file_data, write->start, DATA_SYNC4,
			      write->length, write->buf);
}

int
nfs4_setup_write(struct compound4args *cp, struct nfs4fs_write *write)
{
	unsigned int npages = write->npages;
	loff_t start;
	struct nfs4_page *req;
	int i;
	int status = 0;

	if (!npages || npages > NFS4_MAXIO)
		BUG();
	for (i = 0; i < npages; i++) {
		req = write->pages[i];
		start = ((loff_t)req->wb_page->index << PAGE_CACHE_SHIFT) + req->wb_offset;
		nfs4_kmap_request(req);
		status = __setup_write(cp, req->wb_fdata, start, write->how,
				       req->wb_bytes, req->wb_buf + req->wb_offset);
		nfs4_kunmap_request(req);
		if (status)
			break;
	}
	return status;
}

static int
nfs4_force_getfh(struct dentry *dentry)
{
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_putfh putfh;
	struct nfs4fs_getattr reval;
	struct nfs4fs_getfh getfh;
	int status;

	cargs.sb = dentry->d_sb;
	cargs.tag = "nfs4_force_getfh()";
	putfh.dentry = dentry->d_parent;
	getfh.nonfatal = 0;
	if ((status = nfs4_setup_compound(&cargs)))
		goto out;
	if ((status = nfs4_setup_putfh(&cargs, &putfh)))
		goto out_args;
	if ((status = nfs4_setup_reval(&cargs, &reval)))
		goto out_args;
	if ((status = nfs4_setup_getfh(&cargs, &getfh)))
		goto out_args;
	
	if ((status = nfs4_call_compound(&cargs, &cres, NULL)))
		goto out_args;
	if ((status = nfs4_handle_putfh(&cres, &putfh)))
		goto out_res;
	if ((status = nfs4_handle_reval(&cres, &reval)))
		goto out_res;
	if ((status = nfs4_handle_getfh(&cres, &getfh)))
		goto out_res;
	
out_res:
	nfs4_release_compoundres(&cres);
out_args:
	nfs4_release_compoundargs(&cargs);
out:
	return status;
}
