/*
 * linux/fs/nfs4fs/file.c
 *
 * from linux/fs/nfs/file.c
 *    Copyright (C) 1992  Rick Sladkey
 *    Changes Copyright (C) 1994 by Florian La Roche
 *
 * NFSv4 implementation 
 *    Kendrick Smith <kmsmith@umich.edu>
 *    Andy Adamson   <andros@umich.edu>
 *    Copyright (C) 2001 The Regents of the University of Michigan
 */
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/mm.h>
#include <linux/pagemap.h>
#include <linux/file.h>
#include <asm/uaccess.h>

#include <linux/nfs4/nfs4_prot.h>
#include <linux/nfs4/nfs4_debug.h>
#include <linux/nfs4/nfs4.h>
#include <linux/nfs4/nfs4fs.h>

static ssize_t nfs4_read(struct file *filp, char *buf, size_t count, loff_t *ppos);
static ssize_t nfs4_write(struct file *filp, const char *buf, size_t count, loff_t *ppos);
static int nfs4_flush(struct file *filp);
static int nfs4_fsync(struct file *file, struct dentry *dentry, int datasync);
static int nfs4_mmap(struct file *filp, struct vm_area_struct *vma);
static int nfs4_sync_page(struct page *page);
static int nfs4_prepare_write(struct file *filp, struct page *page, unsigned int offset,
			      unsigned int to);
static int nfs4_commit_write(struct file *filp, struct page *page, unsigned int offset,
			     unsigned int to);

struct inode_operations nfs4_file_inode_operations = {
	revalidate:	nfs4_inode_revalidate_wait,
	setattr:	nfs4_setattr,
#ifdef NFS4ACL
       	permission: nfs4_posix_acl_permission,
       	get_posix_acl: nfs4_acl_get_inode,
       	set_posix_acl: nfs4_acl_set_inode,
       	get_ext_attr: nfs4_get_ext_attr,
       	set_ext_attr: nfs4_set_ext_attr
#endif
};

struct file_operations nfs4_file_operations = {
	read: nfs4_read,
	write: nfs4_write,
	mmap: nfs4_mmap,
	open: nfs4_open,
	flush: nfs4_flush,
	release: nfs4_release,
	fsync: nfs4_fsync,
	lock: nfs4_lock
};

struct address_space_operations nfs4_file_aops = {
	readpage: nfs4_readpage,
	writepage: nfs4_writepage,
	sync_page: nfs4_sync_page,
	prepare_write: nfs4_prepare_write,
	commit_write: nfs4_commit_write
};

/* a_ops->sync_page(): "initiates the completion of any page readahead operations" --NFSv3 */
static int
nfs4_sync_page(struct page *page)
{
	struct address_space *mapping;
	struct inode *inode;
	int status = 0;

	mapping = page->mapping;
	if (!mapping)       /* XXX: can this really happen? */
		goto out;
	inode = mapping->host;
	if (!inode)         /* XXX: can this really happen? */
		goto out;
	
	status = nfs4_pagein_inode(inode, page->index, NFS4_RPAGES(inode->i_sb));
	if (status > 0)
		status = 0;

out:
	return status;
}

static int
nfs4_mmap(struct file *filp, struct vm_area_struct *vma)
{
	struct dentry *dentry = filp->f_dentry;
	struct inode *inode = dentry->d_inode;
	struct nfs4fs_file_data *fdata = (struct nfs4fs_file_data *)filp->private_data;
	int status;

	nfsv4_printk(level2, "nfs4_mmap(): starting\n");
	lock_kernel();
			   
#if NFS4_DEBUG
	/* XXX: isn't the first test done in the VFS? */
	if (!(inode->i_mode & S_IFREG))
		BUG();
	if (!fdata)
		BUG();
#endif

	if ((status = nfs4_inode_revalidate(dentry)))
		goto out;
#if 0
	if (vma->vm_flags & VM_READ) {
		if ((filp->f_flags & O_ACCMODE) == O_WRONLY)
			BUG();
		if (list_empty(&fdata->fi_read_map))
			list_add(&fdata->fi_read_map, &NFS4_READ_MAPPINGS(inode));
	}
	if (vma->vm_flags & VM_WRITE) {
		if ((filp->f_flags & O_ACCMODE) == O_RDONLY)
			BUG();
		if (list_empty(&fdata->fi_write_map))
			list_add(&fdata->fi_write_map, &NFS4_WRITE_MAPPINGS(inode));
	}
#else
	/*
	 * XXX: bakeoff hack!
	 */
	if ((filp->f_flags & O_ACCMODE) != O_WRONLY && list_empty(&fdata->fi_read_map))
		list_add(&fdata->fi_read_map, &NFS4_READ_MAPPINGS(inode));
	if ((filp->f_flags & O_ACCMODE) != O_RDONLY && list_empty(&fdata->fi_write_map))
		list_add(&fdata->fi_write_map, &NFS4_WRITE_MAPPINGS(inode));
#endif
	status = generic_file_mmap(filp, vma);

out:
	unlock_kernel();
	nfsv4_printk(level2, "nfs4_mmap(): returning status %d\n\n", status);
	return status;
}

/* a_ops->prepare_write(): called just before generic_file_write() copies data into the
   page from userspace.  page is already locked.  offset=starting offset within page,
   to=final offset within page. */

static int
nfs4_prepare_write(struct file *filp, struct page *page, unsigned int offset, unsigned int to)
{
	nfsv4_printk(level2, "nfs4_prepare_write: %d:[%d,%d]\n", (int)page->index, offset, to);
	kmap(page);
	return nfs4_flush_incompatible(filp, page);
}

/* a_ops->commit_write(): called just after generic_file_write() copies data into the page
   from userspace.  we still have the page lock and kmap(); we will kunmap() after the write
   is finished, but keep the lock.  the return value from this function is the number of
   bytes written successfully, or a negative error status.  a return value of 0 is treated
   as a successful write of all the data, so be careful! */

static int
nfs4_commit_write(struct file *filp, struct page *page, unsigned int offset, unsigned int to)
{
	loff_t pos = ((loff_t)page->index << PAGE_CACHE_SHIFT) + to;
	struct dentry *dentry = filp->f_dentry;
	struct inode *inode = dentry->d_inode;
	int status;

	nfsv4_printk(level2, "nfs4_commit_write %d:[%d,%d]\n", (int)page->index, offset, to);
	kunmap(page);
	
	lock_kernel();
	status = nfs4_updatepage(filp, page, offset, to-offset);
	/* XXX: not sure whether this is necessary; listed as a CHECKME in v2/v3 */
	if (!status && pos > inode->i_size)
		inode->i_size = pos;
	unlock_kernel();
	
	return status;
}

static ssize_t
nfs4_read(struct file *filp, char *buf, size_t count, loff_t *ppos)
{
	unsigned long start_idx, end_idx;
	ssize_t result;

	nfsv4_printk(level2, "nfs4_read(%.*s/%.*s): starting, reading %Ld-%Ld\n",
		     (int)filp->f_dentry->d_parent->d_name.len,
		     filp->f_dentry->d_parent->d_name.name,
		     (int)filp->f_dentry->d_name.len, filp->f_dentry->d_name.name,
		     *ppos, *ppos + count);

	if ((result = nfs4_inode_revalidate(filp->f_dentry)))
		goto out;

	/*
	 * XXX: how does v3 get away with not doing this??
	 */
	start_idx = *ppos >> PAGE_CACHE_SHIFT;
	end_idx = (*ppos + count) >> PAGE_CACHE_SHIFT;
	result = nfs4_sync_file(filp->f_dentry->d_inode, NULL, 0,
				0, FLUSH_WAIT | FLUSH_STABLE);
	if (result)
		goto out;
	
	result = generic_file_read(filp, buf, count, ppos);

out:
	nfsv4_printk(level2, "nfs4_read: returning status %d\n", result);
	return result;
}

static ssize_t
nfs4_write(struct file *filp, const char *buf, size_t count, loff_t *ppos)
{
	ssize_t result = 0;

	/* VFS checks that file has been opened for write */
	/* "hack for future NFS swap support" has been removed */

	nfsv4_printk(level2, "nfs4_write(%.*s/%.*s): starting, writing %Ld-%Ld\n",
		     (int)filp->f_dentry->d_parent->d_name.len,
		     filp->f_dentry->d_parent->d_name.name,
		     (int)filp->f_dentry->d_name.len, filp->f_dentry->d_name.name,
		     *ppos, *ppos + count);

	if ((result = nfs4_inode_revalidate(filp->f_dentry)))
		goto out;
	if (!count)
		goto out;
	result = generic_file_write(filp, buf, count, ppos);
	
out:
	nfsv4_printk(level2, "nfs4_write: returning status %d\n", result);
	return result;
}

/* f_op->flush(): called when a file descriptor is closed.  the return value here is the
   value that will be returned from the close() system call.  locks held: BKL. */
static int
nfs4_flush(struct file *filp)
{
	struct inode *inode = filp->f_dentry->d_inode;
	int status;

	/* Make sure all async reads have been sent off. We don't bother
	 * waiting on them though... */
	if (filp->f_mode & FMODE_READ)
		nfs4_pagein_inode(inode, 0, 0);

	status = nfs4_sync_file(inode, filp, 0, 0, FLUSH_WAIT);
	if (!status) {
		status = filp->f_error;
		filp->f_error = 0;
	}
	return status;
}

/* f_op->fsync(): called on fsync() or fdatasync().  last arg is 0 or 1, respectively, for
   these cases.  locks held: inode->i_sem. */
static int
nfs4_fsync(struct file *file, struct dentry *dentry, int datasync)
{
	int status;

	nfsv4_printk(level2, "nfs4_fsync(): starting on file \"%.*s/%.*s\"\n",
		     (int)dentry->d_parent->d_name.len, dentry->d_parent->d_name.name,
		     (int)dentry->d_name.len, dentry->d_name.name);

	lock_kernel();
	status = nfs4_sync_file(dentry->d_inode, file, 0, 0, FLUSH_WAIT);
	if (!status) {
		status = file->f_error;
		file->f_error = 0;
	}
	unlock_kernel();

	nfsv4_printk(level2, "nfs4_fsync(): returning status %d\n", status);
	return status;
}

/*
 * The remainder of this file is devoted to "generic" (i.e., not specific to
 * either the read or write case) helper routines for I/O.
 */

spinlock_t nfs4_wreq_lock = SPIN_LOCK_UNLOCKED;
DECLARE_WAIT_QUEUE_HEAD(nfs4_wait);
static atomic_t nfs4_nr_requests = ATOMIC_INIT(0);

struct nfs4_page *
nfs4_create_request(struct file *file, struct nfs4fs_file_data *fdata,
		    struct page *page, unsigned int offset, unsigned int count)
{
	struct nfs4_page *req = NULL;
	unsigned long timeout;

	/* caller is responsible for initializing wb_timeout, but this routine will
	   initialize all other fields */

	if (file && file->private_data != fdata)
		BUG();
	if (!fdata)
		BUG();
	if (!fdata->fi_inode)
		BUG();
	if (!page)
		BUG();

	/* Deal with hard/soft limits. */
	do {
		/* If we're over the global soft limit, wake up all requests */
		if (atomic_read(&nfs4_nr_requests) >= NFS4_MAX_REQUEST_SOFT) {
			nfsv4_printk(async, "ASYNC: hit soft limit (%d requests)\n",
			       atomic_read(&nfs4_nr_requests));
			nfs4_wake_flushd();
		}

		/* If we haven't reached the local hard limit yet,
		 * try to allocate the request struct */
		if (atomic_read(&nfs4_nr_requests) < NFS4_MAX_REQUEST_HARD) {
			req = NFS4_ALLOC(sizeof(struct nfs4_page));
			if (req != NULL)
				break;
		}

		/* We're over the hard limit. Wait for better times */
		nfsv4_printk(async, "ASYNC: hit hard limit (%d requests)\n",
		       atomic_read(&nfs4_nr_requests));
		timeout = 1 * HZ;

		/* XXX: should make this conditional on the MOUNT_INTR option (see v3 code) */
		interruptible_sleep_on_timeout(&nfs4_wait, timeout);
		if (signalled())
			break;
	} while (!req);

	if (req) {
		nfsv4_printk(async, "ASYNC: allocating page\n");
		if (file)
			get_file(file);
		atomic_inc(&nfs4_nr_requests);
		atomic_inc(&fdata->fi_count);
		page_cache_get(page);
			
		INIT_LIST_HEAD(&req->wb_list);
		INIT_LIST_HEAD(&req->wb_writelist);
		init_waitqueue_head(&req->wb_wait);
		req->wb_flags = 0;
		req->wb_file = file;
		req->wb_fdata = fdata;
		req->wb_page = page;
		req->wb_inode = fdata->fi_inode;
		req->wb_offset = offset;
		req->wb_bytes = count;
		req->wb_count = 1;
		req->wb_buf = NULL;
		req->wb_safeguard = 0;
	}
	return req;
}

void
_nfs4_release_request(struct nfs4_page *req)
{
	struct super_block *sb;
	
	/* XXX: will need to expand this routine to release cached credential */

	nfsv4_printk(async, "ASYNC: freeing page\n");
	if (!list_empty(&req->wb_list))
		BUG();
	if (!list_empty(&req->wb_writelist))
		BUG();
	if (test_bit(PG_BUSY, &req->wb_flags))
		BUG();
	if (req->wb_buf)
		BUG();
	sb = req->wb_fdata->fi_dentry->d_sb;
	page_cache_release(req->wb_page);
	nfs4_put_file_data(req->wb_fdata);
	if (req->wb_file)
		fput(req->wb_file);
	NFS4_FREE(req);
	atomic_dec(&nfs4_nr_requests);
	wake_up(&nfs4_wait);
	if (atomic_read(&nfs4_nr_requests) < 0)
		BUG();
}

void
nfs4_list_add_request(struct nfs4_page *req, struct list_head *head)
{
	struct list_head *prev;

	/* called with the wreq spinlock held */
	
	if (!list_empty(&req->wb_list))
		BUG();
	if (!test_bit(PG_BUSY, &req->wb_flags))
		BUG();
	
	prev = head->prev;
	while (prev != head) {
		struct nfs4_page *p = list_entry(prev, struct nfs4_page, wb_list);
		if (p->wb_page->index < req->wb_page->index)
			break;
		prev = prev->prev;
	}
	list_add(&req->wb_list, prev);
}

void
nfs4_list_remove_request(struct nfs4_page *req)
{
	/* called with the wreq spinlock held */
	
	if (list_empty(&req->wb_list))
		BUG();
	if (!test_bit(PG_BUSY, &req->wb_flags))
		BUG();
	list_del_init(&req->wb_list);
}

int
nfs4_scan_list(struct list_head *src, struct list_head *dst, struct file *file,
	      unsigned long idx_start, unsigned int npages, unsigned int *bytecount)
{
	struct list_head	*p;
	struct nfs4_page	*req;
	unsigned long		idx_end;
	int			res;

	/* called with the wreq spinlock held */
	
	if (!list_empty(dst))
		BUG();  /* see comment below */
	
	res = 0;
	if (npages == 0)
		idx_end = ~0;
	else
		idx_end = idx_start + npages - 1;
	p = src->next;
	while (p != src) {
		unsigned long pg_idx;

		req = list_entry(p, struct nfs4_page, wb_list);
		p = p->next;

		if (file && req->wb_file != file)
			continue;

		pg_idx = req->wb_page->index;
		if (pg_idx < idx_start || pg_idx > idx_end)
			continue;
		if (!nfs4_lock_request(req))
			continue;

		/* We put the page at the tail of the dst list.  This keeps the list in
		   sorted order because the dst list is assumed to be initially empty
		   (hence the assertion at the top). */
		list_del(&req->wb_list);
		list_add_tail(&req->wb_list, dst);
		if (bytecount)
			*bytecount -= req->wb_bytes;
		res++;
	}
	nfsv4_printk(async, "ASYNC: scan_list() grabbed list of %d pages\n", res);
	return res;
}

int
nfs4_scan_list_timeout(struct list_head *head, struct list_head *dst, struct inode *inode,
		       unsigned int *bytecount)
{
	struct list_head	*p;
        struct nfs4_page	*req;
        int			pages = 0;
	unsigned long		now = jiffies;

	/* called with the wreq spinlock held */

	if (!list_empty(dst))
		BUG();  /* see comment below */
	
	p = head->next;
        while (p != head) {
		req = list_entry(p, struct nfs4_page, wb_list);
		p = p->next;
		if (time_after(req->wb_timeout, now)) {
			if (time_after(NFS4_FLUSHD_NEXTSCAN(inode), req->wb_timeout))
				NFS4_FLUSHD_NEXTSCAN(inode) = req->wb_timeout;
			continue;
		}
		if (!nfs4_lock_request(req))
			continue;
		
		/* We put the page at the tail of the dst list.  This keeps the list in
		   sorted order because the dst list is assumed to be initially empty
		   (hence the assertion at the top). */
		list_del(p);
		list_add_tail(p, dst);
		if (bytecount)
			*bytecount -= req->wb_bytes;
		pages++;
	}
	nfsv4_printk(async, "ASYNC: scan_list_timeout() grabbed list of %d pages\n", pages);
	return pages;
}
