/*
 * linux/fs/nfs4fs/flushd.c
 * from linux/fs/nfs/flushd.c
 * 
 * NFSv4 implementation
 *    Kendrick Smith  <kmsmith@umich.edu>
 *    Andy Adamson  <andros@umich.edu>
 *    Copyright (C) 2001 The Regents of the University of Michigan
 */

#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/mm.h>
#include <linux/pagemap.h>

#include <linux/nfs4/nfs4_prot.h>
#include <linux/nfs4/nfs4_debug.h>
#include <linux/nfs4/nfs4.h>
#include <linux/nfs4/nfs4fs.h>

/*
 * Implementation of the NFSv4 "flush daemon", which flushes pending async
 * I/O requests to the server.
 */

static struct rpc_wait_queue    flushd_queue = RPC_INIT_WAITQ("nfs4_flushd");

static void flushd(struct rpc_task *task);
static void flushd_exit(struct rpc_task *task);

int
nfs4fs_init_flushd(struct super_block *sb)
{
	struct rpc_task *task;
	int status;
	
	lock_kernel();
	
	status = -ENOMEM;
	if (!(task = rpc_new_task(NFS4_CLNT(sb), NULL, RPC_TASK_ASYNC)))
		goto out;
	NFS4_FLUSHD(sb) = task;
	nfsv4_printk(async2, "FLUSHD: created!\n");
	task->tk_calldata = sb;
	task->tk_action = flushd;
	task->tk_exit = flushd_exit;
	rpc_execute(task);
	status = 0;
	
out:
	unlock_kernel();
	return status;
}

void
nfs4_wake_flushd(void)
{
	rpc_wake_up_status(&flushd_queue, -ENOMEM);
}

static inline void
inode_append_flushd(struct inode *inode)
{
	/* caller must hold wreq spinlock */
	if (list_empty(&NFS4_FLUSHD_LIST(inode))) {
		list_add_tail(&NFS4_FLUSHD_LIST(inode), &NFS4_FLUSHD_INODES(inode->i_sb));
		atomic_inc(&inode->i_count);
	}
}

static inline void
inode_remove_flushd(struct inode *inode)
{
	/* caller must hold wreq spinlock.  (NB: this is a lie; the only caller is
	   flushd, which doesn't need the spinlock since it runs under the BKL.) */
	if (!list_empty(&NFS4_FLUSHD_LIST(inode))) {
		list_del_init(&NFS4_FLUSHD_LIST(inode));
		iput(inode);
	}
}

void
nfs4_schedule_inode(struct inode *inode, unsigned long time)
{
	struct rpc_task		*task;
	unsigned long		mintimeout;
	int			empty;
	
	/* caller must hold wreq spinlock */
	/* Doesn't this routine have a small flaw if the inode has not been scheduled
	 * recently (consider, e.g., the first time an inode is ever scheduled)?  Then
	 * NFS4_FLUSHD_NEXTSCAN(inode) will be set to some ancient value, and we will
	 * wake up flushd and flush the inode, even though there is no need to do so...
	 */
	if (time_after(NFS4_FLUSHD_NEXTSCAN(inode), time))
		NFS4_FLUSHD_NEXTSCAN(inode) = time;		
	mintimeout = jiffies + 1 * HZ;
	if (time_before(mintimeout, NFS4_FLUSHD_NEXTSCAN(inode)))
		mintimeout = NFS4_FLUSHD_NEXTSCAN(inode);
	empty = list_empty(&NFS4_FLUSHD_INODES(inode->i_sb));
	inode_append_flushd(inode);

	task = NFS4_FLUSHD(inode->i_sb);
	if (!task)
		BUG();
	if (empty || time_after(NFS4_FLUSHD_TIMEOUT(inode->i_sb), mintimeout))
		rpc_wake_up_task(task);
}

static void
flushd(struct rpc_task *task)
{
	struct super_block *sb = (struct super_block *) task->tk_calldata;
	LIST_HEAD(head);
	struct list_head *p, *q;
	struct inode *inode;
	unsigned long wake_up;
	int flush = (task->tk_status == -ENOMEM);
	
	/* NB: rpciod runs with the BKL held, so we don't need to acquire the
	   wreq spinlock here. */

	nfsv4_printk(async, "FLUSHD: starting\n");
	
	/*
	 * 1. Move all inodes which are scheduled for flushing onto our own private list
	 */
	for (p = NFS4_FLUSHD_INODES(sb).next; p != &NFS4_FLUSHD_INODES(sb); p = q) {
		q = p->next;
		inode = list_entry(p, struct inode, u.nfs4_i.flushd_list);
		if (flush || time_after(jiffies, NFS4_FLUSHD_NEXTSCAN(inode))) {
			list_del(p);
			list_add_tail(p, &head);
		}
	}

	nfsv4_printk(async, "FLUSHD: part 2\n");
	
	/*
	 * 2. Traverse the private list, and start the I/O for each inode.
	 * XXX: The FLUSH_AGING field doesn't seem to do anything (we inherited
	 * this from v3)
	 */
	for (p = head.next; p != &head; p = q) {
		q = p->next;
		inode = list_entry(p, struct inode, u.nfs4_i.flushd_list);
		if (flush) {
			nfsv4_printk(async, "FLUSHD: pageing in inode\n");
			nfs4_pagein_inode(inode, 0, 0);
			nfsv4_printk(async, "FLUSHD: syncing inode\n");
			nfs4_sync_file(inode, NULL, 0, 0, FLUSH_AGING);
		}
		else {
			NFS4_FLUSHD_NEXTSCAN(inode) = jiffies + NFS4_WRITEBACK_LOCKDELAY;
			/*
			 * NB: the next two routines will set NFS4_FLUSHD_NEXTSCAN(inode).
			 */
			nfsv4_printk(async, "FLUSHD: calling nfs4_pagein_timeout()\n");
			nfs4_pagein_timeout(inode);
			nfsv4_printk(async, "FLUSHD: calling nfs4_flush_timeout()\n");
			nfs4_flush_timeout(inode, FLUSH_AGING);
			nfsv4_printk(async, "FLUSHD: calling nfs4_commit_timeout()\n");
			nfs4_commit_timeout(inode, FLUSH_AGING);
		}
		if (list_empty(&NFS4_FLUSHD_READ(inode)) &&
		    list_empty(&NFS4_FLUSHD_WRITE(inode))) {
			nfsv4_printk(async, "FLUSHD: removing inode\n");
			inode_remove_flushd(inode);
		}
		else {
			nfsv4_printk(async, "FLUSHD: rescheduling inode\n");
			list_del(p);
			list_add_tail(p, &NFS4_FLUSHD_INODES(sb));
		}
	}

	nfsv4_printk(async, "FLUSHD: part 3\n");
	
	/*
	 * 3. Determine our next wakeup time, and go to sleep.  Following NFSv3, we
	 * force at least a 1-second timeout.
	 */
	task->tk_timeout = 0;
	if (!list_empty(&NFS4_FLUSHD_INODES(sb))) {
		wake_up = jiffies + NFS4_WRITEBACK_LOCKDELAY;
		list_for_each(p, &NFS4_FLUSHD_INODES(sb)) {
			inode = list_entry(p, struct inode, u.nfs4_i.flushd_list);
			if (time_after(task->tk_timeout, NFS4_FLUSHD_NEXTSCAN(inode)))
				wake_up = NFS4_FLUSHD_NEXTSCAN(inode);
		}
		task->tk_timeout = 1 * HZ;
		if (time_before(jiffies + 1 * HZ, wake_up))
			task->tk_timeout = wake_up - jiffies;
	}
	NFS4_FLUSHD_TIMEOUT(sb) = jiffies + task->tk_timeout;
	task->tk_status = 0;
	task->tk_action = flushd;
	nfsv4_printk(async, "FLUSHD: back to sleep, will awaken in %ld/%d seconds\n",
		     task->tk_timeout, HZ);
	rpc_sleep_on(&flushd_queue, task, NULL, NULL);
}

static void
flushd_exit(struct rpc_task *task)
{
	struct super_block *sb = (struct super_block *) task->tk_calldata;
	struct inode *inode;
	
	/* XXX: revisit this when I understand the umount process better... */

	while (!list_empty(&NFS4_FLUSHD_INODES(sb))) {
		inode = list_entry(NFS4_FLUSHD_INODES(sb).next,
				   struct inode, u.nfs4_i.flushd_list);
		if (NFS4_FLUSHD_NREAD(inode))
			printk(KERN_ERR "NFS4: pending I/O requests during unmount!\n");
		inode_remove_flushd(inode);
	}
	
	NFS4_FLUSHD(sb) = NULL;
	nfsv4_printk(async2, "FLUSHD: destroyed!\n");
}
