/*
 * linux/fs/nfs4fs/inode.c
 * from linux/fs/nfs/inode.c
 *    Copyright (C) 1992  Rick Sladkey
 *
 * NFSv4 implementation
 *    Kendrick Smith  <kmsmith@umich.edu>
 *    Andy Adamson  <andros@umich.edu>
 *    Copyright (C) 2001 The Regents of the University of Michigan
 *
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/pagemap.h>
#include <linux/net.h>
#include <linux/inet.h>
#include <linux/nfs4/nfs4_prot.h>
#include <linux/nfs4/nfs4_debug.h>
#include <linux/nfs4/nfs4.h>
#include <linux/nfs4/nfs4fs.h>
#include <linux/nfs4_mount.h>

static void nfs4_read_inode(struct inode *inode);
static void nfs4_put_super(struct super_block *sb);
static void nfs4_umount_begin(struct super_block *sb);
static int nfs4_statfs(struct super_block *sb, struct statfs *fs);
static void nfs4_clear_inode(struct inode *inode);

static struct super_operations nfs4_sops = {
	read_inode: nfs4_read_inode,
	put_inode: force_delete,         /* called on every iput(), before decreasing count */
	clear_inode: nfs4_clear_inode,   /* called when inode is deleted */
	put_super: nfs4_put_super,
	statfs: nfs4_statfs,
	umount_begin: nfs4_umount_begin,
};

static void
initialize(void)
{
	static int initialized = 0;
	if (!initialized) {
		nfs4fs_init_callbacks();
		nfs4fs_init_state();
		initialized = 1;
	}
}

/* an unforunately named function, ripped from the v3 code.. */
static inline unsigned long
nfs_block_bits(unsigned long bsize, unsigned char *nrbitsp)
{
	/* make sure blocksize is a power of two */
	if ((bsize & (bsize - 1)) || nrbitsp) {
		unsigned char	nrbits;

		for (nrbits = 31; nrbits && !(bsize & (1 << nrbits)); nrbits--)
			;
		bsize = 1 << nrbits;
		if (nrbitsp)
			*nrbitsp = nrbits;
	}

	return bsize;
}

static inline unsigned long
nfs_block_size(unsigned long bsize, unsigned char *nrbitsp)
{
	if (bsize < 1024)
		bsize = NFS4_DEF_FILE_IO_BUFFER_SIZE;
	else if (bsize >= NFS4_MAX_FILE_IO_BUFFER_SIZE)
		bsize = NFS4_MAX_FILE_IO_BUFFER_SIZE;

	return nfs_block_bits(bsize, nrbitsp);
}

static u32 mount_bmval[] = { 0xc810051a, 0x20823a };
static bitmap4 mount_bitmap = { 2, mount_bmval };

static struct super_block *
nfs4_read_super(struct super_block *sb, void *raw_data, int silent)
{
	struct nfs4_mount_data	*data = (struct nfs4_mount_data *) raw_data;
	struct nfs4_sb_info	*sbinfo = &sb->u.nfs4_sb;
	int			len, plen;
	struct qstr		*pval;
	char			*cptr, *placeholder;
	char			*hostname;
	struct compound4args	cargs;
	struct compound4res	cres;
	struct nfs4fs_ctx ctx;
	struct nfs4fs_putfh	putfh;
	struct nfs4fs_getattr	getattr;
	struct nfs4fs_getfh	getfh;
	int			tcp, authflavor;
	struct rpc_timeout	timeparms;
	struct rpc_xprt		*xprt = NULL;
	struct rpc_clnt		*clnt = NULL;
	struct dentry		*root_dentry = NULL;
	int			i;

	/* XXX : no need to lock_super() in the 2.4 kernel? */

	nfsv4_printk(level2, "nfs4_read_super: starting\n");
	initialize();

	/*
	 *  1.  Check the verson #, hostname, remote address, and mount path passed down by
	 *      the mount program
	 */

	if (!data) {
		printk("nfs4_read_super: no data from mount program!\n");
		goto out_fail;
	}
	if (data->version != NFS4_MOUNT_VERSION) {
		printk("nfs4: mount version %s than kernel\n",
			data->version < NFS4_MOUNT_VERSION ? "older" : "newer");
		printk("nfs4: maybe you need to bring your mount program up-to-date?\n");
		if (data->version < NFS4_MOUNT_VERSION)
			/* XXX : for the time being, this is a fatal error */
			goto out_fail;
	}

	for (i = 0; i < sizeof(data->hostname); i++)
		if (!data->hostname[i])
			break;
#if 0
	if (i != data->namlen) {
		printk("nfs4_read_super: data->namlen=%d, but strlen(hostname)=%d!\n",
		       data->namlen, i);
		goto out_fail;
	}
#else
	/* XXX : The value of data->namlen passed down by mount doesn't appear to be equal to
	   strlen(hostname), so overwrite data->namlen with strlen(hostname) */
	data->namlen = i;
#endif
	hostname = NFS4_ALLOC(data->namlen + 1);
	memcpy(hostname, data->hostname, data->namlen);
	hostname[data->namlen] = 0;

	if (data->addr.sin_addr.s_addr == INADDR_ANY) {
		printk("nfs4_read_super: mount program didn't pass remote address!\n");
		goto out_free_hostname;
	}

	plen = 1;
	for (len = 0; (len < sizeof(data->mnt_path)) && data->mnt_path[len]; len++)
		if (data->mnt_path[len] == '/')
			plen++;
	if (len == sizeof(data->mnt_path)) {
		printk("nfs4_read_super: mount passed non null-terminated mount path!\n");
		goto out_free_hostname;
	}
	if (!(sbinfo->mnt_path = NFS4_ALLOC(len+1))) {
		printk("nfs4_read_super: couldn't alloc mount path!\n");
		goto out_free_hostname;
	}	
	memcpy(sbinfo->mnt_path, data->mnt_path, len);
	sbinfo->mnt_path[len] = 0;
	if (!(pval = NFS4_ALLOC(plen * sizeof(struct qstr)))) {
		printk("nfs4_read_super: couldn't alloc mount path!\n");
		goto out_free_pathname2;
	}
	plen = 0;
	cptr = sbinfo->mnt_path;
	for (;;) {
		while (*cptr == '/')
			cptr++;
		if (!*cptr)
			break;
		pval[plen].name = placeholder = cptr;
		while (*cptr && (*cptr != '/'))
			cptr++;
		pval[plen].len = cptr - placeholder;   /* note: pval->hash not used */
		plen++;
	}

	nfsv4_printk(mount, "nfs4_read_super: arguments read successfully\n");
	nfsv4_printk(mount, "nfs4_read_super: hostname=\"%s\"\n", hostname);
	nfsv4_printk(mount, "nfs4_read_super: remote addr = %s\n",
		     in_ntoa(data->addr.sin_addr.s_addr));
	nfsv4_printk(mount, "nfs4_read_super: mount_path=/");
	for (i = 0; i < plen; i++)
		nfsv4_printk(mount, "%.*s/", (int)pval[i].len, pval[i].name);
	nfsv4_printk(mount, "\n");

	/*
	 *  2.  Set up state for the compound RPC which will get the root filehandle
	 *      and attributes
	 */
	
	/* Which protocol do we use? */
	/* XXX : for now, ignore the NFS4_MOUNT_TCP flag and always use UDP */
	tcp = 0;

	/* Initialize timeout values */
	/* XXX : ripped from v3 code.. come back to this and rethink it! */
	timeparms.to_initval = (data->timeo * HZ / 10);
	timeparms.to_retries = data->retrans ?: 5;
	timeparms.to_maxval  = tcp? RPC_MAX_TCP_TIMEOUT : RPC_MAX_UDP_TIMEOUT;
	timeparms.to_exponential = 1;
	if (!timeparms.to_initval)
		/* XXX : where do these values come from? */
		timeparms.to_initval = (tcp ? 600 : 11) * HZ / 10;

	/* create RPC transport */
	if (!(xprt = xprt_create_proto(tcp? IPPROTO_TCP : IPPROTO_UDP,
				       &data->addr, &timeparms))) {
		nfsv4_printk(mount2, "nfs4_read_super: couldn't create RPC xprt!\n");
		goto out_free_pathname;
	}
	nfsv4_printk(mount, "nfs4_read_super: RPC transport created successfully\n");

	/* Choose authentication flavor */
	/* XXX : for now, always AUTH_NULL */
	authflavor = RPC_AUTH_UNIX;

	/* create RPC client state */
	if (!(clnt = rpc_create_client(xprt, hostname, &nfs4_program, 4, authflavor))) {
		nfsv4_printk(mount2, "nfs4_read_super: couldn't create RPC clnt!\n");
		goto out_destroy_xprt;
	}
	clnt->cl_intr     = (data->flags & NFS4_MOUNT_INTR)? 1 : 0;
	clnt->cl_softrtry = (data->flags & NFS4_MOUNT_SOFT)? 1 : 0;
	clnt->cl_chatty   = 1;
	nfsv4_printk(mount, "nfs4_read_super: RPC client state created successfully\n");

	if (rpciod_up() != 0) {
		nfsv4_printk(mount2, "nfs4_read_super: couldn't start rpciod!\n");
		goto out_destroy_clnt;
	}
	nfsv4_printk(mount, "nfs4_read_super: rpciod running\n");

	/*
	 *  3.  instantiate the root of the filesystem; make the COMPOUND call
	 */

	/* before instantiating root inode, a few fields must be set.
	   sb->s_op:  since s_op->read_inode() will be called by nfs4_handle_getattr()
	   sbinfo->{plen,pval}:  since nfs4_setup_putfh() needs them.
	   sbinfo->client:  must be set before using ctx in nfs4_call_compound() */

	sb->s_op = &nfs4_sops;
	sbinfo->client = clnt;
	sbinfo->plen = plen;
	sbinfo->pval = pval;

	if (!(root_dentry = d_alloc(NULL, &(const struct qstr) { "/", 1, 0 }))) {
		printk("nfs4_read_super: couldn't alloc root dentry!\n");
		goto out_rpciod_down;
	}
	root_dentry->d_sb = sb;
	root_dentry->d_parent = root_dentry;
	root_dentry->d_op = &nfs4_dentry_operations;
	cargs.sb = sb;
	cargs.tag = "nfs4_read_super()";
	ctx.pt_sb = sb; 
	ctx.pt_dentry = root_dentry;
	putfh.dentry = root_dentry;
	getattr.flags = 0;
	getattr.bm = &mount_bitmap;
	getfh.nonfatal = 0;
	if (nfs4_setup_compound(&cargs))
		goto out_dput;
	if (nfs4_setup_putfh(&cargs, &putfh))
		goto out_args;
	if (nfs4_setup_getattr(&cargs, &getattr))
		goto out_args;
	if (nfs4_setup_getfh(&cargs, &getfh))
		goto out_args;
	
	if (nfs4_call_compound(&cargs, &cres, &ctx))
		goto out_args;
	if (nfs4_handle_putfh(&cres, &putfh))
		goto out_res;
	if (nfs4_handle_getattr(&cres, &getattr))
		goto out_res;
	if (nfs4_handle_getfh(&cres, &getfh))
		goto out_res;

	nfs4_release_compoundres(&cres);
	nfs4_release_compoundargs(&cargs);
	nfsv4_printk(mount, "nfs4_read_super: root of filesystem instantiated\n");

	/*
	 *  4.  Start renewd for this superblock, but first set a few fields which
	 *  renewd will need for successful operation.
	 */
	
	if (getattr.fattr.fa_valid & FA_LEASE_TIME) {
		sbinfo->lease_time = getattr.fattr.fa_lease_time * HZ;
		if (sbinfo->lease_time < NFS4FS_MIN_LEASE)
			sbinfo->lease_time = NFS4FS_MIN_LEASE;
		if (sbinfo->lease_time > NFS4FS_MAX_LEASE)
			sbinfo->lease_time = NFS4FS_MAX_LEASE;
	}
	else
		sbinfo->lease_time = NFS4FS_DEFAULT_LEASE;
	INIT_LIST_HEAD(&sbinfo->file_list);
	sbinfo->last_renewal = jiffies;
	nfsv4_printk(mount, "lease time = %d/%d\n", sbinfo->lease_time, HZ);
	
	if (nfs4fs_init_renewd(sb)) {
		printk(KERN_INFO "nfs4: couldn't start renewd for this mountpoint!\n");
		goto out_dput;
	}

	/* ACL Temp code */
#ifdef NFS4ACL
	sb->s_ext_attr_flags = 0;
#ifdef CONFIG_EXT2_FS_EXT_ATTR
	sb->s_ext_attr_flags |= EXT_ATTR_FLAG_EXT_ATTR;
#ifdef CONFIG_FS_USER_EXT_ATTR
	sb->s_ext_attr_flags |= EXT_ATTR_FLAG_EXT_ATTR_USER;
#endif
#ifdef CONFIG_FS_POSIX_ACL
        sb->s_ext_attr_flags |= EXT_ATTR_FLAG_POSIX_ACL;
#endif
#endif
#endif
	/* end of ACL Temp code */

	/*
	 *  5.  Start flushd for this superblock.
	 */
	
	sbinfo->flushd_timeout = jiffies;
	INIT_LIST_HEAD(&sbinfo->flushd_inodes);
	if (nfs4fs_init_flushd(sb)) {
		printk(KERN_INFO "nfs4: couldn't start flushd for this mountpoint!\n");
		goto out_dput;
	}

	/*
	 *  6.  If we made it this far, the mount will succeed.
	 *      Start the callback service for this superblock.
	 *      XXX: POST BAKEOFF: create a mount flag for callbacks!
	 */
	if (nfs4_create_callback(sb)) {
		printk(KERN_WARNING "nfs4: warning: couldn't start callback service\n"
		       "nfs4: warning: Delegations will not be available for this mountpoint\n");
	}
	
	/*
	 *  7.  Finish up by setting the necessary fields in the superblock.
	 */
	sbinfo->flags = data->flags & NFS4_MOUNT_FLAGMASK;  /* XXX : not using all these! */
	sbinfo->rsize = nfs_block_size(data->rsize, NULL);
	sbinfo->wsize = nfs_block_size(data->wsize, NULL);
	if ((getattr.fattr.fa_valid & FA_MAXREAD) &&
	    (sbinfo->rsize > getattr.fattr.fa_maxread)) {
		printk(KERN_INFO "nfs4: rsize=%d too large for server\n", sbinfo->rsize);
		sbinfo->rsize = nfs_block_bits(getattr.fattr.fa_maxread, NULL);
		printk(KERN_INFO "nfs4: setting rsize=%d\n", sbinfo->rsize);
	}
	if ((getattr.fattr.fa_valid & FA_MAXWRITE) &&
	    (sbinfo->wsize > getattr.fattr.fa_maxwrite)) {
		printk(KERN_INFO "nfs4: wsize=%d too large for server\n", sbinfo->wsize);
		sbinfo->wsize = nfs_block_bits(getattr.fattr.fa_maxwrite, NULL);
		printk(KERN_INFO "nfs4: setting wsize=%d\n", sbinfo->wsize);
	}
	sbinfo->rpages = (sbinfo->rsize + PAGE_CACHE_SIZE - 1) >> PAGE_CACHE_SHIFT;
	sbinfo->wpages = (sbinfo->wsize + PAGE_CACHE_SIZE - 1) >> PAGE_CACHE_SHIFT;
	if (sbinfo->rpages > NFS4_MAXIO) {
		printk(KERN_INFO "nfs4: rpages=%d too large; setting rpages=%d\n",
		       sbinfo->rpages, NFS4_MAXIO);
		printk(KERN_INFO "nfs4: to use a larger rpages, increase NFS4_MAXIO\n");
		sbinfo->rpages = NFS4_MAXIO;
	}
	if (sbinfo->wpages > NFS4_MAXIO) {
		printk(KERN_INFO "nfs4: wpages=%d too large; setting wpages=%d\n",
		       sbinfo->wpages, NFS4_MAXIO);
		printk(KERN_INFO "nfs4: to use a larger rpages, increase NFS4_MAXIO\n");
		sbinfo->wpages = NFS4_MAXIO;
	}
	sbinfo->dtsize = (sbinfo->rsize < PAGE_CACHE_SIZE) ? sbinfo->rsize : PAGE_CACHE_SIZE;
	sbinfo->strict_locking = data->strict_locking;
	sbinfo->silly_rename = data->silly_rename;
	sbinfo->lock_mintimeout = 1 * HZ;
	sbinfo->lock_maxtimeout = 30 * HZ;
	sbinfo->pidof_recoverd = 0;
	init_waitqueue_head(&sbinfo->grace_wait_queue);
	
	nfsv4_printk(mount, "nfs4_read_super: rsize=%d wsize=%d dtsize=%d lease=%d\n",
		     sbinfo->rsize, sbinfo->wsize, sbinfo->dtsize, sbinfo->lease_time);
	nfsv4_printk(mount, "nfs4_read_super: rpages=%d wpages=%d\n",
		     sbinfo->rpages, sbinfo->wpages);
	nfsv4_printk(mount, "nfs4_read_super: silly_rename=%s strict_locking=%s\n",
		     sbinfo->silly_rename ? "yes" : "no",
		     sbinfo->strict_locking ? "yes" : "no");
	
	if (data->flags & NFS4_MOUNT_NOAC) {
		data->acregmin = data->acregmax = 0;
		data->acdirmin = data->acdirmax = 0;
	}
	sbinfo->acregmin = data->acregmin*HZ;
	sbinfo->acregmax = data->acregmax*HZ;
	sbinfo->acdirmin = data->acdirmin*HZ;
	sbinfo->acdirmax = data->acdirmax*HZ;
	nfsv4_printk(mount, "nfs4_read_super: regmin=%d regmax=%d dirmin=%d dirmax=%d\n",
		     sbinfo->acregmin, sbinfo->acregmax, sbinfo->acdirmin, sbinfo->acdirmax);
	
	sbinfo->namelen = data->namlen;
	sbinfo->hostname = hostname;
	sbinfo->have_clientid = 0;
	sbinfo->clientid = 0;
	memcpy(NFS4_IP_ADDR(sb), data->ipaddr, 16);

	/* XXX : ugh! setting sb->s_blocksize is kind of a pain.. what is this value even
	   used for, anyway?  for the time being, do the following: if data->bsize is
	   specified, use it.  otherwise, if the server returns both a maxread and a maxwrite
	   attr, use the smaller.  otherwise, use NFS4_DEF_FILE_IO_BUFFER_SIZE */
	sb->s_root = root_dentry;
	sb->s_magic = NFS4_SUPER_MAGIC;
	if (getattr.fattr.fa_valid & FA_MAXFILESIZE)
		sb->s_maxbytes = getattr.fattr.fa_maxfilesize;
	else {
		printk(KERN_INFO "nfs4: warning: couldn't read max. file size from server\n");
		sb->s_maxbytes = NFS4_DEF_MAXFILESIZE;
	}
	if (!data->bsize) {
		if ((getattr.fattr.fa_valid & FA_MAXREAD) &&
		    (getattr.fattr.fa_valid & FA_MAXWRITE))
			data->bsize = (getattr.fattr.fa_maxread < getattr.fattr.fa_maxwrite) ?
				getattr.fattr.fa_maxread : getattr.fattr.fa_maxwrite;
		else
			data->bsize = NFS4_DEF_FILE_IO_BUFFER_SIZE;
	}
	sb->s_blocksize = nfs_block_bits(data->bsize, &sb->s_blocksize_bits);
	nfsv4_printk(mount, "nfs4_read_super: s_blocksize=%ld\n", sb->s_blocksize);
	
out:
	nfsv4_printk(level2, "nfs4_read_super: returning %s\n\n", sb ? "success" : "failure");
	return sb;

out_res:
	nfs4_release_compoundres(&cres);
out_args:
	nfs4_release_compoundargs(&cargs);
out_dput:
	dput(root_dentry);
out_rpciod_down:
	rpciod_down();
out_destroy_clnt:
	rpc_shutdown_client(clnt);
	goto out_free_pathname;
out_destroy_xprt:
	xprt_destroy(xprt);
out_free_pathname:
	NFS4_FREE(pval);
out_free_pathname2:
	NFS4_FREE(sbinfo->mnt_path);
out_free_hostname:
	NFS4_FREE(hostname);
out_fail:
	sb = NULL;
	goto out;
}

static void
nfs4_read_inode(struct inode *inode)
{
	/* XXX : ripped from v3 code.. is this really what we want? */
	inode->i_blksize = inode->i_sb->s_blocksize;
	inode->i_mode = 0;
	inode->i_rdev = 0;
	memset(&inode->u.nfs4_i, 0, sizeof(struct nfs4_inode_info));
}

#if 0
static void
nfs4_delete_inode(struct inode *inode)
{
	struct nfs4fs_close close;
	struct nfs4fs_file_data *read = NFS4_READ_OPENFILE(inode);
	struct nfs4fs_file_data *write = NFS4_WRITE_OPENFILE(inode);
	
	lock_kernel();

	/* XXX: when async writebacks are implemented, will need to wait on them here.. */

	if (read) {
		close.data = read;
		close.read = 1;
		close.write = 0;
		if (nfs4fs_do_close(read->fi_dentry, &close))
			printk("!!! nfs4_delete_inode: couldn't close file !!!\n");
		nfs4_put_file_data(read);
		NFS4_READ_OPENFILE(inode) = NULL;
	}
	if (write) {
		close.data = write;
		close.read = 1;
		close.write = 1;
		if (nfs4fs_do_close(read->fi_dentry, &close))
			printk("!!! nfs4_delete_inode: couldn't close file !!!\n");
		nfs4_put_file_data(write);
		NFS4_WRITE_OPENFILE(inode) = NULL;
	}

	clear_inode(inode);
	unlock_kernel();
}
#endif

static void
nfs4_clear_inode(struct inode *inode)
{
	struct nfs4fs_delegation *dp;

	spin_lock(&nfs4_wreq_lock);
	dp = NFS4_DELEGATION(inode);
	if (dp) {
		list_del(&dp->dl_hash);
		NFS4_DELEGATION(inode) = NULL;
	}
	spin_unlock(&nfs4_wreq_lock);
	
	if (!dp)
		return;
	nfs4_do_delegreturn(inode->i_sb, dp->dl_stateid);
	NFS4_FREE(dp);
}

static void
nfs4_umount_begin(struct super_block *sb)
{
	struct rpc_clnt *clnt = sb->u.nfs4_sb.client;

	if (clnt)
		rpc_killall_tasks(clnt);   /* Note: this will kill renewd and flushd */
}

static int
nfs4_statfs(struct super_block *sb, struct statfs *fs)
{
	static u32 statfs_bmval[] = { 0x20c00000, 0x00001400};
	static bitmap4 statfs_bitmap = { 2, statfs_bmval };
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_getattr getattr;
	struct nfs4fs_putfh putfh;
	int bits;
	int status;

	nfsv4_printk(level2, "nfs4_statfs: starting\n");

	cargs.sb = sb;
	cargs.tag = "nfs4_statfs()";
	putfh.dentry = sb->s_root;
	getattr.flags = FL_NOREVAL;
	getattr.bm = &statfs_bitmap;
	if ((status = nfs4_setup_compound(&cargs)))
		goto out;
	if ((status = nfs4_setup_putfh(&cargs,&putfh)))
		goto out_args;
	if ((status = nfs4_setup_getattr(&cargs, &getattr)))
		goto out_args;

	if ((status = nfs4_call_compound(&cargs, &cres,(struct nfs4fs_ctx *)NULL)))
		goto out_args;
	if ((status = nfs4_handle_putfh(&cres,&putfh)))
		goto out_res;
	if ((status = nfs4_handle_getattr(&cres, &getattr)))
		goto out_res;
	
	bits = sb->s_blocksize_bits;
	fs->f_type = NFS4_SUPER_MAGIC;
	fs->f_bsize = sb->s_blocksize;
	fs->f_blocks = (getattr.fattr.fa_valid & FA_STOTAL) ?
		(getattr.fattr.fa_stotal >> bits) : (1 << 20);
	fs->f_bavail = (getattr.fattr.fa_valid & FA_SAVAIL) ?
		(getattr.fattr.fa_savail >> bits) : fs->f_blocks;
	fs->f_files = (getattr.fattr.fa_valid & FA_FTOTAL) ?
		(getattr.fattr.fa_ftotal >> bits) : (1 << 20);
	fs->f_ffree = (getattr.fattr.fa_valid & FA_FFREE) ?
		(getattr.fattr.fa_ffree >> bits) : fs->f_files;
	fs->f_fsid.val[0] = fs->f_fsid.val[1] = 0;   /* XXX : how can we do this right? */
	fs->f_namelen = (getattr.fattr.fa_valid & FA_MAXNAME) ?
		(getattr.fattr.fa_maxname) : NAME_MAX;

out_res:
	nfs4_release_compoundres(&cres);
out_args:
	nfs4_release_compoundargs(&cargs);
out:
	nfsv4_printk(level2, "nfs4_statfs: returning status %d\n\n", status);
	return status;
}

static void
nfs4_put_super(struct super_block *sb)
{
	struct nfs4_sb_info *sbinfo = &sb->u.nfs4_sb;

	nfs4_destroy_callback(sb);
	if (sbinfo->client)
		rpc_shutdown_client(sbinfo->client);
	rpciod_down();
	NFS4_FREE(sbinfo->pval);
	NFS4_FREE(sbinfo->mnt_path);
	NFS4_FREE(sbinfo->hostname);
}

int
nfs4_inode_revalidate(struct dentry *dentry)
{
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_putfh putfh;
	struct nfs4fs_getattr getattr;
	struct nfs4fs_getfh getfh;
	int do_getfh;
	int status = 0;

#if NFS4_DEBUG
	if (!dentry->d_inode) {
		printk("!!! nfs4_inode_revalidate: dentry is negative !!!\n");
		status = -EIO;
		goto out;
	}
#endif
	
	nfsv4_printk(level2, "nfs4_inode_revalidate: starting on \"%.*s/%.*s\"\n",
		     (int)dentry->d_parent->d_name.len, dentry->d_parent->d_name.name,
		     (int)dentry->d_name.len, dentry->d_name.name);
	lock_kernel();

	nfsv4_printk(reval, "nfs4_inode_revalidate: jiffies=%ld, cjiffies=%ld, timeout=%ld\n",
		     jiffies, NFS4_CACHE_JIFFIES(dentry->d_inode),
		     NFS4_CACHE_TIMEOUT(dentry->d_inode));

	if (!NFS4_NEED_REVAL(dentry->d_inode) && NFS4_HAVE_FH(dentry)) {
		nfsv4_printk(reval, "nfs4_inode_revalidate: inode will not be refreshed\n");
		goto out;
	}
	nfsv4_printk(reval, "nfs4_inode_revalidate: inode will be refreshed\n");
	
	cargs.sb = dentry->d_sb;
	cargs.tag = "nfs4_inode_revalidate()";
	putfh.dentry = dentry;
	getattr.flags = 0;
	getattr.bm = &nfs4_negative_bitmap;
	do_getfh = !NFS4_HAVE_FH(dentry);
	getfh.nonfatal = 1;
	if ((status = nfs4_setup_compound(&cargs)))
		goto out;
	if ((status = nfs4_setup_putfh(&cargs, &putfh)))
		goto out_args;
	if ((status = nfs4_setup_getattr(&cargs, &getattr)))
		goto out_args;
	if (do_getfh && ((status = nfs4_setup_getfh(&cargs, &getfh))))
		goto out_args;

	if ((status = nfs4_call_compound(&cargs, &cres, NULL)))
		goto out_args;
	if ((status = nfs4_handle_putfh(&cres, &putfh)))
		goto out_res;
	if ((status = nfs4_handle_getattr(&cres, &getattr)))
		goto out_res;
	if (do_getfh)
		nfs4_handle_getfh(&cres, &getfh);   /* ignore return value */

out_res:
	nfs4_release_compoundres(&cres);
out_args:
	nfs4_release_compoundargs(&cargs);
out:
	unlock_kernel();
	nfsv4_printk(level2, "nfs4_inode_revalidate: returning status %d\n\n", status);
	return status;
}

int
nfs4_inode_revalidate_wait(struct dentry *dentry)
{
	int status;
	
	/*
	 * We wait on pending write requests, before going through the usual revalidation
	 * logic.  This is to ensure that stat() sees the proper i_size after writes have
	 * been done.  XXX: how does the normal nfs client get around this problem??
	 */
	nfsv4_printk(level2, "nfs4_inode_revalidate_wait: starting on \"%.*s/%.*s\"\n",
		     (int)dentry->d_parent->d_name.len, dentry->d_parent->d_name.name,
		     (int)dentry->d_name.len, dentry->d_name.name);
	
	if (!dentry->d_inode)
		BUG();
	if ((status = nfs4_sync_file(dentry->d_inode, NULL, 0, 0, FLUSH_WAIT|FLUSH_SYNC)))
		goto out;
	status = nfs4_inode_revalidate(dentry);

out:
	nfsv4_printk(level2, "nfs4_inode_revalidate_wait: returning status %d\n\n", status);
	return status;
}

/* i_op->setattr().  locks held: BKL, also inode->i_sem if size change requested */
int
nfs4_setattr(struct dentry *dentry, struct iattr *iap)
{
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_putfh putfh;
	struct nfs4fs_getattr getattr;
	struct nfs4fs_file_data *file_data;
	struct nfs4fs_open open;
	struct nfs4fs_close close;
	int status;
	
	nfsv4_printk(level2, "nfs4_setattr: starting on \"%.*s\", iap->ia_valid=%d\n",
		     (int)dentry->d_name.len, dentry->d_name.name, iap->ia_valid);
	
	/* XXX: Is this really needed?  It's in v3... */
	if (!S_ISREG(dentry->d_inode->i_mode))
		iap->ia_valid &= ~ATTR_SIZE;

	if ((status = nfs4_sync_file(dentry->d_inode, NULL, 0, 0, FLUSH_WAIT)))
		goto out;		
	
	file_data = NULL;
	if (iap->ia_valid & ATTR_SIZE) {
		status = -ENOMEM;
		if (!(file_data = nfs4_get_file_data(dentry)))
			goto out;
		open.claim_type = CLAIM_NULL;
		open.dentry = dentry;
		open.flags = FMODE_WRITE;
		open.mode = 0;
		open.lockowner = file_data->fi_lockowner;
		open.pre_file_data = file_data;
		if ((status = nfs4fs_do_open(&open)))
			goto out_put;
		if (open.post_file_data != file_data) {
			/* XXX: leaking state here! */
			printk(KERN_WARNING "nfs4_setattr: didn't open right file!\n");
			nfs4_put_file_data(open.post_file_data);
			goto out_put;
		}
		nfs4_put_file_data(file_data);    /* we now have 2 references */
	}
	
	cargs.sb = dentry->d_sb;
	cargs.tag = "nfs4_setattr()";
	putfh.dentry = dentry;
	getattr.flags = FL_NONFATAL;
	getattr.bm = &nfs4_negative_bitmap;
	
	if ((status = nfs4_setup_compound(&cargs)))
		goto out_close;
	if ((status = nfs4_setup_putfh(&cargs, &putfh)))
		goto out_args;
	if ((status = nfs4_setup_setattr(&cargs, iap, file_data)))
		goto out_args;
	if ((status = nfs4_setup_getattr(&cargs, &getattr)))
		goto out_args;

	if ((status = nfs4_call_compound(&cargs, &cres,(struct nfs4fs_ctx *)NULL)))
		goto out_args;
	if ((status = nfs4_handle_putfh(&cres, &putfh)))
		goto out_res;
	if ((status = nfs4_handle_setattr(&cres)))
		goto out_res;
	nfs4_handle_getattr(&cres, &getattr);    /* ignore return value */

out_res:
	nfs4_release_compoundres(&cres);
out_args:
	nfs4_release_compoundargs(&cargs);
out_close:
	if (iap->ia_valid & ATTR_SIZE) {
		close.data = file_data;
		close.read = 0;
		close.write = 1;
		if (nfs4fs_do_close(dentry, &close))
			printk(KERN_WARNING "warning: close in nfs4_setattr failed\n");
out_put:
		nfs4_put_file_data(file_data);
	}
out:
	nfsv4_printk(level2, "nfs4_setattr: returning status %d\n\n", status);
	return status;
}

int
nfs4_do_secinfo(struct super_block *sb,struct dentry *dentry)
{
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_putfh putfh;
	struct nfs4fs_secinfo secinfo;
	int status;
	struct rpc_auth *save_auth = NULL;

	nfsv4_printk(level2, "nfs4_do_secinfo: starting\n");

	/* Save old auth handle, replace with AUTH_GSS. */
	printk("nfs4_do_secinfo: 3\n");

	if(sb->u.nfs4_sb.client && sb->u.nfs4_sb.client->cl_auth){
		save_auth = sb->u.nfs4_sb.client->cl_auth;
	}
	printk("nfs4_do_secinfo: 4\n");
	printk("nfs4_do_secinfo: 4.5 save_auth %p\n",save_auth);
	rpcauth_create(RPC_AUTH_GSS, sb->u.nfs4_sb.client);

	cargs.sb = sb;
	printk("nfs4_do_secinfo: 0 cargs.clnt %p cargs.sb->cl_auth %p\n",
	       NFS4_CLNT(cargs.sb), NFS4_CLNT(sb)->cl_auth);
	cargs.tag = "nfs4_do_secinfo()";
	putfh.dentry = dentry;
	secinfo.name = &dentry->d_name;
	secinfo.sb = sb;
	if ((status = nfs4_setup_compound(&cargs)))
		goto out;
	printk("nfs4_do_secinfo: 1\n");
	if ((status = nfs4_setup_putfh(&cargs, &putfh)))
		goto out_args;
	printk("nfs4_do_secinfo: 2\n");
	if ((status = nfs4_setup_secinfo(&cargs, &secinfo)))
		goto out_args;
	
	printk("nfs4_do_secinfo: 5\n");
	if ((status = nfs4_call_compound(&cargs, &cres,(struct nfs4fs_ctx *)NULL)))
		goto out_auth;
	printk("nfs4_do_secinfo: 6\n");
	if ((status = nfs4_handle_putfh(&cres, &putfh)))
		goto out_res;
	printk("nfs4_do_secinfo: 7\n");
	if ((status = nfs4_handle_secinfo(&cres, &secinfo)))
		goto out_res;
	printk("nfs4_do_secinfo: 8\n");

out:
	nfsv4_printk(level2, "nfs4_do_secinfo: returning status %d\n", status);
	return status;
out_res:
	nfs4_release_compoundres(&cres);
out_auth:
	/* restore old auth */
	printk("nfs4_do_secinfo: 9\n");
	sb->u.nfs4_sb.client->cl_auth = save_auth;
out_args:
	nfs4_release_compoundargs(&cargs);
	goto out;
}

int
nfs4_do_secinfo_async(struct nfs4_async_compound *ctx)
{
	printk("!!! nfs4_do_secinfo_async: not implemented yet !!!\n");
	return -EIO;
}

int
nfs4_do_setclientid(struct super_block *sb)
{
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_setclientid setclientid;
	char netid[10];
	char uaddr[30];
	char name[30];
	int try = 0;
	int status;

	nfsv4_printk(level2, "nfs4_do_setclientid: starting\n");
	
	if (NFS4_HAVE_CLIENTID(sb)) {
		nfsv4_printk(clientid2, "nfs4_do_setclientid: already have clientid!\n");
		status = 0;
		goto out;
	}

	nfs4_get_callback(sb, &setclientid.callback_prog, netid, uaddr);
	cargs.sb = sb;
	cargs.tag = "nfs4_do_setclientid()";
	setclientid.callback_netid = netid;
	setclientid.callback_universal_address = uaddr;

try_again:
	/* XXX : how do we really want to choose the name? */
	sprintf(name, "citi-%d", try);
	setclientid.sb = sb;
	setclientid.namlen = strlen(name);
	setclientid.name = name;
	if ((status = nfs4_setup_compound(&cargs)))
		goto out;
	if ((status = nfs4_setup_setclientid(&cargs, &setclientid)))
		goto out_args;
	
	if ((status = nfs4_call_compound(&cargs, &cres,(struct nfs4fs_ctx *)NULL)))
		goto out_args;
	status = nfs4_handle_setclientid(&cres, &setclientid);
	nfs4_release_compoundres(&cres);
	nfs4_release_compoundargs(&cargs);
	if (status) {
		if (setclientid.in_use && (++try < NFS4_SETCLIENTID_MAXTRIES)) {
			nfsv4_printk(clientid2, "do_setclientid: clid in use, retrying\n");
			goto try_again;
		}
		else
			goto out_res;
	}
	
	nfsv4_printk(clientid, "nfs4_do_setclientid: obtained clientid successfully\n");
	NFS4_HAVE_CLIENTID(sb) = 1;
	NFS4_CLIENTID(sb) = setclientid.clientid;
	
out:
	nfsv4_printk(level2, "nfs4_do_setclientid: returning status %d\n", status);
	return status;
out_res:
	nfs4_release_compoundres(&cres);
out_args:
	nfs4_release_compoundargs(&cargs);
	goto out;
}

int
nfs4fs_start_recoverd(struct super_block *sb)
{
	int status = 0;

#if 1
	return 1;
#endif
	
	lock_kernel();
	if (NFS4_PIDOF_RECOVERD(sb))
		goto out;
	/* XXX: make sure that kernel_thread() can't block, to ensure atomicity under the
	   kernel lock!  think about flags (CLONE_FS, etc.) passed to kernel_thread(). */
	if ((status = kernel_thread(nfs4fs_recoverd, sb, 0)))
		goto out;
	NFS4_PIDOF_RECOVERD(sb) = current->pid;

out:
	unlock_kernel();
	return status;
}

int
nfs4fs_wait_on_recoverd(struct super_block *sb)
{
  	int status = 0;

	lock_kernel();
	if (!NFS4_PIDOF_RECOVERD(sb) || (NFS4_PIDOF_RECOVERD(sb) == current->pid))
		goto out;
	nfsv4_printk(lease, "nfs4: thread waiting on grace period\n");
	interruptible_sleep_on(&NFS4_GRACE_WAIT_QUEUE(sb));
	nfsv4_printk(lease, "nfs4: thread done waiting on grace period\n");
	status = -EINTR;
	if (signal_pending(current))
		goto out;
	status = 0;
	
out:
	unlock_kernel();
	return status;
}

static DECLARE_FSTYPE(nfs4_fstype, "nfs4", nfs4_read_super, FS_ODD_RENAME);

#ifdef MODULE

EXPORT_NO_SYMBOLS;
MODULE_AUTHOR("UM-CITI");

int
init_module(void)
{
#if 0
	struct inode dummy_inode;
	struct super_block dummy_sb;
	DECLARE_MUTEX(semaphore);
	
	printk("   ===> please do not remove this printk()!\n");
	printk("        sizeof(inode.u)=%d, sizeof(nfs4_inode_info)=%d\n",
	       sizeof(dummy_inode.u), sizeof(struct nfs4_inode_info));
	printk("        sizeof(sb.u)=%d, sizeof(nfs4_sb_info)=%d\n",
	       sizeof(dummy_sb.u), sizeof(struct nfs4_sb_info));
	printk("        sizeof(struct nfs4_async_compound)=%d\n",
	       sizeof(struct nfs4_async_compound));
	printk("        first down_trylock(): return value=%d\n", down_trylock(&semaphore));
	printk("        second down_trylock(): return value=%d\n", down_trylock(&semaphore));
	printk("   <===\n");
#endif

	/* XXX : fhcache, nfspagecache, readpagecache */
#ifdef CONFIG_PROC_FS
	rpc_proc_register(&nfs4_rpcstat);
#endif

	initialize();

	printk(KERN_INFO "Installing nfs4fs\n");
        return register_filesystem(&nfs4_fstype);
}

void
cleanup_module(void)
{
	/* XXX : fhcache, nfspagecache, readpagecache */
	
#ifdef CONFIG_PROC_FS
	rpc_proc_unregister("nfs4");
#endif
	printk(KERN_INFO "Uninstalling nfs4fs\n");
	nfs4fs_cleanup_state();
	unregister_filesystem(&nfs4_fstype);
}
#endif
