/*
 *  linux/fs/nfs4/lock.c
 *
 *  Copyright (c) 2001 The Regents of the University of Michigan.
 *  All rights reserved.
 *
 *  Kendrick Smith <kmsmith@umich.edu>
 *  Andy Adamson <andros@umich.edu>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/pagemap.h>

#include <linux/nfs4/nfs4_prot.h>
#include <linux/nfs4/nfs4_debug.h>
#include <linux/nfs4/nfs4.h>
#include <linux/nfs4/nfs4fs.h>

static int nfs4_lock_strict(struct file *filp, struct file_lock *lock, int is_blocking);

/* nfs4_testlock(): If a conflicting lock exists, write data for the conflicting lock into
   the file_lock structure.  If no such lock exists, set the 'fl_type' field to FL_UNLCK. */
static int
nfs4_testlock(struct file *filp, struct file_lock *file_lock)
{
	struct dentry *dentry = filp->f_dentry;
	struct nfs4fs_file_data *data = (struct nfs4fs_file_data *)filp->private_data;
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_putfh putfh;
	struct nfs4fs_getattr reval;
	struct nfs4fs_lockt lockt;
	int status;

	nfsv4_printk(lock, "nfs4_testlock: testing byterange %Ld-%Ld\n",
		     file_lock->fl_start, file_lock->fl_end);
	
#if NFS4_DEBUG
	if ((file_lock->fl_type != F_RDLCK) && (file_lock->fl_type != F_WRLCK)) {
		printk("!!! unrecognized lock type in nfs4_testlock !!!\n");
		file_lock->fl_type = F_UNLCK;
		return 0;
	}
#endif

	cargs.sb = dentry->d_sb;
	cargs.tag = "nfs4_testlock()";
	putfh.dentry = dentry;
	lockt.lockowner = data->fi_lockowner;
	lockt.lock = file_lock;

	if ((status = nfs4_setup_compound(&cargs)))
		goto out;
	if ((status = nfs4_setup_putfh(&cargs, &putfh)))
		goto out_args;
	if ((status = nfs4_setup_reval(&cargs, &reval)))
		goto out_args;
	if ((status = nfs4_setup_lockt(&cargs, &lockt)))
		goto out_args;
	
	if ((status = nfs4_call_compound(&cargs, &cres, NULL)))
		goto out_args;
	if ((status = nfs4_handle_putfh(&cres, &putfh)))
		goto out_res;
	if ((status = nfs4_handle_reval(&cres, &reval)))
		goto out_res;
	if ((status = nfs4_handle_lockt(&cres, &lockt)))
		goto out_res;

	/* If the LOCKT handler returns success, it will overwrite the file_lock structure
	   with data about the conflicting lock. */

out_res:
	nfs4_release_compoundres(&cres);
out_args:
	nfs4_release_compoundargs(&cargs);
out:
	nfsv4_printk(lock, "nfs4_testlock: returning status %d, conflict=%s\n",
		     status, (file_lock->fl_type == F_UNLCK) ? "no" : "yes");
	return status;
}

/* nfs4_lock_posix: used to send an F_SETLK or F_SETLKW fcntl to the server.  Assumes that
   the server can handle POSIX locking semantics. */
static int
nfs4_lock_posix(struct file *filp,  struct file_lock *file_lock, int is_blocking)
{
	struct dentry *dentry = filp->f_dentry;
	struct nfs4fs_file_data *file_data = (struct nfs4fs_file_data *)filp->private_data;
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_putfh putfh;
	struct nfs4fs_getattr reval;
	struct nfs4fs_lock lock;
	int lt;
	int status;
	volatile unsigned long timeout = NFS4_LOCK_MINTIMEOUT(dentry->d_sb);
	
	/* XXX:  I (KMS) believe there is an unlikely but nonetheless possible race
	   condition associated with NFS locking which should be looked into at some point.
	   (It looks like NFSv3 suffers from the race condition too.)  For now, pretend it
	   doesn't exist.. */
	
	nfsv4_printk(lock, "nfs4_lock_posix: %s byterange %Ld-%Ld, %s lock\n",
		     (file_lock->fl_type == F_UNLCK) ? "unlocking" : "locking",
		     file_lock->fl_start, file_lock->fl_end,
		     is_blocking ? "blocking" : "nonblocking");

	switch (file_lock->fl_type) {
	case F_RDLCK:
		lt = is_blocking ? READW_LT : READ_LT;
		break;
	case F_WRLCK:
		lt = is_blocking ? WRITEW_LT : WRITE_LT;
		break;
	case F_UNLCK:
		lt = WRITE_LT;      /* server better ignore this... */
		break;
#if NFS4_DEBUG
	default:
		printk("!!! nfs4_lock_posix: bad type !!!\n");
		status = -EIO;
		goto out;
#endif
	}

	cargs.sb = dentry->d_sb;
	cargs.tag = "nfs4_lock_posix()";
	putfh.dentry = dentry;
	lock.reclaim = 0;
	lock.file_data = file_data;
	lock.type = lt;
	lock.start = file_lock->fl_start;
	lock.end = file_lock->fl_end;

retry:
	if ((status = nfs4_setup_compound(&cargs)))
		goto out;
	if ((status = nfs4_setup_putfh(&cargs, &putfh)))
		goto out_args;
	if ((status = nfs4_setup_reval(&cargs, &reval)))
		goto out_args;
	status = (file_lock->fl_type == F_UNLCK) ? nfs4_setup_locku(&cargs, &lock) :
		nfs4_setup_lock(&cargs, &lock);
	if (status)
		goto out_args;

	if ((status = nfs4_call_compound(&cargs, &cres, NULL)))
		goto out_args;
	if ((status = nfs4_handle_putfh(&cres, &putfh)))
		goto out_res;
	if ((status = nfs4_handle_reval(&cres, &reval)))
		goto out_res;
	status = (file_lock->fl_type == F_UNLCK) ? nfs4_handle_locku(&cres, &lock) :
		nfs4_handle_lock(&cres, &lock);
	if (is_blocking && (status == -EAGAIN)) {
		nfsv4_printk(lock, "nfs4_lock_posix: lock denied, sleeping for "
			     "%ld %ld/%d seconds\n", timeout / HZ, timeout % HZ, HZ);
		
		/* XXX: release+realloc is somewhat inefficient here.. */
		nfs4_release_compoundres(&cres);
		nfs4_release_compoundargs(&cargs);
		
		/* sleep, with exponential backoff, and retry the operation. */
		current->state = TASK_INTERRUPTIBLE;
		status = schedule_timeout(timeout);
		status = -ERESTARTSYS;
		if (signal_pending(current))
			goto out;
		current->state = TASK_RUNNING;
		timeout <<= 1;
		if (timeout > NFS4_LOCK_MAXTIMEOUT(dentry->d_sb))
			timeout = NFS4_LOCK_MAXTIMEOUT(dentry->d_sb);
		
		nfsv4_printk(lock, "nfs4_lock_posix: woke up, retrying lock\n");
		goto retry;
	}

out_res:
	nfs4_release_compoundres(&cres);
out_args:
	nfs4_release_compoundargs(&cargs);
out:
	nfsv4_printk(lock, "nfs4_lock_posix: returning status %d\n", status);
	return status;
}

int
nfs4_lock(struct file *filp, int cmd, struct file_lock *file_lock)
{
	struct dentry *dentry = filp->f_dentry;
	struct inode *inode = dentry->d_inode;
	int status;

#if NFS4_DEBUG
	struct nfs4fs_file_data *data = (struct nfs4fs_file_data *)filp->private_data;
	struct nfs4fs_lockowner *lockowner = data->fi_lockowner;
	if (!inode) {
		printk("!!! nfs4_lock: dentry is negative !!!\n");
		return -EIO;
	}
	if (file_lock->fl_file != filp) {
		printk("!!! nfs4_lock: filp mismatch !!!\n");
		return -EIO;
	}
	if (lockowner->lo_sb != inode->i_sb) {
		printk("!!! nfs4_lock: superblock mismatch !!!\n");
		return -EIO;
	}
#endif

	status = -ENOLCK;
	if ((inode->i_mode & (S_ISGID | S_IXGRP)) == S_ISGID) {
		printk("!!! nfs4_lock: no mandatory locks yet !!!\n");
		goto out;
	}
	if (!file_lock->fl_owner || (file_lock->fl_flags & (FL_POSIX|FL_BROKEN)) != FL_POSIX) {
		printk("!!! nfs4_lock: no BSD flocks yet !!!\n");
		goto out;
	}
	
	nfsv4_printk(level2, "nfs4_lock(%.*s/%.*s): starting, cmd=%d type=%d "
		     "range=[%Ld,%Ld]\n", (int)dentry->d_parent->d_name.len,
		     dentry->d_parent->d_name.name, (int)dentry->d_name.len,
		     dentry->d_name.name, cmd, file_lock->fl_type, file_lock->fl_start,
		     file_lock->fl_end);

	/*
	 * Flush all pending writes before doing anything with locks...
	 */
	down(&inode->i_sem);
	status = nfs4_sync_file(inode, NULL, 0, 0, FLUSH_WAIT);
	up(&inode->i_sem);
	if (status < 0)
		goto out;
	
	switch (cmd) {
	case F_GETLK:
		status = nfs4_testlock(filp, file_lock);
		break;
	case F_SETLK:
		if (NFS4_STRICT_LOCKING(inode->i_sb))
			status = nfs4_lock_strict(filp, file_lock, 0);
		else
			status = nfs4_lock_posix(filp, file_lock, 0);
		break;
	case F_SETLKW:
		if (NFS4_STRICT_LOCKING(inode->i_sb))
			status = nfs4_lock_strict(filp, file_lock, 1);
		else
			status = nfs4_lock_posix(filp, file_lock, 1);
		break;
#if NFS4_DEBUG
	default:
		printk("!!! nfs4_lock: bad lock type !!!\n");
		return -EIO;
#endif
	}
	
	/*
	 * Make sure we clear the cache whenever we try to get the lock.
	 * This makes locking act as a cache coherency point.
	 * XXX: Do we really want to zap caches every time?  Something else
	 * disturbing: does nfs4_zap_caches() always have to be protected by
	 * the inode semaphore?
	 */
	if (!status && (cmd == F_SETLK || cmd == F_SETLKW) &&
	    (file_lock->fl_type != F_UNLCK)) {
		down(&inode->i_sem);
		nfs4_sync_file(inode, NULL, 0, 0, FLUSH_WAIT);
		nfs4_zap_caches(inode);
		up(&inode->i_sem);
	}
	
out:
	nfsv4_printk(level2, "nfs4_lock: returning status %d\n\n", status);
	return status;
}

/*  The remainder of this file is devoted to implementing "strict locking".  The code was
 *  written to be as failsafe as possible, and is non-optimal in a few places.  Of course,
 *  clients which use strict locking have more to worry about (inherent race conditions)
 *  than non-optimality anyway..
 */

#define WRITE_LOCKED(fl)   ((NFS4_CURRENT_TYPE(fl) == WRITE_LT) || \
                            (NFS4_CURRENT_TYPE(fl) == WRITEW_LT))
#define READ_LOCKED(fl)    ((NFS4_CURRENT_TYPE(fl) == READ_LT) || \
                            (NFS4_CURRENT_TYPE(fl) == READW_LT))

#if NFS4_DEBUG
static int
check_invariants(struct file_lock *fl)
{
	/* used for debugging purposes only, to check that certain invariants of the
	   private fields in fl are satisfied. */
	
	if (!fl) {
		printk("nfs4: NULL fl in check_invariants()!\n");
		goto bad;
	}

	if (fl->fl_type == F_UNLCK) {
		nfsv4_printk(lock, "check_invariants: warning: F_UNLCK!\n");
		goto base_type;
	}
	
	switch (NFS4_CURRENT_TYPE(fl)) {
	case READ_LT:
	case READW_LT:
		if (fl->fl_type != F_RDLCK) {
			printk("nfs4: check_invariants: type mismatch!\n");
			goto bad;
		}
		break;
	case WRITE_LT:
	case WRITEW_LT:
		if (fl->fl_type != F_WRLCK) {
			printk("nfs4: check_invariants: type mismatch!\n");
			goto bad;
		}
		break;
	default:
		printk("nfs4: check_invariants: bad type %d!\n", NFS4_CURRENT_TYPE(fl));
		goto bad;
	}

base_type:
	switch (NFS4_BASE_TYPE(fl)) {
	case UNLOCK_LT:
		if (NFS4_NR_BLOCKEDWRITE(fl)) {
			if (!WRITE_LOCKED(fl)) {
				printk("nfs4: check_invariants: expected writelock!\n");
				goto bad;
			}
		}
		else if (NFS4_NR_BLOCKEDREAD(fl)) {
			if (!READ_LOCKED(fl)) {
				printk("nfs4: check_invariants: expected readlock!\n");
				goto bad;
			}
		}
		else {
			printk("nfs4: check_invariants: stray lock!\n");
			goto bad;
		}
		break;
	case READ_LT:
	case READW_LT:
		if (NFS4_NR_BLOCKEDWRITE(fl)) {
			if (!WRITE_LOCKED(fl)) {
				printk("nfs4: check_invariants: expected writelock!\n");
				goto bad;
			}
		}
		else {
			if (!READ_LOCKED(fl)) {
				printk("nfs4: check_invariants: expected readlock!\n");
				goto bad;
			}
		}
		break;
	case WRITE_LT:
	case WRITEW_LT:
		if (!WRITE_LOCKED(fl)) {
			printk("nfs4: check_invariants: expected writelock!\n");
			goto bad;
		}
		break;
	default:
		printk("nfs4: check_invariants: bad base type!\n");
		goto bad;
	}

	return 0;

bad:
	printk("!!! [%Ld,%Ld] flock_type=%d NFS4_BASE_TYPE=%d\n"
	       "!!! NFS4_CURRENT_TYPE=%d READ=%d WRITE=%d\n",
	       fl->fl_start, fl->fl_end, fl->fl_type, NFS4_BASE_TYPE(fl),
	       NFS4_CURRENT_TYPE(fl), NFS4_NR_BLOCKEDREAD(fl), NFS4_NR_BLOCKEDWRITE(fl));
	return -EINVAL;
}
#endif

/*
 *  Primitives for strict locking are:
 *    strict_emitlockrequest() - sends a LOCK or LOCKU to the server
 *    strict_newlock() - sends a LOCK and inserts an entry into the lock list
 *    strict_unlock() - sends a LOCKU and removes an entry from the lock list
 *    strict_read_to_write() - tries to "upgrade" a read lock to a write lock, reacquires
 *                             the read lock on failure
 *    strict_write_to_read() - "downgrades" a write lock to a read lock
 *    strict_splitlock() - replaces a lock with two back-to-back locks
 */

static int
strict_emit_lockrequest(struct file *filp, nfs_lock_type4 lt, loff_t start, loff_t end,
			int is_unlock)
{
	struct dentry *dentry = filp->f_dentry;
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_putfh putfh;
	struct nfs4fs_getattr reval;
	struct nfs4fs_lock lock;
	int status;
	
#if NFS4_DEBUG
	status = -EIO;
	if (lt == UNLOCK_LT) {
		printk("nfs4: UNLOCK_LT in strict_emit_lockrequest()!\n");
		goto out;
	}
	if (start > end) {
		printk("nfs4: start > end in strict_emit_lockrequest()!\n");
		goto out;
	}
#endif

	/* called with big kernel lock, NFS4_STRICT_LOCKING_SEMAPHORE(inode) held. */

	cargs.sb = dentry->d_sb;
	cargs.tag = "nfs4_lock_strict()";
	putfh.dentry = dentry;
	lock.reclaim = 0;
	lock.file_data = (struct nfs4fs_file_data *)filp->private_data;
	lock.type = lt;
	lock.start = start;
	lock.end = end;

	if ((status = nfs4_setup_compound(&cargs)))
		goto out;
	if ((status = nfs4_setup_putfh(&cargs, &putfh)))
		goto out_args;
	if ((status = nfs4_setup_reval(&cargs, &reval)))
		goto out_args;
	status = is_unlock ? nfs4_setup_locku(&cargs, &lock) : nfs4_setup_lock(&cargs, &lock);
	if (status)
		goto out_args;

	if ((status = nfs4_call_compound(&cargs, &cres, NULL)))
		goto out_args;
	if ((status = nfs4_handle_putfh(&cres, &putfh)))
		goto out_res;
	if ((status = nfs4_handle_reval(&cres, &reval)))
		goto out_res;
	status = is_unlock ? nfs4_handle_locku(&cres, &lock) : nfs4_handle_lock(&cres, &lock);
	if (status)
		goto out_res;

out_res:
	nfs4_release_compoundres(&cres);
out_args:
	nfs4_release_compoundargs(&cargs);
out:
	return status;
}

static int
strict_newlock(struct file *filp, struct file_lock **flp, int lt, loff_t start, loff_t end)
{
	struct file_lock *newlock;
	int status;

	nfsv4_printk(lock, "strict_newlock: locking [%Ld,%Ld]\n", start, end);

	/* We allocate the lock in advance to avoid error handling in the case where the lock
	   request succeeds on the server but we fail to create a new lock on the client. */
	status = -ENOLCK;
	if (!(newlock = locks_alloc_lock(1))) {
		nfsv4_printk(lock2, "nfs4: couldn't alloc new lock!\n");
		goto out;
	}
	
	if ((status = strict_emit_lockrequest(filp, lt, start, end, 0))) {
		locks_free_lock(newlock);
		goto out;
	}
	locks_init_lock(newlock);
	newlock->fl_file = filp;
	newlock->fl_flags = FL_POSIX;
	newlock->fl_start = start;
	newlock->fl_end = end;
	newlock->fl_owner = current->files;
	newlock->fl_pid = current->pid;
	NFS4_BASE_TYPE(newlock) = UNLOCK_LT;
	NFS4_CURRENT_TYPE(newlock) = lt;
	if ((lt == READ_LT) || (lt == READW_LT)) {
		newlock->fl_type = F_RDLCK;
		NFS4_NR_BLOCKEDREAD(newlock) = 1;
		NFS4_NR_BLOCKEDWRITE(newlock) = 0;
	}
	else {
		newlock->fl_type = F_WRLCK;
		NFS4_NR_BLOCKEDREAD(newlock) = 0;
		NFS4_NR_BLOCKEDWRITE(newlock) = 1;
	}
	locks_insert_lock(flp, newlock);
	check_invariants(newlock);
	
out:
	return status;	
}

static int
strict_unlock(struct file_lock **flp)
{
	struct file_lock *fl = *flp;
	int status;

	nfsv4_printk(lock, "strict_unlock: unlocking [%Ld,%Ld]\n", fl->fl_start, fl->fl_end);
	
#if NFS4_DEBUG
	if ((status = check_invariants(fl)))
		goto out;
	status = -EIO;
	if (NFS4_NR_BLOCKEDREAD(fl) || NFS4_NR_BLOCKEDWRITE(fl)) {
		printk("nfs4: strict_unlock() called, lock is held open!\n");
		goto out;
	}
#endif

	if ((status = strict_emit_lockrequest(fl->fl_file, NFS4_CURRENT_TYPE(fl),
					      fl->fl_start, fl->fl_end, 1)))
		goto out;
	locks_delete_lock(flp, 0);

out:
	return status;
}

static int
strict_read_to_write(struct file *filp, struct file_lock **flp, int lt)
{
	struct file_lock *fl = *flp;
	int status;

	nfsv4_printk(lock, "strict_read_to_write: upgrading [%Ld,%Ld]\n", fl->fl_start,
		     fl->fl_end);
	
#if NFS4_DEBUG
	if ((status = check_invariants(fl)))
		goto out;
	status = -EIO;
	if (WRITE_LOCKED(fl)) {
		printk("nfs4: writelock in strict_read_to_write()!\n");
		goto out;
	}
	if ((lt != WRITE_LT) && (lt != WRITEW_LT)) {
		printk("nfs4: bad type in strict_read_to_write()!\n");
		goto out;
	}
#endif

	/* temporarily drop read lock on server */
	if ((status = strict_emit_lockrequest(fl->fl_file, NFS4_CURRENT_TYPE(fl), fl->fl_start,
					      fl->fl_end, 1)))
		goto out;
	
	/* try to acquire write lock; if successful, upgrade lock type and return success */
	if (!(status = strict_emit_lockrequest(filp, lt, fl->fl_start, fl->fl_end, 0))) {
		fl->fl_file = filp;
		fl->fl_type = F_WRLCK;
		NFS4_CURRENT_TYPE(fl) = lt;
		NFS4_NR_BLOCKEDWRITE(fl) = 1;
		goto out;
	}
	
	/* no write lock available; try to reacquire read lock.
	   on success, return the error returned by the earlier lock request. */
	if (strict_emit_lockrequest(fl->fl_file, READ_LT, fl->fl_start, fl->fl_end, 0)) {
		status = -EIO;   /* XXX: what error should we return? */
		printk(KERN_WARNING "nfs4: raced, lost lock on server!\n");
		locks_delete_lock(flp, 0);
	}
	NFS4_CURRENT_TYPE(fl) = READ_LT;

out:
	check_invariants(fl);
	return status;
}

static int
strict_write_to_read(struct file *filp, struct file_lock **flp)
{
	struct file_lock *fl = *flp;
	int status;

	nfsv4_printk(lock, "strict_write_to_read: downgrading [%Ld,%Ld]\n",
		     fl->fl_start, fl->fl_end);
	
#if NFS4_DEBUG
	if ((status = check_invariants(fl)))
		goto out;
	status = -EIO;
	if (READ_LOCKED(fl)) {
		printk("nfs4: readlock in strict_write_to_read()!\n");
		goto out;
	}
#endif

	/* temporarily drop write lock on server */
	if ((status = strict_emit_lockrequest(fl->fl_file, NFS4_CURRENT_TYPE(fl), fl->fl_start,
					      fl->fl_end, 1)))
		goto out;

	/* try to acquire readlock */
	status = -EIO;
	if (strict_emit_lockrequest(filp, READ_LT, fl->fl_start, fl->fl_end, 0)) {
		printk(KERN_WARNING "nfs4: raced, lost lock on server!\n");
		locks_delete_lock(flp, 0);
		goto out;
	}
	fl->fl_type = F_RDLCK;
	NFS4_CURRENT_TYPE(fl) = READ_LT;
	if (NFS4_BASE_TYPE(fl) != UNLOCK_LT)
		NFS4_BASE_TYPE(fl) = READ_LT;
	status = 0;

out:
	check_invariants(fl);
	return status;
}

static int
strict_splitlock(struct file_lock **flp, loff_t new_end)
{
	struct file_lock *fl = *flp;
	struct file_lock *newlock;
	nfs_lock_type4 lt;
	loff_t start, end;
	int status;

	nfsv4_printk(lock, "strict_splitlock: splitting [%Ld,%Ld]\n",
		     fl->fl_start, fl->fl_end);
	
#if NFS4_DEBUG
	if ((status = check_invariants(fl)))
		goto out;
	status = -EIO;
	if (new_end < fl->fl_start) {
		printk("!!! strict_splitlock: new_end < lock->fl_start !!!\n");
		goto out;
	}
	if (new_end >= fl->fl_end) {
		printk("!!! strict_splitlock: new_end >= lock->fl_end !!!\n");
		goto out;
	}
#endif
	lt = NFS4_CURRENT_TYPE(fl);
	start = fl->fl_start;
	end = fl->fl_end;
	
	status = -ENOMEM;
	if (!(newlock = locks_alloc_lock(0))) {
		nfsv4_printk(lock2, "strict_splitlock: couldn't allocate lock!\n");
		goto out;
	}
	locks_init_lock(newlock);   /* XXX: necessary? */
	locks_copy_lock(newlock, fl);
	fl->fl_end = new_end;
	newlock->fl_start = new_end + 1;
	locks_insert_lock(&fl->fl_next, newlock);
	
	/* temporarily drop the lock on the server */
	if ((status = strict_emit_lockrequest(fl->fl_file, NFS4_CURRENT_TYPE(fl),
					      start, end, 1)))
		goto out;

	/* reacquire first half of lock */
	if (!strict_emit_lockrequest(fl->fl_file, NFS4_CURRENT_TYPE(fl), start, new_end, 0))
		flp = &fl->fl_next;
	else {
		status = -EIO;
		printk(KERN_WARNING "nfs4: raced, lost lock on server!\n");
		locks_delete_lock(flp, 0);
	}

	/* reacquire second half */
	if (strict_emit_lockrequest(fl->fl_file, NFS4_CURRENT_TYPE(fl),
				    new_end + 1, end, 0)) {
		status = -EIO;
		printk(KERN_WARNING "nfs4: raced, lost lock on server!\n");
		locks_delete_lock(flp, 0);
	}
	
out:
	return status;	
}

static int
strict_upgrade(struct file *filp, struct file_lock **flp, int lt)
{
	struct file_lock *fl = *flp;
	int status = 0;

#if NFS4_DEBUG
	if ((status = check_invariants(fl)))
		return status;
#endif
	
	switch (lt) {
	case UNLOCK_LT:    /* no-op */
		break;
	case READ_LT:
	case READW_LT:
		NFS4_NR_BLOCKEDREAD(fl)++;
		break;
	case WRITE_LT:
	case WRITEW_LT:
		if (READ_LOCKED(fl))
			/* strict_read_to_write() will increase the write count */
			status = strict_read_to_write(filp, flp, lt);
		else
			NFS4_NR_BLOCKEDWRITE(fl)++;
		break;
	}

	return status;
}

static int
strict_downgrade(struct file *filp, struct file_lock **flp, int lt)
{
	struct file_lock *fl = *flp;
	int status = 0;

#if NFS4_DEBUG
	if ((status = check_invariants(fl)))
		return status;
#endif
	
	switch (lt) {
	case UNLOCK_LT:
		if (NFS4_NR_BLOCKEDWRITE(fl))
			break;
		/* XXX: what if the unlocking process isn't using a file descriptor which has
		   been opened for read?! */
		if (NFS4_NR_BLOCKEDREAD(fl) && WRITE_LOCKED(fl))
			status = strict_write_to_read(filp, flp);
		else
			status = strict_unlock(flp);
		break;
	case READ_LT:
	case READW_LT:
		if (NFS4_NR_BLOCKEDWRITE(fl))
			break;
		if (!WRITE_LOCKED(fl))
			break;
		status = strict_write_to_read(filp, flp);
	case WRITE_LT:
	case WRITEW_LT:    /* no-op */
		break;
	}

	return status;
}

static int
strict_finalize(struct file_lock *fl, int lt)
{
	int status = 0;
	
#if NFS4_DEBUG
	if ((status = check_invariants(fl)))
		return status;
#endif
	
	switch (lt) {
	case UNLOCK_LT:    /* no-op */
		break;
	case READ_LT:
	case READW_LT:
		if (READ_LOCKED(fl))
			NFS4_BASE_TYPE(fl) = NFS4_CURRENT_TYPE(fl);
		else
			NFS4_BASE_TYPE(fl) = READ_LT;
		NFS4_NR_BLOCKEDREAD(fl)--;
		break;
	case WRITE_LT:
	case WRITEW_LT:
		NFS4_BASE_TYPE(fl) = NFS4_CURRENT_TYPE(fl);
		NFS4_NR_BLOCKEDWRITE(fl)--;
		break;
	}

	return status;
}

static int
strict_rollback(struct file *filp, struct file_lock **flp, int lt)
{
	struct file_lock *fl = *flp;
	int status = 0;

#if NFS4_DEBUG
	if ((status = check_invariants(fl)))
		return status;
#endif
	
	switch (lt) {
	case UNLOCK_LT:
		nfsv4_printk(level1, "strict_rollback: UNLOCK_LT in strict_rollback?!\n");
		break;
	case READ_LT:
	case READW_LT:
		if (--NFS4_NR_BLOCKEDREAD(fl))
			break;
		if (NFS4_NR_BLOCKEDWRITE(fl) || (NFS4_BASE_TYPE(fl) != UNLOCK_LT))
			break;
		status = strict_unlock(flp);
		break;
	case WRITE_LT:
	case WRITEW_LT:
		if (--NFS4_NR_BLOCKEDWRITE(fl))
			break;
		if ((NFS4_BASE_TYPE(fl) == WRITE_LT) || (NFS4_BASE_TYPE(fl) == WRITEW_LT))
			break;
		if (!NFS4_NR_BLOCKEDREAD(fl) && (NFS4_BASE_TYPE(fl) == UNLOCK_LT))
			status = strict_unlock(flp);
		else
			status = strict_write_to_read(filp, flp);
		break;
	}

	return status;
}

static int
nfs4_lock_strict(struct file *filp, struct file_lock *file_lock, int is_blocking)
{
	struct dentry *dentry = filp->f_dentry;
	struct inode *inode = dentry->d_inode;
	struct file_lock *fl, **flp;
	volatile unsigned long timeout = NFS4_LOCK_MINTIMEOUT(dentry->d_sb);
	loff_t start, end;
	int encountered_owner;
	nfs_lock_type4 lt;
	int status = 0;

	nfsv4_printk(lock, "nfs4_lock_strict(): starting\n");
	
	start = file_lock->fl_start;
	end = file_lock->fl_end;
	lock_kernel();

	switch (file_lock->fl_type) {
	case F_RDLCK:
		lt = is_blocking ? READW_LT : READ_LT;
		break;
	case F_WRLCK:
		lt = is_blocking ? WRITEW_LT : WRITE_LT;
		break;
	case F_UNLCK:
		lt = UNLOCK_LT;
		break;
#if NFS4_DEBUG
	default:
		printk("nfs4_lock_strict: bad lock type!\n");
		status = -EINVAL;
		goto out;
#endif
	}

upgrade_pass:
	/*
	 *  Pass 1:  Upgrade pass.  Identify regions where we want to upgrade the locking
	 *  state, either because the region is currently unlocked, or it is readlocked and
	 *  we want a writelock.  If we fail to get an upgrade, return -EAGAIN (non-blocking
	 *  case), or sleep and retry (blocking case).  If we have to sleep, release the
	 *  locking semaphore so as not to block other processes.  If we get an upgrade,
	 *  remember...
	 */

	nfsv4_printk(lock, "nfs4_strict_lock(): in upgrade pass\n");
	down(&NFS4_LOCKING_SEMAPHORE(inode));

	/* Loop invariant: at top of loop, start is the first byte we
	   are trying to lock (this might not be equal to file_lock->fl_start if we slept on a
	   previous pass through the loop and had to restart) */
	
	encountered_owner = 0;
	for (flp = &inode->i_flock; *flp; flp = &(*flp)->fl_next) {
		fl = *flp;
		if ((fl->fl_owner != file_lock->fl_owner) ||
		    (fl->fl_pid != file_lock->fl_pid)) {
			nfsv4_printk(lock, "nfs4_strict_lock: 'fl' has different owner\n");
			if (!encountered_owner)
				continue;
			goto lock_to_end;
		}
		encountered_owner = 1;
		if (fl->fl_end < start) {
			nfsv4_printk(lock, "nfs4_strict_lock: 'fl' to left of range\n");
			continue;
		}
		if (fl->fl_start > end) {
			nfsv4_printk(lock, "nfs4_strict_lock: 'fl' to right of range\n");
			goto lock_to_end;
		}

		/* The locks overlap.
		 * First, if 'start' is in the middle of the range, split the lock.
		 * If 'start' is to the right of the range, grab all space up to the lock.
		 */
		if (start > fl->fl_start) {
			nfsv4_printk(lock, "nfs4_strict_lock: left overlap; splitting\n");
			if ((status = strict_splitlock(flp, start - 1)))
				goto lock_error;
			/* advance past the first lock created by strict_splitlock().. */
			flp = &(*flp)->fl_next;
			fl = *flp;
#if NFS4_DEBUG
			if (fl->fl_start != start) {
				status = -EIO;
				printk("nfs4: bad right-hand split in strict_splitlock()!\n");
				goto out;
			}
#endif
		}
		else if (lt == UNLOCK_LT)
			start = fl->fl_start;
		else if (start < fl->fl_start) {
			nfsv4_printk(lock, "nfs4_strict_lock: grabbing left of 'fl'\n");
			if ((status = strict_newlock(filp, flp, lt, start, fl->fl_start - 1)))
				goto lock_error;
			/* advance past the new lock created by strict_newlock().. */
			flp = &(*flp)->fl_next;
#if NFS4_DEBUG
			if (fl != *flp) {
				status = -EIO;
				printk("nfs4: bad lock insertion in strict_newlock()!\n");
				goto lock_error;
			}
#endif
			start = fl->fl_start;
			timeout = NFS4_LOCK_MINTIMEOUT(dentry->d_sb);
		}
		/* We now have locked up to the start of the existing lock.
		 * Second, split the existing lock in two if it won't be completely
		 * subsumed.
		 */
		if (end < fl->fl_end) {
			nfsv4_printk(lock, "nfs4_strict_lock: right overlap; splitting\n");
			if ((status = strict_splitlock(flp, end)))
				goto lock_error;
#if NFS4_DEBUG
			if ((fl != *flp) || (fl->fl_end != end)) {
				status = -EIO;
				printk("nfs4: bad left-hand split in strict_splitlock()!\n");
				goto out;
			}
#endif
		}
		/* Third, upgrade the existing lock (will no-op if no upgrade needed)
		 */
		nfsv4_printk(lock, "nfs4_strict_lock: upgrading 'fl'\n");
		if ((status = strict_upgrade(filp, flp, lt)))
			goto lock_error;
		start = fl->fl_end + 1;
		timeout = NFS4_LOCK_MINTIMEOUT(dentry->d_sb);
	}

lock_to_end:
	nfsv4_printk(lock, "nfs4_strict_lock(): locking to end\n");
	if ((lt != UNLOCK_LT) && (start <= end)) {
		if ((status = strict_newlock(filp, flp, lt, start, end)))
			goto lock_error;
	}

	/*  Pass 2:  Downgrade pass.  Identify regions where we want to downgrade the locking
	 *  state (this can only happen if a region is writelocked, and we want a readlock.
	 *  A downgrade can only fail as the result of a race condition with another client
	 *  (this is an inherent race condition between Unix locking semantics and the NFSv4
	 *  locking protocol); if this happens, we just printk() a warning and abort
	 *  everything with an error.  Maybe later, we will try to handle a failed downgrade
	 *  in a more sophisticated way.  During the downgrade pass, we will never need to
	 *  sleep or release the semaphore.  The downgrade pass is seperated from the upgrade
	 *  pass because upgrades may sleep, and we don't want to sleep while holding a
	 *  downgrade: if the sleep is interrupted by a signal, we will have to roll back the
	 *  downgrade and exit, and sleeping will increase the chance that the rollback will
	 *  fail because of another client.
	 */

	/* skip downgrade pass if writelocking region */
	if (file_lock->fl_type == F_WRLCK)
		goto finalize;

	nfsv4_printk(lock, "nfs4_strict_lock(): in downgrade pass\n");
	encountered_owner = 0;
	flp = &inode->i_flock;
	while (*flp) {
		fl = *flp;
		if ((fl->fl_owner != file_lock->fl_owner) ||
		    (fl->fl_pid != file_lock->fl_pid)) {
			if (!encountered_owner) {
				flp = &fl->fl_next;
				continue;
			}
			break;
		}
		encountered_owner = 1;
		if (fl->fl_end < file_lock->fl_start) {
			flp = &fl->fl_next;
			continue;
		}
		if (fl->fl_start > file_lock->fl_end)
			break;
#if NFS4_DEBUG
		if ((fl->fl_start < file_lock->fl_start) ||
		    (fl->fl_end > file_lock->fl_end)) {
			status = -EIO;
			printk("nfs4: non-nested locks during downgrade pass!\n");
			goto out;
		}
#endif
		if ((status = strict_downgrade(filp, flp, lt)))
			goto lock_error;
		/* strict_downgrade() might have removed fl from the list */
		if (fl == *flp)
			flp = &fl->fl_next;
	}
	
finalize:	
	/*  Pass 3: Finalization pass.  We "finalize" all of the upgrades acquired
	 *  in Pass 1 (this never requires sending any traffic to the server, just
	 *  updating our own internal rollback state).
	 */
	if (file_lock->fl_type == F_UNLCK)
		goto done;
	nfsv4_printk(lock, "nfs4_strict_lock(): in finalization pass\n");
	encountered_owner = 0;
	for (flp = &inode->i_flock; *flp; flp = &(*flp)->fl_next) {
		fl = *flp;
		if ((fl->fl_owner != file_lock->fl_owner) ||
		    (fl->fl_pid != file_lock->fl_pid)) {
			if (!encountered_owner)
				continue;
			break;
		}
		encountered_owner = 1;
		if (fl->fl_end < file_lock->fl_start)
			continue;
		if (fl->fl_start > file_lock->fl_end)
			break;
#if NFS4_DEBUG
		if ((fl->fl_start < file_lock->fl_start) || (fl->fl_end > file_lock->fl_end)) {
			status = -EIO;
			printk("nfs4: non-nested locks during finalization pass!\n");
			goto out;
		}
#endif
		if ((status = strict_finalize(fl, lt)))
			goto out;
	}

done:
	/* Success! */
	status = LOCK_DONE;  /* inform VFS that locking state doesn't need to be updated */
	up(&NFS4_LOCKING_SEMAPHORE(inode));
out:
	unlock_kernel();
	nfsv4_printk(lock, "nfs4_strict_lock(): returning status %d\n", status);
	return status;
	
lock_error:
	up(&NFS4_LOCKING_SEMAPHORE(inode));
	if (is_blocking && (status == -EAGAIN)) {
		/* Sleep, with exponential backoff, and retry the operation.  Since we sleep
		   without the semaphore held, the locking state may have undergone arbitrary
		   changes when we wake up.  Therefore, we retry the operation from the very
		   top. */
		nfsv4_printk(lock, "nfs4_lock_strict: lock denied, sleeping for "
			     "%ld %ld/%d seconds\n", timeout / HZ, timeout % HZ, HZ);
		current->state = TASK_INTERRUPTIBLE;
		status = schedule_timeout(timeout);
		status = -ERESTARTSYS;
		if (signal_pending(current))
			goto out;
		current->state = TASK_RUNNING;
		timeout <<= 1;
		if (timeout > NFS4_LOCK_MAXTIMEOUT(dentry->d_sb))
			timeout = NFS4_LOCK_MAXTIMEOUT(dentry->d_sb);
		nfsv4_printk(lock, "nfs4_lock_posix: woke up, retrying lock\n");
		goto upgrade_pass;
	}
	
	/* Operation failed, either because an upgrade failed during a non-blocking lock, a
	   signal was caught during a blocking lock, or a miscellaneous fatal error occured.
	   We roll back all locks acquired so far and return the error. */
	
	if (start == file_lock->fl_start)
		goto out;
	start = file_lock->fl_start;
	end = start - 1;
	encountered_owner = 0;
	flp = &inode->i_flock;
	while (*flp) {
		fl = *flp;
		if ((fl->fl_owner != file_lock->fl_owner) ||
		    (fl->fl_pid != file_lock->fl_pid)) {
			if (!encountered_owner) {
				flp = &fl->fl_next;
				continue;
			}
			break;
		}
		encountered_owner = 1;
		if (fl->fl_end < start) {
			flp = &fl->fl_next;
			continue;
		}
		if (fl->fl_start > end)
			break;
#if NFS4_DEBUG
		if ((fl->fl_start < start) || (fl->fl_end > end)) {
			status = -EIO;
			printk("nfs4: non-nested locks during rollback!\n");
			goto out;
		}
#endif
		strict_rollback(filp, flp, lt);
		/* strict_rollback() might have removed fl from the list */
		if (fl == *flp)
			flp = &fl->fl_next;
	}
	goto out;
}
