/*
 *  linux/fs/nfs4/open.c
 *
 *  Copyright (c) 2001 The Regents of the University of Michigan.
 *  All rights reserved.
 *
 *  Kendrick Smith <kmsmith@umich.edu>
 *  Andy Adamson <andros@umich.edu>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* XXX: remove #include spinlock.h?  (This was done for nfsv2/v3 in 2.4.4) */

#include <linux/module.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/spinlock.h>
#include <linux/fs.h>
#include <linux/list.h>
#include <linux/mm.h>
#include <linux/pagemap.h>

#include <linux/nfs4/nfs4_prot.h>
#include <linux/nfs4/nfs4_debug.h>
#include <linux/nfs4/nfs4.h>
#include <linux/nfs4/nfs4fs.h>

#include <linux/quotaops.h>
#define ACC_MODE(x) ("\000\004\002\006"[(x)&O_ACCMODE])

#if 0
static void recover_open(struct super_block *sb, struct nfs4fs_file_data *fdata,
			struct list_head *inode_list);
static void recover_locks(struct super_block *sb, struct inode *inode);
#endif

/*  Lockowner hashtable
 *  XXX : a serious problem: how do we know when an nfs4fs_lockowner can be freed?!
 *  XXX : need lockowner id wraparound protection
 */

#define LOCKOWNER_HASH_BITS	8
#define LOCKOWNER_HASH_SIZE	(1 << LOCKOWNER_HASH_BITS)
#define LOCKOWNER_HASH_MASK	(LOCKOWNER_HASH_SIZE - 1)

static u32 global_lockowner_id = 0;
static struct list_head lockowner_tbl[LOCKOWNER_HASH_SIZE];
static struct list_head lockowner_id_tbl[LOCKOWNER_HASH_SIZE];
static spinlock_t lockowner_spinlock;
static struct nfs4fs_lockowner *preallocated_lockowner;

#define FILEDATA_HASH_BITS	8
#define FILEDATA_HASH_SIZE	(1 << FILEDATA_HASH_BITS)
#define FILEDATA_HASH_MASK	(FILEDATA_HASH_SIZE - 1)

static struct list_head filedata_tbl[FILEDATA_HASH_SIZE];
static spinlock_t filedata_spinlock;

struct nfs4fs_lockowner *
__nfs4_get_lockowner(struct super_block *sb, struct files_struct *fs, pid_t pid)
{
	unsigned int id;
	unsigned int hashval;
	struct list_head *head, *p;
	struct nfs4fs_lockowner *lockowner;

	hashval = opaque_hashval(&fs, sizeof(fs)) + opaque_hashval(&sb, sizeof(sb)) +
		opaque_hashval(&pid, sizeof(pid));
	head = &lockowner_tbl[hashval & LOCKOWNER_HASH_MASK];

retry:
	/* Search the hashtable for the lockowner. */
	spin_lock(&lockowner_spinlock);
	list_for_each(p, head) {
		lockowner = list_entry(p, struct nfs4fs_lockowner, lo_hash);
		if ((lockowner->lo_sb == sb) && (lockowner->lo_fs == fs) &&
		    (lockowner->lo_pid == pid)) {
			spin_unlock(&lockowner_spinlock);
			goto got_it;
		}
	}
	/* Lockowner not found.  Do we have one preallocated? */
	if (preallocated_lockowner) {
		lockowner = preallocated_lockowner;
		id = global_lockowner_id++;
		list_add(&lockowner->lo_hash, head);
		list_add(&lockowner->lo_id_hash, &lockowner_id_tbl[id & LOCKOWNER_HASH_MASK]);
		lockowner->lo_sb = sb;
		lockowner->lo_fs = fs;
		lockowner->lo_pid = pid;
		lockowner->lo_id = id;
		lockowner->lo_seqid = 0;    /* arbitrary */
		init_MUTEX(&lockowner->lo_sema);
		atomic_set(&lockowner->lo_count, 1);   /* refcount==1 means hashed */
		spin_unlock(&lockowner_spinlock);
		nfsv4_printk(lockowner, "__nfs4_get_lockowner creating lockowner %d\n", id);
		preallocated_lockowner = NFS4_ALLOC(sizeof(*lockowner));
		goto got_it;
	}
	spin_unlock(&lockowner_spinlock);

	/* Try to allocate a new lockowner.  If we succeed, retry the operation (since we
	   may have blocked).  If we fail, it is an error. */
	lockowner = NULL;
	if ((preallocated_lockowner = NFS4_ALLOC(sizeof(struct nfs4fs_lockowner))))
		goto retry;
	goto out;
	
got_it:
	atomic_inc(&lockowner->lo_count);
	nfsv4_printk(lockowner, "__nfs4_get_lockowner returning lockowner %d (count %d)\n",
		     lockowner->lo_id, atomic_read(&lockowner->lo_count));
out:
	return lockowner;
}
	
struct nfs4fs_lockowner *
nfs4_get_lockowner_by_id(unsigned int id)
{
	struct list_head *p;
	struct nfs4fs_lockowner *lockowner = NULL;

	spin_lock(&lockowner_spinlock);
	list_for_each(p, &lockowner_id_tbl[id & LOCKOWNER_HASH_MASK]) {
		lockowner = list_entry(p, struct nfs4fs_lockowner, lo_id_hash);
		if (lockowner->lo_id == id) {
			spin_unlock(&lockowner_spinlock);
			atomic_inc(&lockowner->lo_count);
			nfsv4_printk(lockowner, "get_lockowner_by_id: returning lockowner %d "
				     "(count %d)\n", lockowner->lo_id,
				     atomic_read(&lockowner->lo_count));
			goto out;
		}
	}
	spin_unlock(&lockowner_spinlock);
	lockowner = NULL;

out:
	return lockowner;
}

void
nfs4_put_lockowner(struct nfs4fs_lockowner *lockowner)
{
	/* XXX: for now, we never flush unused lockowners from the hashtables.. */
#if NFS4_DEBUG
	if (!atomic_read(&lockowner->lo_count)) {
		printk("!!! nfs4_put_lockowner: zero refcount (lockowner %d)!!!\n",
		       lockowner->lo_id);
		return;
	}
#endif
	atomic_dec(&lockowner->lo_count);
	nfsv4_printk(lockowner, "nfs4_put_lockowner: lockowner %d, count=%d\n",
		     lockowner->lo_id, atomic_read(&lockowner->lo_count));
}

static inline struct nfs4fs_file_data *
file_data_seek(struct list_head *head, struct inode *inode, struct files_struct *fs, pid_t pid)
{
	struct list_head *p;
	struct nfs4fs_file_data *fp;

	/* call with spinlock held */
	list_for_each(p, head) {
		fp = list_entry(p, struct nfs4fs_file_data, fi_hash);
		if ((fp->fi_inode == inode) && (fp->fi_fs == fs) && (fp->fi_pid == pid)) {
			atomic_inc(&fp->fi_count);
			return fp;
		}
	}
	return NULL;
}

#define FILEDATA_HASHVAL(inode, fs, pid)   (opaque_hashval(&(inode), sizeof(inode)) + \
                                            opaque_hashval(&(fs), sizeof(fs)) + \
                                            opaque_hashval(&(pid), sizeof(pid)))
#define FILEDATA_HEAD(inode, fs, pid)    &filedata_tbl[FILEDATA_HASHVAL(inode,fs,pid) & \
                                                       FILEDATA_HASH_MASK]

struct nfs4fs_file_data *
nfs4_get_file_data(struct dentry *dentry)
{
	struct inode *inode = dentry->d_inode;
	struct files_struct *fs = current->files;
	pid_t pid = current->pid;
	struct list_head *head;
	struct nfs4fs_file_data *fp, *new;
	struct nfs4fs_lockowner *lockowner;

#if NFS4_DEBUG
	if (!inode) {
		printk("!!! open_file_data: inode is NULL !!!\n");
		return NULL;
	}
#endif

	/* Search for an existing file_data structure */
	head = FILEDATA_HEAD(inode, fs, pid);
	spin_lock(&filedata_spinlock);
	fp = file_data_seek(head, inode, fs, pid);
	spin_unlock(&filedata_spinlock);
	if (fp)
		goto out;
	
	/* Filedata not found; allocate a new one */
	if (!(lockowner = nfs4_get_lockowner(inode->i_sb)))
		goto out;
	if (!(new = NFS4_ALLOC(sizeof(struct nfs4fs_file_data)))) {
		nfs4_put_lockowner(lockowner);
		goto out;
	}
	
	/* Retry the search, since we may have blocked. */
	spin_lock(&filedata_spinlock);
	fp = file_data_seek(head, inode, fs, pid);
	if (fp) {
		spin_unlock(&filedata_spinlock);
		NFS4_FREE(new);
		goto out;
	}
	
	/* Insert the new filedata structure */
	fp = new;
	list_add(&fp->fi_hash, head);
	fp->fi_lockowner = lockowner;
	fp->fi_dentry = dget(dentry);
	fp->fi_fs = fs;
	fp->fi_pid = pid;
	nfs4_init_rwlock(&fp->fi_sema);
	fp->fi_nopen = 0;
	fp->fi_nread = 0;
	fp->fi_nwrite = 0;
	atomic_set(&fp->fi_count, 1);

	INIT_LIST_HEAD(&fp->fi_read_map);
	INIT_LIST_HEAD(&fp->fi_write_map);

	list_add(&fp->fi_list, &NFS4_FILE_LIST(dentry->d_sb));
	spin_unlock(&filedata_spinlock);
	
out:
	return fp;
}

void
nfs4_put_file_data(struct nfs4fs_file_data *file_data)
{
	/* We drop filedata structs as soon as they are no longer in use, to release the
	   inode reference. */

	if (atomic_dec_and_lock(&file_data->fi_count, &filedata_spinlock)) {
		list_del(&file_data->fi_hash);
		dput(file_data->fi_dentry);                 /* release reference to dentry */
		nfs4_put_lockowner(file_data->fi_lockowner);   /* release ref to lockowner */
		
		list_del(&file_data->fi_read_map);
		list_del(&file_data->fi_write_map);
		
		list_del(&file_data->fi_list);
		spin_unlock(&filedata_spinlock);
#if NFS4_DEBUG
		if (file_data->fi_nopen || file_data->fi_nread || file_data->fi_nwrite)
			printk("!!! nfs4_put_filedata: file is open !!!\n");
#endif
		NFS4_FREE(file_data);
	}
}

void
nfs4fs_init_state(void)
{
	int i;
	static int initialized = 0;

	if (initialized)
		return;
	
	spin_lock_init(&lockowner_spinlock);
	preallocated_lockowner = NULL;
	for (i = 0; i < LOCKOWNER_HASH_SIZE; i++) {
		INIT_LIST_HEAD(&lockowner_tbl[i]);
		INIT_LIST_HEAD(&lockowner_id_tbl[i]);
	}
	
	spin_lock_init(&filedata_spinlock);
	for (i = 0; i < FILEDATA_HASH_SIZE; i++)
		INIT_LIST_HEAD(&filedata_tbl[i]);
	
	initialized = 1;
}

void
nfs4fs_cleanup_state(void)
{
	int i;
	struct nfs4fs_lockowner *lockowner;
	
	for (i = 0; i < FILEDATA_HASH_SIZE; i++) {
		if (list_empty(&filedata_tbl[i]))
			continue;
		/* for now, we just warn about this error without trying to fix it */
		printk(KERN_WARNING "nfs4: stray file_data after unmount!\n");
		INIT_LIST_HEAD(&filedata_tbl[i]);
	}
	for (i = 0; i < LOCKOWNER_HASH_SIZE; i++)
		while (!list_empty(&lockowner_tbl[i])) {
			lockowner = list_entry(lockowner_tbl[i].next, struct nfs4fs_lockowner,
					       lo_hash);
			if (atomic_read(&lockowner->lo_count) != 1)
				printk(KERN_WARNING "nfs4: stray lockowner after unmount!\n");
			list_del(&lockowner->lo_hash);
			list_del(&lockowner->lo_id_hash);
			NFS4_FREE(lockowner);
		}
	if (preallocated_lockowner) {
		NFS4_FREE(preallocated_lockowner);
		preallocated_lockowner = NULL;
	}
}

int
nfs4fs_do_open(struct nfs4fs_open *open)
{
	struct dentry *dentry = open->dentry;
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_putfh putfh;
	struct nfs4fs_getattr reval;	
	int status;

	nfsv4_printk(open, "nfs4fs_do_open: starting\n");
	
#if 0
	/* (This section disabled for testing.) */
	/* Optimize out an OPEN if the share access wouldn't actually change.. */
	struct nfs4fs_file_data *file_data = open->file_data;
	
	status = 0;
	down_write(&file_data->fi_sema);
	if ((file_data->fi_nopen > 1) &&
	    (!(open->flags & FMODE_READ) || (ofp->fi_nread)) &&
	    (!(open->flags & FMODE_WRITE) || (ofp->fi_nwrite))) {
		file_data->fi_nopen++;
		if (open->flags & FMODE_READ)
			file_data->fi_nread++;
		if (open->flags & FMODE_WRITE)
			file_data->fi_nwrite++;
		up_write(&file_data->fi_sema);
		goto out;
	}
	up_write(&file_data->fi_sema);
#endif

	if (dentry->d_inode) {
		status = nfs4_sync_file(dentry->d_inode, NULL,
					0, 0, FLUSH_WAIT | FLUSH_STABLE);
		if (status)
			goto out;
	}
	
	cargs.sb = dentry->d_sb;
	cargs.tag = "nfs4fs_do_open()";
	putfh.dentry = dentry->d_parent;
	
	if ((status = nfs4_setup_compound(&cargs)))
		goto out;
	if ((status = nfs4_setup_putfh(&cargs, &putfh)))
		goto out_args;
	if ((status = nfs4_setup_reval(&cargs, &reval)))
		goto out_args;
	if ((status = nfs4_setup_open(&cargs, open)))
		goto out_args;

	if (open->flags & O_CREAT)
		nfs4_zap_caches(dentry->d_parent->d_inode);
	if (dentry->d_inode && (open->flags & O_TRUNC))
		nfs4_zap_caches(dentry->d_inode);
	if ((status = nfs4_call_compound(&cargs, &cres, NULL)))
		goto out_args;

	if ((status = nfs4_handle_putfh(&cres, &putfh)))
		goto out_res;
	if ((status = nfs4_handle_reval(&cres, &reval)))
		goto out_res;
	/* XXX: The refcount handling here is paranoid... */
	open->post_file_data = NULL;
	if ((status = nfs4_handle_open(&cargs, &cres, open))) {
		if (open->post_file_data) {
			nfs4_put_file_data(open->post_file_data);
			open->post_file_data = NULL;
		}
	}

out_res:
	nfs4_release_compoundres(&cres);
out_args:
	nfs4_release_compoundargs(&cargs);
out:
	return status;
}

int
nfs4fs_do_close(struct dentry *dentry, struct nfs4fs_close *close)
{
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_putfh putfh;
	struct nfs4fs_getattr reval;
	int status;

#if 0
	/* (This section disabled for testing.) */
	/* Optimize out an OPEN_DOWNGRADE if the share access would not actually change. */
	struct nfs4fs_file_data *file_data = close->data;
	
	status = 0;
	down_write(&file_data->fi_sema);
	if ((file_data->fi_nopen > 1) &&
	    (!close->read || (file_data->fi_nread > 1)) &&
	    (!close->write || (file_data->fi_nwrite > 1))) {
		file_data->fi_nread -= close->read;
		file_data->fi_nwrite -= close->write;
		up_write(&file_data->fi_sema);
		goto out;
	}
	up_write(&filedata->fi_sema);
#endif

	cargs.sb = dentry->d_sb;
	cargs.tag = "nfs4fs_do_close()";
	putfh.dentry = dentry;

	if ((status = nfs4_setup_compound(&cargs)))
		goto out;
	if ((status = nfs4_setup_putfh(&cargs, &putfh)))
		goto out_args;
	if ((status = nfs4_setup_close(&cargs, close)))
		goto out_args;
	if ((status = nfs4_setup_reval(&cargs, &reval)))
		goto out_args;

	if ((status = nfs4_call_compound(&cargs, &cres, NULL)))
		goto out_args;
	if ((status = nfs4_handle_putfh(&cres, &putfh)))
		goto out_res;
	if ((status = nfs4_handle_close(&cres, close)))
		goto out_res;
	nfs4_handle_reval(&cres, &reval);   /* ignore return value */

out_res:
	nfs4_release_compoundres(&cres);
out_args:
	nfs4_release_compoundargs(&cargs);
out:
	return status;
}

/* In nfsv4, there is no way to specify O_TRUNC but not O_CREAT on an OPEN.  To
 * properly honor O_TRUNC but not O_CREAT, the open routines call this function
 * to truncate the file in a seperate request after opening it.
 */
static int
nfs4fs_do_truncate(struct dentry *dentry, struct nfs4fs_file_data *fdata)
{
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_putfh putfh;
	struct iattr iattr;
	struct nfs4fs_getattr getattr;
	int status;
	
	cargs.sb = dentry->d_sb;
	cargs.tag = "__nfs4_truncate()";
	putfh.dentry = dentry;
	iattr.ia_valid = ATTR_SIZE;
	iattr.ia_size = 0;
	getattr.flags = FL_NONFATAL;
	getattr.bm = &nfs4_negative_bitmap;

	if ((status = nfs4_setup_compound(&cargs)))
		goto out;
	if ((status = nfs4_setup_putfh(&cargs, &putfh)))
		goto out_args;
	if ((status = nfs4_setup_setattr(&cargs, &iattr, fdata)))
		goto out_args;
	
	if ((status = nfs4_call_compound(&cargs, &cres, NULL)))
		goto out_args;
	if ((status = nfs4_handle_putfh(&cres, &putfh)))
		goto out_res;
	if ((status = nfs4_handle_setattr(&cres)))
		goto out_res;
	nfs4_handle_getattr(&cres, &getattr);    /* ignore return value */

out_res:
	nfs4_release_compoundres(&cres);
out_args:
	nfs4_release_compoundargs(&cargs);
out:
	return status;
}

/* i_op->create(): creates a previously nonexistent file, hopefully without a race condition.
     This is problematic in nfsv4, since we can't create a file without opening it.
     This routine is called two ways from the VFS: via sys_mknod() and sys_open().  In the
   former case, we really do want to create a previously nonexistent file, but not open it,
   so we can just create the file and close it.  In the latter case, things are very
   difficult.  Just opening the file, then closing it, presents this problem: What do we do if
   the create produces the error NFS4ERR_EXIST?  (This is possible if another client creates
   the file at just the right time.)  If we return an error, we break open()'s where O_CREAT
   is set but O_EXCL is not.  If we return no error, we break sys_mknod() for regular files,
   since it assumes exclusive creation semantics.  Additionally, opening the file then closing
   it is a bad solution performancewise because the file will be opened on the server, then
   closed, then opened again.. just to implement a single sys_open().
     The root of the problem is that when i_op->create() is called, we don't know whether an
   exclusive create is called for, or whether we are trying to open the file (vs. just creating
   it without opening).
     We are guaranteed that dentry is a hashed, negative dentry.
   Locks held: BKL, dir_inode->i_sem, dir_inode->i_zombie.
 */

int
nfs4_create(struct inode *dir_inode, struct dentry *dentry, int mode)
{
	struct dentry *dir_dentry = dentry->d_parent;
	struct nfs4fs_open open;
	struct nfs4fs_close close;
	int status;
	
	/* as described above, we can assume that this function is only called when exclusive
	   creation semantics are needed */
	
#if NFS4_DEBUG
	status = -EINVAL;
	if (!dir_dentry) {
		printk("nfs4_create: dir_dentry is NULL!!\n");
		goto out;
	}
	if (dir_dentry->d_inode != dir_inode) {
		printk("nfs4_create: inode mismatch!!\n");
		goto out;
	}
	if (dentry->d_inode) {
		printk("nfs4_create: positive dentry!!\n");
		goto out;
	}
	printk("nfs4: warning: ->create() method called!\n");
#endif

	nfsv4_printk(level2, "nfs4_create: starting on %.*s/%.*s\n",
		     (int)dir_dentry->d_name.len, dir_dentry->d_name.name,
		     (int)dentry->d_name.len, dentry->d_name.name);

	open.claim_type = CLAIM_NULL;
	open.dentry = dentry;
	open.flags = O_CREAT | O_EXCL;
	open.mode = mode;
	open.lockowner = nfs4_get_lockowner(dir_inode->i_sb);
	open.pre_file_data = NULL;
	if ((status = nfs4fs_do_open(&open)))
		goto out_put;
	
#if NFS4_DEBUG
	if (!open.post_file_data) {
		printk("!!! nfs4_create: no file data !!!\n");
		goto out_put;
	}
#endif

	close.data = open.post_file_data;
	close.read = 0;
	close.write = 0;
	status = nfs4fs_do_close(dentry, &close);
	nfs4_put_file_data(open.post_file_data);

out_put:
	nfs4_put_lockowner(open.lockowner);
out:
	nfsv4_printk(level2, "nfs4_create: returning status %d\n", status);
	return status;
}

int
nfs4_lookup_open(struct inode *dir_inode, struct dentry *dentry, struct nameidata *nd)
{
	struct nfs4fs_open open;
	struct inode *inode = dentry->d_inode;
	struct files_struct *fs = current->files;
	pid_t pid = current->pid;
	struct list_head *head;
	struct file *filp;
	int flags;
	int status;

#if NFS4_DEBUG
	if (dentry->d_parent == dentry) {
		printk("!!! lookup_open: at mountpoint !!!\n");
		return -EIO;
	}
	if (dentry->d_parent->d_inode != dir_inode) {
		printk("!!! lookup_open: inode mismatch !!!\n");
		return -EIO;
	}
	if ((dentry->d_name.len != nd->last.len) ||
	    (memcmp(dentry->d_name.name, nd->last.name, nd->last.len))) {
		printk("!!! lookup_open: name mismatch !!!\n");
		return -EIO;
	}
	if (inode && S_ISLNK(inode->i_mode)) {
		printk("!!! lookup_open: symlink?!!!\n");
		return -EIO;
	}
#endif

	nfsv4_printk(level2, "nfs4_lookup_open: starting on (%.*s/%.*s)\n",
		     (int)dentry->d_parent->d_name.len, dentry->d_parent->d_name.name,
		     (int)dentry->d_name.len, dentry->d_name.name);
	
	lock_kernel();          /* XXX: is this necessary? */

	/* XXX: Where is the proper place to put this?  I'm putting it right up front
	   because this seems the safest.. */
	dentry->d_op = &nfs4_dentry_operations;
	
	/* 
	 * By design, ->lookup_open() opens an unvalidated dentry (which may even be
	 * negative).  Since we can't be sure which file we're opening, which stateid
	 * do we lock?  If we guess incorrectly, we may end up with ERR_OLD_STATEID on
	 * concurrent read/write requests (note that there can be no concurrent locking
	 * requests because we can always lock the correct seqid).  We deal with this by
	 * making a "best guess" for the stateid.  We therefore expect to get ERR_OLD_STATEID
	 * occasionally on I/O requests -- in the compound handler, we deal with this error
	 * by printing a warning message and retrying the operation.
	 */
	open.lockowner = NULL;
	open.pre_file_data = NULL;
	open.post_file_data = NULL;
	if (inode) {
		if (!S_ISREG(inode->i_mode)) {
			/* nfs4_dentry_revalidate() returns 0 on failure */
			status = -EAGAIN;
			if (!nfs4_dentry_revalidate(dentry, 0))
				goto out;
			goto create_filp;
		}
		head = FILEDATA_HEAD(inode, fs, pid);
		spin_lock(&filedata_spinlock);
		open.pre_file_data = file_data_seek(head, inode, fs, pid);
		spin_unlock(&filedata_spinlock);
		if (open.pre_file_data) {
			open.lockowner = open.pre_file_data->fi_lockowner;
			goto got_lockowner;
		}
	}
	open.lockowner = __nfs4_get_lockowner(dentry->d_sb, fs, pid);
	
got_lockowner:
	open.claim_type = CLAIM_NULL;
	open.dentry = dentry;
	open.flags = nd->open_flags;
	open.mode = nd->mode;

	status = nfs4fs_do_open(&open);
	switch (status) {
	case 0:
		/* XXX: Is this the right place to put this?  Do we need the
		 *   if (d_unhashed())?  Should this also be placed in other
		 *   branches of the switch?
		 */
		if (d_unhashed(dentry))
			d_rehash(dentry);
#if NFS4_DEBUG
		if (!dentry->d_inode || !S_ISREG(dentry->d_inode->i_mode)) {
			printk("!!! nfs4_lookup_open: success, not regular file?!\n");
			status = -EIO;
			goto out_release;
		}
#endif
		break;

	case -EISDIR:   /* directory. XXX: is this consistent w/spec? */
	case -ELOOP:    /* symlink */
	case -EINVAL:   /* any other non-regular file type */
		/* XXX: this assumes that there is only one set of conditions under which
		 * NFS4ERR_INVAL is returned: if an OPEN is attempted on a non-regular file,
		 * non-symlink.  Is this assumption correct?
		 *
		 * It seems that out OPEN failed because a non-regular file exists by the
		 * same name.  We make the dentry negative if possible, then attempt to look
		 * up the file.
		 */
		if (dentry->d_inode) {
			d_delete(dentry);
			/* Note that d_delete() will make the dentry negative if possible
			 * (i.e., if nobody else is using it).  If not, it will drop it from
			 * the dcache, and we will have to try the whole ->lookup_open()
			 * operation again.
			 */
			status = -EAGAIN;
			if (dentry->d_inode)
				goto out_release;
		}

		/* Dentry is now negative, go ahead and try the file lookup. */
		status = nfs4fs_do_lookup(dentry);
		if (status)
			goto out_release;
		if (d_unhashed(dentry))
			d_rehash(dentry);
		
		/* Confusion! When we tried the lookup, we didn't find anything, even though
		   the original OPEN failed.  Just try the whole thing again.. */
		inode = dentry->d_inode;
		status = -EAGAIN;
		if (!inode)
			goto out_release;

		status = -ELOOP;
		if (S_ISLNK(inode->i_mode))
			goto out_release;
		if (!S_ISREG(inode->i_mode))
			goto create_filp;

		/* Confusion!  When we tried the lookup, we found a regular file, even though
		   the original OPEN failed.  Just try the whole thing again.. */
		status = -EAGAIN;
		goto out_release;
		
	default:
		goto out_release;
	}

	/* The file has been opened, but there is one final wrinkle:  If O_TRUNC was set but
	 * not O_CREAT, we have to truncate the file in a seperate request (just to make life
	 * more interesting, this combination of flags can't be put into an open in nfsv4).
	 *
	 * XXX: If truncate() fails, need to issue a CLOSE or OPEN_DOWNGRADE to avoid
	 * "leaking state" onto the server..
	 */
	if ((open.flags & O_TRUNC) && !(open.flags & O_CREAT) && dentry->d_inode->i_size) {
		if ((status = nfs4fs_do_truncate(dentry, open.post_file_data))) {
			nfs4_put_file_data(open.post_file_data);
			goto out_release;
		}
	}

create_filp:
	/* XXX: If create_filp() fails, we should issue a CLOSE or OPEN_DOWNGRADE to avoid
	   leaking state onto the server.  For now, I'm omitting this code path since it is
	   such an unlikely situation. */
	
	flags = nd->open_flags;
	if (flags & O_ACCMODE)
		flags--;
	filp = create_filp(dentry, nd->mnt, flags);
	status = PTR_ERR(filp);
	if (IS_ERR(filp)) {
		nfs4_put_file_data(open.post_file_data);
		goto out_release;
	}

	filp->private_data = open.post_file_data;
	nd->file = filp;
	status = NFS4_OK;

out_release:
	if (open.pre_file_data)
		nfs4_put_file_data(open.pre_file_data);
	else if (open.lockowner)
		nfs4_put_lockowner(open.lockowner);
out:
	unlock_kernel();
	nfsv4_printk(level2, "nfs4_lookup_open: returning status %d\n", status);
	return status;
}

int
nfs4_open(struct inode *inode, struct file *filp)
{
	struct dentry *dentry = filp->f_dentry;
	struct nfs4fs_open open;
	int status;
	
	/* We have arranged things so that the VFS calls ->lookup_open() instead of
	 * ->open().  The ->open() method is prone to inefficiency and race conditions.
	 * However, there are a few circumstances under which the ->open() method will
	 * be called: (1) Upon opening the root of a mounted filesystem, (2) certain
	 * device drivers will call ->open() directly from within the kernel, (3) in
	 * a similar vein, character device files always use ->open() directly, (4) if
	 * an NFSv4 filesystem is being exported using NFS, ->open() will be called
	 * directly (of course, NFS4->NFS exports will suffer from more serious
	 * problems than this!)
	 *
	 * XXX: assumption: dentry has been revalidated prior to ->open() call
	 */

#if NFS4_DEBUG
	if (dentry->d_inode != inode) {
		printk("!!! nfs4_open: inode mismatch !!!\n");
		return -EIO;
	}
#endif
	
	nfsv4_printk(level2, "nfs4_open: starting on (%.*s/%.*s)\n",
		     (int)dentry->d_parent->d_name.len, dentry->d_parent->d_name.name,
		     (int)dentry->d_name.len, dentry->d_name.name);
	
	lock_kernel();          /* XXX: is this necessary? */

	if (!S_ISREG(inode->i_mode)) {
		filp->private_data = NULL;
		status = 0;
		goto out;
	}
	status = -ENOMEM;
	if (!(open.pre_file_data = nfs4_get_file_data(dentry)))
		goto out;
	open.lockowner = open.pre_file_data->fi_lockowner;
	open.claim_type = CLAIM_NULL;
	open.dentry = dentry;
	
	/* Ugh.  There are two conventions in use in the kernel for the lower two bits of
	 * the open flags: the 'sys_open' convention, and the 'open_namei' convention.  We
	 * want to use 'filp->f_mode', which uses the 'open_namei' convention for the lower
	 * two bits, but drops the upper bits entirely.  Note that we are purposefully
	 * dropping the upper bits, since we want to avoid creating files here.  If O_TRUNC
	 * is specified, we will have to take care of it seperately by calling
	 * nfs4fs_do_truncate() later.
	 */

#if NFS4_DEBUG
	/* check consistency between the lower two bits of f_flags and f_mode... */
	if (filp->f_mode != ((filp->f_flags+1) & 3)) {
		printk("!!! nfs4_open: inconsistent mode bits !!!\n");
		status = -EIO;
		goto out_release;
	}
#endif
	open.flags = filp->f_mode;

	/* We try the OPEN.  Note that this can fail if another client snuck in and replaced
	 * our file by one with the same name.  This race condition is inherent, and indeed
	 * is the whole motivation for introducing the ->lookup_open() method.  If we trigger
	 * this race condition, we just return -EIO.  (I don't think that this will be a
	 * problem in practice, since the ->lookup_open() method, which doesn't suffer from
	 * the race condition, will always be used except in exceptional circumstances.)
	 *
	 * XXX: We currently "leak state" on the server in the event of this race condition..
	 * need a recovery path which cleans up using a CLOSE or OPEN_DOWNGRADE.
	 */
	if ((status = nfs4fs_do_open(&open)))
		goto open_err;

	status = -EIO;
	if (open.post_file_data != open.pre_file_data) {
		nfsv4_printk(open2, "nfs4_open: didn't open the right file!\n");
		nfs4_put_file_data(open.post_file_data);
		goto out_release;
	}
	nfs4_put_file_data(open.post_file_data);  /* already have reference in pre_file_data */
	
	/* The file has been opened, but there is one final wrinkle:  If O_TRUNC was set,
	 * we have to truncate the file in a seperate request.
	 *
	 * XXX: If truncate() fails, need to issue a CLOSE or OPEN_DOWNGRADE to avoid
	 * "leaking state" onto the server..
	 */
	if ((open.flags & O_TRUNC) && dentry->d_inode->i_size) {
		if ((status = nfs4fs_do_truncate(dentry, open.pre_file_data)))
			goto out_release;
	}

	filp->private_data = open.pre_file_data;
	status = 0;

out:
	unlock_kernel();
	nfsv4_printk(level2, "nfs4_lookup_open: returning status %d\n", status);
	return status;
open_err:
	/* A return value of NFS4ERR_SYMLINK or NFS4ERR_INVAL presumably means that the
	 * server encountered a non-regular file.  We map these errors to -EIO because they
	 * fall under the umbrella of the race condition described above.  XXX: is the
	 * assumption about NFS4ERR_INVAL correct (we make this same assumption in
	 * nfs4_lookup_open() below)?
	 */
	if ((status == -ELOOP) || (status == -EINVAL))
		status = -EIO;
out_release:
	nfs4_put_file_data(open.pre_file_data);
	goto out;
}

/* note: the vfs does not check the return value of this function, so we just return 0 even
   if there is an error.  XXX: should we try to decrease latency by dispatching the CLOSE
   asynchronously and returning immediately? */

int
nfs4_release(struct inode *inode, struct file *filp)
{
	struct nfs4fs_file_data *file_data = filp->private_data;
	struct dentry *dentry = filp->f_dentry;
	int accflags = filp->f_flags & O_ACCMODE;
	struct nfs4fs_close close;
	
	lock_kernel();
	
	nfsv4_printk(level2, "nfs4_release: starting on \"%.*s\"\n",
		     (int)dentry->d_name.len, dentry->d_name.name);

	/* XXX: a bad way to catch the S_ISDIR case? */
	if (!file_data)
		goto out;

	/*
	 * XXX: I added this at the bakeoff.  Is it needed?
	 */
	nfs4_sync_file(dentry->d_inode, NULL, 0, 0, FLUSH_WAIT);
	
	close.read = (accflags != O_WRONLY);
	close.write = (accflags != O_RDONLY);
	close.data = file_data;
	if (nfs4fs_do_close(dentry, &close))
		printk(KERN_WARNING "nfs4_release: close failed\n");
	nfs4_put_file_data(file_data);
	filp->private_data = NULL;
	
out:
	unlock_kernel();
	nfsv4_printk(level2, "nfs4_release: done\n\n");
	return 0;
}

/* This kernel thread is invoked when the client believes that the server has rebooted.
   It tries to reestablish lost state on the server, then wakes up threads waiting for
   the grace period to end. */

#define RECOVERD_SLACK    (1 * HZ)

int
nfs4fs_recoverd(void *arg)
{
	BUG();
	return 0;
	
#if 0
	static u32 recoverd_bmval[] = { 0x00000400 };
	static bitmap4 recoverd_bitmap = { 1, recoverd_bmval };
	struct super_block *sb = (struct super_block *)arg;
	struct list_head *l;
	struct nfs4fs_file_data *fdata;
	struct inode *inode;
	struct list_head inode_list;
	unsigned long end_of_grace_period;
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_getattr getattr;
	int status;

	sprintf(current->comm, "recoverd");
	lock_kernel();

	printk(KERN_INFO "nfs4: server has rebooted, trying to reconnect...\n");

	/*  Wait for server to come back up, re-instantiate server's lease time.
	    XXX: do we have to explicitly reconnect in the TCP case, or will the RPC layer
	    take care of this?
	    XXX: need to put more thought into management of timeouts here.. */
	
	cargs.sb = sb;
	cargs.tag = "recoverd";
	getattr.bm = &recoverd_bitmap;

	if ((status = nfs4_setup_compound(&cargs)))
		goto set_lease;
	if ((status = nfs4_setup_putrootfh(&cargs)))
		goto free_args;
	if ((status = nfs4_setup_getattr(&cargs, &getattr)))
		goto free_args;
	if ((status = nfs4_call_compound(&cargs, &cres, NULL)))
		goto free_args;
	if ((status = nfs4_handle_putrootfh(&cres)))
		goto free_res;
	if ((status = nfs4_handle_getattr(&cres, &getattr)))
		goto free_res;

free_res:
	nfs4_release_compoundres(&cres);
free_args:
	nfs4_release_compoundargs(&cargs);
set_lease:
	if (!status && (getattr.fattr.fa_valid & FA_LEASE_TIME)) {
		NFS4_LEASE_TIME(sb) = getattr.fattr.fa_lease_time;
		if (NFS4_LEASE_TIME(sb) < NFS4FS_MIN_LEASE)
			NFS4_LEASE_TIME(sb) = NFS4FS_MIN_LEASE;
		if (NFS4_LEASE_TIME(sb) > NFS4FS_MAX_LEASE)
			NFS4_LEASE_TIME(sb) = NFS4FS_MAX_LEASE;
		end_of_grace_period = jiffies + NFS4_LEASE_TIME(sb) * HZ;
	}
	else {
		printk(KERN_WARNING
		       "nfs4: couldn't determine server lease time, using default");
		NFS4_LEASE_TIME(sb) = NFS4FS_DEFAULT_LEASE;
		end_of_grace_period = jiffies + NFS4FS_DEFAULT_LEASE * HZ + RECOVERD_SLACK;
	}
	
	/*
	 *  Our old clientid presumably won't work, so get a new one now...
	 */
	NFS4_HAVE_CLIENTID(sb) = 0;
	if ((status = nfs4_do_setclientid(sb))) {
		printk(KERN_WARNING "nfs4: couldn't get a clientid from server!\n");
		/* oh well, try to use old clientid and hope for the best... */
		NFS4_HAVE_CLIENTID(sb) = 1;
	}

	/*
	 *  Now recover all open files, and locking state for each
	 */

	INIT_LIST_HEAD(&inode_list);
	list_for_each(l, &NFS4_FILE_LIST(sb)) {
		fdata = list_entry(l, struct nfs4fs_file_data, fi_list);
		recover_open(sb, fdata, &inode_list);
	}
	while (!list_empty(&inode_list)) {
		inode = list_entry(inode_list.next, struct inode, u.nfs4_i.recovery_list);
		recover_locks(sb, inode);       /* will remove inode from inode_list */
	}
	
	/*
	 *  Sleep until end of grace period, then wake up threads which are blocked waiting
	 *  for grace period to end.
	 */
	current->state = TASK_UNINTERRUPTIBLE;
	status = schedule_timeout(end_of_grace_period - jiffies);
	current->state = TASK_RUNNING;
	NFS4_PIDOF_RECOVERD(sb) = 0;
	wake_up_all(&NFS4_GRACE_WAIT_QUEUE(sb));
	unlock_kernel();
	return 0;    /* XXX: who gets this return value? */
#endif
}

#if 0
static void
recover_open(struct super_block *sb, struct nfs4fs_file_data *fdata,
	     struct list_head *inode_list)
{
	printk("!!! recover_open() hit !!!\n");
	return;
	
	struct dentry *dentry = fdata->fi_dentry;
	struct inode *inode = dentry->d_inode;
	struct nfs4fs_lockowner *lockowner = fdata->fi_lockowner;
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_putfh putfh;
	struct nfs4fs_open open;
	int status;
	
	cargs.sb = sb;
	cargs.tag = "recover_open()";
	putfh.dentry = dentry;
	open.claim_type = CLAIM_PREVIOUS;
	open.dentry = dentry;
	open.flags = 0;
	if (fdata->fi_nread)
		open.flags |= FMODE_READ;
	if (fdata->fi_nwrite)
		open.flags |= FMODE_WRITE;
	open.lockowner = lockowner;
	open.file_data = fdata;
		
	if ((status = nfs4_setup_compound(&cargs)))
		goto process_inode;
	if ((status = nfs4_setup_putfh(&cargs, &putfh)))
		goto free_args;;
	if ((status = nfs4_setup_open(&cargs, &open)))
		goto free_args;
	if ((status = nfs4_call_compound(&cargs, &cres, NULL)))
		goto free_args;
	if ((status = nfs4_handle_putfh(&cres, &putfh)))
		goto free_res;
	if ((status = nfs4_handle_open(&cargs, &cres, &open)))
		goto free_res;

free_res:
	nfs4_release_compoundres(&cres);
free_args:
	nfs4_release_compoundargs(&cargs);
process_inode:
	if (status) {
		printk(KERN_WARNING
		       "nfs4: recoverd couldn't reacquire open! (pid %d, ino %ld)\n",
		       lockowner->lo_pid, inode->i_ino);
	}
	else if (inode->i_flock && !list_empty(&NFS4_RECOVERY_LIST(inode)))
		list_add(&NFS4_RECOVERY_LIST(inode), inode_list);
}

static void
recover_locks(struct super_block *sb, struct inode *inode)
{
	struct file_lock **flp, *fl;
	struct nfs4fs_lockowner *lockowner;
	struct nfs4fs_file_data *fdata;
	struct list_head *head;
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_putfh putfh;
	struct nfs4fs_lock lock;
	int status;
	
	if (NFS4_STRICT_LOCKING(sb))
		down(&NFS4_LOCKING_SEMAPHORE(inode));

	flp = &inode->i_flock;
	while (*flp) {
		status = -EIO;
		fl = *flp;
		if (!(lockowner = __nfs4_get_lockowner(sb, fl->fl_owner, fl->fl_pid)))
			goto process_lock;
		head = FILEDATA_HEAD(inode, fl->fl_owner, fl->fl_pid);
		if (!(fdata = file_data_seek(head, inode, fl->fl_owner, fl->fl_pid)))
			goto process_lock;
		
		cargs.sb = sb;
		cargs.tag = "recover_locks()";
		putfh.dentry = fdata->fi_dentry;
		lock.reclaim = 1;
		lock.file_data = fdata;
		lock.type = (fl->fl_type == F_RDLCK) ? READ_LT : WRITE_LT;
		lock.start = fl->fl_start;
		lock.end = fl->fl_end;

		if ((status = nfs4_setup_compound(&cargs)))
			goto process_lock;
		if ((status = nfs4_setup_putfh(&cargs, &putfh)))
			goto free_args;
		if ((status = nfs4_setup_lock(&cargs, &lock)))
			goto free_args;
		if ((status = nfs4_call_compound(&cargs, &cres, NULL)))
			goto free_args;
		if ((status = nfs4_handle_putfh(&cres, &putfh)))
			goto free_res;
		if ((status = nfs4_handle_lock(&cres, &lock)))
			goto free_res;

	free_res:
		nfs4_release_compoundres(&cres);
	free_args:
		nfs4_release_compoundargs(&cargs);
	process_lock:
		if (!status) {
			/* successful reacquisition of lock */
			flp = &fl->fl_next;
			continue;
		}
		printk(KERN_WARNING "nfs4: recoverd couldn't reacquire lock! "
		       "(pid %d, ino %ld [%Ld,%Ld])\n", fl->fl_pid, inode->i_ino,
		       fl->fl_start, fl->fl_end);
		locks_delete_lock(flp, 0);
	}
	
	if (NFS4_STRICT_LOCKING(sb))
		up(&NFS4_LOCKING_SEMAPHORE(inode));
	list_del_init(&NFS4_RECOVERY_LIST(inode));
}
#endif
