/*
 *  linux/fs/nfs4/proc.c
 *
 *  Copyright (c) 2001 The Regents of the University of Michigan.
 *  All rights reserved.
 *
 *  Kendrick Smith <kmsmith@umich.edu>
 *  Andy Adamson <andros@umich.edu>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/mm.h>
#include <linux/pagemap.h>
#include <asm/uaccess.h>

#include <linux/nfs4/nfs4_prot.h>
#include <linux/nfs4/nfs4_debug.h>
#include <linux/nfs4/nfs4.h>
#include <linux/nfs4/nfs4fs.h>

/*
 *  NFSv4 async tasks wait here if waiting to retry a semaphore or waiting for the server's
 *  grace period to expire.
 */
static struct rpc_wait_queue    retry_queue = RPC_INIT_WAITQ("nfs4_retry");

static inline void
finalize_args(struct compound4args *argp)
{
	int i;
	u32 *p;

	FINALIZE_IOVEC(argp);
	*argp->nopsp = htonl(argp->nops);
	if (argp->seqid_holder) {
		if (!(argp->flags & CA_SEQID_LOCKED)) {
			down(&argp->seqid_holder->lo_sema);
			argp->flags |= CA_SEQID_LOCKED;
		}
		nfsv4_printk(seqid, "nfs4_call_compound: using seqid %d, lockowner %d\n",
			     argp->seqid_holder->lo_seqid, argp->seqid_holder->lo_id);
		p = argp->seqidp;
		*p++ = htonl(argp->seqid_holder->lo_seqid);
	}
	if (argp->stateid_holder) {
		if (!(argp->flags & CA_STATEID_LOCKED)) {
			if (argp->flags & CA_STATEID_RELEASE)
				nfs4_down_read(&argp->stateid_holder->fi_sema);
			else
				nfs4_down_write(&argp->stateid_holder->fi_sema);
			argp->flags |= CA_STATEID_LOCKED;
		}
		for (i = 0; i < argp->nstateids; i++)
			memcpy(argp->stateidp[i], argp->stateid_holder->fi_stateid,
			       sizeof(stateid4));
	}
	argp->timestamp = jiffies;
}

/*
 * For asynchronous RPC's, we need a nonblocking version of finalize_args().  Asynchronous
 * RPC's cannot block safely because they are run in rpciod's context, which could deadlock
 * if it blocked waiting for an incoming RPC reply to release a seqid or stateid semaphore.
 */
static inline int
finalize_args_nonblocking(struct compound4args *argp)
{
	int i;
	int status;
	
#if NFS4_DEBUG
	/* The only "interesting" ops that we put in async RPC's are READ,WRITE,RENEW.
	 * These ops satisfy certain restrictions, which we verify as a consistency check...
	 */
	if (argp->seqid_holder) {
		printk("!!! seqid_holder in async compound !!!\n");
		return 1;
	}
	if (argp->stateid_holder && ((argp->flags & CA_STATEID_LOCKED) ||
				     !(argp->flags & CA_STATEID_RELEASE) ||
				     !argp->nstateids)) {
		printk("!!! bad stateid_holder in async compound !!!\n");
		return 1;
	}
#endif
	
	FINALIZE_IOVEC(argp);
	*argp->nopsp = htonl(argp->nops);

	status = 0;
	if (argp->stateid_holder) {
		for (i = 0; i < argp->nstateids; i++)
			memcpy(argp->stateidp[i], argp->stateid_holder->fi_stateid,
			       sizeof(stateid4));
		status = nfs4_down_read_trylock(&argp->stateid_holder->fi_sema);
	}
	argp->timestamp = jiffies;
	return status;
}

int
nfs4_call_compound(struct compound4args *argp, struct compound4res *resp,
		   struct nfs4fs_ctx *ctx)
{
	int status;
	
retry:
	nfsv4_printk(level2, "nfs4_call_compound starting: (\"%s\") ctx: %p\n",
		     argp->tag, ctx);	
	finalize_args(argp);
	
	if ((status = rpc_call(NFS4_CLNT(argp->sb), NFSPROC4_COMPOUND, argp, resp, 0))) {
		nfsv4_printk(level2, "nfs4_call_compound (\"%s\"): err=%d\n",
			     argp->tag, status);
		goto out;
	}
#if 0
	printk("     ===============       \n");
	hex_dump(resp->p, (char *)resp->end - (char *)resp->p);
	printk("     ===============       \n");
#endif

	/*
	 *  If you change any of the special error handlers here, don't forget to change
	 *  the corresponding ones in the async handler! (nfs4_async_handle_compound)
	 */
	
	status = nfs4_handle_compound(argp, resp, ctx);
	switch (status) {
	case 0:
		break;
	case -EREBOOT:
		/* This special return value means that we must start recoverd, wait
		   for the grace period to end, and retry the operation.  An exception
		   is if the caller has specified CA_NORETRY: then we just propagate
		   EREBOOT up to the caller, and let it sort things out... */
		
		status = -EIO;
		if (argp->flags & CA_FAIL_RETRY)
			goto out;
		printk(KERN_WARNING "nfs4: server reboot detected!\n");
		if ((status = nfs4fs_start_recoverd(argp->sb))) {
			printk(KERN_WARNING "nfs4: couldn't start recoverd!\n");
			goto out;
		}
		status = -EREBOOT;
		if (argp->flags & CA_NORETRY)
			goto out;
#if NFS4_DEBUG
		if (argp->flags & (CA_SEQID_LOCKED | CA_STATEID_LOCKED)) {
			printk("!!! call_compound: semaphore held while recovering !!!\n");
			return -EIO;
		}
#endif
		if ((status = nfs4fs_wait_on_recoverd(argp->sb)))
			goto out;
		goto retry;			
	case -EWRONGSEC:
		/* This special return value means that we must negotiate security
		   and retry the operation. */
		
		nfsv4_printk(auth, "nfs4_handle_compound: NFS4ERR_WRONGSEC\n");
		if (!ctx || !ctx->pt_sb || !ctx->pt_dentry)
			goto ctx_error;
		status = nfs4_do_secinfo(ctx->pt_sb, ctx->pt_dentry);
		nfsv4_printk(auth,"nfs4_handle_compound: nfs4_do_secinfo returns %d\n",status);
		if (status)
			goto out;
		goto retry;
	default:
		break;
	}

out:
	nfsv4_printk(level2, "nfs4_proc_compound: returning status %d\n", status);
	return status;
ctx_error:
	nfsv4_printk(auth2, "nfs4_handle_compound: invalid nfs4fs_ctx\n");
	status = -EINVAL;
	goto out;
}


static int
nfs4_enc_void(struct rpc_rqst *req, u32 *p, void *dummy)
{
	req->rq_slen = xdr_adjust_iovec(req->rq_svec, p);
	return 0;
}

static int
nfs4_dec_void(struct rpc_rqst *req, u32 *p, void *dummy)
{
	return 0;
}

static int
nfs4_enc_compound(struct rpc_rqst *req, u32 *p, void *dummy)
{
	struct compound4args *cp = (struct compound4args *)dummy;
	char *buf = (char *)p;
	int i;
	int len, buflen;
	int status;
	
	/* XXX : right now, the rpc layer allocates us a buffer, which we copy the
	   data into.  Sooner or later, figure out how to arrange things so that we
	   do all the memory allocation, and pass the buffer down to the RPC layer. */

	len = (char *)p - (char *)req->rq_svec->iov_base;
	buflen = req->rq_svec->iov_len;
	for (i = 0; i < cp->iov_nr; i++) {
		len += cp->iov[i].iov_len;
		if (len > buflen) {
			nfsv4_printk(xdr2, "nfs4_enc_compound: out of buffer space!\n");
			status = -EIO;
			goto out;
		}
		memcpy(buf, cp->iov[i].iov_base, cp->iov[i].iov_len);
		buf += XDR_LEN(cp->iov[i].iov_len);  /* automatically include XDR "baggage" */
	}

	req->rq_slen = xdr_adjust_iovec(req->rq_svec, (u32 *)buf);
	status = 0;

out:
	return status;
}

static int
nfs4_dec_compound(struct rpc_rqst *req, u32 *p, void *dummy)
{
	struct compound4res *rp = (struct compound4res *) dummy;
	int status;
	int len, buflen;

	/* XXX: bakeoff hack */
	char *bakeoff_buffer;
	
	/* We need to determine how much data was received, to test for buffer
	   overflows before they happen.  Unfortunately, both lengths handed down
	   from the rpc layer, req->rq_rlen and req->rq_rvec[0].iov_len, seem to
	   be garbage!  The only reliable way to get the length appears to be
	   venturing into the "private" part of the request structure, and using
	   req->rq_task->tk_status.  Since this is only an ad hoc procedure which
	   seems to work, not something guaranteed to work by the RPC layer, we
	   subject the length to some consistency checks. */

	if (req->rq_rnr != 1) {
		printk("nfs4_dec_compoundres: multiple iovecs received!\n");
		goto out_bad;
	}
	if (!req->rq_task) {
		printk("nfs4_dec_compoundres: req->rq_task is NULL!\n");
		goto out_bad;
	}
	
	len = (char *)p - (char *)req->rq_rvec->iov_base;
	buflen = req->rq_task->tk_status;
	if ((len < 0) || (len >= buflen)) {
		printk("bad lengths in nfs4_dec_compoundres()!\n");
		goto out_bad;
	}

#if 0
	rp->iov_nr = 1;
	rp->iov_current = 0;
	rp->iov[0].iov_len = buflen - len;
	rp->iov[0].iov_base = (char *)p;
	rp->p = (char *)p;
	rp->end = (char *)p + (buflen - len);
#else
	if (!(bakeoff_buffer = NFS4_ALLOC(buflen - len))) {
		printk("XXX couldn't allocate bakeoff buffer!\n");
		goto out_nomem;
	}
	memcpy(bakeoff_buffer, p, buflen - len);
	rp->iov_nr = 1;
	rp->iov_current = 0;
	rp->iov[0].iov_len = buflen - len;
	rp->iov[0].iov_base = bakeoff_buffer;
	rp->p = bakeoff_buffer;
	rp->end = bakeoff_buffer + (buflen - len);
#if 0
	printk("XXXXXXXXXXXXXXXXXXXXX 0x%8x XXXXXXXXXXXXXXXX\n", req->rq_xid);
	hex_dump(bakeoff_buffer, buflen - len);
#endif
	
#endif
	  
	nfsv4_printk(xdr, "xdr: received %d bytes from server\n", buflen-len);
	status = 0;

out:
	return status;
out_bad:
	status = -EIO;
	goto out;
out_nomem:
	status = -ENOMEM;
	goto out;
}

void
nfs4_release_compoundargs(struct compound4args *cp)
{
	int i;

	for (i = 0; i < cp->iov_nr; i++)
		if (!cp->iov_static[i])
			NFS4_FREE(cp->iov[i].iov_base);
	cp->iov_nr = 0;
	if (cp->seqid_holder && (cp->flags & CA_SEQID_LOCKED)) {
		up(&cp->seqid_holder->lo_sema);
		cp->seqid_holder = NULL;
	}
	if (cp->stateid_holder && (cp->flags & CA_STATEID_LOCKED)) {
		if (cp->flags & CA_STATEID_RELEASE)
			nfs4_up_read(&cp->stateid_holder->fi_sema);
		else
			nfs4_up_write(&cp->stateid_holder->fi_sema);
		cp->seqid_holder = NULL;
	}
}

void
nfs4_release_compoundres(struct compound4res *rp)
{
	/* XXX: This routine ended up being empty! If it's still empty at the end of
	 * the project, make sure to clean up by removing calls to it...
	 */
	NFS4_FREE(rp->iov[0].iov_base);
	rp->iov[0].iov_base = NULL;
}

struct nfs4_async_compound *
nfs4_async_create(struct super_block *sb)
{
	struct nfs4_async_compound *ctx;
	struct rpc_task *task;

	if ((ctx = NFS4_ALLOC(sizeof(struct nfs4_async_compound)))) {
		task = &ctx->ac_task;
		rpc_init_task(task, NFS4_CLNT(sb), NULL, RPC_TASK_ASYNC);
		task->tk_msg.rpc_proc = NFSPROC4_COMPOUND;
		task->tk_msg.rpc_argp = &ctx->ac_arg;
		task->tk_msg.rpc_resp = &ctx->ac_res;
		task->tk_msg.rpc_cred = NULL;            /* XXX */
		task->tk_calldata = ctx;
		task->tk_release = nfs4_async_release;
		ctx->ac_release = NULL;
	}
	return ctx;
}

void
nfs4_async_release(struct rpc_task *task)
{
	struct nfs4_async_compound *ctx = (struct nfs4_async_compound *)task->tk_calldata;
	
#if NFS4_DEBUG
	if (task != &ctx->ac_task) {
		printk("!!! nfs4_async_releasefunc: task mismatch !!!\n");
		return;
	}
#endif
	if (ctx->ac_release)
		ctx->ac_release(ctx);
	NFS4_FREE(ctx);
}

void
nfs4_async_call_compound(struct rpc_task *task)
{
	struct nfs4_async_compound *ctx = (struct nfs4_async_compound *)task->tk_calldata;
	struct compound4args *argp;
	
#if NFS4_DEBUG
	if (task != &ctx->ac_task) {
		printk("!!! task mismatch in nfs4_async_call_compound !!!\n");
		rpc_exit(task, -EIO);
		return;
	}
#endif
	
	argp = &ctx->ac_arg;

	nfsv4_printk(level2, "nfs4_call_compound starting: (\"%s\") ctx: %p\n",
		     argp->tag, ctx);	
	if (finalize_args_nonblocking(argp)) {
		/* XXX: for now, just using a fixed hardcoded delay if we have to wait for
		 * a semaphore.  Should we switch to exponential backoff?
		 */
		task->tk_timeout = NFS4FS_ASYNC_DELAY;
		task->tk_action = nfs4_async_call_compound;
		rpc_sleep_on(&retry_queue, task, NULL, NULL);
	}
	else {
		task->tk_status = 0;
		rpc_call_setup(task, &task->tk_msg, 0);
	}
}

int
nfs4_async_prehandler(struct nfs4_async_compound *cp, struct nfs4fs_ctx *ctx)
{
	struct rpc_task *task = &cp->ac_task;
	struct compound4args *argp = &cp->ac_arg;
	struct compound4res *resp = &cp->ac_res;
	int status;
	
#if 0
	printk("     ===============       \n");
	hex_dump(resp->p, (char *)resp->end - (char *)resp->p);
	printk("     ===============       \n");
#endif

	/* If something went wrong in the RPC layer, task->tk_status will be nonzero... */
	status = task->tk_status;
	if (status < 0) {
		nfsv4_printk(level2, "nfs4_async_prehandler: tk_status < 0!\n");
		goto out;
	}
	
	status = nfs4_handle_compound(argp, resp, ctx);
	switch (status) {
	case 0:
		break;
	case -EREBOOT:
		printk("!!! nfs4: \"STALE\" err on async call - can't handle this yet !!!\n");
		status = -EIO;
		break;
	case -EWRONGSEC:
		printk("!!! nfs4: WRONGSEC on async call -- can't handle this yet !!!\n");
		status = -EIO;
		break;
	default:
		break;
	}
	task->tk_status = status;

out:
	nfsv4_printk(level2, "nfs4_async_handle_compound: returning status %d\n", status);
	return status;
}

static struct rpc_procinfo nfs4_procedures[] = {
	{ "nfs4_null", (kxdrproc_t)nfs4_enc_void, (kxdrproc_t)nfs4_dec_void,
	  NFS4_BUFSIZE, 0 },
	{ "nfs4_compound", (kxdrproc_t)nfs4_enc_compound, (kxdrproc_t)nfs4_dec_compound,
	  NFS4_BUFSIZE, 0 }
};

static struct rpc_version	 nfs_version4 = {
	4,
	2,
	nfs4_procedures
};

static struct rpc_version *	nfs4_version[] = {
	NULL,
	NULL,
	NULL,
	NULL,
	&nfs_version4
};

struct rpc_program		nfs4_program = {
	"nfs4",
	NFS4_PROGRAM,
	sizeof(nfs4_version) / sizeof(nfs4_version[0]),
	nfs4_version,
	&nfs4_rpcstat,
};

struct rpc_stat			nfs4_rpcstat = { &nfs4_program };
