/*
 * linux/fs/nfs4fs/read.c
 * from linux/fs/nfs/read.c
 * 
 * NFSv4 implementation
 *    Kendrick Smith  <kmsmith@umich.edu>
 *    Andy Adamson  <andros@umich.edu>
 *    Copyright (C) 2001 The Regents of the University of Michigan
 */
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/mm.h>
#include <linux/pagemap.h>
#include <asm/uaccess.h>

#include <linux/nfs4/nfs4_prot.h>
#include <linux/nfs4/nfs4_debug.h>
#include <linux/nfs4/nfs4.h>
#include <linux/nfs4/nfs4fs.h>

#ifndef IS_SWAPFILE
# define IS_SWAPFILE(inode)        0
#endif

/* The page locking in the v3 code seems really screwy.  It seems that, on an
   asynchronous readpage(), the page lock is released only on completion of the actual read,
   whereas on an asynchronous writepage(), the page lock is released immediately, and
   reacquired before doing the actual write.  This seems really bizarre; is it correct? */

static int
nfs4_readpage_sync(struct nfs4fs_file_data *fdata, struct page *page)
{
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_putfh putfh;
	struct nfs4fs_simple_read read;
	struct nfs4fs_getattr reval;
	struct dentry *dentry = fdata->fi_dentry;
	int count = PAGE_CACHE_SIZE;
	int rsize = NFS4_RSIZE(dentry->d_inode);
	int status;

	/* XXX : What kind of revalidation do we really want here? */

	nfsv4_printk(read, "nfs4_readpage_sync(%.*s/%.*s): starting, index=%d\n",
		     (int)dentry->d_parent->d_name.len, dentry->d_parent->d_name.name,
		     (int)dentry->d_name.len, dentry->d_name.name,
		     (int)page->index);

	cargs.sb = dentry->d_sb;
	cargs.tag = "nfs4_readpage_sync()";
	putfh.dentry = dentry;
	read.file_data = fdata;
	read.start = (loff_t)page->index << PAGE_CACHE_SHIFT;
	read.buf = kmap(page);   /* XXX : is there any way for this to return NULL? */
	
	do {
		read.length = (count < rsize) ? count : rsize;
		nfsv4_printk(read, "nfs4_readpage_sync: attempting to read %d bytes\n",
			     read.length);

		if ((status = nfs4_setup_compound(&cargs)))
			goto out_unmap;
		if ((status = nfs4_setup_putfh(&cargs, &putfh)))
			goto out_args;
		if ((status = nfs4_setup_simple_read(&cargs, &read)))
			goto out_args;
		if ((status = nfs4_setup_reval(&cargs, &reval)))
			goto out_args;
		
		lock_kernel();   /* XXX : taken from v3 code; why needed? */
		status = nfs4_call_compound(&cargs, &cres, NULL);
		unlock_kernel();
		if (status)
			goto out_args;
		
		if ((status = nfs4_handle_putfh(&cres, &putfh)))
			goto out_res;
		if ((status = nfs4_handle_simple_read(&cres, &read)))
			goto out_res;
		nfs4_handle_reval(&cres, &reval);  /* XXX: ignore return value? */

		nfs4_release_compoundres(&cres);
		nfs4_release_compoundargs(&cargs);
		
		nfsv4_printk(read, "nfs4_read: read %d bytes, eof=%d\n", read.bytes_read,
			     read.eof);
		read.start += read.bytes_read;
		read.buf += read.bytes_read;
		count -= read.bytes_read;
		if (read.eof) {
			memset(read.buf, 0, count);   /* zero out rest of page */
			break;
		}
		if (!read.bytes_read) {
			nfsv4_printk(read, "nfs4_read: no data read\n");
			status = -EIO;
			goto out_unmap;
		}
	} while (count);

	flush_dcache_page(page);   /* XXX : what does this do?  it's a null macro on intel */
	SetPageUptodate(page);
	if (PageError(page))
		ClearPageError(page);
	
out_unmap:
	kunmap(page);
	UnlockPage(page);
	nfsv4_printk(read, "nfs4_readpage_sync: returning status %d\n", status);
	return status;
out_res:
	nfs4_release_compoundres(&cres);
out_args:
	nfs4_release_compoundargs(&cargs);
	goto out_unmap;				       
}

static inline struct nfs4_page *
nfs4_find_read(struct inode *inode, struct page *page)
{
	struct list_head *head, *next;
	struct nfs4_page *req;

	/* call with spinlock held */
	
	head = &NFS4_FLUSHD_READ(inode);
	next = head->next;
	while (next != head) {
		req = list_entry(next, struct nfs4_page, wb_list);
		next = next->next;
		if (req->wb_page->index != page->index)
			continue;
		req->wb_count++;
		goto out;
	}
	req = NULL;

out:
	return req;
}

static inline void
nfs4_mark_request_read(struct nfs4_page *req)
{
	struct inode *inode = req->wb_inode;

	/* call with spinlock held */
	
	if (!test_bit(PG_BUSY, &req->wb_flags))
		BUG();
	
	if (list_empty(&req->wb_list)) {
		nfs4_list_add_request(req, &NFS4_FLUSHD_READ(inode));
		NFS4_FLUSHD_NREAD(inode)++;
	}
	/*
	 * NB: The call to schedule_inode() can lie inside the spinlock,
	 * since it cannot block in NFSv4...
	 */
	nfs4_schedule_inode(inode, req->wb_timeout);
}

static int
nfs4_readpage_async(struct file *file, struct nfs4fs_file_data *fdata, struct page *page)
{
	struct inode *inode;
	struct nfs4_page  *req, *new = NULL;
	int		result;

	if (!fdata)
		BUG();
	if (!fdata->fi_dentry)
		BUG();
	if (!fdata->fi_inode)
		BUG();
	
	inode = fdata->fi_inode;
	
	for (;;) {
		result = 0;
		if (Page_Uptodate(page))
			break;

		nfsv4_printk(read, "nfs4_readpage_async() calling find_read\n");
		
		spin_lock(&nfs4_wreq_lock);
		req = nfs4_find_read(inode, page);
		if (req) {
			nfsv4_printk(read, "nfs4_readpage_async() found page\n");
			if (page != req->wb_page) {
				/* XXX: can this really happen? */
				nfsv4_printk(read2, "nfs4_readpage_async: page mismatch?!\n");
				spin_unlock(&nfs4_wreq_lock);
				nfs4_release_request(req);
				nfs4_pagein_inode(inode, page->index, 0);
				continue;
			}
			spin_unlock(&nfs4_wreq_lock);
			nfs4_release_request(req);
			break;
		}

		if (new) {
			nfsv4_printk(read, "nfs4_readpage_async: scheduling new page...\n");
			new->wb_timeout = jiffies + NFS4_READ_DELAY;
			nfs4_lock_request(new);
			nfs4_mark_request_read(new);
			spin_unlock(&nfs4_wreq_lock);
			nfs4_unlock_request(new);
			new = NULL;
			break;
		}

		spin_unlock(&nfs4_wreq_lock);
		result = -ENOMEM;
		nfsv4_printk(read, "nfs4_readpage_async: creating new page...\n");
		new = nfs4_create_request(file, fdata, page, 0, PAGE_CACHE_SIZE);
		if (!new)
			break;
	}

	/* XXX: This seems like a sub-optimal way of doing things... */
	/* XXX: Need the spinlock to inspect FLUSHD_NREAD()? */
	if (NFS4_FLUSHD_NREAD(inode) >= NFS4_RPAGES(inode->i_sb) ||
	    page->index == (inode->i_size + PAGE_CACHE_SIZE - 1) >> PAGE_CACHE_SHIFT) {
		nfsv4_printk(read, "nfs4_readpage_async: pageing in inode now!\n");
		nfs4_pagein_inode(inode, 0, 0);
	}
	
	if (new)
		nfs4_release_request(new);
	return result;
}

int
nfs4_readpage(struct file *file, struct page *page)
{
	struct nfs4fs_file_data *fdata;
	struct address_space *mapping;
	struct inode *inode;
	int error;

	nfsv4_printk(pagecache, "nfs4_readpage %d: pagecount=%d flags=0x%x\n",
	       (int)page->index, (int)atomic_read(&page->count),
	       (int)page->flags);

	if (file) {
		fdata = (struct nfs4fs_file_data *)file->private_data;
		if (!file->f_dentry)
			BUG();
		if (!file->f_dentry->d_inode)
			BUG();
		inode = file->f_dentry->d_inode;
	}
	else {
		mapping = page->mapping;
		if (!mapping)
			BUG();
		inode = (struct inode *)mapping->host;
		if (!inode)
			BUG();
		lock_kernel();    /* XXX: granulate */
		if (list_empty(&NFS4_READ_MAPPINGS(inode)))
			BUG();
		fdata = list_entry(NFS4_READ_MAPPINGS(inode).next,
				   struct nfs4fs_file_data, fi_read_map);
		atomic_inc(&fdata->fi_count);
		unlock_kernel();
	}

	nfsv4_printk(read, "nfs4_readpage: starting, index=%ld\n", page->index);

	/*
	 * Try to flush any pending writes to the file..
	 *
	 * NOTE! Because we own the page lock, there cannot
	 * be any new pending writes generated at this point
	 * for this page (other pages can be written to).
	 *
	 * XXX: why FLUSH_STABLE here?
	 */
	error = nfs4_sync_file(inode, NULL, page->index, 1, FLUSH_WAIT | FLUSH_STABLE);
	if (error)
		goto out_error;

	/* XXX: how about adding a mount flag to force sync reading? */
	error = -1;
	if (!PageError(page) && NFS4_RSIZE(inode) >= PAGE_CACHE_SIZE)
		error = nfs4_readpage_async(file, fdata, page);
	if (error >= 0)
		goto out;

	error = nfs4_readpage_sync(fdata, page);   /* will release page lock */
	if ((error < 0) && IS_SWAPFILE(inode))
		printk("Aiee.. nfs swap-in of page failed!\n");

out:
	if (!file)
		nfs4_put_file_data(fdata);
	nfsv4_printk(read, "nfs4_readpage: returning status %d\n", error);
	return error;
out_error:
	UnlockPage(page);
	goto out;
}

static void
read_handler(struct rpc_task *task)
{
	struct nfs4_async_compound *cp = (struct nfs4_async_compound *)task->tk_calldata;
	struct compound4args *argp = &cp->ac_arg;
	struct compound4res *resp = &cp->ac_res;
	unsigned int npages = cp->ac_read.npages;
	struct nfs4_page **pages = cp->ac_read.pages;
	unsigned int bytes_read = 0;
	unsigned int pages_read;
	int eof = 0;
	int i;
	struct nfs4_page *req;
	struct page *page;
	
	if (task->tk_status < 0) {
		nfsv4_printk(read2, "read_handler: task->tk_status < 0!\n");
		goto finish;
	}
	if (nfs4_async_prehandler(cp, NULL)) {
		nfsv4_printk(read2, "read_handler: async_prehandler failed!\n");
		goto finish;
	}
	if (nfs4_handle_putfh(resp, &cp->ac_read_putfh)) {
		nfsv4_printk(read2, "read_handler: handle_putfh failed!\n");
		goto finish_res;
	}
	if (nfs4_handle_read(resp, &cp->ac_read)) {
		nfsv4_printk(read2, "read_handler: handle_putfh failed!\n");
		goto finish_res;
	}
	if (nfs4_handle_reval(resp, &cp->ac_read_reval)) {
		nfsv4_printk(read2, "read_handler: handle_reval failed!\n");
		goto finish_res;
	}
	bytes_read = cp->ac_read.bytes_read;
	eof = cp->ac_read.eof;
	
finish_res:
	nfs4_release_compoundres(resp);
finish:
	nfs4_release_compoundargs(argp);
	npages = cp->ac_read.npages;
	pages_read = bytes_read >> PAGE_CACHE_SHIFT;
	if (eof) {
		struct inode *inode = pages[0]->wb_inode;
		loff_t size;
		
		size = ((loff_t)pages[0]->wb_page->index << PAGE_CACHE_SHIFT) + bytes_read;
		
		/* XXX: is this right?  NFSv2/v3 distinguishes between NFS_CACHE_ISIZE(inode)
		   and inode->i_size.. need to understand thoroughly.. */
		/* XXX: do we have to zero out the rest of the page, a la readpage_sync()? */
		if (inode->i_size != size) {
			nfsv4_printk(read2, "read_handler: size mismatch!\n");
			nfs4_zap_caches(inode);
			pages_read = 0;
		}
		else if (bytes_read & (PAGE_CACHE_SIZE - 1))
			pages_read++;
	}
	nfsv4_printk(read, "read_handler(%.*s): successfully read %d pages, "
		     "failed to read %d\n",
		     (int)pages[0]->wb_fdata->fi_dentry->d_name.len,
		     pages[0]->wb_fdata->fi_dentry->d_name.name,
		     pages_read, npages - pages_read);

	/* clean up.  Note: we are holding the BKL here. */
	for (i = 0; i < npages; i++) {
		req = pages[i];
		page = req->wb_page;
		if (i < pages_read)
			SetPageUptodate(page);
		else
			SetPageError(page);
		flush_dcache_page(page);
		UnlockPage(page);
		nfs4_unlock_request(req);
		if (!req->wb_count)
			BUG();
		nfs4_release_request(req);
	}
	cp->ac_read.npages = 0;   /* so read_release() won't try to clean up twice */
}

static void
read_release(struct nfs4_async_compound *cp)
{
	int i;
	struct nfs4_page *req;

	/* clean up.  Note: we are holding the BKL here. */
	for (i = 0; i < cp->ac_read.npages; i++) {
		req = cp->ac_read.pages[i];
		SetPageError(req->wb_page);
		flush_dcache_page(req->wb_page);
		UnlockPage(req->wb_page);
		nfs4_unlock_request(req);
		if (!req->wb_count)
			BUG();
		nfs4_release_request(req);
	}
	cp->ac_read.npages = 0;  /* paranoid */
}

static int
nfs4_pagein_one(unsigned int npages, struct nfs4_page **pages)
{
	int i;
	struct dentry *dentry;
	struct nfs4_async_compound *cp;
	struct rpc_task *task;
	struct compound4args *argp;
	struct nfs4fs_putfh *putfh;
	struct nfs4fs_read *read;
	struct nfs4fs_getattr *reval;
	struct nfs4_page *req;
	int status;

	if (!npages || (npages > NFS4_MAXIO))
		BUG();
	if (!pages[0]->wb_fdata)
		BUG();
	dentry = pages[0]->wb_fdata->fi_dentry;
	if (!dentry)
		BUG();
	if (!dentry->d_inode)
		BUG();
	
	status = -ENOMEM;
	if (!(cp = nfs4_async_create(dentry->d_sb))) {
		nfsv4_printk(level1, "nfs4: couldn't alloc async read data!\n");
		goto out;
	}
	cp->ac_release = read_release;

	task = &cp->ac_task;
	argp = &cp->ac_arg;
	putfh = &cp->ac_read_putfh;
	read = &cp->ac_read;
	reval = &cp->ac_read_reval;
	
	argp->sb = dentry->d_sb;
	argp->tag = "flushd [read]";
	putfh->dentry = dentry;
	read->npages = npages;
	memcpy(read->pages, pages, npages * sizeof(struct nfs4_page));
	
	if ((status = nfs4_setup_compound(argp)))
		goto out_err;
	if ((status = nfs4_setup_putfh(argp, putfh)))
		goto out_args;
	if ((status = nfs4_setup_read(argp, read)))
		goto out_args;
	if ((status = nfs4_setup_reval(argp, reval)))
		goto out_args;
	task->tk_action = nfs4_async_call_compound;
	task->tk_exit = read_handler;
	nfsv4_printk(read, "pagein_one: %d-page read task scheduled\n", npages);
	rpc_execute(task);
	
out:
	return status;
out_args:
	nfs4_release_compoundargs(argp);
out_err:
	nfsv4_printk(read2, "pagein_one: error, failed to schedule task!\n");
	/* clean up */
	for (i = 0; i < npages; i++) {
		req = pages[i];
		SetPageError(req->wb_page);
		flush_dcache_page(req->wb_page);
		UnlockPage(req->wb_page);
		nfs4_unlock_request(req);
		if (!req->wb_count)
			BUG();
		nfs4_release_request(req);
	}
	rpc_release_task(task);
	goto out;
}

static inline int
coalesce_requests(struct list_head *src, struct nfs4_page **dst, unsigned int maxpages)
{
	struct nfs4_page	*req = NULL;
	unsigned int		pages = 0;

	/* We only put a single READ op into a request, in contrast to the write case
	   (where we put several, possibly discontiguous WRITES into a single request).
	   This is because of page-alignment problems if a pathological server decides
	   to return success for each READ, but only a subset of the bytes requested.
	   Therefore, we use the NFSv3 mechanism of coalescing adjacent writes into a
	   single request. */

	if (list_empty(src))
		BUG();

	while (!list_empty(src)) {
		struct nfs4_page  *prev = req;

		req = list_entry(src->next, struct nfs4_page, wb_list);

		if (!list_empty(&req->wb_writelist))
			BUG();
		if (!test_bit(PG_BUSY, &req->wb_flags))
			BUG();
		if (req->wb_offset || (req->wb_bytes != PAGE_CACHE_SIZE))
			BUG();
		
		if (prev) {
			if (req->wb_file != prev->wb_file || req->wb_fdata != prev->wb_fdata)
				break;
			if (req->wb_page->index != prev->wb_page->index + 1)
				break;
		}
		list_del_init(&req->wb_list);
		dst[pages++] = req;
		
		if (pages >= maxpages)
			break;
	}
	nfsv4_printk(read, "coalesced one %d-page read request\n", pages);
	return pages;
}

static int
nfs4_pagein_list(struct inode *inode, struct list_head *head)
{
	int			error = 0;
	unsigned int		npages;
	unsigned int		total_pages = 0;
	unsigned int		rpages = NFS4_RPAGES(inode->i_sb);
	struct nfs4_page	*pages[NFS4_MAXIO];

	/* XXX: make sure we don't need to acquire the spinlock here... */

	if (rpages > NFS4_MAXIO)
		BUG();
	
	while (!list_empty(head)) {
		npages = coalesce_requests(head, pages, rpages);
		total_pages += npages;
		error = nfs4_pagein_one(npages, pages);
		if (error < 0)
			break;
	}
	if (error >= 0)
		return total_pages;

	/* clean up */
	while (!list_empty(head)) {
		struct nfs4_page *req = list_entry(head->next, struct nfs4_page, wb_list);
		list_del_init(&req->wb_list);
		SetPageError(req->wb_page);
		flush_dcache_page(req->wb_page);    /* XXX: needed? */
		UnlockPage(req->wb_page);
		if (!req->wb_count)
			BUG();
		nfs4_release_request(req);
	}
	
	return error;
}

static inline int
nfs4_scan_read(struct inode *inode, struct list_head *dst,
	      unsigned long idx_start, unsigned int npages)
{
	int	res;
	spin_lock(&nfs4_wreq_lock);
	res = nfs4_scan_list(&NFS4_FLUSHD_READ(inode), dst, NULL, idx_start, npages, NULL);
	NFS4_FLUSHD_NREAD(inode) -= res;
	if (NFS4_FLUSHD_NREAD(inode) < 0)
		BUG();
	if ((NFS4_FLUSHD_NREAD(inode) == 0) != list_empty(&NFS4_FLUSHD_READ(inode)))
		BUG();
	spin_unlock(&nfs4_wreq_lock);
	return res;
}

static inline int
nfs4_scan_read_timeout(struct inode *inode, struct list_head *dst)
{
	int	pages;
	spin_lock(&nfs4_wreq_lock);
	pages = nfs4_scan_list_timeout(&NFS4_FLUSHD_READ(inode), dst, inode, NULL);
	NFS4_FLUSHD_NREAD(inode) -= pages;
	if (NFS4_FLUSHD_NREAD(inode) < 0)
		BUG();
	if ((NFS4_FLUSHD_NREAD(inode) == 0) != list_empty(&NFS4_FLUSHD_READ(inode)))
		BUG();
	spin_unlock(&nfs4_wreq_lock);
	return pages;
}

int
nfs4_pagein_inode(struct inode *inode, unsigned long idx_start, unsigned int npages)
{
	LIST_HEAD(head);
	int	res,
		error = 0;

	res = nfs4_scan_read(inode, &head, idx_start, npages);
	if (res)
		error = nfs4_pagein_list(inode, &head);
	if (error < 0)
		return error;
	return res;
}

int
nfs4_pagein_timeout(struct inode *inode)
{
	LIST_HEAD(head);
	int	pages,
		error = 0;

	pages = nfs4_scan_read_timeout(inode, &head);
	if (pages)
		error = nfs4_pagein_list(inode, &head);
	if (error < 0)
		return error;
	return pages;
}
