/*
 *  linux/fs/nfs4fs/renewd.c
 *
 *  Copyright (c) 2001 The Regents of the University of Michigan.
 *  All rights reserved.
 *
 *  Kendrick Smith <kmsmith@umich.edu>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/mm.h>
#include <linux/pagemap.h>

#include <linux/nfs4/nfs4_prot.h>
#include <linux/nfs4/nfs4_debug.h>
#include <linux/nfs4/nfs4.h>
#include <linux/nfs4/nfs4fs.h>

/*
 * Implementation of the NFSv4 "renew daemon", which wakes up periodically to
 * send a RENEW, to keep state alive on the server.  The daemon is implemented
 * as an rpc_task, not a real kernel thread, so it always runs in rpciod's
 * context.  There is one renewd per superblock.
 */

static void renewd_release(struct nfs4_async_compound *cp);
static void renewd(struct rpc_task *task);
static void renewd_handler(struct rpc_task *task);
static void renewd_set_timeout(struct nfs4_async_compound *cp, struct rpc_task *task);

static struct rpc_wait_queue   renewd_queue = RPC_INIT_WAITQ("nfs4_renewd");

/* Note: Before calling nfs4fs_init_renewd(), make sure that the lease period and
 * last renewal time are correctly set in the superblock!
 */
int
nfs4fs_init_renewd(struct super_block *sb)
{
	struct nfs4_async_compound *cp;
	struct rpc_task *task;
	int status;
	
	lock_kernel();	
	status = -ENOMEM;
	if ((cp = nfs4_async_create(sb))) {
		cp->ac_release = renewd_release;
		cp->ac_renew_sb = sb;
		cp->ac_renew_fp = NULL;
		NFS4_RENEWD(sb) = cp;
		
		/* XXX: Am I doing this right?  My understanding of RPC layer says that if
		 * we call rpc_execute() on an async task, it will run in the current context
		 * until it has to sleep, at which point it will end up on a queue somewhere
		 * and wake up in rpciod's context... (I have a similar question for the
		 * flushd startup.)
		 */
		task = &cp->ac_task;
		task->tk_action = renewd;
		nfsv4_printk(lease, "RENEWD: created!\n");
		rpc_execute(task);
		status = 0;
	}
	unlock_kernel();
	return status;
}

static void
renewd_release(struct nfs4_async_compound *cp)
{
	/* Probably unnecessary, but in the interests of programming defensively... */
	NFS4_RENEWD(cp->ac_renew_sb) = NULL;
	if (cp->ac_renew_fp) {
		nfs4_put_file_data(cp->ac_renew_fp);
		cp->ac_renew_fp = NULL;
	}
	nfsv4_printk(lease, "RENEWD: destroyed!\n");
}

static void
renewd(struct rpc_task *task)
{
	struct nfs4_async_compound *cp = (struct nfs4_async_compound *) task->tk_calldata;
	struct super_block *sb = cp->ac_renew_sb;
	struct list_head *l;
	struct nfs4fs_file_data *fp;
	struct compound4args *argp;

	/* NB: rpciod runs with the BKL held, so no need for lock_kernel()... */
	
#if NFS4_DEBUG
	if (task != &cp->ac_task) {
		printk("!!! nfs4: task mismatch in renewd !!!\n");
		rpc_exit(task, -EIO);
		return;
	}
	if (cp->ac_renew_fp) {
		printk("!!! nfs4: non-NULL file pointer in renewd !!!\n");
		rpc_exit(task, -EIO);
		return;
	}
#endif
	
	nfsv4_printk(lease2, "RENEWD: woke up\n");
	if (list_empty(&NFS4_FILE_LIST(sb))) {
		/*
		 * XXX: if we wanted to, could renew delegations here
		 * by calling nfs4_exists_delegation(), but is it worth
		 * the trouble?  The need for this will go away once
		 * RENEW is changed in the spec...
		 */
		task->tk_timeout = 2 * NFS4_LEASE_TIME(sb) / 3;
		task->tk_action = renewd;
		task->tk_exit = NULL;
		nfsv4_printk(lease2, "RENEWD: nothing to do, sleeping for %ld/%d seconds\n",
			     task->tk_timeout, HZ);
		rpc_sleep_on(&renewd_queue, task, NULL, NULL);
	}
	else if (jiffies > NFS4_LAST_RENEWAL(sb) + NFS4_LEASE_TIME(sb)/3) {
		nfsv4_printk(lease2, "RENEWD: making RENEW call\n");
		l = NFS4_FILE_LIST(sb).next;
		fp = list_entry(l, struct nfs4fs_file_data, fi_list);
		atomic_inc(&fp->fi_count);

		/* Set up the RENEW call... */
		argp = &cp->ac_arg;
		argp->sb = sb;
		argp->tag = "renewd";
		if (nfs4_setup_compound(argp))
			goto fail;
		if (nfs4_setup_renew(argp, fp))
			goto fail_args;
		cp->ac_renew_fp = fp;
		task->tk_action = nfs4_async_call_compound;
		task->tk_exit = renewd_handler;
	}
	else
		renewd_set_timeout(cp, task);
	nfsv4_printk(lease2, "RENEWD: back to sleep\n");
	return;

fail_args:
	nfs4_release_compoundargs(argp);
fail:
	nfs4_put_file_data(fp);
	task->tk_timeout = 10 * HZ;      /* hardcoded */
	task->tk_action = renewd;
	task->tk_exit = NULL;
	rpc_sleep_on(&renewd_queue, task, NULL, NULL);
	printk(KERN_WARNING "nfs4: failed to renew lease!\n");
}

static void
renewd_handler(struct rpc_task *task)
{
	struct nfs4_async_compound *cp = (struct nfs4_async_compound *) task->tk_calldata;
	struct compound4args *argp = &cp->ac_arg;
	struct compound4res *resp = &cp->ac_res;

	nfsv4_printk(lease, "RENEWD: in handler\n");
	nfs4_put_file_data(cp->ac_renew_fp);
	cp->ac_renew_fp = NULL;
	
	if (nfs4_async_prehandler(cp, NULL)) {
		nfsv4_printk(lease2, "RENEWD: async_prehandler() failed!\n");
		goto fail_args;
	}
	if (nfs4_handle_renew(resp)) {
		nfsv4_printk(lease2, "RENEWD: handle_renew() failed!\n");
		goto fail_res;
	}
	nfs4_release_compoundres(resp);
	nfs4_release_compoundargs(argp);
	xprt_release(task);
	renewd_set_timeout(cp, task);
	return;
	
fail_res:
	nfs4_release_compoundres(resp);
fail_args:
	nfs4_release_compoundargs(argp);
	xprt_release(task);
	task->tk_timeout = 10 * HZ;     /* hardcoded */
	task->tk_action = renewd;
	task->tk_exit = NULL;
	rpc_sleep_on(&renewd_queue, task, NULL, NULL);
	printk(KERN_WARNING "nfs4: failed to renew lease!\n");
}

static void
renewd_set_timeout(struct nfs4_async_compound *cp, struct rpc_task *task)
{
	struct super_block *sb = cp->ac_renew_sb;
	unsigned long timeout;
	
	timeout = (2 * NFS4_LEASE_TIME(sb))/3 + NFS4_LAST_RENEWAL(sb) - jiffies;

	/* XXX: This is a crude safeguard against various circumstances that can cause
	 * timeout==0: (1) A server which returns pathological values for its lease time
	 * can trigger roundoff error in the timeout calculation, (2) Consecutive failed
	 * RENEW's, (3) pathologically heavy client load.  Need to think about what the
	 * best algorithm really is here!
	 */
	if ((timeout < 1 * HZ) || (timeout > (2 * NFS4_LEASE_TIME(sb) / 3)))
		timeout = 1 * HZ;

	nfsv4_printk(lease, "RENEWD: will awaken in %ld/%d seconds\n", timeout, HZ);
	
	task->tk_timeout = timeout;
	task->tk_action = renewd;
	task->tk_exit = NULL;
	rpc_sleep_on(&renewd_queue, task, NULL, NULL);
}
