/*
 * linux/fs/nfs4fs/silly.c
 * from linux/fs/nfs/unlink.c
 * 
 * NFSv4 implementation
 *    Kendrick Smith  <kmsmith@umich.edu>
 *    Andy Adamson  <andros@umich.edu>
 *    Copyright (C) 2001 The Regents of the University of Michigan
 */

#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/mm.h>
#include <linux/pagemap.h>

#include <linux/nfs4/nfs4_prot.h>
#include <linux/nfs4/nfs4_debug.h>
#include <linux/nfs4/nfs4.h>
#include <linux/nfs4/nfs4fs.h>

#define  SILLY_CHEAT              0

static struct list_head		nfs_deletes = LIST_HEAD_INIT(nfs_deletes);
static struct rpc_wait_queue	nfs_delete_queue = RPC_INIT_WAITQ("nfs_delete_queue");

static void
log_failed_sillydelete(struct dentry *dentry)
{
	struct super_block *sb = dentry->d_sb;
	struct dentry *parents[10];
	int i;
	
	parents[0] = dentry;
	for (i = 1; i < 10; i++) {
		if (parents[i-1] == sb->s_root)
			break;
		parents[i] = parents[i-1]->d_parent;
	}

	/*
	 * XXX: Question: can printk() sleep?
	 */
	lock_kernel();
	printk(KERN_WARNING "nfs4: warning: failed to sillydelete file ");
	while (i >= 0) {
		printk("%.*s%s", (int)parents[i]->d_name.len,
		       parents[i]->d_name.name, i ? "/" : "!\n");
		i--;
	}
	unlock_kernel();
}

/*
 * XXX: Is ->tk_exit() guaranteed to be called?
 */
static void
silly_handler(struct rpc_task *task)
{
	struct nfs4_async_compound *ctx = (struct nfs4_async_compound *) task->tk_calldata;
	struct compound4res *resp = &ctx->ac_res;

	nfsv4_printk(remove, "silly_handler: starting\n");

	if (nfs4_async_prehandler(ctx, NULL)) {
		nfsv4_printk(remove2, "silly_handler: prehandler returned error!\n");
		goto out_fail;
	}
	
	if (nfs4_handle_putfh(resp, &ctx->ac_unlink_putfh)) {
		nfsv4_printk(level1, "silly_handler: handle_putfh failed!\n");
		goto out_res_fail;
	}
	if (nfs4_handle_remove(resp, ctx->ac_unlink_dir_inode)) {
		nfsv4_printk(level1, "silly_handler: REMOVE failed on server!\n");
		goto out_res_fail;
	}
	nfs4_release_compoundres(resp);
	
	nfsv4_printk(remove, "silly_handler: success\n");

out:
	iput(ctx->ac_unlink_dir_inode);
	ctx->ac_unlink_dir_inode = NULL;    /* paranoid.. */
	return;
out_res_fail:
	nfs4_release_compoundres(resp);
out_fail:
	/*
	 * XXX: It would be nice if we could do a log_failed_sillydelete()
	 * here, but we don't have the dentry anymore... (see comment in
	 * silly_setup_unlink()).  Maybe we could save the parent dentry
	 * and the name of the .nfs... file?
	 */
	printk(KERN_WARNING "nfs4: warning: couldn't sillydelete file!\n");
	goto out;
}

static int
silly_setup_unlink(struct dentry *dentry)
{
	struct nfs4_async_compound *ctx;
	struct compound4args *argp;
	int status;

	nfsv4_printk(remove, "silly_setup_unlink: starting\n");

	status = -ENOMEM;
	ctx = nfs4_async_create(dentry->d_sb);
	if (!ctx)
		goto out;

	argp = &ctx->ac_arg;
	argp->sb = dentry->d_sb;
	argp->tag = "silly_unlink";
	
	ctx->ac_unlink_putfh.dentry = dentry->d_parent;
	if ((status = nfs4_setup_compound(argp)))
		goto out_release;
	if ((status = nfs4_setup_putfh(argp, &ctx->ac_unlink_putfh)))
		goto out_args_release;
	if ((status = nfs4_setup_remove(argp, &dentry->d_name)))
		goto out_args_release;

	ctx->ac_unlink_dentry = dentry;
	ctx->ac_unlink_dir_inode = NULL;

	lock_kernel();   /* XXX: do we already have the BKL? */
	list_add(&ctx->ac_unlink_list, &nfs_deletes);
	unlock_kernel();

	dentry->d_flags |= DCACHE_NFSFS_RENAMED;
	rpc_sleep_on(&nfs_delete_queue, &ctx->ac_task, NULL, NULL);
	status = 0;
	
out:
	nfsv4_printk(remove, "silly_setup_unlink: returning status %d\n", status);
	return status;
out_args_release:
	nfs4_release_compoundargs(argp);
out_release:
	nfs4_async_release(&ctx->ac_task);
	goto out;
}

static void
silly_finish_unlink(struct dentry *dentry)
{
	struct list_head *l;
	struct nfs4_async_compound *ctx;

	/* called with BKL held */
	
	list_for_each(l, &nfs_deletes) {
		ctx = list_entry(l, struct nfs4_async_compound, ac_unlink_list);
		if (ctx->ac_unlink_dentry == dentry) {
			nfsv4_printk(remove, "silly_finish_unlink: waking up task\n");
			list_del_init(&ctx->ac_unlink_list);

			/*
			 * It seems safest to set ctx->ac_dentry to NULL,
			 * since we do not hold a reference, so there is
			 * no guarantee that it will still be around after
			 * dentry_iput() returns...
			 */
			ctx->ac_unlink_dentry = NULL;
			ctx->ac_unlink_dir_inode = igrab(dentry->d_parent->d_inode);
			ctx->ac_task.tk_action = nfs4_async_call_compound;
			ctx->ac_task.tk_exit = silly_handler;
			rpc_wake_up_task(&ctx->ac_task);
			return;
		}
	}

	/*
	 * If we reach this point, we failed to find the dentry in the list.
	 */
	nfsv4_printk(level1, "silly_finish_unlink: dentry not found?!\n");
	log_failed_sillydelete(dentry);
}

int
nfs4_sillyrename(struct inode *dir, struct dentry *dentry)
{
	static unsigned int sillycounter;
	const int      i_inosize  = sizeof(dir->i_ino)*2;
	const int      countersize = sizeof(sillycounter)*2;
	const int      slen       = strlen(".nfs") + i_inosize + countersize;
	char           silly[slen+1];
	struct dentry *sdentry;
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_putfh putfh;
	struct nfs4fs_rename rename;
	struct nfs4fs_getattr getattr;
	int status = 0;
	int counter = 0;

	/*
	 * XXX: Need to think carefully about race conditions on the sillyrename...
	 * For one thing, this routine can sleep, so even the BKL does not protect
	 * us from another process concurrently sillyrenaming the file.
	 */
	if (dentry->d_flags & DCACHE_NFSFS_RENAMED)
		goto out;

#if NFS4_DEBUG
	status = -EIO;
	if (!dentry->d_inode) {
		printk("!!! nfs_sillyrename: negative dentry !!!\n");
		goto out;
	}
	if (atomic_read(&dentry->d_count) == 1) {
		printk("!!! nfs_sillyrename: d_count==1 !!!\n");
		goto out;
	}
	if (dentry->d_flags & DCACHE_NFSFS_RENAMED) {
		nfsv4_printk(remove, "!!! nfs_sillyrename: already silly-renamed !!!\n");
		goto out;
	}
#endif

	nfsv4_printk(remove, "nfs_sillyrename: starting\n");

	sprintf(silly, ".nfs%*.*lx", i_inosize, i_inosize, dentry->d_inode->i_ino);
	sdentry = NULL;
	do {
		char *suffix = silly + slen - countersize;
		if (++counter > 1000)
			BUG();
		dput(sdentry);
		sillycounter++;
		sprintf(suffix, "%*.*x", countersize, countersize, sillycounter);
		sdentry = lookup_one(silly, dentry->d_parent);
		status = PTR_ERR(sdentry);
		if (IS_ERR(sdentry)) {
			nfsv4_printk(remove2, "nfs_sillyrename: lookup_one returned error!\n");
			goto out;
		}
	} while (sdentry->d_inode != NULL); /* need negative lookup */

	nfsv4_printk(remove, "nfs_silly_rename: renaming \"%.*s\" -> \"%.*s\"\n",
		     (int)dentry->d_name.len, dentry->d_name.name,
		     (int)sdentry->d_name.len, sdentry->d_name.name);
	
	cargs.sb = dir->i_sb;
	cargs.tag = "nfs_silly_rename()";
	putfh.dentry = dentry->d_parent;
	rename.old_dentry = dentry;
	rename.new_dentry = sdentry;
	getattr.flags = FL_NONFATAL;
	getattr.bm = &nfs4_negative_bitmap;
	if ((status = nfs4_setup_compound(&cargs)))
		goto out;
	if ((status = nfs4_setup_putfh(&cargs, &putfh)))
		goto out_args;
	if ((status = nfs4_setup_savefh(&cargs)))
		goto out_args;
	if ((status = nfs4_setup_rename(&cargs, &rename)))
		goto out_args;
	if ((status = nfs4_setup_getattr(&cargs, &getattr)))
		goto out_args;
	if ((status = nfs4_call_compound(&cargs, &cres, NULL)))
		goto out_args;
	if ((status = nfs4_handle_putfh(&cres, &putfh)))
		goto out_res;
	if ((status = nfs4_handle_savefh(&cres)))
		goto out_res;
	if ((status = nfs4_handle_rename(&cres, &rename))) {
		goto out_res;
	}
	nfs4_handle_getattr(&cres, &getattr);   /* ignore return value */
	d_move(dentry, sdentry);
	d_drop(dentry);    /* so that dput() will get rid of it via ->d_iput() */
	dput(sdentry);
	dentry->d_time = jiffies;

#if SILLY_CHEAT
	dentry->d_flags |= DCACHE_NFSFS_RENAMED;
#else
	/*
	 *  Set up asynchronous RPC to unlink the file.
	 */
	status = silly_setup_unlink(dentry);
#endif

out_res:
	nfs4_release_compoundres(&cres);
out_args:
	nfs4_release_compoundargs(&cargs);
out:
	nfsv4_printk(remove, "nfs_sillyrename: returning status %d\n", status);
	return status;
}

/*
 * Called when the dentry loses inode.
 * We use it to clean up silly-renamed files.
 *
 * NB: The default ->d_iput() method is just iput().
 * In other words, this routine is responsible for decreasing
 * the inode's refcount, in addition to anything else it does.
 */
void
nfs4_silly_dentry_iput(struct dentry *dentry, struct inode *inode)
{
	nfsv4_printk(remove, "nfs4_silly_dentry_iput: starting\n");

	if (dentry->d_flags & DCACHE_NFSFS_RENAMED) {
		lock_kernel();
#if SILLY_CHEAT
		nfsv4_printk(remove, "nfs4_silly_dentry_iput: calling nfs4_remove\n");
		nfs4_remove(dentry->d_parent->d_inode, dentry);
#else
		nfsv4_printk(remove, "nfs4_silly_dentry_iput: calling finish_unlink\n");
		silly_finish_unlink(dentry);
#endif
		unlock_kernel();
	}
	iput(inode);
	nfsv4_printk(remove, "nfs4_silly_dentry_iput: done\n");
}
