/*
 * linux/fs/nfs4fs/symlink.c
 * from linux/fs/nfs/symlink.c
 *    Copyright (C) 1992  Rick Sladkey
 *    Optimization changes Copyright (C) 1994 Florian La Roche
 *
 * NFSv4 implementation
 *    Kendrick Smith  <kmsmith@umich.edu>
 *    Andy Adamson  <andros@umich.edu>
 *    Copyright (C) 2001 The Regents of the University of Michigan
 *
 */
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/mm.h>
#include <linux/pagemap.h>
#include <linux/nfs4/nfs4_prot.h>
#include <linux/nfs4/nfs4_debug.h>
#include <linux/nfs4/nfs4.h>
#include <linux/nfs4/nfs4fs.h>

static int nfs4_readlink(struct dentry *dentry, char *buffer, int buflen);
static int nfs4_follow_link(struct dentry *dentry, struct nameidata *nd);
	
struct inode_operations nfs4_symlink_inode_operations = {
	readlink:	nfs4_readlink,
	follow_link:	nfs4_follow_link,
	revalidate:	nfs4_inode_revalidate,
	setattr:	nfs4_setattr,
#ifdef NFS4ACL
        permission: nfs4_posix_acl_permission,
        get_posix_acl: nfs4_acl_get_inode,
        set_posix_acl: nfs4_acl_set_inode,
        get_ext_attr: nfs4_get_ext_attr,         
        set_ext_attr: nfs4_set_ext_attr
#endif
};

static int
nfs4_getlink(struct dentry *dentry, struct page **ppage, char **bufp)
{
	struct inode *inode = dentry->d_inode;
	struct page *page;
	char *buf;
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_putfh putfh;
	struct nfs4fs_getattr reval;
	struct nfs4fs_readlink readlink;
	int status;

	lock_kernel();
	
	/* Caller revalidated the directory inode already. */
	page = grab_cache_page(inode->i_mapping, 0);
	nfsv4_printk(readlink, "nfs4_getlink: starting, page=%p\n", page);
	status = -ENOMEM;
	if (!page)
		goto out_buf;
	buf = kmap(page);
	status = 0;
	if (Page_Uptodate(page))
		goto out;
	nfsv4_printk(readlink, "nfs4_getlink: buf=%p, locked=%d, uptodate=%d, count=%d\n",
		     buf, (int)PageLocked(page), (int)Page_Uptodate(page),
		     (int)atomic_read(&page->count));

	cargs.sb = dentry->d_sb;
	cargs.tag = "nfs4_getlink()";
	putfh.dentry = dentry;
	readlink.buf = buf;
	readlink.nbytes = PAGE_SIZE - 1;   /* need one byte for null-terminator */
	
	if ((status = nfs4_setup_compound(&cargs)))
		goto out_page;
	if ((status = nfs4_setup_putfh(&cargs, &putfh)))
		goto out_args;
	if ((status = nfs4_setup_reval(&cargs, &reval)))
		goto out_args;
	if ((status = nfs4_setup_readlink(&cargs)))
		goto out_args;

	if ((status = nfs4_call_compound(&cargs, &cres, NULL)))
		goto out_args;
	if ((status = nfs4_handle_putfh(&cres, &putfh)))
		goto out_res;
	if ((status = nfs4_handle_reval(&cres, &reval)))
		goto out_res;
	if ((status = nfs4_handle_readlink(&cres, &readlink)))
		goto out_res;

	nfs4_release_compoundres(&cres);
	nfs4_release_compoundargs(&cargs);
	SetPageUptodate(page);
	
out:
	unlock_kernel();
	*ppage = page;
	*bufp = buf;
	nfsv4_printk(readlink, "nfs4_getlink: returning status %d\n", status);
	return status;
out_res:
	nfs4_release_compoundres(&cres);
out_args:
	nfs4_release_compoundargs(&cargs);
out_page:
	kunmap(page);
	UnlockPage(page);
	page_cache_release(page);
	page = NULL;
out_buf:
	buf = NULL;
	goto out;
}

static int
nfs4_readlink(struct dentry *dentry, char *buf, int buflen)
{
	struct page *page;
	char *link;
	int status;
	
	/* By calling vfs_readlink(), we compute strlen() for a string whose length we
	   already know. probably not worth optimizing out, though.. */

	nfsv4_printk(level2, "nfs4_readlink(%.*s/%.*s): starting\n",
		     (int)dentry->d_parent->d_name.len, dentry->d_parent->d_name.name,
		     (int)dentry->d_name.len, dentry->d_name.name);
	
	if ((status = nfs4_getlink(dentry, &page, &link)))
		goto out;
	status = vfs_readlink(dentry, buf, buflen, link);
	kunmap(page);
	UnlockPage(page);
	page_cache_release(page);

out:
	nfsv4_printk(level2, "nfs4_readlink: returning status %d\n\n", status);
	return status;
}

static int
nfs4_follow_link(struct dentry *dentry, struct nameidata *nd)
{
	struct page *page;
	char *buf;
	int status;

	nfsv4_printk(level2, "nfs4_follow_link(%.*s/%.*s): starting\n",
		     (int)dentry->d_parent->d_name.len, dentry->d_parent->d_name.name,
		     (int)dentry->d_name.len, dentry->d_name.name);
	
	if ((status = nfs4_getlink(dentry, &page, &buf)))
		goto out_release;
	status = vfs_follow_link(nd, buf);
	kunmap(page);
	UnlockPage(page);
	page_cache_release(page);

out:
	nfsv4_printk(level2, "nfs4_follow_link: returning status %d\n\n", status);
	return status;
out_release:
	path_release(nd);
	goto out;
}
