/*
 * linux/fs/nfs4fs/write.c
 * from linux/nfs/nfs/write.c
 *    Copyright (C) 1996, 1997, Olaf Kirch <okir@monad.swb.de>
 *
 * NFSv4 implementation
 *    Kendrick Smith  <kmsmith@umich.edu>
 *    Andy Adamson  <andros@umich.edu>
 *    Copyright (C) 2001 The Regents of the University of Michigan
 *
 * XXX: audit this file carefully for SMP races; there may be some remaining!
 * XXX: also audit for wb refcounts...
 * XXX: also think about "swap-out" comment at top of fs/nfs/write.c
 */
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <asm/uaccess.h>
#include <linux/pagemap.h>
#include <asm/uaccess.h>

#include <linux/nfs4/nfs4_prot.h>
#include <linux/nfs4/nfs4_debug.h>
#include <linux/nfs4/nfs4.h>
#include <linux/nfs4/nfs4fs.h>

static struct nfs4_page *nfs4_update_request(struct file *file, struct nfs4fs_file_data *fdata,
					     struct page *page, unsigned int offset,
					     unsigned int count);
static void nfs4_strategy(struct inode *inode);
static int nfs4_flush_file(struct inode *inode, struct file *file,
			   unsigned long idx_start, unsigned int npages, int how);
static int nfs4_commit_file(struct inode *inode, int how);

/*
 * Write a page synchronously.
 * Offset is the data offset within the page.
 */
static int
nfs4_writepage_sync(struct nfs4fs_file_data *fdata, struct dentry *dentry, struct page *page,
		    unsigned int offset, unsigned int count)
{
	struct inode *inode = dentry->d_inode;
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_putfh putfh;
	struct nfs4fs_simple_write write;
	int wsize;
	int total_bytes_written = 0;
	int status;

	/* XXX: do we want to revalidate? */

	wsize = NFS4_WSIZE(inode);
	
	nfsv4_printk(write, "nfs4_writepage_sync: file=\"%.*s\", index=%d, start=%d, "
		     "count=%d", (int)dentry->d_name.len, dentry->d_name.name,
		     (int)page->index, offset, count);

	/* XXX: I'm not sure whether we need the kmap() here.. the code path leading here from
	   write() already has the page kmap()'ed.  what about the code path from msync()? */
	
	cargs.sb = dentry->d_sb;
	cargs.tag = "nfs4_writepage_sync()";
	putfh.dentry = dentry;
	write.file_data = fdata;
	write.start = ((loff_t)page->index << PAGE_CACHE_SHIFT) + offset;
	write.buf = kmap(page) + offset;
	
	do {
		write.length = (count < wsize) ? count : wsize;
		nfsv4_printk(write, "nfs4_write: attempting to write %d bytes\n",
			     write.length);
		
		if ((status = nfs4_setup_compound(&cargs)))
			goto out_error;
		if ((status = nfs4_setup_putfh(&cargs, &putfh)))
			goto out_args;
		if ((status = nfs4_setup_simple_write(&cargs, &write)))
			goto out_args;

		if ((status = nfs4_call_compound(&cargs, &cres, NULL)))
			goto out_args;
		if ((status = nfs4_handle_putfh(&cres, &putfh)))
			goto out_res;
		if ((status = nfs4_handle_simple_write(&cres, &write)))
			goto out_res;

		nfsv4_printk(write, "nfs4_write: wrote %d bytes\n", write.bytes_written);
		
		nfs4_release_compoundres(&cres);
		nfs4_release_compoundargs(&cargs);
		
		if (!write.bytes_written) {
			status = -EIO;
			goto out_error;
		}
		total_bytes_written += write.bytes_written;
		write.start += write.bytes_written;
		write.buf += write.bytes_written;
		count -= write.bytes_written;

		/*
		 * If we've extended the file, update the inode
		 * now so we don't invalidate the cache.
		 */
		if (write.start > inode->i_size)
			inode->i_size = write.start;
	} while (count);

	if (PageError(page))
		ClearPageError(page);

out_unmap:
	kunmap(page);
	nfsv4_printk(write, "nfs4_writepage_sync: returning %d\n",
		     total_bytes_written ?: status);
	return total_bytes_written ?: status;
out_res:
	nfs4_release_compoundres(&cres);
out_args:
	nfs4_release_compoundargs(&cargs);
out_error:
	ClearPageUptodate(page);
	goto out_unmap;
}

static int
nfs4_writepage_async(struct nfs4fs_file_data *fdata, struct dentry *dentry, struct page *page,
		     unsigned int offset, unsigned int count)
{
	struct nfs4_page *req;
	int status;

	if (!dentry->d_inode)
		BUG();
	nfsv4_printk(write, "writepage_async() starting, calling nfs4_update_request()\n");
	req = nfs4_update_request(NULL, fdata, page, offset, count);
	status = IS_ERR(req) ? PTR_ERR(req) : 0;
	if (status < 0)
		goto out;
	if (!req->wb_count)
		BUG();
	nfs4_release_request(req);
	nfsv4_printk(write, "writepage_async() calling nfs4_strategy()\n");
	nfs4_strategy(dentry->d_inode);
	status = nfs4_sync_file(dentry->d_inode, NULL, 0, 0, FLUSH_WAIT | FLUSH_SYNC);

out:
	return status;
}

int
nfs4_writepage(struct page *page)
{
	struct inode *inode;
	struct nfs4fs_file_data *fdata;
	unsigned long end_index;
	unsigned offset = PAGE_CACHE_SIZE;
	int err;
	struct address_space *mapping = page->mapping;

	nfsv4_printk(write, "nfs4_writepage: starting\n");
	
	if (!mapping)
		BUG();
	inode = (struct inode *)mapping->host;
	if (!inode)
		BUG();
	
	lock_kernel();   /* XXX: granulate */
	if (list_empty(&NFS4_WRITE_MAPPINGS(inode)))
		goto out_nomap;
	fdata = list_entry(NFS4_WRITE_MAPPINGS(inode).next, struct nfs4fs_file_data, fi_write_map);
	atomic_inc(&fdata->fi_count);
	unlock_kernel();
	
	if (!fdata->fi_dentry || fdata->fi_inode != inode)
		BUG();
	end_index = inode->i_size >> PAGE_CACHE_SHIFT;
	
	/* Ensure we've flushed out any previous writes */
	nfs4_sync_file(inode, 0, page->index, 1, FLUSH_WAIT | FLUSH_STABLE);
	
	/* easy case */
	if (page->index < end_index)
		goto do_it;
	/* things got complicated... */
	offset = inode->i_size & (PAGE_CACHE_SIZE-1);

	/* OK, are we completely out? */
	err = -EIO;
	if ((page->index > end_index) || !offset)
		goto out;
	
do_it:
	lock_kernel();
	if (NFS4_WSIZE(inode) >= PAGE_CACHE_SIZE) {
		err = nfs4_writepage_async(fdata, fdata->fi_dentry, page, 0, offset);
		if (err >= 0)
			err = 0;
	}
	else {
		err = nfs4_writepage_sync(fdata, fdata->fi_dentry, page, 0, offset);
		if (err == offset)
			err = 0;
	}
	unlock_kernel();

	nfs4_put_file_data(fdata);
	
out:
	UnlockPage(page);
	return err;
out_nomap:
	/* XXX: nfs4_zap_caches? */
	nfsv4_printk(level1, "nfs4: warning: stranded dirty page!\n");
	err = -EIO;
	goto out;
}

static int
region_locked(struct inode *inode, struct nfs4_page *req)
{
	struct file_lock	*fl;
	loff_t			rqstart, rqend;

	rqstart = ((loff_t)req->wb_page->index << PAGE_CACHE_SHIFT) + req->wb_offset;
	rqend = rqstart + req->wb_bytes;
	for (fl = inode->i_flock; fl; fl = fl->fl_next) {
		if (fl->fl_owner == current->files && (fl->fl_flags & FL_POSIX)
		    && fl->fl_type == F_WRLCK
		    && fl->fl_start <= rqstart && rqend <= fl->fl_end) {
			return 1;
		}
	}

	return 0;
}

/*
 * nfs4_inode_{add,remove}_request(): adds an nfs4_page to an inode's writelist
 *
 * Warning! Callers of nfs4_inode_add_request() must hold the spinlock.
 * Callers of nfs4_inode_remove_request() must NOT hold the spinlock.
 * (Confusing, but this is what nfsv3 does...)
 */
static inline void
nfs4_inode_add_request(struct inode *inode, struct nfs4_page *req)
{
	if (!list_empty(&req->wb_list))
		BUG();
	if (!list_empty(&req->wb_writelist))
		BUG();
	if (!test_bit(PG_BUSY, &req->wb_flags))
		BUG();
	if (req->wb_inode != inode)
		BUG();

	/* Note: We don't need to increment the inode's refcount, since a reference is
	   held through req->wb_fdata->fi_inode. */
	req->wb_count++;
	list_add(&req->wb_writelist, &NFS4_FLUSHD_WRITE(inode));
}

static inline void
nfs4_inode_remove_request(struct nfs4_page *req)
{
	spin_lock(&nfs4_wreq_lock);
	if (list_empty(&req->wb_writelist))
		BUG();
	if (!list_empty(&req->wb_list))
		BUG();
	if (!test_bit(PG_BUSY, &req->wb_flags))
		BUG();
	list_del_init(&req->wb_writelist);
	spin_unlock(&nfs4_wreq_lock);

	if (!req->wb_count)
		BUG();
	nfs4_release_request(req);
}

static inline struct nfs4_page *
nfs4_find_write(struct inode *inode, struct page *page)
{
	struct list_head *head, *next;
	struct nfs4_page *req;

	/* caller must hold spinlock */
	
	head = &NFS4_FLUSHD_WRITE(inode);
	next = head->next;
	while (next != head) {
		req = list_entry(next, struct nfs4_page, wb_writelist);
		next = next->next;
		if (req->wb_page->index != page->index)
			continue;
		/* XXX: Do we need to test to make sure that (req->wb_page == page)?
		   In NFSv3, this condition is tested on read paths but not write paths..
		   why is this?  Note also: It _is_ tested in flush_incompatible()! */
		req->wb_count++;
		goto out;
	}
	req = NULL;

out:
	return req;
}

static inline void
nfs4_mark_request_dirty(struct nfs4_page *req)
{
	struct inode *inode = req->wb_inode;

	if (!list_empty(&req->wb_list))
		BUG();

	spin_lock(&nfs4_wreq_lock);
	nfs4_list_add_request(req, &NFS4_FLUSHD_DIRTY(inode));
	NFS4_FLUSHD_NDIRTY(inode)++;
	NFS4_FLUSHD_DIRTYBYTES(inode) += req->wb_bytes;
	/*
	 * NB: The call to schedule_inode() can lie inside the spinlock,
	 * since it cannot block in NFSv4...
	 */
	nfs4_schedule_inode(inode, req->wb_timeout);
	spin_unlock(&nfs4_wreq_lock);
}

static inline void
nfs4_mark_list_dirty(struct list_head *head)
{
	struct nfs4_page *req;

	/* assumption: all pages on the list are locked by caller */
	
	while (!list_empty(head)) {
		req = list_entry(head->next, struct nfs4_page, wb_list);
		list_del_init(&req->wb_list);
		nfs4_mark_request_dirty(req);
		nfs4_unlock_request(req);
	}
}

static inline void
nfs4_mark_vector_dirty(unsigned int npages, struct nfs4_page **pages)
{
	int i;
	struct nfs4_page *req;

	/* assumption: all pages on the list are locked by caller */
	
	if (!npages || npages > NFS4_MAXIO)
		BUG();
	for (i = 0; i < npages; i++) {
		req = pages[i];
		nfs4_mark_request_dirty(req);
		nfs4_unlock_request(req);
	}
}

static inline void
nfs4_mark_request_commit(struct nfs4_page *req)
{
	struct inode *inode = req->wb_inode;

	/* call with spinlock held */
	
	if (!list_empty(&req->wb_list))
		BUG();

	spin_lock(&nfs4_wreq_lock);
	nfs4_list_add_request(req, &NFS4_FLUSHD_COMMIT(inode));
	NFS4_FLUSHD_NCOMMIT(inode)++;
	/*
	 * NB: The call to schedule_inode() can lie inside the spinlock,
	 * since it cannot block in NFSv4...
	 */
	nfs4_schedule_inode(inode, req->wb_timeout);
	spin_unlock(&nfs4_wreq_lock);
}

static inline void
nfs4_mark_list_commit(struct list_head *head)
{
	struct nfs4_page *req;

	/* assumption: all pages on the list are locked by caller */
	
	if (list_empty(head))
		BUG();

	while (!list_empty(head)) {
		req = list_entry(head->next, struct nfs4_page, wb_list);
		list_del_init(&req->wb_list);
		nfs4_mark_request_commit(req);
		nfs4_unlock_request(req);
	}
}

static inline int
nfs4_wait_on_request(struct nfs4_page *req)
{
	/* XXX: This was taken from NFSv3, but is it really SMP-safe?
	   Will it still work if another CPU clears the bit between the
	   initial test_bit() and the subsequent wait_event()? */
	
	if (!test_bit(PG_BUSY, &req->wb_flags))
		return 0;
	else {
		struct inode *inode = req->wb_inode;
		struct rpc_clnt *clnt = NFS4_CLNT(inode->i_sb);
		
		if (clnt->cl_intr) {
			int retval;
			sigset_t oldmask;
			rpc_clnt_sigmask(clnt, &oldmask);
			retval = wait_event_interruptible(req->wb_wait,
							  !test_bit(PG_BUSY, &req->wb_flags));
			rpc_clnt_sigunmask(clnt, &oldmask);
			return retval;
		}
		wait_event(req->wb_wait, !test_bit(PG_BUSY, &req->wb_flags));
		return 0;
	}
}

static int
nfs4_wait_on_requests(struct inode *inode, struct file *file,
		      unsigned long idx_start, unsigned int npages)
{
	struct list_head *p, *head;
	unsigned long idx_end;
	int res = 0;
	int error;

	if (npages == 0)
		idx_end = ~0;
	else
		idx_end = idx_start + npages - 1;

	spin_lock(&nfs4_wreq_lock);
	head = &NFS4_FLUSHD_WRITE(inode);
	p = head->next;
	while (p != head) {
		unsigned long pg_idx;
		struct nfs4_page *req = list_entry(p, struct nfs4_page, wb_writelist);
		
		p = p->next;
		if (file && req->wb_file != file)
			continue;
		pg_idx = req->wb_page->index;
		if (pg_idx < idx_start || pg_idx > idx_end)
			continue;
		if (!test_bit(PG_BUSY, &req->wb_flags))
			continue;

		req->wb_count++;
		spin_unlock(&nfs4_wreq_lock);
		error = nfs4_wait_on_request(req);

		if (!req->wb_count)
			BUG();
		nfs4_release_request(req);
		if (error < 0)
			return error;
		spin_lock(&nfs4_wreq_lock);
		p = head->next;
		res++;
	}
	spin_unlock(&nfs4_wreq_lock);
	return res;
}

static inline int
nfs4_scan_dirty(struct inode *inode, struct list_head *dst, struct file *file,
		unsigned long idx_start, unsigned int npages)
{
	int pages;
	
	spin_lock(&nfs4_wreq_lock);
	pages = nfs4_scan_list(&NFS4_FLUSHD_DIRTY(inode), dst, file, idx_start, npages,
			       &NFS4_FLUSHD_DIRTYBYTES(inode));
	NFS4_FLUSHD_NDIRTY(inode) -= pages;
	if ((NFS4_FLUSHD_NDIRTY(inode) == 0) != list_empty(&NFS4_FLUSHD_DIRTY(inode)))
		BUG();
	spin_unlock(&nfs4_wreq_lock);
	return pages;
}

static inline int
nfs4_scan_commit(struct inode *inode, struct list_head *dst, struct file *file,
		 unsigned long idx_start, unsigned int npages)
{
	int pages;
	
	spin_lock(&nfs4_wreq_lock);
	pages = nfs4_scan_list(&NFS4_FLUSHD_COMMIT(inode), dst, file, idx_start, npages,
			       &NFS4_FLUSHD_DIRTYBYTES(inode));
	NFS4_FLUSHD_NCOMMIT(inode) -= pages;
	if ((NFS4_FLUSHD_NCOMMIT(inode) == 0) != list_empty(&NFS4_FLUSHD_COMMIT(inode)))
		BUG();
	spin_unlock(&nfs4_wreq_lock);
	return pages;
}

static inline int
nfs4_scan_dirty_timeout(struct inode *inode, struct list_head *dst)
{
	int pages;

	spin_lock(&nfs4_wreq_lock);
	pages = nfs4_scan_list_timeout(&NFS4_FLUSHD_DIRTY(inode), dst, inode, NULL);
	NFS4_FLUSHD_NDIRTY(inode) -= pages;
	if ((NFS4_FLUSHD_NDIRTY(inode) == 0) != list_empty(&NFS4_FLUSHD_DIRTY(inode)))
		BUG();
	spin_unlock(&nfs4_wreq_lock);
	return pages;
}

static inline int
nfs4_coalesce_requests(struct list_head *src, struct nfs4_page **dst,
		       unsigned int wsize, unsigned int wpages)
{
	struct list_head *p, *q;
	struct nfs4_page *req;
	int pages = 0;
	
	if (list_empty(src))
		BUG();
	if (wsize < PAGE_CACHE_SIZE || !wpages || wpages > NFS4_MAXIO)
		BUG();

	for (p = src->next; p != src; p = q) {
		q = p->next;
		req = list_entry(p, struct nfs4_page, wb_list);
		if (!test_bit(PG_BUSY, &req->wb_flags))
			BUG();
		if (pages) {
			if (req->wb_file != dst[0]->wb_file ||
			    req->wb_fdata != dst[0]->wb_fdata)
				continue;
		}
		if (req->wb_bytes > wsize)
			break;
		list_del_init(p);
		dst[pages++] = req;
		if (pages == wpages)
			break;
		wsize -= req->wb_bytes;
	}

	return pages;
}

static struct nfs4_page *
nfs4_update_request(struct file *file, struct nfs4fs_file_data *fdata,
		    struct page *page, unsigned int offset, unsigned int count)
{
	struct inode *inode = fdata->fi_inode;
	struct nfs4_page	*req, *new = NULL;
	unsigned long		rqend, end;

	/* Note: must be called with the page lock held */

	end = offset + count;

	for (;;) {
		
		nfsv4_printk(write, "update_request() searching for page\n");
		spin_lock(&nfs4_wreq_lock);
		
		req = nfs4_find_write(inode, page);
		if (req) {

			if (!nfs4_lock_request(req)) {
				/* XXX: SMP race here? */
				spin_unlock(&nfs4_wreq_lock);
				nfsv4_printk(write, "update_request(): locked, waiting\n");
				nfs4_wait_on_request(req);
				if (!req->wb_count)
					BUG();
				nfs4_release_request(req);
				continue;
			}
			
			rqend = req->wb_offset + req->wb_bytes;
			if ((req->wb_file == file) && (req->wb_fdata == fdata) &&
			    (req->wb_offset <= end) && (rqend >= offset)) {
				/* We can just update the existing request. */
				NFS4_FLUSHD_DIRTYBYTES(inode) -= req->wb_bytes;
				if (NFS4_FLUSHD_DIRTYBYTES(inode) < 0)
					BUG();
				if (offset < req->wb_offset)
					req->wb_offset = offset;
				if (end > rqend)
					rqend = end;
				req->wb_bytes = rqend - req->wb_offset;
				NFS4_FLUSHD_DIRTYBYTES(inode) += req->wb_bytes;
				spin_unlock(&nfs4_wreq_lock);
				nfs4_unlock_request(req);
				nfsv4_printk(write, "update_request() used existing page\n");
				break;
			}

			/* The existing request conflicts with ours, so we return
			   -EBUSY and let the caller sort things out.  XXX: This is
			   the NFSv3 behavior, but wouldn't it be better to release
			   the spinlock, call nfs4_wb_page(), and retry the loop? */

			spin_unlock(&nfs4_wreq_lock);
			nfs4_unlock_request(req);
			nfsv4_printk(write, "update_request() conflict, returning -EBUSY\n");
			if (!req->wb_count)
				BUG();
			nfs4_release_request(req);
			return ERR_PTR(-EBUSY);
		}

		/* No matching request was found... */
		
		req = new;
		if (req) {
			nfs4_lock_request(req);
			nfs4_inode_add_request(inode, req);
			req->wb_count++;
			spin_unlock(&nfs4_wreq_lock);
			nfs4_mark_request_dirty(req);
			nfs4_unlock_request(req);
			nfsv4_printk(write, "update_request() inserting new page\n");
			new = NULL;
			break;
		}

		/* Create a new request... */
		
		spin_unlock(&nfs4_wreq_lock);

		/* XXX: I removed the logic to wake up all requests if over the soft
		   limit, because I was worried about SMP race conditions that might
		   result from dropping the spinlock.  However, I don't think that
		   this logic was necessary anyway, since nfs4_create_request() should
		   take care of this adequately. */

		nfsv4_printk(write, "update_request() allocating new request\n");

		if (file && file->private_data != fdata)
			BUG();
		
		new = nfs4_create_request(file, fdata, page, offset, count);
		if (!new)
			return ERR_PTR(-ENOMEM);

		/* If the region is locked, adjust the timeout */
		/* DELEG: This is where the delegation logic will go.. */
		if (region_locked(inode, new))
			new->wb_timeout = jiffies + NFS4_WRITEBACK_LOCKDELAY;
		else
			new->wb_timeout = jiffies + NFS4_WRITEBACK_DELAY;
	}
	
	if (new) {
		if (!new->wb_count)
			BUG();
		nfs4_release_request(new);
	}
	return req;
}

static void
nfs4_strategy(struct inode *inode)
{
	unsigned int ndirty, nbytes, ncommit;
	
	/* XXX: It seems to me that the spinlock is necessary here (it isn't in v3) */
	spin_lock(&nfs4_wreq_lock);
	ndirty = NFS4_FLUSHD_NDIRTY(inode);
	ncommit = NFS4_FLUSHD_NCOMMIT(inode);
	nbytes = NFS4_FLUSHD_DIRTYBYTES(inode);
	spin_unlock(&nfs4_wreq_lock);

	if (ndirty + ncommit > NFS4_MAX_REQUEST_SOFT) {
		nfsv4_printk(write, "nfs4_strategy(): soft limit hit, flushing whole thing\n");
		nfs4_sync_file(inode, 0, 0, 0, FLUSH_WAIT);
	}
	else {
		if ((ndirty > NFS4_MAXIO) || (nbytes > NFS4_WSIZE(inode))) {
			nfsv4_printk(write, "nfs4_strategy: flushing\n");
			nfs4_flush_file(inode, NULL, 0, 0, 0);
		}
		if (ncommit > NFS4_MAXCOMMIT) {
			nfsv4_printk(write, "nfs4_strategy: committing\n");
			nfs4_commit_file(inode, 0);
		}
	}
}

int
nfs4_flush_incompatible(struct file *file, struct page *page)
{
	struct nfs4_page *req;
	struct inode *inode;

	if (!file)
		BUG();
	if (!file->f_dentry)
		BUG();
	if (!file->f_dentry->d_inode)
		BUG();
	inode = file->f_dentry->d_inode;
	
	spin_lock(&nfs4_wreq_lock);
	req = nfs4_find_write(inode, page);
	if (req) {
		if (req->wb_file == file && req->wb_page == page) {
			spin_unlock(&nfs4_wreq_lock);
			if (!req->wb_count)
				BUG();
			nfs4_release_request(req);
			return 0;
		}
		spin_unlock(&nfs4_wreq_lock);
		if (!req->wb_count)
			BUG();
		nfs4_release_request(req);		
		return nfs4_sync_file(inode, file, page->index, 1, FLUSH_WAIT | FLUSH_STABLE);
	}
	spin_unlock(&nfs4_wreq_lock);
	return 0;
}

int
nfs4_updatepage(struct file *file, struct page *page, unsigned int offset, unsigned int count)
{
	struct nfs4fs_file_data *fdata = (struct nfs4fs_file_data *)file->private_data;
	struct dentry *dentry = file->f_dentry;
	struct inode *inode = dentry->d_inode;
	struct nfs4_page *req;
	int status;

	nfsv4_printk(write, "updatepage(): starting\n");
	if (!fdata)
		BUG();
	
	if (NFS4_WSIZE(inode) < PAGE_CACHE_SIZE)
		return nfs4_writepage_sync(fdata, dentry, page, offset, count);

	for (;;) {
		req = nfs4_update_request(file, fdata, page, offset, count);
		if (!req)
			BUG();
		if (!IS_ERR(req))
			break;
		status = PTR_ERR(req);
		if (status != -EBUSY)
			goto done;
		
		nfsv4_printk(write, "updatepage() found conflict, syncing page\n");
		if ((status = nfs4_sync_file(inode, file, page->index,
					     1, FLUSH_WAIT | FLUSH_STABLE)))
			goto done;
	}

	/* XXX: Could we just do this in nfs4_update_request()? */
	if (req->wb_bytes == PAGE_CACHE_SIZE)
		SetPageUptodate(page);
	
	status = 0;
	if (file->f_flags & O_SYNC) {
		nfsv4_printk(write, "updatepage: O_SYNC set, syncing page\n");
		status = nfs4_sync_file(inode, file, page->index,
					1, FLUSH_SYNC | FLUSH_STABLE);
		if (!status && file->f_error < 0)
			status = file->f_error;
		file->f_error = 0;
	}
	else if (req->wb_offset == 0 && req->wb_bytes == PAGE_CACHE_SIZE) {
		nfsv4_printk(write, "updatepage() calling nfs4_strategy()\n");
		nfs4_strategy(inode);
	}
	if (!req->wb_count)
		BUG();
	nfs4_release_request(req);
	status = nfs4_sync_file(dentry->d_inode, NULL, 0, 0, FLUSH_WAIT | FLUSH_SYNC);
	
done:
	if (status < 0)
		ClearPageUptodate(page);
	return status;
}

static int
write_rpcsetup(unsigned int npages, struct nfs4_page **pages, int how,
	       struct compound4args *argp, struct nfs4fs_putfh *putfh,
	       struct nfs4fs_write *write, struct nfs4fs_getattr *reval)
{
	struct nfs4fs_file_data *fdata;
	struct dentry *dentry;
	int status;

	/* TODO: This routine could be improved to notice when two pages are adjacent,
	   and coalescing the ranges into a single NFS write request.  Is it worth doing
	   this? */
	
	nfsv4_printk(write, "write_rpcsetup: starting, %d pages\n", npages);
	
	if (!npages || (npages > NFS4_MAXIO))
		BUG();
	if (!pages[0])
		BUG();
	fdata = pages[0]->wb_fdata;
	if (!fdata)
		BUG();
	dentry = fdata->fi_dentry;
	if (!dentry)
		BUG();
	if (!dentry->d_inode)
		BUG();

	argp->sb = dentry->d_sb;
	argp->tag = "write";
	putfh->dentry = dentry;
	write->how = (how & FLUSH_STABLE) ? DATA_SYNC4 : UNSTABLE4;
	write->npages = npages;
	memcpy(write->pages, pages, npages * sizeof(struct nfs4_page *));

	if ((status = nfs4_setup_compound(argp))) {
		nfsv4_printk(write2, "write_rpcsetup: setup_compound() failed!\n");
		goto out_err;
	}
	if ((status = nfs4_setup_putfh(argp, putfh))) {
		nfsv4_printk(write2, "write_rpcsetup: setup_putfh() failed!\n");
		goto out_args;
	}
	if ((status = nfs4_setup_write(argp, write))) {
		nfsv4_printk(write2, "write_rpcsetup: setup_write() failed!\n");
		goto out_args;
	}
#if 0
        if ((status = nfs4_setup_reval(argp, reval))) {
                nfsv4_printk(write2, "write_rpcsetup: setup_reval() failed!\n");
                goto out_args;
        }
#else
        /*
         * BAKEOFF HACK: setup_reval() isn't quite right since we always need to
         * get the mtime.  generic_file_write() always updates the mtime on the
         * client, which is wrong: we always use server-side mtimes (otherwise, we
         * can get confused by clock skew).
         */
        reval->flags = FL_NONFATAL;
        reval->bm = &nfs4_negative_bitmap;
        if ((status = nfs4_setup_getattr(argp, reval))) {
                nfsv4_printk(write2, "write_rpcsetup: setup_getattr() failed!\n");
                goto out_args;
        }
#endif

out:
	nfsv4_printk(write, "write_rpcsetup: returning status %d\n", status);
	return status;
out_args:
	nfs4_release_compoundargs(argp);
out_err:
	nfs4_mark_vector_dirty(npages, pages);
	goto out;
}

static inline void
__release_write(struct nfs4fs_write *write)
{
	struct nfs4_page *req;
	int i;

	nfsv4_printk(write, "__release_write() called\n");
	for (i = 0; i < write->npages; i++) {
		req = write->pages[i];
		ClearPageUptodate(req->wb_page);
		SetPageError(req->wb_page);
		nfs4_inode_remove_request(req);
		nfs4_unlock_request(req);
		if (!req->wb_count)
			BUG();
		nfs4_release_request(req);
	}
	write->npages = 0;
}

static int
__handle_write(struct compound4args *argp, struct compound4res *resp,
	       struct nfs4fs_putfh *putfh, struct nfs4fs_write *write,
	       struct nfs4fs_getattr *reval)
{
	int i;
	struct nfs4_page *req;
	int status;
	
	/* TODO: nfs4_handle_write() currently returns an error if the server wrote fewer
	   bytes than requested.  The correct behavior is to retry the request with the
	   missing data.  Fix it! */

	nfsv4_printk(write, "__handle_write: starting\n");
	
	if ((status = nfs4_handle_putfh(resp, putfh))) {
		nfsv4_printk(write2, "__handle_write: handle_putfh() failed!\n");
		goto out_err;
	}
	if ((status = nfs4_handle_write(resp, write))) {
		nfsv4_printk(write2, "__handle_write: handle_write() failed!\n");
		goto out_err;
	}
	
	for (i = 0; i < write->npages; i++) {
		req = write->pages[i];

		if (req->wb_how_written != UNSTABLE4) {
			nfsv4_printk(write, "handle_write: page %d written stably, ok\n", i);
			nfs4_inode_remove_request(req);
			nfs4_unlock_request(req);
			if (!req->wb_count)
				BUG();
			nfs4_release_request(req);
		}
		else {
			nfsv4_printk(write, "handle_write: page %d written unstably, "
				     "marking for commit\n", i);
			req->wb_timeout = jiffies + NFS4_COMMIT_DELAY;
			nfs4_mark_request_commit(req);
			nfs4_unlock_request(req);
		}
	}
	write->npages = 0;

	/* XXX: Unfortunately, handle_reval() will invalidate the cache, since
	   it sees that the file's changeid has changed!  This presents a
	   potential performance problem.  What can we do about this? */
	nfs4_handle_reval(resp, reval);   /* ignore return value */
	
out:
	nfs4_release_compoundres(resp);
	nfs4_release_compoundargs(argp);
	return status;
out_err:
	__release_write(write);
	goto out;
}

static int
do_write(struct compound4args *argp, struct compound4res *resp, struct nfs4fs_putfh *putfh,
	 struct nfs4fs_write *write, struct nfs4fs_getattr *getattr)
{
	int status;

	if ((status = nfs4_call_compound(argp, resp, NULL))) {
		nfsv4_printk(write2, "do_write(): call_compound failed!\n");
		goto out_err;
	}
	status = __handle_write(argp, resp, putfh, write, getattr);

out:
	return status;
out_err:
	nfs4_release_compoundargs(argp);
	__release_write(write);
	goto out;
}

static void
write_handler(struct rpc_task *task)
{
	struct nfs4_async_compound *cp = (struct nfs4_async_compound *) task->tk_calldata;
	struct file *file = cp->ac_write.pages[0]->wb_file;
	int status;
	
	if ((status = nfs4_async_prehandler(cp, NULL))) {
		nfsv4_printk(write2, "write_handler: async_prehandler() failed!\n");
		goto out_err;
	}
	status = __handle_write(&cp->ac_arg, &cp->ac_res, &cp->ac_write_putfh,
				&cp->ac_write, &cp->ac_write_reval);
	
out:
	if (status && file)
		file->f_error = status;
	return;
out_err:
	nfs4_release_compoundargs(&cp->ac_arg);
	__release_write(&cp->ac_write);
	goto out;
}

static void
write_release(struct nfs4_async_compound *cp)
{
	__release_write(&cp->ac_write);
}

static int
nfs4_flush_one(struct inode *inode, unsigned int npages, struct nfs4_page **pages, int how)
{
  /* AAA: move these vars back down */
	struct nfs4_async_compound *cp;
	struct rpc_task *task;
	struct compound4args cargs;
	struct compound4res cres;
	struct nfs4fs_putfh putfh;
	struct nfs4fs_write write;
	struct nfs4fs_getattr getattr;	
	int i, status;

	nfsv4_printk(write, "flush_one() starting, npages=%d, sync=%s\n", npages,
		     (how & FLUSH_SYNC) ? "yes" : "no");
	
	if (!npages || npages > NFS4_MAXIO)
		BUG();
	for (i = 0; i < npages; i++) {
		if (!list_empty(&pages[i]->wb_list))
			BUG();
		if (list_empty(&pages[i]->wb_writelist))
			BUG();
		if (!test_bit(PG_BUSY, &pages[i]->wb_flags))
			BUG();
	}
	
	if (how & FLUSH_SYNC) {
		status = write_rpcsetup(npages, pages, how, &cargs, &putfh, &write, &getattr);
		if (status)
			goto out;
		if ((status = do_write(&cargs, &cres, &putfh, &write, &getattr)))
			goto out;
	}
	else {
		status = -ENOMEM;
		if (!(cp = nfs4_async_create(inode->i_sb)))
			goto out_bad;
		status = write_rpcsetup(npages, pages, how, &cp->ac_arg, &cp->ac_write_putfh,
					&cp->ac_write, &cp->ac_write_reval);
		if (status)
			goto out_bad;
		cp->ac_release = write_release;
		task = &cp->ac_task;
		task->tk_action = nfs4_async_call_compound;
		task->tk_exit = write_handler;

		/* XXX: I know that we need a lock_kernel()..unlock_kernel() sandwich here,
		   but do we also need an rpc_clnt_sigmask()..rpc_clnt_sigunmask()?  Same
		   question applies at several other places on the client! */
		lock_kernel();
		rpc_execute(task);
		unlock_kernel();
	}
	
out:
	nfsv4_printk(write, "flush_one(): returning status %d\n", status);
	return status;
out_bad:
	nfs4_mark_vector_dirty(npages, pages);
	goto out;
}

static int
nfs4_flush_list(struct inode *inode, struct list_head *head, int how)
{
	int status = 0;
	unsigned int npages;
	unsigned int wpages = NFS4_WPAGES(inode->i_sb);
	unsigned int wsize = NFS4_WSIZE(inode);
	struct nfs4_page *pages[NFS4_MAXIO];
	int total_pages = 0;
	
	if (wpages > NFS4_MAXIO)
		BUG();
	if (list_empty(head))
		BUG();
	
	for (;;) {
		if (list_empty(head)) {
			nfsv4_printk(write, "flush_list() all done, returning\n");
			return total_pages;
		}
		
		npages = nfs4_coalesce_requests(head, pages, wsize, wpages);
		if (!npages) {
			nfsv4_printk(write, "flush_list() couldn't lock, returning\n");
			return total_pages;
		}
		
		nfsv4_printk(write, "flush_list() calling flush_one()\n");
		if ((status = nfs4_flush_one(inode, npages, pages, how))) {
			nfsv4_printk(write2, "flush_one() returned error!\n");
			nfs4_mark_list_dirty(head);
			return status;
		}
		total_pages += npages;
	}
}

static int
commit_rpcsetup(struct list_head *src, struct compound4args *argp, struct nfs4fs_putfh *putfh,
		struct nfs4fs_commit *commit, struct nfs4fs_getattr *getattr)
{
	struct list_head *head = &commit->pages;
	struct nfs4_page *first, *last;
	struct dentry *dentry;
	struct inode *inode;
	loff_t start, end, len;
	int status;
	
	INIT_LIST_HEAD(head);
	list_splice(src, head);
	INIT_LIST_HEAD(src);    /* just to be on the safe side */

	if (list_empty(head))
		BUG();
	
	/*
	 * Determine the offset range of requests in the COMMIT call.
	 * We rely on the fact that data->pages is an ordered list...
	 */
	first = list_entry(head->next, struct nfs4_page, wb_list);
	start = ((loff_t)first->wb_page->index << PAGE_CACHE_SHIFT) + first->wb_offset;
	last = list_entry(head->prev, struct nfs4_page, wb_list);
	end = ((loff_t)last->wb_page->index << PAGE_CACHE_SHIFT)
		+ last->wb_offset + last->wb_bytes;
	len = end - start;
	
	/* If 'len' is not a 32-bit quantity, pass '0' in the COMMIT call */
	dentry = first->wb_fdata->fi_dentry;
	inode = dentry->d_inode;
	if (end >= inode->i_size || len < 0 || len > (~((u32)0) >> 1))
		len = 0;
	nfsv4_printk(write, "setting up commit, start=%Ld, length=%Ld\n", start, len);
	
	argp->sb = inode->i_sb;
	argp->tag = "commit";
	putfh->dentry = dentry;
	commit->start = start;
	commit->length = len;
	
	if ((status = nfs4_setup_compound(argp))) {
		nfsv4_printk(write2, "commit_rpcsetup: setup_compound() failed!\n");
		goto out_err;
	}
	if ((status = nfs4_setup_putfh(argp, putfh))) {
		nfsv4_printk(write2, "commit_rpcsetup: setup_putfh() failed!\n");
		goto out_args;
	}
	if ((status = nfs4_setup_commit(argp, commit))) {
		nfsv4_printk(write2, "commit_rpcsetup: setup_commit() failed!\n");
		goto out_args;
	}
	if ((status = nfs4_setup_reval(argp, getattr))) {
		nfsv4_printk(write2, "commit_rpcsetup: setup_reval() failed!\n");
		goto out_args;
	}

out:
	return status;
out_args:
	nfs4_release_compoundargs(argp);
out_err:
	nfs4_mark_list_commit(head);
	goto out;
}

static inline void
__release_commit(struct nfs4fs_commit *commit)
{
	struct list_head *head = &commit->pages;
	struct nfs4_page *req;

	nfsv4_printk(write, "__release_commit() called\n");
	while (!list_empty(head)) {
		req = list_entry(head->next, struct nfs4_page, wb_list);
		list_del_init(&req->wb_list);
		nfs4_inode_remove_request(req);
		nfs4_unlock_request(req);
		if (!req->wb_count)
			BUG();
		nfs4_release_request(req);
	}
}


static inline int
__handle_commit(struct compound4args *argp, struct compound4res *resp,
		struct nfs4fs_putfh *putfh, struct nfs4fs_commit *commit,
		struct nfs4fs_getattr *getattr)
{
	struct list_head *head = &commit->pages;
	struct nfs4_page *req;
	int status;
	
	if ((status = nfs4_handle_putfh(resp, putfh))) {
		nfsv4_printk(write2, "__handle_commit(): handle_putfh() failed!\n");
		goto out_err;
	}
	if ((status = nfs4_handle_commit(resp, commit))) {
		nfsv4_printk(write2, "__handle_commit(): handle_commit() failed!\n");
		goto out_err;
	}

	while (!list_empty(head)) {
		req = list_entry(head->next, struct nfs4_page, wb_list);
		list_del_init(&req->wb_list);
		if (!memcmp(req->wb_verifier, commit->verifier, sizeof(verifier4))) {
			nfsv4_printk(write, "__handle_commit: verifiers agree, commit ok\n");
			nfs4_inode_remove_request(req);
			nfs4_unlock_request(req);
			if (!req->wb_count)
				BUG();
			nfs4_release_request(req);
		}
		else {
			static unsigned long complain = 0;
			nfsv4_printk(level1, "nfs4: verifier mismatch, server rebooted!!\n");
			if (time_before(complain, jiffies)) {
				printk(KERN_INFO
				       "nfs4: server reboot detected, rewriting data\n");
				complain = jiffies + 300 * HZ;
			}
			if (++req->wb_safeguard < 10) {
				nfs4_mark_request_dirty(req);
				nfs4_unlock_request(req);
			}
			else {
				printk(KERN_WARNING "nfs4: warning: pathological server, data "
				       "may have been lost!\n");
				nfs4_inode_remove_request(req);
				nfs4_unlock_request(req);
				nfs4_release_request(req);
				status = -EIO;
			}
		}
	}
	
	nfs4_handle_reval(resp, getattr);   /* ignore return value */

out:
	nfs4_release_compoundres(resp);
	nfs4_release_compoundargs(argp);
	return status;
out_err:
	__release_commit(commit);
	goto out;
}

static int
do_commit(struct compound4args *argp, struct compound4res *resp, struct nfs4fs_putfh *putfh,
	  struct nfs4fs_commit *commit, struct nfs4fs_getattr *getattr)
{
	int status;
	
	if ((status = nfs4_call_compound(argp, resp, NULL))) {
		nfsv4_printk(write2, "do_commit: call_compound() failed!\n");
		goto out_err;
	}
	status = __handle_commit(argp, resp, putfh, commit, getattr);
	
out:
	return status;

out_err:
	nfs4_release_compoundargs(argp);
	__release_commit(commit);
	goto out;
}

static void
commit_handler(struct rpc_task *task)
{
	struct nfs4_async_compound *cp = (struct nfs4_async_compound *)task->tk_calldata;
	struct file *file;
	int status;

	file = list_entry(cp->ac_commit.pages.next, struct nfs4_page, wb_list)->wb_file;
	
	if ((status = nfs4_async_prehandler(cp, NULL))) {
		nfsv4_printk(write2, "commit_handler: async_prehandler() failed!\n");
		goto out_err;
	}
	status = __handle_commit(&cp->ac_arg, &cp->ac_res, &cp->ac_commit_putfh,
				 &cp->ac_commit, &cp->ac_commit_reval);

out:
	if (status && file)
		file->f_error = status;
	return;
out_err:
	nfs4_release_compoundargs(&cp->ac_arg);
	__release_commit(&cp->ac_commit);
	goto out;
}

static void
commit_release(struct nfs4_async_compound *cp)
{
	__release_commit(&cp->ac_commit);
}

static int
nfs4_commit_file(struct inode *inode, int how)
{
	LIST_HEAD(dst);
	int pages;
	int status;

	nfsv4_printk(write, "commit_file() starting, sync=%s\n",
		     (how & FLUSH_SYNC) ? "yes" : "no");
	
	pages = nfs4_scan_commit(inode, &dst, NULL, 0, 0);
	if (!pages) {
		nfsv4_printk(write, "commit_file(): no pages available, returning 0\n");
		return 0;
	}

	if (how & FLUSH_SYNC) {
		struct compound4args cargs;
		struct compound4res cres;
		struct nfs4fs_putfh putfh;
		struct nfs4fs_commit commit;
		struct nfs4fs_getattr getattr;

		if ((status = commit_rpcsetup(&dst, &cargs, &putfh, &commit, &getattr)))
			goto out;
		if ((status = do_commit(&cargs, &cres, &putfh, &commit, &getattr)))
			goto out;
	}
	else {
		struct nfs4_async_compound *cp;
		struct rpc_task *task;

		status = -ENOMEM;
		if (!(cp = nfs4_async_create(inode->i_sb)))
			goto out_err;
		if ((status = commit_rpcsetup(&dst, &cp->ac_arg, &cp->ac_commit_putfh,
					      &cp->ac_commit, &cp->ac_commit_reval)))
			goto out;
		cp->ac_release = commit_release;
		task = &cp->ac_task;
		task->tk_action = nfs4_async_call_compound;
		task->tk_exit = commit_handler;

		lock_kernel();
		rpc_execute(task);
		unlock_kernel();
	}
	status = 1;   /* return 1 if a commit was done, 0 for no commit */

out:
	return status;
out_err:
	nfs4_mark_list_commit(&dst);
	goto out;
}

int
nfs4_commit_timeout(struct inode *inode, int how)
{
	struct list_head *head;
	struct nfs4_page *req;
	unsigned long now = jiffies;
	int commit_me = 0;
	
	spin_lock(&nfs4_wreq_lock);
	list_for_each(head, &NFS4_FLUSHD_COMMIT(inode)) {
		req = list_entry(head, struct nfs4_page, wb_list);
		if (time_before(req->wb_timeout, now))
			commit_me = 1;
		else if (time_before(req->wb_timeout, NFS4_FLUSHD_NEXTSCAN(inode)))
			NFS4_FLUSHD_NEXTSCAN(inode) = req->wb_timeout;
	}
	spin_unlock(&nfs4_wreq_lock);

	if (commit_me) {
		nfsv4_printk(write, "commit_timeout: calling commit_file()\n");
		return nfs4_commit_file(inode, how);
	}
	nfsv4_printk(write, "commit_timeout: no commit needed\n");
	return 0;
}

static int
nfs4_flush_file(struct inode *inode, struct file *file, unsigned long idx_start,
		unsigned int npages, int how)
{
	LIST_HEAD(head);

	if (nfs4_scan_dirty(inode, &head, file, idx_start, npages)) {
		nfsv4_printk(write, "flush_file(): calling flush_list()\n");
		return nfs4_flush_list(inode, &head, how);
	}
	nfsv4_printk(write, "flush_file(): no flush needed\n");
	return 0;
}

int
nfs4_flush_timeout(struct inode *inode, int how)
{
	LIST_HEAD(head);
	
	if (nfs4_scan_dirty_timeout(inode, &head)) {
		nfsv4_printk(write, "flush_timeout() calling flush_list()\n");
		return nfs4_flush_list(inode, &head, how);
	}
	nfsv4_printk(write, "flush_timeout(): no flush needed\n");
	return 0;
}

int
nfs4_sync_file(struct inode *inode, struct file *file,
	       unsigned long idx_start, unsigned int npages, int how)
{
	int	error,
		wait;

	/* XXX: This was taken from NFSv3, but doesn't it have a race condition?
	   What if another process locks some pages between nfs4_wait_on_requests()
	   and nfs4_flush_file()? */
	
	wait = how & FLUSH_WAIT;
	how &= ~FLUSH_WAIT;

	if (!inode)
		BUG();
	if (file && file->f_dentry->d_inode != inode)
		BUG();

	do {
		error = 0;
		if (wait) {
			nfsv4_printk(write, "sync_file() calling waiting_on_requests()\n");
			error = nfs4_wait_on_requests(inode, file, idx_start, npages);
		}
		if (error == 0) {
			nfsv4_printk(write, "sync_file() calling flush_file()\n");
			error = nfs4_flush_file(inode, file, idx_start, npages, how);
		}
		if (error == 0) {
			nfsv4_printk(write, "sync_file() calling commit_file()\n");
			error = nfs4_commit_file(inode, how);
		}
	} while (error > 0);
	return error;
}
