/*
 *  linux/fs/nfs4/compound.c
 *
 *  Copyright (c) 2001 The Regents of the University of Michigan.
 *  All rights reserved.
 *
 *  Kendrick Smith <kmsmith@umich.edu>
 *  Andy Adamson <andros@umich.edu>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
*/

#ifdef __linux__
# include <linux/nfs4/nfs4_prot.h>
# include <linux/nfs4/nfs4_debug.h>
# include <linux/nfs4/nfs4.h>
# include <linux/nfs4/nfsd4.h>
#endif

#ifdef __OpenBSD__
#include <sys/types.h>
#include <nfs/nfsd4_prot.h>
#include <nfs/nfs4.h>
#endif

#define NFS4_SUPPORTED_MINOR_VERSION  0
#define HUMAN_READABLE_TAGS           1

#if NFS4_DEBUG
static char *
opcode_name(nfs_opnum4 op)
{
	static char *name_map[] = {
		"BAD OP",
		"BAD OP",
		"BAD OP",
		"ACCESS",
		"CLOSE",
		"COMMIT",
		"CREATE",
		"DELEGPURGE",
		"DELEGRETURN",
		"GETATTR",
		"GETFH",
		"LINK",
		"LOCK",
		"LOCKT",
		"LOCKU",
		"LOOKUP",
		"LOOKUPP",
		"NVERIFY",
		"OPEN",
		"OPENATTR",
		"OPEN_CONFIRM",
		"OPEN_DOWNGRADE",
		"PUTFH",
		"PUTPUBFH",
		"PUTROOTFH",
		"READ",
		"READDIR",
		"READLINK",
		"REMOVE",
		"RENAME",
		"RENEW",
		"RESTOREFH",
		"SAVEFH",
		"SECINFO",
		"SETATTR",
		"SETCLIENTID",
		"SETCLIENTID_CONFIRM",
		"VERIFY",
		"WRITE",
		"BAD OP"
	};
	static int lastentry = sizeof(name_map) / sizeof(name_map[0]) - 1;

	if (op < 0)
		op = 0;
	if (op > lastentry)
		op = lastentry;
	return name_map[op];
}
#endif

int
nfsd4_proc_compound(struct global_state *stp, COMPOUND4args *argp, COMPOUND4res *resp)
{
	int i;
	int opcnt = argp->OPCNT;
	int status = 0;

	if ((status = utf8dup(&resp->tag, &argp->tag))) {
		nfsv4_printk(level1, "nfsd4_proc_compound: couldn't allocate resp tag!\n");
		goto out_no_tag;
	}
	if (!opcnt) {
		nfsv4_printk(level2, "nfsd4_proc_compound: opcnt 0?!\n");
		goto out_null;
	}
	if (argp->minorversion > NFS4_SUPPORTED_MINOR_VERSION) {
		/* according to draft7 spec, this takes precedence over all other errors */
		nfsv4_printk(level1, "minor version %d not supported!\n",
			     argp->minorversion);
		status = NFS4ERR_MINOR_VERS_MISMATCH;
		goto out_null;
	}
	if (!(resp->RESVAL = NFS4_ALLOC(opcnt * sizeof(nfs_resop4)))) {
		nfsv4_printk(level1, "couldn't allocate resop buffer [%d bytes]!\n",
			     opcnt * sizeof(nfs_resop4));
		status = NFS4ERR_RESOURCE;
		goto out_null;
	}

#if HUMAN_READABLE_TAGS
	nfsv4_printk(level1, "nfsd4_proc_compound() starting, tag=\"%.*s\", opcnt=%d\n",
		     (int)argp->tag.utf8string_len, argp->tag.utf8string_val, opcnt);
#else
	nfsv4_printk(level1, "nfsd4_proc_compound() starting, opcnt=%d\n", opcnt);
#endif

	for (i = 0; !status && (i < opcnt); i++) {
		nfsv4_printk(level2, "processing compound op #%d [%s] [opcode %d]\n", i,
			     opcode_name(argp->OPVAL[i].argop), argp->OPVAL[i].argop);
		switch (argp->OPVAL[i].argop) {
		case OP_ACCESS:
			status = nfs4_access(stp, &argp->OPVAL[i].OPU.opaccess,
					     &resp->RESVAL[i].RESU.opaccess);
			resp->RESVAL[i].RESU.opaccess.status = status;
			break;
		case OP_CLOSE:
			status = nfs4_close(stp, &argp->OPVAL[i].OPU.opclose,
					    &resp->RESVAL[i].RESU.opclose);
			resp->RESVAL[i].RESU.opclose.status = status;
			break;
		case OP_COMMIT:
			status = nfs4_commit(stp, &argp->OPVAL[i].OPU.opcommit,
					     &resp->RESVAL[i].RESU.opcommit);
			resp->RESVAL[i].RESU.opcommit.status = status;
			break;
		case OP_CREATE:
			status = nfs4_create(stp, &argp->OPVAL[i].OPU.opcreate,
					     &resp->RESVAL[i].RESU.opcreate);
			resp->RESVAL[i].RESU.opcreate.status = status;
			break;
		case OP_DELEGPURGE:
			status = nfs4_delegpurge(stp, &argp->OPVAL[i].OPU.opdelegpurge,
						 &resp->RESVAL[i].RESU.opdelegpurge);
			resp->RESVAL[i].RESU.opdelegpurge.status = status;
			break;
		case OP_DELEGRETURN:
			status = nfs4_delegreturn(stp, &argp->OPVAL[i].OPU.opdelegreturn,
						  &resp->RESVAL[i].RESU.opdelegreturn);
			resp->RESVAL[i].RESU.opdelegreturn.status = status;
			break;
		case OP_GETATTR:
			status = nfs4_getattr(stp, &argp->OPVAL[i].OPU.opgetattr,
					      &resp->RESVAL[i].RESU.opgetattr);
			resp->RESVAL[i].RESU.opgetattr.status = status;
			break;
		case OP_GETFH:
			status = nfs4_getfh(stp, &resp->RESVAL[i].RESU.opgetfh);
			resp->RESVAL[i].RESU.opgetfh.status = status;
			break;
		case OP_LINK:
			status = nfs4_link(stp, &argp->OPVAL[i].OPU.oplink,
					   &resp->RESVAL[i].RESU.oplink);
			resp->RESVAL[i].RESU.oplink.status = status;
			break;
		case OP_LOCK:
			status = nfs4_lock(stp, &argp->OPVAL[i].OPU.oplock,
					   &resp->RESVAL[i].RESU.oplock);
			resp->RESVAL[i].RESU.oplock.status = status;
			break;
		case OP_LOCKT:
			status = nfs4_lockt(stp, &argp->OPVAL[i].OPU.oplockt,
					    &resp->RESVAL[i].RESU.oplockt);
			resp->RESVAL[i].RESU.oplockt.status = status;
			break;
		case OP_LOCKU:
			status = nfs4_locku(stp, &argp->OPVAL[i].OPU.oplocku,
					    &resp->RESVAL[i].RESU.oplocku);
			resp->RESVAL[i].RESU.oplocku.status = status;
			break;
		case OP_LOOKUP:
			status = nfs4_lookup(stp, &argp->OPVAL[i].OPU.oplookup,
					     &resp->RESVAL[i].RESU.oplookup);
			resp->RESVAL[i].RESU.oplookup.status = status;
			break;
		case OP_LOOKUPP:
			status = nfs4_lookupp(stp, &resp->RESVAL[i].RESU.oplookupp);
			resp->RESVAL[i].RESU.oplookupp.status = status;
			break;
		case OP_NVERIFY:
			status = nfs4_nverify(stp, &argp->OPVAL[i].OPU.opnverify,
					      &resp->RESVAL[i].RESU.opnverify);
			resp->RESVAL[i].RESU.opnverify.status = status;
			break;
		case OP_OPEN:
			status = nfs4_open(stp, &argp->OPVAL[i].OPU.opopen,
					   &resp->RESVAL[i].RESU.opopen);
			resp->RESVAL[i].RESU.opopen.status = status;
			break;
		case OP_OPENATTR:
			status = nfs4_openattr(stp, &resp->RESVAL[i].RESU.opopenattr);
			resp->RESVAL[i].RESU.opopenattr.status = status;
			break;
		case OP_OPEN_CONFIRM:
			status = nfs4_open_confirm(stp, &argp->OPVAL[i].OPU.opopen_confirm,
						   &resp->RESVAL[i].RESU.opopen_confirm);
			resp->RESVAL[i].RESU.opopen_confirm.status = status;
			break;
		case OP_OPEN_DOWNGRADE:
			status = nfs4_open_downgrade(stp, &argp->OPVAL[i].OPU.opopen_downgrade,
						     &resp->RESVAL[i].RESU.opopen_downgrade);
			resp->RESVAL[i].RESU.opopen_downgrade.status = status;
			break;
		case OP_PUTFH:
			status = nfs4_putfh(stp, &argp->OPVAL[i].OPU.opputfh,
					    &resp->RESVAL[i].RESU.opputfh);
			resp->RESVAL[i].RESU.opputfh.status = status;
			break;
		case OP_PUTPUBFH:
			status = nfs4_putpubfh(stp, &resp->RESVAL[i].RESU.opputpubfh);
			resp->RESVAL[i].RESU.opputpubfh.status = status;
			break;
		case OP_PUTROOTFH:
			status = nfs4_putrootfh(stp, &resp->RESVAL[i].RESU.opputrootfh);
			resp->RESVAL[i].RESU.opputrootfh.status = status;
			break;
		case OP_READ:
			status = nfs4_read(stp, &argp->OPVAL[i].OPU.opread,
					   &resp->RESVAL[i].RESU.opread);
			resp->RESVAL[i].RESU.opread.status = status;
			break;
		case OP_READDIR:
			status = nfs4_readdir(stp, &argp->OPVAL[i].OPU.opreaddir,
					      &resp->RESVAL[i].RESU.opreaddir);
			resp->RESVAL[i].RESU.opreaddir.status = status;
			break;
		case OP_READLINK:
			status = nfs4_readlink(stp, &resp->RESVAL[i].RESU.opreadlink);
			resp->RESVAL[i].RESU.opreadlink.status = status;
			break;
		case OP_REMOVE:
			status = nfs4_remove(stp, &argp->OPVAL[i].OPU.opremove,
					     &resp->RESVAL[i].RESU.opremove);
			resp->RESVAL[i].RESU.opremove.status = status;
			break;
		case OP_RENAME:
			status = nfs4_rename(stp, &argp->OPVAL[i].OPU.oprename,
					     &resp->RESVAL[i].RESU.oprename);
			resp->RESVAL[i].RESU.oprename.status = status;
			break;
		case OP_RENEW:
			status = nfs4_renew(stp, &argp->OPVAL[i].OPU.oprenew,
					    &resp->RESVAL[i].RESU.oprenew);
			resp->RESVAL[i].RESU.oprenew.status = status;
			break;
		case OP_RESTOREFH:
			status = nfs4_restorefh(stp,
						&resp->RESVAL[i].RESU.oprestorefh);
			resp->RESVAL[i].RESU.oprestorefh.status = status;
			break;
		case OP_SAVEFH:
			status = nfs4_savefh(stp, &resp->RESVAL[i].RESU.opsavefh);
			resp->RESVAL[i].RESU.opsavefh.status = status;
			break;
		case OP_SECINFO:
			status = nfs4_secinfo(stp, &argp->OPVAL[i].OPU.opsecinfo,
					      &resp->RESVAL[i].RESU.opsecinfo);
			resp->RESVAL[i].RESU.opsecinfo.status = status;
			break;
		case OP_SETATTR:
			status = nfs4_setattr(stp, &argp->OPVAL[i].OPU.opsetattr,
					      &resp->RESVAL[i].RESU.opsetattr);
			resp->RESVAL[i].RESU.opsetattr.status = status;
			break;
		case OP_SETCLIENTID:
			status = nfs4_setclientid(stp, &argp->OPVAL[i].OPU.opsetclientid,
						  &resp->RESVAL[i].RESU.opsetclientid);
			resp->RESVAL[i].RESU.opsetclientid.status = status;
			break;
		case OP_SETCLIENTID_CONFIRM:
			status = nfs4_setclientid_confirm(stp, &argp->OPVAL[i].OPU.opsetclientid_confirm,
							  &resp->RESVAL[i].RESU.opsetclientid_confirm);
			resp->RESVAL[i].RESU.opsetclientid_confirm.status = status;
			break;
		case OP_VERIFY:
			status = nfs4_verify(stp, &argp->OPVAL[i].OPU.opverify,
					     &resp->RESVAL[i].RESU.opverify);
			resp->RESVAL[i].RESU.opverify.status = status;
			break;
		case OP_WRITE:
			status = nfs4_write(stp, &argp->OPVAL[i].OPU.opwrite,
					     &resp->RESVAL[i].RESU.opwrite);
			resp->RESVAL[i].RESU.opwrite.status = status;
			break;

		default:
			nfsv4_printk(level2, "received bad opcode %d!\n",
				     argp->OPVAL[i].argop);
			status = NFS4ERR_INVAL;
		}
		resp->RESVAL[i].resop = argp->OPVAL[i].argop;	
	}
	resp->RESCNT = i;

out:
	nfsv4_printk(level1, "nfsd4_proc_compound() returning status %d\n", status);
	nfsv4_printk(level2, "\n");
	resp->status = status;
	return status;
out_no_tag:
	resp->tag.utf8string_len = 0;
	resp->tag.utf8string_val = NULL;
out_null:
	resp->RESCNT = 0;
	resp->RESVAL = NULL;
	goto out;
}
