/*
 * linux/fs/nfsd4/nfsctl.c
 * Syscall interface to knfsd.
 *
 * from linux/fs/nfsd/nfsctl.c
 *    Copyright (C) 1995, 1996 Olaf Kirch <okir@monad.swb.de>
 * NFSv4 Implementation
 *    Kendrick Smith <kmsmith@umich.edu>
 *    Andy Adamson   <andros@umich.edu>
 *    Copyright (C) 2001 The Regents of the University of Michigan
 * 
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>

#include <linux/linkage.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/fcntl.h>
#include <linux/net.h>
#include <linux/in.h>
#include <linux/unistd.h>
#include <linux/malloc.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <linux/smp.h>
#include <linux/smp_lock.h>
#include <linux/sunrpc/svc.h>

#include <linux/nfs4/nfs4_prot.h>
#include <linux/nfs4/nfs4_debug.h>
#include <linux/nfs4/nfs4.h>
#include <linux/nfs4/nfsd4_state.h>
#include <linux/nfs4/nfsd4.h>
#include <linux/nfsd/interface.h>

extern long sys_call_table[];
extern int exports_read_proc(char *page, char **start, off_t off, 
					int count, int *eof, void *data);
extern void proc_exports_init(void);

static int	initialized = 0;

/*
 * This semaphore is used to serialize calls to nfsservctl().
 */
DECLARE_RWSEM(nfs4_ctl_sema);

/*
 * Initialize nfsd
 */

static void
nfs4_init(void)
{
	/* There used to be code here to create a procfs entry, but I have removed it
	   temporarily since it seems to contain buffer overflows! */
	nfsv4_printk(ctl, "nfs4_init: calling nfsd_stat_init()\n");
	nfsd_stat_init();	/* Statistics */
	nfsv4_printk(ctl, "nfs4_init: calling nfs4_export_init()\n");
	nfs4_export_init();	/* Exports table */
	nfsv4_printk(ctl, "nfs4_init: calling nfs4_state_init()\n");
	nfs4_state_init();
	initialized = 1;
}

long
asmlinkage handle_sys_nfsservctl(int cmd, void *opaque_argp, void *opaque_resp)
{
	struct nfsctl_arg	arg;
	union nfsctl_res	res;
	unsigned int		namlen;
	char			*name;
	int			err;
	
	MOD_INC_USE_COUNT;
	lock_kernel();
	down_write(&nfs4_ctl_sema);
	
	if (!initialized)
		nfs4_init();
	err = -EPERM;
	if (!capable(CAP_SYS_ADMIN))
		goto done;

	/* XXX : are these access_ok()'s really necessary? */
	err = -EFAULT;
	if (!opaque_argp || !opaque_resp ||
	    !access_ok(VERIFY_READ, opaque_argp, sizeof(struct nfsctl_arg)) ||
	    !access_ok(VERIFY_WRITE, opaque_resp, sizeof(union nfsctl_res)) ||
	    copy_from_user(&arg, opaque_argp, sizeof(struct nfsctl_arg)))
		goto done;
	
	err = -EINVAL;
	if (arg.ca_version != NFSCTL_VERSION) {
		printk(KERN_WARNING "nfs4: incompatible version in syscall.\n");
		printk(KERN_WARNING "nfs4: maybe your export parser isn't up-to-date?\n");
		goto done;
	}
	switch (cmd) {
	case NFS4CTL_ADDPSEUDO:
		namlen = arg.ca_addpseudo.namlen;
		if (namlen > PATH_MAX) {
			nfsv4_printk(ctl2, "nfs4servctl: pathname too long!\n");
			err = -ENAMETOOLONG;
			goto done;
		}
		
		/* For nfs4_addpseudo()'s benefit, replace the user-space pathname with a
		   kernel-space copy */
		if (!(name = NFS4_ALLOC(namlen + 1))) {
			nfsv4_printk(ctl2, "nfs4servctl: couldn't alloc pathname!\n");
			err = -ENOMEM;
			goto done;
		}
		if (!access_ok(VERIFY_READ, arg.ca_addpseudo.name, namlen) ||
		    copy_from_user(name, arg.ca_addpseudo.name, namlen)) {
			err = -EFAULT;
			goto done;
		}
		name[namlen] = 0;
		arg.ca_addpseudo.name = name;
		
		err = nfs4_addpseudo(&arg.ca_addpseudo, &res);
		break;
	case NFS4CTL_EXPORT:
		namlen = arg.ca_export.pathlen;
		if (namlen > PATH_MAX) {
			nfsv4_printk(ctl2, "nfs4servctl: pathname too long!\n");
			err = -ENAMETOOLONG;
			goto done;
		}
		
		/* For nfs4_export()'s benefit, replace the user-space pathname with a
		   kernel-space copy */
		if (!(name = NFS4_ALLOC(namlen + 1))) {
			nfsv4_printk(ctl2, "nfs4servctl: couldn't alloc pathname!\n");
			err = -ENOMEM;
			goto done;
		}
		if (!access_ok(VERIFY_READ, arg.ca_export.path, namlen) ||
		    copy_from_user(name, arg.ca_export.path, namlen)) {
			err = -EFAULT;
			goto done;
		}
		name[namlen] = 0;
		arg.ca_export.path = name;
		
		err = nfs4_export(&arg.ca_export);
		break;
	case NFS4CTL_START:
		err = nfs4_start(&arg.ca_start);
		break;
	case NFS4CTL_REFRESH:
		err = nfs4_refresh(&res);
		break;
	case NFS4CTL_SETPUBFH:
		err = nfs4_setpubfh(arg.ca_setpubfh);
		break;
	default:
		nfsv4_printk(ctl2, "nfs4servctl: bad commmand (%d)!\n", cmd);
		goto done;
	}
	
	if (!err && copy_to_user(opaque_resp, &res, sizeof(union nfsctl_res)))
		err = -EFAULT;

done:
	nfsv4_printk(ctl, "nfsservctl: returning status %d\n", err);
	up_write(&nfs4_ctl_sema);
	unlock_kernel();
	MOD_DEC_USE_COUNT;
	return err;
}

#ifdef MODULE
/* New-style module support since 2.1.18 */
EXPORT_NO_SYMBOLS;
MODULE_AUTHOR("UM-CITI");

struct nfsd_linkage nfsd_linkage_s = {
	do_nfsservctl: handle_sys_nfsservctl,
};

/*
 * Initialize the module
 */
int
init_module(void)
{
	printk(KERN_INFO "Installing knfsd\n");
	nfsd_linkage = &nfsd_linkage_s;
	proc_exports_init();
	return 0;
}

/*
 * Clean up the mess before unloading the module
 */
void
cleanup_module(void)
{
	remove_proc_entry("fs/nfs4", NULL);
	nfsd_linkage = NULL;
	if (initialized) {
		nfsv4_printk(ctl, "calling nfs4_state_shutdown\n");
		nfs4_state_shutdown();
		nfsv4_printk(ctl, "calling nfs4_svc_shutdown\n");
		nfs4_svc_shutdown();
		nfsv4_printk(ctl, "calling nfs4_export_shutdown()\n");
		nfs4_export_shutdown();
		nfsv4_printk(ctl, "calling nfs4_stat_shutdown()\n");
		nfsd_stat_shutdown();
	}
	printk(KERN_INFO "Uninstalling knfsd\n");
}
#endif

void 
proc_exports_init(void)
{
        if (!proc_mkdir("fs/nfs4", 0))
                return;
        create_proc_read_entry("fs/nfs4/exports", 0, 0, exports_read_proc,NULL);
}

