/*
 *  linux/fs/nfs4/export.c
 *
 *  Copyright (c) 2001 The Regents of the University of Michigan.
 *  All rights reserved.
 *
 *  Kendrick Smith <kmsmith@umich.edu>
 *  Andy Adamson   <andros@umich.edu>
 *  Jake Moilanen  <jmoilane@umich.edu>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
*/

#include <linux/list.h>
#include <linux/nfs4/nfs4_prot.h>
#include <linux/nfs4/nfs4_debug.h>
#include <linux/nfs4/nfs4.h>
#include <linux/nfs4/nfsd4.h>


/*
 *  The algorithms here are based on linear search, so will be slow if 
 *  the # of exports becomes large.  For now, I'm assuming that a "large 
 *  pseudofs" is not a case of interest.  Later, it
 *  may be worthwhile to add hashtables or some other form of fast lookup.
 */

static struct nfs4_export	export_table[NFS4_MAX_EXPORTS];
static unsigned int		current_export;
static struct rw_semaphore	export_sema;

static unsigned int		pubfh_id;
static int			pubfh_set;

void
nfs4_export_init(void)
{
	export_table[0].ex_namlen = 0;
	export_table[0].ex_name = NULL;
	export_table[0].ex_id = 0;
	export_table[0].ex_parent = 0;
	INIT_LIST_HEAD(&export_table[0].ex_subdirs);
	export_table[0].ex_dentry = NULL;
	export_table[0].ex_nchild = 0;
	INIT_LIST_HEAD(&export_table[0].ex_child);
	current_export = 1;
	pubfh_id = 0;
	pubfh_set = 0;
	init_rwsem(&export_sema);
}

void
nfs4_export_readlock(void)
{
	down_read(&export_sema);
}

void
nfs4_export_readunlock(void)
{
	up_read(&export_sema);
}

int
nfs4_addpseudo(struct nfs4ctl_addpseudo *arg, union nfsctl_res *res)
{
	unsigned int parent_id = arg->parent_id;
	unsigned int namlen = arg->namlen;
	char *name = arg->name;
	struct list_head *l;
	struct nfs4_export *p;
	int status;

	/* note: caller has allocated name with NFS4_ALLOC(); free if not going to use */
	
	nfsv4_printk(ctl, "nfs4_addpseudo: starting, parent_id=%d, name=\"%.*s\"\n",
		     parent_id, (int)namlen, name);

	/* sorry, calling your export '.' or '..' is not allowed */
	if ( ((namlen == 1) && (name[0] == '.')) ||
	     ((namlen == 2) && (name[0] == '.') && (name[1] == '.')) ) {
		nfsv4_printk(ctl2, "nfsd4: illegal filename");
		status = -EINVAL;
		goto out_free;
	}
	
	down_write(&export_sema);
	if (current_export == NFS4_MAX_EXPORTS) {
		printk(KERN_WARNING "nfsd4: export table full\n");
		status = -ENOMEM;
		goto out_free;
	}
	if ((parent_id >= current_export) ||
	    (export_table[parent_id].ex_dentry != NULL)) {
		nfsv4_printk(ctl2, "nfsd: bad parent id %d in ADDPSEUDO\n", parent_id);
		status = -EINVAL;
		goto out_free;
	}

	/* if this is a duplicate name, return the existing id */
	list_for_each(l, &export_table[parent_id].ex_child) {
		p = list_entry(l, struct nfs4_export, ex_subdirs);
		if ((p->ex_namlen == namlen) && !memcmp(p->ex_name, name, namlen)) {
			nfsv4_printk(ctl, "nfsd: duplicate filename, returning existing id\n");
			res->id = p->ex_id;
			status = 0;
			goto out_free;
		}
	}
	
	export_table[current_export].ex_namlen = namlen;
	export_table[current_export].ex_name = name;
	export_table[current_export].ex_id = current_export;
	export_table[current_export].ex_parent = &export_table[parent_id];
	list_add(&export_table[current_export].ex_subdirs,
		 &export_table[parent_id].ex_child);
	export_table[current_export].ex_dentry = NULL;
	export_table[current_export].ex_nchild = 0;
	INIT_LIST_HEAD(&export_table[current_export].ex_child);
	export_table[parent_id].ex_nchild++;
	res->id = current_export;
	current_export++;
	status = 0;
	
out:
	up_write(&export_sema);
	return status;
out_free:
	NFS4_FREE(name);
	goto out;
}

int
nfs4_export(struct nfs4ctl_export *arg)
{
	unsigned int id = arg->id;
	unsigned int namlen = arg->pathlen;
	char *filename = arg->path;
	struct nameidata nd;
	struct dentry *dentry;
	int i;
	int status = 0;

	/* note: caller has allocated filename with NFS4_ALLOC(); free if not going to use.. */
	/* XXX : make sure refcount handling is right here.. */

	nfsv4_printk(ctl, "nfs4_export: starting, id=%d, pathname=\"%.*s\", krbb5 %d\n", 
	             id, (int)namlen,
		     filename, arg->sec_krb5);

	down_write(&export_sema);
	if ((id >= current_export) || (export_table[id].ex_dentry != NULL) ||
	    (export_table[id].ex_nchild != 0)) {
		nfsv4_printk(ctl2, "nfsd: bad id %d in EXPORT\n", id);
		status = -EINVAL;
		goto out_free;
	}
	if (path_init(filename, LOOKUP_FOLLOW|LOOKUP_POSITIVE|LOOKUP_DIRECTORY, &nd))
		if ((status = path_walk(filename, &nd))) {
			nfsv4_printk(ctl2, "nfsd: lookup \"%s\" failed, status=%d\n",
				     filename, status);
			/* on failure, path_walk() doesn't increase any refcounts */
			goto out_free;
		}
	mntput(nd.mnt);   /* we don't need this */
	
	if (!(nd.dentry->d_inode->i_mode & S_IFDIR)) {
		nfsv4_printk(ctl2, "nfsd: exported object is not a directory!\n");
		status = -ENOTDIR;
		goto out_free;
	}
	for (i = 0; i < current_export; i++)
		if ((dentry = export_table[i].ex_dentry) &&
		    (is_subdir(dentry, nd.dentry) || is_subdir(nd.dentry, dentry))) {
			nfsv4_printk(ctl2, "nfsd: nested exports\n");
			status = -EINVAL;
			dput(nd.dentry);
			goto out;
		}

	/* code before this point just does error-checking, and sets nd.dentry. */
	
	export_table[id].ex_dentry = nd.dentry;
	export_table[id].ex_exportnamlen = namlen;
	export_table[id].ex_exportname = filename;
	export_table[id].ex_readonly = arg->read_only;
	 /* sec modes */
	export_table[id].ex_sec_none = arg->sec_none;
	export_table[id].ex_sec_sys = arg->sec_sys;
	export_table[id].ex_sec_dh = arg->sec_dh;
	export_table[id].ex_sec_krb5 = arg->sec_krb5;
        export_table[id].ex_sec_krb5i = arg->sec_krb5i;
        export_table[id].ex_sec_krb5p = arg->sec_krb5p;
        export_table[id].ex_sec_spkm3 = arg->sec_spkm3;
        export_table[id].ex_sec_spkm3i = arg->sec_spkm3i;
        export_table[id].ex_sec_spkm3p = arg->sec_spkm3p;
        export_table[id].ex_sec_lkey = arg->sec_lkey;
        export_table[id].ex_sec_lkeyi = arg->sec_lkeyi;
        export_table[id].ex_sec_lkeyp = arg->sec_lkeyp;

	nfsv4_printk(ctl2,"nfsd exp index %d ex_exportname: %s\n",id,
            export_table[id].ex_exportname);
	  
out:
	up_write(&export_sema);
	return status;
out_free:
	NFS4_FREE(filename);
	goto out;
}

int
nfs4_setpubfh(unsigned int export_id)
{
	int status;

	if (pubfh_set) {
		status = -EINVAL;
		nfsv4_printk(ctl2, "nfsd4: public filehandle set twice!\n");
	}
	else if (export_id >= current_export) {
		status = -EINVAL;
		nfsv4_printk(ctl2, "nfsd4: bad export id!\n");
	}
	else {
		status = 0;
		pubfh_set = 1;
		pubfh_id = export_id;
	}
	return status;
}

int
nfs4_setexp(unsigned int export_id, struct nfs4_export **expp)
{
	int status;
	
	if (export_id >= current_export)
		status = -EINVAL;
	else {
		*expp = &export_table[export_id];
		status = 0;
	}
	return status;
}

int
nfs4_setroot(struct nfs4_export **expp)
{
	*expp = &export_table[0];
	return 0;
}

int
nfs4_setpub(struct nfs4_export **expp)
{
	*expp = &export_table[pubfh_id];
	return 0;
}

#define RDONLY_ACCESS  (MAY_READ | MAY_EXEC)
#define FULL_ACCESS (MAY_READ | MAY_WRITE | MAY_EXEC | MAY_SATTR | MAY_TRUNC | MAY_LOCK | \
		     MAY_OWNER_OVERRIDE)
int
nfs4_setsec(struct svc_rqst *rqstp, struct nfs4_export *exp, struct nfs4_secattrs *attrs)
{
	struct svc_cred *cred = &rqstp->rq_cred;

	nfsv4_printk(auth,"nfs4_setsec: cred->cr_flavor %d exp %p\n",
	   cred->cr_flavor,exp);


#if NFS4_DEBUG
	if(cred->cr_flavor == 0)
		nfsv4_printk(auth,"cred->cr_flavor: %s\n","RPC_AUTH_NULL");
	else if(cred->cr_flavor == 1 )
		nfsv4_printk(auth,"cred->cr_flavor: %s\n","RPC_AUTH_UNIX");
	else if(cred->cr_flavor == 6)
		nfsv4_printk(auth,"cred->cr_flavor: %s\n","RPC_AUTH_GSS");
	else nfsv4_printk(auth,"INVALID cred->cr_flavor %d",cred->cr_flavor);
#endif 

    /* check for krb5 protected file system */
	if(exp && exp->ex_sec_krb5) {
		if(rqstp->rq_cred.cr_flavor != RPC_AUTH_GSS){
    /* currently, only one mechanism and qop */
			nfsv4_printk(auth,"cred->cr_flavor  %d NOT CORRECT - should be %d\n",
			cred->cr_flavor,RPC_AUTH_GSS);
			return NFS4ERR_WRONGSEC;
		}
	}


	if(rqstp->rq_cred.cr_flavor == RPC_AUTH_NULL){
	/* AUTH_NULL gets nobody access */
		nfsv4_printk(auth, "nfs4_setsec: RPC_AUTH_NULL\n");
		attrs->sa_fsuid = 99;
		attrs->sa_fsgid = 99;
	}
	else {
		/* for now, we let anyone authenticating with 
                 * AUTH_UNIX  use whatever uid/gid they specify.. 
		 *
		 * RPC_AUTH_GSS currently assumes a direct mapping between 
		 * kerberos V5 principals and local unix username and uid's.
		 */
		nfsv4_printk(auth, "nfs4_setsec: %s, uid=%d gid=%d\n", 
	                     (rqstp->rq_cred.cr_flavor == RPC_AUTH_UNIX)?
		             "RPC_AUTH_UNIX":"RPC_AUTH_GSS",
		             cred->cr_uid, cred->cr_gid);
		attrs->sa_fsuid = cred->cr_uid;
		attrs->sa_fsgid = cred->cr_gid;
	}

	/* We derive the capabilities from the uid/gid.  If uid==0 or gid==0, grant CAP_CHOWN.
	   Otherwise, don't grant any capabilities.  XXX : is this right? */
	if ((attrs->sa_fsuid == 0) || (attrs->sa_fsgid == 0))
		attrs->sa_cap_effective = (1 << CAP_CHOWN);
	else
		attrs->sa_cap_effective = 0;

	/* for now, we allow read-only access to "internal" pseudofs nodes, full access to
	   the actual exports.  WARNING: under no circumstances should access less restrictive
	   than RDONLY_ACCESS be given to "internal" nodes; this may cause the kernel to
	   crash! */
	if (exp->ex_dentry)       /* export */
		attrs->sa_access = (exp->ex_readonly) ? RDONLY_ACCESS : FULL_ACCESS;
	else
		attrs->sa_access = RDONLY_ACCESS;
	return NFS4_OK;
}


void
nfs4_export_shutdown(void)
{
	int i;

	down_write(&export_sema);
	for (i = 0; i < current_export; i++) {
		if (export_table[i].ex_dentry) {
			dput(export_table[i].ex_dentry);
			export_table[i].ex_dentry = NULL;
			if (export_table[i].ex_exportname) {
				NFS4_FREE(export_table[i].ex_exportname);
				export_table[i].ex_exportname = NULL;
			}
		}
		if (export_table[i].ex_name) {
			NFS4_FREE(export_table[i].ex_name);
			export_table[i].ex_namlen = 0;
			export_table[i].ex_name = NULL;
		}
	}
	current_export = 1;
	up_write(&export_sema);
}


static int 
proc_calc_metrics(char *page, char **start, off_t off,
                                 int count, int *eof, int len)
{
        if (len <= off+count) *eof = 1;
        *start = page + off;
        len -= off;
        if (len>count) len = count;
        if (len<0) len = 0;
        return len;
}

int
print_exports_parent(struct nfs4_export *entry, char *buf, int len)
{
	int local_len;

	if (entry == entry->ex_parent || !entry->ex_parent){
		return len;
	}
	
	local_len = print_exports_parent(entry->ex_parent, buf, len);

	if ((local_len+sizeof(entry->ex_name)) >= PAGE_SIZE)
       		return local_len;
	local_len += sprintf(buf+local_len, "/%s", entry->ex_name);
	
	return local_len;
}

int
get_exports_list(char *buf)
{
        int len = 0;
	int i;

	if((len+sizeof("# Pseudo path\tExport directory\n#\t\tSecurity Options\n\n")) 
								>= PAGE_SIZE)
        	return len;
        len += sprintf(buf+len, "# Pseudo path\tExport directory\n#\t\tSecurity Options\n\n");

        nfs4_export_readlock();
	for (i = 0; i < current_export; i++){ 
		if (export_table[i].ex_exportname){

			len = print_exports_parent(&export_table[i], buf, len);

			if ((len+sizeof(export_table[i].ex_exportname)+sizeof("/\t\n"))
								 >= PAGE_SIZE)
				return len;
       			len += sprintf(buf+len, "/\t%s\n", 
					export_table[i].ex_exportname); 

			/* read only */
			if (export_table[i].ex_readonly){
				if((len+sizeof("\t\tRead Only\n")) >= PAGE_SIZE)
					return len;
				len += sprintf(buf+len, "\t\tRead Only\n");
			}

			/* none */
			if (export_table[i].ex_sec_none){
                                if((len+sizeof("\t\tAUTH_NONE, RPC security\n")) 
								>= PAGE_SIZE)
                                        return len;
                                len += sprintf(buf+len, "\t\tAUTH_NONE, RPC security\n");
                        }

			/* sys */
                        if (export_table[i].ex_sec_sys){
                                if((len+sizeof("\t\tAUTH_SYS RPC, default uid/gid lists\n")) >= PAGE_SIZE)
                                        return len;
                                len += sprintf(buf+len, "\t\tAUTH_SYS RPC, default uid/gid lists\n");
                        }

			/* diffe-hellman */
                        if (export_table[i].ex_sec_dh){
                                if((len+sizeof("\t\tAUTH_DES, old diffe-hellman encryption\n")) >= PAGE_SIZE)
                                        return len;
                                len += sprintf(buf+len, "\t\tAUTH_DES, old diffe-hellman encryption\n");
                        }

			/* Kerberos 5 authentication */
                        if (export_table[i].ex_sec_krb5){
                                if((len+sizeof("\t\tRPCSEC_GSS Kerberos 5 authentication\n")) >= PAGE_SIZE)
                                        return len;
                                len += sprintf(buf+len, "\t\tRPCSEC_GSS Kerberos 5 authentication\n");
                        }
		
			/* Kerberos 5 integrity */
			if (export_table[i].ex_sec_krb5i){
                                if((len+sizeof("\t\tRPCSEC_GSS Kerberos 5 integrity\n")) >= PAGE_SIZE)
                                        return len;
                                len += sprintf(buf+len, "\t\tRPCSEC_GSS Kerberos 5 integrity\n");
                        }

			/* Kerberos 5 protection */
			if (export_table[i].ex_sec_krb5p){
                                if((len+sizeof("\t\tRPCSEC_GSS Kerberos 5 protection\n")) >= PAGE_SIZE)
                                        return len;
                                len += sprintf(buf+len, "\t\tRPCSEC_GSS Kerberos 5 protection\n");
                        }

			/* SPKM authentication */
			if (export_table[i].ex_sec_spkm3){
                                if((len+sizeof("\t\tSPKM authenication\n")) >= PAGE_SIZE)
                                        return len;
                                len += sprintf(buf+len, "\t\tSPKM authenication\n");
                        }

                        /* SPKM integrity */ 
                        if (export_table[i].ex_sec_spkm3i){
                                if((len+sizeof("\t\tSPKM integrity\n")) >= PAGE_SIZE)
                                        return len;
                                len += sprintf(buf+len, "\t\tSPKM integrity\n");
                        }

                        /* SPKM protection */
                        if (export_table[i].ex_sec_spkm3){
                                if((len+sizeof("\t\tSPKM protection\n")) >= PAGE_SIZE)
                                        return len;
                                len += sprintf(buf+len, "\t\tSPKM protection\n");
                        }

                        /* LIPKEY authentication */
                        if (export_table[i].ex_sec_lkey){
                                if((len+sizeof("\t\tLIPKEY authenication\n")) >= PAGE_SIZE)
                                        return len;
                                len += sprintf(buf+len, "\t\tLIPKEY authenication\n");
                        }

                        /* LIPKEY integrity */
                        if (export_table[i].ex_sec_lkeyi){
                                if((len+sizeof("\t\tLIPKEY integrity\n")) >= PAGE_SIZE)
                                        return len;
                                len += sprintf(buf+len, "\t\tLIPKEY integrity\n");
                        }

                        /* LIPKEY protection */
                        if (export_table[i].ex_sec_lkeyp){
                                if((len+sizeof("\t\tLIPKEY protection\n")) >= PAGE_SIZE)
                                        return len;
                                len += sprintf(buf+len, "\t\tLIPKEY protection\n");
                        }
			
			if((len+sizeof("\n")) >= PAGE_SIZE)
                                        return len;
                        len += sprintf(buf+len, "\n");


		}
        }
        nfs4_export_readunlock();
        return len;
}

int 
exports_read_proc(char *page, char **start, off_t off,
                                 int count, int *eof, void *data)
{
        int len = get_exports_list(page);
        return proc_calc_metrics(page, start, off, count, eof, len);
}
