/*
 *  linux/fs/nfs4/proc.c
 *
 *  Copyright (c) 2001 The Regents of the University of Michigan.
 *  All rights reserved.
 *
 *  Kendrick Smith <kmsmith@umich.edu>
 *  Andy Adamson <andros@umich.edu>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
*/


#include <linux/nfs4/nfs4_prot.h>
#include <linux/nfs4/nfs4_debug.h>
#include <linux/nfs4/nfs4.h>
#include <linux/nfs4/nfsd4.h>

static int
nfs4_proc_null(struct svc_rqst *rqstp, void *argp, void *resp)
{
	nfsv4_printk(level1, "nfsd received null proc\n");
	return 0;
}

static int
nfs4_decode_null(struct svc_rqst *rqstp, u32 *p, void *dummy)
{
	return TRUE;
}

static int
nfs4_encode_null(struct svc_rqst *rqstp, u32 *p, void *dummy)
{
	rqstp->rq_resbuf.len = p - rqstp->rq_resbuf.base;
	return TRUE;
}

static int
nfs4_decode_compound(struct svc_rqst *rqstp, u32 *p, COMPOUND4args *argp)
{
	XDR xdrs;
	
	xdrsvc_create(&xdrs, rqstp, XDR_DECODE);
	return xdr_COMPOUND4args(&xdrs, argp);
}

static int
nfs4_encode_compound(struct svc_rqst *rqstp, u32 *p, COMPOUND4res *resp)
{
	XDR xdrs;
	int status;
	u32 *start, *end;
	
	xdrsvc_create(&xdrs, rqstp, XDR_ENCODE);
	start = rqstp->rq_resbuf.buf;
	status = xdr_COMPOUND4res(&xdrs, resp);
	rqstp->rq_resbuf.len = rqstp->rq_resbuf.buf - rqstp->rq_resbuf.base;
	end = rqstp->rq_resbuf.buf;
	
	return status;
}

void
nfs4_release_compoundargs(COMPOUND4args *argp)
{
	XDR xdrs;
	
	xdrsvc_create(&xdrs, NULL, XDR_FREE);
	if (!xdr_COMPOUND4args(&xdrs, argp))
		printk("!!! nfsd: got error when freeing compound args!\n");
}

static void
release_fattr(fattr4 *fp)
{
	if (fp->attrmask.bitmap4_val)
		NFS4_FREE(fp->attrmask.bitmap4_val);
	if (fp->attr_vals.attrlist4_val)
		NFS4_FREE(fp->attr_vals.attrlist4_val);
}

static int
nfs4_release_compoundres(struct svc_rqst *rqstp, u32 *p, COMPOUND4res *resp)
{
	int i;
	
	for (i = 0; i < resp->RESCNT; i++)
		if ((i < resp->RESCNT - 1) || (!resp->status))
			switch (resp->RESVAL[i].resop) {			
			case OP_GETATTR:
				release_fattr(&resp->RESVAL[i].RESU.opgetattr.GFATTR4);
				break;
			case OP_GETFH:
				NFS4_FREE(resp->RESVAL[i].RESU.opgetfh.GETFH4FH.nfs_fh4_val);
				break;
			case OP_READ:
				NFS4_FREE(resp->RESVAL[i].RESU.opread.READVAL);
				break;
			case OP_READDIR:
				NFS4_FREE(resp->RESVAL[i].RESU.opreaddir.READDIR4res_u.
					  resok4.reply_buf);
				break;
			case OP_READLINK:
				NFS4_FREE(resp->RESVAL[i].RESU.opreadlink.LINKVAL);
			break;
			case OP_SECINFO:
				NFS4_FREE(resp->RESVAL[i].RESU.opsecinfo.SECRES.SECINFO4resok_val);
			break;
			case OP_SETATTR:
				NFS4_FREE(resp->RESVAL[i].RESU.opsetattr.attrsset.bitmap4_val);
				break;
			default:
				break;
			}

	if (resp->tag.utf8string_val)
		NFS4_FREE(resp->tag.utf8string_val);
	if (resp->RESVAL)
		NFS4_FREE(resp->RESVAL);

	return 0;
}

int
nfs4_dummy_compound(struct svc_rqst *rqstp, COMPOUND4args *argp, COMPOUND4res *resp)
{
	printk("!!! nfs4_dummy_compound hit !!!\n");
	resp->status = NFS4ERR_SERVERFAULT;
	resp->tag.utf8string_len = 0;
	resp->tag.utf8string_val = NULL;
	resp->RESCNT = 0;
	resp->RESVAL = NULL;
	return 0;
}

struct svc_procedure nfsd_procedures4[2] = {
	{ (svc_procfunc) nfs4_proc_null, (kxdrproc_t) nfs4_decode_null,
	  (kxdrproc_t) nfs4_encode_null, NULL, 4, 4, 0, 0 },
	{ (svc_procfunc) nfs4_dummy_compound, (kxdrproc_t) nfs4_decode_compound,
	  (kxdrproc_t) nfs4_encode_compound, (kxdrproc_t) nfs4_release_compoundres,
	  NFSSVC_XDRSIZE, NFSSVC_XDRSIZE, 0, 0 }
};
