/*
 *  linux/fs/nfsd4/state.c
 *
 *  Copyright (c) 2001 The Regents of the University of Michigan.
 *  All rights reserved.
 *
 *  Kendrick Smith <kmsmith@umich.edu>
 *  Andy Adamson   <andros@umich.edu>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/*
 * "Generic" implementation of the server-side state for NFSv4.  There are two big issues
 * remaining: (1) finer-grained locking.  Among other things, this will also require adding
 * reference counts so that these structures can't be released while in use.  (2) we
 * currently release nfs4_files and nfs4_openfiles as soon as they are no longer in use;
 * might be better to put them on an "unused" list and let the laundromat thread release
 * them when a timeout expires.  (3) Need reference counts to protect things from deletion
 * properly.
 */

#ifdef __linux__
#include <linux/unistd.h>
#include <linux/socket.h>
#include <linux/net.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/module.h>
#include <linux/quotaops.h>
#include <linux/nfs4/nfs4_prot.h>
#include <linux/nfs4/nfs4_debug.h>
#include <linux/nfs4/nfs4.h>
#include <linux/nfs4/nfsd4_state.h>
#include <linux/nfs4/nfsd4.h>
#endif

static int nfs4_file_upgrade(nfs4_file_desc_t *, unsigned int share_access);
static void nfs4_file_downgrade(nfs4_file_desc_t *, unsigned int share_access);
static int nfs4_file_truncate(struct svc_fh *fhp);
static int nfs4_check_fh(struct svc_fh *fhp, struct nfs4_openfile *ofp);

#ifdef __linux__

int
nfs4_file_open(struct file *filp, struct svc_fh *fhp, unsigned int share_access)
{
	struct dentry *dentry;
	struct inode *inode;
	int is_write = (share_access & OPEN4_SHARE_ACCESS_READ);
	int err;

	dentry = fhp->fh_dentry;
	inode = dentry->d_inode;
	
        /* Disallow access to files with the append-only bit set or
         * with mandatory locking enabled.
	 * XXX: Think about whether this is always necessary in the case of mandatory
	 * locking...
         */
        err = NFS4ERR_PERM;
        if (IS_APPEND(inode) || IS_ISMNDLK(inode))
                goto out;
        if (!inode->i_fop)
                goto out;

	/* XXX: learn about leases, and check for them here, a la fs/nfsd/vfs.c:nfsd_open(). */

	if (is_write && (err = get_write_access(inode)) != 0)
		goto out_nfserr;

	memset(filp, 0, sizeof(*filp));
	filp->f_op    = fops_get(inode->i_fop);
	atomic_set(&filp->f_count, 1);
	filp->f_dentry = DGET(dentry);
	if (is_write) {
		filp->f_flags = O_WRONLY;
		filp->f_mode  = FMODE_WRITE;
		DQUOT_INIT(inode);
	} else {
		filp->f_flags = O_RDONLY;
		filp->f_mode  = FMODE_READ;
	}

	err = 0;
	if (filp->f_op && filp->f_op->open) {
		err = filp->f_op->open(inode, filp);
		if (err) {
			fops_put(filp->f_op);
			if (is_write)
				put_write_access(inode);

			/* I nearly added put_filp() call here, but this filp
			 * is really on callers stack frame. -DaveM
			 */
			atomic_dec(&filp->f_count);
		}
	}
out_nfserr:
	if (err)
		err = nfserrno(err);
out:
	return err;
}

void
nfs4_file_close(struct file *filp)
{
	struct dentry	*dentry = filp->f_dentry;
	struct inode	*inode = dentry->d_inode;

	if (filp->f_op && filp->f_op->release)
		filp->f_op->release(inode, filp);
	fops_put(filp->f_op);
	if (filp->f_mode & FMODE_WRITE)
		put_write_access(inode);
	DPUT(dentry);
	filp->f_dentry = NULL;
}

static inline int
nfs4_file_upgrade(struct file *filp, unsigned int share_access)
{
	int status;
	
	if (share_access & OPEN4_SHARE_ACCESS_WRITE) {
		status = get_write_access(filp->f_dentry->d_inode);
		if (!status)
			filp->f_mode = FMODE_WRITE;
		else
			return nfserrno(status);
	}
	return NFS4_OK;
}

static inline void
nfs4_file_downgrade(struct file *filp, unsigned int share_access)
{
	if (share_access & OPEN4_SHARE_ACCESS_WRITE) {
		put_write_access(filp->f_dentry->d_inode);
		filp->f_mode = FMODE_READ;
	}
}


static inline int
nfs4_file_truncate(struct svc_fh *fhp)
{
	struct iattr iattr;
	iattr.ia_valid = ATTR_SIZE;
	iattr.ia_size = 0;
	return do_setattr(fhp, &iattr);
}

static inline int
nfs4_check_fh(struct svc_fh *fhp, struct nfs4_openfile *ofp)
{
	return (fhp->fh_dentry == ofp->of_vfs_file.f_dentry) ? NFS4_OK : NFS4ERR_BAD_STATEID;
}

#endif

/*
 * Some hash tables for server-side state, and routines for searching, insertion, and
 * deletion for each.  Note that nfs4_clients are hashed twice: once by clientid and
 * once by "nfs_clientid" (the opaque string initially sent by the client to identify
 * itself to the server).
 *
 * Some issues to consider here:
 *   - using simple hash functions for now; might try to do better later
 *   - allocating memory with NFS4_ALLOC(), NFS4_FREE(); might put everything in slab
 *     caches
 */
#define CLIENT_HASH_BITS		 4
#define LOCKOWNER_HASH_BITS		 8
#define FILE_HASH_BITS			 8
#define OPENFILE_HASH_BITS		10

#define CLIENT_HASH_SIZE		(1 << CLIENT_HASH_BITS)
#define CLIENT_HASH_MASK		(CLIENT_HASH_SIZE - 1)
#define LOCKOWNER_HASH_SIZE		(1 << LOCKOWNER_HASH_BITS)
#define LOCKOWNER_HASH_MASK		(LOCKOWNER_HASH_SIZE - 1)
#define FILE_HASH_SIZE			(1 << FILE_HASH_BITS)
#define FILE_HASH_MASK			(FILE_HASH_SIZE - 1)
#define OPENFILE_HASH_SIZE		(1 << OPENFILE_HASH_BITS)
#define OPENFILE_HASH_MASK		(OPENFILE_HASH_SIZE - 1)

static _LIST_HEAD(dummy4, nfs4_client)       clientid_hashtbl[CLIENT_HASH_SIZE];
static _LIST_HEAD(dummy5, nfs4_client)      clientstr_hashtbl[CLIENT_HASH_SIZE];
static _LIST_HEAD(dummy6, nfs4_lockowner)    lockownerid_hashtbl[LOCKOWNER_HASH_SIZE];
static _LIST_HEAD(dummy7, nfs4_lockowner)   lockownerstr_hashtbl[LOCKOWNER_HASH_SIZE];
static _LIST_HEAD(dummy8, nfs4_file)         file_hashtbl[FILE_HASH_SIZE];
static _LIST_HEAD(dummy9, nfs4_openfile)    openfile_hashtbl[OPENFILE_HASH_SIZE];

/*
 * Hash functions.  Note that nfs4_ino_hashval() is defined in state.h.
 * XXX: openfile_hashval() is not very good and should be improved
 */
#define clientid_hashval(id) \
	((id) & CLIENT_HASH_MASK)
#define clientstr_hashval(name, namelen) \
	(opaque_hashval((name), (namelen)) & CLIENT_HASH_MASK)
#define lockownerid_hashval(id) \
	((id) & LOCKOWNER_HASH_MASK)
#define lockownerstr_hashval(clientid, name, namelen) \
	(((clientid) + opaque_hashval((name), (namelen))) & LOCKOWNER_HASH_MASK)
#ifdef __linux__
#define file_hashval(ino) \
	(((unsigned int)((ino)->dev + (ino)->ino + (ino)->generation)) & FILE_HASH_MASK)
#else
#define file_hashval(ino) \
	(opaque_hashval((ino), sizeof(nfs4_ino_desc_t)) & FILE_HASH_MASK)
#endif
#define openfile_hashval(lockowner_id, file_id)  \
	(((lockowner_id) + (file_id)) & OPENFILE_HASH_MASK)

/*
 * A frequently used hash lookup
 */
static inline struct nfs4_openfile *
lookup_openfile(unsigned int hashval, stateid_lockowner_t lockowner_id, stateid_file_t file_id)
{
	struct nfs4_openfile *ofp;
	LIST_FOREACH(ofp, &openfile_hashtbl[hashval], of_hash)
		if ((ofp->of_stateid.st_lockowner == lockowner_id) &&
		    (ofp->of_stateid.st_file == file_id))
			return ofp;
	return NULL;
}

/*
 * LRU lists: these are used by the "laundromat" thread to release state which hasn't
 * been used recently.  The client LRU list contains all clients, ordered by time of
 * last lease renewal; the "laundromat" thread releases a client if it hasn't been renewed
 * within the lease period.  The lockowner LRU list contains all unused lockowners, ordered
 * by time of last use.  A lockowner is considered "unused" if it has no files open, or if
 * the only file open is an unconfirmed OPEN.  Such a lockowner is retained for the lease
 * period, then released.  (Note that we can't release it immediately since we may need to
 * replay the CLOSE operation which put the lockowner into the "unused" state.  So once the
 * lease period expires, we may not be able to correctly replay the CLOSE.  Is it really
 * correct to use the lease period, or should we use a different timeout, to represent the
 * maximum retransmission delay supported by our server?)  Warning: if a lockowner is
 * in use, it is not on the lru list, so don't dereference its "lo_lru" pointers!
 */
static TAILQ_HEAD(dummy10, nfs4_client)    client_lru;
static TAILQ_HEAD(dummy11, nfs4_lockowner) lockowner_unused_lru;
#define LOCKOWNER_UNUSED(lp)   (!lp->lo_confirmed || LIST_EMPTY(&lp->lo_perlockowner))

/*
 * Clientid 0 is reserved; we return it on a failed LOCK or LOCKT when a lock is held by
 * someone other than an NFSv4 client (either a local process, or a different network
 * file system).
 */
static u32 current_clientid = 1;
static u32 current_lockownerid = 0;
static u32 current_fileid = 0;

static time_t boot_time;            /* time when server was last (re)started */
static time_t grace_end = 0;        /* end of grace period */

stateid_t zero_stateid;             /* bits all 0 */
stateid_t one_stateid;              /* bits all 1 */

#if NFS4_DEBUG
static inline int
STALE_CLIENTID(clientid_t *clientid)
{
	if (clientid->cl_boot == boot_time)
		return 0;
	nfsv4_printk(clientid2, "stale clientid (%08x/%08x)!\n",
		     clientid->cl_boot, clientid->cl_id);
	return 1;
}

static inline int
STALE_STATEID(stateid_t *stateid)
{
	if (stateid->st_boot == (stateid_boot_t)boot_time)
		return 0;
	nfsv4_printk(stateid2, "stale stateid (%08x/%08x/%08x/%08x)!\n",
		     stateid->st_boot, stateid->st_lockowner, stateid->st_file,
		     stateid->st_generation);
	return 1;
}

static inline int
STALE_VERIFIER(open_verifier_t *verf)
{
	if (verf->ov_boot == boot_time)
		return 0;
	nfsv4_printk(lockowner2, "stale verifier (%08x/%08x)!\n", verf->ov_boot, verf->ov_id);
	return 1;
}
#else
#  define STALE_CLIENTID(clientid)    ((clientid)->cl_boot != boot_time)
#  define STALE_STATEID(stateid)      ((stateid)->st_boot != boot_time)
#  define STALE_VERIFIER(verf)        ((verf)->ov_boot != boot_time)
#endif

/*
 * Locking: we currently put a giant lock around all of this state, to protect against
 * concurrent access.  When everything is finished, and we know exactly how all of these
 * state structures will need to be interrelated, we'll go back and figure out how to
 * do the locking right.  The giant lock should be claimed at the beginning of
 * state-manipulating operations (OPEN, READ, ...) and released at the end.
 */
static struct semaphore big_sema;

void
nfs4_lock_state(void)
{
	down(&big_sema);
}

void
nfs4_unlock_state(void)
{
	up(&big_sema);
}

/*
 * Routines for allocating and freeing hashtable elements.  These are separated out so that
 * we can easily change the memory allocation scheme, or define platform-dependent
 * routines, if needed.
 */
static inline struct nfs4_client *
alloc_client(const char *name, unsigned int namelen)
{
	struct nfs4_client *clp;
	if ((clp = NFS4_ALLOC(sizeof(struct nfs4_client)))) {
		/* This type of memory management is somewhat inefficient, but we use it
		 * anyway since SETCLIENTID is not a common operation.
		 */
		if ((clp->cl_name = NFS4_ALLOC(namelen))) {
			memcpy(clp->cl_name, name, namelen);
			clp->cl_namelen = namelen;
		}
		else {
			NFS4_FREE(clp);
			clp = NULL;
		}
	}
	return clp;
}

static inline void
free_client(struct nfs4_client *clp)
{
	NFS4_FREE(clp->cl_name);
	NFS4_FREE(clp);
}

static inline struct nfs4_lockowner *
alloc_lockowner(const char *name, unsigned int namelen)
{
	struct nfs4_lockowner *lp;
	if ((lp = NFS4_ALLOC(sizeof(struct nfs4_lockowner)))) {
		memcpy(lp->lo_name, name, namelen);
		lp->lo_namelen = namelen;
	}
	return lp;
}
#define free_lockowner(lp) NFS4_FREE(lp)

#define alloc_file()       NFS4_ALLOC(sizeof(struct nfs4_file))
#define free_file(fp)      NFS4_FREE(fp)
#define alloc_openfile()   NFS4_ALLOC(sizeof(struct nfs4_openfile))
#define free_openfile(ofp) NFS4_FREE(ofp)

/*
 * release_{client,lockowner,file,openfile}: these release all resources associated with
 * the specified object, and removes it from hashtables etc.
 * Note: release_openfile() even releases share reservations.
 *
 * nfsd4_laundromat(): When the laundromat thread wakes up, it calls this routine to release
 * expired state.  The return value is the time when the thread should wake up and call the
 * routine again.
 */

static void
release_file(struct nfs4_file *fp)
{
#if NFS4_DEBUG
	if (!LIST_EMPTY(&fp->fi_perfile)) {
		printk("!!! release_file: file is open, can't release !!!\n");
		return;
	}
#endif
	LIST_REMOVE(fp, fi_hash);
	free_file(fp);
}

static inline void
release_openfile(struct nfs4_openfile *ofp)
{
	struct nfs4_lockowner *lp = ofp->of_lockowner;
	struct nfs4_file *fp = ofp->of_file;
	
	/*
	 * If closing this file puts the lockowner out of use, we move the lockowner to
	 * the unused list.
	 */
	LIST_REMOVE(ofp, of_perlockowner);
	if (lp->lo_confirmed && LIST_EMPTY(&lp->lo_perlockowner)) {
		TAILQ_INSERT_TAIL(&lockowner_unused_lru, lp, lo_lru);
		lp->lo_time = get_time();
	}
	
	/*
	 * If closing this file puts the file out of use, we currently just release the
	 * file.  XXX: should probably keep the file around on a list for a little while,
	 * and let the laundromat thread release it after a timeout, in the event that it's
	 * used again in the near future.
	 */
	LIST_REMOVE(ofp, of_perfile);
	if (LIST_EMPTY(&fp->fi_perfile))
		release_file(fp);
	
	nfs4_file_close(&ofp->of_vfs_file);
	LIST_REMOVE(ofp, of_hash);
	free_openfile(ofp);
}

static void
release_lockowner(struct nfs4_lockowner *lp)
{
	while (!LIST_EMPTY(&lp->lo_perlockowner))
		release_openfile(LIST_FIRST(&lp->lo_perlockowner));
	TAILQ_REMOVE(&lockowner_unused_lru, lp, lo_lru);
	LIST_REMOVE(lp, lo_idhash);
	LIST_REMOVE(lp, lo_strhash);
	LIST_REMOVE(lp, lo_perclient);
	free_lockowner(lp);
}

static void
release_client(struct nfs4_client *clp)
{
	while (!LIST_EMPTY(&clp->cl_perclient))
		release_lockowner(LIST_FIRST(&clp->cl_perclient));
	LIST_REMOVE(clp, cl_idhash);
	LIST_REMOVE(clp, cl_strhash);
	TAILQ_REMOVE(&client_lru, clp, cl_lru);
	free_client(clp);
}

time_t
nfs4_laundromat(void)
{
	struct nfs4_client *clp;
	struct nfs4_lockowner *lp;
	time_t cutoff = get_time() - NFS4_LEASE_TIME;
	time_t t;
	time_t return_val = NFS4_LEASE_TIME;
	
	nfs4_lock_state();
	
#if 0
	/* Do we want to destroy all state for the client, or just its open file
	 * state?  Under the current version of the spec, a client without any files
	 * open has no way to renew its lease (RENEW takes a stateid instead of a
	 * clientid), so destroying all state for the client seems like a bad idea.
	 */
	while (!TAILQ_EMPTY(&client_lru)) {
		clp = TAILQ_FIRST(&client_lru);
		if (time_after(clp->cl_time, cutoff)) {
			t = clp->cl_time - cutoff;
			if (return_val > t)
				return_val = t;
			break;
		}
		nfsv4_printk(level2, "nfsd: purging unused client at %s (clientid %08x)\n",
			     in_ntoa(clp->cl_addr), clp->cl_clientid.cl_id);
		nfsv4_printk(level2, "nfsd: destroying client's state (clientid %08x)\n",
			     clp->cl_clientid.cl_id);
		release_client(clp);
	}
#else
	nfsv4_printk(lease, "LAUNDROMAT THREAD - examining client list\n");
	TAILQ_FOREACH(clp, &client_lru, cl_lru) {
		nfsv4_printk(lease, "LAUNDROMAT THREAD - examining client\n");
		if (time_after(clp->cl_time, cutoff)) {
			nfsv4_printk(lease, "LAUNDROMAT THREAD - client will not be purged\n");
			if (LIST_EMPTY(&clp->cl_perclient)) {
				nfsv4_printk(lease, "LAUNDROMAT THREAD - no lockowners\n");
				continue;
			}
			t = clp->cl_time - cutoff;
			if (return_val > t)
				return_val = t;
			break;
		}
		nfsv4_printk(level2, "nfsd: purging unused client at %s (clientid %08x)\n",
			     in_ntoa(clp->cl_addr), clp->cl_clientid.cl_id);
		nfsv4_printk(level2, "nfsd: destroying client's state (clientid %08x)\n",
			     clp->cl_clientid.cl_id);

		nfsv4_printk(lease, "LAUNDROMAT THREAD - examining client's lockowners\n");
		while (!LIST_EMPTY(&clp->cl_perclient)) {
			nfsv4_printk(lease, "LAUNDROMAT THREAD - releasing lockowner\n");
			release_lockowner(LIST_FIRST(&clp->cl_perclient));
		}
	}
#endif

	nfsv4_printk(lease, "LAUNDROMAT THREAD - examining lockowner list\n");
	while (!TAILQ_EMPTY(&lockowner_unused_lru)) {
		nfsv4_printk(lease, "LAUNDROMAT THREAD - examining lockowner\n");
		lp = TAILQ_FIRST(&lockowner_unused_lru);
#if NFS4_DEBUG
		if (!LOCKOWNER_UNUSED(lp)) {
			printk("!!! nfsd4_laundromat: lockowner on lru list in use !!!\n");
			break;
		}
#endif
		if (time_after(lp->lo_time, cutoff)) {
			nfsv4_printk(lease, "LAUNDROMAT THREAD: lockowner not released\n");
			t = lp->lo_time - cutoff;
			if (return_val > t)
				return_val = t;
			break;
		}
		nfsv4_printk(lease, "nfsd: purging unused lockowner (clientid %08x, "
			     "lockowner %08x\n", lp->lo_client->cl_clientid.cl_id, lp->lo_id);
		release_lockowner(lp);
	}
	if (return_val < NFS4_LAUNDROMAT_MINTIMEOUT)
		return_val = NFS4_LAUNDROMAT_MINTIMEOUT;
	
	nfs4_unlock_state();
	return return_val;
}

/*
 * nfs4_state_init(): called to initialize data structures, currently called when the kernel
 *                    module is loaded.
 * nfs4_state_shutdown(): called to free data structures, currently called when the kernel
 *                        modules is unloaded.
 * nfs4_start_grace(): called to start the grace period; if this is never called, server
 *                     won't use a grace period.  
 */

void
nfs4_state_init(void)
{
	int i;

	for (i = 0; i < CLIENT_HASH_SIZE; i++) {
		LIST_INIT(&clientid_hashtbl[i]);
		LIST_INIT(&clientstr_hashtbl[i]);
	}
	for (i = 0; i < LOCKOWNER_HASH_SIZE; i++) {
		LIST_INIT(&lockownerid_hashtbl[i]);
		LIST_INIT(&lockownerstr_hashtbl[i]);
	}
	for (i = 0; i < FILE_HASH_SIZE; i++)
		LIST_INIT(&file_hashtbl[i]);
	for (i = 0; i < OPENFILE_HASH_SIZE; i++)
		LIST_INIT(&openfile_hashtbl[i]);
	memset(&zero_stateid, 0, sizeof(stateid_t));
	memset(&one_stateid, ~0, sizeof(stateid_t));
	TAILQ_INIT(&client_lru);
	TAILQ_INIT(&lockowner_unused_lru);
	init_MUTEX(&big_sema);
}

static void
__nfs4_state_shutdown(void)
{
	int i;
	struct nfs4_client *clp;
	
	for (i = 0; i < CLIENT_HASH_SIZE; i++) {
		while (!LIST_EMPTY(&clientid_hashtbl[i])) {
			clp = LIST_FIRST(&clientid_hashtbl[i]);
			release_client(clp);
		}
	}
}

void
nfs4_state_shutdown(void)
{
	nfs4_lock_state();
	__nfs4_state_shutdown();
	nfs4_unlock_state();
}

void
nfs4_state_restart(int grace_period)
{
	time_t t = get_time();

	nfs4_lock_state();
	__nfs4_state_shutdown();
	boot_time = t;
	grace_end = grace_period ? (t + NFS4_LEASE_TIME) : 0;
	nfs4_unlock_state();
}

int
nfs4_in_grace_period(void)
{
	return time_before(get_time(), grace_end);
}

void
nfs4_set_verifier(char *verf)
{
	memcpy(verf, &boot_time, 4);
	memset(verf+4, 0, 4);
}

static inline void
renew_client(struct nfs4_client *clp)
{
	/*
	 * Move client to the end of the LRU list.
	 */
	nfsv4_printk(clientid, "renewing client %s (clientid %08x/%08x)\n",
		     in_ntoa(clp->cl_addr), clp->cl_clientid.cl_boot, clp->cl_clientid.cl_id);
	TAILQ_REMOVE(&client_lru, clp, cl_lru);
	TAILQ_INSERT_TAIL(&client_lru, clp, cl_lru);
	clp->cl_time = get_time();
}

/*
 * Here are the routines for actual handling of SETCLIENTID, SETCLIENTID_CONFIRM, RENEW,
 * OPEN, OPEN_CONFIRM, OPEN_DOWGRADE, CLOSE, (truncating) SETATTR, READ, WRITE.  For details
 * on how to use them, see the long comment toward the end of state.h.
 */

int
nfs4_process_setclientid(struct global_state *stp, SETCLIENTID4args *argp,
			 SETCLIENTID4res *resp)
{
	u32 ip_addr = stp->rqstp->rq_addr.sin_addr.s_addr;
	unsigned int namelen = argp->client.id.id_len;
	char *name = argp->client.id.id_val;
	char *verifier = argp->client.verifier;
	unsigned int strhashval;
	unsigned int idhashval;
	struct nfs4_client *clp;
	int status;

	/*
	 * XXX: Callback data is ignored for now.
	 * XXX: figure out cryptic "principal/uid" comments in spec..
	 */
	
	status = NFS4ERR_INVAL;
	if (namelen > NFSD4_CLIENT_MAXNAME) {
		nfsv4_printk(clientid2, "process_setclientid: name too long(%d)!\n", namelen);
		goto out;
	}
	nfsv4_printk(clientid, "process_setclientid: starting, ip=%s, client string=\n",
		     in_ntoa(ip_addr));
	nfsv4_hexdump(clientid, name, namelen);
	nfsv4_printk(clientid, "process_setclientid: verifier=\n");
	nfsv4_hexdump(clientid, verifier, 8);
	
	strhashval = clientstr_hashval(name, namelen);
	LIST_FOREACH(clp, &clientstr_hashtbl[strhashval], cl_strhash) {
		if ((clp->cl_namelen != namelen) || memcmp(clp->cl_name, name, namelen))
			continue;
		/*
		 * If we get here, we already have a record for this client.  If the
		 * verifiers and IP addresses agree, assume this a retransmission, return
		 * success.  If the IP addresses disagree, return ERR_CLID_INUSE.  If the
		 * IP addresses agree but the verifier is different, assume the client has
		 * rebooted, release the old clientid, and return a new one.  This presents
		 * a problem if two clients behind the same NAT firewall use the same client
		 * string, so it is up to the clients to ensure that this doesn't happen.
		 * (This also means that an IP-spoofing client can destroy another's
		 * client state if it can guess its client string, but I don't see any way to
		 * avoid this.  Besides, if a client is really worried about this happening,
		 * it can append a random value to its client string and encrypt its session
		 * with the server.)
		 */
		status = NFS4ERR_CLID_INUSE;
		if (clp->cl_addr != ip_addr) {
			nfsv4_printk(clientid2, "process_setclientid: string in use by client"
				     " %s (clientid %08x/%08x)\n", in_ntoa(clp->cl_addr),
				     clp->cl_clientid.cl_boot, clp->cl_clientid.cl_id);
			goto out;
		}
		if (!memcmp(clp->cl_verifier, verifier, sizeof(verifier4))) {
			nfsv4_printk(clientid2, "process_setclientid: retransmission by client"
				     " %s (clientid %08x/%08x)\n", in_ntoa(clp->cl_addr),
				     clp->cl_clientid.cl_boot, clp->cl_clientid.cl_id);
			goto success;
		}
		nfsv4_printk(clientid2, "process_setclientid: reboot of client %s (clientid "
			     "%08x/%08x)\n", in_ntoa(clp->cl_addr), clp->cl_clientid.cl_boot,
			     clp->cl_clientid.cl_id);
		release_client(clp);
		/* fall through... */
	}

	/*
	 * No record exists for this client, so create a new client structure.
	 */
	status = NFS4ERR_RESOURCE;
	if (!(clp = alloc_client(name, namelen))) {
		nfsv4_printk(clientid2, "process_setclientid: couldn't alloc new client!\n");
		goto out;
	}
	idhashval = clientid_hashval(current_clientid);
	LIST_INSERT_HEAD(&clientid_hashtbl[idhashval], clp, cl_idhash);
	LIST_INSERT_HEAD(&clientstr_hashtbl[strhashval], clp, cl_strhash);
	TAILQ_INSERT_TAIL(&client_lru, clp, cl_lru);
	LIST_INIT(&clp->cl_perclient);
	clp->cl_clientid.cl_boot = boot_time;
	clp->cl_clientid.cl_id = current_clientid;
	clp->cl_addr = ip_addr;
	memcpy(&clp->cl_verifier, verifier, sizeof(verifier4));
	clp->cl_confirmed = 0;
	clp->cl_time = get_time();
	current_clientid++;

success:
	nfsv4_printk(clientid, "process_setclientid: returning clientid %08x/%08x\n",
		     clp->cl_clientid.cl_boot, clp->cl_clientid.cl_id);
	memcpy(&resp->CLIDOK.clientid, &clp->cl_clientid, sizeof(clientid_t));
	memcpy(resp->CLIDOK.setclientid_confirm, &clp->cl_clientid, sizeof(verifier4));
	status = NFS4_OK;
	
out:
	return status;
}

int
nfs4_process_setclientid_confirm(struct global_state *stp, clientid_t *clientid)
{
	u32 ip_addr = stp->rqstp->rq_addr.sin_addr.s_addr;
	unsigned int hashval;
	struct nfs4_client *clp;
	int status;

	nfsv4_printk(clientid, "process_setclientid_confirm(%08x/%08x): starting\n",
		     clientid->cl_boot, clientid->cl_id);
	
	status = NFS4ERR_STALE_CLIENTID;
	if (STALE_CLIENTID(clientid))
		goto out;

	hashval = clientid_hashval(clientid->cl_id);
	LIST_FOREACH(clp, &clientid_hashtbl[hashval], cl_idhash) {
		if (clp->cl_clientid.cl_id != clientid->cl_id )
			continue;
		/*
		 * We found a record for this clientid; if the IP addresses match, return
		 * success.  Otherwise, return ERR_INVAL (just as if the record had not
		 * been found).  We don't return an error if the clientid is already
		 * confirmed; this could be the normal result of a retransmission.
		 */
		if (clp->cl_addr != ip_addr) {
			nfsv4_printk(clientid2, "process_setclientid_confirm: bad ip addr!\n");
			goto not_found;
		}
		nfsv4_printk(clientid, "process_setclientid_confirm: success\n");
		clp->cl_confirmed = 1;
		renew_client(clp);
		status = NFS4_OK;
		goto out;
	}

not_found:
	/*
	 * Couldn't find a record for this clientid.  Presumably this is because the
	 * client took too long between SETCLIENTID and SETCLIENTID_CONFIRM, so return
	 * NFS4ERR_EXPIRED.
	 */
	nfsv4_printk(clientid2, "process_setclientid_confirm: clientid not found!\n");
	status = NFS4ERR_EXPIRED;

out:
	return status;
}

int
nfs4_process_renew(struct global_state *stp, stateid_t *stateid)
{
	stateid_lockowner_t l_id = stateid->st_lockowner;
	stateid_file_t f_id = stateid->st_file;
	stateid_generation_t generation = stateid->st_generation;
	struct nfs4_openfile *ofp;
	stateid_generation_t current_generation;
	int status;

	/*
	 * RENEW currently takes a stateid; this code can be simplified once the
	 * NFSv4 gods decide that it should take a clientid instead (this is
	 * supposedly in the works).
	 */
	nfsv4_printk(stateid, "nfs4_renew: starting, stateid=(%08x/%08x/%08x/%08x)\n",
		     stateid->st_boot, stateid->st_lockowner, stateid->st_file,
		     stateid->st_generation);
	
	status = NFS4ERR_BAD_STATEID;
	if (ZERO_STATEID(stateid) || ONE_STATEID(stateid)) {
		nfsv4_printk(lease2, "process_renew: magic stateid!\n");
		goto out;
	}
	
	status = NFS4ERR_STALE_STATEID;
	if (STALE_STATEID(stateid)) {
		nfsv4_printk(lease2, "process_renew: stale stateid!\n");
		goto out;
	}

	status = NFS4ERR_BAD_STATEID;
	if (!(ofp = lookup_openfile(openfile_hashval(l_id, f_id), l_id, f_id))) {
		nfsv4_printk(lease2, "process_renew: lookup_openfile() failed!\n");
		goto out;
	}

	current_generation = ofp->of_stateid.st_generation;
	if (generation == current_generation) {
		nfsv4_printk(lease, "process_renew: success\n");
		renew_client(ofp->of_lockowner->lo_client);
		status = NFS4_OK;
	}
	else if (generation > current_generation) {
		nfsv4_printk(lease2, "process_renew: future stateid?!\n");
		status = NFS4ERR_BAD_STATEID;
	}
	else {
		nfsv4_printk(lease2, "process_renew: old stateid!\n");
		status = NFS4ERR_OLD_STATEID;
	}

out:
	return status;
}

int
nfs4_process_open1(struct global_state *stp, clientid_t *clientid,
		   char *name, unsigned int namelen, seqid4 seqid)
{
	struct nfs4_client *clp;
	struct nfs4_lockowner *lp;
	unsigned int cl_hashval;
	unsigned int lo_strhashval;
	unsigned int lo_idhashval;
	int status;

	status = NFS4ERR_INVAL;
	if (namelen > NFSD4_LOCKOWNER_MAXNAME) {
		nfsv4_printk(open2, "process_open: lockowner name too long!\n");
		goto out;
	}

	nfsv4_printk(clientid, "process_open: starting, clientid=(%08x/%08x)\n",
		     clientid->cl_boot, clientid->cl_id);
	nfsv4_printk(lockowner, "process_open: starting, seqid=%d, lockowner=\n", seqid);
	nfsv4_hexdump(lockowner, name, namelen);
	
	status = NFS4ERR_STALE_CLIENTID;
	if (STALE_CLIENTID(clientid)) {
		nfsv4_printk(open2, "process_open: clientid(%08x/%08x) is stale!\n",
			     clientid->cl_boot, clientid->cl_id);
		goto out;
	}

	lo_strhashval = lockownerstr_hashval(clientid->cl_id, name, namelen);
	
	/*
	 *  First, try to look up the lockowner, using the string specified by the client.
	 */
	LIST_FOREACH(lp, &lockownerstr_hashtbl[lo_strhashval], lo_strhash) {
		if ((lp->lo_namelen != namelen) || memcmp(lp->lo_name, name, namelen) ||
		    (lp->lo_client->cl_clientid.cl_id != clientid->cl_id))
			continue;
		nfsv4_printk(lockowner, "process_open: lockowner id=%08x\n", lp->lo_id);
		stp->current_lockowner = lp;
		if (lp->lo_confirmed) {
			if (seqid == lp->lo_seqid + 1) {
				nfsv4_printk(lockowner, "process_open: correct seqid %d\n",
					     seqid);
				status = NFS4_OK;
				goto renew;
			}
			status = NFS4ERR_BAD_SEQID;
			nfsv4_printk(open2, "process_open: bad seqid (got %d, expected %d)!\n",
				     seqid, lp->lo_seqid + 1);
			goto out;
		}
		/* If we get here, we received an OPEN for an unconfirmed lockowner.  If
		 * the seqid is the same as the one we originally received, assume that this
		 * is a retransmission, and replay.  If not, then according to spec, we should
		 * assume that the original OPEN was a replay, and purge the original OPEN
		 * before instantiating this one.   However, doesn't this leave us open to
		 * a race condition where:
		 *   1. Client sends an OPEN, server requires OPEN_CONFIRM in reply
		 *   2. Retransmitted OPEN arrives, server purges first OPEN
		 *   3. Client sends OPEN_CONFIRM, which mysteriously fails.
		 */
		if (lp->lo_seqid == seqid) {
			nfsv4_printk(open2, "process_open: seqids equal, retransmission?");
			if (LIST_EMPTY(&lp->lo_perlockowner)) {
				/* oops! the original OPEN failed, which means that we
				 * don't have replay data saved (we are currently only
				 * increasing the seqid on a successful operation).
				 * therefore, we have no choice but to continue processing
				 * this OPEN; presumably, we'll fail again for the same
				 * reason.  (note that this leaves us open to race conditions
				 * of the type pointed out by neil brown on the nfsv4-wg
				 * list)
				 */
				nfsv4_printk(level1, "nfs4_open: warning: weird code path!\n");
				lp->lo_time = get_time();   /* small plug in a large hole */
				status = NFS4_OK;
				goto renew;
			}
			status = NFS4ERR_REPLAY_ME;
			goto out;
		}
		nfsv4_printk(lockowner2, "process_open: purging unconfirmed lockowner!\n");
		clp = lp->lo_client;
		release_lockowner(lp);     /* safest; invalidates old stateid */
		goto instantiate_new_lockowner;
	}
	
	/*
	 * Lockowner not found, so we need to instantiate a new one.  First, we need to
	 * validate the given clientid.
	 */
	cl_hashval = clientid_hashval(clientid->cl_id);
	LIST_FOREACH(clp, &clientid_hashtbl[cl_hashval], cl_idhash) {
		if (clp->cl_clientid.cl_id != clientid->cl_id)
			continue;
		if (clp->cl_addr != stp->rqstp->rq_addr.sin_addr.s_addr) {   /* paranoid? */
			nfsv4_printk(open2, "process_open: bad ip address!\n");
			break;
		}
		status = NFS4ERR_INVAL;    /* XXX: right error? */
		if (!clp->cl_confirmed) {
			nfsv4_printk(open2, "process_open: client is unconfirmed!\n");
			goto out;
		}
		/* successfully validated clientid */
		goto instantiate_new_lockowner;
	}
	
	/*
	 * Couldn't validate the clientid (or IP address mismatch).  Presumably, this is
	 * the result of the client's lease expiring, so we inform the client of this
	 * by sending NFS4ERR_EXPIRED.
	 */
	nfsv4_printk(open2, "process_open: clientid not found!\n");
	status = NFS4ERR_EXPIRED;
	goto out;

	/*
	 * Instantiate the new lockowner. XXX: we currently require an OPEN_CONFIRM whenever
	 * a lockowner is used for the first time!  XXX: to ask the list: what timeout before
	 * we can stop requiring OPEN_CONFIRM?
	 */
instantiate_new_lockowner:
	nfsv4_printk(lockowner, "process_open: instantiating new lockowner, id=%08x\n",
		     current_lockownerid);
	status = NFS4ERR_RESOURCE;
	if (!(lp = alloc_lockowner(name, namelen))) {
		nfsv4_printk(level1, "process_open: couldn't alloc new lockowner!\n");
		goto renew;
	}
	lo_idhashval = lockownerid_hashval(current_lockownerid);
	LIST_INSERT_HEAD(&lockownerid_hashtbl[lo_idhashval], lp, lo_idhash);
	LIST_INSERT_HEAD(&lockownerstr_hashtbl[lo_strhashval], lp, lo_strhash);
	TAILQ_INSERT_TAIL(&lockowner_unused_lru, lp, lo_lru);
	LIST_INSERT_HEAD(&clp->cl_perclient, lp, lo_perclient);	
	LIST_INIT(&lp->lo_perlockowner);
	lp->lo_client = clp;
	lp->lo_boot = boot_time;
	lp->lo_id = current_lockownerid;
	lp->lo_seqid = seqid;
	lp->lo_hashval = lo_idhashval;
	lp->lo_confirmed = 0;
	lp->lo_time = get_time();
	lp->lo_replay_buflen = 0;
	lp->lo_replay_buf = lp->lo_replay_ibuf;
	lp->lo_replay_allocated = sizeof(lp->lo_replay_ibuf);
	stp->current_lockowner = lp;
	current_lockownerid++;
	status = NFS4_OK;

renew:
	renew_client(lp->lo_client);
out:
	return status;
}

int
nfs4_process_open2(struct global_state *stp, nfs4_ino_desc_t *ino, int truncate,
		  unsigned int share_access, unsigned int share_deny, seqid4 seqid)
{
	struct nfs4_lockowner *lp = stp->current_lockowner;
	struct nfs4_file *fp;
	struct nfs4_openfile *ofp = NULL;
	struct nfs4_openfile *ofq;
	unsigned int fi_hashval;
	unsigned int of_hashval;
	int status;

#ifdef __linux__
	nfsv4_printk(open, "process_open: (ino,gen)=(%d,%ld,%d), access=%08x, deny=%08x\n",
		     ino->dev, ino->ino, ino->generation, share_access, share_deny);
#endif
	
	fi_hashval = file_hashval(ino);
	LIST_FOREACH(fp, &file_hashtbl[fi_hashval], fi_hash) {
#if 0
		/* XXX: bakeoff: why doesn't this work?! */
		if (!memcmp(&fp->fi_ino, ino, sizeof(nfs4_ino_desc_t))) {
#else
		if (fp->fi_ino.dev == ino->dev &&
		    fp->fi_ino.ino == ino->ino &&
		    fp->fi_ino.generation == ino->generation) {
#endif
			status = NFS4ERR_SHARE_DENIED;
			nfsv4_printk(stateid, "process_open: fileid=%08x\n", fp->fi_id);
			LIST_FOREACH(ofq, &fp->fi_perfile, of_perfile) {
				if (ofq->of_lockowner != lp) {
					if ((ofq->of_share_access & share_deny) ||
					    (ofq->of_share_deny & share_access)) {
						nfsv4_printk(open2, "process_open: conflicting"
							     " share reservation!\n");
						goto out;
					}
				}
				else
					ofp = ofq;
			}
			goto truncate;
		}
	}

	/* No nfs4_file found in table; allocate a new one...
	 */
	nfsv4_printk(stateid, "process_open: allocating new file, id=%08x\n", current_fileid);
	status = NFS4ERR_RESOURCE;
	if (!(fp = alloc_file())) {
		nfsv4_printk(open2, "process_open: couldn't alloc nfs4_file!\n");
		goto out;
	}
	LIST_INSERT_HEAD(&file_hashtbl[fi_hashval], fp, fi_hash);
	LIST_INIT(&fp->fi_perfile);
	memcpy(&fp->fi_ino, ino, sizeof(nfs4_ino_desc_t));
	fp->fi_id = current_fileid++;
	
truncate:	
	if (truncate && (status = nfs4_file_truncate(&stp->current_fh))) {
		nfsv4_printk(open2, "process_open: truncation failed!\n");
		goto out;
	}
	if (!ofp) {
		nfsv4_printk(stateid, "process_open: allocating new openfile\n");
		status = NFS4ERR_RESOURCE;
		if (!(ofp = alloc_openfile()))
			goto out;
		if ((status = nfs4_file_open(&ofp->of_vfs_file,
					     &stp->current_fh, share_access)))
			goto out_free;
       		/*
		 * If opening this file brings an unused lockowner back into use, remove it
		 * from the "unused_lru" list.
		 */
		if (lp->lo_confirmed && LIST_EMPTY(&lp->lo_perlockowner))
			TAILQ_REMOVE(&lockowner_unused_lru, lp, lo_lru);
		of_hashval = openfile_hashval(lp->lo_id, fp->fi_id);
		LIST_INSERT_HEAD(&openfile_hashtbl[of_hashval], ofp, of_hash);
		LIST_INSERT_HEAD(&lp->lo_perlockowner, ofp, of_perlockowner);
		LIST_INSERT_HEAD(&fp->fi_perfile, ofp, of_perfile);
		ofp->of_lockowner = lp;
		ofp->of_file = fp;
		ofp->of_stateid.st_boot = boot_time;
		ofp->of_stateid.st_lockowner = lp->lo_id;
		ofp->of_stateid.st_file = fp->fi_id;
		ofp->of_stateid.st_generation = 0;
		ofp->of_share_access = share_access;
		ofp->of_share_deny = share_deny;
	}
	else {
		/* This is an upgrade of an existing OPEN.
		 * XXX: Assumption: an OPEN "upgrade" for which the specified bits do not
		 * include the currently-held ones is legal; the bits are simply bitwise-or'ed
		 * together.
		 */
		nfsv4_printk(stateid, "process_open: using existing openfile\n");
		share_access &= ~(ofp->of_share_access);
		if ((status = nfs4_file_upgrade(&ofp->of_vfs_file, share_access)))
			goto out;
		ofp->of_share_access |= share_access;
		ofp->of_share_deny |= share_deny;
		update_stateid(&ofp->of_stateid);
	}
	
	/*
	 * We are currently using the convention that the seqid is only updated if the
	 * operation is successful.
	 */
	nfsv4_printk(lockowner, "process_open: updating seqid to %d\n", seqid);
	nfsv4_printk(stateid, "process_open: stateid=(%08x/%08x/%08x/%08x)\n",
		     ofp->of_stateid.st_boot, ofp->of_stateid.st_lockowner,
		     ofp->of_stateid.st_file, ofp->of_stateid.st_generation);
	status = NFS4_OK;
	stp->current_openfile = ofp;
	lp->lo_seqid = seqid;
	
out:
	return status;
out_free:
	free_openfile(ofp);
	goto out;
}

int
nfs4_process_open_confirm(struct global_state *stp, seqid4 seqid, open_verifier_t *verifier)
{
	unsigned int lo_hashval;
	struct nfs4_lockowner *lp;
	struct nfs4_openfile *ofp;
	int status;

	nfsv4_printk(lockowner, "process_open_confirm: verifier=(%08x/%08x)\n",
		     verifier->ov_boot, verifier->ov_id);
	
	/*
	 * If the verifier is stale, we want to return an error to tell the client that
	 * the server has rebooted between the OPEN and the OPEN_CONFIRM.  ERR_STALE_CLIENTID
	 * or ERR_STALE_STATEID don't make sense (since OPEN_CONFIRM doesn't take either a
	 * clientid or stateid), so the only reasonable choice here is ERR_GRACE (even though
	 * we might not actually be in the grace period!)
	 */
	status = NFS4ERR_GRACE;
	if (STALE_VERIFIER(verifier)) {
		nfsv4_printk(open2, "process_open_confirm: verifier is stale!\n");
		goto out;
	}

	lo_hashval = lockownerid_hashval(verifier->ov_id);
	LIST_FOREACH(lp, &lockownerid_hashtbl[lo_hashval], lo_idhash) {
		if (lp->lo_id != verifier->ov_id)
			continue;
		stp->current_lockowner = lp;
		
		if (!lp->lo_confirmed) {
			if (seqid == lp->lo_seqid + 1) {
				ofp = LIST_FIRST(&lp->lo_perlockowner);
				if ((status = nfs4_check_fh(&stp->current_fh, ofp))) {
					nfsv4_printk(open2, "process_open_confirm: "
						     "fh-stateid mismatch!\n");
					goto out;
				}
				/*
				 * Successful OPEN_CONFIRM; the unconfirmed lockowner
				 * should be unused, and have only one associated openfile.
				 */
				TAILQ_REMOVE(&lockowner_unused_lru, lp, lo_lru);
				lp->lo_confirmed = 1;
				lp->lo_seqid = seqid;
				update_stateid(&ofp->of_stateid);    /* XXX: unnecessary? */
				stp->current_openfile = ofp;
				renew_client(lp->lo_client);
				nfsv4_printk(lockowner, "open_confirm succeeded, seqid=%d\n",
					     seqid);
				nfsv4_printk(stateid, "open_confirm succeeded, "
					     "stateid=(%08x/%08x/%08x/%08x)\n",
					     ofp->of_stateid.st_boot,
					     ofp->of_stateid.st_lockowner,
					     ofp->of_stateid.st_file,
					     ofp->of_stateid.st_generation);
				status = NFS4_OK;
				goto out;
			}
			nfsv4_printk(open2, "process_open_confirm: bad seqid! "
				     "(expected %d, got %d)\n", lp->lo_seqid + 1, seqid);
			status = NFS4ERR_BAD_SEQID;
		}
		else {
			if (seqid == lp->lo_seqid) {
				/* Replayed OPEN_CONFIRM. */
				nfsv4_printk(open2, "process_open_confirm: retransmission?\n");
				status = NFS4ERR_REPLAY_ME;
			}
			nfsv4_printk(open2, "process_open_confirm: already confirmed!\n");
			status = NFS4ERR_INVAL;
		}
		goto out;
	}

	/*
	 * Lockowner not found; presumably this means that the client took too long between
	 * OPEN and OPEN_CONFIRM..
	 */
	nfsv4_printk(open2, "process_open_confirm: lockowner not found!\n");
	status = NFS4ERR_EXPIRED;

out:
	return status;
}

int
nfs4_preprocess_seqid_op(struct global_state *stp, seqid4 seqid, stateid_t *stateid)
{
	stateid_lockowner_t l_id = stateid->st_lockowner;
	stateid_file_t f_id = stateid->st_file;
	unsigned int lo_hashval;
	unsigned int of_hashval;
	struct nfs4_openfile *ofp;
	struct nfs4_lockowner *lp;
	int status;

	nfsv4_printk(stateid, "stateid = (%08x/%08x/%08x/%08x)\n",
		     stateid->st_boot, stateid->st_lockowner, stateid->st_file,
		     stateid->st_generation);
	nfsv4_printk(lockowner, "lockownerid=%08x, seqid=%d\n", stateid->st_lockowner, seqid);
	
	status = NFS4ERR_BAD_STATEID;
	if (ZERO_STATEID(stateid) || ONE_STATEID(stateid)) {
		nfsv4_printk(level2, "preprocess_seqid_op: magic stateid!\n");
		goto out;
	}
	status = NFS4ERR_STALE_STATEID;
	if (STALE_STATEID(stateid)) {
		nfsv4_printk(level2, "preprocess_seqid_op: stale stateid!\n");
		goto out;
	}

	/*
	 * We return BAD_STATEID if filehandle doesn't match stateid.  If there is no
	 * entry in the openfile table for this id, we can't always return BAD_STATEID;
	 * this might be a retransmitted CLOSE which has arrived after the openfile has
	 * been released.
	 */
	of_hashval = openfile_hashval(l_id, f_id);
	if (!(ofp = lookup_openfile(of_hashval, l_id, f_id)))
		goto no_openfile;
	if ((status = nfs4_check_fh(&stp->current_fh, ofp))) {
		nfsv4_printk(level2, "nfs4_check_fh: fh-stateid mismatch!\n");
		goto out;
	}
	stp->current_openfile = ofp;
	stp->current_lockowner = lp = ofp->of_lockowner;
	
	/*
	 *  We now validate the seqid and stateid generation numbers.
	 *  For the moment, we ignore the possibility of generation number wraparound.
	 */
	if (seqid != lp->lo_seqid + 1)
		goto check_replay;
	status = NFS4ERR_BAD_STATEID;
	if (stateid->st_generation > ofp->of_stateid.st_generation) {
		nfsv4_printk(level2, "preprocess_seqid_op: future stateid?!\n");
		goto out;
	}
	status = NFS4ERR_OLD_STATEID;
	if (stateid->st_generation < ofp->of_stateid.st_generation) {
		nfsv4_printk(level2, "preprocess_seqid_op: old stateid!\n");
		goto out;
	}
	renew_client(lp->lo_client);
	status = NFS4_OK;

out:
	return status;
	
no_openfile:
	/*
	 * We determine whether this is a bad stateid or a replay, starting by trying
	 * to look up the lockowner.
	 */
	lo_hashval = lockownerid_hashval(l_id);
	LIST_FOREACH(lp, &lockownerid_hashtbl[lo_hashval], lo_idhash) {
		if (lp->lo_id != l_id)
			continue;
		stp->current_lockowner = lp;
		/*
		 * If the seqid is correct for this lockowner, but no entry is found in
		 * the openfile table, what error should we return?  This probably means
		 * that the client is trying to do I/O on a file which has been closed;
		 * BAD_STATEID seems like the most reasonable choice?
		 */
		status = NFS4ERR_BAD_STATEID;
		if (seqid == lp->lo_seqid + 1) {
			nfsv4_printk(level2, "preprocess_seqid_op: file has been released!\n");
			goto out;
		}
		goto check_replay;
	}
	/*
	 * Yikes, couldn't find a lockowner or an openfile to match this stateid.
	 * XXX: any chance we should return ERR_EXPIRED instead of ERR_BAD_STATEID?
	 */
	nfsv4_printk(level2, "preprocess_seqid_op: no lockowner or openfile!\n");
	status = NFS4ERR_BAD_STATEID;
	goto out;
	
check_replay:
	status = NFS4ERR_BAD_SEQID;
	if (seqid == lp->lo_seqid) {
		nfsv4_printk(level2, "preprocess_seqid_op: retransmission?\n");
		status = NFS4ERR_REPLAY_ME;
	}
	else
		nfsv4_printk(level2, "preprocess_seqid_op: bad seqid (expected %d, got %d)!\n",
			     lp->lo_seqid + 1, seqid);
	goto out;
}

int
nfs4_process_open_downgrade(struct global_state *stp, seqid4 seqid, stateid_t *stateid,
			    unsigned int share_access, unsigned int share_deny)
{
	unsigned int current_access;
	unsigned int current_deny;
	struct nfs4_openfile *ofp;
	struct nfs4_lockowner *lp;
	int status;

	/*  XXX: currently, an OPEN_DOWNGRADE to access=0 and deny=0 keeps the file "open",
	 *  but disallows all I/O until an upgrade is done.  Is this correct?
	 */
	if ((status = nfs4_preprocess_seqid_op(stp, seqid, stateid)))
		goto out;
	lp = stp->current_lockowner;
	ofp = stp->current_openfile;
	current_access = ofp->of_share_access;
	current_deny = ofp->of_share_deny;

	status = NFS4ERR_INVAL;
	if (share_access & ~current_access) {
		nfsv4_printk(open2, "open_downgrade: access not a subset! (current=%08x, desired=%08x)\n",
			     current_access, share_access);
		goto out;
	}
	if (share_deny & ~current_deny) {
		nfsv4_printk(open2, "open_downgrade: deny not a subset! (current=%08x, desired=%08x)\n",
			     current_deny, share_deny);
		goto out;
	}
	nfs4_file_downgrade(&ofp->of_vfs_file, current_access & ~share_access);
	ofp->of_share_access = share_access;
	ofp->of_share_deny = share_deny;

	/*
	 *  Return success, but first update the seqid and stateid.
	 */
	lp->lo_seqid = seqid;
	update_stateid(&ofp->of_stateid);
	status = NFS4_OK;
	
out:
	return status;
}

int
nfs4_process_close(struct global_state *stp, seqid4 seqid, stateid_t *stateid)
{
	struct nfs4_openfile *ofp;
	struct nfs4_lockowner *lp;
	int status;

	if ((status = nfs4_preprocess_seqid_op(stp, seqid, stateid)))
		goto out;
	lp = stp->current_lockowner;
	ofp = stp->current_openfile;

	/* release_openfile() calls nfs4_file_close() */
	release_openfile(ofp);

	/*
	 *  Return success, but first update the seqid and stateid.
	 */
	lp->lo_seqid = seqid;
	update_stateid(&ofp->of_stateid);
	status = NFS4_OK;
	
out:
	return status;
}

int
nfs4_preprocess_stateid_op(struct global_state *stp, stateid_t *stateid)
{
	stateid_lockowner_t l_id = stateid->st_lockowner;
	stateid_file_t f_id = stateid->st_file;
	stateid_generation_t generation = stateid->st_generation;
	stateid_generation_t current_generation;
	unsigned int of_hashval;
	struct nfs4_openfile *ofp;
	struct nfs4_lockowner *lp;
	int status;
	
#if NFS4_DEBUG
	/*
	 *  The "special stateids" should have been trapped before reaching this point.
	 */
	status = NFS4ERR_SERVERFAULT;
	if (ZERO_STATEID(stateid) || ONE_STATEID(stateid)) {
		printk("oops - magic stateid in preprocess_stateid!\n");
		goto out;
	}
#endif

	/*
	 *  XXX: should we return ERR_GRACE during the grace period?  The spec says so,
	 *  but I don't see any harm in allowing I/O during the grace period...
	 */
	
	status = NFS4ERR_STALE_STATEID;
	if (STALE_STATEID(stateid)) {
		nfsv4_printk(level2, "process_stateid: stateid is stale!\n");
		goto out;
	}

	/*
	 *  We return BAD_STATEID if filehandle doesn't match stateid, or there is no
	 *  entry in the openfile table for this id.  
	 */
	of_hashval = openfile_hashval(l_id, f_id);
	status = NFS4ERR_BAD_STATEID;
	if (!(ofp = lookup_openfile(of_hashval, l_id, f_id))) {
		nfsv4_printk(level2, "process_stateid: couldn't find openfile!\n");
		goto out;
	}
	stp->current_openfile = ofp;
	stp->current_lockowner = lp = ofp->of_lockowner;
	if ((status = nfs4_check_fh(&stp->current_fh, ofp)))
		goto out;

	current_generation = ofp->of_stateid.st_generation;
	status = NFS4ERR_BAD_STATEID;
	if (generation > current_generation) {
		nfsv4_printk(level2, "process_stateid: future stateid?!\n");
		goto out;
	}
	status = NFS4ERR_OLD_STATEID;
	if (generation < current_generation) {
		nfsv4_printk(level2, "process_stateid: old stateid!\n");
		goto out;
	}

	status = NFS4_OK;
	renew_client(lp->lo_client);
	
out:
	return status;
}

int
nfs4_share_conflicts(nfs4_ino_desc_t *ino, unsigned int deny_type)
{
	unsigned int fi_hashval;
	struct nfs4_file *fp;
	struct nfs4_openfile *ofp;
	int status;
	
	fi_hashval = file_hashval(ino);
	LIST_FOREACH(fp, &file_hashtbl[fi_hashval], fi_hash) {
		if (memcmp(&fp->fi_ino, ino, sizeof(nfs4_ino_desc_t)))
			continue;
		status = NFS4ERR_SHARE_DENIED;
		LIST_FOREACH(ofp, &fp->fi_perfile, of_perfile) {
			if (ofp->of_share_deny & deny_type)
				goto out;
		}
		break;
	}
	status = NFS4_OK;
	
out:
	return status;
}

int
nfs4_lookup_lockowner(struct global_state *stp, clientid_t *clientid,
		      unsigned int namelen, char *name)
{
	unsigned int hashval;
	struct nfs4_lockowner *lp;
	int status;
	
	status = NFS4ERR_STALE_CLIENTID;
	if (STALE_CLIENTID(clientid)) {
		nfsv4_printk(lock2, "process_lockt: clientid is stale!\n");
		goto out;
	}

	hashval = lockownerstr_hashval(clientid->cl_id, name, namelen);
	LIST_FOREACH(lp, &lockownerstr_hashtbl[hashval], lo_strhash) {
		if ((lp->lo_namelen == namelen) && !memcmp(lp->lo_name, name, namelen) &&
		    (lp->lo_client->cl_clientid.cl_id == clientid->cl_id))
			break;
	}
	stp->current_lockowner = lp;    /* can be NULL */
	status = NFS4_OK;

out:
	return status;
}

int
nfs4_verify_lockowner(struct nfs4_lockowner *lp, unsigned int hashval)
{
	struct nfs4_lockowner *p;
	int status = 0;
	
	if (hashval >= LOCKOWNER_HASH_SIZE)
		goto out;
	LIST_FOREACH(p, &lockownerid_hashtbl[hashval], lo_idhash) {
		if (p == lp) {
			status = 1;
			goto out;
		}
	}
out:
	return status;
}
