/*
 * linux/fs/nfsd4/nfssvc.c
 *
 * from linux/fs/nfsd/nfssvc.c
 *
 * Central processing for nfsd.
 *
 * Authors:	Olaf Kirch (okir@monad.swb.de)
 *
 * Copyright (C) 1995, 1996, 1997 Olaf Kirch <okir@monad.swb.de>
 * NFSv4 Implementation
 *    Kendrick Smith  <kmsmith@umich.edu>
 *    Andy Adamson  <andros@umich.edu>
 *    Copyright (C) 2001 The Regents of the University of Michigan
 */

#define __NO_VERSION__
#include <linux/config.h>
#include <linux/module.h>

#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/errno.h>
#include <linux/nfs.h>
#include <linux/in.h>
#include <linux/uio.h>
#include <linux/version.h>
#include <linux/unistd.h>
#include <linux/malloc.h>
#include <linux/smp.h>
#include <linux/smp_lock.h>

#include <linux/sunrpc/types.h>
#include <linux/sunrpc/stats.h>
#include <linux/sunrpc/svc.h>
#include <linux/sunrpc/svcsock.h>
#include <linux/nfs4/nfs4_prot.h>
#include <linux/nfs4/nfs4_debug.h>
#include <linux/nfs4/nfs4.h>
#include <linux/nfs4/nfsd4_state.h>
#include <linux/nfs4/nfsd4.h>
#include <linux/sunrpc/auth_gss.h>

#define NFSDDBG_FACILITY	NFSDDBG_SVC
#define NFSD_BUFSIZE		(1024 + NFSSVC_MAXBLKSIZE)

#define ALLOWED_SIGS	(sigmask(SIGKILL))
#define SHUTDOWN_SIGS	(sigmask(SIGKILL) | sigmask(SIGINT) | sigmask(SIGQUIT))

/* XXX : bakeoff hack -- needs clean up! */
extern int nfs4_dummy_compound(struct svc_rqst *rqstp, COMPOUND4args *argp,
			       COMPOUND4res *resp);
	
static void			nfsd(struct svc_rqst *rqstp);
static int			laundromat_main(void *not_used);
struct timeval			nfssvc_boot = { 0, 0 };
static struct svc_serv 		*nfsd_serv = NULL;
static int			nfsd_busy = 0;
static unsigned long		nfsd_last_call;

static DECLARE_MUTEX_LOCKED(laundromat_start_sema);
static DECLARE_WAIT_QUEUE_HEAD(laundromat_kill_queue);

static struct task_struct	*laundromat_task = NULL;
static unsigned int		nfsd_nthreads = 0;
struct nfsd_list {
	struct list_head 	list;
	struct task_struct	*task;
};
struct list_head nfsd_list = LIST_HEAD_INIT(nfsd_list);

static int
start_laundromat(void)
{
	int status;
	
#if NFS4_DEBUG
	if (laundromat_task) {
		printk("!!! start_laundromat: already running !!!\n");
		return -EINVAL;
	}
#endif
	
	nfsv4_printk(lease, "LAUNDROMAT - calling kernel_thread()\n");
	if ((status = kernel_thread(laundromat_main, NULL, 0)) < 0) {
		printk("nfsd: couldn't start laundromat thread!\n");
		goto out;
	}
	status = 0;
	down(&laundromat_start_sema);   /* wait for thread to start */

out:
	return status;
}

static void
kill_laundromat(void)
{
	wait_queue_t q;
	unsigned long flags;
	
#if NFS4_DEBUG
	if (!laundromat_task) {
		printk("!!! kill_laundromat: not running !!!\n");
		return;
	}
#endif

	nfsv4_printk(lease, "LAUNDROMAT THREAD - sending SIGKILL\n");

	/* XXX: is there a better way to add q to the wait queue?
	 * same question applies for the wait queue removal below...
	 */
	init_waitqueue_entry(&q, current);
	wq_write_lock_irqsave(&laundromat_kill_queue.lock, flags);
	__add_wait_queue(&laundromat_kill_queue, &q);
	wq_write_unlock(&laundromat_kill_queue.lock);
	current->state = TASK_INTERRUPTIBLE;

	/* XXX: make sure that there is not a race condition here if the
	 * laundromat has been killed by someone else and is blocked on the
	 * lock_kernel()...  Need to make sure that if send_sig() sleeps, it
	 * does so after all references to its second argument.
	 */
	if (!send_sig(SIGKILL, laundromat_task, 1)) {
		schedule();
		wq_write_lock_irq(&laundromat_kill_queue.lock);
		__remove_wait_queue(&laundromat_kill_queue, &q);
		wq_write_unlock_irqrestore(&laundromat_kill_queue.lock, flags);
	}
	else {
		printk(KERN_WARNING "nfsd: couldn't kill laundromat thread!\n");
		current->state = TASK_RUNNING;
		wq_write_lock_irq(&laundromat_kill_queue.lock);
		__remove_wait_queue(&laundromat_kill_queue, &q);
		wq_write_unlock_irqrestore(&laundromat_kill_queue.lock, flags);
	}
}

static int
start_nfsd_thread(void)
{
	return svc_create_thread(nfsd, nfsd_serv);
}

static int
kill_nfsd_thread(void)
{
	int status;
	struct list_head *l;
	struct nfsd_list *p;
	
	status = -EIO;
	if (list_empty(&nfsd_list)) {
		printk(KERN_WARNING "nfsd: couldn't kill thread!\n");
		goto out;
	}
	l = nfsd_list.next;
	p = list_entry(l, struct nfsd_list, list);
	list_del_init(l);
	if ((status = send_sig(SIGKILL, p->task, 1))) {
		printk(KERN_WARNING "nfsd: couldn't kill thread!\n");
		goto out;
	}
	
out:
	return status;
}

int
nfs4_start(struct nfs4ctl_start *arg)
{
	int nrservs = arg->nthreads;
	int	error;

	/* Note: caller holds BKL */

	nfsv4_printk(ctl, "nfs4_start: starting, nthreads=%d\n", nrservs);
	
	error = -EINVAL;
	if (nrservs < 0)
		goto out;
	if (nrservs > NFSD_MAXSERVS)
		goto out;

	/* Create RPC service, if it doesn't already exist */
	if (!nfsd_serv) {
		nfsv4_printk(ctl, "nfsd: svc_create\n");
		error = -ENOMEM;
		nfsd_serv = svc_create(&nfsd_program, NFSD_BUFSIZE, NFSSVC_XDRSIZE);
		if (nfsd_serv == NULL)
			goto out;
		nfsv4_printk(ctl, "nfsd: svc_makesock\n");
		error = svc_makesock(nfsd_serv, IPPROTO_UDP, NFS4_PORT);
		if (error < 0) {
			svc_destroy(nfsd_serv);
			nfsd_serv = NULL;
			goto out;
		}
	}
	
	/* Before starting the first thread, make sure state has been cycled */
	if (!nfsd_nthreads) {
		nfsv4_printk(level1, "CYCLING STATE\n");
		nfs4_state_restart(arg->grace_period);
	}

	/* Start the laundromat thread, if not already running */
	if (nrservs && !laundromat_task) {
		if ((error = start_laundromat()))
			goto out;
	}
	
	/* Spawn or kill nfsd kernel threads, as needed. */
	nrservs -= nfsd_nthreads;
	while (nrservs > 0) {
		nfsv4_printk(level1, "STARTING NFSD THREAD (%d)\n", nrservs);
		if ((error = start_nfsd_thread()))
			goto out;
		nrservs--;
	}
	while (nrservs < 0) {
		nfsv4_printk(level1, "KILLING NFSD THREAD (%d)\n", nrservs);
		if ((error = kill_nfsd_thread()))
			goto out;
		nrservs++;
	}

	/* If we have killed the last thread, go ahead and kill the laundromat thread. */
	if (!arg->nthreads && laundromat_task)
		kill_laundromat();
	error = 0;
	
out:
	return error;
}

int
nfs4_refresh(union nfsctl_res *resp)
{
	int error = 0;
	
	resp->nthreads = nfsd_nthreads;
	
	/* Bring lockowner thread into sync with the nfsd threads... */
	if (nfsd_nthreads && !laundromat_task)
		error = start_laundromat();
	else if (!nfsd_nthreads && laundromat_task)
		kill_laundromat();
		
	return error;
}

void
nfs4_svc_shutdown(void)
{
	if (nfsd_serv) {
		svc_destroy(nfsd_serv);
		nfsd_serv = NULL;
	}
}

static void inline
update_thread_usage(int busy_threads)
{
	unsigned long prev_call;
	unsigned long diff;
	int decile;

	prev_call = nfsd_last_call;
	nfsd_last_call = jiffies;
	decile = busy_threads*10/nfsdstats.th_cnt;
	if (decile>0 && decile <= 10) {
		diff = nfsd_last_call - prev_call;
		if ( (nfsdstats.th_usage[decile-1] += diff) >= NFSD_USAGE_WRAP)
			nfsdstats.th_usage[decile-1] -= NFSD_USAGE_WRAP;
		if (decile == 10)
			nfsdstats.th_fullcnt++;
	}
}

/*
 * This is the NFS server kernel thread
 */
static void
nfsd(struct svc_rqst *rqstp)
{
	struct svc_serv	*serv = rqstp->rq_server;
	int		err = 0;
	struct nfsd_list me;

	/* Lock module and set up kernel thread */
	MOD_INC_USE_COUNT;

	down_write(&nfs4_ctl_sema);
	me.task = current;
	list_add(&me.list, &nfsd_list);
	nfsd_nthreads++;
	up_write(&nfs4_ctl_sema);
	
	lock_kernel();          /* XXX: need to get rid of this! */
	daemonize();
	sprintf(current->comm, "nfsd4");
	current->rlim[RLIMIT_FSIZE].rlim_cur = RLIM_INFINITY; 

	nfsdstats.th_cnt++;

	/* Don't use a questionable "feature" of the RPC layer... */
	rqstp->rq_auth = 0;
	
	nfsv4_printk(ctl, "nfsd kernel thread starting [pid %d]\n", current->pid);

	/*
	 * The main request loop
	 */
	for (;;) {
		/* Block all but the shutdown signals */
		spin_lock_irq(&current->sigmask_lock);
		siginitsetinv(&current->blocked, SHUTDOWN_SIGS);
		recalc_sigpending(current);
		spin_unlock_irq(&current->sigmask_lock);

		/*
		 * Find a socket with data available and call its
		 * recvfrom routine.
		 */
		while ((err = svc_recv(serv, rqstp,
				       MAX_SCHEDULE_TIMEOUT)) == -EAGAIN)
		    ;
		if (err < 0) {
			nfsv4_printk(ctl, "kernel thread got err in svc_recv()!\n");
			break;
		}
		update_thread_usage(nfsd_busy);
		nfsd_busy++;

		/* Validate the client's address. This will also defeat
		 * port probes on port 2049 by unauthorized clients.
		 */

		/* Process request with signals blocked.  */
		spin_lock_irq(&current->sigmask_lock);
		siginitsetinv(&current->blocked, ALLOWED_SIGS);
		recalc_sigpending(current);
		spin_unlock_irq(&current->sigmask_lock);

		nfsv4_printk(ctl, "kernel thread received data, calling svc_process()\n");
		down_read(&nfs4_ctl_sema);
		svc_process(serv, rqstp);
		up_read(&nfs4_ctl_sema);
		nfsv4_printk(ctl, "svc_process() done\n");
		update_thread_usage(nfsd_busy);
		nfsd_busy--;
	}

	if (err != -EINTR) {
		printk(KERN_WARNING "nfsd: terminating on error %d\n", -err);
	} else {
		unsigned int	signo;

		for (signo = 1; signo <= _NSIG; signo++)
			if (sigismember(&current->pending.signal, signo) &&
			    !sigismember(&current->blocked, signo))
				break;
		printk(KERN_WARNING "nfsd: terminating on signal %d\n", signo);
	}

	nfsdstats.th_cnt--;
	
	down_write(&nfs4_ctl_sema);
	list_del_init(&me.list);
	nfsd_nthreads--;
	up_write(&nfs4_ctl_sema);

	/* Release the thread */
	svc_exit_thread(rqstp);

	/* Release module */
	nfsv4_printk(ctl, "nfsd kernel thread exiting [pid %d]\n", current->pid);
	MOD_DEC_USE_COUNT;
}

/*
 * This is the NFS server "laundromat" thread
 */
static int
laundromat_main(void *not_used)
{
	time_t t;

	MOD_INC_USE_COUNT;
	
	/* Inform our parent that we started successfully. */
	lock_kernel();
	laundromat_task = current;
	up(&laundromat_start_sema);
	unlock_kernel();
	
	daemonize();
        sprintf(current->comm, "nfs4_laundromat");
	current->rlim[RLIMIT_FSIZE].rlim_cur = RLIM_INFINITY;

	
	nfsv4_printk(lease, "LAUNDROMAT THREAD - starting\n");
	for (;;) {
		t = nfs4_laundromat();
		nfsv4_printk(lease, "LAUNDROMAT THREAD - sleeping for %ld seconds\n", t);
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(t * HZ);
		if (signal_pending(current))
			break;
		current->state = TASK_RUNNING;
		nfsv4_printk(lease, "LAUNDROMAT THREAD - woke up\n");
	}
	
	lock_kernel();
	laundromat_task = NULL;
	wake_up(&laundromat_kill_queue);   /* wake up the process that killed us */
	unlock_kernel();
	
	nfsv4_printk(lease, "LAUNDROMAT THREAD - done\n");
	
	MOD_DEC_USE_COUNT;
	return 0;    /* where does the return value go, anyway? */
}

static int
nfs4_dispatch(struct svc_rqst *rqstp, u32 *statp)
{
	struct svc_procedure	*proc;
	kxdrproc_t		xdr;
	u32			nfserr;
	struct global_state	st;
	u32 *bufp = rqstp->rq_resp;
	
	proc = rqstp->rq_procinfo;

	/* XXX : let svcauth_gss() handle proc 0 */

	nfsv4_printk(dispatcher, "dispatcher starting, decoding arguments\n");
	nfsv4_printk(dispatcher, "dispatcher proc: %d\n",rqstp->rq_proc);

	if (rqstp->rq_proc == 0 && rqstp->rq_cred.cr_flavor == RPC_AUTH_GSS) {
		nfsv4_printk(dispatcher, "dispatcher encoding GSS NULL PROC \n");

    /* leave room for RPC ACCECPT state set in svc_process */
		bufp++;

		/* XXX GSS NULL_PROC decode stage handled by svcauth_gss() */
		if (!xdr_encode_rpc_gss_init_res(rqstp, rqstp->rq_resbuf.buf,
		      (struct rpc_gss_init_res *)bufp))
			*statp = rpc_system_err;
		return 1;
	}
	/* Decode arguments */
	xdr = proc->pc_decode;
	if (xdr && !xdr(rqstp, rqstp->rq_argbuf.buf, rqstp->rq_argp)) {
		nfsv4_printk(dispatcher2, "failed to xdr decode arguments!\n");
		*statp = rpc_garbage_args;
		return 1;
	}

	if (proc->pc_func != (svc_procfunc)nfs4_dummy_compound) {
		nfserr = proc->pc_func(rqstp, rqstp->rq_argp, rqstp->rq_resp);
		*statp = nfserr;
		xdr = proc->pc_encode;
		if (!nfserr && xdr && !xdr(rqstp, rqstp->rq_resbuf.buf, rqstp->rq_resp)) {
			nfsv4_printk(dispatcher2, "failed to xdr encode result!\n");
			*statp = rpc_system_err;
			return 1;
		}
		return 1;
	}

	nfsv4_printk(dispatcher, "dispatcher calling nfsd4_proc_compound()\n");

	st.rqstp = rqstp;
	fh_init(&st.current_fh);
	fh_init(&st.save_fh);
	st.current_lockowner = NULL;
	nfs4_export_readlock();
	nfserr = nfsd4_proc_compound(&st, rqstp->rq_argp, rqstp->rq_resp);
	nfs4_export_readunlock();
	fh_put(&st.current_fh);
	fh_put(&st.save_fh);
	
	nfsv4_printk(dispatcher, "nfsd4_proc_compound() returned status %d to dispatcher\n",
		     nfserr);

	nfs4_release_compoundargs(rqstp->rq_argp);

	nfsv4_printk(dispatcher, "nfsd4_proc_compound(): release done\n");	
	nfsv4_printk(dispatcher, "dispatcher got status %d, encoding arguments\n", nfserr);

	xdr = proc->pc_encode;
	if (xdr && !xdr(rqstp, rqstp->rq_resbuf.buf, rqstp->rq_resp)) {
		nfsv4_printk(dispatcher, "failed to xdr encode result!\n");
		*statp = rpc_system_err;
		return 1;
	}
	nfsv4_printk(dispatcher, "encoded result successfully, dispatcher done\n");
	return 1;
}

static struct svc_version	nfsd_version4 = {
	4, 2, nfsd_procedures4, nfs4_dispatch
};

static struct svc_version *	nfsd_version[] = {
	NULL,
	NULL,
	NULL,
	NULL,
	&nfsd_version4
};

struct svc_program		nfsd_program = {
	NFS_PROGRAM,		/* program number */
	4, 4,			/* version range */
	5,			/* nr of entries in nfsd_version */
	nfsd_version,		/* version table */
	"nfsd4",	       	/* program name */
	&nfsd_svcstats,		/* version table */
};
