/*
 * linux/fs/nfsd4/vfs.c
 *
 * from linux/fs/nfsd/vfs.c
 *    Copyright (C) 1995-1999 Olaf Kirch <okir@monad.swb.de>
 * NFSv4 Implementation
 *    Kendrick Smith <kmsmith@umich.edu>
 *    Andy Adamson   <andros@umich.edu>
 *    Copyright (C) 2001 The Regents of the University of Michigan
 *
 */

/* XXX : in nfs4_open(), should replace the low-level calls to create(), open() with a higher-
   level call, such as filp_open()!  (this should also make the ->atomic_open() changes carry
   over automatically, in the case of NFSv4->NFSv4 exports)  Note: should also change
   fs/file_table.c:init_private_file() to call the ->atomic_open() op. */

/* XXX: audit usage of lock_kernel() so that it can be removed from nfsd(). */

#define __NO_VERSION__
#include <linux/config.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/quotaops.h>
#include <asm/uaccess.h>

#include <linux/nfs4/nfs4_prot.h>
#include <linux/nfs4/nfs4_debug.h>
#include <linux/nfs4/nfs4.h>
#include <linux/nfs4/nfsd4_state.h>
#include <linux/nfs4/nfsd4.h>
#include <linux/sunrpc/msg_prot.h>
#ifdef NFS4ACL
#include <linux/ext_attr.h>
#include <linux/acl_ea.h>
#endif

#define CURRENT_FH		(stp->current_fh)
#define SAVE_FH			(stp->save_fh)
#define CURRENT_LOCKOWNER	(stp->current_lockowner)

#define isdotent(n, l)			(l < 3 && n[0] == '.' && (l == 1 || n[1] == '.'))
#define set_changeid4(id, inode)	(id) = (inode)->i_mtime

#ifdef NFS4ACL
int nfs4_get_acl(struct nfs4_export *exp, struct dentry *dentry, u32 *p);
#endif

/*
 *  Permission-checking functions
 *  XXX: audit nfs4_inode_permission() carefully
 */
static int
nfs4_inode_permission(struct inode *inode, int acc)
{
	int status;

	nfsv4_printk(level2, "nfs4_inode_permission: acc=0x%x\n",acc);
	
#if NFS4_DEBUG
	if (!inode) {
		printk("nfsd_permission: negative dentry!\n");
		return NFS4ERR_NOENT;
	}
#endif

	/* XXX : how many of these checks really need to be done here?  can't we let the
	   vfs decide permission? */

	/* only care about readonly exports for files and
	 * directories. links don't have meaningful write access,
	 * and all else is local to the client
	 */
	if (S_ISREG(inode->i_mode) || S_ISDIR(inode->i_mode)) {
		if (acc & (MAY_WRITE | MAY_SATTR | MAY_TRUNC)) {
			status = NFS4ERR_ROFS;
			if (IS_RDONLY(inode)) {
				nfsv4_printk(level2, "nfs4_inode_permission: read-only inode!\n");
				goto out;
			}
			status = NFS4ERR_PERM;
			if (IS_IMMUTABLE(inode)) {
				nfsv4_printk(level2, "nfs4_inode_permission: immutable inode!\n");
				goto out;
			}
		}
	}
	status = NFS4ERR_PERM;
	if ((acc & MAY_TRUNC) && IS_APPEND(inode)) {
		nfsv4_printk(level2, "nfs4_inode_permission: truncating append-only inode!\n");
		goto out;
	}
	
#if 0
	/* XXX : is this really right? */
	if (acc & MAY_LOCK) {
		/* If we cannot rely on authentication in NLM requests,
		 * just allow locks, otherwise require read permission, or
		 * ownership
		 */
		if (exp->ex_flags & NFSEXP_NOAUTHNLM)
			return 0;
		else
			acc = MAY_READ | MAY_OWNER_OVERRIDE;
	}
#endif
	
	/*
	 * The file owner always gets access permission for accesses that
	 * would normally be checked at open time. This is to make
	 * file access work even when the client has done a fchmod(fd, 0).
	 *
	 * However, `cp foo bar' should fail nevertheless when bar is
	 * readonly. A sensible way to do this might be to reject all
	 * attempts to truncate a read-only file, because a creat() call
	 * always implies file truncation.
	 * ... but this isn't really fair.  A process may reasonably call
	 * ftruncate on an open file descriptor on a file with perm 000.
	 * We must trust the client to do permission checking - using "ACCESS"
	 * with NFSv3.
	 */
	if ((acc & MAY_OWNER_OVERRIDE) &&
	    inode->i_uid == current->fsuid)
		return 0;

	acc &= ~ MAY_OWNER_OVERRIDE; /* This bit is no longer needed,
                                        and gets in the way later */
	
	status = permission(inode, acc & (MAY_READ|MAY_WRITE|MAY_EXEC));

	/* XXX : is this really right? */
	/* Allow read access to binaries even when mode 111 */
	if (status == -EACCES && S_ISREG(inode->i_mode) && acc == MAY_READ)
	       status = permission(inode, MAY_EXEC);
	
	if (status) {
		nfsv4_printk(level2, "nfs4_inode_permission: permission() failed!\n");
		nfsv4_printk(level2, "acc=0x%x,inode->i_mode=0%o inode->i_uid=%d inode->i_gid=%d "
			     "uid=%d gid=%d\n", acc, inode->i_mode, inode->i_uid, inode->i_gid,
			     current->fsuid, current->fsgid);
		status = nfserrno(status);
	}

out:
	return status;
}

static inline int
FH_INITIALIZED(struct svc_fh *fhp)
{
	if (!fhp->fh_initialized)
		return NFS4ERR_NOFILEHANDLE;
	return NFS4_OK;
}

static inline int
FH_ACCESS(struct svc_fh *fhp, int access)
{
	return (access & ~fhp->fh_access) ? NFS4ERR_ACCES : NFS4_OK;
}

static inline int
FH_EXTERN(struct svc_fh *fhp)
{
	if (!fhp->fh_initialized)
		return NFS4ERR_NOFILEHANDLE;
	if (!fhp->fh_dentry)
		return NFS4ERR_ACCES;
	return NFS4_OK;
}

static inline int
FH_EXTERNDIR(struct svc_fh *fhp)
{
	if (!fhp->fh_initialized)
		return NFS4ERR_NOFILEHANDLE;
	if (!fhp->fh_dentry)
		return NFS4ERR_PERM;
	if (!S_ISDIR(fhp->fh_dentry->d_inode->i_mode))
		return NFS4ERR_NOTDIR;
	return NFS4_OK;
}

static inline int
FH_FILE(struct svc_fh *fhp)
{
	if (!fhp->fh_initialized)
		return NFS4ERR_NOFILEHANDLE;
	if (!fhp->fh_dentry)
		return NFS4ERR_ISDIR;
	if (!S_ISREG(fhp->fh_dentry->d_inode->i_mode))
		return S_ISDIR(fhp->fh_dentry->d_inode->i_mode) ? NFS4ERR_ISDIR
			: NFS4ERR_INVAL;
	return NFS4_OK;
}

static inline int
FH_SYMLINK(struct svc_fh *fhp)
{
	if (!fhp->fh_initialized)
		return NFS4ERR_NOFILEHANDLE;
	if (!fhp->fh_dentry)
		return NFS4ERR_ISDIR;
	if (!S_ISLNK(fhp->fh_dentry->d_inode->i_mode))
		return S_ISDIR(fhp->fh_dentry->d_inode->i_mode) ? NFS4ERR_ISDIR
			: NFS4ERR_INVAL;
	return NFS4_OK;
}

static inline void
nfs4_setuser(struct svc_fh *fhp)
{
	current->fsuid = fhp->fh_fsuid;
	current->fsgid = fhp->fh_fsgid;
	current->cap_effective = fhp->fh_cap_effective;
	nfsv4_printk(auth, "nfs4_setuser: setting uid=%d, gid=%d, cap_effective=0x%x\n",
		     fhp->fh_fsuid, fhp->fh_fsgid, fhp->fh_cap_effective);
}

void
nfsd_sync_dir(struct dentry *dp)
{
	struct inode *inode = dp->d_inode;
	int (*fsync) (struct file *, struct dentry *, int);
	
	if (inode->i_fop && (fsync = inode->i_fop->fsync)) {
		fsync(NULL, dp, 0);
	}
}

/*
 *  Some functions for processing fattr's
 */

static u32      nfs4_ftypes[] = {
        NF4NON,  NF4FIFO, NF4CHR, NF4BAD,
        NF4DIR,  NF4BAD,  NF4BLK, NF4BAD,
        NF4REG,  NF4BAD,  NF4LNK, NF4BAD,
        NF4SOCK, NF4BAD,  NF4LNK, NF4BAD,
};

static int
fattr4_size_vector[] = {
	12,			/* XXX : FATTR4_SUPPORTED_ATTRS */
	sizeof(fattr4_type),
	sizeof(fattr4_fh_expire_type),
	sizeof(fattr4_change),
	sizeof(fattr4_size),
	sizeof(fattr4_link_support),
	sizeof(fattr4_symlink_support),
	sizeof(fattr4_named_attr),
	sizeof(fattr4_fsid),
	sizeof(fattr4_unique_handles),
	sizeof(fattr4_lease_time),
	sizeof(fattr4_rdattr_error),
#ifdef NFS4ACL
        ((sizeof(nfsace4) + 12) * 15),                   /* XXX : FATTR4_ACL, Need to put real value in!*/
#else
        -1,
#endif
	sizeof(fattr4_aclsupport),
	sizeof(fattr4_archive),
	sizeof(fattr4_cansettime),
	sizeof(fattr4_case_insensitive),
	sizeof(fattr4_case_preserving),
	sizeof(fattr4_chown_restricted),
	sizeof(fattr4_fileid),
	sizeof(fattr4_files_avail),
	4 + NFS4_FHSIZE,  /* FATTR4_FILEHANDLE */
	sizeof(fattr4_files_free),
	sizeof(fattr4_files_total),
	-1,			/* XXX : FATTR4_FS_LOCATIONS */
	sizeof(fattr4_hidden),
	sizeof(fattr4_homogeneous),
	sizeof(fattr4_maxfilesize),
	sizeof(fattr4_maxlink),
	sizeof(fattr4_maxname),
	sizeof(fattr4_maxread),
	sizeof(fattr4_maxwrite),
	-1,			/* XXX : FATTR4_MIMETYPE */
	sizeof(fattr4_mode),
	sizeof(fattr4_no_trunc),
	sizeof(fattr4_numlinks),
	4 + XDR_LEN(NFS4_MAXNAMLEN),
	4 + XDR_LEN(NFS4_MAXNAMLEN),
	sizeof(fattr4_quota_hard),
	sizeof(fattr4_quota_soft),
	sizeof(fattr4_quota_used),
	sizeof(fattr4_rawdev),
	sizeof(fattr4_space_avail),
	sizeof(fattr4_space_free),
	sizeof(fattr4_space_total),
	sizeof(fattr4_space_used),
	sizeof(fattr4_system),
	sizeof(fattr4_time_access),
	-1,			/* XXX : FATTR4_TIME_ACCESS_SET */
	sizeof(fattr4_time_backup),
	-1,			/* XXX : FATTR4_TIME_CREATE */
	sizeof(fattr4_time_delta),
	sizeof(fattr4_time_metadata),
	sizeof(fattr4_time_modify),
	-1			/* XXX : FATTR4_TIME_MODIFY_SET */
};

static int
get_attrval4_len(bitmap4 *bm)
{
	/* XXX : cache recently used values? */

	int len = 0;
	int i, j, bit = 0;

	for (i = 0; i < bm->bitmap4_len; i++)
		for (j = 0; j < 32; j++) {
			if (bit == FATTR4_ATTR_MAX)
				goto out;
			if (bm->bitmap4_val[i] & (1<<j)) {
				/* XXX : redundant sanity check */
				if (fattr4_size_vector[bit] < 0) {
					nfsv4_printk(level1,
						     "BAD BIT %d in get_attrval4_len()!\n",
						     bit);
					len = -1;
					goto out;
				}
				else
					len += fattr4_size_vector[bit];
			}
			bit++;
		}

out:
	return len;
}

#ifdef NFS4ACL
static u32 GETATTR_BITMAP[] = { 0xecdf9f7f, 0x0038be3e };
#else
static u32 GETATTR_BITMAP[] = { 0xecdf8f7f, 0x0038be3e };
#endif
#define    GETATTR_BITMAP_LEN 2

static inline int
set_getattr_bitmap(bitmap4 *dst, bitmap4 *src)
{
	int i;
	u32 *bmval;

	if (!(bmval = NFS4_ALLOC(GETATTR_BITMAP_LEN * 4))) {
		nfsv4_printk(level1, "failed malloc in set_getattr_bitmap!\n");
		dst->bitmap4_len = 0;
		dst->bitmap4_val = NULL;
		return NFS4ERR_RESOURCE;
	}
	dst->bitmap4_len = GETATTR_BITMAP_LEN;
	dst->bitmap4_val = bmval;

	for (i = 0; i < GETATTR_BITMAP_LEN; i++)
		bmval[i] = (src->bitmap4_len <= i) ? 0 :
				(src->bitmap4_val[i] & GETATTR_BITMAP[i]);
	return NFS4_OK;
}

static int
__write_attrs(struct nfs4_export *exp, struct dentry *dentry, bitmap4 *bm, u32 *buf)
{
	int i, j, k,  bit = 0;
	struct inode *inode = dentry->d_inode;
	int statfs_status, status;
	struct super_block *sb;
	struct statfs statfs;
	struct gss_cacheent *gssp;
	u64 dummy64;
	u32 bmlen = bm->bitmap4_len;
	u32 *bmval = bm->bitmap4_val;
	u32 *p = (u32 *) buf;
	u32 *lenp;
	mm_segment_t oldfs;

#if NFS4_DEBUG
	if (!exp->ex_dentry) {
		printk("write_attrs: bad export!\n");
		return -NFS4ERR_SERVERFAULT;
	}
#endif
	
	/*
	 *  XXX : need to implement
	 *        NAMED_ATTR (mandatory)
	 *        ACL, ACLSUPPORT, FILES_AVAIL, FS_LOCATIONS, HIDDEN, MAXLINK,
	 *        MIMETYPE, QUOTA_HARD, QUOTA_SOFT, QUOTA_USED, SYSTEM, TIME_BACKUP,
	 *        TIME_METADATA
	 */

	/* XXX : optimize so that statfs is only called when needed */
	if (!(sb = inode->i_sb) || !sb->s_op->statfs)
		statfs_status = -EIO;
	else {
		/* XXX : do we need to set_fs(KERNEL_DS) here? */
		oldfs = get_fs(); set_fs(KERNEL_DS);
		statfs_status = sb->s_op->statfs(sb, &statfs);
		set_fs(oldfs);
	}

	nfsv4_printk(getattr, "write_attrs: starting with bitmap ");
	nfsv4_hexdump(getattr, bmval, bmlen << 2);
	
	for (i = 0; i < bmlen; i++)
		for (j = 0; j < 32; j++) {
			if (bit == FATTR4_ATTR_MAX)
				goto done;
			if (bmval[i] & (1<<j)) {
				nfsv4_printk(getattr, "%d ", bit);
				switch (bit) {
				case FATTR4_SUPPORTED_ATTRS:		/* 0 */
					*p++ = htonl(GETATTR_BITMAP_LEN);
					for (k = 0; k < GETATTR_BITMAP_LEN; k++)
						*p++ = htonl(GETATTR_BITMAP[k]);
					break;
				case FATTR4_TYPE:			/* 1 */
					/* XXX : check for bad file type */
					*p++ = htonl(nfs4_ftypes[ (inode->i_mode >> 12) &
								   0xf ]);
					break;
				case FATTR4_FH_EXPIRE_TYPE:		/* 2 */
					*p++ = htonl(FH4_PERSISTENT);
					break;
				case FATTR4_CHANGE:			/* 3 */
					*p++ = 0;
					*p++ = htonl(inode->i_mtime);
					break;
				case FATTR4_SIZE:			/* 4 */
					*p++ = 0;
					*p++ = htonl(inode->i_size);
					break;
				case FATTR4_LINK_SUPPORT:		/* 5 */
					*p++ = htonl(1);
					break;
				case FATTR4_SYMLINK_SUPPORT:		/* 6 */
					*p++ = htonl(1);
					break;
				case FATTR4_FSID:			/* 8 */
					*p++ = 0;
					*p++ = htonl(MAJOR(inode->i_dev));
					*p++ = 0;
					*p++ = htonl(MINOR(inode->i_dev));
					break;
				case FATTR4_UNIQUE_HANDLES:		/* 9 */
					*p++ = 0;
					break;
				case FATTR4_LEASE_TIME:
					*p++ = htonl(NFS4_LEASE_TIME);	/* 10 */
					break;
				case FATTR4_RDATTR_ERROR:		/* 11 */
					/*
					 * XXX: temporary bakeoff workaround -- fix!
					 */
					*p++ = htonl(NFS4_OK);
					break;
                                case FATTR4_ACL:
					printk("Server case FATTR4_ACL!!\n");
#ifdef NFS4ACL
                                        printk("amount used before: %d\n", (char *) p - (char *) buf);
                                        printk("p : %d\n",(u32) p);
                                        p += nfs4_get_acl(exp, dentry, p);   /* 12 */
                                        printk("p : %d\n",(u32) p);
                                        printk("amount used after: %d\n", (char *) p - (char *) buf);
#endif
                                        break;
				case FATTR4_CANSETTIME:			/* 15 */
					*p++ = htonl(1);
					/* XXX : assumes exported fs is local! */
					break;
				case FATTR4_CASE_INSENSITIVE:		/* 16 */
					*p++ = 0;
					break;
				case FATTR4_CASE_PRESERVING:		/* 17 */
					*p++ = 0;
					break;
				case FATTR4_CHOWN_RESTRICTED:		/* 18 */
					*p++ = htonl(1);
					break;
				case FATTR4_FILEHANDLE:			/* 19 */
					lenp = p++;
					status = fh_write(exp, dentry, p);
					if (status < 0)
						goto out;
					*lenp = htonl(status);
					p += XDR_QUADLEN(status);
					break;
				case FATTR4_FILEID:			/* 20 */
#if 0
					/* This seemed like a good idea to me, but Solaris
					   NFSv4 client choked on it at Connectathon... */
					*p++ = inode->i_generation;
#else
					*p++ = 0;
#endif
					*p++ = (u32)inode->i_ino;
					break;
				case FATTR4_FILES_FREE:			/* 22 */
					if (statfs_status)
						goto out_statfs;
					*p++ = 0;
					*p++ = htonl(statfs.f_ffree);
					break;
				case FATTR4_FILES_TOTAL:		/* 23 */
					if (statfs_status)
						goto out_statfs;
					*p++ = 0;
					*p++ = htonl(statfs.f_files);
					break;
				case FATTR4_HOMOGENEOUS:		/* 26 */
					*p++ = htonl(1);
					break;
				case FATTR4_MAXFILESIZE:		/* 27 */
					*p++ = 0;
					*p++ = ~(u32)0;
					break;
				case FATTR4_MAXNAME:			/* 29 */
					*p++ = htonl(NFS4_MAXNAMLEN);
					break;
				case FATTR4_MAXREAD:			/* 30 */
				case FATTR4_MAXWRITE:			/* 31 */
					*p++ = 0;
					*p++ = htonl(NFSSVC_MAXBLKSIZE);
					break;
				case FATTR4_MODE:			/* 33 */
					*p++ = htonl(inode->i_mode);
					break;
				case FATTR4_NO_TRUNC:			/* 34 */
					*p++ = htonl(1); /* XXX : maybe we should truncate.. */
					break;
				case FATTR4_NUMLINKS:			/* 35 */
					*p++ = htonl(inode->i_nlink);
					break;
				case FATTR4_OWNER:			/* 36 */
					if ((status = gss_get_name(GSS_OWNER, inode->i_uid,
								   &gssp))) {
						status = -nfserrno(status);
						goto out;
					}
					*p++ = htonl(gssp->name.len);
					memcpy(p, gssp->name.name, gssp->name.len);
					p += XDR_QUADLEN(gssp->name.len);
					gss_put(gssp);
					break;
				case FATTR4_OWNER_GROUP:		/* 37 */
					if ((status = gss_get_name(GSS_GROUP, inode->i_gid,
								   &gssp))) {
						status = -nfserrno(status);
						goto out;
					}
					*p++ = htonl(gssp->name.len);
					memcpy(p, gssp->name.name, gssp->name.len);
					p += XDR_QUADLEN(gssp->name.len);
					gss_put(gssp);
					break;
				case FATTR4_RAWDEV:			/* 41 */
					*p++ = htonl(MAJOR(inode->i_rdev));
					*p++ = htonl(MINOR(inode->i_rdev));
					break;
				case FATTR4_SPACE_AVAIL:		/* 42 */
					if (statfs_status)
						goto out_statfs;
					if ((statfs.f_bavail < 0) ||
					    (statfs.f_bavail > statfs.f_bfree))
						statfs.f_bavail = statfs.f_bfree;
					dummy64 = (u64)statfs.f_bavail * (u64)statfs.f_bsize;
					*p++ = htonl((u32) (dummy64 >> 32));
					*p++ = htonl((u32) dummy64);
					break;
				case FATTR4_SPACE_FREE:			/* 43 */
					if (statfs_status)
						goto out_statfs;
					dummy64 = (u64)statfs.f_bfree * (u64)statfs.f_bsize;
					*p++ = htonl((u32) (dummy64 >> 32));
					*p++ = htonl((u32) dummy64);
					break;
				case FATTR4_SPACE_TOTAL:		/* 44 */
					if (statfs_status)
						goto out_statfs;
					dummy64 = (u64)statfs.f_blocks * (u64)statfs.f_bsize;
					*p++ = htonl((u32) (dummy64 >> 32));
					*p++ = htonl((u32) dummy64);
					break;
				case FATTR4_SPACE_USED:			/* 45 */
					dummy64 = inode->i_blksize * inode->i_blocks;
					*p++ = htonl((u32) (dummy64 >> 32));
					*p++ = htonl((u32) dummy64);
					break;
				case FATTR4_TIME_ACCESS:		/* 47 */
					*p++ = 0;
					*p++ = htonl(inode->i_atime);
					*p++ = 0;
					break;
				case FATTR4_TIME_ACCESS_SET:		/* 48 */
					nfsv4_printk(level1,
						     "TIME_ACCESS_SET in write_fattr!\n");
					status = -NFS4ERR_INVAL;
					goto out;
				case FATTR4_TIME_DELTA:			/* 51 */
					*p++ = 0;
					*p++ = htonl(1);
					*p++ = 0;
					break;
				case FATTR4_TIME_METADATA:              /* 52 */
					*p++ = 0;
					*p++ = htonl(inode->i_ctime);
					*p++ = 0;
					break;
				case FATTR4_TIME_MODIFY:		/* 53 */
					*p++ = 0;
					*p++ = htonl(inode->i_mtime);
					*p++ = 0;
					break;
				case FATTR4_TIME_MODIFY_SET:		/* 54 */
					nfsv4_printk(level1,
						     "TIME_MODIFY_SET in write_fattr!\n");
					status = -NFS4ERR_INVAL;
					goto out;
				default:
					/* XXX : redundant sanity check! */
					nfsv4_printk(getattr2,
						     "bad bit %d in write_fattr!\n", bit);
					status = -NFS4ERR_INVAL;
					goto out;
				}
			}
			bit++;
		}	

done:
	status = (char *)p - (char *)buf;
	nfsv4_printk(getattr, "write_attrs wrote fattr\n");
	nfsv4_hexdump(getattr, buf, status);
out:
	return status;
out_statfs:
	nfsv4_printk(getattr, "write_fattr: statfs returned error!\n");
	status = -nfserrno(statfs_status);
	goto out;
}

static int
__write_attrs_pseudo(struct nfs4_export *exp, bitmap4 *bm, u32 *buf)
{
	int i, j, k, bit = 0;
	int status;
	u32 bmlen = bm->bitmap4_len;
	u32 *bmval = bm->bitmap4_val;
	u32 *p = buf;
	u32 *lenp;
	struct gss_cacheent *gssp;
	
	/*
	 *  XXX : need to implement
	 *        NAMED_ATTR (mandatory)
	 *        ACL, ACLSUPPORT, FILEHANDLE, FILES_AVAIL, FS_LOCATIONS, HIDDEN, MAXLINK,
	 *        MIMETYPE, QUOTA_HARD, QUOTA_SOFT, QUOTA_USED, SYSTEM, TIME_BACKUP,
	 *        TIME_METADATA
	 */
	
	nfsv4_printk(getattr, "write_attrs_pseudo: starting with bitmap ");
	nfsv4_hexdump(getattr, bmval, bmlen << 2);

	for (i = 0; i < bmlen; i++)
		for (j = 0; j < 32; j++) {
			if (bit == FATTR4_ATTR_MAX)
				goto done;
			if (bmval[i] & (1<<j)) {
				nfsv4_printk(getattr, "%d ", bit);
				switch (bit) {
				case FATTR4_SUPPORTED_ATTRS:		/* 0 */
					*p++ = htonl(GETATTR_BITMAP_LEN);
					for (k = 0; k < GETATTR_BITMAP_LEN; k++)
						*p++ = htonl(GETATTR_BITMAP[k]);
					break;
				case FATTR4_TYPE:			/* 1 */
					*p++ = htonl(NF4DIR);
					break;
				case FATTR4_FH_EXPIRE_TYPE:		/* 2 */
					/* XXX : not quite right, but nothing else
					   is either.. */
					*p++ = htonl(FH4_PERSISTENT);
					break;
				case FATTR4_CHANGE:			/* 3 */
					*p++ = 0;
					*p++ = htonl(nfssvc_boot.tv_sec);
					break;
				case FATTR4_SIZE:			/* 4 */
					*p++ = 0;
					*p++ = htonl(exp->ex_nchild);
					break;
				case FATTR4_LINK_SUPPORT:		/* 5 */
					*p++ = htonl(1);
					break;
				case FATTR4_SYMLINK_SUPPORT:		/* 6 */
					*p++ = htonl(1);
					break;
				case FATTR4_FSID:			/* 8 */
					/* XXX : pseudo fs - major 0, minor 0 */
					*p++ = 0;
					*p++ = 0;
					*p++ = 0;
					*p++ = 0;
					break;
				case FATTR4_UNIQUE_HANDLES:		/* 9 */
					*p++ = 0;
					break;
				case FATTR4_LEASE_TIME:
					*p++ = htonl(NFS4_LEASE_TIME);	/* 10 */
					break;
				case FATTR4_RDATTR_ERROR:		/* 11 */
					/*
					 * XXX: temporary bakeoff workaround -- fix!
					 */
					*p++ = htonl(NFS4_OK);
					break;
				case FATTR4_CANSETTIME:			/* 15 */
					*p++ = htonl(1);
					/* XXX : assumes exported fs is local! */
					break;
				case FATTR4_CASE_INSENSITIVE:		/* 16 */
					*p++ = 0;
					break;
				case FATTR4_CASE_PRESERVING:		/* 17 */
					*p++ = 0;
					break;
				case FATTR4_CHOWN_RESTRICTED:		/* 18 */
					*p++ = htonl(1);
					break;
				case FATTR4_FILEHANDLE:			/* 19 */
					lenp = p++;
					status = fh_write(exp, NULL, p);
					if (status < 0)
						goto out;
					*lenp = htonl(status);
					p += XDR_QUADLEN(status);
					break;
				case FATTR4_FILEID:			/* 20 */
					*p++ = 0;
					*p++ = htonl(exp->ex_id + 1);
					break;
				case FATTR4_FILES_FREE:			/* 22 */
					*p++ = 0;
					*p++ = 0;
					break;
				case FATTR4_FILES_TOTAL:		/* 23 */
					*p++ = 0;
					*p++ = 0;
					break;
				case FATTR4_HOMOGENEOUS:		/* 26 */
					*p++ = htonl(1);
					break;
				case FATTR4_MAXFILESIZE:		/* 27 */
					*p++ = 0;
					*p++ = ~(u32)0;
					break;
				case FATTR4_MAXNAME:			/* 29 */
					*p++ = htonl(NFS4_MAXNAMLEN);
					break;
				case FATTR4_MAXREAD:			/* 30 */
				case FATTR4_MAXWRITE:			/* 31 */
					*p++ = 0;
					*p++ = htonl(NFSSVC_MAXBLKSIZE);
					break;
				case FATTR4_MODE:			/* 33 */
					*p++ = htonl(S_IFDIR | S_IRUSR | S_IXUSR | S_IRGRP |
						     S_IXGRP | S_IROTH | S_IXOTH);
					break;
				case FATTR4_NO_TRUNC:			/* 34 */
					*p++ = htonl(1); /* XXX : maybe we should truncate.. */
					break;
				case FATTR4_NUMLINKS:			/* 35 */
					*p++ = htonl(exp->ex_nchild + 2);
					break;
				case FATTR4_OWNER:			/* 36 */
					/* XXX : optimize */
					if ((status = gss_get_name(GSS_OWNER, 0, &gssp))) {
						status = -nfserrno(status);
						goto out;
					}
					*p++ = htonl(gssp->name.len);
					memcpy(p, gssp->name.name, gssp->name.len);
					p += XDR_QUADLEN(gssp->name.len);
					gss_put(gssp);
					break;
				case FATTR4_OWNER_GROUP:		/* 37 */
					if ((status = gss_get_name(GSS_GROUP, 0, &gssp))) {
						status = -nfserrno(status);
						goto out;
					}
					*p++ = htonl(gssp->name.len);
					memcpy(p, gssp->name.name, gssp->name.len);
					p += XDR_QUADLEN(gssp->name.len);
					gss_put(gssp);
					break;
				case FATTR4_RAWDEV:			/* 41 */
					/* XXX : pseudo fs - major 0, minor 0 */
					*p++ = htonl(0);
					*p++ = htonl(0);
					break;
				case FATTR4_SPACE_AVAIL:		/* 42 */
					*p++ = 0;
					*p++ = htonl(1 << 30);
					break;
				case FATTR4_SPACE_FREE:			/* 43 */
					*p++ = 0;
					*p++ = htonl(1 << 30);
					break;
				case FATTR4_SPACE_TOTAL:		/* 44 */
					*p++ = 0;
					*p++ = htonl(1 << 30);
					break;
				case FATTR4_SPACE_USED:			/* 45 */
					*p++ = 0;
					*p++ = 0;
					break;
				case FATTR4_TIME_ACCESS:		/* 47 */
					*p++ = 0;
					*p++ = htonl(nfssvc_boot.tv_sec);
					*p++ = 0;
					break;
				case FATTR4_TIME_ACCESS_SET:		/* 48 */
					nfsv4_printk(level1,
						     "TIME_ACCESS_SET in write_fattr!\n");
					status = -NFS4ERR_INVAL;
					goto out;
				case FATTR4_TIME_DELTA:			/* 51 */
					*p++ = 0;
					*p++ = htonl(1);
					*p++ = 0;
					break;
				case FATTR4_TIME_METADATA:              /* 52 */
					*p++ = 0;
					*p++ = htonl(nfssvc_boot.tv_sec);
					*p++ = 0;
					break;
				case FATTR4_TIME_MODIFY:		/* 53 */
					*p++ = 0;
					*p++ = htonl(nfssvc_boot.tv_sec);
					*p++ = 0;
					break;
				case FATTR4_TIME_MODIFY_SET:		/* 54 */
					nfsv4_printk(level1,
						     "TIME_MODIFY_SET in write_fattr!\n");
					status = -NFS4ERR_INVAL;
					goto out;
				default:
					/* XXX : redundant sanity check! */
					nfsv4_printk(getattr2,
						     "bad bit %d in write_fattr!\n", bit);
					status = NFS4ERR_INVAL;
				}
			}
			bit++;
		}
	
done:
	status = (char *)p - (char *)buf;
	nfsv4_printk(getattr, "write_attrs_pseudo wrote fattr\n");
	nfsv4_hexdump(getattr, buf, status);
out:
	return status;
}

static inline int
write_attrs(struct nfs4_export *exp, struct dentry *dentry, bitmap4 *bm, u32 *buf)
{
	return dentry ? __write_attrs(exp, dentry, bm, buf)
		: __write_attrs_pseudo(exp, bm, buf);
}

static u32 SETATTR_BITMAP[] = { 0x00000010, 0x00000032 };
#define    SETATTR_BITMAP_LEN 2

static inline int
set_setattr_bitmap(bitmap4 *dst, bitmap4 *src)
{
	int i;
	u32 *bmval;

	if (!(bmval = NFS4_ALLOC(SETATTR_BITMAP_LEN * 4))) {
		nfsv4_printk(level1, "failed malloc in set_setattr_bitmap!\n");
		dst->bitmap4_len = 0;
		dst->bitmap4_val = NULL;
		return NFS4ERR_RESOURCE;
	}
	dst->bitmap4_len = SETATTR_BITMAP_LEN;
	dst->bitmap4_val = bmval;

	for (i = 0; i < SETATTR_BITMAP_LEN; i++)
		bmval[i] = (src->bitmap4_len <= i) ? 0 :
				(src->bitmap4_val[i] & SETATTR_BITMAP[i]);
	return NFS4_OK;
}

int
read_attrs(struct fattr4 *fp, struct iattr *iap)
{
	int bmlen = fp->attrmask.bitmap4_len;
        u32 *bmval = fp->attrmask.bitmap4_val;
	char *cptr = fp->attr_vals.attrlist4_val;
	char *end = cptr + fp->attr_vals.attrlist4_len;
	u32 len, len2;
	int status = 0;
	int i, j, bit = 0;

	iap->ia_valid = 0;

	nfsv4_printk(setattr, "read_attrs: starting with bm ");
	nfsv4_hexdump(setattr, bmval, bmlen << 2);
	nfsv4_printk(setattr, "read_attrs: starting with attrs\n");
	nfsv4_hexdump(setattr, cptr, fp->attr_vals.attrlist4_len);
	
#define OVERFLOW_CHECK(n)   if (cptr + n > end) goto overrun	

	for (i = 0; i < bmlen; i++)
		for (j = 0; j < 32; j++) {
			if (bmval[i] & (1 << j)) {
				nfsv4_printk(setattr, "%d ", bit);
				if (bit == FATTR4_ATTR_MAX)
					goto out;
				switch (bit) {
				case FATTR4_SUPPORTED_ATTRS:	/* skip */
					OVERFLOW_CHECK(4);
					len = ntohl(*(u32 *)cptr);
					OVERFLOW_CHECK(4 * len);
					cptr += 4 + 4*len;
					break;
				case FATTR4_SIZE:
					OVERFLOW_CHECK(8);
					if (*(u32 *)cptr) {
						status = NFS4ERR_INVAL; /* XXX: right error? */
						goto out;
					}
					cptr += 4;
					iap->ia_valid |= ATTR_SIZE;
					iap->ia_size = ntohl(*(u32 *)cptr);
					cptr += 4;
					nfsv4_printk(setattr, "read_attrs read size=%d\n",
						     (int)iap->ia_size);
					break;
				case FATTR4_FILEHANDLE:		/* skip */
				case FATTR4_MIMETYPE:		/* skip */
					OVERFLOW_CHECK(4);
					len = XDR_LEN(ntohl(*(u32 *)cptr));
					OVERFLOW_CHECK(len);
					cptr += 4 + len;
					break;
				case FATTR4_ACL:		/* skip */
					OVERFLOW_CHECK(4);
					len = ntohl(*(u32 *)cptr);
					cptr += 4;
					while (len--) {
						OVERFLOW_CHECK(16);
						cptr += 12;
						len2 = XDR_LEN(ntohl(*(u32 *)cptr));
						OVERFLOW_CHECK(len2);
						cptr += 4 + len2;
					}
					break;
				case FATTR4_FS_LOCATIONS:	/* skip */
					OVERFLOW_CHECK(4);
					len = ntohl(*(u32 *)cptr);
					cptr += 4;
					while (len--) {
						OVERFLOW_CHECK(4);
						len2 = XDR_LEN(ntohl(*(u32 *)cptr));
						OVERFLOW_CHECK(len2);
						cptr += 4 + len2;
					}
					/* XXX : not finished yet.  yuck! */
					nfsv4_printk(setattr, "read_attrs: FS_LOCATIONS!\n");
					status = NFS4ERR_NOTSUPP;
					goto out;
				case FATTR4_MODE:
					OVERFLOW_CHECK(4);
					iap->ia_valid |= ATTR_MODE;
					iap->ia_mode = ntohl(*(u32 *)cptr) &
						(S_IFMT | S_IALLUGO);
					nfsv4_printk(setattr,
						     "read_attrs() read mode=0%o\n",
						     (unsigned int) iap->ia_mode);
					cptr += 4;
					break;
				case FATTR4_OWNER:
					OVERFLOW_CHECK(4);
					len = ntohl(*(u32 *)cptr);
					len2 = XDR_LEN(len);
					OVERFLOW_CHECK(len2);
					cptr += 4;
					if ((status = gss_get_num(GSS_OWNER, len, cptr,
								  &iap->ia_uid)))
						goto out;
					iap->ia_valid |= ATTR_UID;
					cptr += len2;
					nfsv4_printk(setattr,
						     "read_attrs() read uid=%d\n",
						     (int) iap->ia_uid);
					break;
				case FATTR4_OWNER_GROUP:
					OVERFLOW_CHECK(4);
					len = ntohl(*(u32 *)cptr);
					len2 = XDR_LEN(len);
					OVERFLOW_CHECK(len2);
					cptr += 4;
					if ((status = gss_get_num(GSS_GROUP, len, cptr,
								  &iap->ia_gid)))
						goto out;
					iap->ia_valid |= ATTR_GID;
					cptr += len2;
					nfsv4_printk(setattr,
						     "read_attrs() read gid=%d\n",
						     (int) iap->ia_gid);
					break;
				case FATTR4_TIME_ACCESS_SET:
					OVERFLOW_CHECK(4);
					iap->ia_valid |= ATTR_ATIME;
					if (*(u32 *)cptr == 0) {
						iap->ia_atime = CURRENT_TIME;
						break;
					}
					/* XXX : need to worry about timezone conversion? */
					cptr += 4;
					OVERFLOW_CHECK(12);
					if (*(u32 *)cptr) {
						nfsv4_printk(setattr2,
							     "read_attrs: bad atime!\n");
						status = NFS4ERR_INVAL;
						goto out;
					}
					cptr += 4;
					iap->ia_atime = ntohl(*(u32 *)cptr);
					cptr += 8;
					break;
				case FATTR4_TIME_MODIFY_SET:
					OVERFLOW_CHECK(4);
					iap->ia_valid |= ATTR_MTIME;
					if (*(u32 *)cptr == 0) {
						iap->ia_mtime = CURRENT_TIME;
						break;
					}
					/* XXX : need to worry about timezone conversion? */
					cptr += 4;
					OVERFLOW_CHECK(12);
					if (*(u32 *)cptr) {
						nfsv4_printk(setattr2,
							     "read_attrs: bad atime!\n");
						status = NFS4ERR_INVAL;
						goto out;
					}
					cptr += 4;
					iap->ia_mtime = ntohl(*(u32 *)cptr);
					cptr += 8;
					break;
				default:
					nfsv4_printk(setattr,
						     "read_attrs() read unknown bit %d\n",
						     bit);
					if (fattr4_size_vector[bit] < 0) {
						nfsv4_printk(setattr2,
							     "BAD BIT %d set in read_attrs!\n",
							     bit);
						status = NFS4ERR_NOTSUPP;
						goto out;
					}
					OVERFLOW_CHECK(sizeof(fattr4_size_vector[bit]));
					cptr += fattr4_size_vector[bit];
				}
			}
			bit++;
		}

#undef OVERFLOW_CHECK
	
out:
	nfsv4_printk(setattr, "read_attrs returning %d\n", status);
	return status;
overrun:
	nfsv4_printk(setattr2, "buffer overrun in read_attrs!\n");
	status = NFS4ERR_INVAL;
	goto out;
}

/*
 *  called by LOOKUP, OPEN
 *  XXX : how can we deal with cross-mountpoint lookups?
 */

static int
do_lookup(struct global_state *stp, u32 namlen, char *name)
{
	struct list_head *l;
	struct nfs4_export *exp_new;
	struct dentry *dentry, *d_new;
	struct inode *inode;
	char buf[NFS4_MAXNAMLEN + 1];
	int status;

	/* caller must ensure fh is initialized */
	
	nfsv4_printk(lookup, "do_lookup() starting; name=\"%.*s\"; current fh = ",
		     (int) namlen, name);
	nfsv4_print_svcfh(lookup, &CURRENT_FH);

	if ((status = FH_ACCESS(&CURRENT_FH, MAY_EXEC)))
		goto out;
	
	/* XXX : for now, treat '.' and '..' as errors */
	if (isdotent(name, namlen)) {
		nfsv4_printk(lookup2, "do_lookup: dot entry!\n");
		status = NFS4ERR_INVAL;
		goto out;
	}
	if (namlen > NFS4_MAXNAMLEN) {
		nfsv4_printk(lookup2, "do_lookup: name too long!\n");
		status = NFS4ERR_NAMETOOLONG;
		goto out;
	}

	dentry = CURRENT_FH.fh_dentry;
	if (!dentry) {
		/* pseudofs node -- do linear search through children */
		list_for_each(l, &CURRENT_FH.fh_export->ex_child) {
			exp_new = list_entry(l, struct nfs4_export, ex_subdirs);
			if ((exp_new->ex_namlen == namlen) &&
			    !memcmp(exp_new->ex_name, name, namlen)) {
				/* got it; now update CURRENT_FH and exit */
				CURRENT_FH.fh_export = exp_new;
				CURRENT_FH.fh_dentry = DGET(exp_new->ex_dentry);
				if ((status = nfs4_setsec(stp->rqstp, exp_new,
							  &CURRENT_FH.fh_secattrs)))
					goto out;
				nfs4_setuser(&CURRENT_FH);
				goto out;
			}
		}
		/* name not found */
		nfsv4_printk(lookup2, "do_lookup failed on \"%.*s\" in pseudofs\n",
			     (int) namlen, name);
		status = NFS4ERR_NOENT;
		goto out;		
	}
	else {
		inode = dentry->d_inode;
		if (!S_ISDIR(inode->i_mode)) {
			nfsv4_printk(lookup2, "do_lookup: not a directory!\n");
			status = NFS4ERR_NOTDIR;
			goto out;
		}
		if ((status = nfs4_inode_permission(inode, MAY_EXEC))) {
			nfsv4_printk(lookup2, "do_lookup: permission denied!\n");
			goto out;
		}
		/* XXX : ugh! what can we do about this copy? */
		memcpy(buf, name, namlen);
		buf[namlen] = 0;
		
		/* this block taken from nfsv3 code. XXX : is locking right? */
		down(&inode->i_sem);
		d_new = LOOKUP_ONE(buf, dentry);
		up(&inode->i_sem);
		if (IS_ERR(d_new)) {
			nfsv4_printk(lookup2, "do_lookup(): lookup_one failed!\n");
			status = nfserrno(PTR_ERR(d_new));
			goto out;
		}
		/* XXX : is it possible for d_new to be NULL? */
		if (!d_new || !d_new->d_inode) {
			nfsv4_printk(lookup, "do_lookup(): file not found\n");
			DPUT(d_new);
			status = NFS4ERR_NOENT;
			goto out;
		}
		/* XXX : this code will not successfully cross a mountpoint; fix soon! */
		/* compose fh */
		DPUT(dentry);
		CURRENT_FH.fh_dentry = d_new;
	}

out:
	nfsv4_printk(lookup, "do_lookup() returning status %d\n", status);
	return status;
}

/*
 *  called by SETATTR and OPEN
 */

#if 0
/*
 * For some asinine reason, I felt obligated to write my own setattr routine
 * instead of just copying v3's.  A straightforward port of the tried and
 * true nfsv3 routine follows the #else...
 */
int
do_setattr(struct svc_fh *fhp, struct iattr *iap)
{
	struct dentry *dentry = fhp->fh_dentry;
	struct inode *inode = dentry->d_inode;
	int status;
	int write_access = 0, locked = 0;
	int accmode;

	/* caller must ensure that fhp is initialized, external */
	
	/* XXX : It is possible for this routine to partially succeed, in which case we
	   should find a way to return the set of attributes which did succeed, so that
	   they can be returned from a SETATTR. */

	/* XXX : Is this the right permission check? */

	accmode = (iap->ia_valid & (ATTR_ATIME|ATTR_MTIME)) ?
		(MAY_SATTR | MAY_WRITE | MAY_OWNER_OVERRIDE) : (MAY_SATTR);
	if (iap->ia_valid & ATTR_SIZE)
		accmode |= (MAY_TRUNC | MAY_WRITE);
	if ((status = FH_ACCESS(fhp, accmode))) {
		nfsv4_printk(setattr2, "do_setattr: FH_ACCESS failed\n");
		goto out;
	}
	if ((status = nfs4_inode_permission(inode, accmode))) {
		nfsv4_printk(setattr2, "do_setattr: nfs4_inode_permission failed!\n");
		goto out;
	}
	
	if ((status = inode_change_ok(inode, iap))) {
		nfsv4_printk(setattr2, "do_setattr: inode_change_ok failed!\n");
		status = nfserrno(status);
		goto out;
	}

	/* XXX : put in atime/ctime sanity check, a la nfsv3? */
	/* XXX : fix up mode bits? */
	
	if (iap->ia_valid & ATTR_SIZE) {
		if (iap->ia_size < inode->i_size) {
			if ((status = get_write_access(inode))) {
				nfsv4_printk(setattr2, "do_setattr: write access denied!\n");
				status = nfserrno(status);
				goto out;
			}
			write_access = 1;
			DQUOT_INIT(inode);
		}
		down(&inode->i_sem);
		locked = 1;
	}

	/* iap->ia_ctime is set in notify_change() */
	iap->ia_valid |= ATTR_CTIME;
	
	nfsv4_printk(setattr, "do_setattr setting attributes\n");
#ifdef CONFIG_QUOTA
	/* DQUOT_TRANSFER needs both ia_uid and ia_gid defined */
	if (iap->ia_valid & (ATTR_UID|ATTR_GID)) {
		if (! (iap->ia_valid & ATTR_UID))
			iap->ia_uid = inode->i_uid;
		if (! (iap->ia_valid & ATTR_GID))
			iap->ia_gid = inode->i_gid;
		iap->ia_valid |= ATTR_UID|ATTR_GID;
		status = DQUOT_TRANSFER(dentry, iap);
		if (status)
			nfsv4_printk(setattr2, "do_setattr: DQUOT_TRANSFER failed!\n");
	}
	else {
		status = notify_change(dentry, iap);
		if (status)
			nfsv4_printk(setattr2, "do_setattr: notify_change failed!\n");
	}
#else /* CONFIG_QUOTA */
	status = notify_change(dentry, iap);
	if (status)
		nfsv4_printk(setattr2, "do_setattr: notify_change failed!\n");
#endif
	
	if (locked)
		up(&inode->i_sem);
	if (write_access)
		put_write_access(inode);
#if 0
	if (!status && EX_ISSYNC(exp))
		write_inode_now(inode, 1);
#endif
	
out:
	nfsv4_printk(setattr, "do_setattr returning status %d\n", status);
	return status;
}

#else

int
do_setattr(struct svc_fh *fhp, struct iattr *iap)
{
	struct dentry	*dentry;
	struct inode	*inode;
	int		accmode = MAY_SATTR;
	int		imode;
	int		err;
	int		size_change = 0;

	if (iap->ia_valid & (ATTR_ATIME | ATTR_MTIME | ATTR_SIZE))
		accmode |= MAY_WRITE|MAY_OWNER_OVERRIDE;
	
	if ((err = FH_ACCESS(fhp, accmode))) {
		nfsv4_printk(setattr2, "do_setattr: access denied!\n");
		goto out;
	}
	if (iap->ia_valid & ATTR_SIZE) {
		if ((err = FH_FILE(fhp))) {
			nfsv4_printk(setattr2, "do_setattr: size change, not file?!\n");
			goto out;
		}
	}
	if (!iap->ia_valid)
		goto out;
	
	dentry = fhp->fh_dentry;
	inode = dentry->d_inode;

	err = inode_change_ok(inode, iap);
	/* could be a "touch" (utimes) request where the user is not the owner but does
	 * have write permission. In this case the user should be allowed to set
	 * both times to the current time.  We could just assume any such SETATTR
	 * is intended to set the times to "now", but we do a couple of simple tests
	 * to increase our confidence.
	 */
#define BOTH_TIME_SET (ATTR_ATIME_SET | ATTR_MTIME_SET)
#define	MAX_TOUCH_TIME_ERROR (30*60)
	if (err
	    && (iap->ia_valid & BOTH_TIME_SET) == BOTH_TIME_SET
	    && iap->ia_mtime == iap->ia_ctime
	    ) {
	    /* looks good.  now just make sure time is in the right ballpark.
	     * solaris, at least, doesn't seem to care what the time request is
	     */
	    time_t delta = iap->ia_atime - CURRENT_TIME;
	    if (delta<0) delta = -delta;
	    if (delta < MAX_TOUCH_TIME_ERROR) {
		/* turn off ATTR_[AM]TIME_SET but leave ATTR_[AM]TIME
		 * this will cause notify_change to set these times to "now"
		 */
		iap->ia_valid &= ~BOTH_TIME_SET;
		err = inode_change_ok(inode, iap);
	    }
	}
	    
	if (err) {
		nfsv4_printk(setattr2, "do_setattr: inode_change_ok failed!\n");
		goto out_nfserr;
	}

	/* The size case is special. It changes the file as well as the attributes.  */
	if (iap->ia_valid & ATTR_SIZE) {
		if (iap->ia_size < inode->i_size) {
			if ((err = FH_ACCESS(fhp, MAY_TRUNC|MAY_OWNER_OVERRIDE))) {
				nfsv4_printk(setattr2, "do_setattr: can't truncate (acc)!\n");
				goto out;
			}
			err = nfs4_inode_permission(inode, MAY_TRUNC|MAY_OWNER_OVERRIDE);
			if (err) {
				nfsv4_printk(setattr2, "do_setattr: can't truncate (perm)!\n");
				goto out;
			}
		}

		/*
		 * If we are changing the size of the file, then
		 * we need to break all leases.
		 */
		err = get_lease(inode, FMODE_WRITE);
		if (err) {
			nfsv4_printk(setattr2, "do_setattr: get_lease failed!\n");
			goto out_nfserr;
		}

		err = get_write_access(inode);
		if (err) {
			nfsv4_printk(setattr2, "do_setattr: get_write_access failed!\n");
			goto out_nfserr;
		}

		err = locks_verify_truncate(inode, NULL, iap->ia_size);
		if (err) {
			nfsv4_printk(setattr2, "do_setattr: locks_verify_truncate failed!\n");
			put_write_access(inode);
			goto out_nfserr;
		}
		DQUOT_INIT(inode);
	}

	imode = inode->i_mode;
	if (iap->ia_valid & ATTR_MODE) {
		iap->ia_mode &= S_IALLUGO;
		imode = iap->ia_mode |= (imode & ~S_IALLUGO);
	}

	/* Revoke setuid/setgid bit on chown/chgrp */
	if ((iap->ia_valid & ATTR_UID) && (imode & S_ISUID)
	 && iap->ia_uid != inode->i_uid) {
		iap->ia_valid |= ATTR_MODE;
		iap->ia_mode = imode &= ~S_ISUID;
	}
	if ((iap->ia_valid & ATTR_GID) && (imode & S_ISGID)
	 && iap->ia_gid != inode->i_gid) {
		iap->ia_valid |= ATTR_MODE;
		iap->ia_mode = imode &= ~S_ISGID;
	}

	/* Change the attributes. */


	iap->ia_valid |= ATTR_CTIME;
#ifdef CONFIG_QUOTA
	/* DQUOT_TRANSFER needs both ia_uid and ia_gid defined */
	if (iap->ia_valid & (ATTR_UID|ATTR_GID)) {
		if (! (iap->ia_valid & ATTR_UID))
			iap->ia_uid = inode->i_uid;
		if (! (iap->ia_valid & ATTR_GID))
			iap->ia_gid = inode->i_gid;
		iap->ia_valid |= ATTR_UID|ATTR_GID;
	}
#endif /* CONFIG_QUOTA */

	if (iap->ia_valid & ATTR_SIZE) {
		down(&inode->i_sem);
		size_change = 1;
	}
#ifdef CONFIG_QUOTA
	if (iap->ia_valid & (ATTR_UID|ATTR_GID)) 
		err = DQUOT_TRANSFER(dentry, iap);
	else
#endif
		err = notify_change(dentry, iap);
	if (size_change) {
		up(&inode->i_sem);
		put_write_access(inode);
	}
	if (err) {
		nfsv4_printk(setattr2, "do_setattr: notify_change failed!\n");
		goto out_nfserr;
	}
#if 0
	if (EX_ISSYNC(fhp->fh_export))
		write_inode_now(inode, 1);
#endif
	err = 0;
out:
	return err;

out_nfserr:
	err = nfserrno(err);
	goto out;
}

#endif

/*
 *  Readdir-related stuff
 *  Size of a readdir entry = READDIR_ENTRY_STATIC + (bitmap len) + (attrbuf len) + (namelen)
 */

#define READDIR_ENTRY_STATIC	24
#define MINREADDIRBUF		128
#define MAXREADDIRBUF		4096

struct readdir_cd {
       	u32 *			buffer;
	int			buflen;
	u32 *			old_cookie;
	bitmap4			bm4;
	int			attrval_len;   /* obtained by calling get_attrval4_len() */
	struct nfs4_export	*exp;
	struct dentry		*dir_dentry;   /* NULL if pseudo */
	int			eof;
	int			flag;
};

int
nfs4_filldir(void *opaque, const char *name, int namlen,
	     off_t offset, ino_t ino, unsigned int d_type)
{
	struct readdir_cd *cd = (struct readdir_cd *)opaque;
	struct dentry *dentry = NULL;
	char buf[NFS4_MAXNAMLEN + 1];
	u32 *p = cd->buffer;
	u32 *lenp, *nextentry;
	int i;
	int status;
	
	/* server doesn't return '.' or '..' to client */
	if (!name) {
		printk("oh, so this can happen after all..\n");
		status = 0;
		goto out;
	}
	cd->flag = 1;
	if (isdotent(name, namlen)) {
		nfsv4_printk(readdir, "nfs4_filldir(): disregarding entry\n");
		status = 0;
		goto out;
	}
	if (cd->old_cookie)
		*cd->old_cookie = htonl(offset);

	if (cd->attrval_len) {   /* need dentry, to call write_attrs() */
		if (!cd->dir_dentry)
			dentry = DGET(cd->exp->ex_dentry);   /* can be NULL */
		else {
			/* XXX : ugh! what can we do about this? */
			memcpy(buf, name, namlen);
			buf[namlen] = 0;
			/* no need to get cd->dir_dentry->d_inode->i_sem, since it was
			   already obtained in vfs_readdir() */
			dentry = LOOKUP_ONE(buf, cd->dir_dentry);
			if (IS_ERR(dentry)) {
				nfsv4_printk(readdir2, "nfs4_readdir: lookup_one failed!\n");
				status = PTR_ERR(dentry);
				dentry = NULL;
				goto out;
			}
			if (!dentry) {
				status = -ENOMEM;
				goto out;
			}			
		}
	}

	/* XXX : is silent truncation a good idea? */
	if (namlen > NFS4_MAXNAMLEN)
		namlen = NFS4_MAXNAMLEN;
	if ( (READDIR_ENTRY_STATIC + (cd->bm4.bitmap4_len << 2) +
	      XDR_LEN(namlen) + cd->attrval_len) > cd->buflen) {
		cd->eof = 0;
		status = -EINVAL;
		goto out;
	}

	nextentry = p++;        /* previous "nextentry" is filled in last */
	*p++ = 0;               /* first word of cookie */
	cd->old_cookie = p++;   /* second word of cookie, filled in next time filldir() is
				   called */
      
	*p++ = htonl(namlen);   /* name */
	memcpy(p, name, namlen);
	p += XDR_QUADLEN(namlen);

	*p++ = htonl(cd->bm4.bitmap4_len);              /* bitmap */
	for (i = 0; i < cd->bm4.bitmap4_len; i++)
		*p++ = htonl(cd->bm4.bitmap4_val[i]);
	
	if (cd->attrval_len) {                          /* attribute buffer */
		lenp = p++;
		if ((status = write_attrs(cd->exp, dentry, &cd->bm4, p)) < 0) {
			nfsv4_printk(readdir, "nfs4_filldir: write_attrs failed!\n");
			goto out;
		}
		if (status > cd->attrval_len)
			printk("!!! actual buffer length exceeds calculated maximum !!!\n");
		*lenp = htonl(status);
		p += XDR_QUADLEN(status);
	}
	else
		*p++ = 0;

	*nextentry = htonl(1);  /* backfill previous "nextentry" */
	cd->buflen -= ((char *)p - (char *)cd->buffer);
	cd->buffer = p;
	status = 0;
	
out:
	DPUT(dentry);
	nfsv4_printk(readdir, "nfs4_filldir: returning status %d\n", status);
	return status;
}

/* XXX : the dev-based hashtable in struct svc_client kind of sucks */

static int
pseudofs_readdir(u32 cookie, struct readdir_cd *buf)
{
	int offset = 2;   /* to avoid returning "illegal cookies" */
	struct list_head *l;
	struct nfs4_export *exp = buf->exp;
	struct nfs4_export *exp_child;
	int status = NFS4_OK;

	if (cookie > exp->ex_nchild+2) {
		nfsv4_printk(readdir2, "pseudofs_readdir: cookie too large!\n");
		status = NFS4ERR_BAD_COOKIE;
		goto out;
	}
	nfsv4_printk(readdir, "pseudofs_readdir: cookie=%d\n", cookie);

	/* advance to specified offset in directory */
	list_for_each(l, &exp->ex_child) {
		if (offset >= cookie) {
			exp_child = list_entry(l, struct nfs4_export, ex_subdirs);
			buf->exp = exp_child;
			if (nfs4_filldir(buf, exp_child->ex_name, exp_child->ex_namlen, offset,
					 exp_child->ex_id + 1, DT_DIR))
				goto out;
		}
		offset++;
	}
	if (buf->old_cookie)
		*buf->old_cookie = htonl(offset);

out:
	return status;
}

/*
 *  Some helper functions for locking
 */
#if NFSD4_STRICT_LOCKING
static int
posix_lock_file_strict(struct file *filp, struct file_lock *caller)
{
	struct inode *inode = filp->f_dentry->d_inode;
	struct file_lock **flp, *fl;

	for (flp = &inode->i_flock; (fl = *flp); flp = &fl->fl_next) {
		if ((caller->fl_owner != fl->fl_owner) ||
		    (caller->fl_pid != fl->fl_pid))
			break;
		if ((caller->fl_start > fl->fl_end) ||
		    (fl->fl_start > caller->fl_end))
			continue;
		if ((caller->fl_type != F_UNLCK) || (caller->fl_start != fl->fl_start) ||
		    (caller->fl_end != fl->fl_end))
			return -ESERVERFAULT;     /* translated to NFS4ERR_LOCK_RANGE */
	}
	return posix_lock_file(filp, caller, 0);
}
#endif

/*
 * set_lock4_denied() writes (and XDRs) LOCK4denied data into the buffer pointed to by p.
 * It returns the number of bytes written.
 * Caution: must be called with the state lock held!
 */
static int
set_lock4_denied(struct file_lock *fl, u32 *p)
{
        struct nfs4_lockowner *lp = (struct nfs4_lockowner *) fl->fl_owner;
	unsigned int namelen;
        length4 length;
        u32 *start = p;

        if (nfs4_verify_lockowner(lp, fl->fl_pid)) {
                memcpy(p, &lp->lo_client->cl_clientid, sizeof(clientid_t));
                p += 2;
                namelen = lp->lo_namelen;
                *p++ = htonl(namelen);
                memcpy(p, lp->lo_name, namelen);
                p += XDR_QUADLEN(namelen);
        }
        else {
                /* No NFSv4 lockowner matches this lock, so we return clientid 0 and an empty
                   lockowner string to indicate local access. */
                *p++ = 0;
                *p++ = 0;
                *p++ = 0;
        }

        *p++ = htonl((u32)(fl->fl_start >> 32));
        *p++ = htonl((u32)(fl->fl_start));
        if (fl->fl_end == ~(u64)0)
                length = ~(u64)0;
        else
                length = fl->fl_end - fl->fl_start + 1;                 
        *p++ = htonl((u32)(length >> 32));
        *p++ = htonl((u32)(length));
#if TYPE_IN_LOCK_DENIED
        if (fl->fl_type == F_RDLCK)
                *p++ = htonl(READ_LT);
        else
                *p++ = htonl(WRITE_LT);
#endif

        return (char *)p - (char *)start;
}

int
nfs4_access(struct global_state *stp, ACCESS4args *argp, ACCESS4res *resp)
{
	int status;
	struct svc_fh *fhp = &CURRENT_FH;
	struct inode *inode;
	u32 access = 0;
	
	if ((status = FH_INITIALIZED(fhp)))
		goto out;

	if (fhp->fh_dentry) {
		inode = fhp->fh_dentry->d_inode;
		if ((argp->access & ACCESS4_READ) &&
		    !FH_ACCESS(fhp, MAY_READ) &&
		    !nfs4_inode_permission(inode, MAY_READ))
			access |= ACCESS4_READ;
		if ((argp->access & ACCESS4_LOOKUP) &&
		    S_ISDIR(inode->i_mode) &&
		    !FH_ACCESS(fhp, MAY_EXEC) &&
		    !nfs4_inode_permission(inode, MAY_EXEC))
			access |= ACCESS4_LOOKUP;
		if ((argp->access & ACCESS4_MODIFY) &&
		    !FH_ACCESS(fhp, MAY_WRITE) &&
		    !nfs4_inode_permission(inode, MAY_WRITE))
			access |= ACCESS4_MODIFY;
		if ((argp->access & ACCESS4_EXTEND) &&
		    !FH_ACCESS(fhp, MAY_WRITE) &&
		    !nfs4_inode_permission(inode, MAY_WRITE))
			access |= ACCESS4_EXTEND;
		if ((argp->access & ACCESS4_DELETE) &&
		    !FH_ACCESS(fhp, MAY_REMOVE) &&
		    !nfs4_inode_permission(inode, MAY_WRITE|MAY_REMOVE))
			access |= ACCESS4_DELETE;
		if ((argp->access & ACCESS4_EXECUTE) &&
		    S_ISREG(inode->i_mode) &&
		    !FH_ACCESS(fhp, MAY_EXEC) &&
		    !nfs4_inode_permission(inode, MAY_EXEC))
			access |= ACCESS4_EXECUTE;
	}
	else {   /* "internal" pseudofs */
		if ((argp->access & ACCESS4_READ) && !FH_ACCESS(fhp, MAY_READ))
			access |= ACCESS4_READ;
		if ((argp->access & ACCESS4_LOOKUP) && !FH_ACCESS(fhp, MAY_EXEC))
			access |= ACCESS4_LOOKUP;
	}

	resp->SUPPORTED = argp->access & 0x3f;
	resp->ACCESS = access;
	status = NFS4_OK;
	
out:
	nfsv4_printk(fh, "nfs4_access: in=%x supported=%x out=%x\n",
		     argp->access, resp->SUPPORTED, resp->ACCESS);
	return status;
}

int
nfs4_close(struct global_state *stp, CLOSE4args *argp, CLOSE4res *resp)
{
        stateid_t *stateid = (stateid_t *) &argp->stateid;
        struct nfs4_lockowner *lp;
	u32 *start, *p;
        int status;
        
        nfs4_lock_state();

        status = nfs4_process_close(stp, argp->seqid, stateid);
        lp = stp->current_lockowner;
        if (!status) {
                resp->buf = lp->lo_replay_buf;
                start = p = (u32 *)resp->buf;
                memcpy(p, &stp->current_openfile->of_stateid, sizeof(stateid_t));
		p += XDR_QUADLEN(sizeof(stateid_t));
                resp->buflen = lp->lo_replay_buflen = (char *)p - (char *)start;
        }
        else if (status == NFS4ERR_REPLAY_ME) {
                resp->buflen = lp->lo_replay_buflen;
                resp->buf = lp->lo_replay_buf;
                status = NFS4_OK;
        }
        
        nfs4_unlock_state();
        return status;
}

int
nfs4_commit(struct global_state *stp, COMMIT4args *argp, COMMIT4res *resp)
{
	int status;
	struct dentry *dentry;

	/*
	 * XXX: nfsv3 does this by opening a file, calling the ->fsync() method,
	 * then closing the file.  I deviated from this so that we could pass
	 * the chmod(0)+write connectathon test.  Question: Is file_fsync() the
	 * best choice here?  write_inode_now() is faster, but I'm not sure
	 * if it guarantees that the data makes it to the disk, so I decided
	 * to be conservative for now...
	 */
	
	if ((status = FH_FILE(&CURRENT_FH)))
		goto out;
	dentry = CURRENT_FH.fh_dentry;
	file_fsync(NULL, dentry, 0);
	nfs4_set_verifier(resp->COMMITVERF);
	
out:
	return status;
}

int
nfs4_create(struct global_state *stp, CREATE4args *argp, CREATE4res *resp)
{
	int status;
	int len;
	char *name = argp->objname.utf8string_val;
	char buf[NFS4_MAXNAMLEN+1];
	struct dentry *dentry, *dchild;
	struct inode *inode;

	if ((status = FH_EXTERNDIR(&CURRENT_FH)))
		goto out;
	if ((status = FH_ACCESS(&CURRENT_FH, MAY_CREATE)))
		goto out;
	dentry = CURRENT_FH.fh_dentry;
	inode = dentry->d_inode;
	if ((status = nfs4_inode_permission(inode, MAY_CREATE)))
		goto out;
	
	if ((len = argp->objname.utf8string_len) <= 0) {
		status = NFS4ERR_INVAL;
		goto out;
	}
	if (len > NFS4_MAXNAMLEN)
		len = NFS4_MAXNAMLEN;
	if (isdotent(name, len)) {
		/* XXX : need name mangling! */
		status = NFS4ERR_EXIST;
		goto out;
	}

	nfsv4_printk(create, "nfs4_create: name=\"%.*s\", type=%d\n",
		     len, name, argp->objtype.type);
	
	/* XXX : having to do this copy really sucks! */
	memcpy(buf, name, len);
	buf[len] = 0;
	
	nfsv4_printk(create, "nfs4_create calling lookup_one..\n");	
	down(&inode->i_sem);
	set_changeid4(resp->CREATE4res_u.resok4.cinfo.before, dentry->d_inode);
	dchild = LOOKUP_ONE(buf, dentry);
	if (IS_ERR(dchild)) {
		nfsv4_printk(create, "nfs4_create: lookup_one returned error\n");
		status = nfserrno(PTR_ERR(dchild));
		goto out_unlock;
	}
	if (!dchild) {   /* XXX: can this happen? */
		status = NFS4ERR_RESOURCE;
		goto out_unlock;
	}
	/* XXX : lock child?! */
	if (dchild->d_inode) {
		nfsv4_printk(create, "nfs4_create: lookup_one returned non-neg. dentry\n");
		DPUT(dchild);
		status = NFS4ERR_EXIST;
		goto out_unlock;
	}
	switch (argp->objtype.type) {
	case NF4DIR:
		status = vfs_mkdir(dentry->d_inode, dchild, S_IFDIR);
		break;
	case NF4LNK:
		len = argp->objtype.createtype4_u.linkdata.utf8string_len;
		if (len > NFS4_MAXNAMLEN) {
			/* XXX : is this protocol-compliant? */
			nfsv4_printk(create2, "nfs4_create: symlink text too long!\n");
			status = NFS4ERR_INVAL;
			DPUT(dchild);
			goto out_unlock;
		}
		/* XXX : ugh! another unnecessary copy! */
		memcpy(buf, argp->objtype.createtype4_u.linkdata.utf8string_val, len);
		buf[len] = 0;
		status = vfs_symlink(dentry->d_inode, dchild, buf);
		break;
	default:
		nfsv4_printk(create2, "bad type in nfs4_create!");
		status = -EINVAL;   /* translated to NFS4ERR_INVAL before returned */
		break;
	}
	set_changeid4(resp->CREATE4res_u.resok4.cinfo.after, dentry->d_inode);
	up(&inode->i_sem);
	
	if (status) {
		nfsv4_printk(create2, "nfs4_create: vfs create returned error\n");
		status = nfserrno(status);
		DPUT(dchild);
		goto out;
	}
	/* sanity check */
	if (!dchild->d_inode) {
		nfsv4_printk(level1, "nfs4_create(): dentry is still negative?!\n");
		status = NFS4ERR_SERVERFAULT;
		DPUT(dchild);
		goto out;
	}

#if 0
	/* XXX : this code fragment from nfsv3.. is it needed? */
	if (EX_ISSYNC(exp)) {
		nfsd_sync_dir(dentry);
		write_inode_now(dchild->d_inode, 1);
	}
#endif
	
	/* update fh -- no export change */
	resp->CREATE4res_u.resok4.cinfo.atomic = 1;
	DPUT(dentry);
	CURRENT_FH.fh_dentry = dchild;

out:
	return status;
out_unlock:
	up(&inode->i_sem);
	goto out;
}

int
nfs4_delegpurge(struct global_state *stp, DELEGPURGE4args *argp, DELEGPURGE4res *resp)
{
	return NFS4ERR_NOTSUPP;
}

int
nfs4_delegreturn(struct global_state *stp, DELEGRETURN4args *argp,
		 DELEGRETURN4res *resp)
{
	return NFS4ERR_NOTSUPP;
}

int
nfs4_getattr(struct global_state *stp, GETATTR4args *argp, GETATTR4res *resp)
{
	int status;
	int len;
	u32 *p;

	if ((status = FH_INITIALIZED(&CURRENT_FH)))
		goto out;
	/* XXX : need permission check? */

	nfsv4_printk(getattr, "nfs4_getattr() called with bm ");
	nfsv4_hexdump(getattr, argp->attr_request.bitmap4_val,
		      argp->attr_request.bitmap4_len << 2);
	nfsv4_printk(getattr, "current fh = ");
	nfsv4_print_svcfh(getattr, &CURRENT_FH);
	
	if ((status = set_getattr_bitmap(&resp->GFATTR4.attrmask, &argp->attr_request)))
		goto out;
	if ((len = get_attrval4_len(&resp->GFATTR4.attrmask)) < 0) {
		nfsv4_printk(level1, "nfs4_getattr: get_attrval4_len failed!\n");
		status = NFS4ERR_SERVERFAULT;
		goto out_free_bm;
	}
	if (!len) {
		p = NULL;
		goto done;
	}
	if (!(p = NFS4_ALLOC(len))) {
		nfsv4_printk(level1, "failed to alloc attr buf in nfs4_getattr!\n");
		status = NFS4ERR_RESOURCE;
		goto out_free_bm;
	}

	status = write_attrs(CURRENT_FH.fh_export, CURRENT_FH.fh_dentry,
			     &resp->GFATTR4.attrmask, p);
	if (status < 0) {
		status = -status;
		goto out_free;
	}
	
done:
	resp->GFATTR4.attr_vals.attrlist4_len = status;
	resp->GFATTR4.attr_vals.attrlist4_val = (char *) p;
	status = NFS4_OK;

out:
	nfsv4_printk(getattr, "nfs4_getattr() returning status %d\n", status);
	return status;
out_free:
	NFS4_FREE(p);
out_free_bm:
	NFS4_FREE(resp->GFATTR4.attrmask.bitmap4_val);
	goto out;
}

int
nfs4_getfh(struct global_state *stp, GETFH4res *resp)
{
	int status;
	u32 *p;

	if ((status = FH_INITIALIZED(&CURRENT_FH)))
		goto out;
	if (!(p = NFS4_ALLOC(NONPSEUDO_FH_LEN))) {
		status = NFS4ERR_RESOURCE;
		goto out;
	}
	status = fh_write(CURRENT_FH.fh_export, CURRENT_FH.fh_dentry, p);
	if (status < 0) {
		status = -status;
		NFS4_FREE(p);
		goto out;
	}
	resp->GETFH4FH.nfs_fh4_len = status;
	resp->GETFH4FH.nfs_fh4_val = (char *)p;
	status = 0;
	
out:
	return status;
}

int
nfs4_link(struct global_state *stp, LINK4args *argp, LINK4res *resp)
{
	int status;
	int namlen = argp->newname.utf8string_len;
	char *name = argp->newname.utf8string_val;
	char buf[NFS4_MAXNAMLEN+1];
	struct dentry *d_dir, *d_old, *d_new;
	struct inode *i_dir;
	struct change_info4 *cinfo = &resp->LINK4res_u.resok4.cinfo;

	/* XXX : should we allow cross-export hardlinks to be created? */
	
	if ((status = FH_EXTERN(&SAVE_FH)))
		goto out;
	d_old = SAVE_FH.fh_dentry;
	if (S_ISDIR(d_old->d_inode->i_mode)) {
		status = NFS4ERR_ISDIR;
		goto out;
	}
	if ((status = FH_EXTERNDIR(&CURRENT_FH)))
		goto out;
	if ((status = FH_ACCESS(&CURRENT_FH, MAY_CREATE)))
		goto out;
	d_dir = CURRENT_FH.fh_dentry;
	i_dir = d_dir->d_inode;
	if ((status = nfs4_inode_permission(i_dir, MAY_CREATE)))
		goto out;

	if (namlen < 0) {
		status = NFS4ERR_INVAL;
		goto out;
	}
	if (namlen > NFS4_MAXNAMLEN)
		namlen = NFS4_MAXNAMLEN;
		
	nfsv4_printk(link, "nfs4_link() starting in dir ");
	nfsv4_print_svcfh(link, &CURRENT_FH);
	nfsv4_printk(link, "creating link \"%.*s\" to file ", namlen, name);
	nfsv4_print_svcfh(link, &SAVE_FH);

	if (isdotent(name, namlen)) {
		/* XXX : need name mangling! */
		status = NFS4ERR_EXIST;
		goto out;
	}
	/* XXX : this copy really sucks! */
	memcpy(buf, name, namlen);
	buf[namlen] = 0;
	
	down(&i_dir->i_sem);
	d_new = LOOKUP_ONE(buf, d_dir);
	if (IS_ERR(d_new)) {
		nfsv4_printk(link2, "nfs4_link: lookup_one returned error!\n");
		status = nfserrno(PTR_ERR(d_new));
		goto out_unlock;
	}
	if (!d_new) {   /* XXX : can this happen? */
		status = NFS4ERR_RESOURCE;
		goto out_unlock;
	}
	set_changeid4(cinfo->before, i_dir);
	if ((status = vfs_link(d_old, i_dir, d_new))) {
		nfsv4_printk(link2, "nfs4_link(): vfs_link returned error!\n");
		status = nfserrno(status);
		goto out_dput;
	}
	
#if 0
	if (EX_ISSYNC(CURRENT_FH.fh_export)) {
		nfsd_sync_dir(d_dir);
		write_inode_now(d_new->d_inode, 1);
	}
#endif
	
	set_changeid4(cinfo->after, i_dir);
	cinfo->atomic = 1;
	
out_dput:
	DPUT(d_new);
out_unlock:
	up(&i_dir->i_sem);
out:
	nfsv4_printk(link, "nfs4_link() returning status %d\n", status);
	return status;
}

static inline void
hack_lock(struct file_lock *lock)
{
        /* XXX: for some perverse reason, Linux uses signed fields in struct file_lock, so
           we have to resort to the following ugliness.  Figure out how to fix this! */
        if (lock->fl_start < 0)
                lock->fl_start = OFFSET_MAX;
        if (lock->fl_end < 0)
                lock->fl_end = OFFSET_MAX;
}

int
nfs4_lock(struct global_state *stp, LOCK4args *argp, LOCK4res *resp)
{
        struct inode *inode;
        struct file_lock lock;
        struct nfs4_openfile *ofp;
        struct nfs4_lockowner *lp;
        struct file *filp;
        struct file_lock *conflicting_lock = NULL;
        stateid_t *stateid = (stateid_t *) &argp->stateid;
        u32 *start, *p;
        int status;

        nfs4_lock_state();

        status = NFS4ERR_SERVERFAULT;
        if (argp->reclaim) {
                nfsv4_printk(lock2, "nfs4_lock: reclaim not supported yet!\n");
                goto out;
        }
        status = NFS4ERR_GRACE;
        if (nfs4_in_grace_period()) {
                nfsv4_printk(lease2, "nfs4_lock: in grace period!\n");
                goto out;
        }
        if ((status = nfs4_preprocess_seqid_op(stp, argp->seqid, stateid))) {
		nfsv4_printk(lock2, "nfs4_lock: process_seqid failed!\n");
                goto out_check_replay;
	}
        
        /* XXX: is the permission test here correct? */
        if ((status = FH_ACCESS(&CURRENT_FH, MAY_LOCK))) {
		nfsv4_printk(lock2, "nfs4_lock: access denied!\n");
                goto out;
	}
	inode = CURRENT_FH.fh_dentry->d_inode;
        if ((status = nfs4_inode_permission(inode, MAY_LOCK))) {
		nfsv4_printk(lock2, "nfs4_lock: permission denied!\n");
                goto out;
	}
        
        ofp = stp->current_openfile;
        lp = ofp->of_lockowner;
        filp = &ofp->of_vfs_file;
        
        switch (argp->locktype) {
        case READ_LT:
        case READW_LT:
                lock.fl_type = F_RDLCK;
                break;
        case WRITE_LT:
        case WRITEW_LT:
                lock.fl_type = F_WRLCK;
                break;
        default:
                status = NFS4ERR_INVAL;
                goto out;
        }
        lock.fl_owner = (fl_owner_t) lp;
        lock.fl_pid = lp->lo_hashval;
        lock.fl_file = filp;
        lock.fl_flags = FL_POSIX;
        lock.fl_notify = NULL;
        lock.fl_insert = NULL;
        lock.fl_remove = NULL;
        lock.fl_start = argp->offset;   
        if ((argp->length == ~(u64)0) || LOFF_OVERFLOW(argp->offset, argp->length))
                lock.fl_end = ~(u64)0;
        else
                lock.fl_end = argp->offset + argp->length - 1;

        hack_lock(&lock);

        nfsv4_printk(lock, "nfs4_lock: owner=%p type=%d start=0x%Lx end=0x%Lx\n",
                     lp, lock.fl_type, lock.fl_start, lock.fl_end);
        
        /*
         *  Try to lock the file.
         */
        lock_kernel();
        status = posix_lock_file(filp, &lock, 0);
        if (status == -EAGAIN) {        /* conflicting lock */
		nfsv4_printk(lock, "nfs4_lock: conflicting lock exists\n");
                goto out_conflicting_lock;
	}
        unlock_kernel();
        if (status) {
		nfsv4_printk(lock2, "nfs4_lock: posix_lock_file() failed!\n");
                goto out_nfserr;
	}
	
        /* looks like the LOCK succeeded; update seqid and stateid, and write out the
           response buffer. */
        update_stateid(&ofp->of_stateid);
        resp->buf = CURRENT_LOCKOWNER->lo_replay_buf;
        start = p = (u32 *)resp->buf;
        memcpy(p, &ofp->of_stateid, sizeof(stateid_t));
	p += XDR_QUADLEN(sizeof(stateid_t));
        CURRENT_LOCKOWNER->lo_replay_buflen = resp->buflen = (char *)p - (char *)start;
        CURRENT_LOCKOWNER->lo_seqid = argp->seqid;
        status = NFS4_OK;
        
out:
        nfs4_unlock_state();
        return status;
out_check_replay:
	if (status != NFS4ERR_REPLAY_ME)
		goto out;
	nfsv4_printk(lock2, "nfs4_lock: retransmission?!\n");
        resp->buflen = CURRENT_LOCKOWNER->lo_replay_buflen;
        resp->buf = CURRENT_LOCKOWNER->lo_replay_buf;
        status = NFS4_OK;
        goto out;
out_conflicting_lock:
        conflicting_lock = __posix_test_lock(inode, &lock);
        unlock_kernel();
        nfsv4_printk(lock, "nfs4_lock: conflicting lock exists\n");
        resp->buflen = set_lock4_denied(conflicting_lock, (u32 *)resp->deniedbuf);
        resp->buf = resp->deniedbuf;
        status = NFS4ERR_DENIED;
        goto out;
out_nfserr:
        status = nfserrno(status);
        goto out;
}

int
nfs4_lockt(struct global_state *stp, LOCKT4args *argp, LOCKT4res *resp)
{
        struct file_lock lock;
        struct file_lock *conflicting_lock;
        struct inode *inode;
        clientid_t *clientid = (clientid_t *) &argp->owner.clientid;
        unsigned int namelen = argp->owner.owner.owner_len;
        char *name = argp->owner.owner.owner_val;
        struct nfs4_lockowner *lp;
        int status;

        nfsv4_printk(lock, "nfs4_lockt: starting\n");
        nfs4_lock_state();
        
        status = NFS4ERR_GRACE;
        if (nfs4_in_grace_period()) {
                nfsv4_printk(lock2, "nfs4_lockt: in grace period!\n");
                goto out;
        }
        if ((status = FH_FILE(&CURRENT_FH))) {
                nfsv4_printk(lock2, "nfs4_lockt: filehandle check failed!\n");
                goto out;
        }
        inode = CURRENT_FH.fh_dentry->d_inode;

        switch (argp->locktype) {
        case READ_LT:
        case READW_LT:
                lock.fl_type = F_RDLCK;
                break;
        case WRITE_LT:
        case WRITEW_LT:
                lock.fl_type = F_WRLCK;
                break;
        default:
                nfsv4_printk(lock2, "nfs4_lockt: bad lock type!\n");
                status = -EINVAL;
                goto out;
        }
	
        if ((status = nfs4_lookup_lockowner(stp, clientid, namelen, name))) {
		nfsv4_printk(lock2, "nfs4_lockt: lookup_lockowner failed!\n");
                goto out;
	}
        lp = stp->current_lockowner;
        if (lp) {
                lock.fl_owner = (fl_owner_t) lp;
                lock.fl_pid = lp->lo_hashval;
        }
        else {
                lock.fl_owner = NULL;
                lock.fl_pid = 0;
        }
        lock.fl_flags = FL_POSIX;
        lock.fl_start = argp->offset;
        if (LOFF_OVERFLOW(argp->offset, argp->length))
                lock.fl_end = ~(u64)0;
        else
                lock.fl_end = argp->offset + argp->length - 1;

        hack_lock(&lock);
        
        nfsv4_printk(lock, "nfs4_lockt: owner=%p type=%d start=0x%Lx end=0x%Lx\n",
                     lock.fl_owner, lock.fl_type, lock.fl_start, lock.fl_end);

        status = NFS4_OK;
        conflicting_lock = __posix_test_lock(inode, &lock);
        if (conflicting_lock) {
                nfsv4_printk(lock, "nfs4_lockt: conflicting lock exists\n");
                resp->buflen = set_lock4_denied(conflicting_lock, (u32 *)resp->deniedbuf);
                status = NFS4ERR_DENIED;
                goto out;
        }
        
out:
        nfs4_unlock_state();
        return status;
}

int
nfs4_locku(struct global_state *stp, LOCKU4args *argp, LOCKU4res *resp)
{
        struct nfs4_openfile *ofp;
        struct nfs4_lockowner *lp;
        struct file *filp;
        struct file_lock lock;
	stateid_t *stateid = (stateid_t *) &argp->stateid;
        u32 *start, *p;
        int status;

        nfsv4_printk(stateid, "nfs4_locku: starting\n");
	
        nfs4_lock_state();
        
	/* XXX: is it OK to let LOCKU succeed during the grace period? */
        if ((status = nfs4_preprocess_seqid_op(stp, argp->seqid, stateid))) {
                nfsv4_printk(lock2, "nfs4_locku: preprocess_seqid failed!\n");
                goto out_check_replay;
        }
        lp = stp->current_lockowner;
	ofp = stp->current_openfile;
        filp = &ofp->of_vfs_file;
        
        lock.fl_type = F_UNLCK;
        lock.fl_owner = (fl_owner_t) lp;
        lock.fl_pid = lp->lo_hashval;
        lock.fl_file = filp;
        lock.fl_flags = FL_POSIX;    /* XXX: not sure if all these need to be set... */
        lock.fl_notify = NULL;
        lock.fl_insert = NULL;
        lock.fl_remove = NULL;
        lock.fl_start = argp->offset;
        if ((argp->length == ~(u64)0) || LOFF_OVERFLOW(argp->offset, argp->length))
                lock.fl_end = ~(u64)0;
        else
                lock.fl_end = argp->offset + argp->length - 1;

        hack_lock(&lock);
        
        nfsv4_printk(lock, "nfs4_locku: owner=%p type=%d start=0x%Lx end=0x%Lx\n",
                     lock.fl_owner, lock.fl_type, lock.fl_start, lock.fl_end);
        
        /*
         *  Try to unlock the file.
         */
        status = posix_lock_file(filp, &lock, 0);
        if (status) {
		nfsv4_printk(lock2, "nfs4_locku: posix_lock_file failed!\n");
                goto out_nfserr;
	}

        /* looks like the LOCKU succeeded; update seqid and stateid, and write out the
           response buffer. */
        update_stateid(&ofp->of_stateid);
        resp->buf = CURRENT_LOCKOWNER->lo_replay_buf;
        start = p = (u32 *)resp->buf;
        memcpy(p, &ofp->of_stateid, sizeof(stateid_t));
	p += XDR_QUADLEN(sizeof(stateid_t));
        CURRENT_LOCKOWNER->lo_replay_buflen = resp->buflen = (char *)p - (char *)start;
        CURRENT_LOCKOWNER->lo_seqid = argp->seqid;
        status = NFS4_OK;
        
out:
        nfs4_unlock_state();
        return status;
out_check_replay:
	if (status != NFS4ERR_REPLAY_ME)
		goto out;
	nfsv4_printk(lock2, "nfs4_locku: retransmission?!\n");
        resp->buflen = CURRENT_LOCKOWNER->lo_replay_buflen;
        resp->buf = CURRENT_LOCKOWNER->lo_replay_buf;
        status = NFS4_OK;
        goto out;
out_nfserr:
        status = nfserrno(status);
        goto out;
}

int
nfs4_lookup(struct global_state *stp, LOOKUP4args *argp, LOOKUP4res *resp)
{
	int i;
	int status = 0;

	if ((status = FH_INITIALIZED(&CURRENT_FH)))
		goto out;
	for (i = 0; i < argp->path.pathname4_len; i++)
		if ((status = do_lookup(stp, argp->path.pathname4_val[i].utf8string_len,
					argp->path.pathname4_val[i].utf8string_val)))
			goto out;
	
out:
	return status;
}

int
nfs4_lookupp(struct global_state *stp, LOOKUPP4res *resp)
{
	struct dentry *dentry;
	struct nfs4_export *exp;
	int status;
	
	if ((status = FH_INITIALIZED(&CURRENT_FH)))
		goto out;
	/* XXX : no permission check for 'cd ..' ? */

	exp = CURRENT_FH.fh_export;
	dentry = CURRENT_FH.fh_dentry;
	if (dentry) {
		if (dentry == exp->ex_dentry) {
			DPUT(dentry);
			goto pseudo;
		}
		/* XXX : can't cross a mountpoint this way.. fix soon! */
		CURRENT_FH.fh_dentry = DGET(dentry->d_parent);
		DPUT(dentry);
		status = NFS4_OK;
		goto out;
	}
	
pseudo:
	CURRENT_FH.fh_export = exp->ex_parent;
	CURRENT_FH.fh_dentry = DGET(exp->ex_parent->ex_dentry);
	if ((status = nfs4_setsec(stp->rqstp, exp->ex_parent, &CURRENT_FH.fh_secattrs))) {
		nfsv4_printk(lookup2, "nfs4_lookupp: nfs4_setsec failed!\n");
		goto out;
	}
	nfs4_setuser(&CURRENT_FH);
	
out:
	return status;
}

int
nfs4_nverify(struct global_state *stp, NVERIFY4args *argp, NVERIFY4res *resp)
{
	int status;
	u32 dummy;
	bitmap4 *bm = &argp->obj_attributes.attrmask;
	attrlist4 *attrs = &argp->obj_attributes.attr_vals;
	u32 *buf;
	int i;
	int len;

	if ((status = FH_INITIALIZED(&CURRENT_FH)))
		goto out;
	
	for (i = 0; i < bm->bitmap4_len; i++) {
		dummy = (i < GETATTR_BITMAP_LEN) ? GETATTR_BITMAP[i] : 0;
		if (bm->bitmap4_val[i] & ~dummy) {
			nfsv4_printk(getattr2, "nfs4_nverify: bad bits set!\n");
			status = NFS4ERR_INVAL;
			goto out;
		}
	}
	if ((len = get_attrval4_len(bm)) < 0) {
		nfsv4_printk(level1, "nfs4_nverify: get_attrval4_len failed!\n");
		status = NFS4ERR_SERVERFAULT;
		goto out;
	}
	if (!(buf = NFS4_ALLOC(len))) {
		nfsv4_printk(level1, "nfs4_nverify: failed to alloc buffer!\n");
		status = NFS4ERR_RESOURCE;
		goto out;
	}
	status = write_attrs(CURRENT_FH.fh_export, CURRENT_FH.fh_dentry, bm, buf);
	if (status < 0) {
		nfsv4_printk(getattr2, "nfs4_nverify: write_attrs returned error!\n");
		status = -status;
		goto out_free;
	}
	if (status != attrs->attrlist4_len) {
		/* attrs are different */
		status = NFS4_OK;
		goto out_free;
	}
	if (memcmp(buf, attrs->attrlist4_val, status)) {
		/* attrs are different */
		status = NFS4_OK;
		goto out_free;
	}
	status = NFS4ERR_SAME;

out_free:
	NFS4_FREE(buf);
out:
	return status;
}

int
nfs4_open(struct global_state *stp, OPEN4args *argp, OPEN4res *resp)
{
        clientid_t *clientid = (clientid_t *) &argp->owner.clientid;
        int createtype;
        struct iattr iattr;
        unsigned int plen;
        utf8string *pval;
        unsigned int namelen;
        char *name;
	u32 v_mtime, v_atime;
        char buf[NFS4_MAXNAMLEN + 1];
        struct dentry *dfile;
        struct inode *dir_inode;
	struct inode *inode;
        u32 pre_changeid, post_changeid;
        int imode;
        unsigned int share_access = argp->share_access;
        unsigned int share_deny = argp->share_deny;
	int accmode;
        int truncate = 0;
        nfs4_ino_desc_t ino;
        struct nfs4_lockowner *lp;
        struct nfs4_openfile *ofp;
        u32 *start, *p;
        int i;
        int status;

        nfs4_lock_state();

        /*
         *  1. Parse the file creation settings in argp->openhow.
         *    The variable 'createtype' is set to one of -1 (indicating no creation),
         *    UNCHECKED4, GUARDED4, or EXCLUSIVE4.  Creation attributes, if any, are
         *    stored in the variable 'iattr'.
         */
        switch (argp->openhow.opentype) {
        case OPEN4_NOCREATE:
                createtype = -1;
                iattr.ia_valid = 0;
                break;
        case OPEN4_CREATE:
                createtype = argp->openhow.HOW.mode;
                switch (createtype) {
                case UNCHECKED4:
                case GUARDED4:
                        if ((status = read_attrs(&argp->openhow.HOW.CREATE_ATTR, &iattr))) {
                                nfsv4_printk(open2, "nfs4_open: read_attrs() failed!\n");
                                goto out;
                        }
                        break;
                case EXCLUSIVE4:
			memcpy(&v_atime, &argp->openhow.HOW.CREATE_VERF[0], 4);
			memcpy(&v_mtime, &argp->openhow.HOW.CREATE_VERF[4], 4);
			iattr.ia_valid = 0;
			break;
                default:
                        nfsv4_printk(open2, "nfs4_open: bad createtype %d!\n", createtype);
                        status = NFS4ERR_INVAL;
                        goto out;
                }
                break;
        default:
                nfsv4_printk(open2, "nfs4_open: bad opentype %d\n", argp->openhow.opentype);
                status = NFS4ERR_INVAL;
                goto out;
        }

        /*
         *  2. Validate the clientid, lockowner, and seqid.  This is done in
         *    nfs4_process_open1().  If nfs4_process_open1() detects a replay,
         *    it returns the special "error" NFS4ERR_REPLAY_ME, to inform the
         *    caller.
         */
        status = nfs4_process_open1(stp, clientid, argp->owner.owner.owner_val,
                                    argp->owner.owner.owner_len, argp->seqid);
        if (status)
                goto out_check_replay;

        /*
         *  3. Set 'dfile' to the dentry of the file being opened, and set
         *    pre_changeid, post_changeid.
         *    Note: under a CLAIM_PREVIOUS open, the current filehandle represents
         *    the file being opened, not the containing directory.
         *    XXX: for the list: why not CLAIM_DELEGATE_PREV?
         */
        switch (argp->claim.claim) {
        case CLAIM_PREVIOUS:
                /* XXX: what if one of these comes in well after the end of the
                 * grace period?
                 */
                status = NFS4ERR_INVAL;
                if (createtype >= 0) {
                        nfsv4_printk(open2, "nfs4_open: CLAIM_PREVIOUS, not NOCREATE?!");
                        goto out;
                }
                if ((status = FH_EXTERN(&CURRENT_FH))) {
                        nfsv4_printk(open2, "nfs4_open: no filehandle, or pseudofh!\n");
                        goto out;
                }
                dfile = CURRENT_FH.fh_dentry;
		inode = dfile->d_inode;
                pre_changeid = post_changeid = dfile->d_parent->d_inode->i_mtime;
                goto permission_checks;
                
        case CLAIM_NULL:
                status = NFS4ERR_GRACE;
                if (nfs4_in_grace_period()) {
                        nfsv4_printk(lease2, "nfs4_open: in grace period!\n");
                        goto out;
                }
                if ((status = FH_INITIALIZED(&CURRENT_FH)))
                        goto out;
                
                /* Look up intermediate path components, resetting CURRENT_FH along the way.
                 * Note that this is slightly inefficient for multi-component paths, but those
                 * don't seem to occur in practice anyway.
                 */
		plen = argp->claim.open_claim4_u.file.pathname4_len;
                pval = argp->claim.open_claim4_u.file.pathname4_val;
                for (i = 0; i < plen-1; i++) {
                        namelen = pval[i].utf8string_len;
                        name = pval[i].utf8string_val;
                        if ((status = do_lookup(stp, namelen, name))) {
                                nfsv4_printk(open2, "nfs4_open: do_lookup failed!\n");
                                goto out;
                        }
                }
                namelen = pval[i].utf8string_len;
                name = pval[i].utf8string_val;
                status = NFS4ERR_NAMETOOLONG;
                if (namelen > NFS4_MAXNAMLEN) {
                        nfsv4_printk(open2, "nfs4_open: name too long!\n");
                        goto out;
                }
                if ((status = FH_EXTERNDIR(&CURRENT_FH)))
                        goto out;
                if ((status = FH_ACCESS(&CURRENT_FH, MAY_EXEC)))
                        goto out;
                dir_inode = CURRENT_FH.fh_dentry->d_inode;
                if ((status = nfs4_inode_permission(dir_inode, MAY_EXEC)))
                        goto out;

               /* XXX: ugh! we have to copy the name to get the null-terminator! */
                memcpy(buf, name, namelen);
                buf[namelen] = 0;
                
                /* Now lock the directory, and try to look up the file. */
                down(&dir_inode->i_sem);
                pre_changeid = dir_inode->i_mtime;
                dfile = LOOKUP_ONE(buf, CURRENT_FH.fh_dentry);          
                if (IS_ERR(dfile)) {
                        nfsv4_printk(open2, "nfs4_open: lookup_one failed!\n");
                        goto out_up_ptrerr;
                }
                if (!dfile) {                            /* XXX : can this happen? */
                        nfsv4_printk(open2, "nfs4_open: couldn't alloc dentry!\n");
                        goto out_up;
                }
                nfsv4_printk(open, "nfs4_open: file lookup succeeded, file %s exist\n",
                             dfile->d_inode ? "does" : "does not");
                
               if (dfile->d_inode) {                  /* file exists */
                        if ((createtype <= 0) ||         /* NOCREATE or UNCHECKED4 */
			    (createtype == EXCLUSIVE4 && (dfile->d_inode->i_atime == v_atime &&
							  dfile->d_inode->i_mtime == v_mtime &&
							  dfile->d_inode->i_size == 0))) {
                                post_changeid = dir_inode->i_mtime;
                                up(&dir_inode->i_sem);
                                DPUT(CURRENT_FH.fh_dentry);
                                CURRENT_FH.fh_dentry = dfile;
				inode = dfile->d_inode;
                                if (iattr.ia_valid & ATTR_SIZE) {
                                        status = NFS4ERR_INVAL;   /* XXX: ERR_EXIST? */
                                        if (iattr.ia_size) {
                                                nfsv4_printk(open2, "nfs4_open: size != 0!\n");
                                                goto out;
                                        }
                                        truncate = 1;
                                }
                                goto permission_checks;
                                
                        }
                        nfsv4_printk(open2, "nfs4_open: GUARDED open, file exists!\n");
                        status = NFS4ERR_EXIST;
                        goto out_up_dput;
                }

                /* If we reach this point, the file does not exist */
                status = NFS4ERR_NOENT;
                if (createtype < 0) {
                        nfsv4_printk(open2, "nfs4_open: NOCREATE open, file doesn't exist!\n");
                        goto out_up_dput;
                }
                /* Try to create the file */
                if ((status = FH_ACCESS(&CURRENT_FH, MAY_CREATE))) {
                        nfsv4_printk(open2, "nfs4_open: don't have create access!\n");
                        goto out_up_dput;
                }
                if ((status = nfs4_inode_permission(dir_inode, MAY_CREATE))) {
                        nfsv4_printk(open2, "nfs4_open: don't have create permission!\n");
                        goto out_up_dput;
                }
                imode = S_IFREG;   /* XXX: use mode 000 by default? */
                if (iattr.ia_valid | ATTR_MODE) {
                        imode |= iattr.ia_mode & (S_IALLUGO|S_ISUID|S_ISGID|S_ISVTX);
                        iattr.ia_valid &= ~ATTR_MODE;
                }
                status = vfs_create(dir_inode, dfile, imode);
                post_changeid = dir_inode->i_mtime;
                up(&dir_inode->i_sem);
                if (status) {
                        nfsv4_printk(open2, "nfs4_open: vfs_create() failed!\n");
                        goto out_dput_nfserr;
                }
#if NFS4_DEBUG
                /* sanity check */
                status = NFS4ERR_SERVERFAULT;
                if (!dfile->d_inode) {
                        printk("!!! nfsd4: no inode after file create !!!\n");
                        dput(dfile);
                        goto out;
                }
#endif
		DPUT(CURRENT_FH.fh_dentry);
		CURRENT_FH.fh_dentry = dfile;
		inode = dfile->d_inode;
                nfsv4_printk(open, "nfs4_open: file created successfully\n");

                /* Set additional attributes, if necessary.  XXX: make sure more permission
                 * checks aren't needed here!
                 */
                if (!iattr.ia_size)     /* optimization: newly-created file has size 0 */
                        iattr.ia_valid &= ~ATTR_SIZE;
		if (createtype == EXCLUSIVE4) {
			iattr.ia_atime = v_atime;
			iattr.ia_mtime = v_mtime;
			iattr.ia_valid = (ATTR_MTIME | ATTR_ATIME |
					  ATTR_MTIME_SET | ATTR_MTIME_SET);
		}
                if (iattr.ia_valid) {
                        if ((status = do_setattr(&CURRENT_FH, &iattr))) {
                                nfsv4_printk(open2, "nfs4_open: do_setattr() failed!\n");
                                goto out;
                        }
                }
                goto do_open;    /* skip permission checks.  XXX: make sure this is OK! */
                
        case CLAIM_DELEGATE_CUR:
        case CLAIM_DELEGATE_PREV:
                status = NFS4ERR_SERVERFAULT;
                nfsv4_printk(level1, "nfs4_open: CLAIM_DELEGATE_* not supported yet!\n");
                goto out;

        default:
                status = NFS4ERR_INVAL;
                nfsv4_printk(open2, "nfs4_open: bad claim type!\n");
                goto out;
        }

        /*
         *  4. Do some permission checks before opening the file in the VFS (we will skip
         *    this step if we created the file.  XXX: is this always ok?).  'dfile' must
         *    be a positive dentry, equal to CURRENT_FH.fh_dentry.  Note the similarity
         *    with the permission checks in fs/namei.c:open_namei().
         */
permission_checks:
        imode = inode->i_mode;
        status = NFS4ERR_ISDIR;
        if (S_ISDIR(imode)) {
                nfsv4_printk(open2, "nfs4_open: file is a directory!\n");
                goto out;
        }
        status = NFS4ERR_SYMLINK;
        if (S_ISLNK(imode)) {
                nfsv4_printk(open2, "nfs4_open: file is a symlink!\n");
                goto out;
        }
        /* XXX: for now, we don't allow an OPEN to succeed on a named pipe, socket, or
         * device special file.
         */
        status = NFS4ERR_SERVERFAULT;
	if (!S_ISREG(imode)) {
                nfsv4_printk(open2, "nfs4_open: not a regular file!\n");
                goto out;
        }

        accmode = 0;
        if (share_access & OPEN4_SHARE_ACCESS_READ)
                accmode = MAY_READ;
        if (share_access & OPEN4_SHARE_ACCESS_WRITE)
                accmode |= MAY_WRITE;
        if ((status = FH_ACCESS(&CURRENT_FH, accmode))) {
                nfsv4_printk(open2, "nfs4_open: access denied!\n");
                goto out;
        }
        if ((status = permission(inode, accmode))) {
                nfsv4_printk(open2, "nfs4_open: permission() failed!\n");
                goto out_nfserr;
        }

        /* We don't allow an OPEN to succeed on an append-only inode.
         * XXX: is this right?  how do append-only inodes live and die, anyway? (ask jim)
         */
        status = NFS4ERR_PERM;
        if (IS_APPEND(inode)) {
                nfsv4_printk(open2, "nfs4_open: inode is append-only!\n");
                goto out;
        }
        
        /* XXX: need to learn about kernel leases, and check for them here... */

        if (truncate) {
                status = NFS4ERR_INVAL;
                if (!(share_access & OPEN4_SHARE_ACCESS_WRITE)) {
                        nfsv4_printk(open2, "nfs4_open: truncating without writing!\n");
                        goto out;
                }
#if 0
                /* XXX: should we check for locks?  This means we need to settle the issue
                   of mandatory locks in NFSv4... (ask jim) */
                status = locks_verify_locked(inode);
#endif
        }
        
        /*
         *  5. We check for share reservation conflicts, then open the file in the
         *    VFS.  This is done in nfs4_process_open2(), which sets
         *    stp->current_openfile.  'dfile' must be a positive regular file,
         *    equal to CURRENT_FH.fh_dentry.
         */
do_open:
        ino.dev = inode->i_dev;
        ino.ino = inode->i_ino;
        ino.generation = inode->i_generation;
        if ((status = nfs4_process_open2(stp, &ino, truncate, share_access,
					 share_deny, argp->seqid)))
                goto out;

        /*
         *  6. If we made it this far, the operation has succeeded.  We write the
         *    response data into the lockowner replay buffer and return.
         *    Note: nfs4_process_open2() has already incremented the seqid.
         *    XXX: there are some issues with mandatory locking to think about here!
         */
        lp = stp->current_lockowner;
        ofp = stp->current_openfile;
        resp->buf = lp->lo_replay_buf;
        start = p = (u32 *)resp->buf;
        memcpy(p, &ofp->of_stateid, sizeof(stateid_t));
	p += XDR_QUADLEN(sizeof(stateid_t));
	*p++ = htonl(1);
        *p++ = 0;
        *p++ = htonl(pre_changeid);
        *p++ = 0;
        *p++ = htonl(post_changeid);
        if (lp->lo_confirmed) {
                *p++ = 0;
                *p++ = 0;
                *p++ = 0;
        }
        else {
                *p++ = htonl(OPEN4_RESULT_CONFIRM);
                memcpy(p, &lp->lo_verifier, sizeof(verifier4));
                p += XDR_QUADLEN(sizeof(verifier4));
        }
        *p++ = htonl(OPEN_DELEGATE_NONE);
        lp->lo_replay_buflen = resp->buflen = (char *)p - (char *)start;
        
out:
        nfs4_unlock_state();
        return status;
out_check_replay:
        if (status == NFS4ERR_REPLAY_ME) {
                nfsv4_printk(open, "nfs4_open: replaying retransmission!\n");
                lp = stp->current_lockowner;
                resp->buflen = lp->lo_replay_buflen;
                resp->buf = lp->lo_replay_buf;
                status = NFS4_OK;
        }
        goto out;
out_up_ptrerr:
        status = nfserrno(PTR_ERR(dfile));
        goto out_up;
out_up_dput:
        dput(dfile);
out_up:
        up(&dir_inode->i_sem);
        goto out;
out_dput_nfserr:
        dput(dfile);
out_nfserr:
        status = nfserrno(status);
        goto out;
}

int
nfs4_openattr(struct global_state *stp, OPENATTR4res *resp)
{
	return NFS4ERR_NOTSUPP;
}

int
nfs4_open_confirm(struct global_state *stp, OPEN_CONFIRM4args *argp,
		  OPEN_CONFIRM4res *resp)
{
	open_verifier_t *verifier = (open_verifier_t *) &argp->open_confirm;
	struct nfs4_lockowner *lp;
	u32 *start, *p;
	int status;
	
	nfs4_lock_state();
	
	status = nfs4_process_open_confirm(stp, argp->seqid, verifier);
	lp = stp->current_lockowner;
	if (!status) {
		resp->buf = lp->lo_replay_buf;
		start = p = (u32 *)resp->buf;
		memcpy(p, &stp->current_openfile->of_stateid, sizeof(stateid_t));
		p += XDR_QUADLEN(sizeof(stateid_t));
		resp->buflen = lp->lo_replay_buflen = (char *)p - (char *)start;
	}
	else if (status == NFS4ERR_REPLAY_ME) {
		resp->buflen = lp->lo_replay_buflen;
		resp->buf = lp->lo_replay_buf;
		status = NFS4_OK;
	}

	nfs4_unlock_state();
	return status;
}		  

int
nfs4_open_downgrade(struct global_state *stp, OPEN_DOWNGRADE4args *argp,
		    OPEN_DOWNGRADE4res *resp)
{
	stateid_t *stateid = (stateid_t *) &argp->stateid;
	struct nfs4_lockowner *lp;
	u32 *start, *p;
	int status;
	
	nfs4_lock_state();

	status = nfs4_process_open_downgrade(stp, argp->seqid, stateid,
					     argp->share_access, argp->share_deny);
	lp = stp->current_lockowner;
	if (!status) {
		resp->buf = lp->lo_replay_buf;
		start = p = (u32 *)resp->buf;
		memcpy(p, &stp->current_openfile->of_stateid, sizeof(stateid_t));
		p += XDR_QUADLEN(sizeof(stateid_t));
		resp->buflen = lp->lo_replay_buflen = (char *)p - (char *)start;
	}
	else if (status == NFS4ERR_REPLAY_ME) {
		resp->buflen = lp->lo_replay_buflen;
		resp->buf = lp->lo_replay_buf;
		status = NFS4_OK;
	}
	
	nfs4_unlock_state();
	return status;
}

int
nfs4_putfh(struct global_state *stp, PUTFH4args *argp, PUTFH4res *resp)
{
	int status;
	
	fh_put(&CURRENT_FH);
	status = fh_read(stp->rqstp, &CURRENT_FH, &argp->object);
	if (status)
		nfsv4_printk(level1, "bad filehandle in PUTFH!\n");
	else
		nfs4_setuser(&CURRENT_FH);
	return status;
}

int
nfs4_putpubfh(struct global_state *stp, PUTPUBFH4res *resp)
{
	int status;
	
	fh_put(&CURRENT_FH);
	if ((status = nfs4_setpub(&CURRENT_FH.fh_export)))
		goto out;
	if ((status = nfs4_setsec(stp->rqstp, CURRENT_FH.fh_export, &CURRENT_FH.fh_secattrs)))
		goto out;
	nfs4_setuser(&CURRENT_FH);	
	CURRENT_FH.fh_dentry = DGET(CURRENT_FH.fh_export->ex_dentry);
	CURRENT_FH.fh_initialized = 1;
	
out:
	return status;
}

int
nfs4_putrootfh(struct global_state *stp, PUTROOTFH4res *resp)
{
	int status;
	
	fh_put(&CURRENT_FH);
	if ((status = nfs4_setroot(&CURRENT_FH.fh_export)))
		goto out;
	if ((status = nfs4_setsec(stp->rqstp, CURRENT_FH.fh_export, &CURRENT_FH.fh_secattrs)))
		goto out;
	nfs4_setuser(&CURRENT_FH);
	CURRENT_FH.fh_dentry = DGET(CURRENT_FH.fh_export->ex_dentry);
	CURRENT_FH.fh_initialized = 1;
	
out:
	return status;
}

int
nfs4_read(struct global_state *stp, READ4args *argp, READ4res *resp)
{
	int status;
	struct nfs4_openfile *ofp;
	struct file *filp;
	struct file file;
	struct inode *inode;
	nfs4_ino_desc_t ino;
	mm_segment_t oldfs;
	u32 offset;
	u32 count = argp->count;
	char *buf;
	stateid_t *stateid = (stateid_t *) &argp->stateid;

	/* XXX : no readahead cache yet! */

	nfs4_lock_state();

	/* XXX: for now, LFS support is disabled.. */
	status = NFS4ERR_INVAL;
	if (argp->offset > ~(u32)0) {
		nfsv4_printk(read2, "nfs4_read: offset too large!\n");
		goto out;
	}
	offset = (u32)argp->offset;
	
	if (ZERO_STATEID(stateid))
		goto zero_stateid;
	if (ONE_STATEID(stateid))
		goto one_stateid;
	if ((status = nfs4_preprocess_stateid_op(stp, stateid)))
		goto out;
	ofp = stp->current_openfile;
	status = NFS4ERR_ACCES;
	if (!(ofp->of_share_access & OPEN4_SHARE_ACCESS_READ)) {
		nfsv4_printk(read2, "nfs4_read: file not open for read!\n");
		goto out;
	}
	filp = &ofp->of_vfs_file;

do_read:
	status = NFS4ERR_PERM;
	if (!filp->f_op || !filp->f_op->read) {
		nfsv4_printk(read2, "nfs4_read(): no vfs read op!\n");
		goto out_close;
	}
	/* XXX : is this ok, or do we need an llseek()? */
	filp->f_pos = offset;

	status = NFS4ERR_RESOURCE;
	/* XXX : ugh! get rid of this copy! */
	if (!(buf = NFS4_ALLOC(count))) {
		nfsv4_printk(read2, "nfs4_read(): couldn't alloc buffer!\n");
		goto out_close;
	}
	/* XXX : is the set_fs() really necessary?! */
	oldfs = get_fs(); set_fs(KERNEL_DS);
	status = filp->f_op->read(filp, buf, count, &filp->f_pos);
	set_fs(oldfs);
	if (status < 0) {
		nfsv4_printk(write2, "nfs4_write: f_op->write() failed!\n");
		goto out_read_err;
	}
	
	resp->READLEN = status;
	resp->READVAL = buf;
	resp->READEOF = ((offset + status) >= CURRENT_FH.fh_dentry->d_inode->i_size);
	nfsv4_printk(read, "nfs4_read: read %d bytes, eof=%d\n", status, resp->READEOF);
	status = NFS4_OK;	

out_close:
	if (filp == &file)
		nfs4_file_close(filp);
	
out:
	nfs4_unlock_state();
	return status;
out_read_err:
	status = nfserrno(status);
	NFS4_FREE(buf);
	goto out_close;
	
zero_stateid:
	/*
	 * For stateid 0, the client doesn't have to have the file open, but
	 * we still check for share reservation conflicts.  For stateid -1, we
	 * don't check share reservations.  For both of the "magic stateids",
	 * we do some permission checking (permission checking is done only
	 * for the magic stateids because it is normally done on OPEN).
	 */
	if ((status = FH_FILE(&stp->current_fh))) {
		nfsv4_printk(read2, "nfs4_read: no fh, or not a regular file!\n");
		goto out;
	}
	inode = stp->current_fh.fh_dentry->d_inode;
	ino.dev = inode->i_dev;
	ino.ino = inode->i_ino;
	ino.generation = inode->i_generation;	
	if ((status = nfs4_share_conflicts(&ino, OPEN4_SHARE_DENY_READ))) {
		nfsv4_printk(read2, "nfs4_read: conflicting share reservation!\n");
		goto out;
	}
	goto permission_checks;
	
one_stateid:
	/*
	 * For stateid -1, the client doesn't have to have the file open.  We
	 * do permission checking here, since permissions are normally checked
	 * on OPEN.
	 */
	if ((status = FH_FILE(&stp->current_fh))) {
		nfsv4_printk(read2, "nfs4_read: no fh, or not a regular file!\n");
		goto out;
	}

permission_checks:
	if ((status = FH_ACCESS(&stp->current_fh, MAY_READ))) {
		nfsv4_printk(read2, "nfs4_read: access denied!\n");
		goto out;
	}
	if ((status = nfs4_inode_permission(stp->current_fh.fh_dentry->d_inode, MAY_READ))) {
		nfsv4_printk(read2, "nfs4_read: access denied!\n");
		goto out;
	}
	filp = &file;
	if ((status = nfs4_file_open(filp, &stp->current_fh, OPEN4_SHARE_ACCESS_READ)))
		goto out;
	goto do_read;
}

int
nfs4_readdir(struct global_state *stp, READDIR4args *argp, READDIR4res *resp)
{
	int status;
	struct readdir_cd cd;
	struct dentry *dentry;
	struct inode *inode;
	struct file file;
	u32 *buf, *p;
	
	/* XXX : where is the call to nfsd_close in the v3 code ?! */

	/*
	 *  First, do a lot of checks, and set "dentry" to the directory's dentry.  If the
	 *  directory is an "internal" pseudofs node, indicate that by setting dentry=NULL.
	 */
	if ((status = FH_INITIALIZED(&CURRENT_FH))) {
		nfsv4_printk(readdir2, "nfs4_readdir: no fh!\n");
		goto out;
	}
	if ((status = FH_ACCESS(&CURRENT_FH, MAY_READ))) {
		nfsv4_printk(readdir2, "nfs4_readdir: access denied!\n");
		goto out;
	}

	if (argp->cookie > ~(u32) 0) {
		nfsv4_printk(readdir2, "nfs4_readdir(): cookie too large [%x %x]!\n",
			     (u32) (argp->cookie >> 32), (u32) (argp->cookie & 0xffffffff));
		status = NFS4ERR_BAD_COOKIE;
		goto out;
	}
	if (argp->maxcount < MINREADDIRBUF) {
		status = NFS4ERR_INVAL;   /* XXX : do right */
		goto out;
	}
	if (argp->maxcount > MAXREADDIRBUF)
		argp->maxcount = MAXREADDIRBUF;
	nfsv4_printk(readdir, "nfs4_readdir: cookie=%d, maxcount=%d\n", (int)argp->cookie,
		     (int)argp->maxcount);
	
	dentry = CURRENT_FH.fh_dentry;
	if (dentry) {
		if (!S_ISDIR(dentry->d_inode->i_mode)) {
			nfsv4_printk(readdir2, "nfs4_readdir: not a directory!\n");
			status = NFS4ERR_NOTDIR;
			goto out;
		}
		if ((status = nfs4_file_open(&file, &stp->current_fh,
					     OPEN4_SHARE_ACCESS_READ))) {
			nfsv4_printk(readdir2, "nfs4_readdir: nfsd_open failed!\n");
			goto out;
		}
		file.f_pos = argp->cookie;   /* XXX : should ->llseek() instead? */
	}

	/*
	 *  Second, prepare the buffer.
	 */
	nfsv4_printk(readdir, "nfs4_readdir: preparing buffer\n");
	if ((status = set_getattr_bitmap(&cd.bm4, &argp->attr_request))) {
		nfsv4_printk(readdir2, "nfs4_readdir: set_getattr_bitmap failed!\n");
		goto out_close;
	}
	if ((cd.attrval_len = get_attrval4_len(&cd.bm4)) < 0) {
		nfsv4_printk(readdir2, "nfs4_readdir: get_attrval_len failed!\n");
		status = NFS4ERR_INVAL;
		goto out_bitmap;
	}
	if (!(buf = NFS4_ALLOC(argp->maxcount << 1))) {
		/* XXX : too much space.. */
		nfsv4_printk(readdir2, "nfs4_readdir: couldn't alloc buffer!\n");
		status = NFS4ERR_RESOURCE;
		goto out_bitmap;
	}
	cd.buffer = buf;
	cd.buflen = argp->maxcount - 8;
	cd.old_cookie = NULL;	
	cd.exp = CURRENT_FH.fh_export;
	cd.dir_dentry = dentry;
	cd.eof = 1;

	/*
	 *  Third, call the function which will actually fill up the buffer
	 *  (vfs_readdir for nonpseudo, or pseudofs_readdir for pseudo).
	 */
	if (dentry) {
		nfsv4_printk(readdir, "nfs4_readdir: calling vfs_readdir() in dir \"%.*s\"\n",
			     (int)dentry->d_name.len, dentry->d_name.name);
		/* XXX: is the i_zombie lock needed?  vfs_readdir() obtains it, but
		   nfsd_readdir() does not... */
		inode = dentry->d_inode;
		down(&inode->i_sem);
		down(&inode->i_zombie);
		/*
		 * "This silly loop is necessary because ->readdir() is not guaranteed
		 * to fill up the entire buffer, but may choose to do less."  --NFSv3
		 */
		do {
			cd.flag = 0;
			status = file.f_op->readdir(&file, &cd, nfs4_filldir);
			if (status) {
				nfsv4_printk(readdir2, "nfs4_readdir: vfs_readdir failed!\n");
				goto out_vfserr;
			}
		} while (cd.flag && cd.eof);
		up(&inode->i_zombie);
		up(&inode->i_sem);
		if (cd.old_cookie)
			*cd.old_cookie = htonl((u32)file.f_pos);
	}
	else {
		nfsv4_printk(readdir, "nfs4_readdir: calling pseudofs_readdir() in exp %d\n",
			     cd.exp->ex_id);
		if ((status = pseudofs_readdir((u32)argp->cookie, &cd))) {
			nfsv4_printk(readdir, "nfs4_readdir: pseudofs_readdir failed!\n");
			goto out_buf;
		}
	}
	p = cd.buffer;
	*p++ = 0;
	*p++ = htonl(cd.eof);
	
	memset(resp->READDIR4res_u.resok4.cookieverf, 0, NFS4_VERIFIER_SIZE);
	resp->READDIR4res_u.resok4.reply_len = ((char *)p - (char *)buf);
	resp->READDIR4res_u.resok4.reply_buf = (char *)buf;

out_bitmap:
	NFS4_FREE(cd.bm4.bitmap4_val);
out_close:
	if (dentry)
		nfs4_file_close(&file);
out:
	return status;
out_vfserr:
	up(&inode->i_zombie);
	up(&inode->i_sem);
	status = nfserrno(status);
out_buf:
	NFS4_FREE(buf);
	goto out_bitmap;
}

int
nfs4_readlink(struct global_state *stp, READLINK4res *resp)
{
	int status;
	char *buf;
	mm_segment_t old_fs;
	struct dentry *dentry;
	struct inode *inode;

	if ((status = FH_SYMLINK(&CURRENT_FH)))
		goto out;
	
	nfsv4_printk(readlink, "nfs4_readlink(): starting with fh ");
	nfsv4_print_svcfh(readlink, &CURRENT_FH);
	
	dentry = CURRENT_FH.fh_dentry;
	inode = dentry->d_inode;
	if (!inode->i_op || !inode->i_op->readlink) {
		status = NFS4ERR_INVAL;
		goto out;
	}
	UPDATE_ATIME(inode);

	/* XXX : needs to be NFS4_MAXNAMLEN+1? */
	if (!(buf = NFS4_ALLOC(NFS4_MAXNAMLEN + 1))) {
		nfsv4_printk(level1, "nfs4_readlink(): couldn't alloc buffer!\n");
		status = NFS4ERR_RESOURCE;
		goto out;
	}
	resp->READLINK4res_u.resok4.link.utf8string_val = buf;
	
	/* XXX : why do you need tbe set_fs()?! */
	nfsv4_printk(readlink, "nfs4_readlink(): calling vfs readlink\n");
	old_fs = get_fs();
	set_fs(KERNEL_DS);
	status = inode->i_op->readlink(dentry, buf, NFS4_MAXNAMLEN);
	set_fs(old_fs);

	if (status < 0) {
		nfsv4_printk(readlink2, "nfs4_readlink(): vfs readlink returned error\n");
		NFS4_FREE(buf);
		goto out_nfserr;
	}
	resp->READLINK4res_u.resok4.link.utf8string_len = status;
	status = NFS4_OK;
	
out:
	nfsv4_printk(readlink, "nfs4_readlink(): returning status %d\n", status);
	return status;
out_nfserr:
	status = nfserrno(status);
	goto out;
}

int
nfs4_remove(struct global_state *stp, REMOVE4args *argp, REMOVE4res *resp)
{
	int namlen = argp->target.utf8string_len;
	char *name = argp->target.utf8string_val;
	char buf[NFS4_MAXNAMLEN + 1];
	int status;
	struct dentry *dentry, *dchild;
	struct inode *inode;

	status = NFS4ERR_GRACE;
	if (nfs4_in_grace_period()) {
		nfsv4_printk(lease2, "nfs4_remove: in grace period!\n");
		goto out;
	}
	if ((status = FH_EXTERNDIR(&CURRENT_FH)))
		goto out;
	if ((status = FH_ACCESS(&CURRENT_FH, MAY_REMOVE)))
		goto out;
	dentry = CURRENT_FH.fh_dentry;
	inode = dentry->d_inode;	
	if ((status = nfs4_inode_permission(inode, MAY_REMOVE)))
		goto out;
	
	if (namlen <= 0) {
		nfsv4_printk(remove2, "nfs4_remove(): namlen < 0!\n");
		status = NFS4ERR_NOENT;
		goto out;
	}
	if (namlen > NFS4_MAXNAMLEN)
		namlen = NFS4_MAXNAMLEN;
	if (isdotent(name, namlen)) {
		/* XXX : need name mangling */
		status = NFS4ERR_INVAL;
		goto out;
	}
	memcpy(buf, argp->target.utf8string_val, namlen);  /* XXX : ugh */
	buf[namlen] = 0;
	
	nfsv4_printk(remove, "nfs4_remove(): starting on %.*s, current fh = ",
		     namlen, name);
	nfsv4_print_svcfh(remove, &CURRENT_FH);

	down(&inode->i_sem);
	dchild = LOOKUP_ONE(buf, dentry);
	if (IS_ERR(dchild)) {
		nfsv4_printk(remove2, "nfs4_remove(): lookup_one returned error\n");
		status = nfserrno(PTR_ERR(dchild));
		goto out_unlock;
	}
	if (!dchild) {   /* XXX : can this happen? */
		status = NFS4ERR_RESOURCE;
		goto out_unlock;
	}
	if (!dchild->d_inode) {
		nfsv4_printk(remove, "nfs4_remove(): no such file or directory\n");
		DPUT(dchild);
		status = NFS4ERR_NOENT;
		goto out_unlock;
	}
	
	nfsv4_printk(remove, "nfs4_remove(): calling vfs remove\n");
	if (dchild->d_inode->i_mode & S_IFDIR)
		status = vfs_rmdir(inode, dchild);
	else
		status = vfs_unlink(inode, dchild);
	up(&inode->i_sem);
	DPUT(dchild);
	if (status) {
		nfsv4_printk(remove2, "nfs4_remove(): vfs remove returned error\n");
		status = nfserrno(status);
		goto out;
	}
	
#if 0
	/* XXX */
	if (EX_ISSYNC(CURRENT_FH.fh_export))
		nfsd_sync_dir(dentry);
#endif

out:
	nfsv4_printk(remove, "nfs4_remove() returning status %d\n", status);
	return status;
out_unlock:
	up(&inode->i_sem);
	goto out;
}

int
nfs4_rename(struct global_state *stp, RENAME4args *argp, RENAME4res *resp)
{
	int status;
	int flen = argp->oldname.utf8string_len;
	int tlen = argp->newname.utf8string_len;
	char *fname = argp->oldname.utf8string_val;
	char *tname = argp->newname.utf8string_val;
	char buf[NFS4_MAXNAMLEN+1];
	struct dentry *d_fdir, *d_tdir, *d_old, *d_new;
	struct inode *i_fdir, *i_tdir;
	change_info4 *src_cinfo = &resp->RENAME4res_u.resok4.source_cinfo;
	change_info4 *dst_cinfo = &resp->RENAME4res_u.resok4.target_cinfo;

	status = NFS4ERR_GRACE;
	if (nfs4_in_grace_period()) {
		nfsv4_printk(lease2, "nfs4_rename: in grace period!\n");
		goto out;
	}
	if ((status = FH_EXTERNDIR(&SAVE_FH)))
		goto out;
	if ((status = FH_ACCESS(&SAVE_FH, MAY_REMOVE)))
		goto out;
	d_fdir = SAVE_FH.fh_dentry;
	i_fdir = d_fdir->d_inode;
	if ((status = nfs4_inode_permission(i_fdir, MAY_REMOVE)))
		goto out;
	if ((status = FH_EXTERNDIR(&CURRENT_FH)))
		goto out;
	if ((status = FH_ACCESS(&SAVE_FH, MAY_CREATE)))
		goto out;
	d_tdir = CURRENT_FH.fh_dentry;
	i_tdir = d_tdir->d_inode;
	if ((status = nfs4_inode_permission(i_tdir, MAY_CREATE)))
		goto out;
	
	if ((flen <= 0) || (tlen <= 0)) {
		status = NFS4ERR_INVAL;
		goto out;
	}
	if (flen > NFS4_MAXNAMLEN)
		flen = NFS4_MAXNAMLEN;
	if (tlen > NFS4_MAXNAMLEN)
		tlen = NFS4_MAXNAMLEN;
	nfsv4_printk(rename, "nfs4_rename(): \"%.*s\" in dir ", flen, fname);
	nfsv4_print_svcfh(rename, &SAVE_FH);
	nfsv4_printk(rename, "  ---> \"%.*s\" in dir ", tlen, tname);
	nfsv4_print_svcfh(rename, &CURRENT_FH);

	if (isdotent(fname, flen) || isdotent(tname, tlen)) {
		status = NFS4ERR_INVAL;
		goto out;
	}
	if (i_fdir->i_dev != i_tdir->i_dev) {
		status = NFS4ERR_XDEV;
		goto out;
	}

	double_down(&i_fdir->i_sem, &i_tdir->i_sem);
	set_changeid4(src_cinfo->before, i_fdir);
	set_changeid4(dst_cinfo->before, i_tdir);

	/* XXX : yuck!  unneccessary copies! */
	memcpy(buf, fname, flen);
	buf[flen] = 0;
	d_old = LOOKUP_ONE(buf, d_fdir);
	if (IS_ERR(d_old)) {
		nfsv4_printk(rename2, "nfs4_rename(): lookup_one(fname) returned error!\n");
		status = nfserrno(PTR_ERR(d_old));
		goto out_unlock;
	}
	if (!d_old->d_inode) {
		nfsv4_printk(rename, "nfs4_rename(): file not found\n");
		status = NFS4ERR_NOENT;
		goto out_dput_old;
	}
	memcpy(buf, tname, tlen);
	buf[tlen] = 0;
	d_new = LOOKUP_ONE(buf, d_tdir);
	if (IS_ERR(d_new)) {
		nfsv4_printk(rename2, "nfs4_rename(): lookup_one(tname) returned error!\n");
		status = nfserrno(PTR_ERR(d_new));
		goto out_dput_old;
	}

	nfsv4_printk(rename, "nfs4_rename() calling vfs_rename\n");
	if ((status = vfs_rename(i_fdir, d_old, i_tdir, d_new))) {
		nfsv4_printk(rename2, "nfs4_rename(): vfs_rename returned error!\n");
		status = nfserrno(status);
		goto out_dput;
	}
	set_changeid4(src_cinfo->after, i_fdir);
	set_changeid4(dst_cinfo->after, i_tdir);
	src_cinfo->atomic = dst_cinfo->atomic = 1;

out_dput:
	DPUT(d_new);
out_dput_old:
	DPUT(d_old);
out_unlock:
	/* hand-unlock fdir, tdir */
	double_up(&i_fdir->i_sem, &i_tdir->i_sem);
out:
	nfsv4_printk(rename, "nfs4_rename() returning status %d\n", status);
	return status;
}

int
nfs4_renew(struct global_state *stp, RENEW4args *argp, RENEW4res *resp)
{
	stateid_t *stateid = (stateid_t *) &argp->stateid;
	int status;

	nfs4_lock_state();
	status = nfs4_process_renew(stp, stateid);
	nfs4_unlock_state();
	return status;
}

int
nfs4_restorefh(struct global_state *stp, RESTOREFH4res *resp)
{
	int status;

	status = FH_INITIALIZED(&SAVE_FH);
	if (!status) {
		fh_copy(&CURRENT_FH, &SAVE_FH);
		nfs4_setuser(&CURRENT_FH);
	}
	return status;
}

int
nfs4_savefh(struct global_state *stp, SAVEFH4res *resp)
{
	int status;

	status = FH_INITIALIZED(&CURRENT_FH);
	if (!status)
		fh_copy(&SAVE_FH, &CURRENT_FH);
	return status;
}


void
hard_code_secinfo4res(SECINFO4res *resp)
{
	secinfo4 *secinfo;
	char krb5[] = "1.2.840.113554.1.2.2";
	int len = 0;
	int kstrlen,i;
	char *cptr = NULL;

	nfsv4_printk(secinfo,"hard_code_secinfo4res\n");
	resp->status = 0;

 /* how many secinfo4 structs, each containing one rpcsec_gss_info */
	resp->SECRES.SECINFO4resok_len = 1;
	resp->SECRES.SECINFO4resok_val = NFS4_ALLOC(1*sizeof(secinfo4));

	secinfo = &resp->SECRES.SECINFO4resok_val[0];
	secinfo->flavor = RPC_AUTH_GSS;

	kstrlen = strlen(krb5);
	/* this is opaque data, therfore no pointers. */
	len = sizeof(u_int) + kstrlen + sizeof(qop4) + sizeof(rpc_gss_svc_t);
	nfsv4_printk(secinfo,"hard_code_secinfo4res len %d\n",len);
	secinfo->flavor_info.flavor_info_len = XDR_QUADLEN(len)<<2;

	secinfo->flavor_info.flavor_info_val = kmalloc(len,GFP_KERNEL);

	nfsv4_printk(secinfo,"hard_code_secinfo4res:\n");
	cptr = secinfo->flavor_info.flavor_info_val;
	*(int *)cptr = kstrlen;
	nfsv4_printk(secinfo,"     oid_len %d\n",(int)*cptr);
	cptr += 4;
	memcpy(cptr,krb5,kstrlen);
	nfsv4_printk(secinfo,"     ");
	for(i=0;i<kstrlen;i++) 
		nfsv4_printk(secinfo,"%c",*cptr++);
	nfsv4_printk(secinfo,"\n");

	*(int *)cptr = 42; /* fake qop */
	nfsv4_printk(secinfo,"     qop %d\n",(int)*cptr);
	cptr += 4;
	*(int *)cptr  = RPC_GSS_SVC_NONE;
	nfsv4_printk(secinfo,"     service %d\n",(int)*cptr);
	cptr += 4;
	
	nfsv4_printk(secinfo,
	"hard_code_secinfo4res: resp->SECRES.SECINFO4resok_val.flavor_info.flavor_info_len %d\n", (unsigned int)resp->SECRES.SECINFO4resok_val->flavor_info.flavor_info_len);
}

int
nfs4_secinfo(struct global_state *stp, SECINFO4args *argp, SECINFO4res *resp)
{
	nfsv4_printk(secinfo, "nfs4_secinfo: hard code default secinfo4\n");
	hard_code_secinfo4res(resp);
	return 0;
}

int
nfs4_setattr(struct global_state *stp, SETATTR4args *argp, SETATTR4res *resp)
{
	static u32 zero_bmval[] = { 0, 0 };
	struct iattr iattr;
	struct dentry *dentry;
	struct inode *inode;
	int imode;
	stateid_t *stateid = (stateid_t *) &argp->stateid;
	int status;

	nfsv4_printk(setattr, "nfs4_setattr: starting with bm ");
	nfsv4_print_bm(setattr, &argp->obj_attributes.attrmask);
	nfsv4_printk(setattr, "\n");

	if ((status = FH_EXTERN(&CURRENT_FH)))
		goto out;
	dentry = CURRENT_FH.fh_dentry;
	inode = dentry->d_inode;
	/* do_setattr() will check permissions */
	
	if ((status = read_attrs(&argp->obj_attributes, &iattr))) {
		nfsv4_printk(setattr2, "nfs4_setattr(): read_attrs returned error!\n");
		goto out_bad;
	}
	if (iattr.ia_valid & ATTR_SIZE) {
		nfsv4_printk(setattr, "nfs4_setattr: changing file size\n");
		status = NFS4ERR_BAD_STATEID;
		if (ZERO_STATEID(stateid) || ONE_STATEID(stateid)) {
			nfsv4_printk(setattr2, "nfs4_setattr: magic stateid!\n");
			goto out;
		}
		nfs4_lock_state();
		status = nfs4_preprocess_stateid_op(stp, stateid);
		nfs4_unlock_state();
		if (status)
			goto out;
		status = NFS4ERR_ACCES;
		if (!(stp->current_openfile->of_share_access & OPEN4_SHARE_ACCESS_WRITE)) {
			nfsv4_printk(setattr2, "nfs4_setattr: truncating, no write access!\n");
			goto out;
		}
		/* XXX: should we check for conflicting locks? */
	}

	nfsv4_printk(setattr, "nfs4_setattr(): doing setattr with current fh = ");
	nfsv4_print_svcfh(setattr, &CURRENT_FH);

	/* fix up mode bits */
	imode = dentry->d_inode->i_mode;
	if (iattr.ia_valid & ATTR_MODE) {
		iattr.ia_mode &= S_IALLUGO;
		iattr.ia_mode |= (imode & ~S_IALLUGO);
	}
	else {
		/* revoke setuid/setgid bit on chown/chgrp */
		/* XXX : logic here is a little different from v3.. is it OK? */
		if ((iattr.ia_valid & ATTR_UID) && (iattr.ia_uid != inode->i_uid)) {
			iattr.ia_valid |= ATTR_MODE;
			iattr.ia_mode = imode &= ~S_ISUID;
		}
		if ((iattr.ia_valid & ATTR_GID) && (iattr.ia_gid != inode->i_gid)) {
			iattr.ia_valid |= ATTR_MODE;
			iattr.ia_mode = imode &= ~S_ISGID;
		}
	}
	    
	if ((status = do_setattr(&CURRENT_FH, &iattr)))
		goto out_bad;
	if ((status = set_setattr_bitmap(&resp->attrsset, &argp->obj_attributes.attrmask)))
		goto out_bad;
	
out:
	nfsv4_printk(setattr, "nfs4_setattr() returning status %d\n", status);
	return status;
out_bad:
	resp->attrsset.bitmap4_len = 2;
	resp->attrsset.bitmap4_val = zero_bmval;
	goto out;
}

int
nfs4_setclientid(struct global_state *stp, SETCLIENTID4args *argp,
                 SETCLIENTID4res *resp)
{
        int status;

        /* Don't worry, this unnecessary wrapper routine will be eliminated when I finish
           the XDR.. */
        nfs4_lock_state();
        status = nfs4_process_setclientid(stp, argp, resp);
        nfs4_unlock_state();
        return status;
}

int
nfs4_setclientid_confirm(struct global_state *stp, SETCLIENTID_CONFIRM4args *argp,
                         SETCLIENTID_CONFIRM4res *resp)
{
        clientid_t *clientid = (clientid_t *)argp->setclientid_confirm;
        int status;
        
        /* Don't worry, this unnecessary wrapper routine will be eliminated when I finish
           the XDR.. */
        nfs4_lock_state();
        status = nfs4_process_setclientid_confirm(stp, clientid);
        nfs4_unlock_state();
        return status;
}

int
nfs4_verify(struct global_state *stp, VERIFY4args *argp, VERIFY4res *resp)
{
	int status;
	u32 dummy;
	bitmap4 *bm = &argp->obj_attributes.attrmask;
	attrlist4 *attrs = &argp->obj_attributes.attr_vals;
	u32 *buf;
	int i;
	int len;

	if ((status = FH_INITIALIZED(&CURRENT_FH)))
		goto out;
	/* XXX : permission check? */
	
	for (i = 0; i < bm->bitmap4_len; i++) {
		dummy = (i < GETATTR_BITMAP_LEN) ? GETATTR_BITMAP[i] : 0;
		if (bm->bitmap4_val[i] & ~dummy) {
			nfsv4_printk(getattr2, "nfs4_verify: bad bits set!\n");
			printk("!!! bitmap4_val[%d]=0x%x, GETATTR_BITMAP[%d]=0x%x\n",
			       i, bm->bitmap4_val[i], i, GETATTR_BITMAP[i]);
			status = NFS4ERR_INVAL;
			goto out;
		}
	}
	if ((len = get_attrval4_len(bm)) < 0) {
		nfsv4_printk(level1, "nfs4_verify: get_attrval4_len failed!\n");
		status = NFS4ERR_SERVERFAULT;
		goto out;
	}
	if (!(buf = NFS4_ALLOC(len))) {
		nfsv4_printk(level1, "nfs4_verify: failed to alloc buffer!\n");
		status = NFS4ERR_RESOURCE;
		goto out;
	}
	status = write_attrs(CURRENT_FH.fh_export, CURRENT_FH.fh_dentry, bm, buf);
	if (status < 0) {
		nfsv4_printk(getattr2, "nfs4_verify: write_attrs returned error!\n");
		status = -status;
		goto out_free;
	}
	if (status != attrs->attrlist4_len) {
		/* attrs are different */
		status = NFS4ERR_NOT_SAME;
		goto out_free;
	}
	if (memcmp(buf, attrs->attrlist4_val, status)) {
		/* attrs are different */
		status = NFS4ERR_NOT_SAME;
		goto out_free;
	}
	status = NFS4_OK;

out_free:
	NFS4_FREE(buf);
out:
	return status;
}

int
nfs4_write(struct global_state *stp, WRITE4args *argp, WRITE4res *resp)
{
	int status;
	u32 offset;
	u32 count = argp->WRITELEN;
	char *buf = argp->WRITEVAL;
	struct nfs4_openfile *ofp;
	struct file *filp;
	struct file file;
	struct inode *inode;
	nfs4_ino_desc_t ino;
	mm_segment_t oldfs;
	stateid_t *stateid = (stateid_t *) &argp->stateid;

	nfs4_lock_state();

	/* XXX: for now, LFS support is disabled.. */
	status = NFS4ERR_INVAL;
	if (argp->offset > ~(u32)0) {
		nfsv4_printk(write2, "nfs4_write: offset too large!\n");
		goto out;
	}
	offset = (u32)argp->offset;

	if (ZERO_STATEID(stateid) || ONE_STATEID(stateid))
		goto special_stateid;
	if ((status = nfs4_preprocess_stateid_op(stp, stateid)))
		goto out;
	ofp = stp->current_openfile;
	status = NFS4ERR_ACCES;
	if (!(ofp->of_share_access & OPEN4_SHARE_ACCESS_WRITE)) {
		nfsv4_printk(write2, "nfs4_write: file not open for write!\n");
		goto out;
	}
	filp = &ofp->of_vfs_file;

do_write:
	status = NFS4ERR_PERM;
	if (!filp->f_op || !filp->f_op->write) {
		nfsv4_printk(write2, "nfs4_write: no vfs write op!\n");
		goto out_close;
	}
	/* XXX : is this ok, or do we need an llseek()? */
	filp->f_pos = offset;

	/* XXX : is the set_fs() really necessary?! */
	oldfs = get_fs(); set_fs(KERNEL_DS);
	if (argp->stable == UNSTABLE4) {
		status = filp->f_op->write(filp, buf, count, &filp->f_pos);
	}
	else {
		/* XXX: The v3 code contains some "gathered write" logic for optimizing
		   sync writes.. should we bring this over? */
		filp->f_flags |= O_SYNC;
		status = filp->f_op->write(filp, buf, count, &filp->f_pos);
		filp->f_flags &= ~O_SYNC;
	}
	set_fs(oldfs);
	if (status < 0) {
		nfsv4_printk(write2, "nfs4_write: f_op->write() failed!\n");
		goto out_write_err;
	}
	
	resp->WRITERES.count = status;
	/* XXX: is this right? */
	resp->WRITERES.committed = FILE_SYNC4;
	if (argp->stable == UNSTABLE4)
		resp->WRITERES.committed = UNSTABLE4;
	nfs4_set_verifier(resp->WRITERES.writeverf);
	status = NFS4_OK;

out_close:
	if (filp == &file)
		nfs4_file_close(filp);	
out:
	nfs4_unlock_state();
	return status;
out_write_err:
	status = nfserrno(status);
	goto out_close;
	
special_stateid:
	/*
	 * For stateids 0 and -1, the client doesn't have to have the file open, but
	 * we still check for share reservation conflicts.  We do permission checking
	 * only for the "magic stateids" because permissions are normally checked on
	 * OPEN.
	 */
	if ((status = FH_FILE(&stp->current_fh))) {
		nfsv4_printk(write2, "nfs4_write: no fh, or not a regular file!\n");
		goto out;
	}
	inode = stp->current_fh.fh_dentry->d_inode;
	ino.dev = inode->i_dev;
	ino.ino = inode->i_ino;
	ino.generation = inode->i_generation;	
	if ((status = nfs4_share_conflicts(&ino, OPEN4_SHARE_DENY_WRITE))) {
		nfsv4_printk(write2, "nfs4_write: conflicting share reservation!\n");
		goto out;
	}
	if ((status = FH_ACCESS(&stp->current_fh, MAY_WRITE))) {
		nfsv4_printk(write2, "nfs4_write: access denied!\n");
		goto out;
	}
	if ((status = nfs4_inode_permission(stp->current_fh.fh_dentry->d_inode, MAY_WRITE))) {
		nfsv4_printk(write2, "nfs4_write: permission denied!\n");
		goto out;
	}
	filp = &file;
	if ((status = nfs4_file_open(filp, &stp->current_fh, OPEN4_SHARE_ACCESS_WRITE)))
		goto out;
	goto do_write;
}

#ifdef NFS4ACL

int
nfs4_get_acl(struct nfs4_export *exp, struct dentry *dentry, u32 *buf)
{
        posix_acl_t *acl;
        u32 *p = buf;
        struct inode *inode = dentry->d_inode;
        int status;
        fattr4_acl * f_ace;
        nfsace4         *ace;
        int i;


        if (IS_POSIX_ACL(inode) && inode->i_op && inode->i_op->get_posix_acl){
                acl = inode->i_op->get_posix_acl(inode, ACL_TYPE_ACCESS);
                if (acl){
                        nfsv4_printk(acl, "acl:          %p\n", acl);
                        nfsv4_printk(acl, "acl->a_count: %d\n", acl->a_count);
			printk("((sizeof(nfsace4) + 12) * 15) = %d\n", ((sizeof(nfsace4) + 12) * 15));
			printk("sizeof(1024) = %d\n", sizeof(1024));

                        f_ace = posix_acl_to_nfsv4_acl(acl);
                        if (!f_ace) {
                                nfsv4_printk(acl2, "nfs4_get_acl: No ace returned\n");
                                //goto exit;  // XXX need correct exit
                        }
                        *p++ = htonl(f_ace->fattr4_acl_len);   /* ACE count */
			printk("nfs4_get_acl: f_ace->fattr4_acl_len=%d\n", f_ace->fattr4_acl_len);
                        for (i = 0; i < f_ace->fattr4_acl_len; i++) {
                                printk("--------------------\nnfs4_get_acl: i=%d\n", i);
                                ace = &f_ace->fattr4_acl_val[i];
                                if (!ace)
                                        nfsv4_printk(acl2, "nfs4_get_acl: No ace in fattr4\n");
                                *p++ = htonl(ace->type); /* type */
                                printk("nfs4_get_acl: type=%d\n", ace->type);
                                *p++ = htonl(ace->flag); /* flags */
                                printk("nfs4_get_acl: flag=%d\n", ace->flag);
                                *p++ = htonl(ace->access_mask);  /* mask */
                                printk("nfs4_get_acl: mask=%d\n", ace->access_mask);
                                *p++ = htonl(ace->who.utf8string_len);
                                printk("nfs4_get_acl: string_len=%d\n", ace->who.utf8string_len);
                                memcpy(p, ace->who.utf8string_val, ace->who.utf8string_len);
				/* pad it */
				//memset(p + ace->who.utf8string_len, 0, XDR_QUADLEN(ace->who.utf8string_len) - ace->who.utf8string_len);
                                printk("nfs4_get_acl: string_val=%s\n", ace->who.utf8string_val);
                                printk("nfs4_get_acl: pointer string_val=%s\n", (char *)p);
                                p += XDR_QUADLEN(ace->who.utf8string_len);
                                printk("nfs4_get_acl: XDR_QUADLEN=%d\n",  XDR_QUADLEN(ace->who.utf8string_len));
                        } // for
                } // if (acl)
		else {
			*p++ = htonl(0);   /* ACE count */
	        	printk("nfs4_get_acl: f_ace->fattr4_acl_len=%d\n", 0);
		}
        } // if (IS_POSIX_ACL...
	else {
		*p++ = htonl(0);   /* ACE count */
        	printk("nfs4_get_acl: Not a Posix acl.  Make sure acl's are supported by this local fs.\n"); 
	}

        status = (u32 *)p - (u32 *)buf;
        nfsv4_printk(acl, "nfs4_get_acl: status=%d\n", status);
        return status;
}
#endif

