/*
 * linux/fs/posix_acl.c
 *
 *  Copyright (C) 2000 by Andreas Gruenbacher <a.gruenbacher@computer.org>
 *
 *  Fixes from William Schumacher incorporated on 15 March 2001.
 *     (Reported by Charles Bertsch, <CBertsch@microtest.com>).
 */

/*
 *  This file contains generic functions for manipulating
 *  POSIX 1003.1e draft standard 17 ACLs.
 */

#include <linux/kernel.h>
#include <linux/slab.h>
#include <asm/atomic.h>
#include <linux/posix_acl.h>
#include <linux/module.h>

EXPORT_SYMBOL(posix_acl_alloc);
EXPORT_SYMBOL(posix_acl_dup);
EXPORT_SYMBOL(posix_acl_release);
EXPORT_SYMBOL(posix_acl_valid);
EXPORT_SYMBOL(posix_acl_equiv_mode);
EXPORT_SYMBOL(posix_acl_from_inode);
EXPORT_SYMBOL(posix_acl_check_perm);
EXPORT_SYMBOL(posix_acl_create_masq);
EXPORT_SYMBOL(posix_acl_chmod_masq);
EXPORT_SYMBOL(posix_acl_masq_nfs_v2_mode);
EXPORT_SYMBOL(posix_acl_permission);
EXPORT_SYMBOL(posix_acl_chmod);
EXPORT_SYMBOL(posix_acl_init_cred);
#if 0
EXPORT_SYMBOL(posix_acl_cmp);
EXPORT_SYMBOL(posix_acl_hash);
#endif

/*
 * Allocate a new ACL with the specified number of entries.
 */
posix_acl_t *
posix_acl_alloc(
	int count)
{
	const size_t size = sizeof(posix_acl_t) +
	                    count * sizeof(posix_acl_entry_t);
	posix_acl_t *acl = kmalloc(size, GFP_KERNEL);
	if (acl) {
		atomic_set(&acl->a_refcount, 1);
		acl->a_count = count;
	}
	return acl;
}

/*
 * Duplicate an ACL handle.
 */
posix_acl_t *
posix_acl_dup(
	posix_acl_t *acl)
{
	if (acl)
		atomic_inc(&acl->a_refcount);
	return acl;
}

/*
 * Clone an ACL.
 */
posix_acl_t *
posix_acl_clone(
	const posix_acl_t *acl)
{
	posix_acl_t *clone = NULL;

	if (acl) {
		int size = sizeof(posix_acl_t) + acl->a_count *
		           sizeof(posix_acl_entry_t);
		clone = kmalloc(size, GFP_KERNEL);
		if (clone) {
			memcpy(clone, acl, size);
			atomic_set(&clone->a_refcount, 1);
		}
	}
	return clone;
}

/*
 * Free an ACL handle.
 */
void
posix_acl_release(
	posix_acl_t *acl)
{
	if (acl && atomic_dec_and_test(&acl->a_refcount))
		kfree(acl);
}

/*
 * Check if an acl is valid. Returns 0 if it is, or -E... otherwise.
 */
int
posix_acl_valid(
	const posix_acl_t *acl)
{
	const posix_acl_entry_t *pa, *pe;
	int state = ACL_USER_OBJ;
	unsigned int id = 0;  /* keep gcc happy */
	int needs_mask = 0;

	FOREACH_ACL_ENTRY(pa, acl, pe) {
		if (pa->e_perm & ~(ACL_READ|ACL_WRITE|ACL_EXECUTE))
			return -EINVAL;
		switch (pa->e_tag) {
			case ACL_USER_OBJ:
				if (state == ACL_USER_OBJ) {
					id = 0;
					state = ACL_USER;
					break;
				}
				return -EINVAL;

			case ACL_USER:
				if (state != ACL_USER)
					return -EINVAL;
				if (pa->e_id == ACL_UNDEFINED_ID ||
				    pa->e_id < id)
					return -EINVAL;
				id = pa->e_id + 1;
				needs_mask = 1;
				break;

			case ACL_GROUP_OBJ:
				if (state == ACL_USER) {
					id = 0;
					state = ACL_GROUP;
					break;
				}
				return -EINVAL;

			case ACL_GROUP:
				if (state != ACL_GROUP)
					return -EINVAL;
				if (pa->e_id == ACL_UNDEFINED_ID ||
				    pa->e_id < id)
					return -EINVAL;
				id = pa->e_id + 1;
				needs_mask = 1;
				break;

			case ACL_MASK:
				if (state != ACL_GROUP)
					return -EINVAL;
				state = ACL_OTHER;
				break;

			case ACL_OTHER:
				if (state == ACL_OTHER ||
				    (state == ACL_GROUP && !needs_mask)) {
					state = 0;
					break;
				}
				return -EINVAL;

			default:
				return -EINVAL;
		}
	}
	if (state == 0)
		return 0;
	return -EINVAL;
}

/*
 * Returns 0 if the acl can be exactly represented in the traditional
 * file mode permission bits, or else 1. Returns -E... on error.
 */
int
posix_acl_equiv_mode(
	const posix_acl_t *acl,
	mode_t *modep)
{
	const posix_acl_entry_t *pa, *pe;
	mode_t mode = 0;
	int not_equiv = 0;

	FOREACH_ACL_ENTRY(pa, acl, pe) {
		switch (pa->e_tag) {
			case ACL_USER_OBJ:
				mode |= (pa->e_perm & S_IRWXO) << 6;
				break;
			case ACL_GROUP_OBJ:
				mode |= (pa->e_perm & S_IRWXO) << 3;
				break;
			case ACL_OTHER:
				mode |= pa->e_perm & S_IRWXO;
				break;
			case ACL_MASK:
				mode = (mode & ~S_IRWXG) |
				       ((pa->e_perm & S_IRWXO) << 3);
				not_equiv = 1;
				break;
			case ACL_USER:
			case ACL_GROUP:
				not_equiv = 1;
				break;
			default:
				return -EINVAL;
		}
	}
        if (modep)
                *modep = (*modep & ~S_IRWXUGO) | (mode & S_IRWXUGO);
        return not_equiv;
}

/*
 * Create an ACL representing the file mode permission bits of an inode.
 */
posix_acl_t *
posix_acl_from_inode(
	struct inode *inode)
{
	posix_acl_t *acl = posix_acl_alloc(3);
	if (!acl)
		return ERR_PTR(-ENOMEM);

	acl->a_entries[0].e_tag  = ACL_USER_OBJ;
	acl->a_entries[0].e_id   = inode->i_uid;
	acl->a_entries[0].e_perm = (inode->i_mode & S_IRWXU) >> 6;

	acl->a_entries[1].e_tag  = ACL_GROUP_OBJ;
	acl->a_entries[1].e_id   = inode->i_gid;
	acl->a_entries[1].e_perm = (inode->i_mode & S_IRWXG) >> 3;

	acl->a_entries[2].e_tag  = ACL_OTHER;
	acl->a_entries[2].e_id   = ACL_UNDEFINED_ID;
	acl->a_entries[2].e_perm = (inode->i_mode & S_IRWXO);
	return acl;
}

/*
 * Return 0 if current is granted want access to the inode
 * by the acl. Returns -E... otherwise.
 */
int
posix_acl_check_perm(
	struct inode *inode,
	const posix_acl_t *acl,
	int want)
{
	const posix_acl_entry_t *pa, *pe, *mask_obj;
	int found = 0;

	FOREACH_ACL_ENTRY(pa, acl, pe) {
                switch(pa->e_tag) {
                        case ACL_USER_OBJ:
				/* (May have been checked already) */
                                if (inode->i_uid == current->fsuid)
                                        goto check_perm;
                                break;
                        case ACL_USER:
                                if (pa->e_id == current->fsuid)
                                        goto mask;
				break;
                        case ACL_GROUP_OBJ:
                                if (in_group_p(inode->i_gid)) {
					found = 1;
					if ((pa->e_perm & want) == want)
						goto mask;
                                }
				break;
                        case ACL_GROUP:
                                if (in_group_p(pa->e_id)) {
					found = 1;
					if ((pa->e_perm & want) == want)
						goto mask;
                                }
                                break;
                        case ACL_MASK:
                                break;
                        case ACL_OTHER:
				if (found)
					return -EACCES;
				else
					goto check_perm;
			default:
				return -EIO;
                }
        }
	return -EIO;

mask:
	for (mask_obj = pa+1; mask_obj != pe; mask_obj++) {
		if (mask_obj->e_tag == ACL_MASK) {
			if ((pa->e_perm & mask_obj->e_perm & want) == want)
				return 0;
			return -EACCES;
		}
	}

check_perm:
	if ((pa->e_perm & want) == want)
		return 0;
	return -EACCES;
}

/*
 * Modify acl when creating a new inode. The caller must ensure the acl is
 * only referenced once.
 *
 * modep initially must contain the mode parameter to the open() / creat()
 * system calls. All permissions that are not granted by the acl are removed.
 * The permissions in the acl are changed to reflect the modep parameter.
 */
int
posix_acl_create_masq(
	posix_acl_t *acl,
	mode_t *modep)
{
	posix_acl_entry_t *pa, *pe;
	posix_acl_entry_t *group_obj = NULL, *mask_obj = NULL;
	mode_t mode = *modep;
	int not_equiv = 0;

	/* assert(atomic_read(acl->a_refcount) == 1); */

	FOREACH_ACL_ENTRY(pa, acl, pe) {
                switch(pa->e_tag) {
                        case ACL_USER_OBJ:
				pa->e_perm &= (mode >> 6) | ~S_IRWXO;
				mode &= (pa->e_perm << 6) | ~S_IRWXU;
				break;

			case ACL_USER:
			case ACL_GROUP:
				not_equiv = 1;
				break;

                        case ACL_GROUP_OBJ:
				group_obj = pa;
                                break;

                        case ACL_OTHER:
				pa->e_perm &= mode | ~S_IRWXO;
				mode &= pa->e_perm | ~S_IRWXO;
                                break;

                        case ACL_MASK:
				mask_obj = pa;
				not_equiv = 1;
                                break;

			default:
				return -EIO;
                }
        }

	if (mask_obj) {
		mask_obj->e_perm &= (mode >> 3) | ~S_IRWXO;
		mode &= (mask_obj->e_perm << 3) | ~S_IRWXG;
	} else {
		if (!group_obj)
			return -EIO;
		group_obj->e_perm &= (mode >> 3) | ~S_IRWXO;
		mode &= (group_obj->e_perm << 3) | ~S_IRWXG;
	}

	*modep = mode;
        return not_equiv;
}

/*
 * Modify acl for the chmod syscall.
 */
int
posix_acl_chmod_masq(
	posix_acl_t *acl,
	mode_t mode)
{
	posix_acl_entry_t *group_obj = NULL, *mask_obj = NULL;
	posix_acl_entry_t *pa, *pe;

	/* assert(atomic_read(acl->a_refcount) == 1); */

	FOREACH_ACL_ENTRY(pa, acl, pe) {
		switch(pa->e_tag) {
			case ACL_USER_OBJ:
				pa->e_perm = (mode & S_IRWXU) >> 6;
				break;

			case ACL_USER:
			case ACL_GROUP:
				break;

			case ACL_GROUP_OBJ:
				group_obj = pa;
				break;

			case ACL_MASK:
				mask_obj = pa;
				break;

			case ACL_OTHER:
				pa->e_perm = (mode & S_IRWXO);
				break;

			default:
				return -EIO;
		}
	}

	if (mask_obj) {
		mask_obj->e_perm = (mode & S_IRWXG) >> 3;
	} else {
		if (!group_obj)
			return -EIO;
		group_obj->e_perm = (mode & S_IRWXG) >> 3;
	}

	return 0;
}

/*
 * Adjust the mode parameter so that NFSv2 grants nobody permissions
 * that would not be granted by the ACL. The group file mode permission
 * bits are set to the intersection of the ACL_GROUP_OBJ and ACL_MASK
 * entries. The others file mode permission bits are set to the intersection
 * of all ACL entries except the ACL_USER_OBJ entry. This is necessary
 * because NFSv2 computes access permissions on the client side, and serves
 * cached data whenever it assumes access would be granted.
 *
 * No permissions granted by an ACL are effective if an inode is accessed
 * via NFSv2. Unfortunately this restriction is necessary to ensure file
 * security.
 */
int posix_acl_masq_nfs_v2_mode(
	posix_acl_t *acl, mode_t *modep)
{
	posix_acl_entry_t *pa, *pe; int min_perm = S_IRWXO;

	FOREACH_ACL_ENTRY(pa, acl, pe) {
                switch(pa->e_tag) {
			case ACL_USER_OBJ:
			case ACL_OTHER:
				break;

			case ACL_GROUP_OBJ:
				/* owning group: (ACL_GROUP_OBJ & ACL_MASK) */
				*modep &= ~S_IRWXG | (pa->e_perm << 3);
				break;

			case ACL_USER:
			case ACL_GROUP:
			case ACL_MASK:
				min_perm &= pa->e_perm;
				break;

			default:
				return -EIO;
		}
	}
	/* others: least common permissions, so that permissions that may be
	   denied are never granted */
	*modep &= ~S_IRWXO | min_perm;

	return 0;
}

/*
 * Check for read/write/execute permissions on a file
 * (used as permission inode operation.)
 */
int
posix_acl_permission(
	struct inode *inode,
	int mask)
{
	int mode = inode->i_mode;

	/* Nobody gets write access to a read-only fs */
	if ((mask & MAY_WRITE) && IS_RDONLY(inode) &&
	    (S_ISREG(mode) || S_ISDIR(mode) || S_ISLNK(mode)))
		return -EROFS;
	/* Nobody gets write access to an immutable file */
	if ((mask & MAY_WRITE) && IS_IMMUTABLE(inode))
	    return -EACCES;
	if (current->fsuid == inode->i_uid) {
		mode >>= 6;
#ifdef CONFIG_FS_POSIX_ACL
	} else if (IS_POSIX_ACL(inode) && inode->i_acl) {
		posix_acl_t *acl;

		/* ACL can't contain additional permissions if
		   the ACL_MASK entry is 0 */
		if (!(inode->i_mode & S_IRWXG))
			goto check_mode;
		if (!inode->i_op || !inode->i_op->get_posix_acl)
			goto check_mode;
		acl = inode->i_op->get_posix_acl(inode, ACL_TYPE_ACCESS);
		if (IS_ERR(acl))
			return PTR_ERR(acl);
		else if (acl) {
			int error = posix_acl_check_perm(inode, acl, mask);
			posix_acl_release(acl);
			if (error == -EACCES)
				goto check_capabilities;
			return error;
		} else
			goto check_groups;
#endif
	} else {
check_groups:
		if (in_group_p(inode->i_gid))
			mode >>= 3;
	}
check_mode:
	if ((mode & mask & S_IRWXO) == mask)
		return 0;

check_capabilities:
	/* Allowed to override Discretionary Access Control? */
	if (capable(CAP_DAC_OVERRIDE))
		return 0;
	/* Read and search granted if capable(CAP_DAC_READ_SEARCH) */
	if (capable(CAP_DAC_READ_SEARCH) && ((mask == MAY_READ) ||
	    (S_ISDIR(mode) && !(mask & ~(MAY_READ | MAY_EXEC)))))
		return 0;
	return -EACCES;
}

/*
 * Does chmod for an inode that may have an Access Control List.
 * Returns 0 on success, or a negative error number.
 *
 * We change the ACL rather than storing some ACL entries in the file
 * mode permission bits (which would be more efficient), because that
 * would break once additional permissions (like  ACL_APPEND, ACL_DELETE
 * for directories) are added. There are no more bits available in the
 * file mode.
 */
int
posix_acl_chmod(
        struct inode *inode)
{
	posix_acl_t *acl, *clone;
        int error;

	if (!inode->i_op ||
            !inode->i_op->get_posix_acl || !inode->i_op->set_posix_acl)
		return 0;
	acl = inode->i_op->get_posix_acl(inode, ACL_TYPE_ACCESS);
	if (IS_ERR(acl) || !acl)
		return PTR_ERR(acl);
	clone = posix_acl_clone(acl);
	posix_acl_release(acl);
	if (!clone)
		return -ENOMEM;
	error = posix_acl_chmod_masq(clone, inode->i_mode);
	if (!error)
		error = inode->i_op->set_posix_acl(
			inode, ACL_TYPE_ACCESS, clone);
	posix_acl_release(clone);
	return error;
}

/*
 * Initialize the i_uid, i_gid and i_mode fields of a new inode,
 * and inherit the default ACL if present.
 *
 * On ACL-enabled filesystems, inode->i_op must be initialized
 * before calling this function (the extended attribute
 * operations are needed for ACL copying).
 */
int
posix_acl_init_cred(
	struct inode *inode,
	struct inode *dir,
	mode_t mode,
	int policy)
{
	int error = 0;

	posix_acl_t *acl = NULL;

	if (!S_ISLNK(mode)) {
		if (IS_POSIX_ACL(dir) &&
		    dir->i_op && dir->i_op->get_posix_acl) {
			acl = dir->i_op->get_posix_acl(dir, ACL_TYPE_DEFAULT);
			if (IS_ERR(acl))
				return PTR_ERR(acl);
		}
		if (!acl)
			mode &= ~current->fs->umask;
	}
	init_cred(inode, dir, mode, policy | IPOL_IGNORE_UMASK);
	if (IS_POSIX_ACL(inode) && acl &&
	    inode->i_op && inode->i_op->set_posix_acl) {
               posix_acl_t *clone;

		if (S_ISDIR(inode->i_mode)) {
			error = inode->i_op->set_posix_acl(
				inode, ACL_TYPE_DEFAULT, acl);
			if (error)
				goto cleanup;
		}
		clone = posix_acl_clone(acl);
		if (!clone) {
			error = -ENOMEM;
			goto cleanup;
		}
               error = posix_acl_create_masq(clone, &inode->i_mode);
               if (error > 0) {
                       /* This is an extended ACL */
                       error = inode->i_op->set_posix_acl(
                               inode, ACL_TYPE_ACCESS, clone);
               }
	       posix_acl_release(clone);
       }
cleanup:
       posix_acl_release(acl);
       return error;
}

#if 0
/*
 * Compare two ACLs. Returns 0 if they are equal.
 */
int
posix_acl_cmp(
	const posix_acl_t *acl1,
	const posix_acl_t *acl2)
{
	/* (Also works for acl1->a_count != acl2->a_count) */
	return memcmp(acl1, acl2, sizeof(posix_acl_t) +
                      acl1->a_count * sizeof(posix_acl_entry_t));
}
#endif

#if 0
/*
 * Hash an ACL.
 */
unsigned int
posix_acl_hash(
	posix_acl_t *acl)
{
	unsigned int hash = 0;
	posix_acl_entry_t *pa, *pe;

	FOREACH_ACL_ENTRY(pa, acl, pe) {
		/* The shift values are rough guesses of the significant bits
		   in each field */
		hash = ((hash << 13) | (hash >> 19)) ^
		       (pa->e_tag << (12+3)) ^ (pa->e_id << 3) ^ (pa->e_perm);
	}
	return hash;
}
#endif
