/*
 *  linux/fs/nfs4/nfs4fs.h
 *
 *  Copyright (c) 2001 The Regents of the University of Michigan.
 *  All rights reserved.
 *
 *  Kendrick Smith <kmsmith@umich.edu>
 *  Andy Adamson <andros@umich.edu>
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifndef _LINUX_NFS4_NFS4FS_H
#define _LINUX_NFS4_NFS4FS_H

#include <linux/nfs4_fs_sb.h>
#include <linux/nfs4_fs_i.h>

#define NFS4_MINOR_VERSION		0
#define NFS4_DEF_FILE_IO_BUFFER_SIZE	4096
#define NFS4_MAX_FILE_IO_BUFFER_SIZE	32768
#define NFS4_DEF_MAXFILESIZE		0xffffffff
#define NFS4_SUPER_MAGIC		0xF00BA4

#define NFS4FS_SILLY_RENAME             1
#define NFS4FS_STRICT_LOCKING           1
#define NFS4FS_MIN_LEASE                (1 * HZ)
#define NFS4FS_DEFAULT_LEASE            (30 * HZ)
#define NFS4FS_MAX_LEASE                (120 * HZ)
#define NFS4FS_ASYNC_DELAY              (HZ >> 2)    /* half a second */

#define NFS4_CLNT(sb)			((sb)->u.nfs4_sb.client)
#define NFS4_RPAGES(sb)			((sb)->u.nfs4_sb.rpages)
#define NFS4_WPAGES(sb)			((sb)->u.nfs4_sb.wpages)
#define NFS4_HAVE_CLIENTID(sb)		((sb)->u.nfs4_sb.have_clientid)
#define NFS4_CLIENTID(sb)		((sb)->u.nfs4_sb.clientid)
#define NFS4_PLEN(sb)			((sb)->u.nfs4_sb.plen)
#define NFS4_PVAL(sb)			((sb)->u.nfs4_sb.pval)
#define NFS4_LOCK_MINTIMEOUT(sb)	((sb)->u.nfs4_sb.lock_mintimeout)
#define NFS4_LOCK_MAXTIMEOUT(sb)	((sb)->u.nfs4_sb.lock_maxtimeout)
#define NFS4_STRICT_LOCKING(sb)		((sb)->u.nfs4_sb.strict_locking)
#define NFS4_SILLY_RENAME(sb)		((sb)->u.nfs4_sb.silly_rename)
#define NFS4_LEASE_TIME(sb)             ((sb)->u.nfs4_sb.lease_time)
#define NFS4_PIDOF_RECOVERD(sb)         ((sb)->u.nfs4_sb.pidof_recoverd)
#define NFS4_GRACE_WAIT_QUEUE(sb)       ((sb)->u.nfs4_sb.grace_wait_queue)
#define NFS4_FILE_LIST(sb)		((sb)->u.nfs4_sb.file_list)
#define NFS4_RENEWD(sb)			((sb)->u.nfs4_sb.renewd)
#define NFS4_LAST_RENEWAL(sb)		((sb)->u.nfs4_sb.last_renewal)
#define NFS4_FLUSHD(sb)			((sb)->u.nfs4_sb.flushd)
#define NFS4_FLUSHD_TIMEOUT(sb)		((sb)->u.nfs4_sb.flushd_timeout)
#define NFS4_FLUSHD_INODES(sb)		((sb)->u.nfs4_sb.flushd_inodes)
#define NFS4_FLUSHD_WAIT(sb)		((sb)->u.nfs4_sb.flushd_wait)
#define NFS4_IP_ADDR(sb)		((sb)->u.nfs4_sb.ip_addr)
#define NFS4_CALLBACK_PORT(sb)		((sb)->u.nfs4_sb.callback_port)

#define NFS4_DFSID_MAJOR(dentry)	((dentry)->d_inode->u.nfs4_i.fsid_major)
#define NFS4_DFSID_MINOR(dentry)	((dentry)->d_inode->u.nfs4_i.fsid_minor)

#define NFS4_BUFSIZE			8192
#define NFS4FS_MAX_IOV			10
#define NFS4_IOV_NWORDS			256
#define NFS4_IOV_NBYTES			(NFS4_IOV_NWORDS << 2)
#define NFS4_SETCLIENTID_MAXTRIES	5
#define NFS4_READDIR_MAXTRIES		5
#define NFS4_MAXIO			4
#define NFS4_MAX_REQUEST_SOFT		192
#define NFS4_MAX_REQUEST_HARD		256
#define NFS4_MAXCOMMIT			64
#define NFS4_READ_DELAY			(2 * HZ)
#define NFS4_WRITEBACK_DELAY		(5 * HZ)
#define NFS4_WRITEBACK_LOCKDELAY	(60 * HZ)
#define NFS4_COMMIT_DELAY		(5 * HZ)

/*
 *  nfs4_rwlock_t: We need a read-write semaphore which provides trylock() semantics.
 *  More specifically, we need the following primitives:
 *    nfs4_down_read() - blocking read lock
 *    nfs4_down_write() - blocking write lock
 *    nfs4_up_read() - read unlock
 *    nfs4_up_write() - write unlock
 *    nfs4_down_read_trylock() - nonblocking read lock, returns 1 if lock unavailable
 *    nfs4_down_write_trylock() - nonblocking write lock, returns 1 if lock unavailable
 *  Currently, there is no built-in kernel data structure which does the trick.  Maybe
 *  within the 2.5 timeframe, someone will add trylock() semantics to 'struct rw_semaphore'.
 *  Alternatively, it wouldn't be difficult to build such a data structure out of currently
 *  supported primitives (but with some extra overhead); maybe I'll do this if it
 *  doesn't look like trylock() semantics are going to happen.  In the meantime, I'm just
 *  using a struct semaphore, and losing the improved concurrency properties that would arise
 *  from having shared read locks.
 */
#if 0

typedef struct semaphore	nfs4_rwlock_t;
#define nfs4_down_read		down
#define nfs4_down_write		down
#define nfs4_up_read		up
#define nfs4_up_write		up
#define nfs4_down_read_trylock	down_trylock
#define nfs4_down_write_trylock	down_trylock
#define nfs4_init_rwlock	init_MUTEX

#else

typedef struct {
	int read_count;       /* negative if there are writers */
	struct rw_semaphore sem;
} nfs4_rwlock_t;

static inline void
nfs4_down_read(nfs4_rwlock_t *lock)
{
	lock_kernel();
	down_read(&lock->sem);
	if (lock->read_count < 0)
		BUG();
	lock->read_count++;
	unlock_kernel();
}

static inline void
nfs4_down_write(nfs4_rwlock_t *lock)
{
	lock_kernel();
	down_write(&lock->sem);
	if (lock->read_count != 0)
		BUG();
	lock->read_count--;
	unlock_kernel();
}

static inline void
nfs4_up_read(nfs4_rwlock_t *lock)
{
	lock_kernel();
	up_read(&lock->sem);
	if (lock->read_count <= 0)
		BUG();
	lock->read_count--;
	unlock_kernel();
}

static inline void
nfs4_up_write(nfs4_rwlock_t *lock)
{
	lock_kernel();
	up_write(&lock->sem);
	if (lock->read_count >= 0)
		BUG();
	lock->read_count++;
	unlock_kernel();
}

static inline int
nfs4_down_read_trylock(nfs4_rwlock_t *lock)
{
	lock_kernel();
	if (lock->read_count < 0) {
		unlock_kernel();
		return 1;
	}
	down_read(&lock->sem);   /* can't block */
	lock->read_count++;
	unlock_kernel();
	return 0;
}

static inline int
nfs4_down_write_trylock(nfs4_rwlock_t *lock)
{
	lock_kernel();
	if (lock->read_count > 0) {
		unlock_kernel();
		return 1;
	}
	down_write(&lock->sem);  /* can't block */
	lock->read_count--;
	unlock_kernel();
	return 0;
}

static inline void
nfs4_init_rwlock(nfs4_rwlock_t *lock)
{
	lock->read_count = 0;
	init_rwsem(&lock->sem);
}

#endif

struct nfs4fs_dentry_data {
	u32	fh_allocated;
	u32	fh_len;
	char	fh_val[1];
};
#define DDATA_SIZE(len)    (sizeof(struct nfs4fs_dentry_data) + len - 1)
#define NFS4_HAVE_FH(dentry)  ((dentry)->d_fsdata && \
			       ((struct nfs4fs_dentry_data *)(dentry)->d_fsdata)->fh_len)

static inline void
nfs4_drop_fh(struct dentry *dentry)
{
	struct nfs4fs_dentry_data *ddata = (struct nfs4fs_dentry_data *)dentry->d_fsdata;
	if (ddata)
		ddata->fh_len = 0;
}

static inline void
nfs4_zap_caches(struct inode *inode)
{
	NFS4_CACHE_TIMEOUT(inode) = NFS4_MINTIMEOUT(inode);
	NFS4_CACHE_TIMEOUT_JIFFIES(inode) = jiffies;
	invalidate_inode_pages(inode);    /* XXX: should this be truncate_inode_pages()? */
	NFS4_SNAPSHOT(inode) = 0;
}

/* According to fs/locks.c, locking state is separated by (files_struct, pid)-pair for each
   process.  In addition, an NFSv4 lockowner must maintain a seperate seqid for each mount.
   Therefore, an NFSv4 lockowner is a (superblock, files_struct, pid)-triple. */

struct nfs4fs_lockowner {
	struct list_head	lo_hash;     /* by superblock, files_struct, pid */
	struct list_head	lo_id_hash;  /* by id */
	struct super_block	*lo_sb;
	struct files_struct	*lo_fs;
	pid_t			lo_pid;
	u32			lo_id;
	u32			lo_seqid;
	struct semaphore	lo_sema;      /* held during seqid-mutating ops */
	atomic_t		lo_count;     /* refcount */
};

/* XXX: our state model depends on the assertion that, if two tasks share a struct file,
   they have the same files_struct and pid.  Find out whether this is always true! */

struct nfs4fs_file_data {
	struct list_head	fi_hash;    /* by inode, files_struct, pid */
	struct dentry		*fi_dentry;
	struct files_struct	*fi_fs;
	pid_t			fi_pid;
	struct nfs4fs_lockowner	*fi_lockowner;
	stateid4		fi_stateid;
	nfs4_rwlock_t		fi_sema;
	unsigned int		fi_nopen;   /* count of total opens */
	unsigned int		fi_nread;   /* count of opens for read */
	unsigned int		fi_nwrite;  /* count of opens for write */
	atomic_t		fi_count;   /* refcount */

	/* data needed for mmap().  the dentries in fi_{read,write}_map_dentry hold references. */
	/* Note to self: When I switch to a two-level structure for file_data, put these
	   in the first level, along with the credentials */
	struct list_head	fi_read_map;
	struct list_head	fi_write_map;

	/* this is needed by renewd and recoverd. */
	struct list_head	fi_list;   /* all files on this superblock */
};
#define fi_inode	fi_dentry->d_inode

struct nfs4fs_delegation {
	struct list_head	dl_hash;	/* hashed by superblock, filehandle */
	struct super_block	*dl_sb;
	struct inode		*dl_inode;	/* hold a reference to the inode */
	open_delegation_type4	dl_type;
	stateid4		dl_stateid;
};

struct compound4args {
	char			*tag;
	int			iov_nr;
	struct iovec		iov[NFS4FS_MAX_IOV];
	int			iov_static[NFS4FS_MAX_IOV];
	u32			*p;        /* buffer pos in current iovec */
	u32			*end;      /* end of buffer in current iovec */
	struct dentry		*current_fh;
	struct dentry		*save_fh;
	unsigned int		flags;

	/* Some saved pointers, for data that will be filled in "just in time". */
	u32			*nopsp;
	int			nops;
	u32			*seqidp;
	struct nfs4fs_lockowner *seqid_holder;
	unsigned int		nstateids;
	u32			*stateidp[NFS4_MAXIO];
	struct nfs4fs_file_data *stateid_holder;

	/* Data which the handler needs to do "generic" processing */
	struct super_block	*sb;                      /* needed for a couple of reasons */
	int			seqid_index;   /* nonzero for seqid-incrementing operations */
	int			renew_index;    /* nonzero for clientid-renewing operations */
	unsigned long		timestamp;                                    /* in jiffies */
};
/*
 * Flags for compound4args.
 */
#define	CA_SEQID_LOCKED		0x00000001	/* is seqid currently locked? */
#define CA_SEQID_RELEASE	0x00000002	/* should handler release seqid? */
#define CA_STATEID_LOCKED	0x00000004	/* is stateid currently locked? */
#define CA_STATEID_RELEASE	0x00000008	/* should handler release stateid? */
#define CA_NORETRY		0x00000010	/* on "stale" errors, don't wait & retry */
#define CA_FAIL_RETRY		0x00000020	/* on "stale" errors, fail completely */

#define FINALIZE_IOVEC(cp)    do {					\
	(cp)->iov[(cp)->iov_nr-1].iov_len =				\
		(char *)(cp)->p - (char *)(cp)->iov[(cp)->iov_nr-1].iov_base;	\
	nfsv4_printk(xdr, "xdr: finalizing iovec (len %d)\n",		\
		     (cp)->iov[(cp)->iov_nr-1].iov_len);		\
} while (0)

struct compound4res {
	int			iov_nr;
	int			iov_current;
	struct iovec iov[NFS4FS_MAX_IOV];
	char 			*p;                /* buffer pos in current iovec */
	char 			*end;              /* end of buffer in current iovec */
	int			toplevel_status;
	int			nops;
#if NFS4_DEBUG
	char *			startp;
	char *			savep;
#endif
};

/* context used to perform common error handling in nfs4_call_compound.
 *
 * enters as an input to nfs4_call_compound() who passes it thru to 
 * nfs4_handle_compound.
 * should be expanded as necessary to provide data to handle common 
 * compound rpc error returns. perhaps eventually a union..
 */
struct nfs4fs_ctx {
	struct super_block *pt_sb;
	struct dentry  *pt_dentry;
};

struct nfs4fs_fattr {
	unsigned int	fa_valid;
	nfs_ftype4	fa_type;
	loff_t		fa_size;
	u64		fa_fsid_major;
	u64		fa_fsid_minor;
	u64		fa_fileid;
	umode_t		fa_mode;
	nlink_t		fa_nlink;
	uid_t		fa_uid;
	gid_t		fa_gid;
	u32		fa_rdev_major;
	u32		fa_rdev_minor;
	time_t		fa_atime;
	time_t		fa_ctime;
	time_t		fa_mtime;
	u64		fa_maxread;
	u64		fa_maxwrite;
	u64		fa_ffree;
	u64		fa_ftotal;
	u32		fa_maxname;
	u64		fa_savail;
	u64		fa_sfree;
	u64		fa_stotal;
	u64		fa_changeid;
	u32		fa_lease_time;
	u64		fa_maxfilesize;
	unsigned int 	fa_nacl;	/* number of acl's allocated */
	nfs4_acl	*fa_acl;	/* array of size fa_nacl */
};

/* flags for fa_valid */
#define FA_SIZE		0x00000001
#define FA_FSID		0x00000002
#define FA_FILEID	0x00000004
#define FA_MODE		0x00000008
#define FA_NLINK	0x00000010
#define FA_UID 		0x00000020
#define FA_GID 		0x00000040
#define FA_RDEV		0x00000080
#define FA_ATIME	0x00000100
#define FA_CTIME	0x00000200
#define FA_MTIME	0x00000400
#define FA_MAXREAD	0x00000800
#define FA_MAXWRITE	0x00001000
#define FA_TYPE		0x00002000
#define FA_FFREE       	0x00004000
#define FA_FTOTAL       0x00008000
#define FA_MAXNAME	0x00010000
#define FA_SAVAIL	0x00020000
#define FA_SFREE 	0x00040000
#define FA_STOTAL	0x00080000
#define FA_CHANGEID	0x00100000
#define FA_LEASE_TIME	0x00200000
#define FA_MAXFILESIZE	0x00400000
#define FA_ACL		0x00800000

struct nfs4fs_getattr {
	/* filled in by caller */
	int flags;
	bitmap4 *bm;
	
	/* filled in by setup routine */
	struct dentry *dentry;

	/* filled in by handler */
	struct nfs4fs_fattr fattr;
};
/* flags for struct nfs4fs_getattr */
#define FL_NONFATAL		0x00000001
#define FL_NOREVAL		0x00000002
#define FL_CLOBBER_OK		0x00000004
#define FL_NOENT_OK             0x00000008

struct nfs4fs_getfh {
	/* filled in by caller */
	int nonfatal;

	/* filled in by setup routine */
	struct dentry *dentry;
};

struct nfs4fs_close {
	/* filled in by caller */
	struct nfs4fs_file_data *data;
	int read;
	int write;
	
	/* The CLOSE setup routine will either use a CLOSE or OPEN_DOWNGRADE, as needed.
	   We record the actual operation used here, for the handler's benefit. */
	u32 op;
};

struct nfs4fs_commit {
	/* filled in by caller */
	loff_t start;
	u32 length;
	struct list_head pages;

	/* filled in by handler */
	verifier4 verifier;
};

struct nfs4fs_create {
	/* filled in by caller */
	struct dentry *dentry;
	nfs_ftype4 type;
	union {
		umode_t mode;		/* for directory only */
		const char *linktext;	/* null-terminated, for symlink only */
		/* fields needed for mknod() will also be put in this union */
	} u;

	/* coalesced into the call by the setup routine */
	struct nfs4fs_getattr getattr;
	struct nfs4fs_getfh getfh;
};

struct nfs4fs_link {
	/* filled in by caller */
	struct dentry *dentry;

	/* filled in by setup routine */
	struct inode *inode;
};

struct nfs4fs_lock {
	/* filled in by caller */
	int reclaim;
	struct nfs4fs_file_data *file_data;
	nfs_lock_type4 type;
	loff_t start;
	loff_t end;
};

struct nfs4fs_lockt {
	/* filled in by caller */
	struct nfs4fs_lockowner *lockowner;

	/* filled in by caller, changed by handler */
	struct file_lock *lock;
};

struct nfs4fs_lookup {
	/* filled in by caller.  Note: we keep a separate field for the name, rather than
	   always using dentry->d_name, because nfs4_rename() needs to make them different. */
	int flags;
	struct dentry *dentry;
	struct qstr *name;

	/* coalesced into call */
	int did_getfh;
	struct nfs4fs_getattr getattr;   /* reval, fatal, lookup */
	struct nfs4fs_getfh getfh;       /* nonfatal */
};

/* Note: The 'flags' field in 'struct nfs4fs_open' uses the 'open_namei' convention for the
 * lower two bits (there are two conventions in use in the kernel!).  To quote fs/namei.c:
 *
 * Note that the low bits of "flag" aren't the same as in the open
 * system call - they are 00 - no permissions needed
 *			  01 - read permission needed
 *			  10 - write permission needed
 *			  11 - read/write permissions needed
 *
 * (Besides the lower two bits, the only flags which are actually used in nfs4fs_setup_open()
 *  are O_CREAT, O_EXCL, O_TRUNC.)
 */
struct nfs4fs_open {
	/* filled in by caller */
	open_claim_type4	claim_type;
	struct dentry		*dentry;	/* redundant for CLAIM_PREVIOUS, but still must be set */
	int			flags;
	umode_t			mode;		/* used only if O_CREAT is set */
	struct nfs4fs_lockowner	*lockowner;	
	struct nfs4fs_file_data *pre_file_data;

	/* filled in by handler */
	verifier4 verifier;
	struct nfs4fs_file_data *post_file_data;
	
	/* coalesced into call */
	struct nfs4fs_getattr getattr;
	struct nfs4fs_getfh getfh;
};

struct nfs4fs_putfh {
	/* filled in by caller */
	struct dentry *dentry;

	/* filled in by setup routine */
	nfs_opnum4 op;
	int nlookups;
};

struct nfs4fs_simple_read {
	/* filled in by caller */
	struct nfs4fs_file_data *file_data;
	loff_t start;
	u32 length;
	char *buf;

	/* filled in by handler */
	u32 bytes_read;
	int eof;
};

struct nfs4fs_read {
	/* filled in by caller */
	unsigned int npages;
	struct nfs4_page *pages[NFS4_MAXIO];

	/* filled in by handler */
	u32 bytes_read;   /* in case the server returns less than expected */
	int eof;
};

struct nfs4fs_readdir {
	/* Filled in by caller */
	u32 count;
	bitmap4 *bm;
	struct inode *inode;
	
	/* Filled in by caller, but changed by handler.  The caller must have locked and
	   kmap()'ed 'page'.  If the handler needs to replace 'page' with a new page, it
	   will unlock+kunmap() the old page, and lock+kmap() the new one.  Therefore,
	   after running the handler, the caller should always unlock+kunmap()
	   readdir.page (NOT the original page passed in) to avoid deadlocks. */
	nfs_cookie4 cookie;    /* in network byte order */
	verifier4 verifier;
	struct page *page;
	char *buf;             /* current position on page */
	char *end;             /* end of page */

	/* Filled in by handler */
	int eof;
	int bad_verifier;
};

struct nfs4fs_readlink {
	/* filled in by caller */
	char *buf;
	int nbytes;
};

struct nfs4fs_rename {
	struct dentry *old_dentry;
	struct dentry *new_dentry;
};

struct nfs4fs_secinfo {
	/* filled in by caller */
	struct super_block *sb;
	struct qstr *name;
};

struct nfs4fs_setclientid {
	/* filled in by caller */
	struct super_block *sb;
	int namlen;
	char *name;
	u32 callback_prog;
	char *callback_netid;
	char *callback_universal_address;
	
	/* filled in by handler */
	int in_use;
	clientid4 clientid;
};

struct nfs4fs_simple_write {
	/* filled in by caller */
	struct nfs4fs_file_data *file_data;
	loff_t start;
	u32 length;
	char *buf;

	/* filled in by handler */
	u32 bytes_written;
};

struct nfs4fs_write {
	/* filled in by caller */
	u32 how;    /* UNSTABLE4, DATA_SYNC4, or FILE_SYNC4 */
	unsigned int npages;
	struct nfs4_page *pages[NFS4_MAXIO];
};

/*
 *  Context of an async compound call.  This can always be found in the ->tk_calldata field
 *  of an async RPC task.
 */
struct nfs4_async_compound {
	struct rpc_task		ac_task;
	struct compound4args	ac_arg;
	struct compound4res	ac_res;
	void			(*ac_release)(struct nfs4_async_compound *);
	union {
		struct {
			struct nfs4fs_putfh putfh;
			struct nfs4fs_commit commit;
			struct nfs4fs_getattr reval;
		} commit;
		struct {
			stateid4 stateid;
		} delegreturn;
		struct {
			struct nfs4fs_putfh putfh;
			struct nfs4fs_read read;
			struct nfs4fs_getattr reval;
		} read;
		struct {
			struct super_block *sb;
			struct nfs4fs_file_data *fp;         /* holds reference */
		} renew;
		struct {
			struct rpc_wait_queue wait;
		} secinfo;
		struct {
			struct dentry *dentry;
			struct inode *dir_inode;             /* holds reference */
			struct nfs4fs_putfh putfh;
			struct list_head list;
		} unlink;
		struct {
			struct nfs4fs_putfh putfh;
			struct nfs4fs_write write;
			struct nfs4fs_getattr reval;
		} write;
	} u;
};
#define ac_commit_putfh		u.commit.putfh
#define ac_commit		u.commit.commit
#define ac_commit_reval		u.commit.reval
#define ac_delegreturn_stateid	u.delegreturn.stateid
#define ac_read_putfh		u.read.putfh
#define ac_read			u.read.read
#define ac_read_reval		u.read.reval
#define ac_renew_sb		u.renew.sb
#define ac_renew_fp		u.renew.fp
#define ac_secinfo_wait		u.secinfo.wait
#define ac_unlink_dentry	u.unlink.dentry
#define ac_unlink_dir_inode	u.unlink.dir_inode
#define ac_unlink_putfh		u.unlink.putfh
#define ac_unlink_list		u.unlink.list
#define ac_write_putfh		u.write.putfh
#define ac_write		u.write.write
#define ac_write_reval		u.write.reval

/*
 *  struct nfs4_page: maintains state for a single non-uptodate page
 */

struct nfs4_page {
	struct list_head	wb_list;
	struct list_head	wb_writelist;
	wait_queue_head_t	wb_wait;
	unsigned long		wb_flags;
	unsigned long		wb_timeout;            /* in jiffies */
	struct file		*wb_file;              /* holds reference, can be NULL */
	struct nfs4fs_file_data	*wb_fdata;             /* holds reference */
	struct page		*wb_page;              /* holds reference */
	struct inode		*wb_inode;             /* no reference held */
	unsigned int		wb_offset, wb_bytes;   /* defines byterange in page */
	unsigned int		wb_count;  /* refcount, protected by nfs4_wreq_lock */
	char			*wb_buf;
	u32			wb_how_written;
	verifier4		wb_verifier;
	int			wb_safeguard;   /* protects against pathological servers */
};

/* only the one flag is defined for wb_flags */
#define PG_BUSY 0

extern spinlock_t		nfs4_wreq_lock;   /* actually protects both read and write requests */

#ifndef FLUSH_AGING
/*
 * When flushing a cluster of dirty pages, there can be different
 * strategies:
 */
#define FLUSH_AGING		0	/* only flush old buffers */
#define FLUSH_SYNC		1	/* file being synced, or contention */
#define FLUSH_WAIT		2	/* wait for completion */
#define FLUSH_STABLE		4	/* commit to stable storage */
#endif

/* encode.c */
extern int nfs4_setup_compound(struct compound4args *cp);
extern int nfs4_setup_simple_op(struct compound4args *cp, nfs_opnum4 op);
extern int nfs4_setup_close(struct compound4args *cp, struct nfs4fs_close *close);
extern int nfs4_setup_commit(struct compound4args *cp, struct nfs4fs_commit *commit);
extern int nfs4_setup_create(struct compound4args *cp, struct nfs4fs_create *create);
extern int nfs4_setup_delegreturn(struct compound4args *cp, stateid4 stateid);
extern int nfs4_setup_getattr(struct compound4args *cp, struct nfs4fs_getattr *getattr);
extern int nfs4_setup_getfh(struct compound4args *cp, struct nfs4fs_getfh *getfh);
extern int nfs4_setup_link(struct compound4args *cp, struct nfs4fs_link *link);
extern int nfs4_setup_lock(struct compound4args *cp, struct nfs4fs_lock *lock);
extern int nfs4_setup_lockt(struct compound4args *cp, struct nfs4fs_lockt *lockt);
extern int nfs4_setup_locku(struct compound4args *cp, struct nfs4fs_lock *lock);
extern int nfs4_setup_lookup(struct compound4args *cp, struct nfs4fs_lookup *lookup);
extern int nfs4_setup_open(struct compound4args *cp, struct nfs4fs_open *open);
extern int nfs4_setup_open_confirm(struct compound4args *cp, struct nfs4fs_open *open);
extern int nfs4_setup_putfh(struct compound4args *cp, struct nfs4fs_putfh *putfh);
#define    nfs4_setup_putrootfh(cp)    nfs4_setup_simple_op((cp), OP_PUTROOTFH)
extern int nfs4_setup_read(struct compound4args *cp, struct nfs4fs_read *read);
extern int nfs4_setup_simple_read(struct compound4args *cp, struct nfs4fs_simple_read *read);
extern int nfs4_setup_readdir(struct compound4args *cp, struct nfs4fs_readdir *readdir);
#define    nfs4_setup_readlink(cp)     nfs4_setup_simple_op((cp), OP_READLINK)
extern int nfs4_setup_remove(struct compound4args *cp, struct qstr *name);
extern int nfs4_setup_rename(struct compound4args *cp, struct nfs4fs_rename *rename);
extern int nfs4_setup_renew(struct compound4args *cp, struct nfs4fs_file_data *fp);
extern int nfs4_setup_restorefh(struct compound4args *cp);
extern int nfs4_setup_reval(struct compound4args *cp, struct nfs4fs_getattr *reval);
extern int nfs4_setup_savefh(struct compound4args *cp);
extern int nfs4_setup_setattr(struct compound4args *cp, struct iattr *iap,
			      struct nfs4fs_file_data *fdata);
extern int nfs4_setup_setattr_mode(struct compound4args *cp, umode_t mode);
extern int nfs4_setup_secinfo(struct compound4args *cp, struct nfs4fs_secinfo *secinfo);
extern int nfs4_setup_setclientid(struct compound4args *cp,
				  struct nfs4fs_setclientid *setclientid);
extern int nfs4_setup_setclientid_confirm(struct compound4args *cp, verifier4 verifier);
extern int nfs4_setup_write(struct compound4args *cp, struct nfs4fs_write *write);
extern int nfs4_setup_simple_write(struct compound4args *cp,
				   struct nfs4fs_simple_write *write);
extern bitmap4 nfs4_positive_bitmap;
extern bitmap4 nfs4_negative_bitmap;

/* decode.c */
extern int nfs4_handle_compound(struct compound4args *argp, struct compound4res *resp,
				struct nfs4fs_ctx *ctx);
extern int nfs4_handle_simple_op(struct compound4res *rp, nfs_opnum4 op);
extern int nfs4_handle_close(struct compound4res *rp, struct nfs4fs_close *close);
extern int nfs4_handle_commit(struct compound4res *rp, struct nfs4fs_commit *commit);
extern int nfs4_handle_create(struct compound4res *rp, struct nfs4fs_create *create);
#define    nfs4_handle_delegreturn(rp)	nfs4_handle_simple_op(rp, OP_DELEGRETURN)
extern int nfs4_handle_getattr(struct compound4res *rp, struct nfs4fs_getattr *getattr);
extern int nfs4_handle_getfh(struct compound4res *rp, struct nfs4fs_getfh *getfh);
extern int nfs4_handle_link(struct compound4res *rp, struct nfs4fs_link *link);
extern int nfs4_handle_lock(struct compound4res *rp, struct nfs4fs_lock *lock);
extern int nfs4_handle_lockt(struct compound4res *rp, struct nfs4fs_lockt *lockt);
extern int nfs4_handle_locku(struct compound4res *rp, struct nfs4fs_lock *lock);
extern int nfs4_handle_lookup(struct compound4res *rp, struct nfs4fs_lookup *lookup);
extern int nfs4_handle_open(struct compound4args *argp, struct compound4res *rp,
			    struct nfs4fs_open *open);
extern int nfs4_handle_open_confirm(struct compound4res *rp, struct nfs4fs_open *open);
extern int nfs4_handle_putfh(struct compound4res *rp, struct nfs4fs_putfh *putfh);
#define    nfs4_handle_putrootfh(rp)	nfs4_handle_simple_op(rp, OP_PUTROOTFH)
extern int nfs4_handle_read(struct compound4res *rp, struct nfs4fs_read *read);
extern int nfs4_handle_simple_read(struct compound4res *rp, struct nfs4fs_simple_read *read);
extern int nfs4_handle_readdir(struct compound4res *rp, struct nfs4fs_readdir *readdir);
extern int nfs4_handle_readlink(struct compound4res *rp, struct nfs4fs_readlink *readlink);
extern int nfs4_handle_remove(struct compound4res *rp, struct inode *dir_inode);
extern int nfs4_handle_rename(struct compound4res *rp, struct nfs4fs_rename *rename);
#define    nfs4_handle_renew(rp)        nfs4_handle_simple_op(rp, OP_RENEW)
#define    nfs4_handle_restorefh(rp)	nfs4_handle_simple_op(rp, OP_RESTOREFH)
#define    nfs4_handle_reval(rp, reval) nfs4_handle_getattr(rp, reval)
#define    nfs4_handle_savefh(rp)	nfs4_handle_simple_op(rp, OP_SAVEFH)
extern int nfs4_handle_setattr(struct compound4res *rp);
extern int nfs4_handle_secinfo(struct compound4res *rp, struct nfs4fs_secinfo *secinfo);
extern int nfs4_handle_setclientid(struct compound4res *rp,
				   struct nfs4fs_setclientid *setclientid);
extern int nfs4_handle_setclientid_confirm(struct compound4res *rp,
					   struct nfs4fs_setclientid *setclientid);
extern int nfs4_handle_write(struct compound4res *rp, struct nfs4fs_write *write);
extern int nfs4_handle_simple_write(struct compound4res *rp,
				    struct nfs4fs_simple_write *write);
#if NFS4_DEBUG
extern void debug_dump(struct compound4res *rp, int nbytes);
#endif

/* inode.c */
extern int nfs4_inode_revalidate(struct dentry *dentry);
extern int nfs4_inode_revalidate_wait(struct dentry *dentry);
extern int nfs4_setattr(struct dentry *dentry, struct iattr *iap);
extern int nfs4_do_secinfo(struct super_block *sb,struct dentry *dentry);
extern int nfs4_do_secinfo_async(struct nfs4_async_compound *ctx);
extern int nfs4_do_setclientid(struct super_block *sb);
extern int nfs4fs_start_recoverd(struct super_block *sb);
extern int nfs4fs_wait_on_recoverd(struct super_block *sb);

/* dir.c */
extern struct dentry_operations nfs4_dentry_operations;
extern struct inode_operations nfs4_dir_inode_operations;
extern struct file_operations nfs4_dir_operations;
int nfs4_dentry_revalidate(struct dentry *dentry, int flags);
int nfs4fs_do_lookup(struct dentry *dentry);
int nfs4_remove(struct inode *dir_inode, struct dentry *dentry);

/* open.c */
extern void nfs4fs_init_state(void);
extern void nfs4fs_cleanup_state(void);
extern struct nfs4fs_lockowner *__nfs4_get_lockowner(struct super_block *sb,
						     struct files_struct *fs, pid_t pid);
#define nfs4_get_lockowner(sb)  __nfs4_get_lockowner((sb), current->files, current->pid)
extern struct nfs4fs_lockowner *nfs4_get_lockowner_by_id(unsigned int id);
extern void nfs4_put_lockowner(struct nfs4fs_lockowner *lockowner);
extern struct nfs4fs_file_data *nfs4_get_file_data(struct dentry *dentry);
extern void nfs4_put_file_data(struct nfs4fs_file_data *file_data);
extern int nfs4fs_do_open(struct nfs4fs_open *open);
extern int nfs4fs_do_close(struct dentry *dentry, struct nfs4fs_close *close);
extern int nfs4_create(struct inode *dir_inode, struct dentry *dentry, int mode);
extern int nfs4_lookup_open(struct inode *dir_inode, struct dentry *dentry,
			    struct nameidata *nd);
extern int nfs4_open(struct inode *inode, struct file *filp);
extern int nfs4_release(struct inode *inode, struct file *filp);
extern int nfs4fs_recoverd(void *arg);

/* file.c */
extern struct inode_operations nfs4_file_inode_operations;
extern struct file_operations nfs4_file_operations;
extern struct address_space_operations nfs4_file_aops;
extern wait_queue_head_t nfs4_wait;
extern struct nfs4_page *nfs4_create_request(struct file *file, struct nfs4fs_file_data *fdata,
					     struct page *page, unsigned int offset,
					     unsigned int count);
extern void _nfs4_release_request(struct nfs4_page *req);
extern void nfs4_list_add_request(struct nfs4_page *req, struct list_head *head);
extern int nfs4_scan_list(struct list_head *src, struct list_head *dst,
			  struct file *file, unsigned long idx_start,
			  unsigned int npages, unsigned int *bytecount);
extern int nfs4_scan_list_timeout(struct list_head *head, struct list_head *dst,
				  struct inode *inode, unsigned int *bytecount);

static inline void
nfs4_release_request(struct nfs4_page *req)
{
	spin_lock(&nfs4_wreq_lock);
	if (!req->wb_count)
		BUG();
	if (!--req->wb_count) {
		spin_unlock(&nfs4_wreq_lock);
		_nfs4_release_request(req);
	}
	else
		spin_unlock(&nfs4_wreq_lock);
}

/*
 * Warning! Callers of nfs4_lock_request() must hold the spinlock.
 * Callers of nfs4_unlock_requst() must NOT hold the spinlock.
 * (Confusing, but this is what nfsv3 does...)
 */
static inline int
nfs4_lock_request(struct nfs4_page *req)
{
	if (test_and_set_bit(PG_BUSY, &req->wb_flags))
		return 0;
	req->wb_count++;
	return 1;
}

static inline void
nfs4_unlock_request(struct nfs4_page *req)
{
	if (!test_bit(PG_BUSY, &req->wb_flags))
		BUG();
	if (req->wb_buf)
		BUG();
	smp_mb__before_clear_bit();
	clear_bit(PG_BUSY, &req->wb_flags);
	smp_mb__after_clear_bit();
	if (waitqueue_active(&req->wb_wait))
		wake_up(&req->wb_wait);
	if (!req->wb_count)
		BUG();
	nfs4_release_request(req);
}

static inline void
nfs4_kmap_request(struct nfs4_page *req)
{
	if (req->wb_buf)
		BUG();
	if (!test_bit(PG_BUSY, &req->wb_flags))
		BUG();
	req->wb_buf = kmap(req->wb_page);   /* XXX: can this return NULL? */
	if (!req->wb_buf)
		BUG();
}

static inline void
nfs4_kunmap_request(struct nfs4_page *req)
{
	if (!req->wb_buf)
		BUG();
	if (!test_bit(PG_BUSY, &req->wb_flags))
		BUG();
	kunmap(req->wb_page);
	req->wb_buf = NULL;
}

/* read.c */
extern int nfs4_readpage(struct file *filp, struct page *page);
extern int nfs4_pagein_inode(struct inode *inode, unsigned long idx_start,
			     unsigned int npages);
extern int nfs4_pagein_timeout(struct inode *inode);

/* write.c */
extern int nfs4_writepage(struct page *page);
extern int nfs4_updatepage(struct file *file, struct page *page,
			   unsigned int offset, unsigned int count);
extern int nfs4_flush_incompatible(struct file *file, struct page *page);
extern int nfs4_flush_timeout(struct inode *inode, int how);
extern int nfs4_sync_file(struct inode *inode, struct file *file,
			  unsigned long idx_start, unsigned int npages, int how);
extern int nfs4_commit_timeout(struct inode *inode, int how);

/* lock.c */
extern int nfs4_lock(struct file *filp, int cmd, struct file_lock *file_lock);

/* symlink.c */
extern struct inode_operations nfs4_symlink_inode_operations;

/* proc.c */
extern struct rpc_program nfs4_program;
extern struct rpc_stat nfs4_rpcstat;
extern int nfs4_call_compound(struct compound4args *argp, struct compound4res *resp,
			      struct nfs4fs_ctx *ctx);
extern void nfs4_release_compoundargs(struct compound4args *cp);
extern void nfs4_release_compoundres(struct compound4res *rp);
extern struct nfs4_async_compound *nfs4_async_create(struct super_block *sb);
extern void nfs4_async_call_compound(struct rpc_task *task);
extern int nfs4_async_prehandler(struct nfs4_async_compound *cp, struct nfs4fs_ctx *ctx);
extern void nfs4_async_release(struct rpc_task *task);

/* silly.c */
int nfs4_sillyrename(struct inode *dir, struct dentry *dentry);
void nfs4_silly_dentry_iput(struct dentry *dentry, struct inode *inode);

/* renewd.c */
extern int nfs4fs_init_renewd(struct super_block *sb);

/* acl.c */
//#ifdef CONFIG_EXT2_FS_EXT_ATTR
//#ifdef CONFIG_FS_POSIX_ACL
//#ifdef NFS4ACL
extern int nfs4_posix_acl_permission(struct inode *inode, int mask);
extern posix_acl_t *nfs4_acl_get_inode(struct inode *inode, int type);
extern int nfs4_acl_set_inode(struct inode *inode, int type, posix_acl_t *acl);
//#endif /* CONFIG_EXT2_FS_EXT_ATTR */
extern int nfs4_get_ext_attr(struct inode *inode, const char *name, 
		void *buffer, size_t buffer_size, int flags);
extern int nfs4_set_ext_attr(struct inode *inode, const char *name, 
		void *buffer, size_t buffer_size, int flags);
//#endif
//#endif /* CONFIG_FS_POSIX_ACL */

/* flushd.c */
extern int nfs4fs_init_flushd(struct super_block *sb);
extern void nfs4_wake_flushd(void);
extern void nfs4_schedule_inode(struct inode *inode, unsigned long time);

/* callback.c */
extern int nfs4_create_callback(struct super_block *sb);
extern void nfs4_get_callback(struct super_block *sb, u32 *progp, char *netid, char *uaddr);
extern void nfs4_destroy_callback(struct super_block *sb);
extern int nfs4_callback_sb(struct svc_rqst *rqstp, struct super_block **sbp);
extern void nfs4_accept_delegation(struct inode *inode, u32 type, stateid4 stateid);
extern struct nfs4fs_delegation *nfs4_find_delegation(struct super_block *sb, stateid4 stateid);
extern int nfs4_exists_delegation(struct super_block *sb, stateid4 stateid);
extern void nfs4fs_init_callbacks(void);

/* callback_proc.c */
extern struct svc_version nfs4_cb_version1;
extern int nfs4_do_delegreturn(struct super_block *sb, stateid4 stateid);
extern int nfs4_exists_delegation(struct super_block *sb, stateid4 stateid);

#endif
