/*
 *  linux/fs/nfs4/nfsd4.h
 *
 *  Copyright (c) 2001 The Regents of the University of Michigan.
 *  All rights reserved.
 *
 *  Kendrick Smith <kmsmith@umich.edu>
 *  Andy Adamson <andros@umich.edu>
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifndef _LINUX_NFS4_NFSD4_H
#define _LINUX_NFS4_NFSD4_H

#include <linux/nfs4/syscall.h>

#define  ALLOW_NULL_CLIENTID   1
#define  ALWAYS_OPEN_CONFIRM   0
#define  NFSD4_STRICT_LOCKING  0


#define NFSSVC_XDRSIZE		64        /* XXX : 20? */

#define NFS4_MAXPATHLEN		PATH_MAX  /* defined in limits.h as 4095 */
#define NFS4_MAXNAMLEN		NAME_MAX
#define NFS4_LEASE_TIME		60	  /* seconds */
#define NFS4_LAUNDROMAT_MINTIMEOUT    10   /* seconds */
#define NFS4_MAX_EXPORTS	64

/* v3 artifacts, should go away soon! */
#define NFSSVC_MAXBLKSIZE	8192
#define NFSD_BUFSIZE	    (1024 + NFSSVC_MAXBLKSIZE)

#define IS_ISMNDLK(i)   (S_ISREG((i)->i_mode) && MANDATORY_LOCK(i))

struct nfs4_export {
	/* CODE CLEANUP: use a union */
	unsigned int		ex_namlen;
	char			*ex_name;
	unsigned int		ex_id;
	struct nfs4_export	*ex_parent;
	struct list_head	ex_subdirs;

	unsigned int		ex_exportnamlen;   /* only if ex_dentry is non-NULL */
	char			*ex_exportname;    /* only if ex_dentry is non-NULL */
	unsigned int		ex_nchild;    /* only if ex_dentry is NULL */
	struct list_head	ex_child;     /* only if ex_dentry is NULL */

	/* Jake: you can probably safely ignore the fields before this point. */
	struct dentry		*ex_dentry;   /* non-NULL if export, NULL if "internal" */
	unsigned int		ex_readonly;  /* only if ex_dentry is non-NULL */

	/* Sec modes */	
	unsigned int 		ex_sec_none;	/* AUTH_NONE RPC security */	
	unsigned int 		ex_sec_sys;	/* AUTH_SYS RPC, default uid/gid lists */
	unsigned int 		ex_sec_dh;      /* AUTH_DES, old diffe-hellman encryption */
	unsigned int 		ex_sec_krb5;	/* RPCSEC_GSS Kerberos 5 authenication */
	unsigned int 		ex_sec_krb5i;	/* RPCSEC_GSS Kerberos 5 integrity */
	unsigned int 		ex_sec_krb5p;	/* RPCSEC_GSS Kerberso 5 protection */
	unsigned int 		ex_sec_spkm3;	/* SPKM authenication */
	unsigned int 		ex_sec_spkm3i;  /* SPKM integrity */
	unsigned int 		ex_sec_spkm3p;  /* SPKM protection */
	unsigned int 		ex_sec_lkey;	/* LIPKEY authenication */
	unsigned int 		ex_sec_lkeyi;   /* LIPKEY integrity */
	unsigned int 		ex_sec_lkeyp;   /* LIPKEY protection */
	
	/* ANDROS: list of security flavors and accompanying qop's */
	SECINFO4resok		ex_secinfo;
};

/*
 * Per-(export, client) pair security attributes
 */
struct nfs4_secattrs {
	uid_t			sa_fsuid;
	gid_t			sa_fsgid;
	kernel_cap_t		sa_cap_effective;
	int			sa_access;
};
/*
 * Special flags for sa_access. These must be different from MAY_READ,
 * MAY_WRITE, and MAY_EXEC.
 */
#define MAY_SATTR		8
#define MAY_TRUNC		16
#define MAY_LOCK		32
#define MAY_OWNER_OVERRIDE	64
#define MAY_CREATE		(MAY_EXEC|MAY_WRITE)
#define MAY_REMOVE		(MAY_EXEC|MAY_WRITE|MAY_TRUNC)

extern void nfs4_export_init(void);
extern void nfs4_export_readlock(void);
extern void nfs4_export_readunlock(void);
extern int nfs4_addpseudo(struct nfs4ctl_addpseudo *arg, union nfsctl_res *res);
extern int nfs4_export(struct nfs4ctl_export *arg);
extern int nfs4_setpubfh(unsigned int export_id);
extern int nfs4_setexp(unsigned int export_id, struct nfs4_export **expp);
extern int nfs4_setroot(struct nfs4_export **expp);
extern int nfs4_setpub(struct nfs4_export **expp);
extern int nfs4_setsec(struct svc_rqst *rqstp, struct nfs4_export *exp,
		       struct nfs4_secattrs *attrs);
extern void nfs4_export_shutdown(void);

/*
 * The on-the-wire file handle -- not actually used internally
 */
struct knfs_fh {
	__u32		fb_export_id;
	/* remaining fields are only used if not a pseudo fh */
	__u32		fb_ino;		/* our inode number */
	__u32		fb_generation;  /* our inode generation number */
	__u32		fb_dirino;	/* dir inode number, 0 for directories */
};

/*
 * This is the internal representation of an NFS handle used in knfsd.
 */
typedef struct svc_fh {
	unsigned int		fh_initialized;
	struct nfs4_export	*fh_export;
	struct nfs4_secattrs	fh_secattrs;
	struct dentry		*fh_dentry;   /* NULL if in pseudofs */
} svc_fh;
#define fh_fsuid		fh_secattrs.sa_fsuid
#define fh_fsgid		fh_secattrs.sa_fsgid
#define fh_cap_effective	fh_secattrs.sa_cap_effective
#define fh_access		fh_secattrs.sa_access
#define PSEUDO_FH_LEN		4
#define NONPSEUDO_FH_LEN	sizeof(struct knfs_fh)

/*
 *  This is the context of a single nfsd kernel thread.
 */
struct global_state {
        struct svc_rqst		*rqstp;
        struct svc_fh		current_fh;
        struct svc_fh		save_fh;
        struct nfs4_lockowner	*current_lockowner;
        struct nfs4_openfile	*current_openfile;
};

/*
 * nfs server stats
 */
struct nfsd_stats {
	unsigned int	rchits;		/* repcache hits */
	unsigned int	rcmisses;	/* repcache hits */
	unsigned int	rcnocache;	/* uncached reqs */
	unsigned int	fh_stale;	/* FH stale error */
	unsigned int	fh_lookup;	/* dentry cached */
	unsigned int	fh_anon;	/* anon file dentry returned */
	unsigned int	fh_nocache_dir;	/* filehandle not found in dcache */
	unsigned int	fh_nocache_nondir;	/* filehandle not found in dcache */
	unsigned int	io_read;	/* bytes returned to read requests */
	unsigned int	io_write;	/* bytes passed in write requests */
	unsigned int	th_cnt;		/* number of available threads */
	unsigned int	th_usage[10];	/* number of ticks during which n perdeciles
					 * of available threads were in use */
	unsigned int	th_fullcnt;	/* number of times last free thread was used */
	unsigned int	ra_size;	/* size of ra cache */
	unsigned int	ra_depth[11];	/* number of times ra entry was found that deep
					 * in the cache (10percentiles). [10] = not found */
};
/* thread usage wraps very million seconds (approx one fortnight) */
/* XXX : what is this?! */
#define	NFSD_USAGE_WRAP	(HZ*1000000)

/* nfs4ctl.c */
extern struct rw_semaphore nfs4_ctl_sema;

/* nfs4svc.c */
extern struct svc_program nfsd_program;
extern struct timeval nfssvc_boot;
extern int nfs4_start(struct nfs4ctl_start *arg);
extern int nfs4_refresh(union nfsctl_res *resp);
extern void nfs4_svc_shutdown(void);

/* nfs4fh.c */
extern int fh_read(struct svc_rqst *rqstp, struct svc_fh *fhp, nfs_fh4 *nfs_fh);
extern int fh_write(struct nfs4_export *exp, struct dentry *dentry, u32 *p);
extern void print_svcfh(struct svc_fh *fh);

/* nfs4proc.c */
extern void nfs4_release_compoundargs(COMPOUND4args *argp);
extern struct svc_procedure nfsd_procedures4[];

/* nfs4compound.c */
extern int nfsd4_proc_compound(struct global_state *stp, COMPOUND4args *argp,
			       COMPOUND4res *resp);

/* nfs4vfs.c */
extern int do_setattr(struct svc_fh *fhp, struct iattr *iap);
extern int nfs4_access(struct global_state *stp, ACCESS4args *argp, ACCESS4res *resp);
extern int nfs4_close(struct global_state *stp, CLOSE4args *argp, CLOSE4res *resp);
extern int nfs4_commit(struct global_state *stp, COMMIT4args *argp, COMMIT4res *resp);
extern int nfs4_create(struct global_state *stp, CREATE4args *argp, CREATE4res *resp);
extern int nfs4_delegpurge(struct global_state *stp, DELEGPURGE4args *argp,
			   DELEGPURGE4res *resp);
extern int nfs4_delegreturn(struct global_state *stp, DELEGRETURN4args *argp,
			    DELEGRETURN4res *resp);
extern int nfs4_getattr(struct global_state *stp, GETATTR4args *argp, GETATTR4res *resp);
extern int nfs4_getfh(struct global_state *stp, GETFH4res *resp);
extern int nfs4_link(struct global_state *stp, LINK4args *argp, LINK4res *resp);
extern int nfs4_lock(struct global_state *stp, LOCK4args *argp, LOCK4res *resp);
extern int nfs4_lockt(struct global_state *stp, LOCKT4args *argp, LOCKT4res *resp);
extern int nfs4_locku(struct global_state *stp, LOCKU4args *argp, LOCKU4res *resp);
extern int nfs4_lookup(struct global_state *stp, LOOKUP4args *argp, LOOKUP4res *resp);
extern int nfs4_lookupp(struct global_state *stp, LOOKUPP4res *resp);
extern int nfs4_nverify(struct global_state *stp, NVERIFY4args *argp, NVERIFY4res *resp);
extern int nfs4_open(struct global_state *stp, OPEN4args *argp, OPEN4res *resp);
extern int nfs4_openattr(struct global_state *stp, OPENATTR4res *resp);
extern int nfs4_open_confirm(struct global_state *stp, OPEN_CONFIRM4args *argp,
			     OPEN_CONFIRM4res *resp);
extern int nfs4_open_downgrade(struct global_state *stp, OPEN_DOWNGRADE4args *argp,
			       OPEN_DOWNGRADE4res *resp);
extern int nfs4_putfh(struct global_state *stp, PUTFH4args *argp, PUTFH4res *resp);
extern int nfs4_putpubfh(struct global_state *stp, PUTPUBFH4res *resp);
extern int nfs4_putrootfh(struct global_state *stp, PUTROOTFH4res *resp);
extern int nfs4_read(struct global_state *stp, READ4args *argp, READ4res *resp);
extern int nfs4_readdir(struct global_state *stp, READDIR4args *argp, READDIR4res *resp);
extern int nfs4_readlink(struct global_state *stp, READLINK4res *resp);
extern int nfs4_remove(struct global_state *stp, REMOVE4args *argp, REMOVE4res *resp);
extern int nfs4_rename(struct global_state *stp, RENAME4args *argp, RENAME4res *resp);
extern int nfs4_renew(struct global_state *stp, RENEW4args *argp, RENEW4res *resp);
extern int nfs4_restorefh(struct global_state *stp, RESTOREFH4res *resp);
extern int nfs4_savefh(struct global_state *stp, SAVEFH4res *resp);
extern int nfs4_secinfo(struct global_state *stp, SECINFO4args *argp, SECINFO4res *resp);
extern int nfs4_setattr(struct global_state *stp, SETATTR4args *argp, SETATTR4res *resp);
extern int nfs4_setclientid(struct global_state *stp, SETCLIENTID4args *argp,
			    SETCLIENTID4res *resp);
extern int nfs4_setclientid_confirm(struct global_state *stp, SETCLIENTID_CONFIRM4args *argp,
				    SETCLIENTID_CONFIRM4res *resp);
extern int nfs4_verify(struct global_state *stp, VERIFY4args *argp, VERIFY4res *resp);
extern int nfs4_write(struct global_state *stp, WRITE4args *argp, WRITE4res *resp);

/* nfs4stats.c */
extern struct nfsd_stats	nfsdstats;
extern struct svc_stat		nfsd_svcstats;
extern void nfsd_stat_init(void);
extern void nfsd_stat_shutdown(void);

static inline void
fh_copy(struct svc_fh *dst, struct svc_fh *src)
{
	/* just copy field-by-field; the only catch is that we might have to take care of
	   dentry refcounts.  Caller must ensure that src is initialized. */
	if (dst->fh_initialized && dst->fh_dentry)
		DPUT(dst->fh_dentry);
	memcpy(dst, src, sizeof(struct svc_fh));
	if (dst->fh_dentry)
		DGET(dst->fh_dentry);
}

static inline void
fh_init(struct svc_fh *fhp)
{
	fhp->fh_initialized = 0;
}

static inline void
fh_put(struct svc_fh *fhp)
{
	if (fhp->fh_initialized)
		DPUT(fhp->fh_dentry);
}

/*
 * Conversion macros for the filehandle fields.
 */
extern inline __u32 kdev_t_to_u32(kdev_t dev)
{
	return (__u32) dev;
}

extern inline kdev_t u32_to_kdev_t(__u32 udev)
{
	return (kdev_t) udev;
}

extern inline __u32 ino_t_to_u32(ino_t ino)
{
	return (__u32) ino;
}

extern inline ino_t u32_to_ino_t(__u32 uino)
{
	return (ino_t) uino;
}

static inline int nfsd_racache_init(int n) { return 0; }
static inline void nfsd_racache_shutdown(void) { return; }

#endif
