#ifndef _NFSD4_STATE_H
#define _NFSD4_STATE_H

/*
 * Note: We currently set NFSD4_LOCKOWNER_MAXNAME to a small value, and keep a static buffer
 * of size NFSD4_LOCKOWNER_MAXNAME with every lockowner.  This is because lockowner sizes
 * seem to be small in practice.  However, we may ultimately end up having to increase the
 * value, and switch from static to dynamic buffering.
 */
#define NFSD4_CLIENT_MAXNAME      64
#define NFSD4_LOCKOWNER_MAXNAME   16

/* A reasonable value for REPLAY_ISIZE was estimated as follows:  The OPEN response, typically
 * the largest, requires
 * 4(status) + 8(stateid) + 20(changeinfo) + 4(rflags) + 8(verifier) + 4(deleg. type) +
 * 8(deleg. stateid) + 4(deleg. recall flag) + 20(deleg. space limit) + ~32(deleg. ace)
 * = 112 bytes
 */
#define NFSD4_REPLAY_ISIZE       112

/*
 *  Files state.h and state.c implement the NFSv4 state management protocol, written
 *  with ease of portability in mind.  First, there is a small platform-dependent part
 *  consisting of the following primitives:
 *    get_time() - a routine returning the current time in seconds (used for lease timeouts)
 *    nfs4_ino_desc_t - typedef for a descriptor which represents a unique file on the system.
 *                      These will be memcmp()'ed to determine whether two files are the same.
 *    nfs4_file_desc_t - typedef for a descriptor for an open file on the system.  These are
 *                       opened, file I/O is subsequently done, then they are closed.
 *  Everything after this should be platform-independent, and should compile once the
 *  primitives are defined for a given platform.
 */

#ifdef __linux__

#define get_time()  CURRENT_TIME
typedef struct {
	kdev_t			dev;
	unsigned long		ino;
	u32			generation;
} nfs4_ino_desc_t;
#define nfs4_file_desc_t        struct file

#endif

/*
 *  We are using BSD's list manipulation macros to avoid licensing issues when trying to
 *  move Linux code into the BSD kernel.  Non-BSD systems will need this snippet from
 *  BSD's sys/queue.h.
 */
#ifndef __OpenBSD__

#define _LIST_HEAD(name, type)                                          \
struct name {                                                           \
        struct type *lh_first;  /* first element */                     \
}

#define LIST_ENTRY(type)                                                \
struct {                                                                \
        struct type *le_next;   /* next element */                      \
        struct type **le_prev;  /* address of previous next element */  \
}
#define LIST_FIRST(head)                ((head)->lh_first)
#define LIST_END(head)                  NULL
#define LIST_EMPTY(head)                (LIST_FIRST(head) == LIST_END(head))
#define LIST_NEXT(elm, field)           ((elm)->field.le_next)

#define LIST_FOREACH(var, head, field)                                  \
        for((var) = LIST_FIRST(head);                                   \
            (var)!= LIST_END(head);                                     \
            (var) = LIST_NEXT(var, field))

#define LIST_INIT(head) do {                                            \
        LIST_FIRST(head) = LIST_END(head);                              \
} while (0)

#define LIST_INSERT_HEAD(head, elm, field) do {                         \
        if (((elm)->field.le_next = (head)->lh_first) != NULL)          \
                (head)->lh_first->field.le_prev = &(elm)->field.le_next;\
        (head)->lh_first = (elm);                                       \
        (elm)->field.le_prev = &(head)->lh_first;                       \
} while (0)

#define LIST_REMOVE(elm, field) do {                                    \
        if ((elm)->field.le_next != NULL)                               \
                (elm)->field.le_next->field.le_prev =                   \
                    (elm)->field.le_prev;                               \
        *(elm)->field.le_prev = (elm)->field.le_next;                   \
} while (0)

#define TAILQ_HEAD(name, type)                                          \
struct name {                                                           \
        struct type *tqh_first; /* first element */                     \
        struct type **tqh_last; /* addr of last next element */         \
}

#define TAILQ_ENTRY(type)                                               \
struct {                                                                \
        struct type *tqe_next;  /* next element */                      \
        struct type **tqe_prev; /* address of previous next element */  \
}

#define TAILQ_FIRST(head)               ((head)->tqh_first)
#define TAILQ_END(head)                 NULL
#define TAILQ_EMPTY(head)		(TAILQ_FIRST(head) == TAILQ_END(head))
#define TAILQ_NEXT(elm, field)          ((elm)->field.tqe_next)

#define TAILQ_FOREACH(var, head, field)                                 \
        for((var) = TAILQ_FIRST(head);                                  \
            (var) != TAILQ_END(head);                                   \
            (var) = TAILQ_NEXT(var, field))

#define TAILQ_INIT(head) do {                                           \
        (head)->tqh_first = NULL;                                       \
        (head)->tqh_last = &(head)->tqh_first;                          \
} while (0)

#define TAILQ_INSERT_TAIL(head, elm, field) do {                        \
        (elm)->field.tqe_next = NULL;                                   \
        (elm)->field.tqe_prev = (head)->tqh_last;                       \
        *(head)->tqh_last = (elm);                                      \
        (head)->tqh_last = &(elm)->field.tqe_next;                      \
} while (0)

#define TAILQ_REMOVE(head, elm, field) do {                             \
        if (((elm)->field.tqe_next) != NULL)                            \
                (elm)->field.tqe_next->field.tqe_prev =                 \
                    (elm)->field.tqe_prev;                              \
        else                                                            \
                (head)->tqh_last = (elm)->field.tqe_prev;               \
        *(elm)->field.tqe_prev = (elm)->field.tqe_next;                 \
} while (0)

#endif     /* end of BSD list macros */

#ifdef __OpenBSD__
/*
 *  Argh, it turns out that Linux has a built-in macro called LIST_HEAD.  To avoid the
 *  collision, I had to replace LIST_HEAD everywhere with _LIST_HEAD.  BSD will need the
 *  following #define.  Is there a better solution?
 */
#define _LIST_HEAD(a,b)   LIST_HEAD(a,b)
#endif

/*
 *  I/O operations -- these will need to be defined differently for each platform.
 *    int nfs4_file_open(nfs4_file_desc_t *, struct svc_fh *, unsigned int share_access)
 *               opens an nfs4_file_desc_t with specified filehandle + share access.
 *    int nfs4_file_close(nfs4_file_desc_t *)
 *               closes an nfs4_file_desc_t
 *    int nfs4_file_upgrade(nfs4_file_desc_t *, unsigned int share_access)
 *               called whenever share access is upgraded on an nfs4_file_desc_t.
 *               the parameter 'share_access' represents the bits actually added.
 *    int nfs4_file_downgrade(struct nfs4_file_desc_t *, unsigned int share_access)
 *               called whenever share access is downgraded on an nfs4_file_desc_t.
 *               the parameter 'share_access' represents the bits actually added.
 *    int nfs4_truncate(struct svc_fh *fhp)
 *               truncates a filehandle to size 0
 *    int nfs4_check_fh(struct svc_fh *fhp, struct nfs4_openfile *ofp)
 *               returns an error if filehandle fails to match opened file
 */
struct svc_fh;          /* to be defined in nfsd4.h */
extern int nfs4_file_open(nfs4_file_desc_t *, struct svc_fh *, unsigned int share_access);
extern void nfs4_file_close(nfs4_file_desc_t *);

/*
 *    clientid_t       - our internal representation of a clientid
 *    stateid_t        - our internal representation of a stateid
 *    open_verifier_t  - our internal representation of an OPEN_CONFIRM verifier
 *
 *  Note that stateids are currently defined to be 64 bits, but this supposedly will soon
 *  be changed to 128 bits.  For this reason, our stateid_t's are defined so that one can
 *  flip back and forth between 64 and 128 bits using a #define.
 */
typedef struct {
	u32		cl_boot;
	u32		cl_id;
} clientid_t;

#if STATEID_128_BITS
typedef u32 stateid_boot_t;        /* used to detect stale stateids */
typedef u32 stateid_lockowner_t;   /* lockowner id - used in various places */
typedef u32 stateid_file_t;        /* identifies a unique file per lockowner */
typedef u32 stateid_generation_t;  /* used to update stateids */
#else
typedef u16 stateid_boot_t;
typedef u16 stateid_lockowner_t;
typedef u16 stateid_file_t;
typedef u16 stateid_generation_t;
#endif

typedef struct {
	stateid_boot_t		st_boot;
	stateid_lockowner_t	st_lockowner;
	stateid_file_t		st_file;
	stateid_generation_t	st_generation;
} stateid_t;

typedef struct {
	u32		ov_boot;
	u32		ov_id;          /* actually a stateid_lockowner_t.. */
} open_verifier_t;

extern stateid_t                    zero_stateid;    /* bits all zeros */
extern stateid_t                    one_stateid;     /* bits all ones  */
#define ZERO_STATEID(stateid)       (!memcmp((stateid), &zero_stateid, sizeof(stateid_t)))
#define ONE_STATEID(stateid)        (!memcmp((stateid), &one_stateid, sizeof(stateid_t)))

static inline void
update_stateid(stateid_t *stateid)
{
	stateid->st_generation++;
}

/*
 *  We don't bother putting clientids, etc., in network byte order, since they are
 *  opaque to the client.
 */
#define READ_CLIENTIID(clientid,p)    memcpy((clientid), (p), sizeof(clientid_t))
#define WRITE_CLIENTID(clientid,p)    memcpy((p), (clientid), sizeof(clientid_t))
#define READ_STATEID(stateid,p)       memcpy((stateid), (p), sizeof(stateid_t))
#define WRITE_STATEID(stateid,p)      memcpy((p), (stateid), sizeof(stateid_t))
#define READ_VERIFIER(verifier,p)     memcpy((verifier), (p), sizeof(open_verifier_t))
#define WRITE_VERIFIER(verifier,p)    memcpy((p), (verifier), sizeof(open_verifier_t))

/*
 *    struct nfs4_client - We have one of these for every client.  Clientids live here.
 *        o Each nfs4_client is hashed by clientid.
 *        o Each nfs4_clients is also hashed by name (the opaque quantity initially sent by
 *          the client to identify itself).
 *        o Each nfs4_client is also the head of a list of nfs4_lockowners which are active
 *          for the client.
 *        o Each nfs4_client lives on an LRU list, ordered by time of last lease renewal,
 *          which will be scanned periodically by the laundromat thread.
 *  
 *    struct nfs4_lockowner - We have one of these for every lockowner.  Seqids live here.
 *        o Each nfs4_lockowner contains a pointer back to its nfs4_client.
 *        o Each nfs4_lockowner is hashed by name (the opaque quantity sent by the client
 *          to identify a lockowner), and by lockowner id (lockowner ids are used for
 *          OPEN_CONFIRM).
 *        o Each nfs4_lockowner lives on a list of nfs4_lockowners which are active for
 *          its client.
 *        o Each nfs4_lockowner is the head of a list of nfs4_openfiles which were opened
 *          by the lockowner.
 *        o Each "unconfirmed" lockowner (meaning that an OPEN_CONFIRM is outstanding, or
 *          the lockowner has no files open) lives on an LRU list, ordered by time of last
 *          use, which will be scanned periodically by the laundromat thread.
 *
 *    struct nfs4_file - We have one of these for every active (i.e., opened by one or more
 *               clients) file.  Stateids do NOT live here; they live in nfs4_openfiles.
 *        o Each nfs4_file contains an nfs4_ino_dec_t, and is hashed by it.
 *        o Each nfs4_file is the head of a list of nfs4_openfiles corresponding to OPENs
 *          done on the file.
 *
 *    struct nfs4_openfile:  We have one of these per open (lockowner,file)-pair.  It
 *               represents the state of a single OPEN, done by a specific lockowner on a
 *               specific file.  Stateids live here.
 *        o Each nfs4_openfile contains pointers back to its nfs4_lockowner and nfs4_file.
 *        o Each nfs4_openfile is hashed by stateid.
 *        o Each nfs4_openfile lives on a list of nfs4_openfiles for the same lockowner.
 *        o Each nfs4_openfile lives on a list of nfs4_openfiles for the same file.
 */
struct nfs4_client {
	LIST_ENTRY(nfs4_client)			cl_idhash;
	LIST_ENTRY(nfs4_client)			cl_strhash;
	TAILQ_ENTRY(nfs4_client)		cl_lru;
  	_LIST_HEAD(dummy1, nfs4_lockowner)	cl_perclient;  /* list of lockowners */
	clientid_t				cl_clientid;
	int					cl_namelen;
	char					*cl_name;
	u32					cl_addr;       /* client's IP address */
	verifier4				cl_verifier;   /* client's, not server's! */
	int					cl_confirmed;  /* boolean */
	time_t					cl_time;       /* time of last lease renewal */
};

struct nfs4_lockowner {
	LIST_ENTRY(nfs4_lockowner)		lo_idhash;
	LIST_ENTRY(nfs4_lockowner)		lo_strhash;
	LIST_ENTRY(nfs4_lockowner)		lo_perclient;
	TAILQ_ENTRY(nfs4_lockowner)		lo_lru;
	_LIST_HEAD(dummy2, nfs4_openfile)	lo_perlockowner;  /* list of openfiles */
	struct nfs4_client			*lo_client;
	open_verifier_t				lo_verifier;
	seqid4					lo_seqid;
	unsigned int				lo_namelen;
	char					lo_name[NFSD4_LOCKOWNER_MAXNAME];
	unsigned int				lo_hashval;
	int					lo_confirmed;    /* boolean */
	time_t					lo_time;
	
	/* Replay buffer, where the result of the last seqid-mutating operation is cached */
	unsigned int				lo_replay_buflen;
	char					*lo_replay_buf;
	unsigned int				lo_replay_allocated;
	char					lo_replay_ibuf[NFSD4_REPLAY_ISIZE];
};
#define lo_boot					lo_verifier.ov_boot
#define lo_id					lo_verifier.ov_id

struct nfs4_file {
	LIST_ENTRY(nfs4_file)			fi_hash;
	_LIST_HEAD(dummy3, nfs4_openfile)	fi_perfile;   /* list of openfiles */
	nfs4_ino_desc_t				fi_ino;
	stateid_file_t				fi_id;
	/*
	 * I thought about keeping counts for the different share reservations that had been
	 * placed on the file here, to make the algorithm for checking share conflicts more
	 * scalable, but decided against it (I think that, in practice, a single file opened
	 * by a "large" number of lockowners will be a very rare case.)
	 */
};

struct nfs4_openfile {
	LIST_ENTRY(nfs4_openfile)		of_hash;
	LIST_ENTRY(nfs4_openfile)		of_perlockowner;
	LIST_ENTRY(nfs4_openfile)		of_perfile;
	struct nfs4_lockowner			*of_lockowner;
	struct nfs4_file			*of_file;
	stateid_t				of_stateid;
	nfs4_file_desc_t			of_vfs_file;
	unsigned int				of_share_access;
	unsigned int				of_share_deny;
};

struct global_state;    /* to be defined in nfsd4.h */

/*
 *  Prototypes for functions defined in state.c.
 */
extern void nfs4_state_init(void);
extern void nfs4_state_shutdown(void);
extern void nfs4_state_restart(int grace_period);
extern void nfs4_lock_state(void);
extern void nfs4_unlock_state(void);
extern time_t nfs4_laundromat(void);
extern int nfs4_in_grace_period(void);
extern void nfs4_set_verifier(char *verf);

/* To process a SETCLIENTID, call this function. */
extern int nfs4_process_setclientid(struct global_state *stp, SETCLIENTID4args *argp,
				    SETCLIENTID4res *resp);

/* Our SETCLIENTID_CONFIRM verifiers are really clientids.  To process a SETCLIENTID_CONFIRM,
 * set a (clientid_t *) to point to the verifier, and call this function.
 */
extern int nfs4_process_setclientid_confirm(struct global_state *stp, clientid_t *clientid);

/* To process a RENEW, just call... */
extern int nfs4_process_renew(struct global_state *stp, stateid_t *stateid);

/* To process an OPEN, call nfs4_process_open1(), which validates the lockowner, and sets
 * stp->current_lockowner.  It can return the speical error NFS4ERR_REPLAY_ME, which
 * indicates that the response from a previous OPEN should be replayed from the lockowner's
 * replay buffer.  Then determine which file the client is trying to open, set
 * stp->current_fh accordingly, and call nfs4_process_open2().  If this returns success,
 * it will set stp->current_openfile, and the stateid will live there. */
extern int nfs4_process_open1(struct global_state *stp, clientid_t *clientid,
			      char *name, unsigned int namelen, seqid4 seqid);
extern int nfs4_process_open2(struct global_state *stp, nfs4_ino_desc_t *ino, int truncate,
			      unsigned int share_access, unsigned int share_deny,
			      seqid4 seqid);

/* To process:
 *   { OPEN_CONFIRM, OPEN_DOWNGRADE, CLOSE },
 * call:
 *   { nfs4_process_open_confirm(), nfs4_process_open_downgrade(), nfs4_process_close() },
 * which will set stp->current_lockowner and stp->current_openfile.  NFS4ERR_REPLAY_ME is
 * a possible return value here.
 */
extern int nfs4_process_open_confirm(struct global_state *stp, seqid4 seqid,
				     open_verifier_t *verifier);
extern int nfs4_process_open_downgrade(struct global_state *stp, seqid4 seqid,
				       stateid_t *stateid, unsigned int share_access,
				       unsigned int share_deny);
extern int nfs4_process_close(struct global_state *stp, seqid4 seqid, stateid_t *stateid);

/* To process { READ, WRITE, SETATTR(truncating) }, first test for the "magic stateids"
 * 0 and -1 and handle them accordingly.  Then call nfs4_preprocess_stateid_op() to validate
 * the stateid, which will set stp->current_openfile.  Then test whether the openfile is
 * opened with read or write access, as appropriate, and continue processing the operation.
 */
extern int nfs4_preprocess_stateid_op(struct global_state *stp, stateid_t *stateid);

/* To process { LOCK, LOCKU }, call nfs4_preprocess_seqid_op(), which sets
 * stp->current_lockowner and stp->current_openfile.  NFS4ERR_REPLAY_ME is a possible
 * return value.
 */
extern int nfs4_preprocess_seqid_op(struct global_state *stp, seqid4 seqid,
				    stateid_t *stateid);

/* Miscellanea:
 *   nfs4_share_conflicts() - reports whether any OPENs on a file deny read/write access.
 *                            (READ/WRITE with magic stateids need this.)
 *   nfs4_lookup_lockowner() - looks up a lockowner by name and sets stp->current_lockowner.
 *                             (LOCKT needs this.)
 *   nfs4_verify_lockowner() - checks whether an (nfs4_lockowner *) from an "untrusted"
 *                             source is valid, or a garbage pointer.  (LOCKT needs this.)
 */
extern int nfs4_share_conflicts(nfs4_ino_desc_t *ino, unsigned int deny_type);
extern int nfs4_lookup_lockowner(struct global_state *stp, clientid_t *clientid,
				 unsigned int namelen, char *name);
extern int nfs4_verify_lockowner(struct nfs4_lockowner *lp, unsigned int hashval);

#endif   /* NFSD4_STATE_H */
