/*
 * linux/net/sunrpc/auth_gss.c
 *
 * RPCSEC_GSS client authentication.
 
 * Dug Song <dugsong@monkey.org>
 * Copyright (c) 2000 The Regents of the University of Michigan
 *
 * $Id: auth_gss.c,v 1.2 2001/06/28 18:19:40 andros Exp $
 */

#include <linux/types.h>
#include <linux/malloc.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/sunrpc/clnt.h>
#include <linux/sunrpc/auth.h>
#include <linux/sunrpc/auth_gss.h>

extern void rpcauth_remove_credcache(struct rpc_auth *auth, struct rpc_cred *cred);

#ifdef RPC_DEBUG
# define RPCDBG_FACILITY	RPCDBG_AUTH
#endif

#define NFS_NGROUPS	16
struct gss_cred {
	struct rpc_cred		gc_base;
	u32			gc_established;
	GSS_OID			gc_mech;
	u32			gc_qop;
	GSS_BUFFER_T		gc_service_name;
	GSS_CTX_ID_T		gc_ctx_id;
	u32			gc_win;
	struct rpc_gss_cred	gc_cred;
};
#define gc_uid			gc_base.cr_uid
#define gc_count		gc_base.cr_count
#define gc_flags		gc_base.cr_flags
#define gc_expire		gc_base.cr_expire

#define GSS_CRED_EXPIRE		(60 * HZ)	/* XXX */
#define GSS_CRED_SLACK		1024		/* XXX */
#define GSS_VERF_SLACK		512		/* XXX */

/* dump the buffer in `emacs-hexl' style */
#define isprint(c)      ((c > 0x1f) && (c < 0x7f))

void
print_hexl(u32 *p, u_int length, u_int offset)
{
	u_int i, j, jm;
	u8 c, *cp;
	
	printk("\n");
	cp = (u8 *) p;
	
	for (i = 0; i < length; i += 0x10) {
		printk("  %04x: ", (u_int)(i + offset));
		jm = length - i;
		jm = jm > 16 ? 16 : jm;
		
		for (j = 0; j < jm; j++) {
			if ((j % 2) == 1)
				printk("%02x ", (u_int)cp[i+j]);
			else
				printk("%02x", (u_int)cp[i+j]);
		}
		for (; j < 16; j++) {
			if ((j % 2) == 1)
				printk("   ");
			else
				printk("  ");
		}
		printk(" ");
		
		for (j = 0; j < jm; j++) {
			c = cp[i+j];
			c = isprint(c) ? c : '.';
			printk("%c", c);
		}
		printk("\n");
	}
}

static int
xdr_encode_gss_init_arg(struct rpc_rqst *req, u32 *p,
			struct xdr_netobj *argp)
{
	dprintk("RPC: xdr_encode_gss_init_arg (%p:%d)\n",
		argp->data, argp->len);
	
	p = xdr_encode_netobj(p, argp);
	
	req->rq_slen = xdr_adjust_iovec(req->rq_svec, p);
	dprintk("RPC: xdr_encode_gss_init_arg req->rq_slen %d\n",
	  req->rq_slen);
	return 0;
}

static int
xdr_decode_gss_init_res(struct rpc_rqst *req, u32 *p,
			struct rpc_gss_init_res *resp)
{
	p = xdr_decode_netobj(p, &resp->gr_ctx);
	if(!p) return -EIO;

	resp->gr_major = (u32) ntohl(*p++);
	resp->gr_minor = (u32) ntohl(*p++);
	resp->gr_win = (u32) ntohl(*p++);
	p = xdr_decode_netobj(p, &resp->gr_token);
	if(!p) return -EIO;

	dprintk("RPC: xdr_decode_gss_init_res() = "
		"(%p:%d, %d, %d, %d, %p:%d)\n",
		resp->gr_ctx.data, resp->gr_ctx.len,
		resp->gr_major, resp->gr_minor, resp->gr_win,
		resp->gr_token.data, resp->gr_token.len);
#if 0	
	if (!xdr_ressize_check(req, p))
		return -EIO;
#endif
	return 0;
}	

static struct rpc_auth *
gss_create(struct rpc_clnt *clnt)
{
	struct rpc_auth	*auth;

	dprintk("RPC: creating GSS authenticator for client %p\n", clnt);
	if (!(auth = (struct rpc_auth *) rpc_allocate(0, sizeof(*auth))))
		return NULL;
	auth->au_cslack = GSS_CRED_SLACK;
	auth->au_rslack = GSS_VERF_SLACK;
	auth->au_expire = GSS_CRED_EXPIRE;
	auth->au_ops = &authgss_ops;

	rpcauth_init_credcache(auth);

	return auth;
}

/* XXX - finish */
static void
gss_destroy(struct rpc_auth *auth)
{
	dprintk("RPC: destroying GSS authenticator %p\n", auth);
	/* If we have saved contexts, tell server to destroy them. */

	/* Destroy our local contexts. */
	rpcauth_free_credcache(auth);

	rpc_free(auth);
}

static void
gss_destroy_cred(struct rpc_cred *rc)
{
	struct gss_cred *cred = (struct gss_cred *)rc;
	
	if (cred->gc_mech.len > 0)
		rpc_free(cred->gc_mech.data);
	if (cred->gc_service_name.len > 0)
		rpc_free(cred->gc_service_name.data);
	if (cred->gc_ctx_id.len > 0)
		rpc_free(cred->gc_ctx_id.data);
	if (cred->gc_cred.gc_ctx.len > 0)
		rpc_free(cred->gc_cred.gc_ctx.data);
	rpc_free(cred);
}

/* XXX need to receive service from rpc program */
#define SERVICE_NAME	"nfs"

static struct rpc_cred *
gss_create_cred(struct rpc_task *task)
{
	struct gss_cred	*cred;
	struct rpc_gss_init_res res;
	u32		maj_stat, min_stat, ret_flags;
	GSS_BUFFER_T	recv_token, send_token;
	kxdrproc_t	save_encode, save_decode;
	struct rpc_cred *save_cred;

	dprintk("RPC:      allocating GSS cred for uid %d gid %d\n",
				current->uid, current->gid);

	if (!(cred = (struct gss_cred *) rpc_allocate(task->tk_flags,
						      sizeof(*cred))))
		return NULL;

	memset(cred, 0, sizeof(*cred));
	cred->gc_count = 0;
	cred->gc_flags = RPCAUTH_CRED_UPTODATE;
	cred->gc_uid = current->uid;

	/*cred->gc_established = 0;
	cred->gc_mech = GSS_C_NULL_OID;*/
	cred->gc_qop = GSS_C_QOP_DEFAULT;		/* XXX - configure */

	if (task->tk_client->cl_server == NULL ||
	    strcmp(task->tk_client->cl_server, "localhost") == 0) {
		cred->gc_service_name.len = sizeof(SERVICE_NAME);
		cred->gc_service_name.data =
			rpc_allocate(0, cred->gc_service_name.len);
		memcpy(cred->gc_service_name.data, SERVICE_NAME,
		       sizeof(SERVICE_NAME));
	}
	else {
		cred->gc_service_name.len = sizeof(SERVICE_NAME) + 1 +
			strlen(task->tk_client->cl_server);
		cred->gc_service_name.data =
			rpc_allocate(0, cred->gc_service_name.len);
		sprintf(cred->gc_service_name.data, SERVICE_NAME "@%s",
			task->tk_client->cl_server);
	}
	/*cred->gc_ctx_id = GSS_C_NO_CONTEXT;*/
	
	cred->gc_cred.gc_v = RPC_GSS_VERSION;
	cred->gc_cred.gc_proc = RPC_GSS_PROC_INIT;
	cred->gc_cred.gc_svc = RPC_GSS_SVC_NONE;	/* XXX - configure */

	/* Set temporary credential for RPCSEC_GSS NULLPROC call. */
	save_cred = task->tk_msg.rpc_cred;
	task->tk_msg.rpc_cred = (struct rpc_cred *)cred;
	rpcauth_insert_credcache(task->tk_auth, (struct rpc_cred *)cred);

	/*
	 * XXX - swap out the program's NULLPROC handlers directly,
	 * since there is no clean way to prevent dispatch.
	 */
	save_encode = rpcproc_encode(task->tk_client, 0);
	save_decode = rpcproc_decode(task->tk_client, 0);
	rpcproc_encode(task->tk_client, 0) =
		(kxdrproc_t) xdr_encode_gss_init_arg;
	rpcproc_decode(task->tk_client, 0) =
		(kxdrproc_t) xdr_decode_gss_init_res;
	
	memset(&res, 0, sizeof(res));
	memset(&send_token, 0, sizeof(send_token));
	memset(&recv_token, 0, sizeof(recv_token));
	
	for (;;) {
		dprintk("RPC: gss_create_cred() calling init_sec_context()\n");
		if (gssd_init_sec_context(cred->gc_uid,
					  &maj_stat,
					  &min_stat,
					  &cred->gc_ctx_id,
					  &cred->gc_service_name,
					  &cred->gc_mech,
					  &recv_token,
					  &send_token,
					  &ret_flags) < 0)
			break;
		dprintk("RPC: after init_sec_context()\n");
		if (recv_token.len > 0) {
			dprintk("RPC: freeing recv_token\n");
			rpc_free(recv_token.data);
			memset(&recv_token, 0, sizeof(recv_token));
		}
		if (maj_stat != GSS_S_COMPLETE &&
		    maj_stat != GSS_S_CONTINUE_NEEDED)
			break;
		
		if (send_token.len > 0) {
			dprintk("RPC: sending token (%p:%d)\n",
				send_token.data, send_token.len);

			/* Send token as NULLPROC argument. */
			if (rpc_call(task->tk_client, 0,
				     &send_token, &res, 0) < 0)
				break;

			/* Check results. */
			if (res.gr_major != GSS_S_COMPLETE &&
			    res.gr_major != GSS_S_CONTINUE_NEEDED)
				break;

			/* Save results. */
			if (res.gr_ctx.len > 0) {
				if (cred->gc_cred.gc_ctx.len > 0)
					rpc_free(cred->gc_cred.gc_ctx.data);
				cred->gc_cred.gc_ctx.data =
					rpc_allocate(0, res.gr_ctx.len);
				memcpy(cred->gc_cred.gc_ctx.data,
				       res.gr_ctx.data, res.gr_ctx.len);
				cred->gc_cred.gc_ctx.len = res.gr_ctx.len;
			}
			if (res.gr_token.len > 0) {
				recv_token.data =
					rpc_allocate(0, res.gr_token.len);
				memcpy(recv_token.data,
				       res.gr_token.data, res.gr_token.len);
				recv_token.len = res.gr_token.len;
			}
			cred->gc_cred.gc_proc = RPC_GSS_PROC_CONTINUE_INIT;
		}
		if (maj_stat == GSS_S_COMPLETE) {
			/* Done! Change phase to RPC data exchange. */
			cred->gc_established = 1;
			cred->gc_cred.gc_proc = RPC_GSS_PROC_DATA;
			cred->gc_cred.gc_seq = 1;
			cred->gc_win = res.gr_win;
			break;
		}
	}
	/* Disassociate temporary credentials. */
	task->tk_msg.rpc_cred = save_cred;
	rpcauth_remove_credcache(task->tk_auth, (struct rpc_cred *)cred);
	
	/* XXX - restore target program's original NULLPROC mappings. */
	rpcproc_encode(task->tk_client, 0) = (kxdrproc_t) save_encode;
	rpcproc_decode(task->tk_client, 0) = (kxdrproc_t) save_decode;

	if (!cred->gc_established) {
		dprintk("RPC: GSS context establishment failed.\n");
		gss_destroy_cred((struct rpc_cred *)cred);
    task->tk_status = -EACCES;
		return NULL;
	}
	/* XXX - should check for service match. */
	
	dprintk("RPC: GSS context establishment succeeded.\n");
	return (struct rpc_cred *) cred;
}

/*
 * Match credentials against current process creds.
 */
static int
gss_match(struct rpc_cred *rc, int taskflags)
{
	struct gss_cred *cred = (struct gss_cred *)rc;
	
	return (cred->gc_uid == current->uid);
}

/*
 * Marshal credentials.
 * Maybe we should keep a cached credential for performance reasons.
 */
static u32 *
gss_marshal(struct rpc_task *task, u32 *p, int ruid)
{
	struct gss_cred	*cred = (struct gss_cred *) task->tk_msg.rpc_cred;
	u32		*base;

	/* Marshal credential. */
	*p++ = htonl(RPC_AUTH_GSS);
	base = p++;
	*p++ = htonl((u32) cred->gc_cred.gc_v);
	*p++ = htonl((u32) cred->gc_cred.gc_proc);
	*p++ = htonl((u32) cred->gc_cred.gc_seq);
	*p++ = htonl((u32) cred->gc_cred.gc_svc);
	p = xdr_encode_netobj(p, &cred->gc_cred.gc_ctx);
	*base = htonl((p - base - 1) << 2);	/* cred length */

	/* Marshal verifier. */
#if 0
	if (cred->gc_established) {
		bufin.val = p;
		bufin.len = p - task.tk_buffer;

		*p++ = htonl(RPC_AUTH_GSS);
		base = p++;
		p = gss_sign(cred->gc_ctx_id, cred->gc_qop, &bufin, p);
		*base = htonl((p - base - 1) << 2);	/* verf length */

		return p;
	}
#endif
	*p++ = htonl(RPC_AUTH_NULL);
	*p++ = 0;
	return p;
}

/*
 * Refresh credentials. XXX - finish
 */
static int
gss_refresh(struct rpc_task *task)
{
	task->tk_msg.rpc_cred->cr_flags |= RPCAUTH_CRED_UPTODATE;
	return task->tk_status = -EACCES;
}

/* note: we don't deal with verifiers yet - no crypto 
 * in the kernel. may do a gssd upcall just to interoperate....
 */
static u32 *
gss_validate(struct rpc_task *task, u32 *p)
{
#if 0
	struct rpc_cred *cred = task->tk_msg.rpc_cred;
	u32		seq, qop_state;
	GSS_BUFFER_T	bufin;
#endif
	u32		flav,len;

	/* verifier flavor */
	 flav = ntohl(*p++);
	if (flav != RPC_AUTH_GSS && flav != RPC_AUTH_NULL) {
		printk("RPC: bad verf flavor: %ld\n", (unsigned long) flav);
		return NULL;
	}
	printk("gss_validate: flavor %ld \n",(unsigned long)flav);

	if ((len = ntohl(*p++)) > 400) {
		printk("RPC: giant verf size: %ld\n", (unsigned long) len);
		return NULL;
	}
	printk("len %ld\n", (unsigned long)len);

#if 0
	if (cred->gc_established && cred->gc_cred.gc_svc != RPC_GSS_SVC_NONE) {
		seq = htonl(cred->gc_cred.gc_seq);
		bufin.data = (u8 *) &seq;
		bufin.len = sizeof(seq);
		bufout.data = (u8 *) p;
		bufout.len = n;
		
		if (gss_verify(cred->gc_ctx_id, &bufin, &bufout,
			       &qop_state) < 0)
			return NULL;
	}
#endif
	printk("RPC: GSS gss_validate: skipping verifier: skipping %d bytes\n",
	  XDR_QUADLEN(len));
	task->tk_auth->au_rslack = XDR_QUADLEN(len) + 2;
	return p + XDR_QUADLEN(len);
}

struct rpc_authops	authgss_ops = {
	RPC_AUTH_GSS,
#ifdef RPC_DEBUG
	"RPCSEC_GSS",
#endif
	gss_create,
	gss_destroy,
	gss_create_cred,
	gss_destroy_cred,
	gss_match,
	gss_marshal,
	gss_refresh,
	gss_validate
};
