/*
 * linux/net/sunrpc/gssd_clnt.c
 *
 * GSSD client.
 *
 * Dug Song <dugsong@monkey.org>
 * Copyright (c) 2000 The Regents of the University of Michigan
 *
 * $Id: gssd_clnt.c,v 1.1 2001/05/15 22:30:21 kmsmith Exp $
 */

#include <linux/config.h>
#include <linux/types.h>
#include <linux/socket.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/uio.h>
#include <linux/in.h>
#include <linux/sunrpc/clnt.h>
#include <linux/sunrpc/xprt.h>
#include <linux/sunrpc/sched.h>
#include <linux/sunrpc/auth_gss.h>

#ifdef RPC_DEBUG
# define RPCDBG_FACILITY	RPCDBG_AUTH
#endif

/*
 * GSSD gss.x definitions
 */

#define RPC_GSSD_PROGRAM	100666
#define RPC_GSSD_VERSION	1

enum gssd_proc {
	GSSD_NAME_TO_UID	= 1,
	GSSD_UID_TO_NAME	= 2,
	GSSD_NAME_TO_GID	= 3,
	GSSD_GID_TO_NAME	= 4,
	GSSD_INIT_SEC_CONTEXT	= 5,
	GSSD_ACCEPT_SEC_CONTEXT	= 6
};

struct gssd_init_sec_context_arg {
	u32		uid;
	GSS_CTX_ID_T	ctx_id;
	GSS_BUFFER_T	target_name;
	GSS_OID		mech;
	GSS_BUFFER_T	input_token;
};

struct gssd_init_sec_context_res {
	GSS_CTX_ID_T	ctx_id;
	u32		maj_stat;
	u32		min_stat;
	GSS_BUFFER_T	output_token;
	u32		ret_flags;
};

struct gssd_accept_sec_context_arg {
	GSS_CTX_ID_T	ctx_id;
	GSS_BUFFER_T	input_token;
};

struct gssd_accept_sec_context_res {
	GSS_CTX_ID_T	ctx_id;
	u32		maj_stat;
	u32		min_stat;
	GSS_BUFFER_T	src_name;
	GSS_BUFFER_T	output_token;
	u32		ret_flags;
};

extern struct rpc_program	gssd_program;

static struct rpc_clnt *
gssd_create(void)
{
	struct sockaddr_in	sin;
	struct rpc_xprt	       *xprt;
	struct rpc_clnt	       *clnt;
	int			port;
	uint   saved_fsuid = current->fsuid;
	kernel_cap_t saved_cap = current->cap_effective;

	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	
	if ((port = rpc_getport_external(&sin, RPC_GSSD_PROGRAM,
					 RPC_GSSD_VERSION, IPPROTO_UDP)) < 0)
		return NULL;
	sin.sin_port = ntohs(port);

	/* create rpc socket as root so we get a priv port */
	current->fsuid = 0;
	cap_raise(current->cap_effective, CAP_NET_BIND_SERVICE);
	xprt = xprt_create_proto(IPPROTO_UDP, &sin, NULL);
	current->fsuid = saved_fsuid;
	current->cap_effective = saved_cap;
	if(xprt == NULL)
		return NULL;
	
	clnt = rpc_create_client(xprt, "localhost", &gssd_program,
				 RPC_GSSD_VERSION, RPC_AUTH_UNIX);
	if (!clnt) {
		xprt_destroy(xprt);
	} else {
		clnt->cl_softrtry = 1;
		clnt->cl_chatty   = 1;
		clnt->cl_oneshot  = 1;
	}
	dprintk("RPC: gssd_create()\n");
	return clnt;
}

/*
 *  GSSD RPCSEC_GSS test, just for fun.
 */
int
gssd_client_test(struct xdr_netobj *name, u32 *uidp)
{
	struct sockaddr_in	sin;
	struct rpc_xprt	       *xprt;
	struct rpc_clnt	       *clnt;
	int			port, error;
	struct rpc_auth	       *save_auth;

	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	
	if ((port = rpc_getport_external(&sin, RPC_GSSD_PROGRAM,
					 RPC_GSSD_VERSION, IPPROTO_UDP)) < 0)
		return -1;
	
	if (!(xprt = xprt_create_proto(IPPROTO_UDP, &sin, NULL)))
		return -1;
	xprt->addr.sin_port = htons(port);
	
	clnt = rpc_create_client(xprt, "localhost", &gssd_program,
				 RPC_GSSD_VERSION, RPC_AUTH_UNIX);
	if (!clnt) {
		xprt_destroy(xprt);
		return -1;
	} else {
		clnt->cl_softrtry = 1;
		clnt->cl_chatty   = 1;
		clnt->cl_oneshot  = 1;
	}
	save_auth = clnt->cl_auth;
	rpcauth_create(RPC_AUTH_GSS, clnt);
	
	error = rpc_call(clnt, GSSD_NAME_TO_UID, name, uidp, RPC_TASK_ROOT_CREDS);

	clnt->cl_auth = save_auth;
	
	if (error < 0) {
		printk(KERN_WARNING
		       "RPC: failed to contact gssd (errno %d).\n", error);
	}
	dprintk("RPC: gssd_client_test(%.*s) = %d\n",
		(int) name->len, name->data, *uidp);
	
	/* Client deleted automatically because cl_oneshot == 1 */
	return error;
}	
	
/*
 * Convert Unix/GSS username/principal to Unix UID.
 */
int
gssd_name_to_uid(struct xdr_netobj *name, u32 *uidp)
{
	struct rpc_clnt		*gssd_clnt;
	unsigned int		error = 0;

	dprintk("RPC: gssd_name_to_uid \n");
	if (!(gssd_clnt = gssd_create())) {
		dprintk("RPC: couldn't create gssd client\n");
		return -EACCES;
	}

	error = rpc_call(gssd_clnt, GSSD_NAME_TO_UID, name, uidp, RPC_TASK_ROOT_CREDS);
	
	if (error < 0) {
		printk(KERN_WARNING
		       "RPC: failed to contact gssd (errno %d).\n", error);
	}
	dprintk("RPC: gssd_name_to_uid(%.*s) = %d\n",
		(int) name->len, name->data, *uidp);
	
	/* Client deleted automatically because cl_oneshot == 1 */
	return error;
}

/*
 * Convert Unix UID to Unix username.
 */
int
gssd_uid_to_name(u32 *uid, struct xdr_netobj *namep)
{
	GSS_BUFFER_T		bufout;
	struct rpc_clnt		*gssd_clnt;
	unsigned int		error = 0;

	if (!(gssd_clnt = gssd_create())) {
		dprintk("RPC: couldn't create gssd client\n");
		return -EACCES;
	}
	error = rpc_call(gssd_clnt, GSSD_UID_TO_NAME, uid, (void *)&bufout, RPC_TASK_ROOT_CREDS);
	
	if (error < 0) {
		printk(KERN_WARNING
		       "RPC: failed to contact gssd (errno %d).\n", error);
	}
	if (bufout.len > 0)
		namep->data = rpc_allocate(0, bufout.len);
	memcpy(namep->data, bufout.data, bufout.len);
	namep->len = bufout.len;
	
	dprintk("RPC: gssd_uid_to_name(%d) = %.*s\n",
		*uid, (int) namep->len, namep->data);
	
	/* Client deleted automatically because cl_oneshot == 1 */
	return error;
}

/*
 * Convert Unix group name to Unix GID.
 */
int
gssd_name_to_gid(struct xdr_netobj *name, u32 *gidp)
{
	struct rpc_clnt		*gssd_clnt;
	unsigned int		error = 0;

	if (!(gssd_clnt = gssd_create())) {
		dprintk("RPC: couldn't create gssd client\n");
		return -EACCES;
	}
	error = rpc_call(gssd_clnt, GSSD_NAME_TO_GID, name, gidp, RPC_TASK_ROOT_CREDS);
	
	if (error < 0) {
		printk(KERN_WARNING
		       "RPC: failed to contact gssd (errno %d).\n", error);
	}
	dprintk("RPC: gssd_name_to_gid(%.*s) = %d\n",
		(int) name->len, name->data, *gidp);
	
	/* Client deleted automatically because cl_oneshot == 1 */
	return error;
}

/*
 * Convert Unix GID to Unix name.
 */
int
gssd_gid_to_name(u32 *gid, struct xdr_netobj *namep)
{
	GSS_BUFFER_T		bufout;
	struct rpc_clnt		*gssd_clnt;
	unsigned int		error = 0;

	if (!(gssd_clnt = gssd_create())) {
		dprintk("RPC: couldn't create gssd client\n");
		return -EACCES;
	}
	error = rpc_call(gssd_clnt, GSSD_GID_TO_NAME, gid, (void *)&bufout, RPC_TASK_ROOT_CREDS);
	
	if (error < 0) {
		printk(KERN_WARNING
		       "RPC: failed to contact gssd (errno %d).\n", error);
	}
	if (bufout.len > 0)
		namep->data = rpc_allocate(0, bufout.len);
	memcpy(namep->data, bufout.data, bufout.len);
	namep->len = bufout.len;
	
	dprintk("RPC: gssd_gid_to_name(%d) = %.*s\n",
		*gid, (int) namep->len, namep->data);
	
	/* Client deleted automatically because cl_oneshot == 1 */
	return error;
}

/*
 * Initialize GSS security context.
 */
int
gssd_init_sec_context(u32		uid,
		      u32	       *maj_statp,
		      u32	       *min_statp,
		      GSS_CTX_ID_T     *ctx_id,
		      GSS_BUFFER_T     *service_name,
		      GSS_OID	       *mech,
		      GSS_BUFFER_T     *input_token,
		      GSS_BUFFER_T     *output_tokenp,
		      u32	       *ret_flagsp)
{
	struct gssd_init_sec_context_arg	arg;
	struct gssd_init_sec_context_res	res;
	struct rpc_clnt		*gssd_clnt;
	unsigned int		error = 0;

	dprintk("RPC: gssd_init_sec_context(): ctx_id %p \n",ctx_id);
	if (!(gssd_clnt = gssd_create())) {
		dprintk("RPC: couldn't create gssd client\n");
		return -EACCES;
	}
	arg.uid = uid;
	arg.ctx_id = *ctx_id;
	arg.target_name = *service_name;
	arg.mech = *mech;
	arg.input_token = *input_token;

	dprintk("RPC: in gssd_init_sec_context() doing rpc_call()\n");

	error = rpc_call(gssd_clnt, GSSD_INIT_SEC_CONTEXT, &arg, &res, RPC_TASK_ROOT_CREDS);

	if (error < 0) {
		printk(KERN_WARNING
		       "RPC: failed to contact gssd (errno %d).\n", error);
		return error;
	}

	/* Save our results. */
	dprintk("RPC: in gssd_init_sec_context() res.ctx_id.len %d\n",res.ctx_id.len);

	if (res.ctx_id.len > 0) {
		if(!(ctx_id->data = rpc_allocate(0, res.ctx_id.len))){
			return -ENOMEM;
		}
		memcpy(ctx_id->data, res.ctx_id.data, res.ctx_id.len);
	}
	ctx_id->len = res.ctx_id.len;
	
	*maj_statp = res.maj_stat;
	*min_statp = res.min_stat;

	dprintk("RPC: in gssd_init_sec_context() res.output_token.len %d\n",
	   res.output_token.len);
	if (res.output_token.len > 0) {
		if(!(output_tokenp->data = rpc_allocate(0, res.output_token.len))){
			return -ENOMEM;
		}
		memcpy(output_tokenp->data, res.output_token.data,
		       res.output_token.len);
	}
	output_tokenp->len = res.output_token.len;
	
	*ret_flagsp = res.ret_flags;
	
	dprintk("RPC: gssd_init_sec_ctx()\n");
	
	/* Client deleted automatically because cl_oneshot == 1 */
	return error;
}

/*
 * Accept GSS security context.
 */
int
gssd_accept_sec_context(u32		*maj_statp,
			u32		*min_statp,
			GSS_CTX_ID_T	*ctx_id,
			GSS_BUFFER_T	*input_token,
			GSS_BUFFER_T	*client_namep,
			GSS_BUFFER_T	*output_tokenp,
			u32		*ret_flagsp)
{
	struct gssd_accept_sec_context_arg	arg;
	struct gssd_accept_sec_context_res	res;
	struct rpc_clnt		*gssd_clnt;
	unsigned int		error = 0;

	if (!(gssd_clnt = gssd_create())) {
		dprintk("RPC: couldn't create gssd client\n");
		return -EACCES;
	}
	arg.ctx_id = *ctx_id;
	arg.input_token = *input_token;

	error = rpc_call(gssd_clnt, GSSD_ACCEPT_SEC_CONTEXT, &arg, &res, RPC_TASK_ROOT_CREDS);

	if (error < 0) {
		printk(KERN_WARNING
		       "RPC: failed to contact gssd (errno %d).\n", error);
	}
	/* Save our results. */
	if (res.ctx_id.len > 0) {
		ctx_id->data = rpc_allocate(0, res.ctx_id.len);
		memcpy(ctx_id->data, res.ctx_id.data, res.ctx_id.len);
	}
	ctx_id->len = res.ctx_id.len;
	
	*maj_statp = res.maj_stat;
	*min_statp = res.min_stat;

	if (res.src_name.len > 0) {
		client_namep->data = rpc_allocate(0, res.src_name.len);
		memcpy(client_namep->data, res.src_name.data,
		       res.src_name.len);
	}
	client_namep->len = res.src_name.len;
	
	if (res.output_token.len > 0) {
		output_tokenp->data = rpc_allocate(0, res.output_token.len);
		memcpy(output_tokenp->data, res.output_token.data,
		       res.output_token.len);
	}
	output_tokenp->len = res.output_token.len;

	*ret_flagsp = res.ret_flags;
	
	dprintk("RPC: gssd_accept_sec_ctx()\n");

	/* Client deleted automatically because cl_oneshot == 1 */
	return error;
}

/*
 * XDR encode/decode functions for GSSD
 */
static int
xdr_error(struct rpc_rqst *req, u32 *p, void *dummy)
{
	return -EIO;
}

#if 0
static inline int
xdr_ressize_check(struct rpc_rqst *req, u32 *p)
{
	struct svc_buf	*buf = req->rq_resbuf;
	
	buf->len = p - buf->base;
	dprintk("gssc_clnt: ressize_check p %p base %p len %d\n",
		p, buf->base, buf->buflen);
	return (buf->len <= buf->buflen);
}
#endif

static int
xdr_decode_name(struct rpc_rqst *req, u32 *p, struct xdr_netobj *name)
{
	xdr_decode_netobj(p, name);
	dprintk("RPC: xdr_decode_name() = (%p:%d)\n", name->data, name->len);
	return 0;
}

static int
xdr_encode_name(struct rpc_rqst *req, u32 *p, struct xdr_netobj *name)
{
	dprintk("RPC: xdr_encode_name(%.*s)\n", (int) name->len, name->data);
	p = xdr_encode_netobj(p, name);

	req->rq_slen = xdr_adjust_iovec(req->rq_svec, p);
	return 0;
}
		
static int
xdr_decode_id(struct rpc_rqst *req, u32 *p, u32 *idp)
{
	*idp = (u32) ntohl(*p);
	dprintk("RPC: xdr_decode_id() = (%d)\n", *idp);
	return 0;
}

static int
xdr_encode_id(struct rpc_rqst *req, u32 *p, u32 *idp)
{
	dprintk("RPC: xdr_encode_id(%d)\n", *idp);
	*p++ = (u32) htonl(*idp);
	
	req->rq_slen = xdr_adjust_iovec(req->rq_svec, p);
	return 0;
}

static int
xdr_encode_init_sec_context_arg(struct rpc_rqst *req, u32 *p,
				struct gssd_init_sec_context_arg *argp)
{
	dprintk("RPC: xdr_encode_init_sec_context_arg"
		"(uid %d, ctx %p:%d, name %p:%d, mech %p:%d, token %p:%d)\n",
		argp->uid, argp->ctx_id.data, argp->ctx_id.len,
		argp->target_name.data, argp->target_name.len,
		argp->mech.data, argp->mech.len,
		argp->input_token.data, argp->input_token.len);
	
  	*p++ = (u32) htonl(argp->uid);
	p = xdr_encode_netobj(p, &argp->ctx_id);
	p = xdr_encode_netobj(p, &argp->target_name);
 	p = xdr_encode_netobj(p, &argp->mech);
	p = xdr_encode_netobj(p, &argp->input_token);
	
	req->rq_slen = xdr_adjust_iovec(req->rq_svec, p);
	return 0;
}

static int
xdr_decode_init_sec_context_res(struct rpc_rqst *req, u32 *p,
				struct gssd_init_sec_context_res *resp)
{
	p = xdr_decode_netobj(p, &resp->ctx_id);
	if(!p) return -EIO;
	resp->maj_stat = (u32) ntohl(*p++);
	resp->min_stat = (u32) ntohl(*p++);
	p = xdr_decode_netobj(p, &resp->output_token);
	if(!p) return -EIO;
	resp->ret_flags = (u32) ntohl(*p++);

	dprintk("RPC: xdr_decode_init_sec_context_res() = "
		"(ctx %p:%d, maj_stat %d, min_stat %d, "
		"token %p:%d, flags %d)\n",
		resp->ctx_id.data, resp->ctx_id.len,
		resp->maj_stat, resp->min_stat,
		resp->output_token.data, resp->output_token.len,
		resp->ret_flags);
#if 0	
	if (!xdr_ressize_check(req, p))
		return -EIO;
#endif
	return 0;
}	

static int
xdr_encode_accept_sec_context_arg(struct rpc_rqst *req, u32 *p,
				  struct gssd_accept_sec_context_arg *argp)
{
	dprintk("RPC: xdr_encode_accept_sec_context_arg("
		"ctx %p:%d, token %p:%d)\n",
		argp->ctx_id.data, argp->ctx_id.len,
		argp->input_token.data, argp->input_token.len);
	
	p = xdr_encode_netobj(p, &argp->ctx_id);
	p = xdr_encode_netobj(p, &argp->input_token);
	
	req->rq_slen = xdr_adjust_iovec(req->rq_svec, p);
	return 0;
}

static int
xdr_decode_accept_sec_context_res(struct rpc_rqst *req, u32 *p,
				  struct gssd_accept_sec_context_res *resp)
{
	p = xdr_decode_netobj(p, &resp->ctx_id);
	if(!p) return -EIO;
	resp->maj_stat = (u32) ntohl(*p++);
	resp->min_stat = (u32) ntohl(*p++);
	p = xdr_decode_netobj(p, &resp->src_name);
	if(!p) return -EIO;
	p = xdr_decode_netobj(p, &resp->output_token);
	if(!p) return -EIO;
	resp->ret_flags = (u32) ntohl(*p++);

	dprintk("RPC: xdr_decode_accept_sec_context_res() = "
		"(ctx %p:%d, maj_stat %d, min_stat %d, name %p:%d, "
		"token %p:%d, flags %d)\n",
		resp->ctx_id.data, resp->ctx_id.len,
		resp->maj_stat, resp->min_stat,
		resp->src_name.data, resp->src_name.len,
		resp->output_token.data, resp->output_token.len,
		resp->ret_flags);
#if 0	
	if (!xdr_ressize_check(req, p))
		return -EIO;
#endif
	return 0;
}

static struct rpc_procinfo	gssd_procedures[] = {
	{ "gssd_null",
		(kxdrproc_t) xdr_error,	
		(kxdrproc_t) xdr_error,	0, 0 },
	{ "gssd_name_to_uid",
		(kxdrproc_t) xdr_encode_name,	
		(kxdrproc_t) xdr_decode_id, RPC_MAXNETNAMELEN >> 2, 1 },
	{ "gssd_uid_to_name",
		(kxdrproc_t) xdr_encode_id,	
		(kxdrproc_t) xdr_decode_name, RPC_MAXNETNAMELEN >> 2, 1 },
	{ "gssd_name_to_gid",
		(kxdrproc_t) xdr_encode_name,	
		(kxdrproc_t) xdr_decode_id, RPC_MAXNETNAMELEN >> 2, 1 },
	{ "gssd_gid_to_name",
		(kxdrproc_t) xdr_encode_id,	
		(kxdrproc_t) xdr_decode_name, RPC_MAXNETNAMELEN >> 2, 1 },
	{ "gssd_init_sec_context",
		(kxdrproc_t) xdr_encode_init_sec_context_arg,
		(kxdrproc_t) xdr_decode_init_sec_context_res,
	  	XDR_MAX_NETOBJ >> 2, 1 },
	{ "gssd_accept_sec_context",
		(kxdrproc_t) xdr_encode_accept_sec_context_arg,
		(kxdrproc_t) xdr_decode_accept_sec_context_res,
	  	XDR_MAX_NETOBJ >> 2, 1 }
};

static struct rpc_version	gssd_version1 = {
	1,
	sizeof(gssd_procedures) / sizeof(gssd_procedures[0]),
	gssd_procedures
};

static struct rpc_version *	gssd_version[] = {
	NULL,
	&gssd_version1
};

static struct rpc_stat		gssd_stats;

struct rpc_program	gssd_program = {
	"GSSD",
	RPC_GSSD_PROGRAM,
	sizeof(gssd_version) / sizeof(gssd_version[0]),
	gssd_version,
	&gssd_stats,
};
